// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//

package de.uka.ilkd.asmkey.gui;


import java.io.*;
import java.util.*;

import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

import de.uka.ilkd.asmkey.logic.*;
import de.uka.ilkd.asmkey.parser.AstParser;
import de.uka.ilkd.asmkey.parser.TreeParser;
import de.uka.ilkd.asmkey.parser.ast.AstDeclaration;
import de.uka.ilkd.asmkey.parser.ast.AstProblem;
import de.uka.ilkd.asmkey.parser.env.*;
import de.uka.ilkd.asmkey.proof.ProofParser;
import de.uka.ilkd.asmkey.rule.StepRule;
import de.uka.ilkd.key.gui.*;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.abstraction.PrimitiveType;
import de.uka.ilkd.key.java.abstraction.SLListOfType;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.ldt.*;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.parser.Location;
import de.uka.ilkd.key.parser.ParserException;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.TacletIndex;
import de.uka.ilkd.key.proof.BuiltInRuleIndex;
import de.uka.ilkd.key.rule.*;


/** This is the main class for starting the standalone prover of
 * ASM-KeY. The class extends the default Main class of the core
 * system.
 *
 * @author Hubert Schmid */

public final class Main extends de.uka.ilkd.key.gui.Main {

    protected Main(String title, ProofSettings settings) {
       super(title, settings);
    }

    public static de.uka.ilkd.key.gui.Main getInstance() {
        if (instance == null) instance = 
            new Main("KeY -- Prover", ProofSettings.THE_SETTINGS);
        if (!instance.isVisible()) instance.setVisible(true);
        return instance;
    }


    public static void main(String[] args) {
	Main.evaluateOptions(args);
	de.uka.ilkd.key.gui.Main key = getInstance();
	key.loadCommandLineFile();
    }
/*
        evaluateOptions(args);
        Config.DEFAULT.setDefaultFonts();
        Main key = new Main("KeY -- Prover", ProofSettings.THE_SETTINGS);
        KeYMediator mediator = new KeYMediator(key);
        key.setMediator(mediator);
        key.initGoalList();
        key.initGUIProofTree();
        SwingUtilities.updateComponentTreeUI(key);
        key.show();
*/
//    }


    /** Return a collection of all dynamic function symbols contained
     * in the given collection. */
    private static Collection getDynamicFunctions(Collection c) {
        ArrayList list = new ArrayList();
        for (Iterator i = c.iterator(); i.hasNext();) {
            Object f = i.next();
            if (f instanceof DynamicFunction) {
                list.add(f);
            }
        }
        return list;
    }

    /** Load all taclet (.rules) files in the given directory and add
     * their declarations to the given environment. The files are read
     * in sorted order (by name). Only files with suffix '.rules' are
     * read. */
    private static void loadRulesDirectory(DeclarationEnvironment env, File path)
        throws ParserException
    {
        System.out.print("load rules from " + path + ":");
        File[] files = path.listFiles(new FileFilter() {
                public boolean accept(File file) {
                    return file.isFile() && file.getName().endsWith(".rules");
                }
            });
        if (files == null) {
            throw new RuntimeException("problem reading rules");
        }
        Arrays.sort(files, new Comparator() {
                public int compare(Object lhs, Object rhs) {
                    return ((File) lhs).getName().compareTo(((File) rhs).getName());
                }
            });
        for (int i = 0; i < files.length; ++i) {
            try {
                System.out.print(" " + files[i].getName());
                Reader in = new BufferedReader(new FileReader(files[i]));
                AstDeclaration[] decls = AstParser.parseDeclarations(in);
                TreeParser.parseDeclarations(env, decls);
                in.close();
            } catch (IOException ioe) {
                ioe.printStackTrace(System.err);
            }
        }
        System.out.println();
    }

    /** Load all taclets and add their declartions to the given
     * environment. The taclets are searched in the directory
     * specified by the system property 'asmkey.rules.path'. */
    private static void loadRules(DeclarationEnvironment env)
        throws ParserException
    {
        String path = System.getProperty("asmkey.rules.path");
        if (path == null) {
            path = System.getProperty("asmkey.base.path", System.getProperty("user.dir", "."))
                + "/rules";
        }
        loadRulesDirectory(env, new File(path));
    }

    /** Create an Integer LDT. */
    private static IntegerLDT createIntegerLDT(DefaultEnvironment env) {
        IntegerLDT intLdt = new IntegerLDT();
        Namespace ns = new Namespace();
        try {
            ns.add(env.getOperator("Z"));
            ns.add(env.getOperator("#"));
            ns.add(env.getOperator("neg"));
            for (char ch = '0'; ch <= '9'; ++ch) {
                ns.add(env.getOperator("" + ch));
            }
            intLdt.init(ns, env.getSort("int"),
                        SLListOfType.EMPTY_LIST.append(PrimitiveType.JAVA_INT));
        } catch (EnvironmentException ee) {
            ee.printStackTrace(System.err);
            System.exit(1);
        }
        return intLdt;
    }


    /* Overwrite method from super class. */
    protected void loadProblem(File file) {
        try {
            TacletIndex rules = new TacletIndex();
            /* meta ops */
            MetaJoin join = new MetaJoin();
            MetaEquiImp equiImp = new MetaEquiImp();
            MetaEquiS equiS = new MetaEquiS();
            /* create environment with list of meta operators */
            Object[] metaOps = new Object[] {
                join,
                equiImp,
                equiS,
                new MetaOpEqualArgs(),
                new MetaNamedRule(),
                Op.META_ADD,
                Op.META_SUB,
                Op.META_MUL,
                Op.META_DIV,
                Op.META_MOD,
                Op.META_LESS,
                Op.META_GREATER,
                Op.META_LEQ,
                Op.META_GEQ,
                Op.META_EQ,
            };
            DefaultEnvironment env = new DefaultEnvironment(Arrays.asList(metaOps));
            /* load default rules */
            loadRules(env);
            /* load problem */
            AstProblem problem = AstParser.parseProblem(new FileReader(file));
            TreeParser.parseDeclarations(env, problem.getDeclarations());
            /* set dynamic functions on meta operators */
            Collection dynamicFunctions = getDynamicFunctions(env.getOperators());
            join.setDynamicFunctions(dynamicFunctions);
            equiImp.setDynamicFunctions(dynamicFunctions);
            equiS.setDynamicFunctions(dynamicFunctions);
            /* parse the formula to be proven. */
            Term formula = TreeParser.parseTerm(env, problem.getFormula());
            /* taclets */
            for (Iterator i = env.getTaclets().iterator(); i.hasNext();) {
                Taclet taclet = (Taclet) i.next();
                rules.add(taclet);
            }
            ListOfBuiltInRule builtInRules =
                SLListOfBuiltInRule.EMPTY_LIST.prepend(
                   new SimplifyIntegerRule()).prepend(new StepRule());
            Proof p = new Proof("User defined problem", formula, rules, 
				new BuiltInRuleIndex(builtInRules));
            p.setNamespaces(env.getEnvAsNamespaceSet());
            p.setJavaServices(new Services());
            addProblem(p);

            ProcedurePool.put(mediator.getJavaInfo(), env.getProcedures());

            PresentationFeatures.initialize(mediator.func_ns(),
                                            mediator.getNotationInfo(),
                                            mediator.getSelectedProof());

            /* int integer LDT by hand. */
            IntegerLDT intLdt = createIntegerLDT(env);
            mediator.getJavaInfo().getTypeConverter().init(intLdt,
                                                           new BooleanLDT(),
                                                           new ObjectLDT());

            /* load the proof, if it exists. */
            if (problem.getProof() != null) {
                ProofLoader proofLoader = new ProofLoader(mediator);
                proofLoader.loadProof(ProofParser.parse(problem.getProof()));
            }
        } catch (IOException ioe) {
            JOptionPane.showMessageDialog(this,
                                          "Could not load Problem due "
                                          + "to an IO error. The file to "
                                          + "load does not exist or you do "
                                          + "not have the permission to"
                                          + " read it.\n"
                                          + ioe.getMessage(),
                                          "IO Error",
                                          JOptionPane.ERROR_MESSAGE);
        } catch (ParserException pe) {
            String message = pe.getMessage();
            Location loc = pe.getLocation();
            if (loc != null) {
                message += " in ";
                if (loc.getFilename() != null) {
                    message += loc.getFilename() + ", ";
                }
                message += "line " + loc.getLine() + ", column " + loc.getColumn();
            }
            JOptionPane.showMessageDialog(this,
                                          "Could not load Problem.\n" + message,
                                          "Error",
                                          JOptionPane.ERROR_MESSAGE);
            pe.printStackTrace(System.err);
        } catch (Exception e) {
            /* e.g. IllegalArgumentException in Term.checked(). */
            e.printStackTrace();
            JOptionPane.showMessageDialog(this,
                                          "Could not load problem/proof.\n" + "Cause: \n" + e.getMessage(),
                                          "Error",
                                          JOptionPane.ERROR_MESSAGE);
        }
    }

}
