// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//

package de.uka.ilkd.asmkey.logic;

import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.op.Operator2;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.logic.sort.Sort;

/** This class implements the interface Operator2 and contains static
 * fields for all ASM operators.
 *
 * @author Hubert Schmid
 */

public class AsmOperator implements Operator2 {

    /** modal operator for '[asm-rule] phi'. */
    public static final Operator2 BOX =
        new AsmOperator("Box", Sort.FORMULA, new Sort[] { AsmSort.ASMRULE, null }) {
            public boolean validTopLevel(Term term) {
                return super.validTopLevel(term)
                    && SortUtil.isPrimitiveSort(term.sub(1).sort());
            }
        };

    /** modal operator for '<asm-rule> phi'. */
    public static final Operator2 DIAMOND =
        new AsmOperator("Diamond", Sort.FORMULA, new Sort[] { AsmSort.ASMRULE, null }) {
            public boolean validTopLevel(Term term) {
                return super.validTopLevel(term)
                    && SortUtil.isPrimitiveSort(term.sub(1).sort());
            }
        };

    /** modal operator for 'def(asm-rule)'. */
    public static final Operator2 DEF = new AsmOperator("def", Sort.FORMULA, new Sort[] { AsmSort.ASMRULE });

    /** modal operator for 'upd(s <- f(x) in asm-rule)'. */
    public static final Operator2 UPD =
        new AsmOperator("upd", Sort.FORMULA, new Sort[] { null, null, AsmSort.ASMRULE }) {
            public boolean validTopLevel(Term term) {
                return super.validTopLevel(term)
                    && SortUtil.isPrimitiveSort(term.sub(0).sort())
                    && term.sub(0).sort() == term.sub(1).sort();
            }
        };

    /** modal operator for 'con(asm-rule)'. */
    public static final Operator2 CON = new AsmOperator("Con", Sort.FORMULA, new Sort[] { AsmSort.ASMRULE });

    /** modal operator for 'inv(s in asm-rule)'. */
    public static final Operator2 INV =
        new AsmOperator("inv", Sort.FORMULA, new Sort[] { null, AsmSort.ASMRULE }) {
            public boolean validTopLevel(Term term) {
                return super.validTopLevel(term)
                    && SortUtil.isPrimitiveSort(term.sub(0).sort());
            }
        };

    /** modal operator for 'joinable(asm-rule and asm-rule)'. */
    public static final Operator2 JOINABLE =
        new AsmOperator("joinable", Sort.FORMULA, new Sort[] { AsmSort.ASMRULE, AsmSort.ASMRULE });

    /** modal operator for 'equi(asm-rule and asm-rule)'. */
    public static final Operator2 EQUI =
        new AsmOperator("equi", Sort.FORMULA, new Sort[] { AsmSort.ASMRULE, AsmSort.ASMRULE });

    /** modal operator for 'equiS(asm-rule and asm-rule)'. */
    public static final Operator2 EQUIS =
        new AsmOperator("equiS", Sort.FORMULA, new Sort[] { AsmSort.ASMRULE, AsmSort.ASMRULE });

    /* ASMs */

    /** operator for sequential composition of asm rules. */
    public static final Operator2 SEQ =
        new AsmOperator("seq", AsmSort.ASMRULE, new Sort[] { AsmSort.ASMRULE, AsmSort.ASMRULE });

    /** operator for parallel composition of asm rules. */
    public static final Operator2 PAR =
        new AsmOperator("par", AsmSort.ASMRULE, new Sort[] { AsmSort.ASMRULE, AsmSort.ASMRULE });

    /** operator for asm skip rule. */
    public static final Operator2 SKIP = new AsmOperator("skip", AsmSort.ASMRULE);

    /** operator for asm assignment rule. */
    public static final Operator2 ASSIGN = new AsmOperator("assign", AsmSort.ASMRULE, new Sort[] { null, null }) {
            public boolean validTopLevel(Term term) {
                return super.validTopLevel(term)
                    && SortUtil.isPrimitiveSort(term.sub(0).sort())
                    && term.sub(0).sort() == term.sub(1).sort();
            }
        };

    /** operator for asm if rule. */
    public static final Operator2 IF =
        new AsmOperator("if", AsmSort.ASMRULE, new Sort[] { Sort.FORMULA, AsmSort.ASMRULE, AsmSort.ASMRULE });

    /** operator for asm let rule. */
    public static final Operator2 LET =
        new AsmOperator(
                        "let",
                        AsmSort.ASMRULE,
                        new Sort[] { null, AsmSort.ASMRULE },
                        new boolean[] { false, true }) {
            public boolean validTopLevel(Term term) {
                return super.validTopLevel(term)
                    && SortUtil.isPrimitiveSort(term.sub(0).sort());
            }
        };

    /** operator for asm forall rule. */
    public static final Operator2 FORALL =
        new AsmOperator(
                        "forall",
                        AsmSort.ASMRULE,
                        new Sort[] { Sort.FORMULA, AsmSort.ASMRULE },
                        new boolean[] { true, true });

    /** operator for asm try rule. */
    public static final Operator2 TRY =
        new AsmOperator("try", AsmSort.ASMRULE, new Sort[] { AsmSort.ASMRULE, AsmSort.ASMRULE });


    /** name of operator */
    private final Name name;

    /** (result) sort of operator */
    private final Sort sort;

    private final int arity;

    /** argument sorts of operator. If the value at the ith position
     * is 'null', then the operator may accept a term of arbitrary
     * sort. */
    private final Sort[] args;

    /** if boundPositions[i] is true, then logical variables are bound
     * in the i-th subterm. */
    private final boolean[] boundPositions;


    /** Creates an operator without arguments (constant). @see
     * #AsmOperator(String, Sort, Sort[]) */
    AsmOperator(String name, Sort sort) {
        this(name, sort, null, null);
    }

    /** Creates an operator that doesn't bound variables. @see
     * #AsmOperator(String, Sort, Sort[], boolean[]) */
    AsmOperator(String name, Sort sort, Sort[] args) {
        this(name, sort, args, null);
    }

    /** Creates an operator.
     *
     * @param name name of operator
     * @param sort sort of term with this operator as top level operator
     * @param args count and sorts of arguments. If sort[i] is null,
     * the i-th argument of this operator may be of arbitrary sort.
     * @param boundPositions indices of subterms that bound variables.
     */
    AsmOperator(String name, Sort sort, Sort[] args, boolean[] boundPositions) {
        this.name = new Name(name);
        this.sort = sort;
        this.args = args;
        this.arity = args == null ? 0 : args.length;
        this.boundPositions = boundPositions;
    }


    public final Name name() {
        return name;
    }

    public final Sort sort(Term[] term) {
        return sort;
    }

    public final int arity() {
        return arity;
    }

    public boolean validTopLevel(Term term) {
        if (arity != term.arity()) {
            return false;
        }
        if (args != null) {
            for (int i = 0; i < arity; ++i) {
                if (args[i] != null) {
                    /* I don't know why there is an extra check for
                     * schema variables, but 'Function.validTopLevel'
                     * does it also, -hs. */
                    if (term.sub(i).op() instanceof SchemaVariable) {
                        // nothing
                    } else if (!term.sub(i).sort().extendsTrans(args[i])) {
                        return false;
                    }
                }
            }
        }
        return true;
    }

    public final boolean boundVariables(int index) {
        return boundPositions != null && boundPositions[index];
    }

    public final Term createTerm(ArrayOfQuantifiableVariable vars, Term[] terms) {
        /* create a new term with the given bound variables and subterms. */
        if (terms != null) {
            for (int i = 0; i < terms.length; ++i) {
                if (terms[i] == null) {
                    throw new IllegalArgumentException("terms[" + i + "] may not be null.");
                }
            }
        }
        if (arity() == 0) {
            if (vars != null) {
                throw new IllegalArgumentException("Operator " + name() + " can not have bound variables.");
            }
            if (terms != null && terms.length != 0) {
                throw new IllegalArgumentException("Invalid number of subterms.");
            }
        } else {
            if (terms == null || terms.length != arity()) {
                throw new IllegalArgumentException("Arity of operator " + name() + " does not match.");
            }
            boolean bound = false;
            for (int i = 0; i < terms.length; ++i) {
                if (boundVariables(i)) {
                    bound = true;
                    break;
                }
            }
            if (bound == (vars == null || vars.size() == 0)) {
                throw new IllegalArgumentException("Operator " + name() + " does not match with bound variables.");
            }
        }
        Term result = new AsmTerm(this, terms, vars);
        if (!validTopLevel(result)) {
            throw new IllegalArgumentException("Term not valid.");
        }
        return result;
    }

    /** for debug only */
    public String toString() {
        return "[Operator:name=" + name + ",sort=" + sort + "]";
    }

    /**
     * @return true if the value of "term" having this operator as
     * top-level operator and may not be changed by modalities
     */
    public boolean isRigid (Term term) {
	// TODO
 	return false;
    }

}
