// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//

package de.uka.ilkd.asmkey.logic;


import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.Operator2;
import de.uka.ilkd.key.logic.sort.Sort;


/** Instances of this class represent named ASM rules (procedures).
 *
 * @author Hubert Schmid
 */

public final class AsmProcedure implements Named {

    /** Returns an array with the sorts of the formal parameters in
     * the same order. */
    private static Sort[] getSorts(FormalParameter[] params) {
        Sort[] sorts = new Sort[params.length];
        for (int i = 0; i < params.length; ++i) {
            sorts[i] = params[i].sort(null);
        }
        return sorts;
    }


    /** The identifier (name) of the named asm rule. */
    private final String id;

    /** The list of formal parameters. */
    private final FormalParameter[] params;

    /** The definining body of the named asm rule. */
    private Term body;

    /** Returns the operator representing invocations of this named
     * asm rule. */
    private final Operator2 op;


    /** creates a new named asm rule with uninitialized body. */
    public AsmProcedure(String id, FormalParameter[] params) {
        this.id = id;
        this.params = params;
        op = new Operator(id, getSorts(params));
    }


    /** returns the name of this named rule. This method is equal to
     * 'new Name(getId())'. */
    public Name name() {
        return new Name(id);
    }

    /** returns the name of this named rule. This method is equals to
     * 'name().toString()'. */
    public String getId() {
        return id;
    }

    /** returns the array of formal parameters. The array may not be
     * changed. */
    public FormalParameter[] getFormalParameters() {
        return params;
    }

    /** Sets the body of this named asm rule. */
    public void setBody(Term body) {
        this.body = body;
    }

    /** Returns the body of this named asm rule. */
    public Term getBody() {
        return body;
    }

    /** Returns the operator representing invocations of this named
     * asm rule. */
    public Operator2 getOp() {
        return op;
    }


    /** The inner class Operator extends AsmOperator and contains a
     * reference to the defintion of the named asm rule.  This
     * operator is used as operator for rule invocations, i.e. the
     * operator of 'call foo(1)'. */

    public class Operator extends AsmOperator {

        Operator(String id, Sort[] argSorts) {
            super(id, AsmSort.ASMRULE, argSorts);
        }


        /** Returns the defintion of the named asm rule. */
        AsmProcedure getProcedure() {
            return AsmProcedure.this;
        }

    }
}
