// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//

package de.uka.ilkd.asmkey.logic;


import java.util.Collection;
import java.util.Iterator;

import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.TermFactory;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.rule.inst.SVInstantiations;


/** Meta operator expands the joinable operator. This meta operator
 * expands a joinable term with its definition.
 *
 * @author Hubert Schmid
 */

public final class MetaJoin extends AsmMetaOperator {

    /** Short refernce to TermFactory. */
    private static final TermFactory tf = TermFactory.DEFAULT;

    /** The MetaJoin operator has to know all dynamic function
     * symbols.  This list should be initialized on problem loading,
     * because all dynamic function symbols are declared in the
     * problem file.  The elements of this collection are instances of
     * type 'Function'. */
    private Collection functions;


    /** creates a new MetaJoin operator (with uninitialized list of
     * dynamic function symbols). */
    public MetaJoin() {
        super("#ASM_JOIN", Sort.FORMULA,
              new Sort[] { AsmSort.ASMRULE, AsmSort.ASMRULE });
    }


    /** Set the list of dynamic function symbols that are used by this
     * meta operator. This method can only be called once.  The list
     * of dynamic function symbols is typically known on problem
     * loading. The elements of the collection have to be instances of
     * type 'Function'. */
    public void setDynamicFunctions(Collection c) {
        if (c == null) {
            throw new NullPointerException();
        } else if (functions != null) {
            throw new IllegalStateException("Dynamic functions on ASM_JOIN already set.");
        }
        functions = c;
    }

    /** This method replaces a joinable operator with two asm rules
     * (subterms) with the defintion of joinable. */
    public Term calculate(Term term, SVInstantiations svInst) {
        if (functions == null) {
            throw new IllegalStateException("Dynamic functions are not initialized.");
        }
        Term result = tf.createJunctorTerm(Op.TRUE);
        for (Iterator i = functions.iterator(); i.hasNext();) {
            Function f = (Function) i.next();
            if (f instanceof DynamicFunction) {
                LogicVariable y = TermUtil.createNewVariable(term, "y", f.sort());
                LogicVariable z = TermUtil.createNewVariable(term, "z", f.sort());
                Term ty = tf.createVariableTerm(y);
                Term tz = tf.createVariableTerm(z);
                LogicVariable[] x = new LogicVariable[f.arity()];
                Term[] terms = new Term[f.arity()];
                for (int j = 0; j < f.arity(); ++j) {
                    x[j] = TermUtil.createNewVariable(term, "x", f.argSort(j));
                    terms[j] = tf.createVariableTerm(x[j]);
                }
                Term tx = tf.createFunctionTerm(f, terms);
                Term c1 = AsmOperator.UPD.createTerm(null, new Term[] { tx, ty, term.sub(0)});
                Term c2 = AsmOperator.UPD.createTerm(null, new Term[] { tx, tz, term.sub(1)});
                Term t =
                    tf.createJunctorTerm(Op.IMP, tf.createJunctorTerm(Op.AND, c1, c2), tf.createEqualityTerm(ty, tz));
                t = tf.createQuantifierTerm(Op.ALL, z, t);
                t = tf.createQuantifierTerm(Op.ALL, y, t);
                for (int j = x.length - 1; j >= 0; --j) {
                    t = tf.createQuantifierTerm(Op.ALL, x[j], t);
                }
                result = tf.createJunctorTerm(Op.AND, result, t);
            }
        }
        return result;
    }

}
