// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//

package de.uka.ilkd.asmkey.parser;


import java.io.IOException;
import java.io.Reader;

import de.uka.ilkd.asmkey.parser.antlr.AsmKeyLexer;
import de.uka.ilkd.asmkey.parser.antlr.AsmKeyParser;
import de.uka.ilkd.asmkey.parser.ast.*;
import de.uka.ilkd.key.parser.Location;
import de.uka.ilkd.key.parser.ParserException;


/** This class contains static methods for parsing terms with the
 *  generated (antlr) parser. It allows parsing of terms, ASM rules
 *  and problem files. The methods return abstract syntax tree, that
 *  represent the context * free structure of the grammer. The
 *  analysis of static semantic (like variable binding and types) is
 *  not done.
 *
 * @author Hubert Schmid
 */

public final class AstParser {

    /** The used AstFactory. */
    private static final AstFactory af = new AstFactoryImpl();


    /** Parse a problem file. */
    public static AstProblem parseProblem(Reader reader) throws ParserException {
        try {
            AstProblem problem = parser(reader).top_problem();
            checkEof(reader);
            return problem;
        } catch (antlr.TokenStreamException tse) {
            tse.printStackTrace();
            throw new ParserException(tse.getMessage(), null);
        } catch (antlr.RecognitionException re) {
            throw new ParserException(re.getMessage(), new Location(re.getFilename(), re.getLine(), re.getColumn()));
        }
    }

    /** Parse term. */
    public static AstTerm parseTerm(Reader reader) throws ParserException {
        try {
            AstTerm term = parser(reader).top_term();
            checkEof(reader);
            return term;
        } catch (antlr.TokenStreamException tse) {
            throw new ParserException(tse.getMessage(), null);
        } catch (antlr.RecognitionException re) {
            throw new ParserException(re.getMessage(), new Location(re.getFilename(), re.getLine(), re.getColumn()));
        }
    }

    /** Parse ASM rule. */
    public static AstAsmRule parseAsmRule(Reader reader) throws ParserException {
        try {
            AstAsmRule asmRule = parser(reader).top_asm_rule();
            checkEof(reader);
            return asmRule;
        } catch (antlr.TokenStreamException tse) {
            throw new ParserException(tse.getMessage(), null);
        } catch (antlr.RecognitionException re) {
            throw new ParserException(re.getMessage(), new Location(re.getFilename(), re.getLine(), re.getColumn()));
        }
    }

    /** Parse problem declarations (variables, predicates, ...). */
    public static AstDeclaration[] parseDeclarations(Reader reader) throws ParserException {
        try {
            AstDeclaration[] decls = parser(reader).top_decls();
            checkEof(reader);
            return decls;
        } catch (antlr.TokenStreamException tse) {
            throw new ParserException(tse.getMessage(), null);
        } catch (antlr.RecognitionException re) {
            throw new ParserException(re.getMessage(), new Location(re.getFilename(), re.getLine(), re.getColumn()));
        }
    }

    /** returns an AsmKeyParser instance. */
    private static AsmKeyParser parser(Reader reader) {
        return new AsmKeyParser(new AsmKeyLexer(reader), af);
    }

    /** Checks if the end-of-file has been reached. If eof has not
     * been reached, this method throws a ParserException. */
    private static void checkEof(Reader reader) throws ParserException {
        try {
            if (reader.read() != -1) {
                throw new ParserException("AstParser exited before end of file.", null);
            }
        } catch (IOException ioe) {
            ioe.printStackTrace(System.err);
        }
    }

}
