// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//

package de.uka.ilkd.asmkey.parser;


import java.util.List;

import de.uka.ilkd.asmkey.parser.ast.*;
import de.uka.ilkd.asmkey.parser.env.DeclarationEnvironment;
import de.uka.ilkd.asmkey.parser.env.TermEnvironment;
import de.uka.ilkd.asmkey.parser.tree.DeclarationParser;
import de.uka.ilkd.asmkey.parser.tree.TermParser;
import de.uka.ilkd.key.logic.Sequent;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.parser.ParserException;


/** Static methods to parse AST nodes in environments. The methods
 * throw a ParserException if an error occurs.
 *
 * @author Hubert Schmid
 */

public final class TreeParser {

    /** parse the given array of declarations in the environment 'env'
     * and add the declarations to this environment. */
    public static void parseDeclarations(DeclarationEnvironment env, AstDeclaration[] decls) throws ParserException {
        DeclarationParser parser = new DeclarationParser(env);
        parser.parse(decls);
    }

    /** parse the given list of declarations in the environment 'env'
     * and add the declarations to this environment.  The elements of
     * 'list' have to be of type {@link AstDeclaration}. */
    public static void parseDeclarations(DeclarationEnvironment env, List list) throws ParserException {
        parseDeclarations(env, (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]));
    }

    /** Parse the given abstract term in the environment 'env' and
     * return the concrete term. */
    public static Term parseTerm(TermEnvironment env, AstTerm term) throws ParserException {
        return parseTerm(env, term, null);
    }

    /** Parse term in the given environment and the bounds logic
     * variables. The elements of 'logicVariables' have to be of type
     * {@link de.uka.ilkd.key.logic.op.LogicVariable}. */
    public static Term parseTerm(TermEnvironment env, AstTerm term, List logicVariables) throws ParserException {
        TermParser parser = new TermParser(env);
        return parser.parse(term, logicVariables);
    }

    /** Parse the asm rule in the given environment. */
    public static Term parseAsmRule(TermEnvironment env, AstAsmRule rule) throws ParserException {
        return parseAsmRule(env, rule, null);
    }

    /** Parse ASM rule in the context of a list of logic
     * variables. The elements of 'logicVariables' have to be of type
     * {@link de.uka.ilkd.key.logic.op.LogicVariable}. */
    public static Term parseAsmRule(TermEnvironment env, AstAsmRule rule, List logicVariables) throws ParserException {
        TermParser parser = new TermParser(env);
        return parser.parse(rule, logicVariables);
    }

    /** Parse a sequent in the given environment. */
    public static Sequent parseSequent(TermEnvironment env, AstSequent sequent) throws ParserException {
        TermParser parser = new TermParser(env);
        return parser.parse(sequent);
    }

}
