// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

header {

    package de.uka.ilkd.asmkey.parser.antlr;

    import de.uka.ilkd.asmkey.parser.*;
    import de.uka.ilkd.asmkey.parser.ast.*;
    import de.uka.ilkd.key.parser.Location;

    import java.util.*;

}


class AsmKeyParser extends Parser;


options {
	importVocab = AsmKeyLexer;
	k = 2;
	defaultErrorHandler = false;
}


{

    /* --- static methods --- */

    private static final Location getLocation(AstDeclaration[] decls) {
        if (decls.length == 0) {
            return null;
        } else {
            return decls[0].getLocation();
        }
    }

    private static final Location getLocation(AstTerm[] terms) {
        if (terms.length == 0) {
            return null;
        } else {
            return terms[0].getLocation();
        }
    }

    /** Return a new empty list. */
    private static List createList() {
        return new ArrayList();
    }

    private static Location createLocation(Token token) {
        return new Location(null, token.getLine(), token.getColumn());
    }

    /** return a new array 't # terms'. */
    private static AstTerm[] prependTerm(AstTerm t, AstTerm[] terms) {
        AstTerm[] a = new AstTerm[terms.length + 1];
        System.arraycopy(terms, 0, a, 1, terms.length);
        a[0] = t;
        return a;
    }


    /* --- static fields --- */

    private static final AstTerm[] EMPTY_TERMS = new AstTerm[]{ };

    private static final Identifier[] EMPTY_IDS = new Identifier[]{ };

    private static final AstCondition[] EMPTY_CONDITIONS = new AstCondition[]{ };

    private static final AstParameter[] EMPTY_PARAMETERS = new AstParameter[]{ };

    
    /* --- constructors --- */

    /** Create a parser with the given input and the given ast factory. */
    public AsmKeyParser(TokenStream lexer, AstFactory af) {
        this(lexer);
        this.af = af;
    }


    /* --- fields --- */

    /** The used AstFactory. */
    private AstFactory af = null;


    /* --- inner class --- */

    /** This class stores either a term or a sequent. It is needed
     * because some productions can't distinguish between terms and 
     * sequents without infinte lookahead.  */
    private static final class TermSequent {
        
        final AstTerm term;
        
        final AstSequent sequent;


        TermSequent(AstTerm term) {
            this(term, null);
        }

        TermSequent(AstSequent sequent) {
            this(null, sequent);
        }

        TermSequent(AstTerm term, AstSequent sequent) {
            this.term = term;
            this.sequent = sequent;
        }

    }
    
}


top_formula returns [AstTerm t]
    :
        t=term EOF
    ;

top_term returns [AstTerm t]
    :
        t=term EOF
    ;

top_problem returns [AstProblem p]
    :
        p=problem EOF
    ;

top_decls returns [AstDeclaration[] decls]
    :
        decls=decls EOF
    ;

top_asm_rule returns [AstAsmRule r]
    :
        r=asm_rule EOF
    ;

protected
ident returns [Identifier s]
	:
		id:IDENT
        {
			s = new Identifier(createLocation(id), id.getText());
		}
    |
        /* '+' is also a valid identifier, because it is used as
         * function symbol. This is a very ugly. */
        plus:PLUS
        {
            s = new Identifier(createLocation(plus), "+");
        }
	;


protected
id_list returns [Identifier[] ids]
{
    Identifier id;
    List list = createList();
}
    :
        id=ident
        {
            list.add(id);
        }
        (
            COMMA id=ident
            {
                list.add(id);
            }
        )*
        {
            ids = (Identifier[]) list.toArray(new Identifier[list.size()]);
        }
    ;


/**
 * In the special but important case of Taclets, we don't yet know
 * whether we are going to have a term or a formula, and it is hard
 * to decide a priori what we are looking at.  So the `term'
 * non-terminal will recognize a term or a formula.  The `formula'
 * reads a formula/term and throws an error if it wasn't a formula.
 * This gives a rather late error message. */


protected
formula returns [AstTerm t]
    :
        t=term
    ;


protected
term returns [AstTerm t]
    :
        t=term20
    ;


protected
term20 returns [AstTerm t1]
{
    AstTerm t2;
}
    :
        t1=term30
        (
            eqv:EQV t2=term20
            {
                t1 = af.createTerm(t1.getLocation(), 
                    new AstOperator(createLocation(eqv), Operator.EQV),
                    new AstTerm[]{ t1, t2 });
            }
        )?
    ;


protected
term30 returns [AstTerm t1]
{
    AstTerm t2;
}
    :
        t1=term40
        (
            imp:IMP t2=term30
            {
                t1 = af.createTerm(t1.getLocation(), 
                    new AstOperator(createLocation(imp), Operator.IMP), 
                    new AstTerm[]{ t1, t2 });
            }
        )?
    ;


protected
term40 returns [AstTerm t1]
{
    AstTerm t2;
}
    :
        t1=term50
        (
            or:OR t2=term40
            {
                t1 = af.createTerm(t1.getLocation(), 
                    new AstOperator(createLocation(or), Operator.OR), 
                    new AstTerm[]{ t1, t2 });
            }
        )?
    ;


protected
term50 returns [AstTerm t1]
{
    AstTerm t2;
}
    :
        t1=term60
        (
            and:AND t2=term50
            {
                t1 = af.createTerm(t1.getLocation(), 
                    new AstOperator(createLocation(and), Operator.AND), 
                    new AstTerm[]{ t1, t2});
            }
        )?
    ;


protected
term60 returns [AstTerm t]
    :
        t=term70
    |
        t=bindingterm
    |
        t=dl_term
    |
        not:NOT t=term60
        {
            t = af.createTerm(createLocation(not), 
                new AstOperator(createLocation(not), Operator.NOT), 
                new AstTerm[]{ t });
        }
    ;


protected
term70 returns [AstTerm t1]
{
    AstTerm t2;
}
    :
        t1=term80
        (
            equals:EQUALS t2=term80
            {
                t1 = af.createTerm(t1.getLocation(), 
                    new AstOperator(createLocation(equals), Operator.EQUALS), 
                    new AstTerm[]{ t1, t2 });
            }
        )?
    ;


protected
term80 returns [AstTerm t1]
{
    AstTerm t2;
}
    :
        t1=term90
        (
            (
                asterix:ASTERIX t2=term80
                {
                    t1 = af.createTerm(t1.getLocation(), 
                        new AstOperator(createLocation(asterix), Operator.MULT),
                        new AstTerm[]{ t1, t2 });
                }
            )
        |
            (
                slash:SLASH t2=term80
                {
                    t1 = af.createTerm(t1.getLocation(), 
                        new AstOperator(createLocation(slash), Operator.DIV),
                        new AstTerm[]{ t1, t2 });
                }
            )
        )?
    ;


protected
term90 returns [AstTerm t1]
{
    AstTerm t2;
}
    :
        t1=term100
        (
            (
                plus:PLUS t2=term90
                {
                    t1 = af.createTerm(t1.getLocation(), 
                        new AstOperator(createLocation(plus), Operator.ADD),
                        new AstTerm[]{ t1, t2 });
                }
            )
        |
            (
                minus:MINUS t2=term90
                {
                    t1 = af.createTerm(t1.getLocation(), 
                        new AstOperator(createLocation(minus), Operator.SUB),
                        new AstTerm[]{ t1, t2 });
                }
            )
        )?
    ;


protected
term100 returns [AstTerm t]
{
    Location location;
    Identifier abbr;
}
    :
        t=funcpredvarterm
    |
        LPAREN t=term RPAREN
    |
        true_:"true"
        {
            t = af.createTerm(createLocation(true_), 
                new AstOperator(createLocation(true_), Operator.TRUE),
                null);
        }
    |
        false_:"false"
        {
            t = af.createTerm(createLocation(false_), 
                new AstOperator(createLocation(false_), Operator.FALSE), 
                null);
        }
    |
        at:AT abbr=ident
        {
            t = af.createAbbreviationTerm(createLocation(at), abbr);
        }
    ;


protected
bindingterm returns [AstTerm t]
{
    AstQuantifier q;
    Identifier var;
    AstTerm t1, t2;
    Identifier sort = null;
}
    :
        (
            q=quantifier var=ident ( COLON sort=ident )? DOT t1=term60
            {
                t = af.createTerm(q.getLocation(), q, var, sort, t1);
            }
        |
            LBRACE var=ident ( COLON sort=ident )? t1=term RBRACE t2=term60
            {
                t = af.createSubstitutionTerm(var.getLocation(), var, sort, t1, t2);
            }
        )
    ;


protected
quantifier returns [AstQuantifier q]
    :
        all:"all"
        {
            q = new AstQuantifier(createLocation(all), Quantifier.ALL);
        }
    |
        ex:"ex"
        {
            q = new AstQuantifier(createLocation(ex), Quantifier.EX);
        }
    ;


protected
dl_term returns [AstTerm t]
{
    AstAsmRule r, s;
    AstTerm t1, t2;
}
    :
        lbrack:LBRACK r=asm_rule RBRACK t=term60
        {
            t = af.createDlBox(createLocation(lbrack), r, t);
        }
    |
        lt:LT r=asm_rule GT t=term60
        {
            t = af.createDlDiamond(createLocation(lt), r, t);
        }
    |
        def:"def" LPAREN r=asm_rule RPAREN
        {
            t = af.createDlDef(createLocation(def), r);
        }
    |
        upd:"upd" LPAREN t1=term UPDATE t2=term "in" r=asm_rule RPAREN
        {
            t = af.createDlUpd(createLocation(upd), t1, t2, r);
        }
    |
        con:"Con" LPAREN r=asm_rule RPAREN
        {
            t = af.createDlCon(createLocation(con), r);
        }
    |
        inv:"inv" LPAREN t=term "in" r=asm_rule RPAREN
        {
            t = af.createDlInv(createLocation(inv), t, r);
        }
    |
        joinable:"joinable" LPAREN r=asm_rule "and" s=asm_rule RPAREN
        {
            t = af.createDlJoinable(createLocation(joinable), r, s);
        }
    |
        equi:"equi" LPAREN r=asm_rule "and" s=asm_rule RPAREN
        {
            t = af.createDlEqui(createLocation(equi), r, s);
        }
    |
        equiS:"equiS" LPAREN r=asm_rule "and" s=asm_rule RPAREN
        {
            t = af.createDlEquiS(createLocation(equiS), r, s);
        }
    ;


protected
term_list returns [AstTerm[] terms]
{
    List list = createList();
    AstTerm t;
}
    :
        t=term
        {
            list.add(t);
        }
        (
            COMMA t=term
            {
                list.add(t);
            }
        )*
        { 
            terms = (AstTerm[]) list.toArray(new AstTerm[list.size()]);
        }
    ;


protected
funcpredvarterm returns [AstTerm a]
{
    Identifier id;
    AstTerm[] l;
}
    :
        id=ident l=term_args
        {
            a = af.createTerm(id.getLocation(), id, l);
        }
    ;


protected
term_args returns [AstTerm[] terms = null]
    :
        ( 
            LPAREN
            (
                terms=term_list RPAREN
            |
                RPAREN
                {
                    terms = EMPTY_TERMS;
                }
            )
        )?
    ;


protected
seq returns [AstSequent s]
{
    AstTerm[] ant = EMPTY_TERMS, suc = EMPTY_TERMS;
} 
    : 
        (
            ant=term_list 
        )?
        SEQARROW
        (
            suc=term_list
        )?
        { 
            s = new AstSequent(getLocation(ant), ant, suc);
        }
    ;


protected
termorseq returns [TermSequent ts]
{
    AstTerm t;
    AstTerm[] ant, suc = EMPTY_TERMS;
}
    :
        (
            t=term
            (
                /* empty */
                {
                    ts = new TermSequent(t);
                }
            |
                SEQARROW
                (
                    suc=term_list
                )?
                {
                    ts = new TermSequent(new AstSequent(t.getLocation(), new AstTerm[]{ t }, suc));
                }
            |
                COMMA ant=term_list SEQARROW
                (
                    suc=term_list
                )?
                {
                    ts = new TermSequent(new AstSequent(t.getLocation(), prependTerm(t, ant), suc));
                }
            )
        |
            seqarrow:SEQARROW 
            (
                suc=term_list
            )?
            {
                ts = new TermSequent(new AstSequent(createLocation(seqarrow), EMPTY_TERMS, suc));
            }
        )
    ;


protected
taclet returns [AstTaclet r]
{ 
    AstSequent s = null;
    TermSequent ts;
    AstCondition[] conds = EMPTY_CONDITIONS;
    AstGoal[] goals;
    AstTacletModifier[] mods;
    Location location = null;
}
	: 
		( 
            if_:"if" LPAREN s=seq RPAREN 
            {
                location = createLocation(if_);
            }
        )?
		( 
            find:"find" LPAREN ts=termorseq RPAREN 
            {
                if (location == null) {
                    location = createLocation(find);
                }
            }
        )
        (
            ( "varcond" | "cond" ) LPAREN conds=conditions RPAREN
        )?
		goals=goalspecs
        mods=modifiers
        {
            if (ts.term != null) {
                r = new AstTaclet(location, s, ts.term, conds, goals, mods);
            } else {
                r = new AstTaclet(location, s, ts.sequent, conds, goals, mods);
            }
        }
    ;


protected
conditions returns [AstCondition[] conds]
{
    AstCondition cond;
    List list = createList();
}
    :
        cond=condition
        {
            list.add(cond);
        }
        (
            COMMA cond=condition
            {
                list.add(cond);
            }
        )*
        {
            conds = (AstCondition[]) list.toArray(new AstCondition[list.size()]);
        }
    ;


protected
condition returns [AstCondition cond]
{
    Identifier v, w;
}
    :
        op:"op" v=ident 
        (
            EQUALS w=ident
            {
                cond = af.createOperatorEqualsCondition(createLocation(op), true, v, w);
            }
        |
            NOTEQUALS w=ident
            
            {
                cond = af.createOperatorEqualsCondition(createLocation(op), false, v, w);
            }
        )
    |
        v=ident "not" "free" "in" w=ident
        {
            cond = af.createNotFreeCondition(v.getLocation(), v, w);
        }
    |
		v=ident "new" "depending" "on" w=ident
        { 
            cond = af.createSkolemCondition(v.getLocation(), v, w);
        }
    |
        pure:"pure" v=ident
        {
            cond = af.createPureCondition(createLocation(pure), v);
        }
    |
        static_:"static" v=ident
        {
            cond = af.createStaticCondition(createLocation(static_), v);
        }
    |
        staticargs:"staticargs" v=ident
        {
            cond = af.createStaticArgsCondition(createLocation(staticargs), v);
        }
    |
        dynamic:"dynamic" v=ident
        {
            cond = af.createDynamicCondition(createLocation(dynamic), v);
        }
    |
        call:"call" v=ident
        {
            cond = af.createCallCondition(createLocation(call), v);
        }
    ;


protected
goalspecs returns [AstGoal[] goals]
    :
		"close" "goal"
        {
            goals = new AstGoal[]{ };
        }
    | 
        goals=goalspec_list
    ;


protected
goalspec_list returns [AstGoal[] goals]
{
    List list = createList();
    AstGoal g;
}
    :
        g=goalspec
        {
            list.add(g);
        }
        (
            SEMI g=goalspec
            {
                list.add(g);
            }
        )*
        { 
            goals = (AstGoal[]) list.toArray(new AstGoal[list.size()]);
        }
    ;


protected
goalspec returns [AstGoal goal] 
    :
        goal=replacewith
    ;


protected
replacewith returns [AstGoal goal]
{
    Location location = null;
    TermSequent ts = null;
    AstSequent s = null;
    AstTacletDeclaration[] l = null;
    String name = null;
}
    :
		(
            rw:"replacewith" 
            (
                str:STRING_LITERAL 
                {
                    name = str.getText();
                }
            )?
            LPAREN ts=termorseq RPAREN
            {
                if (location == null) {
                    location = createLocation(rw);
                }
            }
        )?
        (
            add:"add" LPAREN s=seq RPAREN
            {
                if (location == null) {
                    location = createLocation(add);
                }
            }
        )?
        (
            addrules:"addrules" LPAREN l=taclet_decl_list RPAREN
            {
                if (location == null) {
                    location = createLocation(addrules);
                }
            }
        )?
        {
            if (ts == null) {
                goal = af.createReplaceGoal(location, name, (AstTerm) null, s, l);
            } else if (ts.term != null) {
                goal = af.createReplaceGoal(location, name, ts.term, s, l);
            } else {
                goal = af.createReplaceGoal(location, name, ts.sequent, s, l);
            }
        }
    ;


protected
taclet_decl_list returns [AstTacletDeclaration[] l]
{
    AstTacletDeclaration decl;
    List list = createList();
}
    :
        decl=taclet_decl
        {
            list.add(decl);
        }
        (
            COMMA decl=taclet_decl
            {
                list.add(decl);
            }
        )*
        {
            l = (AstTacletDeclaration[]) list.toArray(new AstTacletDeclaration[list.size()]);
        }
    ;


protected
modifiers returns [AstTacletModifier[] l]
{
    List list = createList();
    AstTacletModifier tm;
}
    :
        (
            tm=modifier
            {
                list.add(tm);
            }
        )*
        {
            l = (AstTacletModifier[]) list.toArray(new AstTacletModifier[list.size()]);
        }
    ;


protected
modifier returns [AstTacletModifier tm]
{
    Identifier[] ids;
}
    :
        heurs:"heuristics" LPAREN ids=id_list RPAREN
        {
            tm = af.createTacletHeuristics(createLocation(heurs), ids);
        }
    |
        ni:"noninteractive"
        {
            tm = af.createTacletNonInteractiveModifier(createLocation(ni));
        }
    |
        re:"recursive"
        {
            tm = af.createTacletRecursiveModifier(createLocation(re));
        }
    |
        dn:DISPLAYNAME dname:STRING_LITERAL
        {
            tm = af.createTacletDisplayNameModifier(createLocation(dn), dname.getText());
        }
    |
        ht:HELPTEXT htext:STRING_LITERAL
        {
            tm = af.createTacletHelpTextModifier(createLocation(ht), htext.getText());
            // b.setHelpText(htext.getText());
        }
    ;
    

protected
problem returns [AstProblem p]
{
    AstDeclaration[] l;
    AstTerm a;
    AstProofExpression pe = null;
}
	:
 		l=decls
        "problem" LBRACE a=formula RBRACE
        (
            pe=proof
        )?
        {
            p = new AstProblem(getLocation(l), l, a, pe);
        }
	;


protected
decls returns [AstDeclaration[] decls]
{
    List list = createList();
    AstDeclaration[] l;
}
    :
		(
            (
                l=sort_decls
            |
                l=schema_var_decls
            |
                l=pred_decls
            |
                l=func_decls
            |
                l=heur_decls
            |
                l=taclet_decls
            |
                l=procedure_decls
            )
            {
                list.addAll(Arrays.asList(l));
            }
		)*
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
	;


protected
sort_decls returns [AstDeclaration[] decls] 
{
	List list = createList();
    AstDeclaration decl;
}
    :
		"sorts" 
		LBRACE
		(
			decl=sort_decl
            {
                list.add(decl);
            }
		)*
		RBRACE
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
	;


protected
sort_decl returns [AstDeclaration decl]
{
    Identifier id;
}
    :
        id=ident SEMI
        {
            decl = af.createSortDeclaration(id.getLocation(), id);
        }
    ;


protected
schema_var_decls returns [AstDeclaration[] decls]
{
	List list = createList();
    AstDeclaration[] decl;
}
    :
		"schema" "variables"
		LBRACE
		(
			decl=schema_var_decl
            {
                list.addAll(Arrays.asList(decl));
            }
		)*
		RBRACE
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
    ;


protected
schema_var_decl returns [AstDeclaration[] decls]
{
	AstSchemaType r;
    Identifier[] l;
}
    :
        r=schema_type_ref l=id_list SEMI
        {
            decls = new AstDeclaration[l.length];
            for (int i = 0; i < l.length; ++i) {
                decls[i] = af.createSchemaVariableDeclaration(r.getLocation(), l[i], r);
            }
        }
    ;


protected
schema_type_ref returns [AstSchemaType r]
{
    Identifier sort;
    boolean variable = false;
    boolean depending = false;
}
    :
        (
            asm:"asm" "rule"
            {
                r = af.createSchemaType(createLocation(asm), PrimitiveSchemaType.ASMRULE);
            }
        | 
            formula:"formula"
            { 
                r = af.createSchemaType(createLocation(formula), PrimitiveSchemaType.FORMULA);
            }			
        | 
            sort=ident 
            ( 
                "variables" 
                { 
                    variable = true; 
                }
            |
                "depending"
                {
                    depending = true;
                }
            )?
            { 
                if (depending) {
                    r = af.createDependingSchemaType(sort.getLocation(), sort);
                } else {
                    r = af.createSchemaType(sort.getLocation(), sort, variable);
                }
            }
        )

    ;


protected
pred_decls returns [AstDeclaration[] decls]
{
    List list = createList();
    AstDeclaration decl;
}
	:
		"predicates"
		LBRACE
		(
            decl=pred_decl
            {
                list.add(decl);
            }
        )*
        RBRACE
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
    ;


protected
pred_decl returns [AstDeclaration decl]
{
    Identifier id;
    Identifier[] args;
}
    :

        id=ident args=arg_ids SEMI
        {
            decl = af.createPredicateDeclaration(id.getLocation(), id, args);
        }
	;


protected
func_decls returns [AstDeclaration[] decls]
{
    List list = createList();
    AstDeclaration decl;
}
	:
		"functions"
		LBRACE
		(
            decl=func_decl
            {
                list.add(decl);
            }
        )*
        RBRACE
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
    ;


protected
func_decl returns [AstDeclaration decl]
{
    Identifier ret;
    Identifier id;
	Identifier[] args;
    boolean dynamic = false;
    Location location = null;
}
	:
        (
            d:"dynamic" 
            { 
                dynamic = true;
                location = createLocation(d);
            }
        )?
        ret=ident id=ident args=arg_ids SEMI
        {
            decl = af.createFunctionDeclaration(location != null ? location : ret.getLocation(),
                id, args, ret, dynamic);
        }
	;


protected
heur_decls returns [AstDeclaration[] decls]
{
    List list = createList();
    AstDeclaration decl;
}
	:
		"heuristics"
		LBRACE
		(
            decl=heur_decl
            {
                list.add(decl);
            }
        )*
        RBRACE
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
    ;


protected
heur_decl returns [AstDeclaration decl]
{
    Identifier id;
}
    :
        id=ident SEMI
        {
            decl = af.createHeuristicDeclaration(id.getLocation(), id);
        }
	;


protected
arg_ids returns [Identifier[] sorts = null] 
	:
		(
			LPAREN
            (
                sorts=id_list RPAREN
            |
                RPAREN
                {
                    sorts = EMPTY_IDS;
                }
            )
		)?
	;


protected
taclet_decls returns [AstDeclaration[] decls] 
{
	List list = createList();
    AstDeclaration d;
}
    :
		"rules" LBRACE
		(
			d=taclet_decl SEMI
            {
                list.add(d);
            }
		)*
		RBRACE
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
	;


protected
taclet_decl returns [AstTacletDeclaration decl]
{
    Identifier id;
    AstTaclet t;
}
    :
		id=ident LBRACE
        t=taclet
        RBRACE
        {
            decl = new AstTacletDeclaration(id.getLocation(), id, t);
        }
    ;


protected
procedure_decls returns [AstDeclaration[] decls]
{
	List list = createList();
    AstDeclaration d;
}
    :
		"procedures" LBRACE
		(
			d=procedure_decl SEMI
            {
                list.add(d);
            }
		)*
		RBRACE
        {
            decls = (AstDeclaration[]) list.toArray(new AstDeclaration[list.size()]);
        }
	;


protected
procedure_decl returns [AstDeclaration d]
{
    Identifier id;
    AstParameter[] params = EMPTY_PARAMETERS;
    AstAsmRule r;
}
    :
        id=ident 
        (
            LPAREN 
            (
                params=param_list 
            )?
            RPAREN 
        )?
        LBRACE r=asm_rule RBRACE
        {
            d = af.createProcedureDeclaration(id.getLocation(), id, params, r);
        }
    ;


protected
asm_rule returns [AstAsmRule r]
    :
        r=asm_rule_seq
    ;

protected
asm_rule_seq returns [AstAsmRule r1]
{
    AstAsmRule r2;
}
    :
        r1=asm_rule_par 
        ( 
            SEMI r2=asm_rule_seq 
            { 
                r1 = af.createAsmSeq(r1.getLocation(), r1, r2);
            }
        )?
    ;

protected
asm_rule_par returns [AstAsmRule r1]
{
    AstAsmRule r2;
}
    :
        r1=asm_rule_atom 
        ( 
            COMMA r2=asm_rule_par 
            {
                r1 = af.createAsmPar(r1.getLocation(), r1, r2);
            }
        )?
    ;

protected
asm_rule_atom returns [AstAsmRule r]
{
    Identifier id;
}
    :
        (
            r=asm_rule_skip
        |
            r=asm_rule_assign
        |
            r=asm_rule_call
        |
            r=asm_rule_if
        |
            r=asm_rule_let
        |
            r=asm_rule_forall
        |
            r=asm_rule_try
        |
            r=asm_rule_subst
        |
            r=asm_rule_meta
        |
            id=ident
            {
                r = af.createAsmSchemaVariable(id.getLocation(), id);
            }
        |
            LPAREN r=asm_rule RPAREN
        )
    ;


protected
asm_rule_skip returns [AstAsmRule r]
    :
        skip:"skip"
        {
            r = af.createAsmSkip(createLocation(skip));
        }
    ;


protected
asm_rule_assign returns [AstAsmRule r]
{
    AstTerm t1, t2;
}
    :
        t1=funcpredvarterm ASSIGN t2=term
        {
            r = af.createAsmAssign(t1.getLocation(), t1, t2);
        }
    ;


protected
asm_rule_call returns [AstAsmRule r]
{
    Identifier id;
    AstTerm[] l;
}
    :
        "call" id=ident l=term_args
        {
            r = af.createAsmCall(id.getLocation(), id, l);
        }
    ;


protected
asm_rule_if returns [AstAsmRule r]
{
    AstTerm t;
    AstAsmRule r1, r2;
}
    :
        if_:"if" t=formula "then" r1=asm_rule "else" r2=asm_rule "end"
        {
            r = af.createAsmIf(createLocation(if_), t, r1, r2);
        }
    ;


protected
asm_rule_let returns [AstAsmRule r]
{
    Identifier v;
    Identifier sort = null;
    AstTerm t;
}
    :
        let:"let" v=ident ( COLON sort=ident )? EQUALS t=funcpredvarterm "in" r=asm_rule "end"
        {
            r = af.createAsmLet(createLocation(let), v, sort, t, r);
        }
    ;


protected
asm_rule_forall returns [AstAsmRule r]
{
    Identifier v;
    Identifier sort = null;
    AstTerm t;
}
    :
        forall:"forall" v=ident ( COLON sort=ident )? "with" t=formula "do" r=asm_rule "end"
        {
            r = af.createAsmForall(createLocation(forall), v, sort, t, r);
        }
    ;


protected
asm_rule_try returns [AstAsmRule r]
{
    AstAsmRule r1, r2;
}
    :
        try_:"try" r1=asm_rule "else" r2=asm_rule "end"
        {
            r = af.createAsmTry(createLocation(try_), r1, r2);
        }
    ;


protected
asm_rule_subst returns [AstAsmRule r]
{
    Identifier var;
    Identifier sort = null;
    AstTerm t;
}
    :
        lbrace:LBRACE var=ident ( COLON sort=ident )? t=term RBRACE r=asm_rule_atom
        {
            r = af.createAsmSubstitution(createLocation(lbrace), var, sort, t, r);
        }
    ;


protected
asm_rule_meta returns [AstAsmRule r]
{
    Identifier id;
}
    :
        at:AT id=ident LPAREN r=asm_rule RPAREN
        {
            r = af.createAsmRuleMeta(createLocation(at), id, r);
        }
    ;


protected
param_list returns [AstParameter[] l]
{
    AstParameter p;
    List list = createList();
}
    :
        p=param
        {
            list.add(p);
        }
        (
            COMMA p=param
            {
                list.add(p);
            }
        )*
        {
            l = (AstParameter[]) list.toArray(new AstParameter[list.size()]);
        }
    ;


protected
param returns [AstParameter p]
{
    Identifier id;
    Identifier sort;
}
    :   
        id=ident COLON sort=ident
        {
            p = new AstParameter(id.getLocation(), id, sort);
        }
    ;



proof returns [AstProofExpression pe]
    :
        PROOF LBRACE pe=proof_expr RBRACE
    ;


protected
proof_expr returns [AstProofExpression pe = null]
{
    AstProofExpression child;
    AstProofExpressionType type;
    List list = createList();
    String id = null;
}
    :
        paren:LPAREN
        (
            type=proof_expr_type
            (
                str:STRING_LITERAL
                {
                    id = str.getText();
                }
            )?
            (
                child=proof_expr
                {
                    list.add(child);
                }
            )*
            {
                AstProofExpression[] children = (AstProofExpression[]) list.toArray(new AstProofExpression[list.size()]);
                pe = new AstProofExpression(createLocation(paren), type, id, children);
            }
        )?
        RPAREN
    ;


/* Das ist aus der Parser-Spezifikation des Java-KeY-Systems. Ich
verstehe allerdings nicht, was der allgemeine (und ungenutzte) IDENT
hier soll. Er wird beispielsweise bei "opengoal" verwendet. */
protected
proof_expr_type returns [AstProofExpressionType type]
    :
        (
            FORMULA
            {
                type = AstProofExpressionType.FORMULA;
            }
        |
            TERM
            {
                type = AstProofExpressionType.TERM;
            }
        |
            RULE
            {
                type = AstProofExpressionType.RULE;
            }
        | 
            BRANCH
            {
                type = AstProofExpressionType.BRANCH;
            }
        |   
            INST
            {
                type = AstProofExpressionType.INST;
            }
        |
            IFSEQFORMULA
            { 
                type = AstProofExpressionType.IFSEQFORMULA;
            }
        |
            INTERACTIVE
            { 
                type = AstProofExpressionType.INTERACTIVE;
            }
        |
            HEUR
            {
                type = AstProofExpressionType.HEURISTICS;
            }
        |
            IDENT
            {
                type = AstProofExpressionType.IDENT;
            }
        )
    ;
