// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//

package de.uka.ilkd.asmkey.parser.env;


import java.util.*;

import de.uka.ilkd.asmkey.logic.AsmProcedure;
import de.uka.ilkd.key.logic.Named;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.NamespaceSet;
import de.uka.ilkd.key.rule.Heuristic;
import de.uka.ilkd.key.rule.Taclet;



/** Implementation for SortEnvironment, TermEnvironment and
 * DeclarationEnvironment.
 *
 * @author Hubert Schmid
 */

public class DefaultEnvironment implements DeclarationEnvironment {

    /** Creates an approtiate map and puts the elements (instanceof
     * Name) in it. */
    private static Map createMap(Collection c) {
        Map map = new HashMap();
        if (c != null) {
            for (Iterator i = c.iterator(); i.hasNext();) {
                Object element = i.next();
                //assert element instanceof Named;
                map.put(((Named) element).name().toString(), element);
            }
        }
        return map;
    }

    /** Returns the element with the name 'id' from the map. Throws an
     * EnvironmentException with the text name, if no such element
     * exists.
     *
     * @throws EnvironmentException if no such element exists.
     */
    private static Object get(String name, String id, Map map) throws EnvironmentException {
        if (map.containsKey(id)) {
            return map.get(id);
        } else {
            throw new EnvironmentException(name + " " + id + " not found.");
        }
    }

    /** Creates an unmodifiable collection containing all the elements
     * of the map. */
    private static Collection createCollection(Map map) {
        return Collections.unmodifiableCollection(new ArrayList(map.values()));
    }


    /** Meta operators. Mapping from {@link String} to {@link
     * MetaOperator}. */
    private final Map metaOps;

    /** Declared sorts. Mapping from {@link String} to {@link
     * Sort}. */
    private final Map sorts;

    /** Schema variables. Mapping from {@link String} to {@link
     * SchemaVariable}. */
    private final Map vars;

    /** Operations, functions and predications. Mapping from {@link
     * String} to {@link Operator}. */
    private final Map ops;

    /** Named asm rules (procedures). Mapping from {@link String} to
     * {@link AsmProcedure}. */
    private final Map procs;

    /** Declared heuristics. Mapping from {@link String} to {@link
     * Heuristic}. */
    private final Map heurs;

    /** Declared taclets. Mapping from {@link String} (taclet name) to
     * {@link Taclet}. */
    private final Map taclets;

    /** Abbreviations. */
    private final AbbreviationMap abbreviations;


    /** Create an empty environment. */
    public DefaultEnvironment() {
        this(null, null, null, null, null, null, null, null);
    }

    /** Create an empty environment with the given meta operators. */
    public DefaultEnvironment(Collection metaOps) {
        this(metaOps, null, null, null, null, null, null, null);
    }

    /** Creates an environment with the given meta operators, sorts,
     * vars, operators (functions, predicates), named asm rules
     * (procedures), heuristics, taclets and abbreviations. Each
     * argument may be null. */
    public DefaultEnvironment(Collection metaOps,
                              Collection sorts,
                              Collection vars,
                              Collection ops,
                              Collection procs,
                              Collection heurs,
                              Collection taclets,
                              AbbreviationMap abbreviations) {
        this.metaOps = createMap(metaOps);
        this.sorts = createMap(sorts);
        this.vars = createMap(vars);
        this.ops = createMap(ops);
        this.procs = createMap(procs);
        this.heurs = createMap(heurs);
        this.taclets = createMap(taclets);
        this.abbreviations = abbreviations;
    }


    public boolean containsMetaOp(String id) {
        return metaOps.containsKey(id);
    }

    public MetaOperator getMetaOp(String id) throws EnvironmentException {
        return (MetaOperator) get("Meta operator", id, metaOps);
    }

    public Sort getSort(String id) throws EnvironmentException {
        return (Sort) get("Sort", id, sorts);
    }

    public void addSort(Sort sort) throws EnvironmentException {
        String id = sort.name().toString();
        if (sorts.containsKey(id)) {
            throw new EnvironmentException("Sort " + id + " is already defined.");
        } else {
            sorts.put(id, sort);
        }
    }

    /** Returns all declared (and initial) sorts. */
    public Collection getSorts() {
        return createCollection(sorts);
    }

    public boolean containsSchemaVariable(String id) {
        return vars.containsKey(id);
    }

    public SchemaVariable getSchemaVariable(String id) throws EnvironmentException {
        return (SchemaVariable) get("Schema variable", id, vars);
    }

    /** Returns all declared (and initial) schema variables. */
    public Collection getSchemaVariables() {
        return createCollection(vars);
    }

    public void addSchemaVariable(SchemaVariable sv) throws EnvironmentException {
        String id = sv.name().toString();
        if (vars.containsKey(id)) {
            throw new EnvironmentException("SchemaVariable " + id + " is already defined.");
        } else {
            vars.put(id, sv);
        }
    }

    public Operator getOperator(String id) throws EnvironmentException {
        return (Operator) get("Operator", id, ops);
    }

    public void addOperator(Operator op) throws EnvironmentException {
        String id = op.name().toString();
        if (ops.containsKey(id)) {
            throw new EnvironmentException("Operator " + id + " is already defined.");
        } else {
            ops.put(id, op);
        }
    }

    /** Returns all declared (and initial) operators. */
    public Collection getOperators() {
        return createCollection(ops);
    }

    public AsmProcedure getProcedure(String id) throws EnvironmentException {
        return (AsmProcedure) get("Procedure", id, procs);
    }

    public void addProcedure(AsmProcedure p) throws EnvironmentException {
        String id = p.getId();
        if (procs.containsKey(id)) {
            throw new EnvironmentException("Procedure " + id + " is already defined.");
        } else {
            procs.put(id, p);
        }
    }

    /** Returns all declared procedures (named ASM rules). */
    public Collection getProcedures() {
        return createCollection(procs);
    }

    public Heuristic getHeuristic(String id) throws EnvironmentException {
        if (heurs.containsKey(id)) {
            return (Heuristic) heurs.get(id);
        } else {
            throw new EnvironmentException("Heuristic " + id + " not found.");
        }
    }

    public void addHeuristic(Heuristic h) throws EnvironmentException {
        String id = h.name().toString();
        if (heurs.containsKey(id)) {
            throw new EnvironmentException("Heuristic " + id + " is already defined.");
        } else {
            heurs.put(id, h);
        }
    }

    /** Returns all declared (and initial) heuristics. */
    public Collection getHeuristics() {
        return createCollection(heurs);
    }

    public void addTaclet(Taclet t) throws EnvironmentException {
        String id = t.name().toString();
        if (taclets.containsKey(id)) {
            throw new EnvironmentException("Taclet " + id + " is already defined.");
        } else {
            taclets.put(id, t);
        }
    }

    /** Returns all declared (and initial) taclets. */
    public Collection getTaclets() {
        return createCollection(taclets);
    }

    public Term getAbbreviatedTerm(String name) throws EnvironmentException {
        if (abbreviations.containsAbbreviation(name)) {
            return abbreviations.getTerm(name);
        } else {
            throw new EnvironmentException("abbreviation " + name + " not found.");
        }
    }
    

    /** Create a namespace from a collection a Named objects. The
     * elements of 'c' have to be of type {@link Named}. */
    private static Namespace toNamespace(Collection c) {
        Namespace ns = new Namespace();
        for (Iterator i = c.iterator(); i.hasNext();) {
            ns.add((Named) i.next());
        }
        return ns;
    }
    
    
    public NamespaceSet getEnvAsNamespaceSet() {
       return new NamespaceSet(
          toNamespace(getSchemaVariables()),
          toNamespace(getOperators()), // XXX: filter functions
          toNamespace(getSorts()),
          toNamespace(getHeuristics()));
    }

}
