// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//

package de.uka.ilkd.asmkey.parser.tree;


import java.util.*;

import de.uka.ilkd.asmkey.logic.*;
import de.uka.ilkd.asmkey.parser.TreeParser;
import de.uka.ilkd.asmkey.parser.ast.*;
import de.uka.ilkd.asmkey.parser.env.*;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.logic.sort.PrimitiveSort;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.parser.ParserException;
import de.uka.ilkd.key.rule.Heuristic;


/** Class to parse all kinds of declarations.
 *
 * @author Hubert Schmid
 */

public final class DeclarationParser implements AstDeclarationVisitor {

    /** Create and return an array of sorts. The i-th entry references
     * the sort identified by id[i]. */
    private static Sort[] getSorts(SortEnvironment env, Identifier[] id) throws ParserException {
        Sort[] sorts = new Sort[id.length];
        for (int i = 0; i < sorts.length; ++i) {
            sorts[i] = EnvironmentUtil.getSort(env, id[i]);
        }
        return sorts;
    }


    /** The environment that is used for parsing and to which new
     * declarations are added. */
    protected final DeclarationEnvironment env;

    /** Instance to parse terms. */
    private final TermParser termParser;

    /** Procedures (named asm rules) are handled special, because the
     * body of the procedure can use procedures, that are declared
     * later. First, the signatures of the procedures are
     * collected. Then the bodies of all procedures are parsed. The
     * map contains mappings from the identifier of the procedure to
     * the body ({@link AstAsmRule}). */
    private final Map procedures = new HashMap();

    /** Temporary stores the identifier of the declaration. The
     * simplifies access to this identifier. */
    private transient Identifier id;


    /** Create a declaration parser that uses the given declaration
     * environment. All new declarations are added to this
     * environment. */
    public DeclarationParser(DeclarationEnvironment env) {
        this.env = env;
        this.termParser = new TermParser(env);
    }


    /** Parse an array of declarations and add them to the
     * environment. */
    public void parse(AstDeclaration[] decls) throws ParserException {
        try {
            for (int i = 0; i < decls.length; ++i) {
                parse(decls[i]);
            }
            for (Iterator i = procedures.entrySet().iterator(); i.hasNext();) {
                try {
                    Map.Entry entry = (Map.Entry) i.next();
                    AsmProcedure p = env.getProcedure((String) entry.getKey());
                    CallEnvironment callEnv = new CallEnvironment(env, p.getFormalParameters());
                    p.setBody(TreeParser.parseAsmRule(callEnv, (AstAsmRule) entry.getValue()));
                } catch (EnvironmentException ee) {
                    throw new ParserException(ee.getMessage(), null /* XXX */
                                              );
                }
            }
            procedures.clear();
        } catch (ParserExceptionWrapper ew) {
            throw ew.getParserException();
        }
    }

    /** Parse a single declaration and add it the to environment. */
    protected void parse(AstDeclaration decl) throws ParserException {
        id = decl.getId();
        decl.accept(this);
        id = null;
    }

    /* --- Interface: DeclarationVisitor --- */

    public void visitSort() {
        try {
            env.addSort(new PrimitiveSort(new Name(id.getText())));
        } catch (EnvironmentException ee) {
            throw new ParserExceptionWrapper(new ParserException(ee.getMessage(), id.getLocation()));
        }
    }

    public void visitSchemaVariable(final AstSchemaType type) {
        SchemaVariableDeclarationParser.parse(env, id, type);
    }

    public void visitPredicate(Identifier[] args) {
        addFunction("Predicate", id, args, Sort.FORMULA, false);
    }

    public void visitFunction(Identifier[] args, Identifier ret, boolean dynamic) {
        try {
            Sort retSort = EnvironmentUtil.getSort(env, ret);
            addFunction("Function", id, args, retSort, dynamic);
        } catch (ParserException pe) {
            throw new ParserExceptionWrapper(pe);
        }
    }

    private void addFunction(String text, Identifier id, Identifier[] args, Sort retSort, boolean dynamic) {
        try {
            Sort[] sorts = args == null ? new Sort[] {}
                : getSorts(env, args);
            Name name = new Name(id.getText());
            Function f = dynamic ? new DynamicFunction(name, retSort, sorts) : new Function(name, retSort, sorts);
            env.addOperator(f);
        } catch (ParserException pe) {
            throw new ParserExceptionWrapper(pe);
        } catch (EnvironmentException ee) {
            throw new ParserExceptionWrapper(new ParserException(ee.getMessage(), id.getLocation()));
        }
    }

    public void visitHeuristic() {
        try {
            env.addHeuristic(new Heuristic(new Name(id.getText())));
        } catch (EnvironmentException ee) {
            throw new ParserExceptionWrapper(new ParserException(ee.getMessage(), id.getLocation()));
        }
    }

    public void visitTaclet(AstTaclet taclet) {
        TacletDeclarationParser.parse(env, id, taclet);
    }

    public void visitProcedure(AstParameter[] params, AstAsmRule rule) {
        try {
            FormalParameter[] fps = new FormalParameter[params.length];
            for (int i = 0; i < fps.length; ++i) {
                fps[i] =
                    new FormalParameter(params[i].getId().getText(), EnvironmentUtil.getSort(env, params[i].getSort()));
            }
            AsmProcedure p = new AsmProcedure(id.getText(), fps);
            env.addProcedure(p);
            procedures.put(id.getText(), rule);
        } catch (ParserException pe) {
            throw new ParserExceptionWrapper(pe);
        } catch (EnvironmentException ee) {
            throw new ParserExceptionWrapper(new ParserException(ee.getMessage(), id.getLocation()));
        }
    }

}
