// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//

package de.uka.ilkd.asmkey.parser.tree;


import de.uka.ilkd.asmkey.logic.TermUtil;
import de.uka.ilkd.asmkey.parser.ast.*;
import de.uka.ilkd.asmkey.parser.ast.Operator;
import de.uka.ilkd.asmkey.parser.ast.Quantifier;
import de.uka.ilkd.asmkey.parser.env.EnvironmentUtil;
import de.uka.ilkd.asmkey.parser.env.TermEnvironment;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.parser.ParserException;


/** Helper class to create new terms and convert operators and
 * quantifiers.
 *
 * @author Hubert Schmid
 */

final class TermParserUtil {

    /** Create a nullary term with the given operator. */
    protected static Term createTerm(Operator2 op) {
        return op.createTerm(null, TermUtil.EMPTY_TERMS);
    }

    /** Create an unary term with the given operator. */
    protected static Term createTerm(Operator2 op, Term term) {
        return op.createTerm(null, new Term[] { term });
    }

    /** Create a binary term with the given operator. */
    protected static Term createTerm(Operator2 op, Term term1, Term term2) {
        return op.createTerm(null, new Term[] { term1, term2 });
    }

    /** Create a ternary term with the given operator. */
    protected static Term createTerm(Operator2 op, Term term1, Term term2, Term term3) {
        return op.createTerm(null, new Term[] { term1, term2, term3 });
    }

    /** Create a term with a bound variable. */
    protected static Term createTerm(Operator2 op,
                                     QuantifiableVariable v,
                                     Term term1,
                                     Term term2) {
        return op.createTerm(new ArrayOfQuantifiableVariable(new QuantifiableVariable[] { v }),
                             new Term[] { term1, term2 });
    }

    /** Converts an AST operator into a logic operator. */
    protected static de.uka.ilkd.key.logic.op.Operator convertOp(TermEnvironment env,
                                                                 AstOperator token)
        throws ParserException
    {
        Operator op = token.getOperator();
        if (op == Operator.IMP) {
            return Op.IMP;
        } else if (op == Operator.OR) {
            return Op.OR;
        } else if (op == Operator.AND) {
            return Op.AND;
        } else if (op == Operator.NOT) {
            return Op.NOT;
        } else if (op == Operator.TRUE) {
            return Op.TRUE;
        } else if (op == Operator.FALSE) {
            return Op.FALSE;
        } else if (op == Operator.ADD) {
            return EnvironmentUtil.getOperator(env, new Identifier(token.getLocation(), "+"));
        } else if (op == Operator.SUB) {
            return EnvironmentUtil.getOperator(env, new Identifier(token.getLocation(), "~d"));
        } else if (op == Operator.MULT) {
            return EnvironmentUtil.getOperator(env, new Identifier(token.getLocation(), "mul"));
        } else if (op == Operator.DIV) {
            return EnvironmentUtil.getOperator(env, new Identifier(token.getLocation(), "div"));
        } else if (op == Operator.NEG) {
            return EnvironmentUtil.getOperator(env, new Identifier(token.getLocation(), "~m"));
        } else {
            throw new ParserException("Unknown operator " + op + ".", token.getLocation());
        }
    }

    /** Converts an AST quantifier into a logic quantifier. */
    protected static de.uka.ilkd.key.logic.op.Quantifier convertQuantifier(AstQuantifier token)
        throws ParserException
    {
        Quantifier q = token.getQuantifier();
        if (q == Quantifier.ALL) {
            return Op.ALL;
        } else if (q == Quantifier.EX) {
            return Op.EX;
        } else {
            throw new ParserException("Unknown quantifier " + q + ".", token.getLocation());
        }
    }

}
