// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//

package de.uka.ilkd.asmkey.pp;


import java.io.IOException;

import de.uka.ilkd.key.gui.Notation;
import de.uka.ilkd.key.gui.SequentPrinter;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.op.QuantifiableVariable;
import de.uka.ilkd.key.util.pp.UnbalancedBlocksException;


/** This class contains the {@link de.uka.ilkd.key.gui.Notation}s for ASM terms and ASM rules. */

public final class AsmNotation {

        /** Notation for modality BOX. */
        public static Notation box() {
                return new Notation(60) {
                        public void print(Term term, SequentPrinter sp) throws IOException {
                                AsmPrinter.printAsmModalityTerm(sp, "[", "]", term, 60);
                        }
                };
        }

        /** Notation for modality DIAMOND. */
        public static Notation diamond() {
                return new Notation(60) {
                        public void print(Term term, SequentPrinter sp) throws IOException {
                                AsmPrinter.printAsmModalityTerm(sp, "<", ">", term, 60);
                        }
                };
        }

        /** Notation for modality DEF. */
        public static Notation def() {
                return new Notation(100) {
                        public void print(Term term, SequentPrinter sp)
                                throws UnbalancedBlocksException, IOException {
                                AsmPrinter.printAsmUnary(sp, "def(", ")", term);
                        }
                };
        }

        /** Notation for modality UPD. */
        public static Notation upd() {
                return new Notation(0) {
                        public void print(Term term, SequentPrinter sp)
                                throws UnbalancedBlocksException, IOException {
                                AsmPrinter.printAsmUpd(sp, term);
                        }
                };
        }

        /** Notation for modality CON. */
        public static Notation Con() {
                return new Notation(100) {
                        public void print(Term term, SequentPrinter sp)
                                throws UnbalancedBlocksException, IOException {
                                AsmPrinter.printAsmUnary(sp, "Con(", ")", term);
                        }
                };
        }

        /** Notation for modality INV. */
        public static Notation inv() {
                return new Notation(0) {
                        public void print(Term term, SequentPrinter sp)
                                throws UnbalancedBlocksException, IOException {
                                AsmPrinter.printAsmInv(sp, term);
                        }

                };
        }

        /** Notation for modality JOINABLE. */
        public static Notation joinable() {
                return new Notation(100) {
                        public void print(Term term, SequentPrinter sp)
                                throws UnbalancedBlocksException, IOException {
                                AsmPrinter.printAsmJoinable(sp, term);
                        }
                };
        }

        /** Notation for modality EQUI. */
        public static Notation equi() {
                return new Notation(100) {
                        public void print(Term term, SequentPrinter sp)
                                throws UnbalancedBlocksException, IOException {
                                AsmPrinter.printAsmEqui(sp, term, false);
                        }
                };
        }

        /** Notation for modality EQUI. */
        public static Notation equiS() {
                return new Notation(100) {
                        public void print(Term term, SequentPrinter sp)
                                throws UnbalancedBlocksException, IOException {
                                AsmPrinter.printAsmEqui(sp, term, true);
                        }
                };
        }

        /** Notation for sequential ASM program composition. */
        public static Notation seq() {
                return new Notation(0) {
                        public void print(Term term, SequentPrinter sp)
                                throws UnbalancedBlocksException, IOException {
                                AsmPrinter.printAsmSeq(sp, ";", term);
                        }
                };
        }

        /** Notation for parallel ASM program composition. */
        public static Notation par() {
                return new Notation(0) {
                        public void print(Term term, SequentPrinter sp)
                                throws UnbalancedBlocksException, IOException {
                                AsmPrinter.printAsmSeq(sp, ",", term);
                        }
                };
        }

        /** Notation for ASM program SKIP. */
        public static Notation skip() {
                return new Notation.Constant("skip", 80);
        }

        /** Notation for ASM assignment rule. */
        public static Notation assign() {
                return new Notation.Infix(":=", 50, 60, 60);
        }

        /** Notation for ASM program construct IF-THEN-ELSE. */
        public static Notation if_() {
                return new Notation(0) {
                        public void print(Term term, SequentPrinter sp)
                                throws UnbalancedBlocksException, IOException {
                                AsmPrinter.printAsmIf(sp, term);
                        }

                };
        }

        /** Notation for ASM program construct LET-IN. */
        public static Notation let() {
                return new Notation(0) {
                        public void print(Term term, SequentPrinter sp)
                                throws UnbalancedBlocksException, IOException {
                                QuantifiableVariable v = term.varsBoundHere(1).getQuantifiableVariable(0);
                                AsmPrinter.printAsmLet(sp, v.toString(), term);
                        }

                };
        }

        /** Notation for ASM program construct FORALL-WITH-DO. */
        public static Notation forall() {
                return new Notation(0) {
                        public void print(Term term, SequentPrinter sp)
                                throws UnbalancedBlocksException, IOException {
                                QuantifiableVariable v = term.varsBoundHere(1).getQuantifiableVariable(0);
                                AsmPrinter.printAsmForAll(sp, v.toString(), term);
                        }

                };
        }

        /** Notation for ASM program construct TRY-ELSE. */
        public static Notation try_() {
                return new Notation(0) {
                        public void print(Term term, SequentPrinter sp)
                                throws UnbalancedBlocksException, IOException {
                                AsmPrinter.printAsmTry(sp, term);
                        }

                };
        }

        /** Notation for ASM named rules calls. */
        public static Notation named() {
                return new Notation(100) {
                        public void print(Term term, SequentPrinter sp)
                                throws UnbalancedBlocksException, IOException {
                                AsmPrinter.printAsmNamedRule(sp, term);
                        }

                };
        }

}
