// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//

package de.uka.ilkd.asmkey.pp;

import java.io.IOException;

import de.uka.ilkd.key.gui.AbstractAsmPrinter;
import de.uka.ilkd.key.gui.SequentPrinter;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.util.pp.UnbalancedBlocksException;


/** SequentPrinter for ASM terms and ASM rules. This class contains the actual
 * code to print the ASM terms and rules. The methods are called from the ASM notations
 * ({@link de.uka.ilkd.asmkey.pp.AsmNotation}).
 *
 * @author Hubert Schmid */

final class AsmPrinter extends AbstractAsmPrinter {

        /** Indentation level for if, try and joinable. */
        private static final int INDENT = 2;

        /** Used for DL modalities BOX and DIAMOND. */
        protected static void printAsmModalityTerm(
                SequentPrinter sp,
                String left,
                String right,
                Term phi,
                int ass)
                throws IOException {
                startTerm(sp, 2);
                layouter(sp).beginC(0).print(left);
                markStartSub(sp);
                sp.printTerm(phi.sub(0));
                markEndSub(sp);
                layouter(sp).print(right).brk(1, 2);
                maybeParens(sp, phi.sub(1), ass);
                layouter(sp).end();
        }

        /** Used for modalities Con() and def(). */
        protected static void printAsmUnary(SequentPrinter sp, String left, String right, Term phi)
                throws UnbalancedBlocksException, IOException {
                startTerm(sp, 1);
                layouter(sp).beginC(4).print(left);
                markStartSub(sp);
                sp.printTerm(phi.sub(0));
                markEndSub(sp);
                layouter(sp).print(right).end();
        }

        /** Used for modality upd. */
        protected static void printAsmUpd(SequentPrinter sp, Term phi)
                throws UnbalancedBlocksException, IOException {
                startTerm(sp, 3);
                layouter(sp).beginC(4).print("upd(");
                markStartSub(sp);
                sp.printTerm(phi.sub(0));
                markEndSub(sp);
                layouter(sp).print(" <- ");
                markStartSub(sp);
                sp.printTerm(phi.sub(1));
                markEndSub(sp);
                layouter(sp).print(" in").brk();
                markStartSub(sp);
                sp.printTerm(phi.sub(2));
                markEndSub(sp);
                layouter(sp).print(")").end();
        }

        /** Used for modality inv. */
        protected static void printAsmInv(SequentPrinter sp, Term phi)
                throws UnbalancedBlocksException, IOException {
                startTerm(sp, 2);
                layouter(sp).print("inv(");
                markStartSub(sp);
                sp.printTerm(phi.sub(0));
                markEndSub(sp);
                layouter(sp).print(" in ");
                markStartSub(sp);
                sp.printTerm(phi.sub(1));
                markEndSub(sp);
                layouter(sp).print(")");
        }

        /** Used for ASM modality joinable. */
        protected static void printAsmJoinable(SequentPrinter sp, Term phi)
                throws UnbalancedBlocksException, IOException {
                startTerm(sp, 2);
                layouter(sp).beginC(AsmPrinter.INDENT).print("joinable(").brk(0);
                markStartSub(sp);
                sp.printTerm(phi.sub(0));
                markEndSub(sp);
                layouter(sp).brk(1, -2).print("and").brk(1, 2);
                markStartSub(sp);
                sp.printTerm(phi.sub(1));
                markEndSub(sp);
                layouter(sp).brk(0, -2).print(")").end();
        }

        /** Used for ASM modality equi and equiS. */
        protected static void printAsmEqui(SequentPrinter sp, Term phi, boolean strong)
                throws UnbalancedBlocksException, IOException {
                String name = strong ? "equiS" : "equi";
                startTerm(sp, 2);
                layouter(sp).beginC(AsmPrinter.INDENT).print(name).print("(").brk(0);
                markStartSub(sp);
                sp.printTerm(phi.sub(0));
                markEndSub(sp);
                layouter(sp).brk(1, -2).print("and").brk(1, 2);
                markStartSub(sp);
                sp.printTerm(phi.sub(1));
                markEndSub(sp);
                layouter(sp).brk(0, -2).print(")").end();
        }

        /** Used for sequential and parallel program composition. */
        protected static void printAsmSeq(SequentPrinter sp, String separator, Term phi)
                throws UnbalancedBlocksException, IOException {
                startTerm(sp, 2);
                layouter(sp).beginC(0);
                markStartSub(sp);
                sp.printTerm(phi.sub(0));
                markEndSub(sp);
                layouter(sp).print(separator).brk();
                markStartSub(sp);
                sp.printTerm(phi.sub(1));
                markEndSub(sp);
                layouter(sp).end();
        }

        /** Used for ASM if program construct. */
        protected static void printAsmIf(SequentPrinter sp, Term phi)
                throws UnbalancedBlocksException, IOException {
                startTerm(sp, 3);
                layouter(sp).beginC(AsmPrinter.INDENT).print("if ");
                markStartSub(sp);
                sp.printTerm(phi.sub(0));
                markEndSub(sp);
                layouter(sp).print(" then").brk();
                markStartSub(sp);
                sp.printTerm(phi.sub(1));
                markEndSub(sp);
                layouter(sp).brk(1, -AsmPrinter.INDENT).print("else").brk();
                markStartSub(sp);
                sp.printTerm(phi.sub(2));
                markEndSub(sp);
                layouter(sp).brk(1, -AsmPrinter.INDENT).print("end").end();
        }

        /** Used for ASM try program construct. */
        protected static void printAsmTry(SequentPrinter sp, Term phi)
                throws UnbalancedBlocksException, IOException {
                startTerm(sp, 3);
                layouter(sp).beginC(AsmPrinter.INDENT).print("try").brk();
                markStartSub(sp);
                sp.printTerm(phi.sub(0));
                markEndSub(sp);
                layouter(sp).brk(1, -AsmPrinter.INDENT).print("else").brk();
                markStartSub(sp);
                sp.printTerm(phi.sub(1));
                markEndSub(sp);
                layouter(sp).brk(1, -AsmPrinter.INDENT).print("end").end();
        }

        /** Used for ASM let program construct. */
        protected static void printAsmLet(SequentPrinter sp, String var, Term phi)
                throws UnbalancedBlocksException, IOException {
                startTerm(sp, 2);
                layouter(sp).beginC(2).print("let " + var + " = ");
                markStartSub(sp);
                sp.printTerm(phi.sub(0));
                markEndSub(sp);
                layouter(sp).print(" in").brk();
                markStartSub(sp);
                sp.printTerm(phi.sub(1));
                markEndSub(sp);
                layouter(sp).brk().print("end").end();
        }

        /** Used for ASM forall program construct. */
        protected static void printAsmForAll(SequentPrinter sp, String var, Term phi)
                throws UnbalancedBlocksException, IOException {
                startTerm(sp, 2);
                layouter(sp).beginC(2).print("forall " + var + " with ");
                markStartSub(sp);
                sp.printTerm(phi.sub(0));
                markEndSub(sp);
                layouter(sp).print(" do").brk();
                markStartSub(sp);
                sp.printTerm(phi.sub(1));
                markEndSub(sp);
                layouter(sp).brk().print("end").end();
        }

        /** Print ASM named rule calls. */
        protected static void printAsmNamedRule(SequentPrinter sp, Term t)
        throws UnbalancedBlocksException, IOException {
                startTerm(sp, t.arity());
                layouter(sp).print("call ").print(t.op().name().toString());
                if (t.arity() > 0) {
                        layouter(sp).print("(").beginC(0);
                        for (int i = 0; i < t.arity(); ++i) {
                                markStartSub(sp);
                                sp.printTerm(t.sub(i));
                                markEndSub(sp);

                                if (i < t.arity() - 1) {
                                        layouter(sp).print(",").brk(1, 0);
                                }
                        }
                        layouter(sp).print(")").end();
                }
        }

}
