// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//

package de.uka.ilkd.asmkey.rule;


import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.rule.VariableCondition;
import de.uka.ilkd.key.rule.inst.*;


/** Taclet condition that checks if two terms have the same top level
 * operator.
 *
 * @author Hubert Schmid
 */

public final class OperatorEqualsCondition implements VariableCondition {

    /** Returns the corresponding term for the given schema variable
     * from the instantiations. */
    private static Term getInstantiation(SVInstantiations instantiations, SchemaVariable v) {
        InstantiationEntry entry = instantiations.getInstantiationEntry(v);
        if (entry == null) {
            return null;
        } else if (entry instanceof TermInstantiation) {
            return ((TermInstantiation) entry).getTerm();
        } else {
            throw new RuntimeException("instantiation for " + v + " is not valid. (" + entry + ")");
        }
    }


    /** this condition returns 'true' iff (the top level operators
     * equals iff 'equal') */
    private final boolean equal;

    /** first schema variable that is checked */
    private final SchemaVariable x;

    /** second schema variable that is checked */
    private final SchemaVariable y;


    /** Creates condition for the given schema variables. If 'equal ==
     * true', the operators must match, else the operators must
     * differ. */
    public OperatorEqualsCondition(boolean equal, SchemaVariable x, SchemaVariable y) {
        this.equal = equal;
        this.x = x;
        this.y = y;
    }


    /** Check if the instantiations for schema variables x and y have
     * the same top level operator. */
    public boolean check(SchemaVariable var, Term term, SVInstantiations svInst) {
        /* This method is called each time a schema variable is
         * instantiated. We have to check, if the schema variable
         * matches 'x' or 'y' and the other schema variable is already
         * instantiated. Else we return 'true' and check the condition
         * on a later invocation of this method. */
        if (var == x) {
            Term ty = getInstantiation(svInst, y);
            if (ty == null) {
                /* The condition can only be checked, if both
                 * variables are instantiated. */
                return true;
            } else {
                return (term.op() == ty.op()) == equal;
            }
        } else if (var == y) {
            Term tx = getInstantiation(svInst, x);
            if (tx == null) {
                /* The condition can only be checked, if both
                 * variables are instantiated. */
                return true;
            } else {
                return (term.op() == tx.op()) == equal;
            }
        } else {
            return true;
        }
    }

}
