// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/*
 * TogetherReprModelClass provides easy access to the
 * properties of a class in a Together model.
 *
 * Bugs: - Each class is assumed to have only one parent
 *        (interfaces are not considered yet, because the OCL
 *        parser does not work with interfaces correctly)
 *      - Only parents in the same packages are considered,
 *        because OCL parser does not work with items from
 *         different packages correctly.
 */

package de.uka.ilkd.key.casetool.together;

import java.util.Vector;
import java.io.File;

import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.casetool.ReprModelClass;
import de.uka.ilkd.key.casetool.HashMapOfAssociations;
import de.uka.ilkd.key.casetool.HashMapOfClassifier;
import de.uka.ilkd.key.casetool.UMLOCLClassifier;
import de.uka.ilkd.key.casetool.UMLOCLMultiplicity;
import de.uka.ilkd.key.casetool.UMLOCLAssociation;

import com.togethersoft.openapi.rwi.RwiNode;
import com.togethersoft.openapi.rwi.RwiModel;
import com.togethersoft.openapi.rwi.RwiDiagram;
import com.togethersoft.openapi.rwi.RwiProperty;
import com.togethersoft.openapi.rwi.RwiPropertyMap;
import com.togethersoft.openapi.rwi.RwiNodeReference;
import com.togethersoft.openapi.rwi.RwiPackage;
import com.togethersoft.openapi.rwi.RwiReference;
import com.togethersoft.openapi.rwi.RwiShapeType;
import com.togethersoft.openapi.rwi.RwiElement;
import com.togethersoft.openapi.rwi.RwiMember;
import com.togethersoft.openapi.rwi.RwiLink;
import com.togethersoft.openapi.rwi.RwiContainer;
import com.togethersoft.openapi.sci.SciElement;
import com.togethersoft.openapi.sci.SciModel;
import com.togethersoft.openapi.sci.SciModelAccess;
import com.togethersoft.openapi.rwi.enum.RwiPropertyEnumeration;
import com.togethersoft.openapi.rwi.enum.RwiMemberEnumeration;
import com.togethersoft.openapi.rwi.enum.RwiNodeEnumeration;
import com.togethersoft.openapi.rwi.enum.RwiContainerEnumeration;
import com.togethersoft.openapi.rwi.enum.RwiPackageEnumeration;
import com.togethersoft.openapi.rwi.enum.RwiLinkEnumeration;
import com.togethersoft.openapi.ide.project.*;

public class TogetherReprModelClass extends TogetherReprModel 
    implements ReprModelClass{

    private static final String CLASS="Class";

    private RwiNode orig;
    private RwiModel currRwiModel;
    private RwiDiagram currRwiDiagram;

    private RwiNode origParent;
    private boolean origParentAlreadySet = false;

    public TogetherReprModelClass(RwiNode theOrig, RwiModel model, 
				  RwiDiagram diag) {
	orig = theOrig;
	currRwiModel = model;
	currRwiDiagram = diag;
    }


    public String getClassName(){
	return orig.getProperty(RwiProperty.NAME);
    }

    public String getContainingPackage() {
	String fullName = getFullClassName();
	if (fullName.lastIndexOf(".")==-1) {
	    return null;
	} else {
	    return fullName.substring
		(0,fullName.lastIndexOf("."));
	}
    }

    public String getFullClassName(){
	return orig.getProperty(RwiProperty.FULL_NAME);
    }

    // looks if orig has a parent and set it if needed
    private RwiNode getOrigParent(){
	if (origParentAlreadySet){
	    return origParent;
	}else{
	    if (orig.getProperty(RwiProperty.EXTENDS)==null){
		origParent = null;
	    }else{
		RwiPropertyEnumeration extended 
		    = orig.properties(RwiProperty.EXTENDS);
		while (extended.hasMoreElements()){
		    RwiProperty nextExtended=extended.nextRwiProperty();
		    RwiPropertyMap subpropertiesContainer 
			= nextExtended.getSubproperties();
		    RwiElement foundClassOrInterface 
			= currRwiModel.findElement
			(subpropertiesContainer.getProperty
			 (RwiProperty.REFERENCED_ELEMENT));
                    if (foundClassOrInterface != null 
			&& foundClassOrInterface.getProperty
			(RwiProperty.SHAPE_TYPE).equals(RwiShapeType.CLASS)){
			// we have found from the parents the 
			// (unique iff it exists)
			// element which is a class
			RwiReference foundRwiReference 
			    = currRwiDiagram.findReference
			    (foundClassOrInterface);
                        if (foundRwiReference == null){
			    // there is a superclass but not in the 
			    // current diagram
                            origParentAlreadySet = true;
                            return null;
                        }
                        RwiNodeReference foundRwiNodeReference 
			    = (RwiNodeReference) foundRwiReference;
                        if (foundRwiNodeReference == null){
                            System.err.println("KeYError: foundRwiReference not"
					       +" of type RwiNodeReference");
                            origParentAlreadySet = true;
                            return null;
                        }
                        origParent = foundRwiNodeReference.getNode();
		    }
		} //while
	    }
	    origParentAlreadySet = true;
	}
	return origParent;
    }


    // Now, a service for ReprModelMeth

    // @param aUniqueMethName is a name getting from aRwiMember.getUniqueName()
    // @return the method declaration of param in the parents of the current
    //         class
    public RwiMember  findParentMeth(String aUniqueMethName){
	String origClassName = getFullClassName();
	String candMethName = substituteCNinUniqueMethName(aUniqueMethName, 
							   origClassName);
	if (candMethName.equals(aUniqueMethName)){
	    // a Method of aUniqueMethName exists within orig
	    // So, we have to ask the parent
	    if (getOrigParent() == null){
		return null;
	    }else{
		TogetherReprModelClass parentReprClass 
		    = new TogetherReprModelClass(getOrigParent(), 
						 currRwiModel, 
						 currRwiDiagram);
		return (parentReprClass.findParentMeth(aUniqueMethName));
	    }
	}else{
	    // we have to look if candMethName really exists in the model
	    // If not, we ask our parent
	    RwiMember returnvalue = (RwiMember) currRwiModel.findElement
		(candMethName);
	    if (returnvalue == null){
		if (getOrigParent() != null){
		    TogetherReprModelClass parentReprClass 
			= new TogetherReprModelClass(getOrigParent(), 
						     currRwiModel, 
						     currRwiDiagram);
		    returnvalue 
			= parentReprClass.findParentMeth(aUniqueMethName);
		}
	    }
	    return returnvalue;
	}
    }
		
	    
	    
	                
	    
    private String substituteCNinUniqueMethName(String uName, 
						String newClassName){
	// region between 2nd and 3rd '#' is substituted by newClassName
	int firstOcc = uName.indexOf('#', 0);
	int secondOcc = uName.indexOf('#', firstOcc + 1);
	int thirdOcc = uName.indexOf('#', secondOcc +1 );
	String beginning = uName.substring(0, secondOcc + 1);
	String ending = uName.substring(thirdOcc, uName.length() );
	return beginning + newClassName + ending;
    }




    public String getMyInv() {
	if (orig.getProperty("invariants")==null){
	    return "";
	}else{
	    return orig.getProperty("invariants");
	}
    }
    
    // @author Kristofer Johannisson
    public void setMyInv(String inv) {
	orig.setProperty("invariants", inv);
    }
	
    // @author Kristofer Johannisson
    public void setMyInvGFAbs(String inv) {
	orig.setProperty("invariantGFAbstractSyntax", inv);
    }
    
    // @author Kristofer Johannisson
    public String getMyInvGFAbs() {
	String result = orig.getProperty("invariantGFAbstractSyntax");
	if (result == null)
	    return "";
	else
	    return result;
    }
    

    // @return the invariant of the parent
    public String getParentInv() { 
	if (getOrigParent() != null){
	    ReprModelClass parentRepr 
		= new TogetherReprModelClass(getOrigParent(), 
					     currRwiModel, 
					     currRwiDiagram);
		return parentRepr.getMyInv();
	}
	else{
	    return "";
	}
    }

    // @return the classname of the parent
    public String getParentClassName() {
		if (getOrigParent() != null){
	    	ReprModelClass parentRepr 
		    = new TogetherReprModelClass(getOrigParent(), 
						 currRwiModel, 
						 currRwiDiagram);
			return parentRepr.getClassName();
		}
		else{
	    	return "";
		}
    }
     public boolean hasOrigParent() {
        return (getOrigParent() != null);
    }

    // @return a vector of String
    public Vector getParentOpNames() {
	if (getOrigParent() != null){
	    ReprModelClass parentRepr 
		= new TogetherReprModelClass(getOrigParent(), 
					     currRwiModel, 
					     currRwiDiagram);
		return parentRepr.getOpNames();
	}
	else{
	    return null;
	}

     }

    // @return a vector of ReprModelMethod
    public Vector getParentOps() {
	if (getOrigParent() != null){
	    ReprModelClass parentRepr 
		= new TogetherReprModelClass(getOrigParent(), 
					     currRwiModel, 
					     currRwiDiagram);
		return parentRepr.getOps();
	}
	else{
	    return null;
	}

     }

     // @return a vector of String
     public Vector getOpNames() {
        RwiMemberEnumeration allMembers = orig.members();
        Vector opMemberNames = new Vector();
        while (allMembers.hasMoreElements()){
            RwiMember nextMember = allMembers.nextRwiMember();
            if(RwiShapeType.OPERATION.equals(nextMember.getProperty
					     (RwiProperty.SHAPE_TYPE))){
                 opMemberNames.add(nextMember.getProperty(RwiProperty.NAME));
            }
        }
        return opMemberNames;
     }


     // @return a vector of ReprModelMethod
     public Vector getOps() {
        RwiMemberEnumeration allMembers = orig.members();
        Vector opVector = new Vector();
        while (allMembers.hasMoreElements()){
            RwiMember nextMember = allMembers.nextRwiMember();
            if(RwiShapeType.OPERATION.equals(nextMember.getProperty
					     (RwiProperty.SHAPE_TYPE))){
                 opVector.add(new TogetherReprModelMethod(nextMember, 
							  currRwiModel, 
							  currRwiDiagram));
            }
        }
        return opVector;
     }

    public void getAssociations(HashMapOfClassifier 
						 classifiers) {

	HashMapOfAssociations setOfAssociations = 
	    new HashMapOfAssociations();
	SciModel model = SciModelAccess.getModel();
	RwiLinkEnumeration linkEnum = orig.outgoingLinks();
	RwiLink link;
	UMLOCLMultiplicity sourceMult = null, targetMult = null;
	String sourceRole = null, targetRole = null;
	RwiMemberEnumeration rwiMemberEnum;
	RwiMember rwiMember;
	while (linkEnum.hasMoreElements()){
	    link = linkEnum.nextRwiLink();
	    //filter out associations
	    if (link.getProperty(RwiProperty.SHAPE_TYPE).
		equals(RwiShapeType.ASSOCIATION)){
		// get cardinality and client role
		targetMult = null;
		targetRole = null;
		rwiMemberEnum = orig.members();
		rwiMember = null;
		while (rwiMemberEnum.hasMoreElements()){
		    rwiMember = rwiMemberEnum.nextRwiMember();
		    if (rwiMember.getUniqueName().indexOf(link.toString())!=-1){
			SciElement e = model.findElement(rwiMember.getUniqueName());
			String card = model.findElement
			    (rwiMember.getUniqueName()).getTagList().
			    getTagValue("supplierCardinality");
			targetMult=UMLOCLMultiplicity.STANDARD;
			if (card!=null)
			    targetMult= UMLOCLMultiplicity.
				getMultiplicityFromString(card);
			
			targetRole = getAssociationRole
			    (rwiMember, link, "supplierRole");
			
			break;
		    }		    
		}

		// associations are bi-directional. So we have to add the 
		// association also to the target classifier 

		sourceMult=null;
		sourceRole=null;
		
		String card = model.findElement
		    (rwiMember.getUniqueName()).getTagList().
		    getTagValue("clientCardinality");
		sourceMult=UMLOCLMultiplicity.STANDARD;
		if (card!=null)		    
		    sourceMult = UMLOCLMultiplicity.
			getMultiplicityFromString(card);
		
		sourceRole = getAssociationRole(rwiMember, link, "clientRole");

		UMLOCLClassifier sourceClassifier = 
		    classifiers.getClassifier(orig.toString());
		UMLOCLClassifier targetClassifier = 
		    classifiers.getClassifier(link.getDestination().toString());
		if (sourceClassifier != null && targetClassifier != null) {
		    sourceClassifier.addAssociation
			(targetRole, new UMLOCLTogetherAssociation
			(sourceClassifier, sourceRole, sourceMult, 
			 targetClassifier, targetRole, targetMult));
		    targetClassifier.addAssociation
			(sourceRole, new UMLOCLTogetherAssociation
			(targetClassifier, targetRole, targetMult,
			 sourceClassifier, sourceRole, sourceMult));

		}
	    }
	}
    }



    private String getAssociationRole(RwiElement rwiMember, 
				      RwiLink link, String which) {
	SciModel model = SciModelAccess.getModel();
	String role;
	role = model.findElement
	    (rwiMember.getUniqueName()).getTagList().
	    getTagValue(which);
	if (role==null){
	    if ("clientRole".equals(which))
		role=link.getSource().toString();
	    else if ("supplierRole".equals(which))
		role=link.getDestination().toString();
	    else 
		return null;
	    role=role.substring(0,1).toLowerCase()+
		role.substring(1,role.length());
	}
	return role;
    }

    public String getText() {
	return orig.getProperty(RwiProperty.TEXT);
    }

    private Vector getAllJavaFiles(RwiPackage pac, Vector v) {
	RwiPackageEnumeration subpackages=pac.subpackages();
	while (subpackages.hasMoreElements()) {
	    v=getAllJavaFiles(subpackages.nextRwiPackage(), v);
	}
	RwiNodeEnumeration enum=pac.nodes();
	while (enum.hasMoreElements()) {
	    RwiElement rwin=enum.nextRwiElement();
	    if (rwin.getProperty(RwiProperty.SHAPE_TYPE).equals(CLASS)){
		String filename=rwin.getProperty(RwiProperty.FILE);
		if (filename!=null && filename.endsWith(".java")) {
		    v.add(filename);
		    Debug.out("Let recoder read file: "+filename);
		}	    
	    }
	}
	return v;
    }

    public String[] getClassesInPackage() {	

	Vector v=new Vector();	
	RwiPackage rootPac=orig.getContainingPackage();	
	while (rootPac.getContainingPackage()!=null) {
	    rootPac=rootPac.getContainingPackage();
	}
	v=getAllJavaFiles(rootPac, v);
	String[] strings = new String[v.size()];
	for (int i=0; i<strings.length; i++) {
	    strings[i]=(String)v.elementAt(i);
	}
	return strings;
    }

    public String getRootDirectory() {
	IdeProjectManager pm=IdeProjectManagerAccess.getProjectManager();
	IdeProject prj=pm.getActiveProject();
	File f = new File(prj.getFileName());
	return f.getParentFile().getAbsolutePath();
    }

}
