// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//




/*

Uses code by Hans Muller and Kathy Walrath from
http://java.sun.com/products/jfc/tsc/articles/threads/threads2.html

*/


package de.uka.ilkd.key.gui;

import java.util.HashSet;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.proof.incclosure.Sink;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.util.Debug;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.rule.strategy.*;
import de.uka.ilkd.key.rule.strategy.proofevent.*;

/**
 * Applies heuristic Taclets in an own thread.
 */
public class ApplyHeuristics {

    private TacletApp lastApp=null;
    private InteractiveProver iaProver;
    /** the proof that is worked with */ 
    private Proof proof;
    /** the maximum of allowed heuristics applications */
    private int maxApplications;
    /** the java information object */
    private JavaInfo javaInfo;
    /** list of goals on which the heuristics should be applied */
    private ListOfGoal goalList;
    /** part of goalList that should be worked on next */
    private ListOfGoal nextGoals;
    /** set of goals where no rule is applicable at the current run */
    //    private HashSet nonApplicableGoals;
    
    /** Subset of the goals to which currently taclets are applied.
     * First this is the list of goals with unsatisfiable constraints,
     * later this is a subset of the goals having inconsistent
     * constraints */
    private ListOfGoal selectedList;

    /** true iff all goals have satisfiable constraints */
    private boolean allGoalsSatisfiable = false;

    private Node currentSubtreeRoot = null;

    private SwingWorker worker;
    
    /** number of rules automatically applied */
    private int countApplied = 0;
    private long time;
    
    /** interrupted by the user? */
    private boolean userTerminated = false;


    private StrategyManager strategy = null;


    private ProofListener proofListener = new ProofListener ();


    /**
     * creates a new ApplyHeuristics with a given IaProver that provides
     * the necessary data and notification mechanisms and with a given
     * StopHeuristics, i.e. a (graphical) user dialog that may stop this
     * ApplyHeuristics and may itself be finished by this class.
     * @param iaProver the InteractiveProver with the necessary data and methods
     * @param maxApp an int bounding the number of heuristics
     * applications before an interactive step is required
     */
    public ApplyHeuristics(InteractiveProver iaProver, Proof proof, 
                           StrategyManager strategy, int maxApp) {
	this.iaProver = iaProver;
	this.proof = proof;
	javaInfo = iaProver.mediator().getJavaInfo();
	maxApplications = maxApp;

	this.strategy = strategy;
    }


    /**
     * @return the next goal a taclet should be applied to
     */    
    protected Goal getNextGoal () {
	if ( allGoalsSatisfiable ) {
	    if ( nextGoals == SLListOfGoal.EMPTY_LIST )
		nextGoals = selectedList;
	    if ( nextGoals == SLListOfGoal.EMPTY_LIST )
		return null;
	    else {
		Goal first = nextGoals.head ();
		nextGoals  = nextGoals.tail ();
	    
		return first;
	    }
	} else
	    return
		selectedList == SLListOfGoal.EMPTY_LIST ?
		null :
		selectedList.head ();
	// Never reached
    }


    /**
     * Remove p_goal from selectedList (e.g. no taclet can be applied to
     * p_goal)
     */
    protected void removeGoal ( Goal p_goal ) {
	selectedList = selectedList.removeAll ( p_goal );
	nextGoals    = SLListOfGoal.EMPTY_LIST;
    }


    /** applies Taclets that belongs to an active heuristic 
     * @return true iff a rule has been applied, false otherwise
     */
    public synchronized boolean applyHeuristicTaclet() {
	/** Look for the strategy ... */
	if ( strategy != null ) {
	    RuleAppRecommendation rec = null;
	    Goal                  g;

	    while ( ( g = getNextGoal () ) != null ) {
		rec = strategy.recommendRuleApp
		    ( iaProver.activeHeuristics(),
		      SLListOfGoal.EMPTY_LIST.prepend ( g ),
		      javaInfo,
		      iaProver.mediator()
		      .getUserConstraint().getConstraint() );

		if ( rec == null )
		    removeGoal ( g );
		else
		    break;
	    }

	    if ( rec != null ) {

		//		System.out.println ( "Recommendation: " + rec.getRuleApp () );
		if ( rec.getRuleApp () instanceof BuiltInRuleApp &&
		     rec.getRuleApp ().rule () instanceof UpdateSimplificationRule )
		    applyUpdateSimplificationRule ( rec );
		else
		    iaProver.apply ( rec.getRuleApp (),
				     rec.getGoal    (),
				     false );
		//		System.out.println ( "... done" );
		return true;
	    }
	}

	return false;
    }


    ////////////// HACK
    private void applyUpdateSimplificationRule ( RuleAppRecommendation p_rec ) {
	Goal           goal = p_rec.getGoal ();
	BuiltInRuleApp rapp = (BuiltInRuleApp)p_rec.getRuleApp ();

	if ( iaProver.getUpdateSimplificationMode () ) {
	    PosInOccurrence    pio        = rapp.posInOccurrence ();
	    ConstrainedFormula conFormula = pio.constrainedFormula ();
	    ConstrainedFormula simplified = goal.simplifier().simplify
		( conFormula,
		  iaProver.mediator ().getUserConstraint ().getConstraint () );
	    goal.changeFormula ( simplified, pio ); 
	} else
	    iaProver.apply ( rapp, goal, false );
    }


    private void updateGoalList( Node       node, 
				 ListOfGoal newGoals ) {
	if ( newGoals == SLListOfGoal.EMPTY_LIST ||
	     ( newGoals.tail () == SLListOfGoal.EMPTY_LIST &&
	       newGoals.head ().node () == node ) ) {
	    // Goals (may) have been closed, remove them from the goal
	    // lists
	    Goal           goal;
	    boolean        changed = false;

	    IteratorOfGoal it      = goalList.iterator ();
	    goalList               = SLListOfGoal.EMPTY_LIST;

	    while ( it.hasNext () ) {
		goal = it.next ();
		if ( proof.openGoals ().contains ( goal ) )
		    // order of goals is not relevant
		    goalList = goalList.prepend ( goal );
	    }
	    
	    it                     = selectedList.iterator ();
	    ListOfGoal newList     = SLListOfGoal.EMPTY_LIST;
	    
	    while ( it.hasNext () ) {
		goal = it.next ();
		if ( proof.openGoals ().contains ( goal ) ) {
		    if ( !allGoalsSatisfiable &&
			 goal.getClosureConstraint ().isSatisfiable () ) {
			goalList = goalList.prepend ( goal );
			changed  = true;
		    } else
			newList  = newList.prepend ( goal );
		} else
		    changed      = true;
	    }

	    if ( changed ) {
		nextGoals    = SLListOfGoal.EMPTY_LIST;

		// for "selectedList", order does matter
		it           = newList.iterator ();
		selectedList = SLListOfGoal.EMPTY_LIST;;
		while ( it.hasNext () )
		    selectedList = selectedList.prepend ( it.next () );
	    }
	} else {
	    ListOfGoal prevGoalList     = SLListOfGoal.EMPTY_LIST;
	    boolean    newGoalsInserted = false;
	    nextGoals                   = SLListOfGoal.EMPTY_LIST;

	    // Remove "node" and goals contained within "newGoals"
	    while ( selectedList != SLListOfGoal.EMPTY_LIST ) {
		if ( selectedList.head ().node () == node ||
		     newGoals.contains ( selectedList.head () ) ) {
		    // continue taclet apps at the next goal in list
		    nextGoals = selectedList = selectedList.tail ();
		    if ( !newGoalsInserted ) {
			IteratorOfGoal it = newGoals.iterator ();
			Goal           g;
			while ( it.hasNext () ) {
			    g = it.next ();
			    if ( proof.openGoals ().contains ( g ) ) {
				if ( !allGoalsSatisfiable &&
				     g.getClosureConstraint ().isSatisfiable () )
				    goalList = goalList.prepend ( g );
				else
				    prevGoalList = prevGoalList.prepend ( g );
			    }
			}
			newGoalsInserted = true;
		    }
		} else {
		    prevGoalList = prevGoalList.prepend ( selectedList.head () );
		    selectedList = selectedList.tail ();
		}
	    }

	    while ( prevGoalList != SLListOfGoal.EMPTY_LIST ) {
		selectedList = selectedList.prepend ( prevGoalList.head () );
		prevGoalList = prevGoalList.tail ();
	    }
	}

	if ( proof.openGoals () == SLListOfGoal.EMPTY_LIST )
	    // proof has been closed
	    nextGoals = selectedList = goalList = SLListOfGoal.EMPTY_LIST;
	else {
	    if ( selectedList == SLListOfGoal.EMPTY_LIST ||
		 ( currentSubtreeRoot != null &&
		   isSatisfiableSubtree ( currentSubtreeRoot ) ) )
		setupGoalLists ( goalList.prepend ( selectedList ) );
	}
    }


    private void setupGoalLists ( ListOfGoal p_goals ) {
	goalList     = SLListOfGoal.EMPTY_LIST;
	selectedList = SLListOfGoal.EMPTY_LIST;
	nextGoals    = SLListOfGoal.EMPTY_LIST;

	if ( allGoalsSatisfiable ) {
	    goalList = p_goals;
	    findMinimalSubtree ( currentSubtreeRoot );
	} else {
	    IteratorOfGoal it   = p_goals.iterator ();
	    Goal           goal;

	    while ( it.hasNext () ) {
		goal = it.next ();
		if ( goal.getClosureConstraint ().isSatisfiable () )
		    goalList     = goalList    .prepend ( goal );
		else
		    selectedList = selectedList.prepend ( goal );
	    }

	    allGoalsSatisfiable = ( selectedList == SLListOfGoal.EMPTY_LIST );

	    if ( allGoalsSatisfiable )
		findMinimalSubtreeBelow ( proof.root () );
	}
    }


    /**
     * Search for a minimal subtree of the proof tree which is not
     * closable (constraints of its goals are inconsistent) below
     * p_startNode
     *
     * PRECONDITION: * !isSatisfiableSubtree ( p_startNode )
     *               * all goals have satisfiable constraints
     *
     * @return true iff a non-empty subtree was found
     */
    private boolean findMinimalSubtreeBelow ( Node p_startNode ) {

	Node node = p_startNode;
	Node child;
	
	while ( node.childrenCount () == 1 )
	    node = node.child ( 0 );

	IteratorOfNode childrenIt = node.childrenIterator ();

	while ( childrenIt.hasNext () ) {
	    child = childrenIt.next ();
	    if ( !isSatisfiableSubtree   ( child ) &&
		 findMinimalSubtreeBelow ( child ) )
		return true;
	}

	Goal goal;

	currentSubtreeRoot = p_startNode;
	childrenIt         = node.leavesIterator ();
	
	while ( childrenIt.hasNext () ) {
	    child = childrenIt.next ();
	    goal  = proof.getGoal ( child );
	    if ( goalList.contains ( goal ) ) {
		selectedList = selectedList.prepend   ( goal );
		goalList     = goalList    .removeAll ( goal );
	    }
	}

	return selectedList != SLListOfGoal.EMPTY_LIST;

    }

    
    /**
     * Search for a minimal subtree of the proof tree which is not
     * closable (constraints of its goals are inconsistent) starting
     * at p_startNode
     *
     * PRECONDITION: all goals have satisfiable constraints
     */
    private void findMinimalSubtree ( Node p_startNode ) {
	while ( isSatisfiableSubtree ( p_startNode ) )
	    p_startNode = p_startNode.parent ();

	if ( !findMinimalSubtreeBelow ( p_startNode ) )
	    findMinimalSubtreeBelow ( proof.root () );
    }


    private boolean isSatisfiableSubtree ( Node p_root ) {
	return p_root.getBranchSink ().getResetConstraint ()
	    .isSatisfiable ();
    }


    void setupStatus() {
	MainStatusLine sl = Main.getInstance ().getStatusLine ();
	sl.reset ();
	sl.setProgressPanelVisible ( true );
	sl.setAbortButtonEnabled ( true );
	sl.setProgressBarMaximum ( maxApplications - 1 );
	sl.setStatusText ( "Processing Heuristics" );
	sl.addAbortListener ( interruptListener );
    }


    void removeStatus() {
	MainStatusLine sl = Main.getInstance ().getStatusLine ();
	sl.reset ();
    }


    void updateStatus() {
	Main.getInstance ().setProgress ( countApplied );
    }
    

    /**
     * applies heuristics until this is no longer
     * possible or the thread is interrupted.
     */
    Object doWork() {
        time = System.currentTimeMillis();
        try{
	   Debug.out("Heuristics started.");
	   //	   nonApplicableGoals = new HashSet();
	   while (countApplied < maxApplications &&
	          applyHeuristicTaclet()) {
               countApplied++;
               updateStatus();
               if (Thread.interrupted()) throw new InterruptedException();
	   }
	   //	   applyCloseRules ();
        } catch (InterruptedException e) {
            return "Interrupted";  // SwingWorker.get() returns this
        } catch (RuntimeException rte) { // treated later in finished()
            System.err.println(rte);
            rte.printStackTrace();
            return "Error";
	} finally{
	    time = System.currentTimeMillis()-time;
	    Debug.out("Heuristics stopped.");
	    Debug.out("Applied ", countApplied);
	    Debug.out("Time elapsed: ", time);
        }
        return "Done.";
    }
    
    
    
    public void start(ListOfGoal goals) {
	setupGoalLists ( goals );

	iaProver.addInteractiveProofListener ( proofListener );
	strategy.settingsChange ( iaProver.activeHeuristics(),
				  proof,
				  javaInfo,
				  iaProver.mediator()
				  .getUserConstraint().getConstraint() );

        /* Invoking start() on the SwingWorker causes a new Thread
         * to be created that will call construct(), and then
         * finished().  Note that finished() is called even if
         * the worker is interrupted because we catch the
         * InterruptedException in doWork().
         */
        worker = new SwingWorker() {
            public Object construct() {
                Object res = doWork();
		return res;
            }
            public void finished() {
		iaProver.removeInteractiveProofListener ( proofListener );

		Main main = Main.getInstance ();

		removeStatus ();

		String timeString = "" + (time/1000)+"."+((time%1000)/100);

                if("Error".equals(get()))
		    JOptionPane.showMessageDialog(
                   iaProver.mediator ().mainFrame (),
		   "An exception occured while executing heuristics.",
                   "Error", JOptionPane.ERROR_MESSAGE);
                iaProver.setInteractive(true);
		iaProver.fireAutoModeStopped(new ProofEvent(iaProver));

		String message;
		int closed = iaProver.mediator().getNrGoalsClosedByHeuristics();

		// display message in the status bar

		if ( countApplied != 0 ) {
		    message = "Heuristics: Applied " + countApplied + " rule";
		    if ( countApplied != 1 )
			message += "s";
		    if ( closed > 0 ) {
			message += ", closed " + closed + " goal";
			if ( closed != 1 )
			    message += "s";			    
		    }
		    message += " (" + timeString + " sec)";
		    main.setStatusLine ( message );
		}

		if ( iaProver.getProof().openGoals().size()==0 ) {
		    // show message dialog
		
		    // inform that heuristics have been stopped
		    message = "Heuristics stopped"+
			(userTerminated ? "by user.\n" : ".\n");
		    if (countApplied == maxApplications) {
			message = "Maximal number of heuristic applications" +
			    "has been reached.\n";
		    }

		    // Show message ...
		    if (closed>0) {
			message += closed+
			    ((closed>1) ? " goals have" : " goal has") +
			    " been closed.\n";
		    }
		    message += "Applied " + countApplied + " rules in " +
			timeString + " seconds.";
		    if (iaProver.getProof().openGoals().size()==0) {
			message="Proved";
		    }
		    if ((closed>0) || (countApplied == maxApplications) 
			|| "Proved".equals(message)) {
			JOptionPane.showMessageDialog
			    (iaProver.mediator ().mainFrame (), message, "Note ...", 
			     JOptionPane.INFORMATION_MESSAGE);
		    }
		}

		iaProver.mediator().resetNrGoalsClosedByHeuristics();		
            }
        };
	iaProver.fireAutoModeStarted(new ProofEvent(iaProver));
        iaProver.setInteractive(false);
	setupStatus ();
        worker.start();
    }



    /**
     * This action listener, called by the "Stop" button, interrupts
     * the worker thread which is running this.doWork().  Note that
     * the doWork() method handles InterruptedExceptions cleanly.
     */
    final ActionListener interruptListener = new ActionListener() {
        public void actionPerformed(ActionEvent event) {
            userTerminated = true;
            worker.interrupt();
        }
    };
    
    
    private class ProofListener implements InteractiveProofListener {

	/** invoked if automatic execution has started
	 */
	public void autoModeStarted(ProofEvent e) {}

	/** invoked if automatic execution has stopped
	 */
	public void autoModeStopped(ProofEvent e) {}
    
	/** invoked when a rule has been applied */
	public void ruleApplied(ProofEvent e) {
	    RuleAppInfo rai = e.getRuleAppInfo ();
	    if ( rai == null )
		return;

	    synchronized ( ApplyHeuristics.this ) {
		ListOfGoal                newGoals = SLListOfGoal.EMPTY_LIST;
		IteratorOfNodeReplacement it       = rai.getReplacementNodes ();
		Node                      node;
		Goal                      goal;

		while ( it.hasNext () ) {
		    node = it.next ().getNode ();
		    goal = proof.getGoal ( node );
		    if ( goal != null )
			newGoals = newGoals.prepend ( goal );
		}

		updateGoalList ( rai.getOriginalNode (),
				 newGoals );
	    }
	}
	
    }    

}
