// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;

import java.util.List;
import java.util.ArrayList;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.ProofTreeListener;
import de.uka.ilkd.key.proof.ProofTreeAdapter;
import de.uka.ilkd.key.proof.ProofTreeEvent;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.SLListOfGoal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.logic.Sequent;

public class GoalList extends JList implements java.io.Serializable {
    
    static final ImageIcon keyIcon=IconFactory.keyHole(20,20);

    private KeYMediator mediator;   

    /** the proof where the goals belongs to */
    private Proof proof;
    /** the model used by this view */
    private GoalListModel goalListModel;

    /** interactive prover and KeYSelection-Listener */
    private GoalListInteractiveListener interactiveListener;
    
    /** listens to gui events */
    private GoalListGUIListener guiListener; 

    MouseListener mouseListener = new GoalListMouseListener();


    class GoalListMouseListener extends MouseAdapter 
	implements java.io.Serializable {
	public void mouseClicked(MouseEvent e) {
	    if (e.getClickCount() == 1) {
		goalChosen();
	    }
	}
    }
    

    public GoalList(KeYMediator mediator){
	interactiveListener = new GoalListInteractiveListener();
	guiListener         = new GoalListGUIListener();

	setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	setMediator(mediator);
	goalListModel = new GoalListModel(mediator.getSelectedProof());
	if (goalListModel.getSize() > 0) {
	    setSelectedIndex(0);
	}       	
	setModel(goalListModel);
	setCellRenderer(new IconCellRenderer());
	addMouseListener(mouseListener);		
	updateUI();
    }

    /** set the KeYMediator */
    private void setMediator(KeYMediator m) {
	if (mediator != null) {
	    unregister();
	}
	mediator = m;
	register();
    }
    

    public void updateUI() {
	super.updateUI();
	Font myFont = UIManager.getFont(Config.KEY_FONT_GOAL_LIST_VIEW);
	if (myFont != null) {
	    setFont(myFont);
	} else {
	    Debug.out("goallist: Warning: Use standard font. Could not find font:", 
		      Config.KEY_FONT_GOAL_LIST_VIEW);
	}
    }

    private void register() {	
	mediator.addKeYSelectionListener(interactiveListener);
	mediator.addInteractiveProofListener(interactiveListener);
	mediator.addGUIListener(guiListener);
    }

    private void unregister() {
	mediator.removeKeYSelectionListener(interactiveListener);
	mediator.removeInteractiveProofListener(interactiveListener);
	mediator.removeGUIListener(guiListener);
    }

    public void removeNotify () {
	unregister ();
	super.removeNotify ();
    }

    public KeYMediator mediator() {
	return mediator;
    }

    
    private void goalChosen() {
	Goal goal = (Goal)getSelectedValue();
	if (goal != null) {
	    mediator.goalChosen(goal);
	}
    }

    /** overrides setVisible from JFrame 
     * takes care that the view item is in the right state
     */
    public void setVisible(boolean b) {
	super.setVisible(b);	
    }


    //INNER CLASSES
        

    class GoalListGUIListener implements GUIListener,
                                         java.io.Serializable {
	/** invoked if a frame that wants modal access is opened */
	public void modalDialogOpened(GUIEvent e) {
	    setEnabled(false);
	}

	/** invoked if a frame that wants modal access is closed */
	public void modalDialogClosed(GUIEvent e) {
	    setEnabled(true);
	}

	public void shutDown(GUIEvent e) {
	}

    }


// This is never used? -VK
    class GoalListListener extends WindowAdapter 
                           implements java.io.Serializable {
	public void windowClosing() {
	    setVisible(false);
	}    
    } 


    class GoalListInteractiveListener implements InteractiveProofListener, 
                                                 KeYSelectionListener,
						 java.io.Serializable {


	/** focused node has changed */
	public void selectedNodeChanged(KeYSelectionEvent e) {
	    if (e.getSource().isGoal()) {
		setSelectedValue(e.getSource().getSelectedNode(), 
				 true);
	    } else {
		clearSelection();
	    }
	    validate();
	}

	/** the selected proof has changed (e.g. a new proof has been
	 * loaded) */ 
	public void selectedProofChanged(KeYSelectionEvent e) {
	    Debug.out("GoalList: initialize with new proof");
	    goalListModel.setProof(e.getSource().getSelectedProof());
	    if (goalListModel.getSize() > 0) {
		setSelectedIndex(0);
	    }       	
	    validate();
	}

	/** invoked if automatic executiion of heuristics has started
	 */
	public void autoModeStarted(ProofEvent e) {	    
	    if (goalListModel.isAttentive()) {
		mediator.removeKeYSelectionListener(interactiveListener);
	    }
	    goalListModel.setAttentive(false);
	}
	
	/** invoked if automatic execution of heuristics has stopped
	 */
	public void autoModeStopped(ProofEvent e) {
	    if (!goalListModel.isAttentive()) {
		mediator.addKeYSelectionListener(interactiveListener);
	    }
	    goalListModel.setAttentive(true);
	}	

	/** invoked when a rule has been applied */
	public void ruleApplied(ProofEvent e) {
	}


    }    

    static class GoalListModel extends AbstractListModel {
    
	/** the proof the model belongs to */	
	private Proof proof;
	/** */
	private List goals;
	/** is used to indicate if the model has to be updated */
	private boolean attentive;
	/** listens to the proof */
	private final ProofTreeListener proofTreeListener
	    = new GoalListProofTreeListener();


    	GoalListModel(Proof p) {
	    goals = new ArrayList(10);
	    setProof(p);
	}

	/** 
	 * the proof this view belongs to has changed
	 */
	private void setProof(Proof p) {
	    if (proof != null) {
		proof.removeProofTreeListener(proofTreeListener);
		clear();
	    } 
	    proof = p;
	    if (proof != null) {
		proof.addProofTreeListener(proofTreeListener);
		add(proof.openGoals());
	    }
	    attentive = true;
	}

	/** Sets whether this object should respond to changes in the
	 * the proof immediately. */
	private void setAttentive(boolean b) {
            if (proof == null) return;
	    if ( b != attentive ) {
		if (b) {
		    proof.addProofTreeListener(proofTreeListener);
		    clear();
		    add(proof.openGoals());
		} else {
		    proof.removeProofTreeListener(proofTreeListener);
		}
		attentive = b;
	    }
	}

	/** returns true if the model respond to changes in the proof
	 * immediately */ 
	public boolean isAttentive() {
	    return attentive;
	}

	public void add(ListOfGoal g) {
	    if (g != SLListOfGoal.EMPTY_LIST) {
		IteratorOfGoal it = g.iterator();
		while (it.hasNext()) {
		    goals.add(it.next());
		}	
		fireIntervalAdded(this, goals.size()-g.size(), 
				  goals.size()-1);
	    }
	}

	public void add(Goal g) {
	    if (g != null) {
		goals.add(g);
		int index = goals.indexOf(g);
		fireIntervalAdded(this, index, index);
	    }
	}

	public void remove(Goal g) {
	    int index = goals.indexOf(g);
	    if (index > -1) {
		goals.remove(g);
		fireIntervalRemoved(this, index, index+1);
	    }
	}

	public void clear() {
	    int size = goals.size();
	    if (size > 0) {
		goals.clear();	    
		fireIntervalRemoved(this, 0, size);
	    }
	}

	public int getSize() {
	    return goals.size();
	}

	public Object getElementAt(int i) {
	    return goals.get(i);
	} 

	class GoalListProofTreeListener extends ProofTreeAdapter 
	                                implements java.io.Serializable {

	    /** invoked if all goals of the proof are closed
	     */
	    public void proofClosed(ProofTreeEvent e) {	   	    
		setAttentive(true);
		clear();
	    }

	    /** The proof tree has been pruned under the node mentioned in the
	     * ProofTreeEvent.  In other words, that node should no longer
	     * have any children now.  Any nodes that were not descendants of
	     * that node are unaffected.*/
	    public void proofPruned(ProofTreeEvent e) {
		clear();
		add(e.getSource().openGoals());
	    }


	    /** invoked if the list of goals changed (goals were added,
	     * removed etc.
	     */
	    public void proofGoalRemoved(ProofTreeEvent e) {
		remove(e.getGoal());
	    }
	
	    /** invoked if the current goal of the proof changed */
	    public void proofGoalsAdded(ProofTreeEvent e) {
		add(e.getGoals());
	    }

	    /** invoked if the current goal of the proof changed */
	    public void proofGoalsChanged(ProofTreeEvent e) {
		clear();		
		add(e.getGoals());
	    }
	}
    }

    class IconCellRenderer extends DefaultListCellRenderer 
	                           implements ListCellRenderer,
				             java.io.Serializable { 
	
	public IconCellRenderer() {
	    GoalList.this.setToolTipText("GOAL");
	}
	
	public Component getListCellRendererComponent
	    (JList list,
	     Object value,            // value to display
	     int index,               // cell index
	     boolean isSelected,      // is the cell selected
	     boolean cellHasFocus)    // the list and the cell have the focus
	{
	    String valueStr;
	    Color  col = Color.black;
	    if (value instanceof Goal) {
		Sequent seq = ((Goal)value).node().sequent();
		SequentPrinter sp = new PureSequentPrinter
		 (new ProgramPrinter(null), 
		  ((GoalList)list).mediator().getNotationInfo(), seq);
		seq.prettyprint(sp);	    
		valueStr = sp.toString().replace('\n',' ');

		if ( mediator () != null &&
		     mediator ().getUserConstraint ().displayClosed
		     ( ((Goal)value).node () ) )
		    col = Color.green;
		else if ( ((Goal)value).getClosureConstraint ().isSatisfiable () )
		    col = Color.blue;
	    } else {
		valueStr = ""+value;
	    }
	    DefaultListCellRenderer sup=
		(DefaultListCellRenderer)
		super.getListCellRendererComponent(list, valueStr, 
						   index, isSelected, 
						   cellHasFocus);
	    sup.setIcon(keyIcon);

	    // set color according to closure status
	    sup.setForeground ( col );

	    return sup;
	}
    }
}
