// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import javax.swing.border.*;
import java.util.*;

import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.rule.Heuristic;
import de.uka.ilkd.key.rule.ListOfHeuristic;
import de.uka.ilkd.key.rule.SLListOfHeuristic;
import de.uka.ilkd.key.rule.IteratorOfHeuristic;
import de.uka.ilkd.key.logic.IteratorOfNamed;

public class HeuristicSelector extends JDialog {

    private HeuristicListModel notSelectedListModel;
    private HeuristicListModel activeHeuristicsListModel;
    private JList notSelectedHeuristics;
    private JList activeHeuristics;
    private List selected;
    private List notSelected;
    private JButton toRight;
    private JButton toLeft;
    private JButton allToRight;
    private JButton allToLeft;
    private JTextField maxHeuristicStepsInputField;

    private ListOfHeuristic originalSelectedList;
    private KeYMediator mediator;
    private Proof proof = null;
    private Namespace available_heur;

    /** creates a new HeuristicSelector */
    public HeuristicSelector(JFrame parent, KeYMediator mediator) {  
	super(parent, "Active Heuristics", true);
        this.mediator = mediator;
        proof = mediator.getProof();
        available_heur = proof.getNamespaces().heuristics();

	createSelectedList(proof.activeHeuristics());
	createNotSelectedList(available_heur);
	
	layoutHeuristicSelector(); 
	
	pack();
	setLocation(70,70);
	setVisible(true);
    }

    /** 
     * creates a linked list with same content like sel
     * @param sel ListOfHeuristic to be converted
     */
    private void createSelectedList(ListOfHeuristic sel) {
        selected = new LinkedList();
	IteratorOfHeuristic it = sel.iterator();
	while (it.hasNext()) {
	    Heuristic h = it.next();
	    selected.add(h);	    
	}
    }

    /** creates a linked list from the heuristics in the given
     * namespace without the heuristics that are found in the list of
     * selected(active) heuristics
     * @param available_heur the Namespace with all known heuristics
     */
    private void createNotSelectedList(Namespace available_heur) {
        notSelected = new LinkedList();
	Namespace h_nsp = available_heur;
	while (h_nsp != null) {
	    IteratorOfNamed it = h_nsp.elements().iterator();
	    while (it.hasNext()) {
		Heuristic next = (Heuristic)it.next();
		if (!selected.contains(next)) {
		    notSelected.add(next);
		}
	    }
	    h_nsp = h_nsp.parent();
	}
    }

    private void updateButtons() {
	toRight.setEnabled(!notSelectedHeuristics.isSelectionEmpty());	
	toLeft.setEnabled(!activeHeuristics.isSelectionEmpty());	  
	allToRight.setEnabled(notSelectedListModel.getSize()>0);	
	allToLeft.setEnabled(activeHeuristicsListModel.getSize()>0);	  
    }


    
    private void makeActive(Object[] notSelected) {
	if (notSelected==null || notSelected.length==0) {
	    return;
	}
	activeHeuristicsListModel.add(notSelected);
	activeHeuristics.setSelectedValue(notSelected[0],true);
	notSelectedListModel.remove(notSelected);	
	updateButtons();
    }

    private void makeInactive(Object[] selected) {
	if (selected==null || selected.length==0) {
	    return;
	}
	notSelectedListModel.add(selected);
	notSelectedHeuristics.setSelectedValue(selected[0],true);

	activeHeuristicsListModel.remove(selected);
	updateButtons();
    }

    /** lays out the selector */
    protected void layoutHeuristicSelector() { 
	getContentPane().setLayout(new BorderLayout());

	notSelectedListModel = new HeuristicListModel(notSelected);
	notSelectedHeuristics = new JList(notSelectedListModel);
	notSelectedHeuristics.setSelectionMode(ListSelectionModel.
					       MULTIPLE_INTERVAL_SELECTION);
	notSelectedHeuristics.setBackground(Color.white);

	activeHeuristicsListModel = new HeuristicListModel(selected);
	activeHeuristics = new JList(activeHeuristicsListModel); 
	activeHeuristics.setSelectionMode(ListSelectionModel.
					  MULTIPLE_INTERVAL_SELECTION);
	activeHeuristics.setBackground(Color.white);	

	notSelectedHeuristics.addMouseListener(new MouseAdapter() {
	
		public void mouseClicked(MouseEvent e) {
		    updateButtons();
		    if (e.getClickCount() == 2) {
			makeActive(notSelectedHeuristics.getSelectedValues());	      
		    }
		}
	    });

	activeHeuristics.addMouseListener(new MouseAdapter() {

		public void mouseClicked(MouseEvent e) {
		    updateButtons();
		    if (e.getClickCount() == 2) {
			makeInactive(activeHeuristics.getSelectedValues());
		    }
		}
	    });
	
	JScrollPane notSelScrollPane = new
	    JScrollPane(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, 
			JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	notSelScrollPane.getViewport().setView(notSelectedHeuristics);
	notSelScrollPane.setBorder(new TitledBorder("Possible Heuristics"));

	JScrollPane activeScrollPane = new 	    
	    JScrollPane(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, 
			JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);

	activeScrollPane.getViewport().setView(activeHeuristics);
	activeScrollPane.setBorder(new TitledBorder("Active Heuristics"));


 	JPanel maxInput=new JPanel();
 	maxInput.setLayout(new BoxLayout(maxInput, BoxLayout.X_AXIS));
 	maxHeuristicStepsInputField = new NumberInputField(mediator.getMaxHeuristicSteps(), 4);
	maxHeuristicStepsInputField.setMaximumSize(new Dimension(40,30));
 	maxInput.add(new JLabel("Maximum heuristics applications: "));
	maxInput.add(maxHeuristicStepsInputField);

 	JButton okButton = new JButton("OK");
	okButton.setMnemonic(KeyEvent.VK_ENTER);

 	okButton.addActionListener(new ActionListener() {
 		public void actionPerformed(ActionEvent e) {
		    int maxSteps = 
                        Integer.parseInt(maxHeuristicStepsInputField.getText());
		    mediator.setMaxHeuristicSteps(maxSteps);
		    proof.setActiveHeuristics(activeHeuristics());		    
		    setVisible(false);
 		}
	    });
 	JButton saveButton = new JButton("Save as Default");

 	saveButton.addActionListener(new ActionListener() {
 		public void actionPerformed(ActionEvent e) {
		    int maxSteps = 
                        Integer.parseInt(maxHeuristicStepsInputField.getText());
		    mediator.setMaxHeuristicSteps(maxSteps);
		    proof.setActiveHeuristics(activeHeuristics());
		    ProofSettings dflt=ProofSettings.THE_SETTINGS;		    
		    dflt.getHeuristicSettings().setMaxSteps(maxSteps);
		    dflt.getHeuristicSettings().setHeuristics(activeHeuristics());
		    //temporary solution, stores more than wanted %%%% 
		    dflt.saveSettings();
		    setVisible(false);
 		}
	    });
 	JButton cancelButton = new JButton("Cancel");
	cancelButton.setMnemonic(KeyEvent.VK_ESCAPE);
 	cancelButton.addActionListener(new ActionListener() {
 		public void actionPerformed(ActionEvent e) {
		    setVisible(false);
		    dispose();
 		}
 	    });

 	JPanel buttonPanel = new JPanel();
 	buttonPanel.add(okButton);
 	buttonPanel.add(saveButton);
 	buttonPanel.add(cancelButton);
	
	JPanel labelPanel = new JPanel();
//	labelPanel.setLayout(new GridLayout(5,1));
	toRight=new JButton(">");
	toRight.setToolTipText("Activate heuristic");
	toRight.addActionListener(new ActionListener() {
 		public void actionPerformed(ActionEvent e) {
		    makeActive(notSelectedHeuristics.getSelectedValues());
 		}
 	    });
	toLeft=new JButton("<");
	toLeft.setToolTipText("Deactivate heuristic");
	toLeft.addActionListener(new ActionListener() {
 		public void actionPerformed(ActionEvent e) {
		    makeInactive(activeHeuristics.getSelectedValues());
 		}
 	    });



	allToRight=new JButton(">>");
	allToRight.setToolTipText("Activate all");
	allToRight.addActionListener(new ActionListener() {
 		public void actionPerformed(ActionEvent e) {
		    makeActive(notSelectedListModel.getAllElements());
 		}
 	    });
	allToLeft=new JButton("<<");
	allToLeft.setToolTipText("Deactivate all");
	allToLeft.addActionListener(new ActionListener() {
 		public void actionPerformed(ActionEvent e) {
		    makeInactive(activeHeuristicsListModel.getAllElements());
 		}
 	    });




	labelPanel.add(allToRight);
	labelPanel.add(new JLabel(""));
	labelPanel.add(toRight);
	labelPanel.add(new JLabel(""));
	labelPanel.add(toLeft);
	labelPanel.add(new JLabel(""));
	labelPanel.add(allToLeft);
	labelPanel.setSize(30,100);

	JPanel listPanel=new JPanel();

	listPanel.setLayout(new BoxLayout(listPanel,BoxLayout.X_AXIS));
	notSelScrollPane.setPreferredSize(new Dimension(250, 150));
        notSelScrollPane.setMinimumSize(new Dimension(250, 150));

	labelPanel.setPreferredSize(new Dimension(80, 150));
        labelPanel.setMinimumSize(new Dimension(80, 150));
	labelPanel.setMaximumSize(new Dimension(80, 150));
	activeScrollPane.setPreferredSize(new Dimension(250, 150));
        activeScrollPane.setMinimumSize(new Dimension(250, 150));


	listPanel.add(notSelScrollPane);
	listPanel.add(labelPanel);
	listPanel.add(activeScrollPane);

	getContentPane().setLayout(new BoxLayout(getContentPane(), 
						 BoxLayout.Y_AXIS));
	getContentPane().add(listPanel);
	getContentPane().add(maxInput);
	getContentPane().add(buttonPanel);

	updateButtons();

    }
    
    
    /** converts the list with all active heuristics to a
     * ListOfHeuristic Representation 
     * @return ListOfHeuristic with the active heuristics
     */
    private ListOfHeuristic activeHeuristics() {
	ListOfHeuristic result = SLListOfHeuristic.EMPTY_LIST;
	for (int i = selected.size()-1; i >= 0 ; i--) {
	    result = result.prepend((Heuristic)selected.get(i)); 
	}
	return result;
     }

    // INNER CLASS
    
    /** model of the heuristic lists used for JList*/
    static class HeuristicListModel extends AbstractListModel {
    
	private List heur_list;
	private static final Comparator heuristicComparator = 
	    new Comparator() {
		    public int compare(Object o1, Object o2) {
			return (""+o1).compareTo(""+o2);
		    }
		};
	

	HeuristicListModel(List heur_list) {
	    this.heur_list = heur_list;
	    Collections.sort(this.heur_list, heuristicComparator);
	}
	
	public int getSize() {
	    return heur_list.size();
	}
	public void remove(Object obj) {
	    int index = heur_list.indexOf(obj);
	    heur_list.remove(obj);

	    fireIntervalRemoved(this,
				index,
				index+1);
	}

	public void add(Object obj) {
	    int s = heur_list.size();
	    heur_list.add(obj);
	    Collections.sort(heur_list, heuristicComparator);
	    fireIntervalAdded(this, s, s + 1);
	    fireContentsChanged(this, 0, s);
	}

	public void add(Object[] obj) {
	    int s = heur_list.size();
	    for (int i=0; i<obj.length; i++) {
		heur_list.add(obj[i]);
	    }
	    Collections.sort(heur_list, heuristicComparator);
	    fireIntervalAdded(this, s, s + obj.length);
	    fireContentsChanged(this, 0, s);
	}


	public void remove(Object[] obj) {
	    for (int i=0; i<obj.length; i++) {
		remove(obj[i]);
	    }
	}


	public Object getElementAt(int i) {
	    return heur_list.get(i);
	}
        

	public Object[] getAllElements() {
	    return heur_list.toArray();
	}
        

    }

    // INNER CLASS TO READ ONLY NUMBERS FOR MAX APPs
    static class NumberDocument extends PlainDocument {

	public void insertString(int offs, String str, AttributeSet a) 
	    throws BadLocationException {
	    if (str == null) {
		return;
	    }
	    char[] upper = str.toCharArray();
	    for (int i = 0; i < upper.length; i++) {
		if (upper[i] < '0' || upper[i] > '9') {
		    return;
		} 
	    }
	    super.insertString(offs, new String(upper), a);
	}
	
    }

    static class NumberInputField extends JTextField {
	public NumberInputField(int number, int cols) {
	    super(cols);
	    setText(""+number);
	}

	protected Document createDefaultModel() {
	    return new NumberDocument();
	}
    }



}    
