// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.gui;

import de.uka.ilkd.key.rule.Heuristic;
import de.uka.ilkd.key.rule.ListOfHeuristic;
import de.uka.ilkd.key.rule.SLListOfHeuristic;
import de.uka.ilkd.key.rule.IteratorOfHeuristic;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.util.Debug;
import java.util.Properties;
import java.util.LinkedList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.StringTokenizer;

/** This class encapsulates the information about the active
 * Heuristics and the maximum amount of heuristics steps before an
 * interactive step is required.
 */
public class HeuristicSettings implements Settings {

    private static final String HEURISTICS_KEY = "[Heuristic]ActiveHeuristics";
    private static final String STEPS_KEY = "[Heuristic]MaximumNumberOfHeuristcsApplications";

    private LinkedList listenerList = new LinkedList();
    private HashSet activeHeuristics;
    private int maxSteps;
    
    public HeuristicSettings(int mSteps) {
	activeHeuristics = new HashSet();
	maxSteps = mSteps;
    }

    /** same like
     * <code> 
     * new HeuristicSettings(10) 
     * </code>
     */
    public HeuristicSettings() {
	this(10);
    }

    /** returns the maximal amount of heuristics steps before an user
     * interaction is required 
     * @return the maximal amount of heuristics steps before an user
     * interaction is required
     */
    public int getMaxSteps(){
	return maxSteps;
    }
        
    /** returns the active heuristics
     * @return the active heuristics
     */
    public ListOfHeuristic getActiveHeuristics(Namespace heur_nsp){
	ListOfHeuristic result = SLListOfHeuristic.EMPTY_LIST;
	Iterator it = activeHeuristics.iterator();
	while (it.hasNext()) {
	    Heuristic heuristic = (Heuristic)heur_nsp.lookup((Name)it.next());
	    if (heuristic != null) {
		result = result.prepend(heuristic);
	    }
	}
	return result;
    }
       
    /** sets the active Heuristics 
     * @param heuristics ListOfHeuristic that have been set to active
     */
    public void setHeuristics(ListOfHeuristic heuristics) {
	activeHeuristics = new HashSet();	
	IteratorOfHeuristic it = heuristics.iterator();
	while (it.hasNext()) {
	    activeHeuristics.add(it.next().name());	    
	}
	fireSettingsChanged();
    }

    /** sets the maximal amount of heuristic steps before a user
     * interaction is required
     * @param mSteps maximal amount of heuristic steps
     */
     public void setMaxSteps(int mSteps) {
	 maxSteps = mSteps;
	 fireSettingsChanged();
     }
    
    /** gets a Properties object and has to perform the necessary
     * steps in order to change this object in a way that it
     * represents the stored settings
     */
    public void readSettings(Properties props) {
	activeHeuristics = new HashSet();
	String numString = props.getProperty(STEPS_KEY);
	String heuSequence = props.getProperty(HEURISTICS_KEY);
	int numSteps = 10;

	if (numString != null){
	    try {
		numSteps = Integer.parseInt(numString);
	    } catch(NumberFormatException e) { 
		Debug.out("heuristicssettings: failure while converting the string "+
			  "with the allowed steps of heuristics applications to int."+
			  "Use default value 10 instead."+
			  "\nThe String that has been tried to convert was", numString);
	    }
	}

	// set heuristics
	if (heuSequence != null) {
	    StringTokenizer st = new StringTokenizer(heuSequence, ",");
	    while (st.hasMoreTokens()) {
		activeHeuristics.add(new Name(st.nextToken().trim()));
	    }
	}
	// set max steps
	setMaxSteps(numSteps);
	
    }


    /** implements the method required by the Settings interface. The
     * settings are written to the given Properties object. Only entries of the form 
     * <key> = <value> (,<value>)* are allowed.
     * @param props the Properties object where to write the settings as (key, value) pair
     */
    public void writeSettings(Properties props) {
	String heuristicSequence = "";
	Iterator it = activeHeuristics.iterator();
	while(it.hasNext()) {
	    heuristicSequence += it.next().toString();
	    if(it.hasNext()){
		heuristicSequence += " , ";
	    }
	}

	props.setProperty(HEURISTICS_KEY, heuristicSequence);
	props.setProperty(STEPS_KEY,  String.valueOf(getMaxSteps()));
    }

    /** sends the message that the state of this setting has been
     * changed to its registered listeners (not thread-safe)
     */
    protected void fireSettingsChanged() {
	Iterator it = listenerList.iterator();
	while (it.hasNext()) {

	    ((SettingsListener)it.next()).settingsChanged(new GUIEvent(this));
	}
    }

    /** adds a listener to the settings object 
     * @param l the listener
     */
    public void addSettingsListener(SettingsListener l) {
	listenerList.add(l);
    }


}
