// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collections;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.SLListOfGoal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.rule.strategy.proofevent.RuleAppInfo;
import de.uka.ilkd.key.rule.strategy.proofevent.NodeChangeJournal;
import de.uka.ilkd.key.proof.Node;

public class InteractiveProver implements java.io.Serializable {

    /** the proof the interactive prover works on */
    private Proof proof;

    /** the user focused goal */
    private Goal focusedGoal;

    /** true iff in interactive mode */

    private boolean interactive = true;

    /** 
     * limits the maximal amount heuristics application before an
     * interaction is needed. 
     */
    private int maxHeuristicsSteps = ProofSettings.THE_SETTINGS
	                             .getHeuristicSettings().getMaxSteps();

    /** the central heuristics processor including GUI signalling */
    private ApplyHeuristics applyHeuristics;
    // list of proof listeners and interactive proof listeners
    private List listenerList = Collections.synchronizedList(new ArrayList(5));
    /** a cached ProofEvent that is fired if heuristics have been
     * started or stopped
     */
    private final ProofEvent stockProofEvent = new ProofEvent(this);

    /** listens to the current selected proof and node */
    private KeYSelectionListener selListener;

    /** the mediator */
    private KeYMediator mediator;
    
    private boolean autoHeuristicsOn = false;

    /** 
     * mode of automatic update simplification: true: runs in implicit mode,
     * this means no new proof nodes are created for simplification steps
     */
    private boolean implicitUpdateSimplification = ProofSettings.THE_SETTINGS.
	getSimultaneousUpdateSimplifierSettings().implicitExecution();
    
    

    /** creates a new interactive prover object 
     */
    public InteractiveProver(KeYMediator mediator) {
	selListener = new InteractiveProverKeYSelectionListener();
	this.mediator = mediator;
	mediator.addKeYSelectionListener(selListener);
    }

    /** returns the KeYMediator */
    KeYMediator mediator() {
	return mediator;
    }

    /** sets up a new proof 
     * @param p a Proof that contains the root of the proof tree
     */
    public void setProof(Proof p) {
	proof = p;
    }

    /** 
     * sets the update simplification mode   
     * @param implicit a boolean that if it is true, inhibits the creation of
     * proof nodes for update simplification steps
     */
    public void setUpdateSimplificationMode(boolean implicit) {
	this.implicitUpdateSimplification = implicit;
    }

    /**
     * @return true iff updates should be simplified implicitly
     */
    public boolean getUpdateSimplificationMode () {
	return implicitUpdateSimplification;
    }

    /** set maximal amount of heuristics steps before an interactive
     * one is needed
     */
    public void setMaxHeuristicSteps(int steps) {
	maxHeuristicsSteps = steps;
    }

    /** returns the maximal amount of heuristics steps before an interactive
     * one is needed
     * @return  the maximal amount of heuristics steps before an interactive
     * one is needed
     */
    public int getMaxHeuristicSteps() {
	return maxHeuristicsSteps;
    }

    /** adds an interactive ProofListener
     * @param p the InteractiveProofListener to be added
     */
    public synchronized void addInteractiveProofListener
	(InteractiveProofListener p) { 
	synchronized(listenerList) {
	    listenerList.add(p);
	}
    }

    /** removes an interactive ProofListener
     * @param p the InteractiveProofListener to be removed
     */
    public synchronized void removeInteractiveProofListener
	(InteractiveProofListener p) { 
	synchronized(listenerList) {	
	    listenerList.remove(p);
	}
    }

    /** fires the event that automatic execution has started */
    protected synchronized void fireAutoModeStarted(ProofEvent e) {
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((InteractiveProofListener)it.next()).
		    autoModeStarted(e);
	    }
	}
    }

    /** fires the event that automatic execution has stopped */
    protected synchronized void fireAutoModeStopped(ProofEvent e) {
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((InteractiveProofListener)it.next()).
		    autoModeStopped(e);
	    }
	}
    }
    
    

    public void setAutoHeuristics(boolean b) {
       autoHeuristicsOn = b;
    }
 

    public boolean autoHeuristics() {
	return autoHeuristicsOn;
    }

    public boolean interactiveMode() {
	return interactive;
    }

    public void setInteractive(boolean interact) {
	interactive=interact;
    }

    /** fires the event that a rule has been applied */
    protected synchronized void fireRuleApplied( ProofEvent p_e ) {
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((InteractiveProofListener)it.next()).
		    ruleApplied(p_e);
	    }
	}
    }    


    /** The rule application is performed on the given goal
     * @param app the RuleApp to apply 
     * @param goal the goal where to apply the rule application
     */
    public void apply(RuleApp app, Goal goal) {
	apply ( app, goal, true,  interactive, activeHeuristics());
    }

    public void apply(RuleApp app, Goal goal, boolean applyAutoHeuristics) {
	apply( app, goal, applyAutoHeuristics, interactive, activeHeuristics());
    }


    public void apply(RuleApp app, Goal goal, boolean applyAutoHeuristics,
                      boolean interactive, ListOfHeuristic activeHeur) {
	Debug.out(" >>>>> INTERACTIVE APPLICATION <<<< ", app);
	Debug.outIfEqual("auto_insert_eq applied", 
			 app.rule().name(), "auto_insert_eq");

	JavaInfo javaInfo = mediator().getJavaInfo();
	Node     oldNode  = goal.node ();

        oldNode.setAppliedHeurInfo(interactive, activeHeur);

	if (app instanceof TacletApp) {
	    app = ((TacletApp)app).instantiateWithMV ( mediator () );
	    app = ((TacletApp)app).createDependingFunctions 
		( mediator ().func_ns () );
	} 

	NodeChangeJournal journal     = new NodeChangeJournal ( getProof (),
								goal );

	ListOfGoal        goalList    = app.execute(goal, javaInfo); 

	if (app instanceof BuiltInRuleApp) {
	    // A goal specific rule application does not always
	    // return a result
	    BuiltInRuleApp gsApp = (BuiltInRuleApp)app;
	    if (!gsApp.success()) { 
		if (!"".equals(gsApp.errorMessage())) {
		    mediator().popupInformationMessage
			(gsApp.errorMessage(), "Note...");
		}
		return;
	    }
	}
	// application was successful

	if ( goalList == SLListOfGoal.EMPTY_LIST ) {
	    proof.closeGoal ( goal, app.constraint () );	    	 
	}
	else {
	    proof.replace   ( goal, goalList );
	    if ( app instanceof TacletApp &&
		 ((TacletApp)app).taclet ().closeGoal () )
		// the first new goal is the one to be closed
		proof.closeGoal ( goalList.head (),
				  app.constraint () );
	}


	RuleAppInfo ruleAppInfo = journal.getRuleAppInfo ( app );

	fireRuleApplied ( new ProofEvent ( this, ruleAppInfo ) );

	if ( interactive && !getProof ().closed () )
	    startHeuristics ( !( applyAutoHeuristics && autoHeuristicsOn ) );
    }


    /** returns the proof the interactive prover is working on
     * @return the proof the interactive prover is working on
     */
    public Proof getProof() {
	return proof;
    }    

    /** returns active heuristics 
     */
    public ListOfHeuristic activeHeuristics() {
	return proof.activeHeuristics();
    }

    /** starts the execution of rules with active heuristics 
     * @param maxApplicationOfHeuristics maximal allowed heuristcal
     * applications of Taclets
     */
    public void startHeuristics ( boolean simplifierOnly ) {
	startHeuristics( proof.openGoals (),
			 simplifierOnly );
    }

    public void startHeuristics () {
	startHeuristics ( false );
    }


    /** starts the execution of rules with active heuristics. The
     * heuristics will only be applied on the goals of the list that
     * is handed over and on the new goals an applied rule adds
     */
    public void startHeuristics ( ListOfGoal goals,
				  boolean    simplifierOnly ) {
	if (activeHeuristics() == SLListOfHeuristic.EMPTY_LIST) {
	    mediator().popupWarning("No heuristics selected", "Oops...");
           return;
        }
	if (goals == SLListOfGoal.EMPTY_LIST) { 
	    mediator().popupWarning("No (more) goals", "Oops...");
	    return;
	}

	int steps;

	if ( simplifierOnly ) {
	    // only apply the update simplifier
	    mediator ().getHeuristicStrategy ().setSimplifierOnly ( true );
	    steps = 100000;
	} else {
	    // standard heuristic application
	    mediator ().getHeuristicStrategy ().setSimplifierOnly ( false );
	    steps = getMaxHeuristicSteps ();
	}	

	applyHeuristics = new ApplyHeuristics(this, proof,
					      mediator ().getHeuristicStrategy (),
					      steps);
	applyHeuristics.start(goals);
    }

    public void startHeuristics ( ListOfGoal goals ) {
	startHeuristics ( goals, false );
    }


    /**
     * collects all built-in rules that are applicable at the
     * given sequent position 'pos'.
     * @param pos the PosInSequent where to look for applicable rules 
     * @param userConstraint the user defined constraint
     */
    public ListOfBuiltInRule getBuiltInRule(PosInSequent pos, 
						 Constraint   userConstraint) {
	ListOfBuiltInRule rules = SLListOfBuiltInRule.EMPTY_LIST;
	IteratorOfRuleApp it = 
	    focusedGoal.ruleAppIndex().getBuiltInRule
	    (focusedGoal, pos.getPosInOccurrence(), userConstraint).iterator();

	while (it.hasNext()) {
	    BuiltInRule r = (BuiltInRule) it.next().rule();
	    if (!rules.contains(r)) {
		rules = rules.prepend(r);
	    }
	}
	return rules;
    }

    /** 
     * collects all built-in rule applications for the given
     * rule that are applicable at position 'pos' and the current user
     * constraint
     * @param rule the BuiltInRule for which the applications are
     * collected 
     * @param pos the PosInSequent the position information
     * @param userConstraint the user defined constraint
     * @return a SetOfRuleApp with all possible rule applications
     */
    public SetOfRuleApp getBuiltInRuleApp(BuiltInRule rule, 
					  PosInSequent     pos,
					  Constraint       userConstraint) {
	
	SetOfRuleApp result = SetAsListOfRuleApp.EMPTY_SET;
	IteratorOfRuleApp it = focusedGoal.ruleAppIndex().
	    getBuiltInRule(focusedGoal, 
			   pos.getPosInOccurrence(),
			   userConstraint).iterator();

	while (it.hasNext()) {
	    RuleApp app = it.next();
	    if (app.rule() == rule) {
		result = result.add(app);
	    }
	}

	return result;
    }
    
    
    /** true iff pos describes a term not a sequent or semisequent */
    private boolean isTerm(PosInSequent pos) {
	return !pos.isSequent() && 
	    !pos.isAntecedent() && 
	    !pos.isSuccedent();
    }

    /** collects all applicable NoFindTaclets of the current goal
     * (called by the SequentViewer)
     * @return a list of Taclets with all applicable NoFindTaclets
     */
    ListOfTaclet getNoFindTaclet() {	
	return filterTaclet(focusedGoal.ruleAppIndex().
		       getNoFindTaclet(interactive,
				       activeHeuristics(), 
				       mediator.getJavaInfo(),
				       mediator.getUserConstraint().getConstraint()));
    }    

    /** collects all applicable FindTaclets of the current goal
     * (called by the SequentViewer)
     * @return a list of Taclets with all applicable FindTaclets
     */

    ListOfTaclet getFindTaclet(PosInSequent pos) {
	if (isTerm(pos) && (focusedGoal!=null) && (pos!=null)) {
	    return filterTaclet(focusedGoal.ruleAppIndex().
			      getFindTaclet(interactive,
					    pos.getPosInOccurrence(),
					    activeHeuristics(),
					    mediator.getJavaInfo(),
					    mediator.getUserConstraint().getConstraint()));
	}
	return SLListOfTaclet.EMPTY_LIST;
    }
    
    /** collects all applicable RewriteTaclets of the current goal
     * (called by the SequentViewer)
     * @return a list of Taclets with all applicable RewriteTaclets
     */
    ListOfTaclet getRewriteTaclet(PosInSequent pos) {
	if (isTerm(pos))  {
	    return filterTaclet(focusedGoal.ruleAppIndex().
		   getRewriteTaclet(interactive,
				    pos.getPosInOccurrence(),
				    activeHeuristics(),
				    mediator.getJavaInfo(),
				    mediator.getUserConstraint().getConstraint())); 
	}

	return SLListOfTaclet.EMPTY_LIST;
    }

    /**
     * takes NoPosTacletApps as arguments and returns a duplicate free list of
     * the contained Taclets
     */
    private ListOfTaclet filterTaclet(ListOfNoPosTacletApp tacletInstances) {
	java.util.HashSet applicableRules = new java.util.HashSet();
	ListOfTaclet result = SLListOfTaclet.EMPTY_LIST;
	IteratorOfNoPosTacletApp it = tacletInstances.iterator();		
	while (it.hasNext()) {
	    TacletApp app = it.next ();
	    // for the moment, just remove taclets which are
	    // inconsistent with user constraint
	    if ( mediator ().getUserConstraint ().getConstraint ()
		 .join ( app.constraint () ).isSatisfiable () ) {
		Taclet taclet = app.taclet();
		if (!applicableRules.contains(taclet)) {
		    applicableRules.add(taclet);
		    result = result.prepend(taclet);
		}
	    }
	}
	
       	return result;
    }


    /** collects all Taclet applications at the given position of the specified
     * taclet
     * @param pos the PosInOccurrence describing the position
     * @param taclet the Taclet
     * @return a list of all found rule applications of the given rule at
     * position pos  
     */
    protected SetOfTacletApp getAppsForName(Goal goal, String name, 
                                            PosInOccurrence pos) {
        return getAppsForName(goal, name, pos, interactive, activeHeuristics());
    }


    protected SetOfTacletApp getAppsForName(Goal goal, String name, 
                                            PosInOccurrence pos,
                                            boolean interactive,
                                            ListOfHeuristic heur) {


	SetOfTacletApp result = SetAsListOfTacletApp.EMPTY_SET;  
        ListOfTacletApp fittingApps = 
           goal.ruleAppIndex().getTacletAppAt(heur, interactive, pos, 
			                      mediator.getJavaInfo(),
					      mediator.getUserConstraint().getConstraint());
	IteratorOfTacletApp it = fittingApps.iterator();
	// filter fitting applications
	while (it.hasNext()) {
	    TacletApp app = it.next();
	    if (app.rule().name().toString().equals(name) ){
		result = result.add(app);
	    }
	}
	return result;
    }
    
    /** listener to KeYSelection Events in order to be informed if the
     * current proof changed
     */
    private class InteractiveProverKeYSelectionListener 
	implements KeYSelectionListener {

	/** focused node has changed */
	public void selectedNodeChanged(KeYSelectionEvent e) {
	    focusedGoal = e.getSource().getSelectedGoal();
	}

	/** the selected proof has changed (e.g. a new proof has been
	 * loaded) */ 
	public void selectedProofChanged(KeYSelectionEvent e) {
	    Debug.out("InteractiveProver: initialize with new proof");
	    setProof(e.getSource().getSelectedProof());
	    focusedGoal = e.getSource().getSelectedGoal();
	}

    }
}
