// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.util.Debug;
import javax.swing.*;
import javax.swing.event.EventListenerList;
import java.awt.*;
import java.awt.event.*;
import java.util.Enumeration;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.rule.strategy.StrategyManager;

public class KeYMediator {
    
    private Main mainFrame;
    
    
    private InteractiveProver interactiveProver;

    /** the update simplifier (may be moved to nodes)*/
    private SimultaneousUpdateSimplifier upd_simplifier;

    /** the notation info used to print sequents */
    private final NotationInfo notationInfo;

    /** class handling the abbreviation of skolem function */
    private SkolemAbbreviator skolemAbbreviator;

    /** the proof the mediator handles with */
    private Proof proof;

    /** listenerList with to gui listeners */  
    private EventListenerList listenerList = new EventListenerList();
        
    /** listens to the proof */
    private KeYMediatorProofListener proofListener;

    /** listens to the ProofTree */
    private KeYMediatorProofTreeListener proofTreeListener;

    /** current proof and node the user works with. All user
     * interaction is relative to this model
     */
    private KeYSelectionModel keySelectionModel;


    /** creates the KeYMediator with a reference to the applications
     * main frame and the current proof settings 
    */
    public KeYMediator(Main mainFrame) {	
	this.mainFrame = mainFrame;
	notationInfo = NotationInfo.createInstance();
	proofListener       = new KeYMediatorProofListener();
	proofTreeListener   = new KeYMediatorProofTreeListener();
	keySelectionModel   = new KeYSelectionModel();
	interactiveProver   = new InteractiveProver(this);	
	skolemAbbreviator   = new SkolemAbbreviator(this);
	interactiveProver.addInteractiveProofListener(proofListener);	
    }

    /** returns the used NotationInfo 
     * @return the used NotationInfo 
     */
    public NotationInfo getNotationInfo() {
	return notationInfo;
    }

    /** returns the variable namespace 
     * @return the variable namespace 
     */
    public Namespace var_ns() {
	return getProof().getNamespaces().variables();
    }

    /** returns the function namespace 
     * @return the function namespace 
     */
    public Namespace func_ns() {
	return getProof().getNamespaces().functions();
    }

    /** returns the sort namespace 
     * @return the sort namespace 
     */
    public Namespace sort_ns() {
	return getProof().getNamespaces().sorts();
    }

    /** returns the heuristics namespace 
     * @return the heuristics namespace 
     */
    public Namespace heur_ns() {
	return getProof().getNamespaces().heuristics();
    }

    /** returns the JavaInfo with the java type information */
    public JavaInfo getJavaInfo() {
       return getProof().getJavaInfo();
    }

    /** returns the user constraint (table model)
     * @return the user constraint
     */
    public ConstraintTableModel getUserConstraint() {
	if ( getProof () == null )
	    return null;
	else
	    return getProof ().getUserConstraint ();
    }

    /** @return Deliverer of new metavariables (with unique names)*/
    public MetavariableDeliverer getMetavariableDeliverer () {
	if ( getProof () == null )
	    return null;
	else
	    return getProof ().getMetavariableDeliverer ();
    }

//     /** returns the proof settings
//      * @return  the proof settings
//      */
//     public ProofSettings proofSettings() {
// 	return proofSettings;
//     }
   

    /** @return the strategy manager for automatic rule application */
    public StrategyManager getHeuristicStrategy () {
	if ( getProof () == null )
	    return null;
	else
	    return getProof ().getHeuristicStrategy ();
    }

    public boolean ensureProofLoaded() {
	if (proof == null) {
	    popupWarning("No proof.", "Oops...");
	    return false;
	}
	return true;
    }
    
    /** Undo.
     * @author VK
     */
    public void setBack() {
	if (ensureProofLoaded()) {
	    setBack(getSelectedGoal());
	}
    }

    public void setBack(Node node) {
	if (ensureProofLoaded()) {
	    if (!proof.setBack(node)) {
		popupWarning("Setting back at the chosen node is not possible.",
			     "Oops...");	    
	    }
	}
    }    
    
    public void setBack(Goal goal) {
	if (ensureProofLoaded()) {
	    if (proof == null || !proof.setBack(goal))
		popupWarning("Setting back the current goal is not possible.", 
			     "Oops...");
	}
    }

    /** 
     * initializes proof (this is Swing thread-safe) 
     */
    public void setProof(Proof p) {
        final Proof pp = p;    
        if (SwingUtilities.isEventDispatchThread()) {
	    setProofHelper(pp);
        } else {
            Runnable swingProzac = new Runnable() {
               public void run() { setProofHelper(pp); }
            };
	    invokeAndWait(swingProzac);
        }        
    }


    private void setProofHelper(Proof p) {
	if (proof != null) {
	    proof.removeProofTreeListener(proofTreeListener);
	}
	proof = p;
	keySelectionModel.setSelectedProof(proof);
	if (proof != null) {
	    proof.addProofTreeListener(proofTreeListener);
	    StrategyManager sm = getHeuristicStrategy ();
	    if ( sm != null )
		sm.setMediator ( this );
	}
	proof.setSimplifier(upd_simplifier);
    }
    

    /**
     * Get the interactive prover.
     */
    public InteractiveProver getInteractiveProver() {
	return interactiveProver;
    }




    public Proof getProof() {
	return proof;
    }
    
    
    /** 
     * sets the update simplification mode   
     * @param implicit a boolean that if it is true, inhibits the creation of
     * proof nodes for update simplification steps
     */
    public void setUpdateSimplificationMode(boolean implicit) {
	interactiveProver.setUpdateSimplificationMode(implicit);    
    }

    /** sets the maximum amount of heuristics applied before the next
     * step has to be interactive
     * @param steps an int setting the limit
     */
    public void setMaxHeuristicSteps(int steps) {
	interactiveProver.setMaxHeuristicSteps(steps);
    }

    /** returns the maximal amount of heuristics steps before an interactive
     * one is needed
     * @return  the maximal amount of heuristics steps before an interactive
     * one is needed
     */
    public int getMaxHeuristicSteps() {
	return interactiveProver.getMaxHeuristicSteps();
    }


    public SetOfTacletApp getTacletApplications(Goal g, String name,
                                                PosInOccurrence p) {
       return interactiveProver.getAppsForName(g, name, p);
    }


    public SetOfTacletApp getTacletApplications(Goal            goal, 
						String          name,
                                                PosInOccurrence pos,
                                                boolean         interactive,
                                                ListOfHeuristic heur) {
       return interactiveProver.getAppsForName(goal, name, pos, 
					       interactive, heur);
    }


    /** selected rule to apply; opens a dialog 
     * @param taclet the selected Taclet
     * @param pos the PosInSequent describes the position where to apply the 
     * rule 
     */
    public void selectedTaclet(Taclet taclet, PosInSequent pos) {
	Goal goal = keySelectionModel.getSelectedGoal();
	if (goal == null) {
	    Debug.assertTrue(false);
	}
        selectedTaclet(taclet, goal, pos.getPosInOccurrence());
    }

    public boolean selectedTaclet(Taclet taclet, Goal goal, PosInOccurrence pos) {
	SetOfTacletApp applics = 
           getTacletApplications(goal, taclet.name().toString(), pos);
        if (applics.size() == 0) {
	   JOptionPane.showMessageDialog(mainFrame, "Taclet application failed." 
					 + "\n" + taclet.name(), "Bummer!",
					 JOptionPane.ERROR_MESSAGE);
           return false;
        }
	IteratorOfTacletApp it = applics.iterator();	
	Namespace varNamespace=goal.getVariableNamespace(var_ns());
	if (applics.size() == 1) {
	    TacletApp firstApp = it.next();
	    if (firstApp.taclet().ifSequent() != Sequent.EMPTY_SEQUENT ||
		! firstApp.complete()) {
		new TacletMatchCompletionDialog(
                    new ApplyTacletDialogModel(
                        firstApp, goal.sequent(), getJavaInfo(),
		        getUserConstraint ().getConstraint(),
		        varNamespace, func_ns(), sort_ns(), 
                        notationInfo.getAbbrevMap()),
                    goal, this);   
	    } else {
		apply(firstApp, goal);
	    }
	} else if (applics.size() > 1) {
	    ApplyTacletDialogModel[] applicationModel = 
		new ApplyTacletDialogModel[applics.size()];

	    for (int i = 0; i < applicationModel.length; i++) {
		applicationModel[i] = new ApplyTacletDialogModel(
		    it.next(), goal.sequent(), getJavaInfo(), 
                    getUserConstraint().getConstraint(),
		    varNamespace, func_ns(), sort_ns(), 
                    notationInfo.getAbbrevMap());
	    }	    

	    TacletMatchCompletionDialog compDlg =
		new TacletMatchCompletionDialog(applicationModel, goal, this);
	}
        return true;
    }


    /** selected rule to apply
     * @param rule the selected built-in rule
     * @param pos the PosInSequent describes the position where to apply the
     * rule 
     */
    public void selectedBuiltInRule(BuiltInRule rule, PosInSequent pos) {
	Goal goal = keySelectionModel.getSelectedGoal();
	Debug.assertTrue(goal != null);

	SetOfRuleApp set = interactiveProver.
	    getBuiltInRuleApp(rule, pos, getUserConstraint().getConstraint());
	if (set.size() > 1) {
	    System.err.println("keymediator:: Expected a single app. If " +
		      "it is OK that there are more than one " +
		      "built-in rule apps. You have to add a " +
		      "selection dialog here");
	    System.err.println("keymediator:: Ambigous applications, " +
		      "taking the first in list.");
	}

	RuleApp app = set.iterator().next();
	if (app != null && app.rule() == rule) {
		interactiveProver.apply(app, goal);
		return;
	}
    }
                    
        
    /** performs the application of the given RuleApp
     * @param app the RuleApp to apply 
     * @param goal the Goal where to apply the rule
     */
    public void apply(RuleApp app, Goal goal) {
	interactiveProver.apply(app, goal);
    }
    
    
    public void apply(RuleApp app, Goal goal, 
                      boolean interactive, ListOfHeuristic avctiveHeuristic) {
	interactiveProver.apply(app, goal, false, interactive, avctiveHeuristic);
    }
    

    /** collects all applicable FindTaclets of the current goal
     * (called by the SequentViewer)
     * @return a list of Taclets with all applicable FindTaclets
     */

    protected ListOfTaclet getFindTaclet(PosInSequent pos) {
	return interactiveProver.getFindTaclet(pos);
    }

    /** collects all applicable RewriteTaclets of the current goal
     * (called by the SequentViewer)
     * @return a list of Taclets with all applicable RewriteTaclets
     */
    protected ListOfTaclet getRewriteTaclet(PosInSequent pos) {
	return interactiveProver.getRewriteTaclet(pos);    
    }

    /** collects all applicable NoFindTaclets of the current goal
     * (called by the SequentViewer)
     * @return a list of Taclets with all applicable NoFindTaclets
     */
    protected ListOfTaclet getNoFindTaclet() {	
	return interactiveProver.getNoFindTaclet();
    }

    /** collects all built-in rules 
     * @return a list of all applicable built-in rules 
     */
    public ListOfBuiltInRule getBuiltInRule(PosInSequent pos) {
	return interactiveProver.getBuiltInRule
	    (pos, getUserConstraint().getConstraint());
    }

    /** adds a listener to the KeYSelectionModel, so that the listener
     * will be informed if the proof or node the user has selected
     * changed 
     * @param listener the KeYSelectionListener to add
     */
    public synchronized void addKeYSelectionListener(KeYSelectionListener listener) {
	keySelectionModel.addKeYSelectionListener(listener);
    }

    /** removes a listener from the KeYSelectionModel
     * @param listener the KeYSelectionListener to be removed
     */
    public synchronized void removeKeYSelectionListener(KeYSelectionListener listener) {
	keySelectionModel.removeKeYSelectionListener(listener);
    }

    /** adds a listener to GUI events 
     * @param listener the GUIListener to be added
     */
    public void addGUIListener(GUIListener listener) {
	listenerList.add(GUIListener.class, listener);
    }

    /** adds a listener to GUI events 
     * @param listener the GUIListener to be added
     */
    public void removeGUIListener(GUIListener listener) {
	listenerList.remove(GUIListener.class, listener);
    }

    /** adds an InteractiveProofListener 
     * @param listener the InteractiveProofListener to be added
     */
    public synchronized void 
	addInteractiveProofListener(InteractiveProofListener listener) {
	interactiveProver.addInteractiveProofListener(listener);
    }

    /** adds an InteractiveProofListener 
     * @param listener the InteractiveProofListener to be added
     */
    public synchronized void 
	removeInteractiveProofListener(InteractiveProofListener listener) {  
	interactiveProver.removeInteractiveProofListener(listener);
    }

    /** sets the current goal 
     * @param goal the Goal being displayed in the view of the sequent
     */
    public void goalChosen(Goal goal) {
	keySelectionModel.setSelectedGoal(goal);
    }

    /** returns the main frame
     * @return the main frame 
     */
    public JFrame mainFrame() {
	return mainFrame;
    }

    /** notifies that a node that is not a goal has been chosen
     * @param node the node being displayed in the view of the sequent
     */
    public void nonGoalNodeChosen(Node node) {
	keySelectionModel.setSelectedNode(node);
    }

    /** called to ask for modal access 
     * @param src Object that is the asking component 
     */
    public synchronized void requestModalAccess(Object src) {
	fireModalDialogOpened(new GUIEvent(src));	
    }

    /** called if no more modal access is needed
    * @param src Object that is the asking component 
     */
    public synchronized void freeModalAccess(Object src) {
	fireModalDialogClosed(new GUIEvent(src));	
    }
    
    /** fires the request of a GUI component for modal access
     * this can be used to disable all views even if the GUI component
     * has no built in modal support 
     */
    public synchronized void fireModalDialogOpened(GUIEvent e) {
	Object[] listeners = listenerList.getListenerList();
	for (int i = listeners.length-2; i>=0; i-=2) {
	    if (listeners[i] == GUIListener.class) {
		((GUIListener)listeners[i+1]).modalDialogOpened(e);
	    }
	}
    }

    /** fires that a GUI component that has asked for modal access
     * has been closed, so views can be enabled again
     */
    public synchronized void fireModalDialogClosed(GUIEvent e) {
	Object[] listeners = listenerList.getListenerList();
	for (int i = listeners.length-2; i>=0; i-=2) {
	    if (listeners[i] == GUIListener.class) {
		((GUIListener)listeners[i+1]).modalDialogClosed(e);
	    }
	}
    }

    /** Fires the shut down event.
     */
    public synchronized void fireShutDown(GUIEvent e) {
	Object[] listeners = listenerList.getListenerList();
	for (int i = listeners.length-2; i>=0; i-=2) {
	    if (listeners[i] == GUIListener.class) {
		((GUIListener)listeners[i+1]).shutDown(e);
	    }
	}
    }

    
    /** sets the simultaneous update simplifier */
    public void setSimplifier(SimultaneousUpdateSimplifier s) {
	upd_simplifier = s;
	if (proof != null) proof.setSimplifier(s);
    }
        
    /** returns the current selected proof 
     * @return the current selected proof 
     */
    public Proof getSelectedProof() {
 	return keySelectionModel.getSelectedProof();
    }

    /** returns the current selected goal 
     * @return the current selected goal 
     */
    public Goal getSelectedGoal() {
 	return keySelectionModel.getSelectedGoal();
    }

    /** returns the current selected node
     * @return the current selected node 
     */
    public Node getSelectedNode() {
 	return keySelectionModel.getSelectedNode();
    }
    
    public void startHeuristics() {
	if (ensureProofLoaded()) {
	    startHeuristics(proof.openGoals());
	}
    }

    public void startHeuristics(ListOfGoal goals) {
       interactiveProver.startHeuristics(goals);
    }
    
    public void setAutoHeuristics(boolean b) {
       interactiveProver.setAutoHeuristics(b);
    }

    public void popupInformationMessage(Object message, String title) {
	JOptionPane.showMessageDialog
	    (mainFrame, message,
	     title, JOptionPane.INFORMATION_MESSAGE);
    }

   public void popupWarning(Object message, String title) {
      JOptionPane.showMessageDialog(mainFrame, message, title, 
                                    JOptionPane.WARNING_MESSAGE);
   }

    /**
     * Brings up a dialog displaying a message.
     * @param modal whether or not the message should be displayed in a modal dialog.
     */
    public void popupInformationMessage(Object message, String title, boolean modal) {
	if (modal) {
	    popupInformationMessage(message, title);
	} else {
	    if (!(message instanceof Component))
		throw new InternalError("only messages of type " + Component.class + " supported, yet");
	    // JFrame dlg = new JDialog(mainFrame(),title, modal);
	    JFrame dlg = new JFrame(title);
	    dlg.setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
	    dlg.getContentPane().add((Component)message);
	    dlg.pack();
	    setCenter(dlg, mainFrame());
	    dlg.setVisible(true);
	}
    }

    /**
     * Center a component within a parental component.
     * @param comp the component to be centered.
     * @param parent center relative to what. <code>null</code> to center relative to screen.
     * @see #setCenter(Component)
     */
    public static void setCenter(Component comp, Component parent) {
	if (parent == null) {
	    setCenter(comp);
	    return;
	} 
	Dimension dlgSize = comp.getPreferredSize();
	Dimension frmSize = parent.getSize();
	Point	  loc = parent.getLocation();
	if (dlgSize.width < frmSize.width && dlgSize.height < frmSize.height)
	    comp.setLocation((frmSize.width - dlgSize.width) / 2 + loc.x, (frmSize.height - dlgSize.height) / 2 + loc.y);
	else
	    setCenter(comp);
    } 

    /**
     * Center a component on the screen.
     * @param comp the component to be centered relative to the screen.
     *  It must already have its final size set.
     * @preconditions comp.getSize() as on screen.
     */
    public static void setCenter(Component comp) {
	Dimension screenSize = comp.getToolkit().getScreenSize();
	Dimension frameSize = comp.getSize();
	if (frameSize.height > screenSize.height)
	    frameSize.height = screenSize.height;
	if (frameSize.width > screenSize.width)
	    frameSize.width = screenSize.width;
	comp.setLocation((screenSize.width - frameSize.width) / 2, (screenSize.height - frameSize.height) / 2);
    } 

    private int goalsClosedByHeuristics=0;

    public void closedAGoal() {

	if (!interactiveProver.interactiveMode()) 
	    goalsClosedByHeuristics++;
	else
	    JOptionPane.showMessageDialog 
		(mainFrame, "A goal has been closed", "Note ...", 
		 JOptionPane.INFORMATION_MESSAGE);
    }

    public int getNrGoalsClosedByHeuristics() {
	return goalsClosedByHeuristics;
    }

    public void resetNrGoalsClosedByHeuristics() {
	goalsClosedByHeuristics=0;
    }


   public void stopInterface(boolean fullStop) {
      final boolean b = fullStop;
      Runnable interfaceSignaller = new Runnable() {
         public void run() {
	     mainFrame.setCursor
		 (new java.awt.Cursor(java.awt.Cursor.WAIT_CURSOR));
            if (b) {
               interactiveProver.fireAutoModeStarted(
                  new ProofEvent(interactiveProver));
            }
         }
      };
      invokeAndWait(interfaceSignaller);
   }

   public void startInterface(boolean fullStop) {
      final boolean b = fullStop;
      Runnable interfaceSignaller = new Runnable() {
         public void run() {
            if (b) {
               interactiveProver.fireAutoModeStopped(
                  new ProofEvent(interactiveProver));
            }
	    mainFrame.setCursor
		(new java.awt.Cursor(java.awt.Cursor.DEFAULT_CURSOR));
         }
      };
      invokeAndWait(interfaceSignaller);
   }


    private void invokeAndWait(Runnable runner) {
      try{
         SwingUtilities.invokeAndWait(runner);
      } catch(InterruptedException e) {
	  System.err.println(e);
	  e.printStackTrace();
      } catch(java.lang.reflect.InvocationTargetException ite) {
         System.err.println(ite);
         ite.printStackTrace();
      }
    }


    class KeYMediatorProofTreeListener extends ProofTreeAdapter 
                                       implements java.io.Serializable {
	public void proofClosed(ProofTreeEvent e) {
	    if (interactiveProver.interactiveMode()) {
		JOptionPane.showMessageDialog 
		    (mainFrame, "Proved", "Proof Completed", 
		     JOptionPane.INFORMATION_MESSAGE);
	    }       
	}

	public void proofPruned(ProofTreeEvent e) {
	    ProofTreeRemovedNodeEvent ev = (ProofTreeRemovedNodeEvent) e;
	    if (ev.getRemovedNode() == getSelectedNode()) {
		keySelectionModel.setSelectedNode(e.getNode());
	    }
	}

	public void proofGoalsAdded(ProofTreeEvent e) {
	    ListOfGoal newGoals = e.getGoals();
	    // Check for a closed goal ...
	    if (newGoals.size() == 0){
		// No new goals have been generated ...

		closedAGoal();

	    }
	}
	

	public void proofStructureChanged(ProofTreeEvent e) {
	    Proof p = e.getSource();
	    if (p == getSelectedProof()) {
		Node sel_node = getSelectedNode();
		if (!(p.find(sel_node))) {
		    keySelectionModel.defaultSelection();
		} else {
		    // %%% hack does need to be done proper
		    // needed top update that the selected node nay have
		    // changed its status 
		    keySelectionModel.setSelectedNode(sel_node);
		}
	    }
	}

    }


    //INNER CLASS
    class KeYMediatorProofListener implements InteractiveProofListener,
                                              java.io.Serializable {

	/** invoked when a rule has been applied */
	public void ruleApplied(ProofEvent e) {
	    keySelectionModel.defaultSelection();	   
	}


	/** invoked if automatic execution has started
	 */
	public void autoModeStarted(ProofEvent e) {
//            if (proof == null) return; // there is no selection or anything
	}
	
	/** invoked if automatic execution has stopped
	 */
	public void autoModeStopped(ProofEvent e) {
            if (proof == null) return; // there is no selection or anything
	    keySelectionModel.defaultSelection();	    
	}
    }

    class KeYMediatorSelectionListener implements KeYSelectionListener {
	/** focused node has changed */
	public void selectedNodeChanged(KeYSelectionEvent e) {
	    // empty
	}

	/** the selected proof has changed (e.g. a new proof has been
	 * loaded) 
	 */ 
	public void selectedProofChanged(KeYSelectionEvent e) {
	    setProof(e.getSource().getSelectedProof());
	}
	
    }


}
