// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import java.awt.Dimension;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;

import de.uka.ilkd.key.logic.ldt.LDT;
import de.uka.ilkd.key.logic.ldt.NumberLDT;
import de.uka.ilkd.key.logic.ldt.IntegerLDT;
import de.uka.ilkd.key.logic.ldt.ArithIntLDTIgnoringOF;
import de.uka.ilkd.key.logic.ldt.ArithIntLDT;
import de.uka.ilkd.key.logic.ldt.BooleanLDT;
import de.uka.ilkd.key.logic.ldt.ObjectLDT;

import de.uka.ilkd.key.util.Debug;

public class LDTSelector extends JDialog {

    /** encapsulates the configuration */
    private LDTSettings settings;
    /** the list with the different parts that can be modelled */
    private String[] types = {"Natural Numbers", "Boolean", "Objects"};
    /** the list with the LDT that can be used to model the parts above
     * the first entry belongs to the first part above, the seconde to the
     * sencond and so on  
     */
    private String[][] models = {{"Java Semantics", 
				  "Arithmetic Semantics (prohibiting overflow)", 
				  "Arithmetic Semantics (ignoring overflow)"}, 
				 {"Boolean"}, {"Object"}};

    /** the JList with the types that can be modeled */
    private JList toModel;
    /** the JList with the LDT names */
    private JList ldtList;
    /** the KeYMediator */

    /** creates a new HeuristicSelector */
    public LDTSelector(KeYMediator mediator, LDTSettings settings) {  
	super(new JFrame(), "Select Datatype Model", true);
	
	this.settings = settings;
	settings.restore();
	layoutLDTSelector();
	init();

	pack();
	setLocation(70, 70);
	setVisible(true);
    }

    /** layout */
    protected void layoutLDTSelector() {
	toModel = new JList(types);
	toModel.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	toModel.addListSelectionListener(new ListSelectionListener() {
	    public void valueChanged(ListSelectionEvent e) {
		setLDTList();				
	    }});
	ldtList = new JList(models[0]);
	ldtList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	ldtList.addListSelectionListener(new ListSelectionListener() {
	    public void valueChanged(ListSelectionEvent e) {
		setLDTModel(ldtList.getSelectedIndex());				
	    }});

	JScrollPane toModelScroll = new
	    JScrollPane(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, 
			JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	toModelScroll.getViewport().setView(toModel);
	toModelScroll.setBorder(new TitledBorder("Aspect"));

	JScrollPane modelScrollPane = new 	    
	    JScrollPane(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, 
			JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
	modelScrollPane.getViewport().setView(ldtList);
	modelScrollPane.setBorder(new TitledBorder("Use LDT"));

	JButton ok = new JButton("OK");
	ok.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    int res = JOptionPane.showOptionDialog
			(LDTSelector.this,
			 "You will have to save the settings and restart KeY "+
			 "to activate the new LDT models.\n"
			 +"You can save now or later.", 
			 "LDT Settings", JOptionPane.DEFAULT_OPTION,
			 JOptionPane.QUESTION_MESSAGE, null,
                         new Object[]{"Save now", "Dismiss"}, "Save now");
		    settings.store();
		    if (res==0) {
			ProofSettings.THE_SETTINGS.saveSettings();
		    }
		    setVisible(false);
		    dispose();
		}
	    });
	JButton cancel = new JButton("Cancel");
	cancel.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    setVisible(false);
		    // restore changes
		    settings.restore();
		    dispose();
		}
	    });
 	JPanel buttonPanel = new JPanel();
 	buttonPanel.add(ok);
 	buttonPanel.add(cancel);

	Dimension paneDim = new Dimension(300, 150);
	modelScrollPane.setPreferredSize(paneDim);
	modelScrollPane.setMinimumSize(paneDim);
	paneDim = new Dimension(150, 150);
	toModelScroll.setPreferredSize(paneDim);
        toModelScroll.setMinimumSize(paneDim);

	JPanel listPanel=new JPanel();
	listPanel.setLayout(new BoxLayout(listPanel, BoxLayout.X_AXIS));
	listPanel.add(toModelScroll);
	listPanel.add(modelScrollPane);

	getContentPane().setLayout(new BoxLayout(getContentPane(), 
						 BoxLayout.Y_AXIS));
	getContentPane().add(listPanel);
	getContentPane().add(buttonPanel);
    }

    /** initializes the selector so that it conforms to the settings stored in
     * the LDTSettings object
     */
    private void init() {
	toModel.setSelectedIndex(0);
	setLDTList();
    }

    /** is called to set the chosen LDT */
    private void setLDTModel(int sel) {
	switch (toModel.getSelectedIndex()) { 
	case 0: 
	    if (sel == 1) {
		settings.setNaturalNumberLDT(new ArithIntLDT());
	    } else if (sel ==2) {
		settings.setNaturalNumberLDT(new ArithIntLDTIgnoringOF());
	    } else {
		settings.setNaturalNumberLDT(new IntegerLDT());
	    }
	    break;
	case 1: settings.setBooleanLDT(new BooleanLDT());
	    break;
	case 2: settings.setObjectLDT(new ObjectLDT());
	    break;
	}	
    }

    /** is called if the selection of left list has changed, and causes the
     * right one to display the possible LDTs that can be used to model the
     * aspect chosen on the left side
     */
    private void setLDTList() {
	ldtList.setListData(models[toModel.getSelectedIndex()]);
	LDT setLDT = null;	
	switch (toModel.getSelectedIndex()) { 
	case 0: setLDT = settings.getNaturalNumberLDT();
	    break;
	case 1: setLDT = settings.getBooleanLDT();
	    break;
	case 2: setLDT = settings.getObjectLDT();
	    break;
	}
	ldtList.setSelectedIndex(index(setLDT));
    }

    /** returns the index of the models array for the given LDT 
     * @return  the index of the models array for the given LDT      
     */
    private int index(LDT ldt) {
	if (ldt instanceof ArithIntLDT) {
	    return 1;
	} else if (ldt instanceof ArithIntLDTIgnoringOF){
	    return 2;
	} else {
	    return 0;
	}
    }
   
    

}    
