// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.gui;

import de.uka.ilkd.key.logic.ldt.LDT;
import de.uka.ilkd.key.logic.ldt.NumberLDT;
import de.uka.ilkd.key.logic.ldt.IntegerLDT;
import de.uka.ilkd.key.logic.ldt.ArithIntLDT;
import de.uka.ilkd.key.logic.ldt.ArithIntLDTIgnoringOF;
import de.uka.ilkd.key.logic.ldt.BooleanLDT;
import de.uka.ilkd.key.logic.ldt.ObjectLDT;
import de.uka.ilkd.key.util.Debug;
import java.util.Properties;
import java.util.LinkedList;
import java.util.Iterator;

/** This class encapsulates the information about the used
 * LDTs to model java data types. There are two differet LDTs available 
 * If the used LDT have been changed the new ones are stroed in <NAME>_toSet
 * these entries are written to the configuration file. The get-Methods return
 * always the current valid models.
 */
public class LDTSettings implements Settings {

    /** the key values for the properties */
    private static final String NATURAL_NUMBER_LDT = "[LDT]NaturalNumberLDT";
    private static final String BOOLEAN_LDT = "[LDT]BooelanLDT";
    private static final String OBJECT_LDT = "[LDT]ObjectLDT";

    
    /** the listener to this settings entry */
    private LinkedList listenerList = new LinkedList();
    /** the different LDTs that are set*/
    private LDT naturalNumberLDT;
    private LDT booleanLDT;
    private LDT objectLDT;
    /** the different LDTs that have to be set */
    private LDT naturalNumberLDT_toSet;
    private LDT booleanLDT_toSet;
    private LDT objectLDT_toSet;
    
    /** creates a default settings object for LDTs */
    public LDTSettings() {
	naturalNumberLDT = new ArithIntLDTIgnoringOF();
	booleanLDT = new BooleanLDT();
	objectLDT = new ObjectLDT();
	
	restore();
    }
        
    /** restores old values */
    public void restore() {
	naturalNumberLDT_toSet = naturalNumberLDT;
	booleanLDT_toSet = booleanLDT;
	objectLDT_toSet = objectLDT;
    }

    /** stores new values */
    public void store() {
	naturalNumberLDT = naturalNumberLDT_toSet;
	booleanLDT = booleanLDT_toSet;
	objectLDT = objectLDT_toSet;
    }


    /** returns the LDT used to represent natural numbers 
     * @return the LDT used to represent natural numbers 
     */
    public LDT getNaturalNumberLDT() {
	return naturalNumberLDT;
    }

    /** returns the LDT used to represent boolean 
     * @return the LDT used to represent boolean 
     */
    public LDT getBooleanLDT() {
	return booleanLDT;
    }

    /** returns the LDT used to represent boolean 
     * @return the LDT used to represent boolean 
     */
    public LDT getObjectLDT() {
	return objectLDT;
    }

    /** sets the LDT used to represent natural numbers 
     * @param ldt the LDT to use
     */
    public void setNaturalNumberLDT(LDT ldt) {	
	naturalNumberLDT_toSet = ldt;
	fireSettingsChanged();
    }

    /** sets the LDT used to represent boolean 
     * @param ldt the LDT to use
     */
    public void setBooleanLDT(LDT ldt) {
	booleanLDT_toSet = ldt;
	fireSettingsChanged();
    }

    /** sets the LDT used to handle objects
     * @param ldt the LDT to use
     */
    public void setObjectLDT(LDT ldt) {
	objectLDT_toSet = ldt;
	fireSettingsChanged();
    }

    /** gets a Properties object and has to perform the necessary
     * steps in order to change this object in a way that it
     * represents the stored settings
     */
    public void readSettings(Properties props) {
	String naturalNumberLDTName  = props.getProperty(NATURAL_NUMBER_LDT);
	String booleanLDTName  = props.getProperty(BOOLEAN_LDT);
	String objectLDTName   = props.getProperty(OBJECT_LDT);
	if ("integer".equals(naturalNumberLDTName)) {
	    naturalNumberLDT = new IntegerLDT();
	} else if ("arithmeticInteger".equals(naturalNumberLDTName)) {
	    naturalNumberLDT = new ArithIntLDT();	    
	} else {
	    naturalNumberLDT = new ArithIntLDTIgnoringOF();	    
	}
	// as there are no other possibilities
	booleanLDT = new BooleanLDT();
	objectLDT = new ObjectLDT();		
	restore();
    }

    /*
    /** The settings to store are written to the given Properties object.
     * @param props the Properties object where to write the settings as (key, value) pair
     */
    public void writeSettings(Properties props) {
	props.setProperty(NATURAL_NUMBER_LDT, naturalNumberLDT_toSet.name().toString());
	props.setProperty(BOOLEAN_LDT,  booleanLDT_toSet.name().toString());
	props.setProperty(OBJECT_LDT,  objectLDT_toSet.name().toString());
    }

    /** sends the message that the state of this setting has been
     * changed to its registered listeners (not thread-safe)
     */
    protected void fireSettingsChanged() {
	Iterator it = listenerList.iterator();
	while (it.hasNext()) {
	    ((SettingsListener)it.next()).settingsChanged(new GUIEvent(this));
	}
    }

    /** adds a listener to the settings object 
     * @param l the listener
     */
    public void addSettingsListener(SettingsListener l) {
	listenerList.add(l);
    }


}
