// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.proof.decproc.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.KeYResourceManager;
import de.uka.ilkd.key.rule.SimplifyIntegerRule;
import de.uka.ilkd.key.rule.BuiltInRuleApp;
import java.net.URL;

import javax.swing.*;
import javax.swing.event.MouseInputAdapter;
import javax.swing.border.*;
import java.awt.*;
import java.awt.Cursor;
import java.awt.event.*;

// just for loading remove if load moves to another class 
import java.io.*;

public class Main extends JFrame {

//DO NOT CHANGE THE FOLLOWING 2 LINES, UNLESS YOU KNOW ABOUT PRCS-REPLACEMENT
    /* $Format: "    private static final String PRCSVERSION = \"$ProjectVersion$\";"$ */
    private static final String PRCSVERSION = "release09.7";


    private static final String VERSION="0.9 \"Pisa\" (internal: "+PRCSVERSION+")";
    private static final String COPYRIGHT="(C) Copyright 2001-2003 Universit\u00e4t"
	+" Karlsruhe and Chalmers University of Technology";

    /**
     * The maximum number of recent files displayed.
     */
    private static final int MAX_RECENT_FILES = 8;
    /**
     * In which file to store the recent files.
     */
    private static final String RECENT_FILES_STORAGE = 
	System.getProperty("user.home")+
	 File.separator+".key"+File.separator+"recentFiles.props";

    static {
	//@xxx preliminary: better store along with other settings.
	PresentationFeatures.ENABLED = true;
    }
    
    /** the tab bar at the left */
    private JTabbedPane	tabbedPane;

    /** the toolbar */
    private JPanel toolBar;

    /** the current goal view */
    private JScrollPane goalView;

    /** the current proof tree*/
    private JScrollPane proofView;

    /** the list of current open goals*/
    private JScrollPane openGoalsView;

    /** the view of a sequent */
    private SequentView sequentView;

    /** the user constraint view */
    private UserConstraintView userConstraintView = null;

    /** the current user constraint */
    private ConstraintTableModel userConstraint = null;

    /** contains a list of all proofs */
    private JScrollPane proofListView;
    private ProofList proofList;

    /** the mediator is stored here */
    protected KeYMediator mediator;
    
    /** the status line */
    MainStatusLine statusLine;
    
    /** listener to global proof events */
    private MainProofListener proofListener;

    /** listener to gui events */
    private MainGUIListener guiListener;
    private RecentFileMenu recentFiles;

    private boolean frozen = false;

    /** listener to changes of the user constraint */
    private MainConstraintTableListener constraintListener;
    protected JMenuBar menubar;

    private JFileChooser fileChooser=null;
    
    private final String PARA = 
       "<p style=\"font-family: lucida;font-size: 12pt;font-weight: bold\">";
    
    private JCheckBox resumeHeuristics = 
       new JCheckBox("<html>"+PARA+"Autoresume"+PARA+"heuristics");
    private JButton applyHeuristicsButton = new JButton("Apply Heuristics");

    private JButton runSimplifyButton = new JButton("Run SIMPLIFY");

    /** the proofsettings */
    private ProofSettings proofSettings;

    private static String fileNameOnStartUp=null;

    /** are we in stand-alone mode? (or with TCC?) */
    public static boolean standalone=  
        System.getProperty("key.together") == null;

    protected static Main instance = null;


    protected Main() {
    }

    /** creates prover -- private, use getInstance()
     * @param title frame title
     * @param proofSettings the settings the proof uses (such as active
     * heuristics, LDTs etc.) 
     */ 
    protected Main(String title, ProofSettings settings) {
	super(title);
	setIconImage(IconFactory.keyLogo());
	setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
	proofSettings      = settings;	
	proofListener      = new MainProofListener();
	guiListener        = new MainGUIListener();
	constraintListener = new MainConstraintTableListener();
	layoutMain();
        setMediator(new KeYMediator(this));
//	Config.DEFAULT.setDefaultFonts();
	initGoalList();
        initProofList();
	initGUIProofTree();
	SwingUtilities.updateComponentTreeUI(this);
	setSize(700,500);
	addWindowListener(new MainListener());
        setVisible(true);
    }
    
    
    public static Main getInstance() {
        if (instance == null) instance = 
            new Main("KeY -- Prover", ProofSettings.THE_SETTINGS);
        if (!instance.isVisible()) instance.setVisible(true); //XXX: enough?
        return instance;
    }

    /** returns the proof settings 
     * @return the proof settings 
     */
    public ProofSettings proofSettings() {
	return proofSettings;
    }

    public void updateUI() {
	if (goalView != null) goalView.updateUI();
	if (proofView != null) proofView.updateUI();
	if (openGoalsView != null) openGoalsView.updateUI();
	if (userConstraintView != null) userConstraintView.updateUI();
    	if (proofListView != null) proofListView.updateUI();
    }

    /** sets the mediator to corresspond with other gui elements
     * @param m the KeYMediator
     */
    private void setMediator(KeYMediator m) {
	if (mediator !=  null) unregister();
	mediator = m;
	mediator.setSimplifier(
            proofSettings.getSimultaneousUpdateSimplifierSettings().getSimplifier());
	sequentView = new SequentView(mediator);
	if ( userConstraintView != null ) userConstraintView.setMediator(m);
	register();
    }

    /** register several listeners */
    private void register() {	
	mediator.addKeYSelectionListener(proofListener);
	mediator.addInteractiveProofListener(proofListener);
	mediator.addGUIListener(guiListener);	
    }

    /** unregister several listeners */
    private void unregister() {
	mediator.removeKeYSelectionListener(proofListener);
	mediator.removeInteractiveProofListener(proofListener);
	mediator.removeGUIListener(guiListener);	
    }
   

    /** return the mediator
     * @return the mediator
     */
    public static KeYMediator mediator() {
	return instance.mediator;
    }

    /** paints empty view */
    private void paintEmptyViewComponent(JScrollPane pane, String name) { 	
	pane.setBorder(new TitledBorder(name));
	pane.setBackground(Color.white);
	pane.getViewport().setBackground(Color.white);	
	pane.setMinimumSize(new Dimension(150,100));
    }

    /** adds a new tab */
    public void addTab(String tabname, JScrollPane view, String description) {
	tabbedPane.addTab(tabname, null, view, description);
    }

    /** lays out the main frame */
    protected void layoutMain() {
	setJMenuBar(createMenuBar());
	getContentPane().setLayout(new BorderLayout());	

	goalView = new JScrollPane();	
	paintEmptyViewComponent(goalView, "Current Goal");

	proofView = new JScrollPane();
	paintEmptyViewComponent(proofView, "Proof");

	openGoalsView = new JScrollPane();
	paintEmptyViewComponent(openGoalsView, "Open Goals");

	userConstraintView = new UserConstraintView ();

	tabbedPane = new JTabbedPane();

	addTab("Proof", proofView, 
	       "The current state of the "
	       +"proof as tree");

	addTab("Goals", openGoalsView, 
	       "The currently open goals");

	tabbedPane.addTab("User Constraint", null, userConstraintView, 
			  "Currently choosen metavariable instantiations");
	tabbedPane.setSelectedIndex(0);
	tabbedPane.setPreferredSize ( new Dimension ( 100, 400 ) );

	JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
					      tabbedPane, goalView) {
            public void setUI(javax.swing.plaf.SplitPaneUI ui) {
                try{ super.setUI(ui); } catch(NullPointerException e) { }
            }
        }; // work around bug in 
        // com.togethersoft.util.ui.plaf.metal.OIMetalSplitPaneUI
	splitPane.setResizeWeight(0); // the right pane is more important
	getContentPane().add(splitPane, BorderLayout.CENTER);
	
	statusLine = new MainStatusLine
	    ("<html>"+PARA+COPYRIGHT+
	     PARA+
	     "KeY is free software and comes with ABSOLUTELY NO WARRANTY."+
	     " See Help | License.",
	     getFont () );
        getContentPane().add(statusLine, BorderLayout.SOUTH);
        
        toolBar = new JPanel();
	toolBar.add(applyHeuristicsButton);
	toolBar.add(resumeHeuristics);
	toolBar.add(runSimplifyButton);
	toolBar.setLayout(new FlowLayout(FlowLayout.LEFT));
	getContentPane().add(toolBar, BorderLayout.NORTH);
        resumeHeuristics.addItemListener(new ItemListener() {
           public void itemStateChanged(ItemEvent e) {
              if (e.getStateChange() == ItemEvent.SELECTED) {
                 mediator().setAutoHeuristics(true);
              } 
              else
              if (e.getStateChange() == ItemEvent.DESELECTED) {
                 mediator().setAutoHeuristics(false);
              }
              else System.err.println("Heuristics checkbox undefined state: "+
                                      e.getStateChange());
              
           }
        });
        applyHeuristicsButton.addActionListener(new ActionListener() {
           public void actionPerformed(ActionEvent e) {
		      mediator().startHeuristics();
	   }
        });	

        runSimplifyButton.addActionListener(new ActionListener() {
           public void actionPerformed(ActionEvent e) {
	       if (!mediator().ensureProofLoaded()) return;
	       int counter = 0;
	       IteratorOfGoal i = mediator.getProof().openGoals().iterator();
	       while (i.hasNext()) {
		   // ConstraintTableModel mediator.getUserConstraint()
		   SimplifyIntegerRule simplifyTaclet = new SimplifyIntegerRule(false);

		   BuiltInRuleApp birApp = new BuiltInRuleApp(simplifyTaclet, 
			null, mediator.getUserConstraint().getConstraint());
		   mediator.getInteractiveProver().apply ( birApp,
							  i.next(),
							  false );

	       }
	   }
        });	
        
        JButton goalBackButton = new JButton("Goal Back");
	toolBar.add(goalBackButton);
        goalBackButton.addActionListener(new ActionListener() {
	   public void actionPerformed(ActionEvent e) {
	       setBack();
	   }
        });
	goalView.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW ).put(
           KeyStroke.getKeyStroke(KeyEvent.VK_C, ActionEvent.CTRL_MASK), 
           "copy");
        goalView.getActionMap().put("copy", new AbstractAction() {
           public void actionPerformed(ActionEvent e) {
              copyHighlightToClipboard(sequentView);
           }
        });

    }

    /**
     * @return the status line object
     */
    public MainStatusLine getStatusLine () {
	return statusLine;
    }

    private void setStandardStatusLineImmediately () {
	statusLine.reset ();
    }

    /**
     * Make the status line display a standard message, make progress
     * bar and abort button invisible
     */
    public void setStandardStatusLine() {
	if ( SwingUtilities.isEventDispatchThread () )
	    setStandardStatusLineImmediately ();
	else {
	    Runnable lineUpdater = new Runnable() {
		    public void run() {
			setStandardStatusLineImmediately ();
		    }
		};
	    SwingUtilities.invokeLater(lineUpdater);
	}
    }

    private void setStatusLineImmediately(String s) {
	statusLine.reset ();
	statusLine.setStatusText(s);
	statusLine.setProgressPanelVisible(false);
    }

    private void setStatusLineImmediately(String s, int totalChars) {
	statusLine.reset ();
	statusLine.setStatusText(s);
	statusLine.setProgressBarMaximum(totalChars);
	statusLine.setProgressPanelVisible(true);
	statusLine.setAbortButtonEnabled(false);
    }

    /**
     * Display the given message in the status line, make progress bar
     * and abort button visible, set the progress bar range to the
     * given value, set the current progress to zero
     */    
    public void setStatusLine(String s, int totalChars) {
	if ( SwingUtilities.isEventDispatchThread () )
	    setStatusLineImmediately ( s, totalChars );
	else {
	    final String str = s;
	    final int max = totalChars;
	    Runnable lineUpdater = new Runnable() {
		    public void run() {
			setStatusLineImmediately ( str, max );
		    }
		};
	    SwingUtilities.invokeLater(lineUpdater);
	}
    }

    /**
     * Display the given message in the status line, make progress bar
     * and abort button invisible
     */    
    public void setStatusLine(String s) {
	if ( SwingUtilities.isEventDispatchThread () )
	    setStatusLineImmediately ( s );
	else {
	    final String str = s;
	    Runnable lineUpdater = new Runnable() {
		    public void run() {
			setStatusLineImmediately ( str );
		    }
		};
	    SwingUtilities.invokeLater(lineUpdater);
	}
    }

    /**
     * Set the current progress as displayed in the status line, don't
     * change anything else
     */
    public void setProgress(int value) {
	if ( SwingUtilities.isEventDispatchThread () )
	    statusLine.setProgress ( value );
	else {
	    final int v = value;
	    Runnable lineUpdater = new Runnable() {
		    public void run() {
			statusLine.setProgress ( v );
		    }
		};
	    SwingUtilities.invokeLater(lineUpdater);
	}
    }

    /**
     * Freeze the main window by blocking all input events, except
     * those for the status line (i.e. the abort button within the
     * status line)
     */
    public void freezeExceptStatusLine () {
	if ( !frozen ) {
	    frozen = true;

	    Component glassPane = new BlockingGlassPane ( statusLine,
							  getContentPane () );
	    setGlassPane ( glassPane );
	    glassPane.setVisible ( true );
	}
    }

    public void unfreezeExceptStatusLine () {
	if ( frozen ) {
	    getGlassPane ().setVisible ( false );
	    frozen = false;
	}
    }

    /** ensure settings have been saved */
    protected void settingsSaved() {
	if (proofSettings().changed()) {
	    int option = JOptionPane.showConfirmDialog
		(Main.this, "Settings have changed. Save them?", 
		 "Save Settings", 
		 JOptionPane.YES_NO_OPTION);
	    if (option == JOptionPane.YES_OPTION) {
		proofSettings().saveSettings();
	    }
	}
	recentFiles.store(RECENT_FILES_STORAGE);
    }
    
    /** exit  */
    protected void exitMain() {
	settingsSaved(); 
	int option = JOptionPane.showConfirmDialog
	    (Main.this, "Really Quit?", "Exit", 
	     JOptionPane.YES_NO_OPTION);
	if (option == JOptionPane.YES_OPTION) {
            mediator.fireShutDown(new GUIEvent(this));
            if (standalone) {
	        System.out.println("Have a nice day.");
    	        System.exit(-1);
            }
	}
    }

    /** opens selection dialog for heuristics */
    protected void selectHeuristics() {
        if (mediator().ensureProofLoaded()) {
	    HeuristicSelector selector = 
                new HeuristicSelector(this, mediator());
        }
    }

    /** opens selection dialog for LDT models */
    protected void selectLDT() {
	LDTSelector selector = new LDTSelector
	    (mediator(), 
	     proofSettings().getLDTSettings());
    }

    /** opens configuration dialog for the simultaneous update simplifier */
    protected void configSimultaneousUpdateSimplifier() {
	SimultaneousUpdateSimplifierConfiguration config = 
	    new SimultaneousUpdateSimplifierConfiguration
	    (mediator(), 
	     proofSettings().getSimultaneousUpdateSimplifierSettings());
    }

    protected void setBack() {
       mediator.setBack();
    }

    protected void makePrettyView() {
	if (mediator().ensureProofLoaded()) {
	    if (!PresentationFeatures.ENABLED) {
		mediator().getNotationInfo().setBackToDefault();
	    } else {	    
		PresentationFeatures.modifyNotationInfo(mediator.getNotationInfo(),
							mediator.func_ns());
	    }
	    mediator().getSelectedProof().updateProof();	
	}

    }

    public void showLicense() {
			     
	URL lic = 
	KeYResourceManager.getManager().getResourceFile(Main.class,
							"LICENSE.TXT"); 
	StringBuffer sb=new StringBuffer();
	try {
	    FileInputStream inp=new FileInputStream(lic.getFile());
	    while (inp.available()>0) sb.append((char)inp.read());	   
	} catch (IOException ioe) {
	    System.out.println("License file cannot be loaded or is missing: \n"+
			       COPYRIGHT+"\nKeY is protected by the "
			       +"GNU General Public License");
	    sb=new StringBuffer(COPYRIGHT+"\nKeY is protected by the "
				+"GNU General Public License");
	} 
	String s=sb.toString();
	JScrollPane scroll = new JScrollPane();
	JTextArea text = new JTextArea(s,20,40);
	text.setEditable(false);
	text.setCaretPosition(0);
	scroll.setViewportView(text);
	JFrame fr = new JFrame("KeY License");
	fr.getContentPane().setLayout(new BorderLayout());
	fr.getContentPane().add(scroll,BorderLayout.CENTER);
	JButton ok = new JButton("OK");
	ok.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    //		    ((JFrame)((JRootPane)((JLayeredPane)((JPanel)((JButton)e.getSource()).getParent()).getParent()).getParent()).getParent()).dispose();
		    ((JFrame)((JButton)e.getSource())
		     .getTopLevelAncestor()).dispose();
		}});
	fr.getContentPane().add(ok, BorderLayout.SOUTH);
	fr.setSize(600,900);
	fr.getContentPane().add(scroll);
	fr.show();
    }



    public void showAbout() {
	JOptionPane pane =new JOptionPane(
            COPYRIGHT+"\n\nWWW: http://key-project.org\n\nVersion "+
            VERSION,
	    JOptionPane.INFORMATION_MESSAGE, JOptionPane.DEFAULT_OPTION,
	    IconFactory.keyLogo(105,60));
	JDialog dialog = pane.createDialog(this, "The KeY Project");
	dialog.show();
    }

    /** registers a new menuitem at the given menu
     * @param menu the JMenu where to register the new item
     * @param item the JMenuItem to register
     */
    public void registerAtMenu(JMenu menu, JMenuItem item) {
	menu.add(item);
    }

   /** adds a separator to the given menu
     */
    public void addSeparator(JMenu menu) {	
	menu.addSeparator();
    }


    /** starts the goal choice frame */
    protected void initGoalList() {
	GoalList goalList = new GoalList(mediator);
	goalList.setSize(goalList.getPreferredSize());
	openGoalsView.setViewportView(goalList);
    } 
    
    protected void initProofList() {
        proofList = new ProofList(mediator);
        proofListView = new JScrollPane();	
        addTab("Problems", proofListView, "Currently loaded proofs");
    }

    protected void addToProofList(Proof proof) {
        proofList.addProof(proof);	
	// GUI 
	proofList.setSize(proofList.getPreferredSize());
	proofListView.setViewportView(proofList);
    }


    /** starts the gui proof tree frame */
    protected void initGUIProofTree() {
	ProofTreeView guiProofTree = new ProofTreeView(mediator);
	guiProofTree.setSize(guiProofTree.getPreferredSize());
	guiProofTree.setVisible(true);
	proofView.setViewportView(guiProofTree);
    }
    
    static JFrame f = null;
    static java.awt.TextArea a;
    
    public static void copyHighlightToClipboard(SequentView view) {
       if (f == null) {
          f = new JFrame("KeY Clipboard");
          a = new java.awt.TextArea();
          f.getContentPane().add(a);
       }
       String s = view.getHighlightedText();
       // now CLIPBOARD
       java.awt.datatransfer.StringSelection ss = 
          new java.awt.datatransfer.StringSelection(s);
       f.getToolkit().getSystemClipboard().setContents(ss,ss);
       // now PRIMARY
       a.setText(s);
       f.pack();
       f.setState(java.awt.Frame.ICONIFIED);
       f.show();
       a.selectAll();
    }
    
    

    protected JMenu createFileMenu() {
	JMenu fileMenu = new JMenu("File");
	JMenuItem load = new JMenuItem("Load");
	load.setAccelerator(KeyStroke.getKeyStroke
			    (KeyEvent.VK_O, ActionEvent.CTRL_MASK));

	load.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    JFileChooser fileChooser=getFileChooser();
		    int loaded = fileChooser.showOpenDialog(Main.this);
		    if (loaded == JFileChooser.APPROVE_OPTION) {
			File file = fileChooser.getSelectedFile(); 
			loadProblem(file);
		    }
		}
	    });

	JMenuItem save = new JMenuItem("Save");
	save.setAccelerator(KeyStroke.getKeyStroke
			    (KeyEvent.VK_S, ActionEvent.CTRL_MASK));
	save.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    saveProof();
		}});



	JMenuItem exit = new JMenuItem("Exit");
	exit.setAccelerator(KeyStroke.getKeyStroke
			    (KeyEvent.VK_Q, ActionEvent.CTRL_MASK));
	exit.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    exitMain();		    
		}});	

	registerAtMenu(fileMenu, load);
	registerAtMenu(fileMenu, save);

	addSeparator(fileMenu);

	recentFiles = new RecentFileMenu
	    (new ActionListener() {
		    public void actionPerformed(ActionEvent e) {
			loadProblem
			    (new File(recentFiles.
				      getAbsolutePath((JMenuItem)e.getSource())));
		    }
		}, MAX_RECENT_FILES, null);

	recentFiles.load(RECENT_FILES_STORAGE);

	registerAtMenu(fileMenu, recentFiles.getMenu());

	addSeparator(fileMenu);
	registerAtMenu(fileMenu, exit);
        return fileMenu;
    }


    protected JMenu createViewMenu() {
	JMenu view = new JMenu("View");

	JMenuItem pretty = new JCheckBoxMenuItem("Pretty&Untrue");
	pretty.setSelected(PresentationFeatures.ENABLED);
	pretty.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
                PresentationFeatures.ENABLED=((JCheckBoxMenuItem)e.getSource()).
		    isSelected();
		makePrettyView();
		}});

	final JMenuItem smaller = new JMenuItem("Smaller");
	smaller.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    Config.DEFAULT.smaller();
		}});
	final JMenuItem larger  = new JMenuItem("Larger");
	larger.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    Config.DEFAULT.larger();
		}});

	Config.DEFAULT.addConfigChangeListener(
	    new ConfigChangeListener() {
		    public void configChanged(ConfigChangeEvent e) {
			smaller.setEnabled(!Config.DEFAULT.isMinimumSize());
			larger.setEnabled(!Config.DEFAULT.isMaximumSize());
		    }
		});
	
	registerAtMenu(view, pretty);
	addSeparator(view);
	registerAtMenu(view, smaller);
	registerAtMenu(view, larger);
        return view;
    }
    
    protected JMenu createProofMenu() {
        JMenu proof = new JMenu("Proof");
	JMenuItem heuristicItem = new JMenuItem("Heuristics...");
	heuristicItem.setAccelerator(KeyStroke.getKeyStroke
			    (KeyEvent.VK_H, ActionEvent.CTRL_MASK));
	heuristicItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    selectHeuristics();
		}});
	registerAtMenu(proof, heuristicItem);

	JMenuItem close = new JMenuItem("Abandon");
	close.setAccelerator(KeyStroke.getKeyStroke
			    (KeyEvent.VK_W, ActionEvent.CTRL_MASK));
	close.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    closeProof();
		}});
	registerAtMenu(proof, close);
        return proof;
    }


    protected JMenu createOptionsMenu() {
	JMenu options = new JMenu("Options");
	JMenuItem dataTypeItem = new JMenuItem("LDT Models...");
	dataTypeItem.setAccelerator(KeyStroke.getKeyStroke
			    (KeyEvent.VK_L, ActionEvent.CTRL_MASK));

	dataTypeItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    selectLDT();
		}});
	registerAtMenu(options, dataTypeItem);	

	JMenuItem updateSimplifierItem = new JMenuItem("Update Simplifier...");
	updateSimplifierItem.setAccelerator(KeyStroke.getKeyStroke
			    (KeyEvent.VK_U, ActionEvent.CTRL_MASK));

	updateSimplifierItem.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    configSimultaneousUpdateSimplifier();
		}});
	registerAtMenu(options, updateSimplifierItem);	

	JMenuItem computeSpecificationOptions = 
	    ComputeSpecificationView.createOptionMenuItems();
	registerAtMenu(options, computeSpecificationOptions);

	addSeparator(options);

	proofSettings().item().addActionListener
	    (new ActionListener() {
		    public void actionPerformed(ActionEvent e) {
			proofSettings().saveSettings();
		    }});

	registerAtMenu(options, proofSettings().item());
        return options;
    }


    protected JMenu createHelpMenu() {
	JMenu help = new JMenu("Help");
	JMenuItem about = new JMenuItem("About");
	about.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    showAbout();
		}});
	help.add(about);	

	JMenuItem license = new JMenuItem("License");
	license.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    showLicense();
		}});
	help.add(license);
        return help;
    }

    protected JMenu createToolsMenu() {
	JMenu tools = new JMenu("Tools");
	menubar.add(tools);

	JMenuItem extractSpecification = new JMenuItem("Extract Specification");
	extractSpecification.setAccelerator(KeyStroke.getKeyStroke
			    (KeyEvent.VK_E, ActionEvent.CTRL_MASK));
	extractSpecification.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    if (mediator().ensureProofLoaded()) {
			//@internal we don't want to block UI just
			//because we are about to calculate a lot of
			//things, now. Also the interactive prover
			//might want to run during the execution of
			//ComputeSpecification
			
			new Thread(new Runnable() {
				public void run() {
				    ComputeSpecificationView.show(mediator());
				}
			    }).start();
		    }
		}
	  });
	tools.add(extractSpecification);
	return tools;
    }

    /** creates menubar */
    protected JMenuBar createMenuBar() {
	menubar = new JMenuBar();
	menubar.add(createFileMenu());
	menubar.add(createViewMenu());
        menubar.add(createProofMenu());
	menubar.add(createOptionsMenu());
        menubar.add(createToolsMenu());
        menubar.add(Box.createHorizontalGlue());
	menubar.add(createHelpMenu());
	
	return menubar;
    }


    /** updates the view of the sequent being displayed in the main
     * frame 
     */
    public synchronized void updateGoalView(Sequent sequent) {
	SequentPrintFilter filter =
	    new ConstraintSequentPrintFilter ( sequent,
					       mediator ().getUserConstraint ()
					       .getConstraint () );
	SequentPrinter printer = new SequentPrinter
	    (new ProgramPrinter(null), 
	     mediator().getNotationInfo(),
	     filter);
	sequentView.setPrinter(printer, filter);
	sequentView.printSequent();

	goalView.setViewportView(sequentView);
	goalView.setBorder(new TitledBorder("Current Goal"));
        sequentView.setCaretPosition(0); // at least until a better solution
    }


    private JFileChooser getFileChooser() {
	if (fileChooser==null) {
	    fileChooser = new JFileChooser(new File("."));
	    fileChooser.setFileFilter(new javax.swing.filechooser.FileFilter() {
		    public boolean accept(File f) {
			return 
			    f.isDirectory() 
			    || f.toString().endsWith(".key") 
			    || f.toString().endsWith(".proof");
		    }

		    public String getDescription() {
			return "KeY files";
		    }
		});
	}
        File selFile = fileChooser.getSelectedFile();
        if ((selFile != null) && selFile.isFile()) { // present & not dir.
            String filename = selFile.getAbsolutePath();    
	    if (!filename.endsWith(".proof")) 
                fileChooser.setSelectedFile(new File(filename+".proof")); 
        }
	return fileChooser;
    }


    /** saves a proof */
    protected void saveProof() {
	JFileChooser jFC = getFileChooser();
	int saved = jFC.showSaveDialog(this);
	if (saved == JFileChooser.APPROVE_OPTION) {
	    String filename = jFC.getSelectedFile().getAbsolutePath();    
            ProofSaver saver;
            if (filename.endsWith(".tex"))
               saver = new ProofSaverLatex(this, filename);
             else saver = new ProofSaver(this, filename);
            String errorMsg = saver.save();
            if (errorMsg != null) {
	       JOptionPane.showMessageDialog(this, errorMsg, "Oops...", 
                                             JOptionPane.ERROR_MESSAGE);
            }

	}	
    }


    protected void loadProblem(File file) {
	recentFiles.addRecentFile(file.getAbsolutePath());
	Thread t = new Thread(new ProblemLoader(file, this));
	t.start();
    }
    
    protected void closeProof() {
        JOptionPane.showMessageDialog(this, "Impement me!", "Oops...", 
                                      JOptionPane.PLAIN_MESSAGE);
    }


    /**
     * brings window in front and request focus
     */
    private void popup() {
	setState(Frame.NORMAL);
	setVisible(true);
	requestFocus();
    }

    public void addProblem(Proof p) {
        setUpNewProof(p);
	addToProofList(p);
	popup();
//        initGoalList();
//        initGUIProofTree();
    }


    protected Proof setUpNewProof(Proof proof) {
	KeYMediator mediator = mediator();
	mediator.setProof(proof);
	return proof;
    }


    private void setToolBarEnabled ( boolean p_enabled ) {
	Component[] cs = toolBar.getComponents ();
	int         i  = cs.length;
	while ( i-- != 0 )
	    cs[i].setEnabled ( p_enabled );
    }


    class MainListener extends WindowAdapter {
	public void windowClosing(WindowEvent e) {
	    exitMain();	    
	}
    }


    class MainGUIListener implements GUIListener {
	/** invoked if a frame that wants modal access is opened */

	private void enableMenuBar(JMenuBar  m, boolean b){
	    for (int i=0; i < m.getMenuCount(); i++){
		JMenu currentMenu = m.getMenu(i);
		if (currentMenu instanceof JMenu) 
                   currentMenu.setEnabled(b); // otherwise it's a spacer
	    }
	}

	public void modalDialogOpened(GUIEvent e) {
	 
	    if (e.getSource() instanceof ApplyTacletDialog){
		// disable all elements except the sequent window (drag'n'drop !) ...
		enableMenuBar(Main.this.getJMenuBar(), false); 
		Main.this.goalView.setEnabled(false);
		Main.this.proofView.setEnabled(false); 
		Main.this.openGoalsView.setEnabled(false);
		Main.this.userConstraintView.setEnabled(false);
		setToolBarEnabled ( false );
	    } else {
		// disable the whole main window ...
		Main.this.setEnabled(false);
	    }
	}

	/** invoked if a frame that wants modal access is closed */
	public void modalDialogClosed(GUIEvent e) {
	    if (e.getSource() instanceof ApplyTacletDialog){
		// enable all previously diabled elements ...
		enableMenuBar(Main.this.getJMenuBar(), true); 
		Main.this.goalView.setEnabled(true);
		Main.this.proofView.setEnabled(true); 
		Main.this.openGoalsView.setEnabled(true);
		Main.this.userConstraintView.setEnabled(true);
		setToolBarEnabled ( true );
	    } else {
		// enable the whole main window ...
	       Main.this.setEnabled(true);
	    }
	}
	
	public void shutDown(GUIEvent e) {
	    Main.this.setVisible(false);
	}

    }

    /** set to true if the view of the current goal
     * should not be updated */       
    private boolean disableCurrentGoalView = false;

    private synchronized void setProofNodeDisplay() {
	if (!disableCurrentGoalView) {
	    Goal goal;
	    try {
		goal = mediator().getSelectedGoal();
	    } catch(IllegalStateException e) { // there is no proof (yet)
		return;
	    }
	    if ( goal != null &&
		 !mediator.getUserConstraint ().displayClosed ( goal.node () ) ) {
		updateGoalView(goal.sequent());
	    } else {
		NonGoalInfoView innerNodeView = 
		    new NonGoalInfoView(mediator().getSelectedNode(), 
					mediator());
		goalView.setBorder(new TitledBorder("Inner Node"));
		goalView.setViewportView(innerNodeView);	    
		goalView.setViewportView(innerNodeView);
		innerNodeView.setCaretPosition(0);
	    }
	    goalView.validate();	    
	    validate();
	}
    }

    class MainProofListener implements InteractiveProofListener, 
                                       KeYSelectionListener {

	/** focused node has changed */
 	public synchronized void selectedNodeChanged(KeYSelectionEvent e) {
	    setProofNodeDisplay();	    
 	}

 	/** the selected proof has changed (e.g. a new proof has been
 	 * loaded) */ 
 	public synchronized void selectedProofChanged(KeYSelectionEvent e) {
	    Debug.out("Main: initialize with new proof");
 	    disableCurrentGoalView = false;	    
 	    goalView.setViewportView(null);

	    if ( userConstraint != null )
		userConstraint
		    .removeConstraintTableListener ( constraintListener );
	    userConstraint = mediator ().getUserConstraint ();
	    if ( userConstraint != null )
		userConstraint
		    .addConstraintTableListener ( constraintListener );
	
	    setProofNodeDisplay();
	    makePrettyView();
 	}

	/** invoked if automatic execution has started
	 */
	public synchronized void autoModeStarted(ProofEvent e) {
	    disableCurrentGoalView = true;
	    mediator().removeKeYSelectionListener(proofListener);
	    freezeExceptStatusLine ();
	}

	/** invoked if automatic execution has stopped
	 */
	public synchronized void autoModeStopped(ProofEvent e) {
	    unfreezeExceptStatusLine ();
	    disableCurrentGoalView = false;
	    setProofNodeDisplay();
	    mediator().addKeYSelectionListener(proofListener);
	}

	/** invoked when a rule has been applied */
	public synchronized void ruleApplied(ProofEvent e) {
	    //	    setProofNodeDisplay();
	}
    }

    
    class MainConstraintTableListener implements ConstraintTableListener {
	public void constraintChanged ( ConstraintTableEvent e ) {
	    setProofNodeDisplay ();
	}
    }
	

    public static void evaluateOptions(String[] opt) {
	int index = 0;
	while (opt.length > index) {	    
	    if ((new File(opt[index])).exists()) {
		fileNameOnStartUp=opt[index];
	    } else {
		opt[index] = opt[index].toUpperCase();		
		if (opt[index].equals("NO_DEBUG")) {
		    de.uka.ilkd.key.util.Debug.ENABLE_DEBUG = false;
		} else if (opt[index].equals("DEBUG")) {
		    de.uka.ilkd.key.util.Debug.ENABLE_DEBUG = true;
		} else if (opt[index].equals("NO_ASSERTION")) {
		    de.uka.ilkd.key.util.Debug.ENABLE_ASSERTION = false;
		} else if (opt[index].equals("ASSERTION")) {
		    de.uka.ilkd.key.util.Debug.ENABLE_ASSERTION = true;
		} else {
		    System.out.println("Possible parameters are (* = default): ");
		    System.out.println("  no_debug        : disables debug mode (*)");
		    System.out.println("  debug           : enables debug mode");
		    System.out.println("  no_assertion    : disables assertions");
		    System.out.println("  assertion       : enables assertions (*)");
		    System.out.println("  <filename>      : loads a .key file");
		    System.exit(-1);
		}		
	    }
	    index++;
	}	

	if (Debug.ENABLE_DEBUG) {
	    System.out.println("Running in debug mode ...");
	} else {
	    System.out.println("Running in normal mode ...");
	}
	if (Debug.ENABLE_ASSERTION) {
	    System.out.println("Using assertions ...");	   
	} else {
	    System.out.println("Not using assertions ...");	   
	}

    }
    
    
    /**
     * Glass pane that only delivers events for the status line
     * (i.e. the abort button)
     *
     * This has been partly taken from the GlassPaneDemo of the Java
     * Tutorial
     */
    class BlockingGlassPane extends JComponent {
	GlassPaneListener listener;

	public BlockingGlassPane ( Component liveComponent,
				   Container contentPane ) {
	    setCursor(new Cursor(Cursor.WAIT_CURSOR));

	    listener = new GlassPaneListener ( liveComponent,
					       this,
					       contentPane );
	    addMouseListener       ( listener );
	    addMouseMotionListener ( listener );
	}
    }


    /**
     * Mouse listener for the glass pane that only delivers events for
     * the status line (i.e. the abort button)
     *
     * This has been partly taken from the GlassPaneDemo of the Java
     * Tutorial
     */
    class GlassPaneListener extends MouseInputAdapter {
	Component liveComponent;
	Component currentComponent = null;
	Component glassPane;
	Container contentPane;

	public GlassPaneListener ( Component liveComponent,
				   Component glassPane,
				   Container contentPane ) {
	    this.liveComponent = liveComponent;
	    this.glassPane     = glassPane;
	    this.contentPane   = contentPane;
	}

	public void mouseMoved(MouseEvent e) {
	    redispatchMouseEvent(e);
	}

	/*
	 * We must forward at least the mouse drags that started
	 * with mouse presses over the check box.  Otherwise,
	 * when the user presses the check box then drags off,
	 * the check box isn't disarmed -- it keeps its dark
	 * gray background or whatever its L&F uses to indicate
	 * that the button is currently being pressed.
	 */
	public void mouseDragged(MouseEvent e) {
	    redispatchMouseEvent(e);
	}

	public void mouseClicked(MouseEvent e) {
	    redispatchMouseEvent(e);
	}

	public void mouseEntered(MouseEvent e) {
	    redispatchMouseEvent(e);
	}

	public void mouseExited(MouseEvent e) {
	    redispatchMouseEvent(e);
	}

	public void mousePressed(MouseEvent e) {
	    redispatchMouseEvent(e);
	}

	public void mouseReleased(MouseEvent e) {
	    redispatchMouseEvent(e);
	    currentComponent = null;
	}

	private void redispatchMouseEvent(MouseEvent e) {
	    if ( currentComponent != null ) {
		dispatchForCurrentComponent ( e );
	    } else {
		int       eventID        = e.getID();
		Point     glassPanePoint = e.getPoint();

		Point     containerPoint =
		    SwingUtilities.convertPoint(glassPane,
						glassPanePoint, 
						contentPane);
		Component component      =
		    SwingUtilities.getDeepestComponentAt(contentPane,
							 containerPoint.x,
							 containerPoint.y);

		if ( eventID == MouseEvent.MOUSE_PRESSED &&
		     isLiveComponent ( component ) ) {
		    currentComponent = component;
		    dispatchForCurrentComponent ( e );		
		}
	    }
	}

	private boolean isLiveComponent ( Component c ) {
	    while ( c != null ) {
		if ( liveComponent == c )
		    return true;
		c = c.getParent ();
	    }
	    return false;
	}

	private void dispatchForCurrentComponent ( MouseEvent e ) {
	    Point glassPanePoint = e.getPoint();
	    Point componentPoint =
		SwingUtilities.convertPoint( glassPane,
					     glassPanePoint, 
					     currentComponent );
	    currentComponent.dispatchEvent(new MouseEvent(currentComponent,
							  e.getID(),
							  e.getWhen(),
							  e.getModifiers(),
							  componentPoint.x,
							  componentPoint.y,
							  e.getClickCount(),
							  e.isPopupTrigger()));
	}
    }
    


    public void loadCommandLineFile() {
	if (fileNameOnStartUp!=null) {
	    loadProblem(new File(fileNameOnStartUp));
	}
    }

    
    public static void main(String[] args) {	
	Main.evaluateOptions(args);
	Main key = getInstance();
	key.loadCommandLineFile();
    }
    
    
}
