// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.util.pp.UnbalancedBlocksException;
import java.io.IOException;

/** Encapsulate the concrete syntax used to print a term.  The {@link
 * NotationInfo} class associates a Notation with every {@link
 * de.uka.ilkd.key.logic.op.Operator}.  The various inner classes of
 * this class represent different kinds of concrete syntax, like prefix,
 * infix, postfix, function style, attribute style, etc.
 * 
 */

public abstract class Notation {

    /** The priority of this operator in the given concrete syntax. 
     * This is used to determine whether parentheses are required around
     * a subterm.
     */
    protected int priority;

    /** Create a Notation with a given priority. */
    protected Notation(int priority) {
	this.priority=priority;
    }

    /** get the priority of the term */
    public int getPriority() {
	return priority;
    }

    /** Print a term to a {@link SequentPrinter}. 
     * Concrete subclasses override this to call one of the
     * <code>printXYZTerm</code> of {@link SequentPrinter}, which
     * do the layout.
     */
    public void print(Term t,SequentPrinter sp) 
	throws UnbalancedBlocksException,IOException
    {
	if(sp.getNotationInfo().getAbbrevMap().isEnabled(t)){
	    sp.printTerm(t);
	}else{
	    sp.printConstant(t.toString());
	    System.err.println("There is no registered Notation "
			       +"for the operator\n"+t.op());
	}
    }

    /** Print a term without beginning a new block.  See
     * {@link SequentPrinter#printTermContinuingBlock(Term)} for the
     * idea behind this.  The standard implementation just delegates to
     * {@link #print(Term,SequentPrinter)}
     */
    public void printContinuingBlock(Term t,SequentPrinter sp)
	throws UnbalancedBlocksException,IOException
    {
	print(t,sp);
    }

    /**
     * The standard concrete syntax for constants like true and false.
     */
    public static class Constant extends Notation {
	String name;

	public Constant(String name,int prio) {
	    super(prio);
	    this.name = name;
	}

	public void print(Term t,SequentPrinter sp)
	    throws UnbalancedBlocksException,IOException
	{
	    sp.printConstant(name);
	}
    }

    /**
     * The standard concrete syntax for prefix operators.
     */
    public static class Prefix extends Notation {
	String name;
	int ass;

	public Prefix(String name,int prio,int ass) {
	    super(prio);
	    this.name = name;
	    this.ass = ass;
	}


	public void print(Term t,SequentPrinter sp)
	    throws UnbalancedBlocksException,IOException
	{
	if(sp.getNotationInfo().getAbbrevMap().isEnabled(t)){
	    sp.printTerm(t);
	}else{
	    sp.printPrefixTerm(name,t.sub(0),ass);
	}
	}

    }

    /**
     * The standard concrete syntax for infix operators.
     */
    public static class Infix extends Notation {
	String name;
	int assLeft,assRight;

	public Infix(String name,int prio,int assLeft,int assRight) {
	    super(prio);
	    this.name = name;
	    this.assLeft = assLeft;
	    this.assRight = assRight;
	}


	public void print(Term t,SequentPrinter sp) 
	    throws UnbalancedBlocksException,IOException
	{
	    if(sp.getNotationInfo().getAbbrevMap().isEnabled(t)){
		sp.printTerm(t);
	    }else{
	    	sp.printInfixTerm(t.sub(0),assLeft,name,t.sub(1),assRight);
	    }
	}

	/** Print a term without beginning a new block.  This calls
	 * the {@link SequentPrinter#printTermContinuingBlock(Term)}
	 * method. */
	public void printContinuingBlock(Term t,SequentPrinter sp) 
	    throws UnbalancedBlocksException,IOException
	{
	    sp.printInfixTermContinuingBlock(
			      t.sub(0),assLeft,name,t.sub(1),assRight);
	}

    }
    
    /** 
     * The standard concrete syntax for arrays.
     */
    public static class ArrayNot extends Notation{
        
        String pre,post;
        
        public ArrayNot(String pre,String post,int prio){
            super(prio);
            this.pre=pre;
            this.post=post;
        }
        
        public void print(Term t, SequentPrinter sp)
	    throws UnbalancedBlocksException,IOException{
	    if(sp.getNotationInfo().getAbbrevMap().isEnabled(t)){
		sp.printTerm(t);
	    }else{
		sp.printArray(pre, post, t.sub(0), t.sub(1));
	    }
	}
    }
            
    /** 
     * The standard concrete syntax for quantifiers.
     */
    public static class Quantifier extends Notation {
	String name;
	int ass;

	public Quantifier(String name,int prio,int ass) {
	    super(prio);
	    this.name = name;
	    this.ass = ass;
	}


	public void print(Term t,SequentPrinter sp) 
	    throws UnbalancedBlocksException,IOException
	{
	    if(sp.getNotationInfo().getAbbrevMap().isEnabled(t)){
		sp.printTerm(t);
	    }else{
	    	QuantifiableVariable v =
		    t.varsBoundHere(0).getQuantifiableVariable(0);
	    	sp.printQuantifierTerm(name,
				   ""+v.name()+":"+v.sort().name(),
				   ".",
				   t.sub(0),
				   ass);
	    }
	}

    }

    /**
     * The standard concrete syntax for DL modalities box and diamond.
     */
    public static class Modality extends Notation {
	String left,right;
	int ass;

	public Modality(String left,String right,int prio,int ass) {
	    super(prio);
	    this.left = left;
	    this.right = right;
	    this.ass = ass;
	}


	public void print(Term t,SequentPrinter sp) 
	    throws UnbalancedBlocksException,IOException
	{
	    if(sp.getNotationInfo().getAbbrevMap().isEnabled(t)){
		sp.printTerm(t);
	    }else{
	    	sp.printModalityTerm(left,t.javaBlock(),right,t.sub(0),ass);
	    }
	}
    }

    /**
     * The standard concrete syntax for terms with updates.
     */
    public static class SimultaneousUpdate extends Notation {

	public SimultaneousUpdate() {
	    super(60);
	}


	public void print(Term t,SequentPrinter sp)
	    throws UnbalancedBlocksException,IOException
	{
	    if(sp.getNotationInfo().getAbbrevMap().isEnabled(t)){
		sp.printTerm(t);
	    } else{
		sp.printUpdateTerm("{", ":=", "}",
				   t, 60, 0, 60);

	    }
	}
    }

    /**
     * The standard concrete syntax for conjunctive conditional terms
     */
    public static class ConjCond extends Notation {

	public ConjCond() {
	    super(40);
	}


	public void print(Term t,SequentPrinter sp)
	    throws UnbalancedBlocksException,IOException
	{
	    if(sp.getNotationInfo().getAbbrevMap().isEnabled(t)){
		sp.printTerm(t);
	    } else{
		sp.printConjCondTerm("(","?=","\u2022",":",t,")", 60);
	    }
	}
    }

    /**
     * The standard concrete syntax for substitution terms.
     */
    public static class Subst extends Notation {
	public Subst () {
	    super(60);
	}

	public void print(Term t,SequentPrinter sp) 
	    throws UnbalancedBlocksException,IOException
	{
	    if(sp.getNotationInfo().getAbbrevMap().isEnabled(t)){
		sp.printTerm(t);
	    }else{
	    	QuantifiableVariable v =
			t.varsBoundHere(1).getQuantifiableVariable(0);
	    	sp.printSubstTerm("{",
			      v,
			      "/",
			      t.sub(0),0,
			      ( t.op () instanceof WarySubstOp ) ?
				  ( ((WarySubstOp)t.op ()).getAllQuantifier () ?
				    " waryAll }" :
				    " waryEx }" ) :
				  "}",
			      t.sub(1),60);
	    }
	}
    }


    /**
     * The standard concrete syntax for function and predicate terms.
     */
    static class Function extends Notation {

	public Function() {
	    super(100);
	}

	public void print(Term t,SequentPrinter sp)
	    throws UnbalancedBlocksException,IOException
	{
	    if(sp.getNotationInfo().getAbbrevMap().isEnabled(t)){
		sp.printTerm(t);
	    }else{
	        sp.printFunctionTerm(t.op().name().toString(),t);
	    }
	}
    }


    /**
     * The standard concrete syntax for query terms <code>o.q(x)</code>.
     */
    static class ProgramMethod extends Notation {
        public ProgramMethod() {
	    super(80);
	}

	public void print(Term t,SequentPrinter sp)
	    throws UnbalancedBlocksException,IOException
	{
	    if(sp.getNotationInfo().getAbbrevMap().isEnabled(t)){
		sp.printTerm(t);
	    }else{
	        sp.printQueryTerm(t.op().name().toString(),t);
	    }
	}
    }


    /**
     * The standard concrete syntax for attribute terms <code>o.a</code>.
     */
    static class Attribute extends Notation {

	private int associativity;

	public Attribute(int priority, int associativity) {
	    super(priority);
	    this.associativity = associativity;
	}

	public void print(Term t,SequentPrinter sp)
	    throws UnbalancedBlocksException,IOException
	{
	    if(sp.getNotationInfo().getAbbrevMap().isEnabled(t)){
		sp.printTerm(t);
	    }else{
	        sp.printPostfixTerm(t.sub(0),associativity,
				    "."+t.sub(1).op().name());
	    }
	}
    }

    /**
     * The standard concrete syntax for conditional terms
     * <code>{o1 =? o2}.a/->t</code>.
     */
    static class IfElse extends Notation {
        public IfElse() {	   
	    super(60);
	}

	public void print(Term t,SequentPrinter sp)
	    throws UnbalancedBlocksException,IOException
	{
	    if(sp.getNotationInfo().getAbbrevMap().isEnabled(t)){
		sp.printTerm(t);
	    } else {
	    	sp.printConditionalTerm("(",
					t.sub(0),80,
					"?=",
					t.sub(1),80,
					")", t.sub(2), ":",
					t.sub(3),60);
	    }
	}
    }

    /**
     * The standard concrete syntax for all kinds of variables.
     */
    static class Variable extends Notation {
        public Variable() {
	    super(1000);
	}

	public void print(Term t,SequentPrinter sp)
	    throws UnbalancedBlocksException,IOException
	{
	    sp.printConstant(t.op().name().toString());
	}
    }


    static class Metavariable extends Notation {
        public Metavariable() {
	    super(1000);
	}

	public void print(Term t,SequentPrinter sp)
	    throws UnbalancedBlocksException,IOException
	{
	    sp.printMetavariable
		( (de.uka.ilkd.key.logic.op.Metavariable)t.op () );
	}
    }


    /**
     * The standard concrete syntax for the number literal indicator `Z'.
     * This is only used in the `Pretty&amp;Untrue' syntax.
     */
    static class NumLiteral extends Notation {
        public NumLiteral() {
	    super(1000);
        }

        private static StringBuffer printlastfirst(Term t) {
	        if (t.op().arity()==0) {
	            return new StringBuffer();
	        } else {
	            return printlastfirst(t.sub(0)).
		        append(t.op().name().toString());
	        }
        }

        public static String printNumberTerm(Term t)
        {
	        StringBuffer sb=new StringBuffer();
	        if (t.sub(0).op().name().toString().equals("neg")) {
	            return "-"+
		        printNumberTerm(t.sub(0));
	        } else {
	        return printlastfirst(t.sub(0)).toString();
	        }
        }
	

	public void print(Term t,SequentPrinter sp)
	    throws UnbalancedBlocksException,IOException
	{
	    sp.printConstant(printNumberTerm(t));
	}
    }

    /**
     * The standard concrete syntax for the character literal indicator `C'.
     */
    static class CharLiteral extends Notation {
        public CharLiteral() {
	    super(1000);
        }

        public static String printCharTerm(Term t)
        {
	    
	    char charVal=0;
	    int intVal=0;

	    String result = NumLiteral.printlastfirst
		(t.sub(0)).toString();
	    try {
		intVal = Integer.parseInt(result);
		charVal = (char)intVal;
		if (intVal-charVal!=0)
		    throw new NumberFormatException(); //overflow!
		
	    } catch (NumberFormatException ex) {
		System.out.println("Oops. "+result+
				   " is not of type char");
	    } 		
	    return ("'"+new Character(charVal)).toString()+"'";
        }
	

	public void print(Term t,SequentPrinter sp)
	    throws UnbalancedBlocksException,IOException
	{
	    sp.printConstant(printCharTerm(t));
	}
    }

}

