// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;

import java.util.HashMap;
import java.util.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.gui.KeYMediator;
import de.uka.ilkd.key.util.Service;


/** Stores the mapping from operators to {@link Notation}s.  Each 
 * {@link Notation} represents the concrete syntax for some 
 * {@link de.uka.ilkd.key.logic.op.Operator}.  The {@link SequentPrinter}
 * asks the NotationInfo to find out which Notation to use for a given term.
 * 
 * The Notation associated with an operator might change.  New Notations can
 * be added.
 */
public class NotationInfo {

    /** Factory method: creates a new NotationInfo instance. The
     * actual implementation depends on system properties or service
     * entries. */
    public static final NotationInfo createInstance() {
	return (NotationInfo) Service.find(NotationInfo.class.getName(),
					   NotationInfo.class.getName());
    }
    
    
    /** This maps operators and classes of operators to {@link
     * Notation}s.  The idea is that we first look wether the operator has
     * a Notation registered.  Otherwise, we see if there is one for the
     * <em>class</em> of the operator.
     */
    private HashMap tbl;

    /**
     * Maps terms to abbreviations and reverse.
     */
    private AbbrevMap scm;

    /** the KeYMediator */
    private KeYMediator mediator;


    /** Create a new NotationInfo. Do not call this constructor
     * directly. Use the factory method {@link createInstance}
     * instead. */
    public NotationInfo() {
	createDefaultNotationTable();
    }

    /** Set all notations back to default. */
    public void setBackToDefault() {
	createDefaultNotationTable();
    }
    
    /** Register the standard set of notations.  This means no
     * abbreviations, and a set of Notations for the built-in operators
     * which corresponds to the parser syntax. 
     */
    private void createDefaultNotationTable() {
	tbl=new HashMap();
	createDefaultOpNotation();
	createDefaultTermSymbolNotation();
	scm = new AbbrevMap();
    }

    /**
     * Registers notations for the built-in operators.  The priorities
     * and associativities correspond to the parser syntax.  
     */
    private void createDefaultOpNotation() {
	tbl.put(Op.TRUE ,new Notation.Constant("true",80));
	tbl.put(Op.FALSE,new Notation.Constant("false",80));
	tbl.put(Op.NOT,new Notation.Prefix("!" ,60,60));
	tbl.put(Op.AND,new Notation.Infix("&"  ,50,60,50));
	tbl.put(Op.OR, new Notation.Infix("|"  ,40,50,40));
	tbl.put(Op.IMP,new Notation.Infix("->" ,30,40,30));
	tbl.put(Op.EQV,new Notation.Infix("<->",20,30,20));
	tbl.put(Op.ARRAY, new Notation.ArrayNot("[","]",120));
    	tbl.put(Op.ALL,new Notation.Quantifier("all",60,60));
	tbl.put(Op.EX, new Notation.Quantifier("ex", 60,60));
	tbl.put(Op.DIA,new Notation.Modality("<",">",60,60));
	tbl.put(Op.BOX,new Notation.Modality("[","]",60,60));
	tbl.put(Op.IF_ELSE, new Notation.IfElse());

	tbl.put(Op.CTRL,new Notation.Prefix("^" ,60,60));

	createNumLitNotation(((de.uka.ilkd.key.logic.ldt.NumberLDT)ProofSettings.
			      THE_SETTINGS.getLDTSettings().
			      getNaturalNumberLDT()).getNumberSymbol());
	tbl.put(Op.SUBST,new Notation.Subst());
	tbl.put(Op.WARY_SUBST_EX,new Notation.Subst());
	tbl.put(Op.WARY_SUBST_ALL,new Notation.Subst());
    }
    

    /** 
     * Register notations for standard classes of operators.  This
     * includes Function operators, all kinds of variables, etc.
     */
    private void createDefaultTermSymbolNotation() {
	tbl.put(Function.class, new Notation.Function());
	tbl.put(LogicVariable.class, new Notation.Variable());
	tbl.put(SchemaVariable.class, new Notation.Variable());
	//	tbl.put(Metavariable.class, new Notation.Variable());
	tbl.put(Metavariable.class, new Notation.Metavariable());
	tbl.put(ProgramVariable.class, new Notation.Variable());
	tbl.put(ProgramMethod.class, new Notation.ProgramMethod());
	tbl.put(Equality.class, new Notation.Infix("=",70,80,80)); 
	tbl.put(SimultaneousUpdateOp.class, new Notation.SimultaneousUpdate());
	tbl.put(AttributeOp.class, new Notation.Attribute(120,80));
	tbl.put(ConjCond.class, new Notation.ConjCond());

    }

    public AbbrevMap getAbbrevMap(){
	return scm;
    }

    /** Get the Notation for a given Operator.  
     * If no notation is registered, a Function notation is returned.
     */
    public Notation getNotation(Operator op) {
	createNumLitNotation(((de.uka.ilkd.key.logic.ldt.NumberLDT)ProofSettings.
			      THE_SETTINGS.getLDTSettings().getNaturalNumberLDT()).getNumberSymbol());
	createCharLitNotation(((de.uka.ilkd.key.logic.ldt.NumberLDT)ProofSettings.
			       THE_SETTINGS.getLDTSettings().getNaturalNumberLDT()).getCharSymbol());
	if (tbl.containsKey(op)) {
	    return (Notation) tbl.get(op);
	} else if (tbl.containsKey(op.getClass())) {
	    return (Notation) tbl.get(op.getClass());
	} else {
	    return new Notation.Function();
	}
    }

    /** Registers an infix notation for a given operator
     * @param op    the operator
     * @param token the string representing the operator
     */
    public void createInfixNotation(Operator op, String token) {
	tbl.put(op, new Notation.Infix(token,90,100,100));
    }

    /** Registers a prefix notation for a given operator 
     * @param op    the operator
     * @param token the string representing the operator
     */
    public void createPrefixNotation(Operator op, String token) {
	tbl.put(op, new Notation.Prefix(token, 110, 0));
    }

    /** Registers a number literal notation for a given operator.
     * This is done for the `Z' operator which marks number literals.
     * A term <code>Z(3(2(#)))</code> gets printed simply as
     * <code>23</code>.
     * @param op the operator */
    public void createNumLitNotation(Operator op) {
	tbl.put(op, new Notation.NumLiteral());
    }


    /** Registers a character literal notation for a given operator.
     * This is done for the `C' operator which marks character literals.
     * A term <code>C(3(2(#)))</code> gets printed simply as
     * the character corresponding to the unicode value 23 (really 23
     * and not 32, see integer literals)
     * @param op the operator */
    public void createCharLitNotation(Operator op) {
	tbl.put(op, new Notation.CharLiteral());
    }


}
