// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;
import de.uka.ilkd.key.logic.*;

/**
 * describes a position in a sequent including the bounds within a string
 * representation of the sequent. In contrast to PosInOccurrence and
 * PosInTerm of package de.uka.ilkd.key.logic, this class is mutable,
 * i.e the bounds may be set later in an already existing PosInSequent
 * instance. Apart from the bounds, PosInSequent has the following kind
 * of states: It marks the whole sequent, the whole antecedent, the whole
 * succedent or includes a PosInOccurrence if a position within a
 * constrained formula of the sequent is referred to. In the latter case
 * it contains also information whether the whole constrained formula
 * is referred to or the formula or the constraint.
 */
public class PosInSequent {

    private int start=0;
    private int end=0;
    private boolean sequent;
    private boolean antecedent;
    private boolean succedent;    
    private PosInOccurrence posInOcc=null;
    private boolean constraint=false;
    private boolean cfma=false;
    private boolean javablock=false;

    private Range firstJavaStatementRange = null;
    /**
     * creates a PosInSequent that points to the whole sequent.
     */
    public static PosInSequent createSequentPos() {
	return new PosInSequent(null, false, false, true, false, false, false);
    }

    /**
     * creates a PosInSequent that points to the antecedent.
     */
    public static PosInSequent createAntecedentPos() {
	return new PosInSequent(null, false, false, false, true, false, false); 
    }

    /**
     * creates a PosInSequent that points to the succedent.
     */
    public static PosInSequent createSuccedentPos() {
	return new PosInSequent(null, false, false, false, false, true, false); 
    }

    /**
     * creates a PosInSequent that points to a ConstrainedFormula described by
     * a PosInOccurrence. Additionally two booleans indicate whether the
     * constraint or the whole ConstrainedFormula is meant.
     * @param posInOcc the PositionInOccurrence describing the
     * ConstrainedFormula and maybe a subterm of its formula.
     * @param constraint true iff the constraint of the ConstrainedFormula is
     * meant.
     * @param cfma true iff the whole ConstrainedFormula (i.e. formula AND
     * constraint) is meant.
     */
    public static PosInSequent createCfmaPos(PosInOccurrence posInOcc, 
					     boolean constraint, 
					     boolean cfma) {
	return new PosInSequent(posInOcc, constraint, cfma, false, false, 
				false, false);
    }

    public static PosInSequent createJavaBlockPos(PosInOccurrence posInOcc) {
	return new PosInSequent(posInOcc, false, false, false, false, 
				false, true);
    }

    // use the create... above for getting instances of PosInSequent
    private PosInSequent(PosInOccurrence posInOcc, boolean constraint, 
			 boolean cfma, boolean sequent, boolean antecedent, 
			 boolean succedent, boolean javablock){
	this.posInOcc=posInOcc;
	this.constraint=constraint;
	this.cfma=cfma;
	this.sequent=sequent;
	this.antecedent=antecedent;
	this.succedent=succedent;
	this.javablock=javablock;
    }

    /**
     * sets the bounds, i.e. the start and end positions of the PosInSequent
     * in a string representation of a sequent.
     * @param s the start position in a string representation of a sequent
     * @param e the end position in a string representation of a sequent
     */
    public void setBounds(int s, int e) {
	start=s;
	end=e;
    }


    /**
     * returns the start position in a string representation of a sequent
     * @return start position
     */
    public int getStart() {
	return start;
    }

    /**
     * returns the end position in a string representation of a sequent
     * @return end position
     */
    public int getEnd() {
	return end;
    }

    /**
     * sets the bounds, i.e. the start and end positions of the first Java statement, 
     * of a corresponding Java program in a string representation of the sequent.
     * @param r the range for the first statement in the corresponding program
     */
    public void setFirstJavaStatementRange(Range r) {
	firstJavaStatementRange = r;
    }

    /**
     * returns the bounds, i.e. the start and end positions of the first Java statement, 
     * of a corresponding Java program in a string representation of the sequent.
     * @return the range specifying the first statement in the corresponding program
     */
    public Range getFirstJavaStatementRange() {
	return firstJavaStatementRange;
    }
    

    /**
     * returns the PosInOccurrence if the PosInSequent marks a
     * ConstrainedFormula or parts of it, null otherwise.
     */	
    public PosInOccurrence getPosInOccurrence() {
	return posInOcc;
    }
	
    /**
     * returns true iff the PosInSequent points to the constraint of a 
     * ConstrainedFormula
     */
    public boolean isConstraint() {
	return constraint;
    }
	
    /**
     * returns true iff the PosInSequent points to a whole 
     * ConstrainedFormula
     */
    public boolean isConstrainedFormula() {
	return cfma;
    }

    /**
     * returns true if the PosInSequent points to a whole Sequent
     */
    public boolean isSequent() {
	return sequent;
    }

    /**
     * returns true if the PosInSequent points to a whole Antecedent
     */
    public boolean isAntecedent() {
	return antecedent;
    }

    public boolean isJavaBlock() {
	return javablock;
    }
    
    /**
     * returns true if the PosInSequent points to a whole Succedent
     */
    public boolean isSuccedent() {
	return succedent;
    }
	
    /**
     * returns a string representation of this PosInSequent
     */
    public String toString() {
	if (isSequent()) return "Whole Sequent";
	if (isAntecedent()) return "Whole Antecedent";
	if (isSuccedent()) return "Whole Succedent";
	if (isJavaBlock()) return "JavaBlock of "+posInOcc + 
			          "\n Range of first statement == " + firstJavaStatementRange ;
	if (isConstraint()) return "Constraint of "+posInOcc;
	if (isConstrainedFormula()) return "Constraint Formula: "+posInOcc;
	return ""+posInOcc;
    }
	
}
