// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;
import de.uka.ilkd.key.logic.*;
import java.util.LinkedList;
import java.util.Iterator;

/**
 * A PositionTable describes the start and end positions of substrings of
 * a String in order to get a PosInSequent from an int describing
 * a position in a string representing a Term or a Sequent, etc. A 
 * PositionTable therefore
 * represents a table consisting of two columns of type int (start and
 * end position) and a reference to another PositionTable representing
 * the position information for that substring.
 * A PositionTable is valid (in order to support efficient putting of new
 * entries to the table and an efficient search for a Position, if the
 * last entry that has been set is (s, e, x) and the next entry is (s', e', x')
 * with s'>e.
 * 
 */
public class PositionTable {

    // the start positions of the direct subterms (or parts of sequent, etc.)
    protected int[] startPos;

    //the end positions of the direct subterms (or parts of sequent, etc.)   
    protected int[] endPos;

    // the PositionTables for the direct subterms (or parts of sequent, etc.)
    protected PositionTable[] child;    

    // the entry number where the next entry in endPos will be set
    private int currentEndEntry=0;

    // the entry number where the next entry in startPos will be set
    private int currentStartEntry=0;

    // the number of rows in the above arrays. Equals the number of direct 
    // subterms (or parts of sequent, etc.)
    private final int rows;

    /**
     * creates a new PositionTable with the number of subterms (or number of
     * ConstrainedFormula in a Semisequent, or the number of Semisequents in a
     * Sequent, etc.)
     * @param rows the number of direct sub-elements in the term whose
     * position information is represented by the constructed object.
     */
    public PositionTable(int rows) {
	this.rows=rows;
	startPos=new int[rows];
	endPos=new int[rows];
	child=new PositionTable[rows];
	for (int i=0; i<rows; i++) {
	    startPos[i]=-1;
	    endPos[i]=-1;
	    child[i]=null;
	}
    }

    /**
     * returns the m with startPos[m]<=index<=endPos[m]. 
     * -1 if no such m exists.
     */
    private int searchEntry(int index) {

	//linear search:
	for (int m=0; m<rows; m++) {
	    if ((startPos[m]<=index) && (endPos[m]>=index)) {
		return m;
	    }
	}


	//binary search (ordered arrays are precondition!), NOT CHECKED SO FAR:
/*	int l=0;
	int r=rows-1;
	int m;
	while (r<l) { 
	    m=(l+r)/2; 
	    if ((startPos[m]<=index) && (endPos[m]>=index)) {
		return m; 
	    } 
	    if (index<startPos[m]) {
		r=m; 
	    } else { 
		l=m; 
	    }
	    }*/
	return -1;
    }

    /**
     * returns a LinkedList of Integers describing the position in a sequent
     * starting from this PositionTable. A list (i0,i1,i2) is added to the
     * given LinkedList and returned if i0 is within startPos[i0] and
     * endPos[i0] AND i1 is within child[i0].startPos[i1] and
     * child[i0].endPos[i1] AND i2 is within child[i0].child[i1].startPos[i2] 
     * and
     * child[i0].child[i1].endPos[i2]. "within" means with respect to
     * startPositions of the fromer childs,
     * i.e. x is within A=child[i0].child[i1].startPos[i2] and 
     * B=child[i0].child[i1].endPos[i2] if
     * A+offset <= x <=B+offset, 
     * with offset=child[i0].startPos[i1]+startPos[i0].
     */
    protected LinkedList getPositionList(int index, LinkedList posList) {
	int sub=searchEntry(index);
	if (sub!=-1) {
	    posList.add(new Integer(sub));	    
	    return child[sub].getPositionList(index-startPos[sub], posList);
	}
	return posList;
    }

    /**
     * sets end in the position table to the next free end entry in the 
     * position
     * table and sets the given PositionTable as child of the sub-element
     * finished by putting this end position
     * @param end char position that ends the sub-element started by the 
     * corresponding start entry in the position table
     * @param child PositionTable for the sub-element from start to end
     */
    public void setEnd(int end, PositionTable child) {
	endPos[currentEndEntry]=end;
	this.child[currentEndEntry]=child;
	currentEndEntry++;
    }

    /**
     * sets start in the position table to the next free start entry in the 
     * position table 
     * @param start char position that starts a sub-element
     */
    public void setStart(int start) {
	startPos[currentStartEntry]=start;
	currentStartEntry++;
    }

    /**
     * Return of the children of this PositionTable
     */
    public PositionTable getChild(int i) {
	
	return child[i];
    }

    /**
     * returns a String representation of the position table
     */
    public String toString() {
	String result="[";
	for (int i=0; i<rows; i++) {
	    result=result+"<"+startPos[i]+","+endPos[i]+","+child[i]+">";
	    if (rows-1!=i) result=result+",";
	}
	return result+"]";
    }




    // This stuff has to do with highlighting the first executable 
    // statement in a possible program that corresponds to this term.
    // THIS SHOULD BE REMOVED in the future ...
    
    Range firstStatementRange = null;
    
    public void setFirstStatementRange(int start, int end){
	firstStatementRange = new Range(start,end);
    }
    
    public void setFirstStatementRange(Range r){
	firstStatementRange = r;
    }
    
    
    public Range getFirstStatementRange(){
	return firstStatementRange;
    }

}


