// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.gui;

/** this class is used to create a list from which the user can select the proof
 * he/she wants to work with
 */

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import java.util.Vector;

import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.Sequent;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.ProofTreeListener;
import de.uka.ilkd.key.proof.ProofTreeAdapter;
import de.uka.ilkd.key.proof.ProofTreeEvent;

public class ProofList extends JList {

    /** the KeYMediator */
    private KeYMediator mediator;
    
    /** listener for mouse events of this gui component */
    private MouseListener mouseListener = new ProofListMouseListener();

    /** listener to the prof tree events */
    private ProofTreeListener proofTreeListener = new ProofListProofTreeListener();

    /** the list model to be used */
    private ProofListModel model = new ProofListModel();

    /** creates a  prooflist filles with proofs. The entries shown in the list 
     * are the names of the proofs or if a proof has no name it is the sequent
     * of its root.
     */
    public ProofList(KeYMediator mediator) {
	super();
	this.mediator = mediator;
	setModel(model);   
	setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
	setCellRenderer(new ProofListIconCellRenderer());
//	setSelectedIndex(0);
//	proofChosen();
	addMouseListener(mouseListener);
    }
    
    
    public void addProof(Proof proof) {
        model.add(proof);
        proof.addProofTreeListener(proofTreeListener);
    }
    

    public void updateUI() {
	super.updateUI();
	Font myFont = UIManager.getFont(Config.KEY_FONT_PROOF_LIST_VIEW);
	if (myFont != null) {
	    setFont(myFont);
	} else {
	    Debug.out(Config.KEY_FONT_PROOF_LIST_VIEW + 
		      " not available, use standard font.");
	}
    }

    /** get the KeYMediator */
    private KeYMediator mediator() {
	return mediator;
    }


    /** returns the n-th proof of the list 
     * @param n index selecting the n-the proof
     * @return the n-th proof of the list 
     */
    public Proof getProof(int n) {
	return model.getProof(n);
    }

    /** returns the selected proof 
     * @return the selected proof
     */
    public Proof getSelectedProof() {
	return getProof(getSelectedIndex());
    }
    

    /** called if the user has clicked on a proof */
    private void proofChosen() {
	Proof proof = getProof(getSelectedIndex());
	if (proof != null && mediator != null) {
	    mediator.setProof(proof);
	}
    }


    
    /** inner class implementing the mouse listener that is responsible for 
     * this gui component
     */
    class ProofListMouseListener extends MouseAdapter {
	public void mouseClicked(MouseEvent e) {
	    if (e.getClickCount() == 1) {
		proofChosen();
	    }
	}
    }

    /** implements the list model */
    static class ProofListModel 
	extends AbstractListModel {

	private Vector data = new Vector();

	/**
	 * @return true iff the list contains the given proof 
	 */
	public boolean contains(Proof p) {
	    for (int i = 0; i<data.size(); i++) {
		if (data.get(i) == p) {
		    return true;
		}
	    }
	    return false;
	}

	/** adds a proof to the list */
	public void add(Proof p) {
	    if (p != null) {
		data.add(p);
		int index = data.indexOf(p);
		fireIntervalAdded(this, index, index);
	    }
	}

	public void remove(Proof p) {
	    int index = data.indexOf(p);
	    if (index > -1) {
		data.remove(p);
		fireIntervalRemoved(this, index, index+1);
	    }
	}

	public void clear() {
	    int size = data.size();
	    if (size > 0) {
		data.clear();	    
		fireIntervalRemoved(this, 0, size);
	    }
	}

	/** returns the i-th model element 
	 * @param i int the index of the model element
	 * @return the i-th model element 
	 */
	public Object getElementAt(int i) {
	    return data.get(i);
	}

	/** returns the i-th proof 
	 * @return the i-th proof 
	 */
	public Proof getProof(int i) {
	    return (Proof)data.get(i);
	}
	
	
	/** returns size of the model
	 * @return number of the list entries
	 */
	public int getSize() {
	    return data.size();
	}
	
    } // end of ProofListMouseListener

    
    /** a prooftree listener, so that it is known when the proof has closed 
     */
    class ProofListProofTreeListener extends ProofTreeAdapter 
	implements java.io.Serializable {
	
	/** invoked if all goals of the proof are closed
	 */
	public void proofClosed(ProofTreeEvent e) {
	    validate();
	}
	
	/** invoked if the list of goals changed (goals were added,
	 * removed etc.
	 */
	public void proofGoalRemoved(ProofTreeEvent e) {
	}
	
	/** invoked if the current goal of the proof changed */
	public void proofGoalsAdded(ProofTreeEvent e) {
	}
	
	/** invoked if the current goal of the proof changed */
	public void proofGoalsChanged(ProofTreeEvent e) {
	}
    } // end of prooftreelistener


    static class ProofListIconCellRenderer extends DefaultListCellRenderer 
	                          implements ListCellRenderer,
				             java.io.Serializable { 
	
	static final ImageIcon keyIcon=IconFactory.keyHole(20,20);
	static final ImageIcon keyClosedIcon=IconFactory.keyHoleClosed(20,20);
	
	public ProofListIconCellRenderer() {
	    setToolTipText("Proof");
	}
	
	public Component getListCellRendererComponent
	    (JList list,
	     Object value,            // value to display
	     int index,               // cell index
	     boolean isSelected,      // is the cell selected
	     boolean cellHasFocus)    // the list and the cell have the focus
	{
	    String valueStr = "unnamed";
	    Proof p = (Proof) value;
	    if (!"".equals(p.name().toString())) {
		valueStr=p.name().toString();
	    } else {
		Sequent seq = p.root().sequent();
		if (seq != null) {
		    SequentPrinter sp = new PureSequentPrinter
			(new ProgramPrinter(null), 
			 ((ProofList)list).mediator().getNotationInfo(), seq);
		    seq.prettyprint(sp);	    
		    valueStr = sp.toString().replace('\n',' ');
		} 
		
	    } 
	    DefaultListCellRenderer sup=(DefaultListCellRenderer)
		super.getListCellRendererComponent(list, valueStr, 
						   index, isSelected, 
						   cellHasFocus);
	    sup.setIcon(p.closed() ? keyClosedIcon : keyIcon);
	    return sup;
	}
    } // IconCellRenderer



} // end of ProofList


