// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;

import java.awt.*;
import java.awt.event.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.util.KeYResourceManager;
import de.uka.ilkd.key.logic.Sequent;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.rule.Taclet;
import de.uka.ilkd.key.rule.Rule;
import de.uka.ilkd.key.rule.RuleApp;
import de.uka.ilkd.key.logic.Name;

public class ProofTreeView extends JPanel {


    private JCheckBoxMenuItem viewMenuItem;

    /** the mediator is stored here */
    private KeYMediator mediator;

    /** The JTree that is used for actual display and interaction */
    private JTree delegateView;

    /** the model that is displayed by the delegateView */
    private GUIProofTreeModel delegateModel;

    /** the proof this view shows */
    private Proof proof;

    /** listener */
    private GUIProofTreeProofListener proofListener;
    private GUITreeSelectionListener treeSelectionListener;

    /** creates a new proof tree */
    public ProofTreeView (KeYMediator mediator) {

	proofListener = new GUIProofTreeProofListener();
	delegateView = new JTree(
		     new DefaultMutableTreeNode("No proof loaded"));

	delegateView.setInvokesStopCellEditing(true);
	delegateView.getSelectionModel().setSelectionMode
	    (TreeSelectionModel.SINGLE_TREE_SELECTION);
	treeSelectionListener = new GUITreeSelectionListener(); 
	delegateView.addTreeSelectionListener(treeSelectionListener);
	delegateView.setScrollsOnExpand(true);

	
	MouseListener ml = new MouseAdapter() {
		public void mousePressed(MouseEvent e) {
		    if (e.isPopupTrigger()) {
			TreePath selPath = delegateView.getPathForLocation
			    (e.getX(), e.getY());
			if (selPath!=null && selPath.getLastPathComponent() 
			                     instanceof GUIProofTreeNode) {
			    JPopupMenu popup = new ProofTreePopupMenu
				(((GUIProofTreeNode)selPath.getLastPathComponent())
				 .node());
			    popup.show(e.getComponent(),
				       e.getX(), e.getY());
			}
		    }
		}
	    };
	
	delegateView.addMouseListener(ml);

	setMediator(mediator);

	UIManager.addPropertyChangeListener(
	    new PropertyChangeListener() {
		    public void propertyChange(PropertyChangeEvent e) {
			if (Config.KEY_FONT_PROOF_TREE.
			    equals(e.getPropertyName())) {
			    setProofTreeFont();
			}
		    }
		});

	setProofTreeFont();
	delegateView.setLargeModel(true);

	updateUI();
	this.setLayout(new BorderLayout());
	this.add(delegateView,BorderLayout.CENTER);
	layoutKeYComponent();

	Proof selProof = this.mediator.getSelectedProof();
	if (selProof != null) {
	    setProof(selProof);
	}
	
    }

    private void setProofTreeFont() {
	Font myFont = UIManager.getFont(Config.KEY_FONT_PROOF_TREE);
	if (myFont != null) {
	    delegateView.setFont(myFont);
	} else {
	    Debug.out("KEY-PROOF_TREE_FONT not available, " +
		      "use standard font.");
	}
    }

    /** layout the component */
    protected void layoutKeYComponent() {
	delegateView.setBackground(Color.white);
	ProofRenderer renderer= new ProofRenderer();	
	delegateView.setCellRenderer(renderer);
	delegateView.putClientProperty("JTree.lineStyle", "Angled");
	delegateView.setVisible(true);	
    }


    /** returns the mediator to communicate with the model 
     * @return the mediator to communicate with the model 
     */
    public KeYMediator mediator() {
	return mediator;
    }

    /** sets the mediator to communicate with the model */
    private void setMediator(KeYMediator m) {
	if ( mediator != null )
	    unregister ();
	mediator = m;
	register();
    }

    private void register() {
	mediator.addKeYSelectionListener(proofListener);
	mediator.addInteractiveProofListener(proofListener);
	mediator.addGUIListener(new GUIProofTreeGUIListener());
    }

    private void unregister() {
	mediator.removeKeYSelectionListener(proofListener);
	mediator.removeInteractiveProofListener(proofListener);
	mediator.removeGUIListener(new GUIProofTreeGUIListener());
    }


    public void removeNotify () {
	unregister ();
	try{ delegateModel.unregister (); 
        } catch(NullPointerException e) {
        }
	super.removeNotify ();
    }


    /** sets up the proof tree view if a proof has been loaded 
     * @param p the Proof that has been loaded
     */
    private void setProof(Proof p) {
	if (delegateModel != null) {
	    delegateModel.unregister();
	}

	proof = p;

	delegateModel = new GUIProofTreeModel(proof);
	delegateView.setModel(delegateModel);

    }


    /** moves the scope of the tree view to the given node so that it
     *	is visible  
     */
    public void makeNodeVisible(Node n) {
 	GUIProofTreeNode node = delegateModel.getProofTreeNode(n);	
 	if (node==null) return;
 	Object[] obs=node.getPath();
 	TreePath tp = new TreePath(obs);
	treeSelectionListener.ignoreChange = true;
 	delegateView.getSelectionModel().setSelectionPath(tp);
	delegateView.scrollPathToVisible(tp);
 	delegateView.validate();
	treeSelectionListener.ignoreChange = false;
    }


    // INNER CLASSES 

    // listens to gui events
    class GUIProofTreeGUIListener implements GUIListener,
                                             java.io.Serializable {
	/** invoked if a frame that wants modal access is opened */
	public void modalDialogOpened(GUIEvent e) {
	    delegateView.setEnabled(false);
	}

	/** invoked if a frame that wants modal access is closed */
	public void modalDialogClosed(GUIEvent e) {
	    delegateView.setEnabled(true);
	}
	public void shutDown(GUIEvent e) {

	}
    }    

    class GUIProofTreeProofListener implements InteractiveProofListener, 
	                                       KeYSelectionListener {
	
	/** node of the last known current goal */
	private Node lastGoalNode;

	/** makes selected node visible of lastGoalNode */
	public void makeSelectedNodeVisible(Node selectedNode) {
	    if (selectedNode != null) {
		lastGoalNode = selectedNode;
	    }	    
	    makeNodeVisible(lastGoalNode);
	    delegateView.validate();
	}
	
	/** invoked when a rule has been applied */
	public void ruleApplied(ProofEvent e) {
	}

	/** focused node has changed */
	public void selectedNodeChanged(KeYSelectionEvent e) {	    
	    makeSelectedNodeVisible(mediator.getSelectedNode());	    
	}

	/** the selected proof has changed (e.g. a new proof has been
	 * loaded) */ 
	public void selectedProofChanged(KeYSelectionEvent e) {
	    Debug.out("ProofTreeView: initialize with new proof");
	    lastGoalNode = null;
	    setProof(e.getSource().getSelectedProof());
	    delegateView.validate();	    
	}
		
	/** invoked if automatic execution has started
	 */
	public void autoModeStarted(ProofEvent e) {	    
	    if (delegateModel == null) {
                Debug.out("delegateModel is null");
                return;
	    }
	    if (delegateModel.isAttentive()) {
		mediator.removeKeYSelectionListener(proofListener);
	    }
	    delegateModel.setAttentive(false);
	}
	
	/** invoked if automatic execution of heuristics has stopped
	 */
	public void autoModeStopped(ProofEvent e) {
            if (mediator.getSelectedProof() == null) return; // no proof (yet)
	    delegateView.removeTreeSelectionListener(treeSelectionListener);
	    if (delegateModel == null) {
		setProof(mediator.getSelectedProof());
	    } 
	    if (!delegateModel.isAttentive()) {
		delegateModel.setAttentive(true);	    
		mediator.addKeYSelectionListener(proofListener);
	    }
 	    makeSelectedNodeVisible(mediator.getSelectedNode());
	    delegateView.addTreeSelectionListener(treeSelectionListener);
 	    delegateView.validate();
	}
    }    

    class GUITreeSelectionListener implements TreeSelectionListener,
                                              java.io.Serializable {
	// hack to reduce duplicated repaints
	public boolean ignoreChange = false;

	public void valueChanged(TreeSelectionEvent e) {
	    if ( ignoreChange )
		return;
	    if (e.getNewLeadSelectionPath()==null) {
		return;
	    }
	    GUIAbstractTreeNode treeNode = 
		((GUIAbstractTreeNode)e.getNewLeadSelectionPath().
		 getLastPathComponent());
	    if (treeNode instanceof GUIProofTreeNode) {		
		Node node = ((GUIProofTreeNode)treeNode).node();
		Goal selected = proof.getGoal(node);
		if (selected != null) {
		    mediator.goalChosen(selected);
		} else {
		    mediator.nonGoalNodeChosen(node);
		}
	    } 
	    if (treeNode instanceof GUIBranchNode) {
		delegateView.setEditable(true);
	    } else {
		delegateView.setEditable(false);
	    }
	}
    }

    class ProofRenderer extends DefaultTreeCellRenderer 
	                implements TreeCellRenderer,
			           java.io.Serializable {

    

	/** this method calculates the text displayed in the trees node
	 * it abbreviates the string by replacing some chars in the middle 
	 * with exact one ~
	 */
	private StringBuffer calculateDisplayedText
	    (StringBuffer seq_string_buffer, int max_display_width,
	     FontMetrics metric) {
	    
	    String abbreviation_string = "..~..";
	    StringBuffer seq_text = seq_string_buffer;

	    int seq_text_width = metric.stringWidth(abbreviation_string);
	    for (int i = 0; i < seq_text.length(); i++) {
		seq_text_width += metric.charWidth(seq_text.charAt(i));
	    }

	    int skip_from = seq_text.length() / 2;
	    int skip_to   = skip_from;
	    while (seq_text_width >= max_display_width && 
		   !(skip_from == 0 && 
		     skip_to == seq_text.length())) {

		skip_from = (skip_from < 1 ? 0 : skip_from - 1); 
		skip_to   = (skip_to > seq_text.length() - 1 ? 
			     seq_text.length() : skip_to + 1); 

		seq_text_width = seq_text_width - 
		    metric.charWidth(seq_text.charAt(skip_from + 1));

		seq_text_width = seq_text_width - 
		    metric.charWidth(seq_text.charAt(skip_to - 1));
	    }

	    return (skip_from != skip_to ? 
		    seq_text.replace(skip_from, skip_to, 
				     abbreviation_string) :
		    seq_text);
	}


	public Component getTreeCellRendererComponent(JTree tree,
						      Object value,
						      boolean sel,
						      boolean expanded,
						      boolean leaf,
						      int row,
						      boolean hasFocus) {
	    if (!(value instanceof GUIProofTreeNode)) {
		super.getTreeCellRendererComponent
		    (tree, value, sel, expanded, leaf, row, hasFocus);
		if (value instanceof GUIBranchNode)
		    if ( mediator ().getUserConstraint ().displayClosed
			 ( ((GUIBranchNode)value).getNode() ) ) {
			// all goals below this node are closed
			this.setIcon(IconFactory.provedFolderIcon());
		    } else if ( ((GUIBranchNode)value).getNode().getBranchSink ()
				.getResetConstraint ().isSatisfiable () ) {
			// a instantiation of the metavariables does
			// exist that makes the remaining goals of
			// this subtree closed
			this.setIcon(IconFactory.closableFolderIcon());			
		    }
		return this;
	    }	    
	    Node node = ((GUIProofTreeNode)value).node();

	    String nodeText = "";
	    
	    RuleApp rap = node.getAppliedRuleApp();
	    if (rap != null) {
		Rule r =  rap.rule();
		if(r != null) {
		    nodeText = r.displayName();
		    if (nodeText == null) {
			nodeText = "rule without name";
		    }
		} else {
		    nodeText = "rule application without rule";
		}
	    } else {
		Goal goal = proof.getGoal(node);
		if(goal == null)
		    nodeText = "Closed goal";
		else
		    nodeText = "OPEN GOAL";
	    }

	    DefaultTreeCellRenderer tree_cell =  
		(DefaultTreeCellRenderer) super.getTreeCellRendererComponent
		(tree, nodeText, sel, expanded, leaf, row, hasFocus);
		
	    tree_cell.setFont(tree.getFont());	    	    
	    tree_cell.setText(nodeText);

	    if (node.leaf()) {
		Goal goal = proof.getGoal(node);
		if ( goal == null ||
		     mediator ().getUserConstraint ().displayClosed ( node ) ) {
		    tree_cell.setForeground(Color.green);
		    tree_cell.setIcon(IconFactory.keyHoleClosed(20,20));
		    ProofTreeView.this.setToolTipText("Closed Goal");
		    tree_cell.setToolTipText("A closed goal");
		} else {
		    if ( goal.getClosureConstraint ().isSatisfiable () ) {
			tree_cell.setForeground(Color.blue);
			tree_cell.setIcon(IconFactory.keyHole(20, 20));
			ProofTreeView.this.setToolTipText("Closable Goal");
			tree_cell.setToolTipText("An goal that can be closed");
		    } else {
			tree_cell.setForeground(Color.red);
			ProofTreeView.this.setToolTipText("Open Goal");
			tree_cell.setToolTipText("An open goal");
		    }
		    tree_cell.setIcon(IconFactory.keyHole(20, 20));
		}
	    } else {
		/*
		if ( node.getBranchSink ().getResetConstraint ().isSatisfiable () )
		    tree_cell.setForeground(Color.blue);
		else
		*/
		tree_cell.setForeground(Color.black);
		tree_cell.setIcon(null);
		tree_cell.setToolTipText("An inner node of the proof");
	    }

	    if (sel) {
		tree_cell.setBackground(Color.blue);
	    } else {
		tree_cell.setBackground(Color.white);
	    }	    
	    
	    return tree_cell;
	}
    }

    class ProofTreePopupMenu extends JPopupMenu 
	implements ActionListener {	
	
	private JMenuItem goalBack=new JMenuItem("Prune proof");
	private JMenuItem runHeuristics=
	    new JMenuItem("Apply Heuristics");
	
	private Node invokedNode;

	public ProofTreePopupMenu(Node node) {
	    super("Choose Action");
	    invokedNode = node;
	    create();
	}

	private void create() {	    
	    ((ProofTreePopupMenu)this).add(goalBack);
	    goalBack.addActionListener(this);
	    goalBack.setEnabled(false);
	    ((ProofTreePopupMenu)this).add(runHeuristics);
	    runHeuristics.addActionListener(this);
	    runHeuristics.setEnabled(false);
	    if (proof != null) {
		if (proof.activeHeuristics().size()>0) {
		    runHeuristics.setEnabled(true);
		}
		if (proof.isGoal(invokedNode) || 
		    proof.getSubtreeGoals(invokedNode).size()>0) {
		    goalBack.setEnabled(true);
		}
	    }
	}

	public void actionPerformed(ActionEvent e) {
	    if (e.getSource() == goalBack) {
		delegateModel.setAttentive(false);
		mediator().setBack(invokedNode);
		delegateModel.setAttentive(true);
	    } else if (e.getSource() == runHeuristics) {
		mediator().startHeuristics
		    (proof.getSubtreeGoals(invokedNode));
	    }	    
	}

    }



}
