// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.util.pp.*;
import de.uka.ilkd.key.logic.ListOfConstrainedFormula;
import de.uka.ilkd.key.logic.ListOfConstraint;

import java.util.Stack;
import java.util.StringTokenizer;
import java.io.IOException;
import java.util.ArrayList;
import java.util.ListIterator;
import java.util.List;

/**
 * The front end for the Sequent pretty-printer.  It prints a sequent
 * and its parts and computes the PositionTable, which is needed for
 * highliting.
 *
 * <P>The actual layouting/formatting is done using the {@link
 * de.uka.ilkd.key.util.pp.Layouter} class.  The concrete syntax for
 * operators is given by an instance of {@link NotationInfo}.  The
 * SequentPrinter is responsible for the concrete <em>layout</em>,
 * e.g. how terms with infix operators are indented, and it binds the
 * various needed components together.
 *
 * @see NotationInfo
 * @see Notation
 * @see de.uka.ilkd.key.util.pp.Layouter
 *
 *   
 */
public class SequentPrinter {

    /** The max. number of characters to put in one line */
    public static final int LINE_WIDTH = 55;
    
    /**
     * The ProgramPrinter used to pretty-print Java blocks in 
     * formulae.
     *
     * @link aggregation
     * @supplierCardinality 1 
     */
    private ProgramPrinter prgPrinter;

    /** Contains information on the concrete syntax of operators. */
    private NotationInfo notationInfo;

    /** The sequent we are pretty-printing */
    private Sequent            seq;
    private SequentPrintFilter filter;

    /** This chooses the layout. */
    private Layouter layouter;

    /** The backend <code>layouter</code> will write to. */
    private PosTableStringBackend backend;

    /** The constraint used for metavariable instantiations of the
     * current formula */
    private Constraint formulaConstraint = null;


    /**
     * Creates a SequentPrinter.  Sets the sequent to be printed, as
     * well as a ProgramPrinter to print Java programs and a
     * NotationInfo which determines the concrete syntax.
     *
     * @param prgPrinter   the ProgramPrinter that pretty-prints Java programs
     * @param notationInfo the NotationInfo for the concrete syntax
     * @param seq	    the sequent that this SequentPrinter will print
     */
    public SequentPrinter(ProgramPrinter prgPrinter, 
			  NotationInfo notationInfo, 
			  Sequent seq) {
	backend           = new PosTableStringBackend(LINE_WIDTH);
	layouter          = new Layouter(backend,2);
	this.prgPrinter   = prgPrinter;
	this.notationInfo = notationInfo;
	this.seq          = seq;
	this.filter       = null;
    }

    public SequentPrinter(ProgramPrinter prgPrinter, 
			  NotationInfo notationInfo, 
			  SequentPrintFilter filter) {
	backend           = new PosTableStringBackend(LINE_WIDTH);
	layouter          = new Layouter(backend,2);
	this.prgPrinter   = prgPrinter;
	this.notationInfo = notationInfo;
	this.seq          = null;
	this.filter       = filter;
    }
 
    public NotationInfo getNotationInfo(){
	return notationInfo;
    }

    /** Reprints the sequent.  This can be useful if settings like
     * PresentationFeatures or abbreviations have changed.
     */
    public void update() {
 	backend = new PosTableStringBackend(LINE_WIDTH);
 	layouter = new Layouter(backend,2);
 	prgPrinter.reset();
 	printSequent();
     }


    /**
     * Pretty-print a sequent.  The sequent <code>s</code> must be the
     * same as the one given to the constructor.
     */
    public void printSequent(Sequent s) {
	if (s==seq) {
	    printSequent();
	} else {
	    throw new IllegalStateException("SequentPrinter is initialized with"
					    +"another Sequent.");
	}
    }


    /**
     * Pretty-print the sequent given to the constructor.   
     * The sequent arrow is rendered as <code>=&gt;</code>.  If the 
     * sequent doen't fit in one line, a line break is inserted after each
     * formula, the sequent arrow is on a line of its own, and formulae
     * are indented w.r.t. the arrow.
     */
    public void printSequent() {
	try {
	    if ( seq == null ) {
		if ( filter != null ) {
		    int i;
		    markStartSub();
		    layouter.beginC(1).ind();
		    printSemisequent(filter.getAntec ());
		    markEndSub();
		    layouter.brk(1,-1).print("==>").brk(1);
		    markStartSub();
		    printSemisequent(filter.getSucc ());
		    layouter.brk(0).end();
		    markEndSub();
		}
	    } else {
		markStartSub();
		layouter.beginC(1).ind();
		printSemisequent(seq.antecedent());
		markEndSub();
		layouter.brk(1,-1).print("==>").brk(1);
		markStartSub();
		printSemisequent(seq.succedent());
		layouter.brk(0).end();
		markEndSub();	
	    }
	} catch (IOException e) {
	    throw new RuntimeException (
	      "IO Exception in pretty printer:\n"+e);
   	} catch (UnbalancedBlocksException e) {
	    throw new RuntimeException (
	      "Unbalanced blocks in pretty printer:\n"+e);
	}
    }

    /**
     * Pretty-prints a Semisequent.  Formulae are separated by commas.
     *
     * @param semiseq the semisequent to be printed
     */
    public void printSemisequent(Semisequent semiseq) 
	throws UnbalancedBlocksException,IOException
    {
	startTerm(semiseq.size());
	for (int i=0;i<semiseq.size();i++) {
	    markStartSub();
	    printConstrainedFormula(semiseq.get(i));
	    markEndSub();
	    if (i!=semiseq.size()-1) {
		layouter.print(",").brk(1);
	    }
	}
    }
    public void printSemisequent ( ListOfSequentPrintFilterEntry p_formulas )
	throws UnbalancedBlocksException,IOException
    {
	IteratorOfSequentPrintFilterEntry it   = p_formulas.iterator ();
	SequentPrintFilterEntry           entry;
	int                               size = p_formulas.size     ();
	startTerm(size);
        while ( size-- != 0 ) {
	    entry = it.next ();
	    markStartSub();
	    formulaConstraint = entry.getDisplayConstraint ();
	    printConstrainedFormula( entry.getFilteredFormula () );
	    markEndSub();
	    if ( size != 0 ) {
		layouter.print(",").brk(1);
	    }
	}
	formulaConstraint = null;
    }

    /**
     * Pretty-prints a constrained formula. The constraint
     * "Constraint.BOTTOM" is suppressed
     *
     * @param semiseq the constrained formula to be printed
     */
    public void printConstrainedFormula(ConstrainedFormula cfma) 
	throws UnbalancedBlocksException,IOException
    {
	if ( cfma.constraint().isBottom() ) {
	    startTerm(1);
	    markStartSub();
	    printTerm(cfma.formula());
	    markEndSub();
	} else {
	    layouter.beginC(0);
	    startTerm(2);
	    layouter.ind();
	    markStartSub();
	    printTerm(cfma.formula());	
	    markEndSub();
	    layouter.brk(1,3).print("<<").ind(1,6);
	    markStartSub();
	    printConstraint(cfma.constraint());
	    markEndSub();
	    layouter.end();
	}
    }
    
    /**
     * Pretty-prints an array expression
     *
     * @param pre usually a <code>[ </code> 
     * @param post usually a <code>] </code> 
     * @param t1 the array reference
     * @param t2 the index expression
     */
    public void printArray(String pre, String post, Term t1, Term t2)
	throws java.io.IOException {
	startTerm(2);
	markStartSub();
	printTerm(t1);
	markEndSub();
	layouter.print(pre);
	markStartSub();
	printTerm(t2);
	markEndSub();	
        layouter.print(post);
    }

    /**
     * Pretty-print a constraint
     *
     * This does currently only work well for "EqualityConstraint"s,
     * which are printed as a list of unifications. The bottom
     * constraint doesn't get special handling in here, i.e. this
     * method should not be called for p == Constraint.BOTTOM
     */
    public void printConstraint(Constraint p) 
	throws UnbalancedBlocksException,IOException
    {
	if ( p instanceof EqualityConstraint ) {

	    // Within the constraint metavariables should not be
	    // instantiated
	    Constraint frmC   = formulaConstraint;
	    formulaConstraint = null;

	    EqualityConstraint eqc = (EqualityConstraint)p;
	    List vars = new ArrayList ();
	    {
		IteratorOfMetavariable it =
		    eqc.restrictedMetavariables ();

		while ( it.hasNext () )
		    vars.add ( it.next () );
	    }

	    startTerm ( vars.size () );
	    layouter.print("[ ").beginI(0);

	    ListIterator it = vars.listIterator ();
	    Metavariable mv;
	    Term         inst;
	    while ( it.hasNext () ) {
		mv   = (Metavariable)it.next ();
		inst = eqc.getDirectInstantiation ( mv );
		if ( inst == null )
		    inst = TermFactory.DEFAULT.createFunctionTerm ( mv );

		markStartSub();
		printInfixTerm ( TermFactory.DEFAULT.createFunctionTerm ( mv ), 15,
				 "=",
				 inst, 10);
		markEndSub();
		
		if ( it.hasNext () ) {
		    layouter.print(",").brk(1,0);
		}
	    }
	    layouter.print(" ]").end();

	    formulaConstraint = frmC;

	} else
	    // Don't know how to pretty-print this constraint class
	    layouter.print ( p.toString () );
    }


    /**
     * Pretty-prints a term or formula.  How it is rendered depends on 
     * the NotationInfo given to the constructor.
     *
     * @param t the Term to be printed
     */
    public void printTerm(Term t)
	throws UnbalancedBlocksException,IOException
    {
	if(notationInfo.getAbbrevMap().isEnabled(t)){
	    startTerm(0);
	    layouter.print(notationInfo.getAbbrevMap().getAbbrev(t));
	} else {
	    notationInfo.getNotation(t.op()).print(t,this);
	}
    }

    /**
     * Pretty-prints a term or formula in the same block.  How it is
     * rendered depends on the NotationInfo given to the constructor.
     * `In the same block' means that no extra indentation will be
     * added if line breaks are necessary.  A formula <code>a &amp; (b
     * &amp; c)</code> would print <code>a &amp; b &amp; c</code>, omitting
     * the redundant parentheses.  The subformula <code>b &amp; c</code>
     * is printed using this method to get a layout of 
     *
     * <pre>
     *   a 
     * &amp; b 
     * &amp; c
     * </pre> 
     * instead of 
     * <pre> 
     *   a 
     * &amp;   b 
     *   &amp; c 
     * </pre>
     *
     *
     * @param t the Term to be printed */
    public void printTermContinuingBlock(Term t)
	throws UnbalancedBlocksException,IOException
    {
	notationInfo.getNotation(t.op()).printContinuingBlock(t,this);
    }


    /** Print a term in <code>f(t1,...tn)</code> style.  If the
     * operator has arity 0, no parentheses are printed, i.e.
     * <code>f</code> instead of <code>f()</code>.  If the term
     * doesn't fit on one line, <code>t2...tn</code> are aligned below
     * <code>t1</code>.
     * 
     * @param name the name to be printed before the parentheses.
     * @param t the term to be printed.  */
    public void printFunctionTerm(String name,
				  Term t) 
	throws UnbalancedBlocksException,IOException
    {
	startTerm(t.arity());
	layouter.print(name);
	if(t.arity()>0) {
	    layouter.print("(").beginC(0);
	    for (int i=0;i<t.arity();i++) {
		markStartSub();
		printTerm(t.sub(i));
		markEndSub();
		
		if (i<t.arity()-1) {
		    layouter.print(",").brk(1,0);
		}
	    }
	    layouter.print(")").end();
	}
    }

    /** Print a term in <code>f(t1,...tn)</code> style.  If it doesn't
     * fit on one line, <code>t2...tn</code> are aligned below t1.
     * Print a term in <code>o.q(t1,...tn)</code> style.
     *
     * @param name the name of the query
     * @param the term to be printed
     */
    public void printQueryTerm(String name,
			       Term t) 
	throws UnbalancedBlocksException,IOException
    {
	startTerm(t.arity());
	markStartSub();
	printTerm(t.sub(0));
	markEndSub();
	layouter.print(".").print(name).print("(").beginC(0);
	for (int i=1;i<t.arity();i++) {
	    markStartSub();
	    printTerm(t.sub(i));
	    markEndSub();
	    if (i<t.arity()-1) {
		layouter.print(",").brk(1,0);
	    }
	}
	layouter.print(")").end();
    }


     
    /** Print a unary term in prefix style.  For instance 
     * <code>!a</code>.  No line breaks are possible.
     *
     * @param name the prefix operator
     * @param t    the subterm to be printed
     * @param ass  the associativity for the subterm
     */
    public void printPrefixTerm(String name,
				Term t,int ass) 
	throws UnbalancedBlocksException,IOException
    {
	startTerm(1);
	layouter.print(name);
	maybeParens(t, ass);
    }


    /** Print a unary term in postfix style.  For instance 
     * <code>t.a</code>, where <code>.a</code> is the postfix operator.  
     * No line breaks are possible.
     *
     * @param name the postfix operator
     * @param t    the subterm to be printed
     * @param ass  the associativity for the subterm
     */
     public void printPostfixTerm(Term t,int ass,String name) 
	throws UnbalancedBlocksException,IOException
    {
	startTerm(1);
	maybeParens(t, ass);
	layouter.print(name);
    }

    /** Print a binary term in infix style.  For instance <code>p
     * &amp; q</code>, where <code>&amp;</code> is the infix
     * operator.  If line breaks are necessary, the format is like
     *
     * <pre>
     *   p
     * & q
     * </pre>
     *
     * The subterms are printed using 
     * {@link #printTermContinuingBlock(Term)}.
     *
     * @param l    the left subterm
     * @param assLeft associativity for left subterm
     * @param name the infix operator
     * @param r    the right subterm
     * @param assRight associativity for right subterm
     */
    public void printInfixTerm(Term l,int assLeft,
			       String name,
			       Term r,int assRight) 
	throws UnbalancedBlocksException,IOException
    {
	int indent = name.length()+1;
	layouter.beginC(indent);
	printInfixTermContinuingBlock(l,assLeft,
				      name,
				      r,assRight);
	layouter.end();
    }

    /** Print a binary term in infix style, continuing a containing
     * block.  See {@link #printTermContinuingBlock(Term)} for the
     * idea.  Otherwise like 
     * {@link #printInfixTerm(Term,int,String,Term,int)}.
     *
     * @param l    the left subterm
     * @param assLeft associativity for left subterm
     * @param name the infix operator
     * @param r    the right subterm
     * @param assRight associativity for right subterm
     * */
    public void printInfixTermContinuingBlock(Term l,int assLeft,
					      String name,
					      Term r,int assRight) 
	throws UnbalancedBlocksException,IOException
    {
	int indent = name.length()+1;
	startTerm(2);
	layouter.ind();
	maybeParens(l, assLeft);
	layouter.brk(1,-indent).print(name).ind(1,0);
	maybeParens(r, assRight);
    }

    /** Print a term with an update.  This looks like
     * <code>{loc1 := val1, ..., locn := valn} t</code>.  If line breaks are necessary, the
     * format is
     *
     * <pre>
     * {loc1:=val1, ..., locn:=valn}
     *   t
     * </pre>
     *
     * @param l       the left brace
     * @param asgn    the assignment operator (including spaces)
     * @param r       the right brace
     * @param t       the update term
     * @param ass1    associativity for the locations
     * @param ass2    associativity for the new values
     * @param ass3    associativity for phi
     */ 

    public void printUpdateTerm(String l, String asgn, String r,
				Term t, int ass1, int ass2, int ass3) 
	throws UnbalancedBlocksException,IOException
    {

	startTerm(t.arity());
	layouter.beginC(2).print(l);
	for (int i = 0; i<t.arity()-1; i+=2) {
	    maybeParens(t.sub(i), ass1);
	    layouter.print(asgn);
	    maybeParens(t.sub(i+1), ass2);
	    if (i < t.arity()-3) {
		layouter.print(",").brk(1,0);
	    }
	}
	layouter.print(r).brk(1);
	maybeParens(t.sub(t.arity()-1), ass3);
	layouter.end();
    }

    
    /** Print a conditional term.  This looks like
     * <code>{obj1 =? obj2}.a/->t</code>.  If line breaks are 
     * necessary, the format is
     *
     * <pre>
     * {   obj1
     *  =? obj2}.a
     *  /->t
     * </pre>
     *
     * @param l       the left brace
     * @param obj1    one object
     * @param ass1    associativity for obj1
     * @param eql     the equality check operator
     * @param obj2    the other object
     * @param ass2    associativity for obj2
     * @param r       the right brace
     * @param _then   the _then part of the object
     * @param bcms    the alternative operator <code>/-></code>
     * @param _else   the alternative term/formuala
     * @param ass3    associativity for _then and _else
     */ 
    public void printConditionalTerm(String l,
				     Term obj1,int ass1,
				     String eql,
				     Term obj2,int ass2,
				     String r, Term _then, 
				     String bcms, Term _else,
				     int ass3) 
	throws UnbalancedBlocksException,IOException
    {
	int ind = eql.length()+1;
	startTerm(4);
	layouter.print(l).beginC(ind).ind();
	maybeParens(obj1, ass1);
	layouter.brk(1,-ind).print(eql).ind(1,0);
	maybeParens(obj2, ass2);
	layouter.print(r).brk(0,-ind);
	maybeParens(_then, ass3);
	layouter.brk(1,-ind).print(bcms).ind(1,0);
	maybeParens(_else, ass3);
	layouter.end();
    }


    /** 
     * Print a conjunction conditional term.  This looks like
     * <code>a * j ?= (r_1*i_1: t_1, ..., r_n*i_n:t_n) : e </code>.  If
     * line breaks are necessary, the format is
     *
     * @param l       the left brace
     * @param conj    the conjunction operator (including spaces)
     * @param guard   the guard operator (including spaces)
     * @param branch  the branch operator ':'
     * @param t       the Term to be printed
     * @param r       the right brace
     * @param ass1    the associativity of the conjunction
     */ 
    public void printConjCondTerm(String l, 
				  String guard, 
				  String conj, 
				  String branch, 
				  Term t, String r,
				  int ass1) 
	throws UnbalancedBlocksException,IOException
    {
	startTerm(t.arity());

	layouter.beginC(2);
	maybeParens(t.sub(0), ass1);
	layouter.print(conj);
	maybeParens(t.sub(1), ass1);
	layouter.print(branch).brk(0,2);
	maybeParens(t.sub(2), ass1);	
	layouter.print(guard).brk(0,0);

	for (int i = 3; i<t.arity(); i+=3) {
	    maybeParens(t.sub(i), ass1);
	    layouter.print(conj).brk(0,2);
	    maybeParens(t.sub(i+1), ass1);
	    layouter.print(branch).brk(0,2);
	    maybeParens(t.sub(i+2), ass1);
	    if (i+3 < t.arity()) {
		layouter.brk(1,0).print(",");
	    }
	}


	layouter.end();
    }
    

    /** Print a substitution term.  This looks like
     * <code>{var/t}s</code>.  If line breaks are necessary, the
     * format is
     *
     * <pre>
     * {var/t}
     *   s
     * </pre>
     *
     * @param l       the left brace
     * @param v       the substituted variable
     * @param asgn    the assignment operator (including spaces)
     * @param t       the new value
     * @param ass2    associativity for the new value
     * @param r       the right brace
     * @param phi     the substituted term/formuala
     * @param ass3    associativity for phi
     */ 
    public void printSubstTerm(String l,
			       QuantifiableVariable v,
			       String asgn,
			       Term t,int ass2,
			       String r,
			       Term phi,int ass3) 
	throws UnbalancedBlocksException,IOException
    {
	layouter.beginC(2).print(l).print(v.name().toString());
	startTerm(2);
	layouter.print(asgn);
	maybeParens(t, ass2);
	layouter.print(r).brk(1);
	maybeParens(phi, ass3);
	layouter.end();
    }



    /** Print a quantified term.  Normally, this looks like
     * <code>all x:s.phi</code>.  If line breaks are necessary,
     * the format is
     *
     * <pre>
     * all x:s.
     *   phi
     * </pre>
     *
     * Note that the parameter <code>var</code> has to contain the
     * variable name with colon and sort.
     *
     * @param name the name of the quantifier
     * @param var  the quantified variable (+colon and sort)
     * @param sep  the separator (usually a dot)
     * @param phi  the quantified formula
     * @param ass  associativity for phi
     */
    public void printQuantifierTerm(String name,
				    String var,
				    String sep,
				    Term phi, int ass)
	throws UnbalancedBlocksException,IOException 
    {
	layouter.beginC(2);
	layouter.print(name).print(" ").print(var).print(sep);
	layouter.brk();
	startTerm(1);
	maybeParens(phi,ass);
	layouter.end();
    }

    

    /** Print a constant.  This just prints the string <code>s</code> and
     * marks it as a nullary term.
     *
     * @param s the constant
     */
    public void printConstant(String s) 
	throws UnbalancedBlocksException,IOException
    {
	startTerm(0);
	layouter.print(s);
    }


    /** Print a constant.  This just prints the string <code>s</code> and
     * marks it as a nullary term.
     *
     * @param s the constant
     */
    public void printMetavariable ( Metavariable p_mv )
	throws UnbalancedBlocksException,IOException
    {
	if ( formulaConstraint != null ) {
	    Term t = formulaConstraint.getInstantiation ( p_mv );
	    if ( t.op () != p_mv ) {
		printTerm ( t );
		return;
	    }
	}

	printConstant ( p_mv.name().toString() );
    }


    /**
     * Print a Java block.  This is formatted using the ProgramPrinter
     * given to the constructor.  The result is indented according to
     * the surrounding material.  The first `executable' statement is
     * marked for highlighting.
     *
     * @param j the Java block to be printed
     */
    public void printJavaBlock(JavaBlock j) 
	throws UnbalancedBlocksException,IOException
    {
	java.io.StringWriter sw=new java.io.StringWriter();
	prgPrinter.reset();
	prgPrinter.setWriter(sw);
	Range r=null;
	try {
	    j.program().prettyPrint(prgPrinter);
	    r = prgPrinter.getRangeOfFirstExecutableStatement();
	} catch (java.io.IOException e) {
	    layouter.print("ERROR");
	    System.err.println("Error while printing Java program \n"+e);
	    throw new RuntimeException(
                               "Error while printing Java program \n"+e);
	}
	layouter.mark(MARK_PRGPOSTBL);
	// send first executable statement range
	printMarkingFirstStatement(sw.toString(),r);

    }

    /** Print a string marking a range as first statement.  The range
     * <code>r</code> indicates the `first statement' character range in
     * string <code>s</code>.  This is sent to the layouter by decomposing
     * <code>s</code> into parts and using the appropriate 
     * {@link de.uka.ilkd.key.util.pp.Layouter#mark(Object)} calls.
     * This solves the problem that the material in <code>s</code> might
     * be further indented.
     *
     * @param s   the string containing a program
     * @param r   the range of the first statement
     */
    private void printMarkingFirstStatement(String s,Range r) 
	throws UnbalancedBlocksException,IOException
    {
	int iEnd   = r.end()<=s.length()?r.end():s.length();
	int iStart = r.start()<=iEnd?r.start():iEnd;
	String start = s.substring(0,iStart);
	String firstStmt = s.substring(iStart,iEnd);
	String end = s.substring(iEnd);
	layouter.beginC(0);
	printVerbatim(start);
	layouter.mark(MARK_START_FIRST_STMT);
	printVerbatim(firstStmt);
	layouter.mark(MARK_END_FIRST_STMT);
	printVerbatim(end);
	layouter.end();
    }
    
    /** Print a string containing newlines to the layouter.  This is like 
     * {@link de.uka.ilkd.key.util.pp.Layouter#pre(String)}, but
     * no block is opened.
     */
    private void printVerbatim(String s) 
	throws UnbalancedBlocksException,IOException
    {
	StringTokenizer st = new StringTokenizer(s,"\n",true);
	while(st.hasMoreTokens()) {
	    String line = st.nextToken();
	    if ("\n".equals(line)) {
		layouter.nl();
	    } else {
		layouter.print(line);
	    }
	}
    }

    /** Print a DL modality formula.    Normally, this looks like
     * <code>&lt;Program&gt;phi</code>.  No line breaks are inserted, as 
     * the program itself is always broken.
     */
    public void printModalityTerm(String left,
				  JavaBlock jb,
				  String right,
				  Term phi,
				  int ass) 
	throws UnbalancedBlocksException,IOException
    {
	startTerm(2);
	layouter.print(left);
	markStartSub();
	printJavaBlock(jb);
	markEndSub();

	// set the range of the first statement in the corresponding
	// diamond for the whole term also in order to garantuee the
	// highlightning of this statement ...
	layouter.mark(MARK_FIRST_STATEMENT_FROM_CHILD);

	layouter.print(right);
	maybeParens(phi, ass);
    }

    /**
     * Returns the pretty-printed sequent.  This should only be called
     * after {@link #printSequent()} returns.
     *
     * @return the pretty-printed sequent.
     */
    public String toString() {
	try {
	    layouter.flush();
	} catch (IOException e) {
	    throw new RuntimeException (
	      "IO Exception in pretty printer:\n"+e);
	}
	return backend.getString()+"\n";
    }

    /**
     * Returns the pretty-printed sequent in a StringBuffer.  This
     * should only be called after {@link #printSequent()} returns.
     *
     * @return the pretty-printed sequent.  */
    public StringBuffer result() {
	try {
	    layouter.flush();
	} catch (IOException e) {
	    throw new RuntimeException (
	      "IO Exception in pretty printer:\n"+e);
	}
	return new StringBuffer(backend.getString()).append("\n");
    }

    /**
     * Returns the sequent assigned to this SequentPrinter.
     *
     * @return the sequent assigned to this SequentPrinter.
     */
    public Sequent getSequent() {
	return seq;
    }    

    /**
     * returns the PositionTable representing position information on
     * the sequent of this SequentPrinter. Subclasses may overwrite
     * this method with a null returning body if position information
     * is not computed there.  
     */
    public InitialPositionTable getPositionTable() {
	return backend.getPositionTable();
    }

    /** Returns the ProgramPrinter
     * @return the ProgramPrinter
     */
    public ProgramPrinter programPrinter() {
	return prgPrinter;
    }

    /** Returns the Layouter
     * @return the Layouter
     */
    protected Layouter getLayouter() {
	return layouter;
    }


    /**
     * Prints a subterm, if needed with parentheses.  Each subterm has
     * a Priority. If the priority is less than the associativity for
     * that subterm fixed by the Notation/NotationInfo, parentheses
     * are needed.
     *
     * <p>If prio and associativity are equal, the subterm is printed
     * using {@link #printTermContinuingBlock(Term)}.  This currently
     * only makes a difference for infix operators.
     *
     * @param t   the the subterm to print
     * @param ass the associativity for this subterm */
    protected void maybeParens(Term t, int ass) 
	throws UnbalancedBlocksException,IOException
    {
	if (notationInfo.getNotation(t.op()).getPriority() < ass){	    
	    markStartSub();
	    layouter.print("(");   
	    printTerm(t);	   
	    layouter.print(")");
	    markEndSub();
	} else {
	    markStartSub();
	    if (notationInfo.getNotation(t.op()).getPriority() == ass) {
		printTermContinuingBlock(t);
	    } else {
		printTerm(t);
	    }
	    markEndSub();
	}
    }



    /** Mark the start of a subterm.  Needed for PositionTable construction.*/
    private static final Object MARK_START_SUB = new Object();
    /** Mark the end of a subterm.  Needed for PositionTable construction.*/
    private static final Object MARK_END_SUB = new Object();
    /** Mark the start of a term with a number of subterms.  
     * Needed for PositionTable construction.*/
    private static final Object MARK_START_TERM = new Object();
    /** Mark the need for a ProgramPositionTable.  
     * Needed for PositionTable construction.*/
    private static final Object MARK_PRGPOSTBL = new Object();
    /** Mark the start of the first executable statement.
     * Needed for PositionTable construction.*/
    private static final Object MARK_START_FIRST_STMT = new Object();
    /** Mark the end of the first executable statement.
     * Needed for PositionTable construction.*/
    private static final Object MARK_END_FIRST_STMT = new Object();
    /** Mark the opportunity to get the first statement range from a
     * child PositionTable.  Needed for PositionTable construction.*/
    private static final Object MARK_FIRST_STATEMENT_FROM_CHILD = new Object();

    /**
     * Called before a substring is printed that has its own entry in
     * a position table.  The method sends a mark to the layouter,
     * which will make the backend set a start entry in posTbl, push a
     * new StackEntry with the current posTbl and current pos on the
     * stack and set the current pos to the length of the current
     * string result. Subclasses may overwrite this method with an
     * empty body if position information is not needed there.
     */
    protected void markStartSub() {
	layouter.mark(MARK_START_SUB);
    }

    /**
     * Called after a substring is printed that has its own entry in a
     * position table.  The backend will finishes the position table
     * on the top of the stack and set the entry on the top of the
     * stack to be the current position/position table. Subclasses may
     * overwrite this method with an empty body if position
     * information is not needed there.
     */
    protected void markEndSub() {
	layouter.mark(MARK_END_SUB);
    }
    
    /**
     * Start a term with subterms.  The backend will set the current
     * posTbl to a newly created position table with the given number
     * of rows. Subclasses may overwrite this method with an empty
     * body if position information is not needed there.
     *
     * @param size the number of rows of the new position table
     */
    protected void startTerm(int size) {
	layouter.mark(new Integer(size));
    }


    /** Utility class for stack entries containing the position table
     * and the position of the start of the subterm in the result.  */
    private class StackEntry {

	PositionTable posTbl;
	int p;

	StackEntry(PositionTable posTbl, int p) {
	    this.posTbl=posTbl;
	    this.p=p;
	}
	
	PositionTable posTbl() {
	    return posTbl;
	}

	int pos() {
	    return p;
	}
	
    }

    /** A {@link de.uka.ilkd.key.util.pp.Backend} which puts its
     * result in a StringBuffer and builds a PositionTable.  Position
     * table construction is done using the {@link
     * de.uka.ilkd.key.util.pp.Layouter#mark(Object)} facility of the
     * layouter with the various static <code>MARK_</code> objects
     * declared {@link SequentPrinter}.
     */
    private class PosTableStringBackend extends StringBackend {

	/** The top PositionTable */
	private InitialPositionTable initPosTbl = 
	    new InitialPositionTable(2,1,1);

	/** The resulting position table or an intermediate result */
	private PositionTable posTbl= initPosTbl;
	
	/** The position in result where the current subterm starts */
	private int pos = 0;
	
	/** The stack of StackEntry representing the nodes above 
	 * the current subterm */
	private Stack stack = new Stack();

	/** These two remember the range corresponding to the first
	 * executable statement in a JavaBlock */
	private int firstStmtStart;
	private Range firstStmtRange;

	PosTableStringBackend(int lineWidth) {
	    super(lineWidth);
	}

	/** Returns the constructed position table.
	 *  @return the constructed position table 
	 */
	public InitialPositionTable getPositionTable() {
	    return initPosTbl;
	}

	/** Receive a mark and act appropriately.
	 */
	public void mark(Object o) {

	    // IMPLEMENTATION NOTE
	    //
	    // This if-cascade is really ugly.  In paricular the part
	    // which says <code>instanceof Integer</code>, which stand
	    // for a startTerm with given arity.
	    //
	    // The alternative would be to 1.: spread these
	    // mini-functionalties across several inner classes in a
	    // visitor-like style, effectively preventing anybody from
	    // finding out what happens, and 2.: allocate separate
	    // objects for each startTerm call to wrap the arity.
	    //
	    // I (MG) prefer it this way.

	    if ( o==MARK_START_SUB ) {
		posTbl.setStart(count()-pos);
		stack.push(new StackEntry(posTbl, pos));
		pos=count();
	    } else if ( o==MARK_END_SUB ) {
		StackEntry se=(StackEntry)stack.peek();
		stack.pop();
		pos=se.pos();
		se.posTbl().setEnd(count()-pos-1, posTbl);
		posTbl=se.posTbl();
	    } else if ( o instanceof Integer ) {
		// This is sent by startTerm
		posTbl = new PositionTable(((Integer)o).intValue());
	    } else if ( o==MARK_PRGPOSTBL ) {
		posTbl = new ProgramPositionTable(0);
	    } else if ( o==MARK_START_FIRST_STMT ) {
		firstStmtStart = count();
	    } else if ( o==MARK_END_FIRST_STMT ) {
		firstStmtRange = new Range(firstStmtStart, count());
		posTbl.setFirstStatementRange(firstStmtRange);
	    } else if ( o==MARK_FIRST_STATEMENT_FROM_CHILD ) {
		posTbl.setFirstStatementRange(firstStmtRange);
	    }
	}
    }
}
