// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.gui;

/** This class displays a DL-formula. It 
 * (a) highlights the subpart of the sequent that is nearest to the mouse
 * cursor 
 * (b) pops up a menu with all applicable Taclets at the highlighted
 * position 
 */
import java.awt.Font;
import java.awt.Color;
import java.awt.Point;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import java.awt.dnd.*;
import java.awt.datatransfer.*;

import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.util.Debug;


public class SequentView extends JEditorPane {

    // the current tag of the highlight
    private Object currentHighlight;

    private Object additionalHighlight;
    private Highlighter.HighlightPainter additionalHighlightPainter;

    private Object additionalHighlight2;
    private Highlighter.HighlightPainter additionalHighlightPainter2;

    // the used sequentprinter
    private SequentPrinter printer;

    // the used sequent filter
    private SequentPrintFilter filter;

    // the mediator
    private KeYMediator mediator;

    // the mouse/mouseMotion listener
    private SequentViewListener listener;
    
    // an object that detects opening and closing of an Taclet instantiation dialog
    private GUIListener guiListener;

    // enables this component to be a Drag Source
    DragSource dragSource = null;

    /** last registered mouse position */
    PosInSequent pos = PosInSequent.createSequentPos();

    /** 
     * creates a viewer for a sequent 
     * @param printer the Sequentprinter to use
     */
    public SequentView(KeYMediator mediator) {
	/* setting this to text/html causes the position tables to go wrong */
	super("text/plain","");
	setMediator(mediator);
	// view cannot be edited
	setEditable(false);
	// disables selection
	setSelectionColor(getBackground());
	// sets the painter for the highlightning
	setHighlighter(new DefaultHighlighter());	
	//sets initial highlight (not visible) and stores its tag
	try {
	    // REMARK: THE ORDER FOR ADDING HIGHLIGHTS IS CRUCIAL
	    // WHEN USING OVERLAPPING HIGHLIGHTS:
	    // Adding highlight H1 before highlight H2 ensures that 
	    // H1 can overwrite overlapping parts of H2 !!!

	    additionalHighlightPainter = 
		new DefaultHighlighter.DefaultHighlightPainter(Color.lightGray);
	    additionalHighlight = getHighlighter().
		addHighlight(0, 0, additionalHighlightPainter); 
	    
	   
	    currentHighlight = getHighlighter().
		addHighlight(0, 0,
			     new DefaultHighlighter.
				 DefaultHighlightPainter(Color.yellow)); 

	   
	} catch(BadLocationException badLocation) {
	    System.err.println("SequentView tried to highlight an area"
			       +"that is not existing.");
	    System.err.println("Exception:"+badLocation);
	}	

	Config.DEFAULT.addConfigChangeListener(
	    new ConfigChangeListener() {
		    public void configChanged(ConfigChangeEvent e) {
			updateUI();
		    }
		});
	setSequentViewFont();

	listener = new SequentViewListener();
	addMouseMotionListener(listener);
	addMouseListener(listener);

	
	guiListener = new GUIListener(){
		/** invoked if a frame that wants modal access is opened */
		public void modalDialogOpened(GUIEvent e){
		 

		    if (e.getSource() instanceof ApplyTacletDialog){
			// enable drag'n'drop ...
			listener.setDragNDropEnabled(true);
			listener.setRefreshHighlightning(true);

		    } else {
			// disable drag'n'drop and highlightning ...
			listener.setDragNDropEnabled(false);
			listener.setRefreshHighlightning(false);
		    }
		}
		
		/** invoked if a frame that wants modal access is closed */
		public void modalDialogClosed(GUIEvent e){
		    if (e.getSource() instanceof ApplyTacletDialog){
			// disable drag'n'drop ...
			listener.setDragNDropEnabled(false);
			listener.setRefreshHighlightning(true);
		    } else {
			listener.setRefreshHighlightning(true);
		    }
		}
		
		/** invoked if the user wants to abort the proving session */
		public void shutDown(GUIEvent e){
		}	
	    };
	
	
	mediator().addGUIListener(guiListener); 

	// and here comes the drag'n'drop stuff that allows the user to copy
	// the currently highlighted subterm/formula by mere drag'n'dop actions
	
	dragSource = new DragSource();
	
	dragSource.createDefaultDragGestureRecognizer( this, 
						       DnDConstants.ACTION_MOVE,
						       listener );
    }


    public void updateUI() {
	super.updateUI();
	setSequentViewFont();
    }

    public void setSequentViewFont() {
	Font myFont = UIManager.getFont(Config.KEY_FONT_CURRENT_GOAL_VIEW);
	if (myFont != null) {
	    setFont(myFont);
	} else {
	    Debug.out("KEY_FONT_CURRENT_GOAL_VIEW not available. Use standard font.");
	}
    }


    /** sets the text being printed */
    public synchronized void printSequent() {
	removeMouseMotionListener(listener);
	removeMouseListener(listener);
	printer.update();
	boolean errorocc;
	do {
	    errorocc=false;
	    try {
		setText(printer.toString());		
	    } catch (Error e) {
		System.err.println("Error occurred while printing Sequent!");
		errorocc=true;
	    }
	} while (errorocc);
	repaint();
	addMouseMotionListener(listener);
	addMouseListener(listener);
    }
        
    /** sets the SequentPrinter to use
     * @param sp the SequentPrinter that is used
     */
    public void setPrinter(SequentPrinter sp) {
	setPrinter ( sp, null );
    }

    /** sets the SequentPrinter to use
     * @param sp the SequentPrinter that is used
     * @param f the SequentPrintFilter that is used
     */
    public void setPrinter(SequentPrinter sp, SequentPrintFilter f) {
	printer = sp;
	filter  = f;
    }

    /** return used SequentPrinter
     * @returns the SequentPrinter that is used
     */
    public SequentPrinter printer() {
	return printer;
    }

    /** sets the mediator
     * @param mediator the KeYMediator used to communicate with other
     * components 
     */
    private void setMediator(KeYMediator mediator) {
	Debug.assertTrue(mediator != null, "Mediator must be set");
	this.mediator = mediator;
    }
    

    /** the startposition and endposition+1 of the string to be
     * highlighted
     * @param startPos offset where the substring to highlight
     * begins 
     * @param endPos offset where the substring to highlight ends +1
     * @param firstJavaStatementRange range of a possible Java statement
     *        to be additionally highlighted.
     */
    public void paintHighlight(int startPosition, int endPosition, Range firstJavaStatementRange) {
	try {	
	   
	    int javaStmtStartPos;
	    int javaStmtEndPos;
	   
	    if(firstJavaStatementRange != null) {
		// There is a Java statement that should additionaly be highlighted ...
		javaStmtStartPos = firstJavaStatementRange.start();
		javaStmtEndPos = firstJavaStatementRange.end();
	    } else {
		javaStmtStartPos = 0;
		javaStmtEndPos = 0;
	    }
	
	    // Change highlight for additional Java statement ...
	    getHighlighter().changeHighlight(additionalHighlight, javaStmtStartPos, javaStmtEndPos);
					     


	    // Change Highlighter for currently selected sequent part ...	
	    getHighlighter().changeHighlight(currentHighlight, startPosition,
	    				     endPosition);
		
	} catch(BadLocationException badLocation) {
	    System.err.println("SequentView tried to highlight an area"
			       +"that is not existing.");
	    System.err.println("Exception:"+badLocation);
	}	
    }

    /** returns the mediator of this view
     * @return the KeYMediator
     */
    public KeYMediator mediator() {
	return mediator;
    }

    /** selected rule to apply 
     * @param taclet the selected Taclet
     * @param pos the PosInSequent describes the position where to apply the 
     * rule 
     */
    void selectedTaclet(Taclet taclet, PosInSequent pos) {
	mediator().selectedTaclet(taclet, pos);
    }

    /**
     * Enables drag'n'drop of highlighted subterms in the sequent window.
     * @param allowDragNDrop enables drag'n'drop iff set to true.
     */
    synchronized public void setDragNDropEnabled(boolean allowDragNDrop){
	    listener.setDragNDropEnabled(allowDragNDrop);
	}
    
    /**
     * Checks whether drag'n'drop of highlighted subterms in the sequent window
     * currently is enabled..
     * @param return true iff drag'n'drop is enabled.
     */	
    synchronized public boolean dragNDropEnabled(){
	return listener.dragNDropEnabled();
    }
    
    public String getHighlightedText() {
       String s;
       try{
          s = getText(pos.getStart(),pos.getEnd()+1-pos.getStart());
       } catch (Exception ble) { // whatever it is -- forget about it
          s="";
       }
       return s;
    }



    /** reacts on mouse events to highlight the selected part of the
     * sequent and it pops up a menu showing all applicable Taclet 
     * at the highlighted position if the mouse is pressed
     *
     * Additionally it performs all necessary actions for draging some
     * highlighted sequent part into some other GUI component (e.g.
     * some Taclet rule instantiation dialog)
     */
    class SequentViewListener extends MouseInputAdapter 
                              implements DragGestureListener  {  
	private TacletMenu menu;
	private boolean refreshHighlightning;
	private boolean dragNDropEnabled;

	SequentViewListener() {
	    menu = new TacletMenu();
	    addMouseListener(this);
	    setRefreshHighlightning(true);
	    setDragNDropEnabled(false);
	}

	
	synchronized private PosInSequent getPosInSequent(int x, int y) {
	    // in the first half of the character the correct character,
	    // in the second half the floowing character is returned.
	    // Therefore get width of character before the returned
	    // one and substract it from the given x value. And get
	    // the character at the new position. That is the correct one. 
	    
	    String seqText = getText();
	    int cursorPosition = viewToModel(new Point(x,y));

	    // previous characterposition in string
	 
	    if (seqText.length() > 0) {
		cursorPosition -= ( cursorPosition > 0 ? 1 : 0);
	 
		cursorPosition = (cursorPosition >= seqText.length() 
				  ? seqText.length()-1 :  
				  cursorPosition);
		cursorPosition = ( cursorPosition >= 0 ? cursorPosition : 0);
		int previousCharacterWidth =
		    (SequentView.this).getFontMetrics(getFont()).charWidth
		    (seqText.charAt(cursorPosition));	    	    

		int characterIndex = viewToModel(new
		    Point( x - (previousCharacterWidth/2), y)); 


		PosInSequent posInSequent = printer().getPositionTable().
		    getPosInSequent(characterIndex, filter);

		return posInSequent;
	    }
	    
	    return null;
	}

	
	public void mouseMoved(MouseEvent me) {	    
	    if (me.getSource()==SequentView.this && 
		!menu.getPopupMenu().isVisible() && refreshHighlightning())  { 
		pos=getPosInSequent(me.getX(), me.getY()); 
		
		if (pos != null) {
		    paintHighlight(pos.getStart(), pos.getEnd()+1, 
                                   pos.getFirstJavaStatementRange() ); 
		} else {
		    paintHighlight(0, 0, null); 
		}
	    }
	}
	    
	public void mousePressed(MouseEvent me) {

	    if  (!menu.getPopupMenu().isVisible() && !dragNDropEnabled()) {
		pos = getPosInSequent(me.getX(), me.getY()); 
		if (mediator()!= null && pos != null) {
		    menu = new TacletMenu(SequentView.this,
					  mediator().getFindTaclet(pos), 
					  mediator().getRewriteTaclet(pos),
					  mediator().getNoFindTaclet(),
					  mediator().getBuiltInRule(pos),
					  pos);
		} else {
		    menu = new TacletMenu(SequentView.this,
					  SLListOfTaclet.EMPTY_LIST,
					  SLListOfTaclet.EMPTY_LIST,
					  SLListOfTaclet.EMPTY_LIST,
					  SLListOfBuiltInRule.EMPTY_LIST,
					  pos);
		}
		
		 menu.getPopupMenu().show(SequentView.this,
					 me.getX(),
					 me.getY());	
	    }
	}

	public void mouseReleased(MouseEvent me) {	
	    if  (!dragNDropEnabled()) {
		menu.setVisible(false);
	    }
	}


	public void hideMenu(){
	    menu.setVisible(false); 
	}


	synchronized public void setRefreshHighlightning(boolean doRefresh){
	    refreshHighlightning = doRefresh;
	}
	
	synchronized public boolean refreshHighlightning(){
	    return refreshHighlightning;
	}

	
	synchronized public void setDragNDropEnabled(boolean allowDragNDrop){
	    dragNDropEnabled = allowDragNDrop;
	}
	
	synchronized public boolean dragNDropEnabled(){
	    return dragNDropEnabled;
	}
	
	/**
	 * a drag gesture has been initiated
	 */
	
	public void dragGestureRecognized(DragGestureEvent dgEvent) {
	    if (dragNDropEnabled()){
	
		// prevent from changing the current highlightning (until the drag operation has been finished !)
		setRefreshHighlightning(false);
	       
		hideMenu();
		
		Point dragOrigin = dgEvent.getDragOrigin();
		PosInSequent pos = getPosInSequent((int) dragOrigin.getX(), (int) dragOrigin.getY());
		
		String subtermString = null;
		if (pos != null) {
		    // a subterm has been selected ...
		    try {
			subtermString = SequentView.this.getText(pos.getStart(), pos.getEnd()+1 - pos.getStart());		
		    } catch (BadLocationException ble){
			ble.printStackTrace();
		    }
		    
		}
		
		if ( subtermString != null ){
		    StringSelection text = new StringSelection( subtermString ); 
		    
		    // as the name suggests, starts the dragging
		    dragSource.startDrag (dgEvent, 
					  DragSource.DefaultCopyDrop, 
					  text, 
					  new DragSourceListener(){
						  
						  /**
						   * is invoked when you are dragging over the DropSite
						   * 
						   */
						  
						  public void dragEnter (DragSourceDragEvent event) {
						  }
						  
						  /**
						   * is invoked when you are exit the DropSite without dropping
						   *
						   */
						  
						  public void dragExit (DragSourceEvent event) {
						  }
						  
						  /**
						   * is invoked when a drag operation is going on
						   * 
						   */
						  
						  public void dragOver (DragSourceDragEvent event) {
						  }
						  
						  
						  /**
						   * this message goes to DragSourceListener, informing it that the dragging 
						   * has ended
						   * 
						   */
						  
						  public void dragDropEnd (DragSourceDropEvent event) {   
						      // Enable updating the subterm highlightning ...
						      setRefreshHighlightning(true);
						     
						      if ( event.getDropSuccess()){
							  // System.err.println("Drag and Dopped successfully !!!");
						      }
						      
						  }						     
						  
						  /**
						   * is invoked if the use modifies the current drop gesture
						   * 
						   */
						  
						  public void dropActionChanged ( DragSourceDragEvent event ) {
						  }
						  
					      });
		} else {
		    // System.out.println( "Nothing has been selected");   
		}
	    }
	}
	
    }
	
    
}






