// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.gui;

import de.uka.ilkd.key.logic.SimultaneousUpdateSimplifier;
import de.uka.ilkd.key.logic.DeletionStrategy;
import de.uka.ilkd.key.logic.DefaultDeletionStrategy;
import java.util.Properties;
import java.util.LinkedList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.StringTokenizer;

/** This class encapsulates the information about the active
 * Heuristics and the maximum amount of heuristics steps before an
 * interactive step is required.
 */
public class SimultaneousUpdateSimplifierSettings implements Settings {


    private static final String USE_SIM_UPDATES_KEY = "[SimultaneousUpdateSimplifier]UseSimultaneousUpdates";
    private static final String DELETE_SIMPLE_KEY = "[SimultaneousUpdateSimplifier]DeleteSimpleLocations";
    private static final String DELETE_UNUSED_KEY = "[SimultaneousUpdateSimplifier]DeleteUnusedLocations";
    private static final String IMPLICIT_MODE_KEY = "[SimultaneousUpdateSimplifier]Implicit Execution";

    private boolean use_sim_updates = true;
    private boolean delete_simple_updates = true;
    private boolean delete_unused_updates = true;
    private boolean implicit_execution = true;

    private SimultaneousUpdateSimplifier sus;


    private LinkedList listenerList = new LinkedList();

    public SimultaneousUpdateSimplifierSettings() {
	updateSimplifier();
    }
    
    // 
    private void updateSimplifier() {
	sus = new SimultaneousUpdateSimplifier
	    (new DefaultDeletionStrategy(delete_simple_updates,
					 delete_unused_updates),
	     use_sim_updates);
    }

    // getter
    public boolean useSimultaneousUpdates() {
	return use_sim_updates;
    }


    public boolean deleteSimpleUpdates() {
	return delete_simple_updates;
    }

    public boolean deleteUnusedUpdates() {
	return delete_unused_updates;
    }

    public boolean implicitExecution() {
	return implicit_execution;
    }

    // setter

    public void setUseSimultaneousUpdates(boolean b) {
	use_sim_updates = b;
	fireSettingsChanged();
    }

    public void setDeleteSimpleUpdates(boolean b) {
	delete_simple_updates = b;
	fireSettingsChanged();
    }

    public void setDeleteUnusedUpdates(boolean b) {
	delete_unused_updates = b;
	fireSettingsChanged();
    }

    public void setImplicitExecution(boolean b) {
	implicit_execution = b;
	fireSettingsChanged();
    }

    // 
    public SimultaneousUpdateSimplifier getSimplifier() {
	return sus;
    }

    /** gets a Properties object and has to perform the necessary
     * steps in order to change this object in a way that it
     * represents the stored settings
     */
    public void readSettings(Properties props) {
	String simultaneous_updates_allowed_string = props.getProperty(USE_SIM_UPDATES_KEY);
	String delete_simple_string = props.getProperty(DELETE_SIMPLE_KEY);	
	String delete_unused_string = props.getProperty(DELETE_UNUSED_KEY);	
	String implicit_execution_string = props.getProperty(IMPLICIT_MODE_KEY);	
	
	setUseSimultaneousUpdates(Boolean.valueOf(simultaneous_updates_allowed_string)
			      .booleanValue());
	setDeleteSimpleUpdates(Boolean.valueOf(delete_simple_string)
			       .booleanValue());
	setDeleteUnusedUpdates(Boolean.valueOf(delete_unused_string)
			       .booleanValue());
	
	setImplicitExecution(Boolean.valueOf(implicit_execution_string)
			     .booleanValue());
    }


    /** implements the method required by the Settings interface. The
     * settings are written to the given Properties object. Only entries of the form 
     * <key> = <value> (,<value>)* are allowed.
     * @param props the Properties object where to write the settings as (key, value) pair
     */
    public void writeSettings(Properties props) {
	props.setProperty(USE_SIM_UPDATES_KEY, "" + use_sim_updates);
	props.setProperty(DELETE_SIMPLE_KEY,   "" + delete_simple_updates);	
	props.setProperty(DELETE_UNUSED_KEY,   "" + delete_unused_updates);	       
	props.setProperty(IMPLICIT_MODE_KEY,  "" + implicit_execution);	
    }

    /** sends the message that the state of this setting has been
     * changed to its registered listeners (not thread-safe)
     */
    protected void fireSettingsChanged() {
	updateSimplifier();

	Iterator it = listenerList.iterator();
	while (it.hasNext()) {
	    ((SettingsListener)it.next()).settingsChanged(new GUIEvent(this));
	}
    }

    /** adds a listener to the settings object 
     * @param l the listener
     */
    public void addSettingsListener(SettingsListener l) {
	listenerList.add(l);
    }

}
