// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.gui;

import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JLabel;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import java.awt.event.*;
import java.util.*;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.NonTerminalProgramElement;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.Node;

/** This class creates a menu with Taclets as entries. The invoker has
 * to be of type SequentView because of the method call selectedTaclet
 * that hands over the selected Taclet. The class is used to get all
 * Taclet that are applicable at a selected position in a sequent.
 */ 
class TacletMenu extends JMenu {

    private PosInSequent pos;
    private SequentView sequentView;
    private KeYMediator mediator;

    private TacletComparator comp = new TacletComparator();
    
    
    /** creates empty menu 
     */    
    TacletMenu() {
    }


    /** creates a new menu that displays all applicable rules at the given
     * position  
     * @param sequentView the SequentView that is the parent of this menu
     * @param findList ListOfTaclet with all applicable FindTaclets
     * @param rewriteList ListOfTaclet with all applicable RewriteTaclets
     * @param noFindList ListOfTaclet with all applicable noFindTaclets
     * @param builtInList ListOfBuiltInRule with all applicable BuiltInRules
     * @param pos the PosInSequent
     */ 
    TacletMenu(SequentView sequentView,
	       ListOfTaclet findList, ListOfTaclet rewriteList, 
	       ListOfTaclet noFindList, ListOfBuiltInRule builtInList, 
	       PosInSequent pos) { 
	super();
	this.sequentView = sequentView;
	this.mediator = sequentView.mediator();
 	this.pos = pos;	
	// delete RewriteTaclet from findList because they will be in
	// the rewrite list and concatenate both lists
	createTacletMenu(removeRewrites(findList).prepend(rewriteList),
			 noFindList, builtInList, new MenuControl());  
    }

    /** removes RewriteTaclet from list
     * @param list the ListOfTaclet from where the RewriteTaclet are
     * removed 
     * @return list without RewriteTaclets
     */
    private ListOfTaclet removeRewrites(ListOfTaclet list) {  
	ListOfTaclet result = SLListOfTaclet.EMPTY_LIST;
	IteratorOfTaclet it = list.iterator();

	while(it.hasNext()) {
	    Taclet taclet=it.next();
	    result = (taclet instanceof RewriteTaclet ? result : 
		      result.prepend(taclet));	     
	}
	return result;
    }


    /** creates the menu by adding all submenus and items */
    private void createTacletMenu(ListOfTaclet find,
				  ListOfTaclet noFind,
				  ListOfBuiltInRule builtInList,
				  MenuControl control) {	 
	addActionListener(control);
	boolean rulesAvailable=(addSection("Find", sort(find), control));
	if (pos != null && pos.isSequent()) {
	    rulesAvailable=addSection("NoFind", noFind, control)
		| rulesAvailable;
	}
	if (!rulesAvailable) {
	    createSection("No rules applicable.");
	}

	createBuiltInRuleMenu(builtInList, control);

	addClipboardItem(control);

	if (pos != null) {
	    PosInOccurrence occ = pos.getPosInOccurrence();	    
	    if (occ != null && occ.posInTerm() != null) {
		Term t = occ.subTerm ();
		createAbbrevSection(t, control);
	    }
	}

    }

    private void createBuiltInRuleMenu(ListOfBuiltInRule builtInList,
				       MenuControl            control) {

	if (builtInList != SLListOfBuiltInRule.EMPTY_LIST) {
	    addSeparator();	
	    IteratorOfBuiltInRule it = builtInList.iterator();
	    while (it.hasNext()) {
		addBuiltInRuleItem(it.next(), control);
	    }
	}
    }
				      
    /** 
     * adds an item for built in rules (e.g. Run Simplify or
     * Update Simplifier) 
     */
    private void addBuiltInRuleItem(BuiltInRule builtInRule,
				MenuControl control) {
        JMenuItem item = new BuiltInRuleMenuItem(builtInRule);
        item.addActionListener(control);
	add(item);
    }

    
    private ListOfTaclet sort(ListOfTaclet finds) {
	ListOfTaclet result = SLListOfTaclet.EMPTY_LIST;
	List list = new ArrayList(finds.size());

	IteratorOfTaclet it = finds.iterator();
	while (it.hasNext()) {
	    list.add(it.next());
	}

	Collections.sort(list, comp);

	Iterator it2 = list.iterator();
	while (it2.hasNext()) {
	    result = result.prepend((Taclet)it2.next());
	}

	
	return result;
    }

    
    private void createAbbrevSection(Term t, MenuControl control){
	AbbrevMap scm = mediator.getNotationInfo().getAbbrevMap();
	JMenuItem sc = null;
	if(scm.containsTerm(t)){
	    sc = new JMenuItem("Change abbreviation");
	    sc.addActionListener(control);
	    add(sc);
	    if(scm.isEnabled(t)){
		sc = new JMenuItem("Disable abbreviation");
	    }else{
		sc = new JMenuItem("Enable abbreviation");
	    }
	}else{
	    sc = new JMenuItem("Create abbreviation");
	}
	sc.addActionListener(control);
	add(sc);
    }



    /** creates a non selectable label with the specified name and adds it to
     * the component followed by the entries of this section 
     * @param title a String the title of the section
     * @param taclet ListOfTaclet that contains the Taclets belonging to this section
     * @return true if section has been added (empty sections are not added)
     */ 
    private boolean addSection(String title, ListOfTaclet taclet, 
			       MenuControl control) {
	if (taclet.size() > 0) {
	    //uncomment if you want submenus with subtitels
	    //	    insert(createSubMenu(taclet, title, control), 1);
	    //	    createSection(title);
	    add(createMenuItems(taclet, control));
	    return true;
	}
	return false;
    }

    /** inserts separator followed from the section's title 
     * @param title a String that contains the title of the section
     */
    private void createSection(String title) {
	//addSeparator();
	add(new JLabel(title));
    }
    
    
    private void addClipboardItem(MenuControl control) {
	addSeparator();
        JMenuItem item = new JMenuItem("to clipboard");
        item.addActionListener(control);
	add(item);
    }
    

    private void addRunSimplifyItem(MenuControl control) {
	addSeparator();
        JMenuItem item = new JMenuItem("Run \"Simplify\"");
        item.addActionListener(control);
	add(item);
    }


    /** adds array of TacletMenuItem to itself*/
    private void add(TacletMenuItem[] items) {
	for (int i = 0; i < items.length; i++) {
	    add(items[i]);
	}
    }

    /** creates new TacletMenuItems for each taclet in the list and set
     * the given MenuControl as their ActionListener
     * @param taclet ListOfTaclet with the Taclets the items represent
     * @param control the ActionListener
     * @return the new MenuItems
     */
    private TacletMenuItem[] createMenuItems(ListOfTaclet taclet, 
					     MenuControl  control) {
	TacletMenuItem[] items = new TacletMenuItem[taclet.size()];
	IteratorOfTaclet it = taclet.iterator();

	for (int i = 0; it.hasNext(); i++) {
	    items[i] = new TacletMenuItem(it.next());
	    items[i].addActionListener(control);
	}
	return items;
    }
    

    /** creates new submenu with the given title whose entries are
     * the taclets in the list and set 
     * the given MenuControl as its ActionListener
     * @param taclet a ListOfTaclet with the Taclets the items of the submenu
     * @param title a String that is the submenu's title
     * @param control the ActionListener
     * @return the new submenu
     */
    private JMenu createSubMenu(ListOfTaclet taclet, String name, 
				MenuControl  control) {	
	JMenu menu=new JMenu(name);
	menu.addActionListener(control);
	JMenuItem[] items=createMenuItems(taclet, control);

	for(int i = 0; i < items.length; i++) {
	    menu.add(items[i]);	    
	}

	return menu;
    }

    
    /** makes submenus invisible */
    void invisible() {
	for (int i = 0; i < getMenuComponentCount(); i++) {
	    if (getMenuComponent(i) instanceof JMenu) 
		((JMenu)getMenuComponent(i)).getPopupMenu().setVisible(false);
	}
    }

    /** ActionListener */
    class MenuControl implements ActionListener{

	private boolean validabbreviation(String s){
	    if(s==null || s.length()==0) return false;
	    for(int i=0; i<s.length(); i++){
		if(!((s.charAt(i)<='9' && s.charAt(i)>='0')||
		     (s.charAt(i)<='z' && s.charAt(i)>='a')||
		     (s.charAt(i)<='Z' && s.charAt(i)>='A')||
		     s.charAt(i)=='_')) return false;
	    }
	    return true;
	}

	public void actionPerformed(ActionEvent e) {
	    if (e.getSource() instanceof TacletMenuItem) {
		((SequentView)(getPopupMenu().getInvoker()))
		    .selectedTaclet(((TacletMenuItem) e.getSource()).connectedTo(), 
				  pos);
	    } else if (e.getSource() instanceof BuiltInRuleMenuItem) {
		mediator.selectedBuiltInRule
		    (((BuiltInRuleMenuItem) e.getSource()).connectedTo(), 
		     pos);
	    } else {
		if (((JMenuItem)e.getSource()).getText()
			   .startsWith("to clipboard")){
                    Main.copyHighlightToClipboard(sequentView);
		} else if(((JMenuItem)e.getSource()).getText().
				 startsWith("Disable abbreviation")){
			PosInOccurrence occ = pos.getPosInOccurrence();	    
			if (occ != null && occ.posInTerm() != null) {
				mediator.getNotationInfo().getAbbrevMap().setEnabled(occ.subTerm(),false);
				sequentView.printSequent();
			}
		}else if(((JMenuItem)e.getSource()).getText().
				 startsWith("Enable abbreviation")){
			PosInOccurrence occ = pos.getPosInOccurrence();	    
			if (occ != null && occ.posInTerm() != null) {
				mediator.getNotationInfo().
				    getAbbrevMap().setEnabled(occ.subTerm(),true);
				sequentView.printSequent();
			}
		}else if(((JMenuItem)e.getSource()).getText().
				 startsWith("Create abbreviation")){
			PosInOccurrence occ = pos.getPosInOccurrence();	 			
			if (occ != null && occ.posInTerm() != null) {
				String abbreviation = (String)JOptionPane.showInputDialog
					(new JFrame(),
					 "Enter abbreviation for term: \n"+occ.subTerm().toString(), 
					 "New Abbreviation",
					 JOptionPane.QUESTION_MESSAGE,
					 null,
					 null,
					 "");
				    
				try{
				    if(abbreviation!=null){
					if(!validabbreviation(abbreviation)){
					    JOptionPane.showMessageDialog(new JFrame(),
									  "Only letters, numbers and '_' are allowed for Abbreviations", 
									  "Sorry",
									  JOptionPane.INFORMATION_MESSAGE);
					}else{
					    mediator.getNotationInfo().
						getAbbrevMap().put(occ.subTerm(),abbreviation,true);
					    sequentView.printSequent();
					}
				    }
				}catch(AbbrevException sce){
				    JOptionPane.showMessageDialog(new JFrame(), sce.getMessage(), "Sorry", 
							      JOptionPane.INFORMATION_MESSAGE);					    
				}
			}
		
		}else if(((JMenuItem)e.getSource()).getText().
			 startsWith("Change abbreviation")){
			PosInOccurrence occ = pos.getPosInOccurrence();	 			
			if (occ != null && occ.posInTerm() != null) {
			    String abbreviation = (String)JOptionPane.showInputDialog
				(new JFrame(),
				 "Enter abbreviation for term: \n"+occ.subTerm().toString(),
				 "Change Abbreviation",
				 JOptionPane.QUESTION_MESSAGE,
				 null,
				 null,
				 mediator.getNotationInfo().
					    getAbbrevMap().getAbbrev(occ.subTerm()).substring(1));
				try{
				    if(abbreviation!=null){
					if(!validabbreviation(abbreviation)){
					    JOptionPane.showMessageDialog(new JFrame(),
									  "Only letters, numbers and '_' are allowed for Abbreviations", 
									  "Sorry",
									  JOptionPane.INFORMATION_MESSAGE);
					}else{					    
					    mediator.getNotationInfo().
						getAbbrevMap().changeAbbrev(occ.subTerm(),abbreviation);
					    sequentView.printSequent();
					}
				    }
				}catch(AbbrevException sce){
				    JOptionPane.showMessageDialog(new JFrame(), sce.getMessage(), "Sorry", 
							      JOptionPane.INFORMATION_MESSAGE);
				}
			}
		}
	    }    
	}
    }



    /** this class extends JMenuItem. The objective is to store
     * the Taclet of each item in the item for easier access to the Taclet
     * if the item has been selected 
     */
    static class TacletMenuItem extends JMenuItem {
	
	private Taclet connectedTo;
	
	/** creates TacletMenuItem attached to a Taclet 
	 * @param connectedTo the Taclet that is reprsented by the item 
	 */
	public TacletMenuItem(Taclet connectedTo) {
	    super(connectedTo.displayName());
	    this.connectedTo = connectedTo;	    	    
	    char[] taclet_char = connectedTo.toString().toCharArray();
	    StringBuffer taclet_sb = new StringBuffer();
	    boolean whitespace = false;
	    for (int i = 0; i < taclet_char.length; i++) {
		if (taclet_char[i] == '\n' || taclet_char[i] == ' ') {
		    whitespace = true;
		} else {
		    if (whitespace) {
			taclet_sb.append(" ");
		    }
		    whitespace = false;
		    taclet_sb.append(taclet_char[i]);
		}
	    }		
	    setToolTipText(taclet_sb.toString());
	} 

	/** gets the Taclet that is attached to this item
	 * @return the attached Taclet
	 */
	public Taclet connectedTo() {
	    return connectedTo;
	}

    }

    /** 
     * equal to TacletMenuItem but for BuiltInRules
     */
    static class BuiltInRuleMenuItem extends JMenuItem {
	
	private BuiltInRule connectedTo;
	
	public BuiltInRuleMenuItem(BuiltInRule connectedTo) {
	    super(connectedTo.name().toString());
	    this.connectedTo = connectedTo;	    	    
	} 

	public BuiltInRule connectedTo() {
	    return connectedTo;
	}

    }
    
    
    static class TacletComparator implements Comparator {

	private int countFormulaSV(TacletVariableCollector c) {
	    int formulaSV = 0;
	    IteratorOfSchemaVariable it = c.varIterator();
	    while (it.hasNext()) {
		if (it.next().sort() == Sort.FORMULA) {
		    formulaSV++;
		}
	    }

	    return formulaSV;
	}

	/** this is a rough estimation about the goal complexity. The
	 * complexity depends on the depth of the term to be replaced.
	 * If no such term exists we add a constant (may be refined in
	 * future)
	 */
	private int measureGoalComplexity(ListOfTacletGoalTemplate l) {
	    int result = 0;
	    IteratorOfTacletGoalTemplate it = l.iterator();
	    while (it.hasNext()) {
		TacletGoalTemplate gt = it.next();
		if (gt instanceof RewriteTacletGoalTemplate) {
		    if (((RewriteTacletGoalTemplate)gt).replaceWith() != null) {
			result += ((RewriteTacletGoalTemplate)gt).replaceWith().depth();
		    }
		} 
		if (gt.sequent() != Sequent.EMPTY_SEQUENT) {
		    result += 10;
		}
	    }
	    return result;
	}


	/**
	 * rough approximation of the program complexity
	 */
	public int programComplexity(JavaBlock b) {
	    if (b == JavaBlock.EMPTY_JAVABLOCK) {
		return 0;
	    }
	    return new de.uka.ilkd.key.java.visitor.JavaASTWalker(b.program()) { 
		    private int counter = 0;
		    
		    protected void doAction(ProgramElement pe) {
			counter++;
		    }
		    
		    public int getCounter() {
			counter = 0;
			start();
			return counter;
		    }
		}.getCounter();
	}

	public int compare(Object o1, Object o2) {
	    FindTaclet taclet1 = (FindTaclet) o1;
	    FindTaclet taclet2 = (FindTaclet) o2;
	    
	    int findComplexity1 = taclet1.find().depth();
	    int findComplexity2 = taclet2.find().depth();
	    findComplexity1 += programComplexity(taclet1.find().javaBlock());
	    findComplexity2 += programComplexity(taclet2.find().javaBlock());

	    int result;
	    if ( findComplexity1 < findComplexity2 ) {
		return -1;
	    } else if (findComplexity1 > findComplexity2) {
		return 1;
	    }
	    
	    
	    

	    // depth are equal. Number of schemavariables decides
	    TacletVariableCollector coll1 = new TacletVariableCollector();
	    taclet1.find().execPostOrder(coll1);
	    int formulaSV1 = countFormulaSV(coll1);
	    
	    TacletVariableCollector coll2  = new TacletVariableCollector();
	    taclet2.find().execPostOrder(coll2);
	    int formulaSV2 = countFormulaSV(coll2);

	    int cmpVar1 = -coll1.size();
	    int cmpVar2 = -coll2.size();
	    
	    if (cmpVar1 == cmpVar2) {
		cmpVar1 = -formulaSV1;
		cmpVar2 = -formulaSV2;
	    }
	    
	    if ( cmpVar1 < cmpVar2 ) {
		return -1;
	    } else if (cmpVar1 > cmpVar2) {
		return 1;
	    }

	    if (taclet1.ifSequent() == Sequent.EMPTY_SEQUENT && 
		taclet2.ifSequent() != Sequent.EMPTY_SEQUENT) {
		return 1;
	    } else if (taclet1.ifSequent() == Sequent.EMPTY_SEQUENT && 
		       taclet2.ifSequent() != Sequent.EMPTY_SEQUENT) {
		return -1;
	    }
	    
	    int goals1 = -taclet1.goalTemplates().size();
	    int goals2 = -taclet2.goalTemplates().size();

	    if (goals1 == goals2) {
		goals1 = -measureGoalComplexity(taclet1.goalTemplates());
		goals2 = -measureGoalComplexity(taclet2.goalTemplates());		
	    } 

	    if (goals1 < goals2) {
		return -1;
	    } else if (goals1 > goals2) {
		return 1;
	    }

	    return 0;
	}
	
    }
    
}
