// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.java;

import de.uka.ilkd.key.rule.metaconstruct.MethodCall;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.expression.literal.NullLiteral;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.ldt.*;
import java.util.Iterator;
import de.uka.ilkd.key.util.Debug;
import java.util.Vector;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;

/**
 * an instance serves as representation of a Java model underlying a DL
 * formula. This class provides calls to access the elements of the Java
 * model using the KeY datastructures only. Implementation specific
 * details like the use of Recoder is hidden in the field of type 
 * {@link KeYProgModelInfo}. This class can be extended to provide further 
 * services.
 */
public class JavaInfo {


    private Services services;
    private KeYProgModelInfo  kpmi;
    private Namespace programVariables = new Namespace();

    /**
     * the type of null
     */
    private KeYJavaType nullType=null;

    /**
     * a term with the constant 'null'
     */
    private Term nullConst=null;

	
     /**
      * creates a new JavaInfo object by giving a KeYProgModelInfo to access
      * the Recoder SourceInfo and using the given {@link Services} object.
      */
     JavaInfo(KeYProgModelInfo kpmi, Services s) {
         this.kpmi 	= kpmi;
         services	= s;	  	
     }
	
    /**
     * this constructor is only for tests
     */
    public JavaInfo() {
        services	= new Services(this);
	kpmi=new KeYProgModelInfo(new KeYRecoderMapping(),
				  services.getTypeConverter());
    }
	
    /**
     * returns the underlying KeYProgModelInfo providing access to the
     * Recoder structures.
     */
    public KeYProgModelInfo getKeYProgModelInfo(){
        return kpmi;
    }

    void setKeYProgModelInfo(KeYProgModelInfo kpmi){
        this.kpmi = kpmi;
    }

    /**
     * convenience method that returns the Recoder-to-KeY mapping underlying
     * the KeYProgModelInfo of this JavaInfo
     */
    public KeYRecoderMapping rec2key() {
	return getKeYProgModelInfo().rec2key();
    }

    private JavaInfo(JavaInfo proto, Services s) {
	this.services = s;
	nullType=proto.getNullType();
	nullConst=proto.getNullConst();
	setKeYProgModelInfo(proto.getKeYProgModelInfo().copy());
	programVariables=proto.programVariables().copy();
    }

    /**
     * copies this JavaInfo and uses the given Services object as the
     * Services object of the copied JavaInfo
     * @param serv
     * @return
     */
    public JavaInfo copy(Services serv) {
	return new JavaInfo(this, serv);
    }

    /**
     * returns a (deep) copy of this JavaInfo by copying the services of the
     * JavaInfo and returning the JavaInfo of the copied Services
     */
    public JavaInfo copy() {
	Services s = getServices().copy();
	return s.getJavaInfo();
    }

    /** returns the TypeConverter to translate program parts to their
     * logic equivalent
     */
    public TypeConverter getTypeConverter() {
        return services.getTypeConverter();
    }

    /**
     * returns the services associated with this JavaInfo
     */
    public Services getServices(){
        return services;
    }

    /**
     * only to be used by Services
     * @param s the services object to be set
     */
    void setServices(Services s) {
	services=s;
    }

    //------------------- common services ----------------------


    /**
     * returns the full name of a given {@link
     * de.uka.ilkd.key.java.abstraction.KeYJavaType}. 
     * @param t the KeYJavaType including the package prefix
     * @return the full name
     */
    public String getFullName(KeYJavaType t) {
        return kpmi.getFullName(t);
    }

    /**
     * creates a new TypeReference for the given KeYJavaType
     */
    public TypeReference createTypeReference(KeYJavaType kjt) {
        return new TypeRef(kjt);
    }

    /**
     * looks up the fully qualifying name given by a String 
     * in the list of all available
     * KeYJavaTypes in the Java model
     * @param fullName the String
     * @return the KeYJavaType with the name of the String
     */
    public KeYJavaType getTypeByName(String fullName) {
        Iterator it=(kpmi.allElements()).iterator();
        while (it.hasNext()) {
            Object o=it.next();
            if (o instanceof KeYJavaType) {
                if (getFullName((KeYJavaType)o).equals(fullName)) {
                    return (KeYJavaType)o;
                }
            }
        }
        return null;
    }


    /**
     * looks up a KeYJavaType with given name. If the name is a fully
     * qualifying name with package prefix an element with this full name is
     * taken. If the name does not contain a full package prefix some
     * KeYJavaType with this short name is taken.
     * @param className the name to look for (either full or short)
     * @return a class matching the name
     */
    public KeYJavaType getTypeByClassName(String className) {
	String shortName;
	int dotpos = className.lastIndexOf(".");
	shortName=(dotpos>=0) ? className.substring(dotpos+1) : className;	
        Iterator it=(kpmi.allElements()).iterator();;
        while (it.hasNext()) {
            Object o=it.next();     
            if (o!=null) {
                if (o instanceof KeYJavaType) {                     
                    KeYJavaType t = (KeYJavaType)o;
		    String name=getFullName(t);
		    String shortn = name;
		    int pos=name.lastIndexOf(".");
		    if (pos!=-1)
			shortn=name.substring(pos+1);
                    if (name.equals(className) || shortn.equals(shortName)) {           
                        Debug.out("javaInfo: type found (className, type):",
                                  className, o);
                        return (KeYJavaType)o;
                    }
                }           
            }
        }       
        Debug.out("javaInfo: type not found. Looked for:",className);
	throw new RuntimeException("TYPE NOT FOUND: "+className);
    }


    /**
     * returns a type declaration with the full name of the given String fullName
     */
    public TypeDeclaration getTypeDeclaration(String fullName) {
        return (TypeDeclaration)getTypeByName(fullName).getJavaType();
    }

    /**
     * returns a primitive KeYJavaType matching the given typename making use
     * of the LDTs of the current type converter in the services.
     */
    public KeYJavaType getPrimitiveKeYJavaType(String typename) {
        LDT[] models = services.getTypeConverter().getModels();
        for (int i = 0; i<models.length ; i++) {
            IteratorOfType it = models[i].javaType().iterator();
            while (it.hasNext()) {
                Type type = (Type) it.next();
                if (type.getFullName().equals(typename)) {
                    return models[i].getKeYJavaType(type);
                }
            }
        }
        return null;
    }

    /**
     * returns a KeYJavaType (either primitive of object type) having the
     * full name of the given String fullName 
     */
    public KeYJavaType getKeYJavaType(String fullName) {
        KeYJavaType result = getPrimitiveKeYJavaType(fullName);
        return (result == null ?
            (KeYJavaType)getTypeByName(fullName) :
            result);
    }


    /**
     * this is an alias for getTypeByClassName
     */
    public KeYJavaType getKeYJavaTypeByClassName(String className) {
        return (KeYJavaType)getTypeByClassName(className);
    }


    /**
     * returns true iff the given subType KeYJavaType is a sub type of the
     * given KeYJavaType superType.
     */
    public boolean isSubtype(KeYJavaType subType, KeYJavaType superType) {
        return kpmi.isSubtype(subType, superType);
    }

    /** 
     * returns a KeYJavaType having the given sort
     */
     public KeYJavaType getKeYJavaType(Sort sort) {
         Iterator it = (kpmi.allElements()).iterator();
         while (it.hasNext()) {
             Object o = it.next();
             if (o != null && o instanceof KeYJavaType &&
             ((KeYJavaType)o).getSort() == sort) {
		 return (KeYJavaType)o;
             }
         }
         return null;
     }

    /** returns the KeYJavaType of the expression if it can be
     * determined else null is returned
     */
    public KeYJavaType getPrimitiveKeYJavaType(Expression e) {
        return services.getTypeConverter().getKeYJavaType(e);
    }

    /**
     * returns all ObjectSorts of the underlying Java model
     * @return a namespace containing the object sorts
     */
    public Namespace getObjectSorts() {
        Iterator it = kpmi.allObjectSorts().iterator();
        Namespace ns = new Namespace();
        while (it.hasNext()) {
            Named n=(Named)it.next();
            if (n!=services.getTypeConverter().getObjectSort()) {
                ns.add(n);
            }
        }
        return ns;
    }


    /**
     * returns the KeYJavaType belonging to the given Type t
     */
    public KeYJavaType getKeYJavaType(Type t) {
        if (t instanceof PrimitiveType) {
	    return services.getTypeConverter().getKeYJavaType(t);
        } else {
            Iterator it = (kpmi.allElements()).iterator();
            while (it.hasNext()) {
                Object o = it.next();
                if (o instanceof KeYJavaType) {
                    if (((KeYJavaType)o).getJavaType()==t) {
                        return (KeYJavaType)o;
                    }
                }
            }
        }
        return null;
    }

    /**
     * returns all methods from the given Type Declaration
     */
    public ListOfMethod getAllMethods(TypeDeclaration td) {
        return kpmi.getAllMethods(td);
    }

    /**
     * returns all methods from the given Type Declaration as ProgramMethods
     */
    public ListOfProgramMethod getAllProgramMethods(TypeDeclaration td) {
        return kpmi.getAllProgramMethods(td);
    }

    /**
     * returns the program methods defined in the given KeYJavaType with name
     * m and the list of types as signature of the method 
     */
    public ProgramMethod getProgramMethod(KeYJavaType ct, String m,
					  ListOfType signature) {
        return kpmi.getProgramMethod(ct, m, signature);
    }


    /**
     * retrieves the direct extended superclass for the given class
     * @param type the KeYJavaType of the type whose superclass
     * has to be determined 
     * @return KeYJavaType of the extended supertype
     */
    public KeYJavaType getSuperclass(KeYJavaType type) {
	KeYJavaType result = null;
	ClassType javaType = (ClassType) type.getJavaType();

	if (javaType.isInterface()) {
	    return null;
	}

	ListOfKeYJavaType localSupertypes = javaType.getSupertypes();
	IteratorOfKeYJavaType it = localSupertypes.iterator();
	while (result == null && it.hasNext()) {
	    KeYJavaType keYType = it.next();
	    if (!((ClassType)keYType.getJavaType()).isInterface()) {
		result = keYType;
	    }
	}

	if (result == null) {
	    result = getKeYJavaType
		(ClassDeclaration.objectClassDecl);
	} 

        return result;	
    }
    
    public Namespace programVariables() {
	return programVariables;
    }


    /**
     * returns the program method defined in the KeYJavaType of the program
     * variable clv, with the name m, and the KeYJavaTypes of the given array
     * of program variables as signatures.
     * @param clv a program variable defining the KeYJavaType to look at
     * @param m the name of the method
     * @param args an array of program variables defining the KeYJavaTypes of
     * the parameters
     * @return a matching program method
     */
    public ProgramMethod getProgramMethod
	(ProgramVariable clv, String m, ProgramVariable[] args){
        ListOfType types = SLListOfType.EMPTY_LIST;
        for (int i=args.length-1; i>=0; i--) {
            types=types.prepend(args[i].getKeYJavaType());
        }
        return getProgramMethod(clv.getKeYJavaType(), m, types);
    }

    /** gets an array of expression and returns a list of types */
    private ListOfKeYJavaType getKeYJavaTypes(ArrayOfExpression args) {
	ListOfKeYJavaType result = SLListOfKeYJavaType.EMPTY_LIST; 
	if (args != null) {
	    for (int i = args.size()-1; i >= 0 ; i--) {
		Expression argument = args.getExpression(i);
		result = result.prepend
		    (getTypeConverter().getKeYJavaType(argument));
	    }
	}
	return result;
    }

    /**
     * retrieves the signature of the given method reference
     * @param method the MethodReference of which the signature has to be determined 
     * @return the signature of the given method reference
     */
    public ListOfKeYJavaType getSignature(MethodReference ref) {
	return getKeYJavaTypes(ref.getArguments());
    }

    /**
     * retrieves the signature of the given constructor reference
     * @param constructor the ConstructorReference of which the signature has to
     * be determined 
     * @return the signature of the given constructor reference
     */
    public ListOfKeYJavaType getSignature(ConstructorReference ref) {
	return getKeYJavaTypes(ref.getArguments());
    }

    /**
     * retrieves the signature according to the given expressions
     * @param arguments ArrayOfExpression of which we try to construct a
     * signature 
     * @return the signature 
     */
    public ListOfKeYJavaType createSignature(ArrayOfExpression arguments) {
	return getKeYJavaTypes(arguments);
// 	return arguments == null ? SLListOfKeYJavaType.EMPTY_LIST : 
// 	    
    }

    /**
     * retrieves all attributes declared in the given class 
     * (including the implicit ones)
     * @param classDecl the ClassDeclaration where to look for the attributes
     * @return all attributes declared in the given class 
     * (including the implicit ones)
     */
    public ListOfField getAllFields(ClassDeclaration classDecl) {
	ListOfField fields = SLListOfField.EMPTY_LIST;
	ArrayOfMemberDeclaration members = classDecl.getMembers();
	for (int i = members.size()-1; i>=0; i--) {
	    MemberDeclaration member = members.getMemberDeclaration(i);
	    if (member instanceof FieldDeclaration) {
		ArrayOfFieldSpecification specs = 
		    ((FieldDeclaration)member).getFieldSpecifications();
		for (int j = specs.size()-1; j>=0; j--) {
		    fields = fields.prepend(specs.getFieldSpecification(j));
		}
	    }
	}
	return fields;
    }


    /**
     * retrieves all implicit attributes of the given class 
     * @param classDecl the ClassDeclaration where to look for the implicit
     * attributes 
     * @return all implicit attributes of the given class
     */
    public ListOfField getImplicitFields(ClassDeclaration classDecl) {
	ListOfField fields = SLListOfField.EMPTY_LIST;
	ArrayOfMemberDeclaration members = classDecl.getMembers();
	for (int i = members.size()-1; i>=0; i--) {
	    if (members.getMemberDeclaration(i) instanceof FieldDeclaration) {
		ArrayOfFieldSpecification specs = 
		    ((FieldDeclaration)members.getMemberDeclaration(i)).
		    getFieldSpecifications();
		for (int j = specs.size()-1; j>=0; j--) {
		    if (specs.getFieldSpecification(j) instanceof ImplicitFieldSpecification) {
			fields = fields.prepend(specs.getFieldSpecification(j));
		    }
		}
	    }
	}
	return fields;
    }
    

    //----------------- parsing services --------------------------

    /**
     * reads a Java block given as a string java as it was in the given
     * TypeDeclaration asIn.
     */
     public JavaBlock readJavaBlock(String java, TypeDeclaration asIn) {
        ClassDeclaration cd=null;
        if (asIn instanceof ClassDeclaration) {
            cd = (ClassDeclaration)asIn;
        } else {
            Debug.out("Reading Java Block from an InterfaceDeclaration:"
                      +" Not yet implemented.");
        }
        return kpmi.readBlock(java,cd);
    }

    /**
     * reads a Java block given as a String
     */
    public JavaBlock readJavaBlock(String java) {
        return kpmi.readJavaBlock(java);
    }

    /**
     * reads a Java statement not necessarily a block
     */
    public ProgramElement readJava(String java) {
        return ((StatementBlock)readJavaBlock("{"+java+"}")
		.program()).getChildAt(0);
    }

    /**
     * returns a term representing 'null' in logic
     */
    public Term getNullTerm() {
        return services.getTypeConverter().convertToLogicElement(NullLiteral.NULL);
    }


    /**
     * retrieves a field with the given name out of the list
     * @param name a String with the name of the field to be looked for
     * @param fields the ListOfField where we have to look for the field
     * @return the program variable of the given name or null if not
     * found
     */
    private final ProgramVariable find(String name, ListOfField fields) {
	IteratorOfField it = fields.iterator();
	while (it.hasNext()) {
	    Field field = it.next();
	    if (name.equals(field.getName())) {
		return (ProgramVariable)
		    ((FieldSpecification)field).getProgramVariable();
	    }
	}
	return null;
    }

    /**
     * extracts all fields out of fielddeclaration
     * @param field the FieldDeclaration of which the field
     * specifications have to be extracted
     * @return a ListOfField the includes all field specifications found
     * int the field declaration of the given list
     */
    private final ListOfField filterField(FieldDeclaration field) {
	ListOfField result = SLListOfField.EMPTY_LIST;
	ArrayOfFieldSpecification spec = field.getFieldSpecifications();
	for (int i = spec.size()-1; i>=0; i--) {
	    result = result.prepend(spec.getFieldSpecification(i));
	} 
	return result;
    }

    /**
     * extracts all field specifications out of the given
     * list. Therefore it descends into field declarations.
     * @param list the ArrayOfMemberDeclaration with the members of a
     * type declaration
     * @return a ListOfField the includes all field specifications found
     * int the field declaration of the given list
     */
    private ListOfField filterField(ArrayOfMemberDeclaration list) {
	ListOfField result = SLListOfField.EMPTY_LIST;
	for (int i = list.size()-1; i >= 0; i--) {
	    MemberDeclaration pe = list.getMemberDeclaration(i);
	    if (pe instanceof FieldDeclaration) {
		result = result.append
		    (filterField((FieldDeclaration)pe));
	    }
	}
	return result;
    }

    /**
     * returns an attribute program variable with the given name defined in
     * the given KeYJavaType 
     */
    public ProgramVariable getAttribute(String name,
					KeYJavaType referenceType) {
	if (referenceType.getJavaType() instanceof ArrayDeclaration) {
	    return find(name, filterField
			(((ArrayDeclaration)referenceType.
			  getJavaType()).getMembers()));
	} else {
	    ListOfField list = kpmi.getAllFields(referenceType);
	    IteratorOfField it = list.iterator();
	    while(it.hasNext()) {
		Field f = it.next();
		if(f.getName().equals(name)){
		    return (ProgramVariable)((VariableSpecification)f).
                    getProgramVariable();
		}
	    }
	}
	return null;
    }

    /**
     * returns an attribute declared in a KeYJavaType represented by the
     * given ObjectSort and with the given name
     */
    public ProgramVariable getAttribute(String name, ObjectSort s) {
        return getAttribute(name, (KeYJavaType)getKeYJavaType(s));
    }

    /**
     * returns the KeYJavaType for class java.lang.Object
     */
    public KeYJavaType getObjectKeYJavaType() {
	return getTypeConverter().getObjectKeYJavaType();
    }

    /**
     * returns the KeYJavaType  representing the type of 'null' 
     */
    public KeYJavaType getNullType() {
	if (nullType==null) {
	    nullType=new KeYJavaType(NullType.JAVA_NULL);
	    nullType.setSort(Sort.NULL);
	}
	return nullType;
    }

    /**
     * returns a term representing the null constant in logic
     */
    public Term getNullConst() {
	if (nullConst==null) {
	    nullConst=TermFactory.DEFAULT.createFunctionTerm(Op.NULL);
	}
	return nullConst;
    }


// ********** Methodcall Service ************************************



   /**
    * Constructs the if-cascade for branching over possible implementations
    * for a given method. This is part of the run-time method binding 
    * procedure.
    * @param mc MethodCall object for performing construction callbacks
    * @param ct Declared type of the method invocation target
    * @param name method name
    * @param signature method signature for lookups
    * @author VK
    */

   public void findImplementations(MethodCall mc, 
                                   KeYJavaType ct, String name, 
				   ListOfKeYJavaType signature) {
       kpmi.findImplementations(mc, ct, name, signature);
   }


    

}
