// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.java;


import java.util.HashMap;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;


public class KeYRecoderMapping{


    /** have special classes been parsed in */
    private boolean parsedSpecial = false;

    /** maps a recoder programelement (or something similar, e.g. Type)
    * to the KeY-equivalent
    */
    private HashMap map;

    /** maps a KeY programelement to the Recoder-equivalent */
    private HashMap revMap;

    /** maps a fully qualified type name to the KeYJavaType */
    private HashMap typeNames;

    public KeYRecoderMapping() {
	this.map = new HashMap();
	this.revMap = new HashMap();
	this.typeNames = new HashMap();
    }


    /**
    * creates a KeYRecoderMapping object
    * @param map a HashMap mapping ProgramElements in Recoder to
    * ProgramElements in KeY
    * @param revMap the reverse map (KeY->Recoder)
    * @param boolean indicating if the special classes have been parsed in
    */
    public KeYRecoderMapping(HashMap map, HashMap revMap, HashMap typeNames, 
			     boolean parsedSpecial){
        this.map      = map;
        this.revMap   = revMap;
	this.typeNames = typeNames;
	this.parsedSpecial = parsedSpecial;
    }

    /**
    * creates a KeYRecoderMapping object
    * @param map a HashMap mapping ProgramElements in Recoder to
    * ProgramElements in KeY
    * @param revMap the reverse map (KeY->Recoder)
    */
    public KeYRecoderMapping(HashMap map, HashMap revMap, HashMap typeNames){
        this.map      = map;
        this.revMap   = revMap;
	this.typeNames = typeNames;
    }

    /**
    * creates a KeYRecoderMapping object
    * @param map a HashMap mapping ProgramElements in Recoder to
    * ProgramElements in KeY
    * @param revMap the reverse map (KeY->Recoder)
    */
    public KeYRecoderMapping(HashMap map, HashMap revMap){
        this.map      = map;
        this.revMap   = revMap;
	this.typeNames = new HashMap();
    }

    /**
    * returns a matching ProgramElement (KeY) to a given
    * ProgramElement (Recoder)
    * @param pe a recoder.java.ProgramElement
    */
    public ProgramElement toKeY(recoder.java.ProgramElement pe) {
        return (ProgramElement)map.get(pe);
    }

    /**
    * returns a matching ModelElement (KeY) to a given recoder.ModelElement
    * @param pe a recoder.ModelElement
    */
    public ModelElement toKeY(recoder.ModelElement pe) {
        return (ModelElement)map.get(pe);
    }

    /**
    * returns the Recoder-equivalent to a given ProgramElement (KeY).
    * If there's no Recoder-equivalent to the ProgramElement pe
    * a debug message "Program Element <pe> not known" is printed.
    * @param pe a JavaProgramElement
    */
    public recoder.java.ProgramElement toRecoder(ProgramElement pe) {
        Object res=revMap.get(pe);
        Debug.assertTrue(res!=null, "Program Element "+pe+" not known");
        return (recoder.java.ProgramElement)res;
    }


    /**
    * returns the Recoder-equivalent to a given ModelElement (KeY).
    * If there's no Recoder-equivalent to the ModelElement pe a
    * debug message "Model Element <pe> not known" is printed.
    * @param pe a ModelElement
    */
    public recoder.ModelElement toRecoder(ModelElement pe) {
        Object res=revMap.get(pe);
        Debug.assertTrue(res!=null, "Model Element "+pe+" not known");	

        return (recoder.ModelElement)res;
    }

    public void put(Object rec, Object key) {
	if (key instanceof KeYJavaType 
	    && typeNames.containsKey
	         (((recoder.abstraction.Type)rec).getFullName())) {
	    KeYJavaType oldkjt=(KeYJavaType)typeNames.get
		(((recoder.abstraction.Type)rec).getFullName());
	    map.put(rec, oldkjt);
	    revMap.put(oldkjt, rec);
	} else {
	    map.put(rec, key);
	    revMap.put(key, rec);
	    if (key instanceof KeYJavaType) {
		typeNames.put(((recoder.abstraction.Type)rec).getFullName(), key);
	    }
	}
    }

    public boolean mapped(Object rec) {
	return map.containsKey(rec);
    }

    public java.util.Set elemsKeY() {
	return revMap.keySet();
    }

    public java.util.Set elemsRec() {
	return map.keySet();
    }

    public KeYRecoderMapping copy() {
	return new KeYRecoderMapping((HashMap)map.clone(), 
				     (HashMap)revMap.clone(),
				     (HashMap)typeNames.clone(),
				     parsedSpecial);
    }

    /**
     * As long as we do not support lemmata we need the source code of
     * some 'java.lang' classes. These are parsed in using method
     * parseSpecial of {@link Recoder2KeY}. To avoid multiple readings
     * this method indicates whether the special have been parsed in or
     * not. 
     * @return true if special classes have been parsed in
     */
    public boolean parsedSpecial() {
	return parsedSpecial;
    }
    

    /**
     * As long as we do not support lemmata we need the source code of
     * some 'java.lang' classes. These are parsed in using method
     * parseSpecial of {@link Recoder2KeY}. To avoid multiple readings
     * this method sets a flag whether the special have been parsed in or
     * not
     * @param b boolean indicating if the special classes have been
     * parsed in
     */
    public void parsedSpecial(boolean b) {
	parsedSpecial = b;
    }

    public String toString() {
	return "reverse map: "+revMap+"\nmap: "+map+"\ntype names: "+typeNames;
    }

}
