// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.java;

import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.abstraction.*;

import java.util.Enumeration;
import de.uka.ilkd.key.java.visitor.Visitor;

/**
 *    Statement block.
 * taken from COMPOST and changed to achieve an immutable structure
 */

public class StatementBlock
    extends JavaStatement
    implements StatementContainer, TypeDeclarationContainer,
               VariableScope, TypeScope {

    /**
     *      Body.
     */

    private final ArrayOfStatement body;

    public StatementBlock() {
	body = new ArrayOfStatement(new Statement[0]);
    }

    /**
     *      Statement block.
     *  @param children an ExtList that contains the children
     */

    public StatementBlock(ExtList children) {
	super(children);
	body=new
	    ArrayOfStatement((Statement[])children.collect(Statement.class)); 
    }



    public StatementBlock(ArrayOfStatement as) {
	body=as;
    }

    public StatementBlock(Statement as) {
	this(new ArrayOfStatement(new Statement[]{as}));
    }

    public StatementBlock(Statement[] body) {
	this(new ArrayOfStatement(body));
    }

    /**
 *      Get body.
 *      @return the statement array wrapper.
     */

    public ArrayOfStatement getBody() {
        return body;
    }

    public boolean isEmpty() {
	return body.size()==0;
    }


    /**
 *      Returns the number of children of this node.
 *      @return an int giving the number of children of this node
    */

    public int getChildCount() {
        return (body != null) ? body.size() : 0;
    }

    /**
 *      Returns the child at the specified index in this node's "virtual"
 *      child array
 *      @param index an index into this node's "virtual" child array
 *      @return the program element at the given position
 *      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
 *                 of bounds
    */

    public ProgramElement getChildAt(int index) {
        if (body != null) {
            int len = body.size();
            if (len > index) {
                return body.getStatement(index);
            }
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    public int getChildPositionCode(ProgramElement child) {
        // role 0 (IDX): statements
        if (body != null) {
            int index = getArrayPos(body,child);
            if (index >= 0) {
                return (index << 4) | 0;
            }
        }
        return -1;
    }

    /**
     *      Get the number of statements in this container.
     *      @return the number of statements.
     */

    public int getStatementCount() {
        return (body != null) ? body.size() : 0;
    }

    /*
      Return the statement at the specified index in this node's
      "virtual" statement array.
      @param index an index for a statement.
      @return the statement with the given index.
      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
      of bounds.
    */

    public Statement getStatementAt(int index) {
        if (body != null) {
            return body.getStatement(index);
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    /**
 *      Get the number of type declarations in this container.
 *      @return the number of type declarations.
     */

    public int getTypeDeclarationCount() {
        int count = 0;
        if (body != null) {
            for (int i = body.size() - 1; i >= 0; i -= 1) {
                if (body.getStatement(i) instanceof TypeDeclaration) {
                    count += 1;
                }
            }
        }
        return count;
    }

    /*
      Return the type declaration at the specified index in this node's
      "virtual" type declaration array.
      @param index an index for a type declaration.
      @return the type declaration with the given index.
      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
      of bounds.
    */

    public TypeDeclaration getTypeDeclarationAt(int index) {
        if (body != null) {
            int s = body.size();
            for (int i = 0; i < s && index >= 0; i += 1) {
                Statement st = body.getStatement(i);
                if (st instanceof TypeDeclaration) {
                    if (index == 0) {
                        return (TypeDeclaration)st;
                    }
                    index -= 1;
                }
            }
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnStatementBlock(this);
    }

    public void prettyPrint(PrettyPrinter p) throws java.io.IOException {
        p.printStatementBlock(this);
    }
}
