// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.java;

import java.util.HashMap;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.expression.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.expression.literal.*;
import de.uka.ilkd.key.java.expression.operator.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.logic.ldt.LDT;
import de.uka.ilkd.key.logic.ldt.IntegerLDT;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.SortDefiningSymbols;
import de.uka.ilkd.key.logic.sort.ArraySort;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.ldt.*;
import de.uka.ilkd.key.gui.ProofSettings;
import de.uka.ilkd.key.java.recoderext.KeYCrossReferenceServiceConfiguration;
import recoder.service.DefaultConstantEvaluator;
import recoder.service.ConstantEvaluator;


public class TypeConverter {


    /** counts number of created object used to create unique names. 
     * %%%% HACK! 
     */
    private static long OBJECT_COUNTER = 0;

    private Services services;

    private TermFactory termFab = TermFactory.DEFAULT;

    private LDT integerLDT;
    private LDT booleanLDT;
    private LDT objectLDT;
    private LDT[] models = new LDT[0];


    public TypeConverter(Services s){
        services = s;
    }

    public void init(LDT integerLDT, LDT booleanLDT, LDT objectLDT) {
	this.integerLDT = integerLDT;
	this.booleanLDT = booleanLDT;
	this.objectLDT = objectLDT;
	this.models = new LDT[]{booleanLDT, integerLDT, objectLDT};
    }

    public LDT[] getModels() {
	return models;
    }

    public Services getServices() {
	return services;
    }

    public LDT getModelFor(Type t) {
	if (integerLDT == null || booleanLDT == null || objectLDT == null) {
	    return  null;
	}
	if (integerLDT.javaType().contains(t)) {
	    return integerLDT;
	} else if (booleanLDT.javaType().contains(t)) {
	    return booleanLDT;
	} else if (objectLDT.javaType().contains(t)) {
	    return objectLDT;
	}
	Debug.out("typeconverter: No LDT found for ", t);
	return null;
    }

    public LDT getModelFor(Sort s) {
	if (integerLDT == null || booleanLDT == null || objectLDT == null) {
	    return  null;
	}
	if (integerLDT.targetSort() == s) {
	    return integerLDT;
	} else if (booleanLDT.targetSort() == s) {	    
	    return booleanLDT;
	} else if (objectLDT.targetSort() == s) {
	    return booleanLDT;
	}
	Debug.out("No LDT found for ", s);
	return null;
    }

    public NumberLDT getIntegerLDT() {
	return (NumberLDT)integerLDT;
    }

    public Sort getObjectSort() {
	if (objectLDT==null) return null;
	return objectLDT.targetSort();
    }


    public KeYJavaType getObjectKeYJavaType() {
	if (objectLDT==null) return null;
	return objectLDT.getKeYJavaType(objectLDT.javaType().head());
    }
 
		     
    private Term translateOperator(de.uka.ilkd.key.java.expression.Operator op,
				   LDT intLDT, LDT booleanLDT, LDT objectLDT) {
	Term[] subs  = new Term[op.getArity()];
	if (op.getArity() >= 1) {
	    subs[0] = convertToLogicElement(op.getExpressionAt(0));
	}
	if (op.getArity() == 2) {
	    subs[1] = convertToLogicElement(op.getExpressionAt(1));	  
	}	
	
	LDT responsibleLDT = null;
	if (intLDT.isResponsible(op, subs)) {
	    responsibleLDT = intLDT;
	} else if (booleanLDT.isResponsible(op, subs)) {
	    responsibleLDT = booleanLDT;
	} else if (objectLDT.isResponsible(op, subs)) {
	    responsibleLDT = objectLDT;
	} else{
	    Debug.out("typeconverter: no data type model "+
		      "available to convert:", op, op.getClass());		
	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this");
	}
	return termFab.createFunctionTerm
	    (responsibleLDT.getFunctionFor(op), subs);
    }


    private Term convertReferencePrefix(ReferencePrefix prefix) {
	Debug.out("typeconverter: (prefix, class)", prefix, 
		  (prefix != null ? prefix.getClass() : null));		
	if (prefix instanceof EmbededVariableReference) {
	    return termFab.createVariableTerm
		(((EmbededVariableReference) prefix).getProgramVariable());
	} else if (prefix instanceof FieldReference) {
	    return convertToAttributeTerm
		((FieldReference) prefix);
	} else if (prefix instanceof ThisReference) {
	    Debug.out("typeconverter: "+
		      "WARNING: this-references not supported yet");		
	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this");
	} else if (prefix instanceof SuperReference) {
	    Debug.out("typeconverter: "+
		      "WARNING: super-references not supported yet");

	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this");
	} else if (prefix instanceof MetaClassReference) {
	    Debug.out("typeconverter: "+
		      "WARNING: metaclass-references not supported yet");
	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this");
	} else 	if (prefix instanceof ProgramVariable) {
	    // the base case: the leftmost item is a local variable
	    return termFab.createVariableTerm
		((ProgramVariable)prefix);
	} else 	if (prefix instanceof VariableReference) {
	    Debug.out("typeconverter: "+
		      "variablereference:", (((VariableReference)prefix).getProgramVariable()));
	    return termFab.createVariableTerm
		(((VariableReference)prefix).getProgramVariable());
	}  else if (prefix instanceof TypeReference) {	   
	    Function typeConst 
		= ((TypeReference)prefix).getKeYJavaType().typeConst();
	    return termFab.createFunctionTerm(typeConst);
	}  else if (prefix instanceof ArrayReference) {	   
	    return convertToArrayTerm(prefix);
	}else {
	    System.out.println("typeconverter: "+
		      "WARNING: unknown reference prefix:"+ prefix+ 
		      prefix.getClass());
	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this");
	}	
    }


    public Term convertToAttributeTerm(VariableReference fr) {	       
	ReferencePrefix prefix = fr.getReferencePrefix();
	Debug.out("TypeConverter: FieldReference: ",fr);
	Term sub = null;
	if (prefix==null && fr instanceof EmbededVariableReference) {
	    return termFab.createVariableTerm
		(((EmbededVariableReference)fr).getProgramVariable());
	}
	if (prefix != null && !(prefix instanceof PackageReference) ) {
	    sub = convertReferencePrefix(prefix);
	}
	if (sub != null) {
	    return termFab.createAttributeTerm(fr.getProgramVariable(), sub);
	} else {
	    System.out.println("typeconverter: Not supported reference type (fr, class):"+
		      fr+ fr.getClass());
	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this");
	    //return errorTerm;
	}
    }
    
    private Term convertNewOperator(New create) {
	OBJECT_COUNTER = OBJECT_COUNTER + 1;
        String typeName = create.getTypeReference().getName();
	Term t = termFab.createFunctionTerm
	    (new ProgramObject
		(new Name("obj_"+OBJECT_COUNTER+"|:"+typeName),
		 create.getTypeReference().getKeYJavaType()));
	return t;
    }
    
    private Term convertNewArrayOperator(NewArray create) {
	OBJECT_COUNTER = OBJECT_COUNTER + 1;
        String typeName = create.getTypeReference().getName();
	String arrayName = "(array_"+OBJECT_COUNTER+"_"+create.getTypeReference().getKeYJavaType().getSort();
        if (create.getArguments() != null) {
            for (int i=0; i < create.getArguments().size(); i ++) {
                arrayName +="[";
                arrayName += create.getArguments().getExpression(i);
                arrayName += "]";
            }
        }
        for (int i=0; i < create.getDimensions()-create.getArguments().size(); i += 1) {
            arrayName +="[]";
        }
        if (create.getArrayInitializer() != null) {
            arrayName += create.getArrayInitializer();
        }
	arrayName +=")";
	Term t = termFab.createFunctionTerm
	    (ProgramArray.createProgramArray
		(new Name(arrayName), create.getKeYJavaType(), create));
	return t;
    }

    private Term convertArrayInitializer(ArrayInitializer ai) {
	OBJECT_COUNTER = OBJECT_COUNTER + 1;
        String typeName = ai.getKeYJavaType(services).getSort().toString();
	String arrayName = "(array_"+OBJECT_COUNTER+"_"+ai.getKeYJavaType(services).getSort();
	arrayName += ai;
	arrayName +=")";
	Term t = termFab.createFunctionTerm
	    (ProgramArray.createProgramArray
		(new Name(arrayName),
		 ai.getKeYJavaType(services), ai));
	return t;
    }

    public Term convertToArrayTerm(ProgramElement pe){
        ArrayReference a = (ArrayReference) pe;
        Term[] index = new Term[a.getDimensionExpressions().size()];
        Term t = convertToLogicElement(a.getReferencePrefix());
         for(int i=0;i<index.length;i++) index[i]=
             convertToLogicElement(
             a.getDimensionExpressions().getExpression(i));
        return termFab.createArrayTerm(Op.ARRAY, t, index);
    }

    private Term convertToInstanceofTerm(Instanceof io) {
	Term obj=convertToLogicElement(io.getChildAt(0));
	KeYJavaType type=((TypeReference)io.getChildAt(1)).getKeYJavaType();
	SortDefiningSymbols s=(SortDefiningSymbols)(type.getSort());
	InstanceofSymbol func
	    = (InstanceofSymbol)s.lookupSymbol(InstanceofSymbol.NAME);
	return termFab.createFunctionTerm(func, obj);
    }

    public Term convertToLogicElement(ProgramElement pe) {
	return convertToLogicElement(pe, integerLDT, booleanLDT, objectLDT);
    }

    public Term convertToLogicElement(ProgramElement pe, 
				      LDT intLDT, LDT booleanLDT,
				      LDT objectLDT) {
	Debug.out("typeconverter: called for:", pe, pe.getClass());
	if (pe instanceof ProgramVariable) {	    
	    return termFab.createFunctionTerm((ProgramVariable)pe);
	} else if (pe instanceof NullLiteral) {
	    return services.getJavaInfo().getNullConst();
	} else if (pe instanceof EmbededVariableReference) {
	    return convertToAttributeTerm((VariableReference)pe);
	//     return termFab.createFunctionTerm
// 		(((EmbededVariableReference)pe).getProgramVariable());
	} else if (pe instanceof FieldReference) {
	    return convertToAttributeTerm((FieldReference)pe);
	} else if (pe instanceof ArrayReference){
	    return convertToArrayTerm(pe);
	} else if (pe instanceof VariableReference) {
	    return termFab.createFunctionTerm
		(((VariableReference)pe).getProgramVariable());	      
	} else if (pe instanceof IntLiteral) {
	    return intLDT.translateLiteral((IntLiteral)pe);
	} else if (pe instanceof ParenthesizedExpression) {
	    return convertToLogicElement
		(((ParenthesizedExpression)pe).getChildAt(0), 
		 intLDT, booleanLDT, objectLDT);
	} else if (pe instanceof New) {
	    return convertNewOperator((New) pe);
	} else if (pe instanceof NewArray) {
	    return convertNewArrayOperator((NewArray) pe);
	} else if (pe instanceof ArrayInitializer){
	    return convertArrayInitializer((ArrayInitializer) pe); 
	} else if (pe instanceof LongLiteral) {
	    return intLDT.translateLiteral((LongLiteral)pe);
	} else if (pe instanceof CharLiteral) {
	    return intLDT.translateLiteral((CharLiteral)pe);
	} else if (pe instanceof Negative 
		   && ((Negative)pe).getChildAt(0) instanceof IntLiteral) {
	    String val = ((IntLiteral)((Negative)pe).getChildAt(0)).getValue();
	    if (val.charAt(0)=='-') {
		return intLDT.translateLiteral
		    (new IntLiteral(val.substring(1)));
	    } else {
		return intLDT.translateLiteral
		    (new IntLiteral("-"+val));
	    }
	} else if (pe instanceof Negative 
		   && ((Negative)pe).getChildAt(0) instanceof LongLiteral ) {
	    String val = ((LongLiteral)((Negative)pe).getChildAt(0)).getValue();
	    if (val.charAt(0)=='-') {
		return intLDT.translateLiteral
		    (new LongLiteral(val.substring(1)));
	    } else {
		return intLDT.translateLiteral
		    (new LongLiteral("-"+val));
	    }
	} else if (pe instanceof Instanceof) {
	    return convertToInstanceofTerm((Instanceof)pe);
	} else if (pe instanceof de.uka.ilkd.key.java.expression.Operator) {
	    return translateOperator
		((de.uka.ilkd.key.java.expression.Operator)pe,
		 intLDT, booleanLDT, objectLDT);
	} else if (pe instanceof BooleanLiteral) {
	    return booleanLDT.translateLiteral((BooleanLiteral) pe);
	} else if (pe instanceof PrimitiveType) {
	    //	    return convertNewOperator((PrimitiveType) pe);
	    throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this primitive type");
	    //return errorTerm;
	}
	throw new IllegalArgumentException("TypeConverter could not handle"
					       +" this logic element.");
	//return errorTerm;
    }

    private boolean isArithmeticOperator
	(de.uka.ilkd.key.java.expression.Operator op) {
	if (op instanceof Divide || op instanceof Times || 
	    op instanceof Plus || op instanceof Minus || 
	    op instanceof Modulo || op instanceof ShiftLeft ||
	    op instanceof ShiftRight || op instanceof BinaryAnd || 
	    op instanceof BinaryNot || op instanceof BinaryOr || 
	    op instanceof BinaryXOr || op instanceof Negative || 
	    op instanceof PreIncrement || op instanceof PostIncrement ||
	    op instanceof PreDecrement || op instanceof PostDecrement) {
	    return true;
	} 
	return false;
    }


    private boolean isCompareOperator
	(de.uka.ilkd.key.java.expression.Operator op) {
	if (op instanceof GreaterThan || op instanceof GreaterOrEquals ||
	    op instanceof LessThan || op instanceof LessOrEquals || 
	    op instanceof Equals || op instanceof LogicalNot
	    || op instanceof LogicalAnd || op instanceof LogicalOr) {
	    return true;
	}
	return false;
    }


    private boolean isBooleanOperator
	(de.uka.ilkd.key.java.expression.Operator op) {
	if (op instanceof LogicalAnd || op instanceof LogicalOr ||
	    op instanceof LogicalNot) {
	    return true;
	}
	return false;
    }


    /**
     * performs binary numeric promotion on the argument types
     */
    public KeYJavaType getPromotedType(KeYJavaType type1, 
				       KeYJavaType type2) {
	if ((type1.getName().equals("boolean") &&
	     type2.getName().equals("boolean")))
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_BOOLEAN);
	if ((type1.getName().equals("byte") ||
	     type1.getName().equals("short")||
	     type1.getName().equals("char") ||
	     type1.getName().equals("int")) &&
	    (type2.getName().equals("byte") ||
	     type2.getName().equals("short")||
	     type2.getName().equals("char") ||
	     type2.getName().equals("int")))
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_INT);
	if ((type2.getName().equals("long")) &&
	    (type1.getName().equals("byte") ||
	     type1.getName().equals("short")||
	     type1.getName().equals("int")  ||
	     type1.getName().equals("char") ||
	     type1.getName().equals("long"))) 
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_LONG);
	if ((type1.getName().equals("long")) &&
	    (type2.getName().equals("byte") ||
	     type2.getName().equals("short")||
	     type2.getName().equals("char") ||
	     type2.getName().equals("int")  ||
	     type2.getName().equals("long")))
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_LONG);	    
	throw new RuntimeException("Could not determine promoted type "+
				   "of "+type1+" and "+type2);
    }


    // this method performs unary numeric promotion on the arguments
    public KeYJavaType getPromotedType(KeYJavaType type1) {
	if (type1.getName().equals("boolean"))
	    // not really numeric ...
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_BOOLEAN);
	if (type1.getName().equals("byte") ||
	    type1.getName().equals("short")||
	    type1.getName().equals("char") ||
	    type1.getName().equals("int"))
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_INT);
	if (type1.getName().equals("long"))
	    return services.getJavaInfo().getKeYJavaType(PrimitiveType.JAVA_LONG);
	throw new RuntimeException("Could not determine promoted type "+
				   "of "+type1);
    }



    public Type resolveType
	(de.uka.ilkd.key.java.expression.Operator op) {	
	if (isArithmeticOperator(op)) {
	    return resolveArithmeticType(op);
	} else if (isCompareOperator(op)) {
	    return booleanLDT.javaType().head();
	} else if (isBooleanOperator(op)) {
	    return booleanLDT.javaType().head();
	}
	Debug.out("typeconverter: the return type of the given "+
		  "operator cannot be determined: ", op);
	return null;
    }

    private Type resolveArithmeticType
	(de.uka.ilkd.key.java.expression.Operator op) {	
	if (op instanceof BinaryAnd || 
	    op instanceof BinaryOr) {
	    return (getKeYJavaType((Expression)op.
				   getChildAt(0)).getJavaType());
	}
	// works to simple needs to be adjust in order 
	// to support real, double
	return integerLDT.javaType().head();
    }

    public Sort getPrimitiveSort(Type t) {
        LDT result = getModelFor(t);
        return (result == null ? null : result.targetSort());
    }

    public KeYJavaType getKeYJavaType(Expression e) {
	return e.getKeYJavaType(services);
    }

    public Expression convertToProgramElement(Term term) {
	if (term.op()==Op.NULL) {
	    return NullLiteral.NULL;
	} else if (term.op() instanceof Function) {
	    for (int i=0; i<models.length; i++) {
		if (models[i].hasLiteralFunction((Function)term.op())) {
		    return models[i].translateTerm(term, null);	       
		}
	    }
	}
	ExtList children=new ExtList();
	for (int i=0; i<term.arity(); i++) {
	    children.add(convertToProgramElement(term.sub(i)));
	}
	if (term.op() instanceof ProgramInLogic) {
	    return ((ProgramInLogic)term.op()).convertToProgram(term, children);
	} else if (term.op() instanceof Function) {
	    for (int i=0; i<models.length; i++) {
		if (models[i].containsFunction((Function)term.op())) {
		    return models[i].translateTerm(term, children);
		}
	    }
	} 
	throw new RuntimeException("Cannot convert term to program: "+term
				   +" "+term.op().getClass());
    }

    public KeYJavaType getKeYJavaType(Term t) {
	if (t.op() instanceof ProgramInLogic) {
	    return ((ProgramInLogic)t.op()).getKeYJavaType(t);
	}
	return getKeYJavaType(convertToProgramElement(t));
    }

    public KeYJavaType getKeYJavaType(Type t) {
	if ( t instanceof KeYJavaType )
	    return (KeYJavaType)t;
	LDT model = getModelFor(t);
	if (model != null) {
	    return model.getKeYJavaType(t);
	} else {
	    Debug.out("javainfo: no predefined model for type ",t);
	    return null;
	}	
    }


    /** These methods are taken from recoder (and modified) */

    public boolean isWidening(PrimitiveType from, PrimitiveType to) {
	// we do not handle null's
	if (from == null || to == null) return false;
	// equal types can be coerced
	if (from == to ) return true;
	// boolean types cannot be coerced into something else
	if (from == PrimitiveType.JAVA_BOOLEAN ||
	    to == PrimitiveType.JAVA_BOOLEAN)	    
	    return false;
	// everything else can be coerced to a double
	if (to == PrimitiveType.JAVA_DOUBLE) return true;
	// but a double cannot be coerced to anything else
	if (from == PrimitiveType.JAVA_DOUBLE) return false;
	// everything except doubles can be coerced to a float
	if (to == PrimitiveType.JAVA_FLOAT) return true;
	// but a float cannot be coerced to anything but float or double
	if (from == PrimitiveType.JAVA_FLOAT) return false;	
	// everything except float or double can be coerced to a long
	if (to == PrimitiveType.JAVA_LONG) return true;
	// but a long cannot be coerced to anything but float, double or long
	if (from == PrimitiveType.JAVA_LONG) return false;
	// everything except long, float or double can be coerced to an int
	if (to == PrimitiveType.JAVA_INT) return true;
	// but an int cannot be coerced to the remaining byte, char, short
	if (from == PrimitiveType.JAVA_INT) return false;
	// between byte, char, short, only one conversion is admissible
	return (from == PrimitiveType.JAVA_BYTE &&
		to == PrimitiveType.JAVA_SHORT);
    }

    public boolean isWidening(ClassType from, ClassType to) {
	return isWidening ( getKeYJavaType ( from ),
			    getKeYJavaType ( to   ) );
    }

    public boolean isWidening(ArrayType from, ArrayType to) {
	KeYJavaType toBase   =  to.getBaseType().getKeYJavaType();
	if (toBase == getObjectKeYJavaType() ) { // seems incorrect
	    return true;
	}
	KeYJavaType fromBase = from.getBaseType().getKeYJavaType();
	if (toBase.getJavaType () instanceof PrimitiveType) {
	    return toBase == fromBase;
	}
	return isWidening(fromBase, toBase);
    }
    
    public boolean isWidening(Type from, Type to) {
	if ( from instanceof KeYJavaType )
	    return isWidening ( (KeYJavaType)from,
				getKeYJavaType ( to ) );
	if ( to   instanceof KeYJavaType )
	    return isWidening ( getKeYJavaType ( from ),
				(KeYJavaType)to );

	if (from instanceof ClassType) {
	    return isWidening ( getKeYJavaType ( from ),
				getKeYJavaType ( to   ) );
	} else if (from instanceof PrimitiveType) {
	    if (to instanceof PrimitiveType) {
		return isWidening((PrimitiveType)from, (PrimitiveType)to);
	    }
	} else if (from instanceof ArrayType) {
	    if (to instanceof ClassType) {
		if (getKeYJavaType ( to ) == getObjectKeYJavaType() ) {
		    return true;
		}
		if (to.getFullName().equals("java.lang.Cloneable") || 
		    to.getFullName().equals("Cloneable")) {
		    return true;
		}
		if (to.getFullName().equals("java.io.Serializable") || 
		    to.getFullName().equals("Serializable")) {
		    return true;
		}
		return false;
	    } else if (to instanceof ArrayType) {
		return isWidening((ArrayType)from, (ArrayType)to);
	    }
	}
	return false;
    }    

    // this also handles class types
    public boolean isWidening(KeYJavaType from, KeYJavaType to) {
	Type a = from.getJavaType ();
	Type b = to  .getJavaType ();

	if ( a instanceof ClassType || a == null ) {
	    return
		from.getSort ().extendsTrans ( to.getSort () ) ||
		( a == NullType.JAVA_NULL &&
		  b instanceof ArrayType );
	} else {
	    if ( b == null )
		return
		    to == getObjectKeYJavaType () &&
		    a instanceof ArrayType;
	    else
		return isWidening ( a, b );
	}
    }


    public boolean isImplicitNarrowing
	(Expression expr, PrimitiveType to) {
	
	int minValue, maxValue;
	if (to == PrimitiveType.JAVA_BYTE) {
	    minValue = Byte.MIN_VALUE;
	    maxValue = Byte.MAX_VALUE;
	} else if (to == PrimitiveType.JAVA_CHAR) {
	    minValue = Character.MIN_VALUE;
	    maxValue = Character.MAX_VALUE;
	} else if (to == PrimitiveType.JAVA_SHORT) {
	    minValue = Short.MIN_VALUE;
	    maxValue = Short.MAX_VALUE;
	} else {
	    return false;
	}

	ConstantExpressionEvaluator cee = 
	    services.getConstantExpressionEvaluator();

	ConstantEvaluator.EvaluationResult res = 
	    new ConstantEvaluator.EvaluationResult();

	if (!cee.isCompileTimeConstant(expr, res) || 
	    res.getTypeCode() != ConstantEvaluator.INT_TYPE) {
	    return false;
	}
	int value = res.getInt();
	return (minValue <= value) && (value <= maxValue);	
    }

    public boolean isIdentical ( Type from, Type to ) {
	from = getKeYJavaType ( from );
	to   = getKeYJavaType ( to   );
	return from == to;
    }

    public boolean isAssignableTo ( Type from, Type to ) {
	return isIdentical ( from, to ) || isWidening ( from, to );
    }

    public boolean isAssignableTo ( Expression expr, Type to ) {
	return
	    ( to instanceof PrimitiveType &&
	      isImplicitNarrowing ( expr, (PrimitiveType)to ) ) ||
	    isIdentical ( expr.getKeYJavaType ( getServices () ), to ) ||
	    isWidening  ( expr.getKeYJavaType ( getServices () ), to );
    }

    // this also handles class types
    public boolean isNarrowing(KeYJavaType from, KeYJavaType to) {
	Type a = from.getJavaType ();
	Type b = to  .getJavaType ();

	if ( a instanceof ClassType || a == null ) {
	    return
		to.getSort ().extendsTrans ( from.getSort () ) ||
		( from == getObjectKeYJavaType () &&
		  a instanceof ArrayType );
	} else {
	    if ( b == null )
		return false;
	    else
		return isNarrowing ( a, b );
	}
    }

    public boolean isNarrowing(PrimitiveType from, PrimitiveType to) {
	// we do not handle null's
	if (from == null || to == null) return false;

	if (from == PrimitiveType.JAVA_BYTE) {
	    return (to == PrimitiveType.JAVA_CHAR );
	}
	if (from == PrimitiveType.JAVA_SHORT) {
	    return (to == PrimitiveType.JAVA_BYTE ||
		    to == PrimitiveType.JAVA_CHAR);
	}
	if (from == PrimitiveType.JAVA_CHAR) {
	    return (to == PrimitiveType.JAVA_BYTE ||
		    to == PrimitiveType.JAVA_SHORT);
	}
	if (from == PrimitiveType.JAVA_INT) {
	    return (to == PrimitiveType.JAVA_BYTE ||
		    to == PrimitiveType.JAVA_SHORT ||
		    to == PrimitiveType.JAVA_CHAR);
	}
	if (from == PrimitiveType.JAVA_LONG) {
	    return (to == PrimitiveType.JAVA_BYTE ||
		    to == PrimitiveType.JAVA_SHORT ||
		    to == PrimitiveType.JAVA_CHAR ||
		    to == PrimitiveType.JAVA_INT);
	}
	if (from == PrimitiveType.JAVA_FLOAT) {
	    return (to == PrimitiveType.JAVA_BYTE ||
		    to == PrimitiveType.JAVA_SHORT ||
		    to == PrimitiveType.JAVA_CHAR ||
		    to == PrimitiveType.JAVA_INT ||
		    to == PrimitiveType.JAVA_LONG);
	}
	if (from == PrimitiveType.JAVA_DOUBLE) {
	    return (to == PrimitiveType.JAVA_BYTE ||
		    to == PrimitiveType.JAVA_SHORT ||
		    to == PrimitiveType.JAVA_CHAR ||
		    to == PrimitiveType.JAVA_INT ||
		    to == PrimitiveType.JAVA_LONG ||
		    to == PrimitiveType.JAVA_FLOAT);
	}
	return false;
    }

    public boolean isNarrowing(ClassType from, ClassType to) {
	return isNarrowing ( getKeYJavaType ( from ),
			     getKeYJavaType ( to   ) );
    }

    public boolean isNarrowing(ArrayType from, ArrayType to) {
	KeYJavaType toBase   = to.getBaseType().getKeYJavaType();
	KeYJavaType fromBase = from.getBaseType().getKeYJavaType();
	return
	    isReferenceType ( toBase   ) &&
	    isReferenceType ( fromBase ) &&
	    isNarrowing(fromBase, toBase);
    }

    public boolean isNarrowing(Type from, Type to) {
	if ( from instanceof KeYJavaType )
	    return isNarrowing ( (KeYJavaType)from,
				getKeYJavaType ( to ) );
	if ( to   instanceof KeYJavaType )
	    return isNarrowing ( getKeYJavaType ( from ),
				(KeYJavaType)to );

	if (from instanceof ClassType) {
	    return isNarrowing ( getKeYJavaType ( from ),
				 getKeYJavaType ( to   ) );
	} else if (from instanceof PrimitiveType) {
	    if (to instanceof PrimitiveType) {
		return isNarrowing((PrimitiveType)from, (PrimitiveType)to);
	    }
	} else if (from instanceof ArrayType) {
	    if (to instanceof ArrayType) {
		return isNarrowing((ArrayType)from, (ArrayType)to);
	    }
	}
	return false;
    }

    public boolean isCastingTo ( Type from, Type to ) {
	// there is currently no interface handling

	// identity conversion
	if ( isIdentical ( from, to ) )
	    return true;

	// conversions between numeric types are always possible
	if ( isArithmeticType ( from ) &&
	     isArithmeticType ( to   ) )
	    return true;

	// all widening conversions
	if ( isWidening ( from, to ) )
	    return true;

	// narrowing
	return isNarrowing ( from, to );
    }

    public boolean isArithmeticType ( Type t ) {
	if ( t instanceof KeYJavaType )
	    t = ((KeYJavaType)t).getJavaType ();
	return
	    t == PrimitiveType.JAVA_BYTE   ||
	    t == PrimitiveType.JAVA_SHORT  ||
	    t == PrimitiveType.JAVA_INT    ||
	    t == PrimitiveType.JAVA_CHAR   ||
	    t == PrimitiveType.JAVA_LONG   ||
	    t == PrimitiveType.JAVA_FLOAT  ||
	    t == PrimitiveType.JAVA_DOUBLE;
    }

    public boolean isIntegralType ( Type t ) {
	if ( t instanceof KeYJavaType )
	    t = ((KeYJavaType)t).getJavaType ();
	return
	    t == PrimitiveType.JAVA_BYTE   ||
	    t == PrimitiveType.JAVA_SHORT  ||
	    t == PrimitiveType.JAVA_INT    ||
	    t == PrimitiveType.JAVA_CHAR   ||
	    t == PrimitiveType.JAVA_LONG;
    }

    public boolean isReferenceType ( Type t ) {
	if ( t instanceof KeYJavaType )
	    t = ((KeYJavaType)t).getJavaType ();
	return
	    // there is currently no interface handling
	    t == null ||
	    ( t instanceof ClassType && !( t instanceof NullType ) ) ||
	    t instanceof ArrayType;
    }

    public boolean isNullType ( Type t ) {
	if ( t instanceof KeYJavaType )
	    t = ((KeYJavaType)t).getJavaType ();
	return
	    t == NullType.JAVA_NULL;
    }

    public boolean isBooleanType ( Type t ) {
	if ( t instanceof KeYJavaType )
	    t = ((KeYJavaType)t).getJavaType ();
	return
	    t == PrimitiveType.JAVA_BOOLEAN;
    }

    public TypeConverter copy(Services services) {
	TypeConverter tc = new TypeConverter(services);
	tc.init(integerLDT, booleanLDT, objectLDT);
	return tc;
    }

}
