// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.java.declaration;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.java.visitor.Visitor;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.declaration.modifier.Public;
import de.uka.ilkd.key.java.expression.Literal;
import de.uka.ilkd.key.java.expression.literal.NullLiteral;
import de.uka.ilkd.key.java.reference.TypeReference;
/**
 * KeY used to model arrays using only the {@link
 * de.uka.ilkd.key.java.abstraction.ArrayType}. As the only attribute of
 * an array has been the length attribute, it has been handled in a
 * different way than members of usual classes. As we need some implicit
 * fields for array creation and initialisation, the special handling of
 * arrays is not any longer practicable. 
 *  So, this class introduce a 'virtual' declaration for array types
 * containing all required members. Please not the array fields accessed
 * by an index are not included, so arrays of different lengths with same base
 * type belong to the same array declaration.
 * Attention: In contrast to the other type declaration, array
 * declarations may be added at runtime.
 */

public class ArrayDeclaration 
    extends TypeDeclaration implements ArrayType {


    /** 
     * reference to the type the elements of this array must subclass
     */
    private final TypeReference basetype;

    /**
     * dimension of the array 
     */
    private final int dim;


    /**
     * Array Declaration
     * @param basetype a TypeReference to the element type of the array
     * @param members an array containing the memberdeclarations of
     * this type
     */
    public ArrayDeclaration(TypeReference basetype,
			    MemberDeclaration[] members) {
	super(new Modifier[]{new Public()}, createName(basetype), members, false);
	this.basetype = basetype;
	this.dim      = dimension();
    }

    /**
     * ArrayDeclaration
     * @param children an ExtList with the basetype and member
     * declarations of this type
     */
    public ArrayDeclaration(ExtList children) { 
	super(children, createName
	      ((TypeReference)children.get(TypeReference.class)));
	this.basetype = (TypeReference)children.get(TypeReference.class);
	this.dim      = dimension();
    } 

    /**
     * Returns the number of children of this node.
     * @return an int giving the number of children of this node
     */
    public int getChildCount() {
        int result = 0;
        if (modArray != null) result += modArray.size();
        if (name     != null) result ++;
	if (basetype != null) result ++;
        if (members  != null) result += members.size();
        return result;
    }

    /**
     * Returns the child at the specified index in this node's "virtual"
     * child array
     * @param index an index into this node's "virtual" child array
     * @return the program element at the given position
     * @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *    of bounds
     */
    public ProgramElement getChildAt(int index) {
        int len;
        if (modArray != null) {
            len = modArray.size();
            if (len > index) {
                return modArray.getModifier(index);
            }
            index -= len;
        }
        if (name != null) {
            if (index == 0) return name;
            index--;
        }
        if (basetype != null) {
            if (index == 0) return basetype;
            index--;
        }
        if (members != null) {
            return members.getMemberDeclaration(index);
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    public int getChildPositionCode(ProgramElement child) {
        // role 0 (IDX): modifier
        // role 1: identifier
        // role 2: basetype
        // role 3: (IDX): members
        if (modArray != null && modArray.size()>0) {
	    int index = getArrayPos(modArray,child);
            if (index >= 0) {
                return (index << 4) | 0;
            }
        }
        if (name     == child) return 1;
        if (basetype == child) return 2;
        if (members  != null) {
            int index = getArrayPos(members,child);
            if (index >= 0) {
                return (index << 4) | 3;
            }
        }
        return -1;
    }

    /**
     * Get the element/base type.
     * @return refernce to the base type .
     */
    public TypeReference getBaseType() {
        return basetype;
    }

    /**
     * Arrays are never strictfp.
     */
    public boolean isStrictFp() {
        return false;
    }

    /**
     * Arrays are never transient.
     */
    public boolean isTransient() {
        return false;
    }

    /**
     * Arrays are never volatile.
     */
    public boolean isVolatile() {
        return false;
    }

    /**
     * Arrays are never interfaces (but may have interface types as
     * element types)
     */
    public boolean isInterface() {
        return false;
    }
    
    /**
     * return the dimension of this array
     */
    public int getDimension() {
	return dim;
    }

    /** 
     * returns the default value of the given type 
     * according to JLS 4.5.5
     * @return the default value of the given type 
     * according to JLS 4.5.5
     */
    public Literal getDefaultValue() {
	return NullLiteral.NULL;
    }

    /**
     * computes the dimension of this array
     */
    private int dimension() {
	Type javaType = basetype.getKeYJavaType().getJavaType(); 
	if (javaType instanceof ArrayType) {
	    return 1+((ArrayType)javaType).getDimension();
	} else {
	    return 1;
	}
    }

    private static ProgramElementName createName(TypeReference basetype) {
	// xxx hack! Because the mapping from sort "int" or "arithInt" to
	// JavaType is not unique
	Type javaBasetype = basetype.getKeYJavaType().getJavaType();
	if (javaBasetype == null) {
	    // entered only if base type is class type
	    return new ProgramElementName
		("[L"+basetype.getName());
	    
	}
	if (javaBasetype instanceof TypeDeclaration) {
	    return new ProgramElementName("[L"+javaBasetype);
	} else if (javaBasetype instanceof PrimitiveType) {
	    PrimitiveType t = (PrimitiveType)javaBasetype;
	    if (t.getName().equals("byte"))
		return new ProgramElementName("[B");
	    else if (t.getName().equals("char"))
		return new ProgramElementName("[C");
	    else if (t.getName().equals("double"))
		return new ProgramElementName("[D");
	    else if (t.getName().equals("float"))
		return new ProgramElementName("[F");
	    else if (t.getName().equals("int"))
		return new ProgramElementName("[I");
	    else if (t.getName().equals("long"))
		return new ProgramElementName("[J");
	    else if (t.getName().equals("short"))
		return new ProgramElementName("[S");
	    else if (t.getName().equals("boolean"))
		return new ProgramElementName("[Z");
	} // must be "array of array"
	return new ProgramElementName
	    ("["+((ArrayType)javaBasetype).getFullName());
    }


    /** 
     * returns the local declared supertypes
     */
    public ListOfKeYJavaType getSupertypes() {
	return SLListOfKeYJavaType.EMPTY_LIST;
    }


    /** 
     * calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnArrayDeclaration(this);
    }

    /**
     * pretty prints an array declaration
     */
    public void prettyPrint(PrettyPrinter p) throws java.io.IOException {
        p.printArrayDeclaration(this);
    }

    /**
     * toString
     */
    public String toString() {
	return name.toString().intern();
    }

}
