// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.java.declaration;


import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.java.visitor.Visitor;

/**
 *  Interface declaration.
 * 
 */

public class InterfaceDeclaration extends TypeDeclaration {

    /**
     *      Extending.
     */

    protected final Extends extending;

    /**
     *      Interface declaration.
     */

    public InterfaceDeclaration() {
	extending=null;
    }

    /** Construct a new outer or member interface class. */
    public InterfaceDeclaration(Modifier[] modifiers, ProgramElementName name,
				Extends extended, MemberDeclaration[] members){
        super(modifiers, name, members, false);
	extending = extended;
    }

    /**
     * uses children list to create non-anonymous class 
     *  @param parentIsInterfaceDeclaration boolean true iff parent is
     * an InterfaceDeclaration.  May contain: a Extends (as pointer to a
     * class), 
     * ProgramElementName (as name), several MemberDeclaration (as members of
     * the type), a parentIsInterfaceDeclaration (indicating if parent is
     * interface), several Modifier (as modifiers of the type decl), a Comment
     */
    public InterfaceDeclaration(ExtList children) { 
	super(children);
	extending=(Extends)children.get(Extends.class);
    } 


    public InterfaceDeclaration(String name) { 
	this (new de.uka.ilkd.key.java.declaration.Modifier[] {}, 
	      new ProgramElementName(name), 
	      null,  
	      new de.uka.ilkd.key.java.declaration.MemberDeclaration[]{});
    }


    /**
     *      Returns the number of children of this node.
     *      @return an int giving the number of children of this node
     */
    public int getChildCount() {
        int result = 0;
        if (modArray != null) result += modArray.size();
        if (name != null)      result++;
        if (extending != null) result++;
        if (members != null)   result += members.size();
        return result;
    }

    /**
     *      Returns the child at the specified index in this node's "virtual"
     *      child array
     *      @param index an index into this node's "virtual" child array
     *      @return the program element at the given position
     *      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *                 of bounds
     */
    public ProgramElement getChildAt(int index) {
        int len;
        if (modArray != null) {
            len = modArray.size();
            if (len > index) {
                return modArray.getModifier(index);
            }
            index -= len;
        }
        if (name != null) {
            if (index == 0) return name;
            index--;
        }
        if (extending != null) {
            if (index == 0) return extending;
            index--;
        }
        if (members != null) {
            len = members.size();
            if (len > index) {
                return members.getMemberDeclaration(index);
            }
            index -= len;
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    public int getChildPositionCode(ProgramElement child) {
        // role 0 (IDX): modifier
        // role 1: identifier
        // role 2: extends
        // role 3: implements (no occurance in interfaces)
        // role 4 (IDX): members
        if (modArray != null) {
            int index = getArrayPos(modArray,child);
            if (index >= 0) {
                return (index << 4) | 0;
            }
        }
        if (name == child) return 1;
        if (extending == child) return 2;
        if (members != null) {
            int index = getArrayPos(members,child);
            if (index >= 0) {
                return (index << 4) | 4;
            }
        }
        return -1;
    }

    /**
     *      Get extended types.
     *      @return the extends.
     */

    public Extends getExtendedTypes() {
        return extending;
    }

    /**
     *      Interfaces are always abstract.
     */
    public boolean isAbstract() {
        return true;
    }

    /**
     * Interfaces are never native.
     */
    public boolean isNative() {
        return false;
    }

    /**
     * Interfaces are never protected.
     */
    public boolean isProtected() {
        return false;
    }

    /**
     * Interfaces are never private.
     */
    public boolean isPrivate() {
        return false;
    }

    /**
     * Interfaces are never strictfp.
     */

    public boolean isStrictFp() {
        return false;
    }

    /**
     * Interfaces are never synchronized.
     */
    public boolean isSynchronized() {
        return false;
    }

    /**
     * Interfaces are never transient.
     */
    public boolean isTransient() {
        return false;
    }

    /**
     * Interfaces are never volatile.
     */
    public boolean isVolatile() {
        return false;
    }

    public boolean isInterface() {
        return true;
    }

    /** 
     * returns the local declared supertypes
     */
    public ListOfKeYJavaType getSupertypes() {
	ListOfKeYJavaType types = SLListOfKeYJavaType.EMPTY_LIST;
	if (extending != null) {
	    for (int i = extending.getTypeReferenceCount()-1; i>=0; i--) {		
		types = types.prepend
		    (extending.getTypeReferenceAt(i).getKeYJavaType());
	    }
	}
	return types;
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnInterfaceDeclaration(this);
    }

    public void prettyPrint(PrettyPrinter p) throws java.io.IOException {
        p.printInterfaceDeclaration(this);
    }
}
