// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

// This file is part of the COMPOST library, (c) 1999 Universitaet Karlsruhe.

package de.uka.ilkd.key.java.declaration;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.java.visitor.Visitor;

/**
 *  Formal parameters require a VariableSpecificationList of size() <= 1
 *  (size() == 0 for abstract methods) without initializer (for Java).
 */

public class ParameterDeclaration extends VariableDeclaration {


    /**
 *      Var spec.
     */
    protected final ArrayOfVariableSpecification varSpec;

    /**
 *      Parameter declaration.
     */

    public ParameterDeclaration() {
	this.varSpec=null;
    }

    /**
     *  Parameter declaration.
     * @param mods a modifier array.
     * @param typeRef a type reference.
     * @param name an identifier.
     * @param parentIsInterfaceDeclaration a boolean set true iff
     * the parent is an InterfaceDeclaration 
     */

    public ParameterDeclaration(Modifier[] mods,
				TypeReference typeRef,
				VariableSpecification var,
				boolean parentIsInterfaceDeclaration)
    {  	
       super(mods,typeRef,parentIsInterfaceDeclaration);
       this.varSpec=
	   new ArrayOfVariableSpecification(new VariableSpecification[]{var});
    }

    /**
     * Parameter declaration.
     * @param children an ExtList of children. May contain:	
     *  a VariableSpecification (specifying the parameter)
     *  a TypeReference (as reference to the type of the declared variable)
     * 	several Modifier (taken as modifiers of the declaration), 
     * 	a Comment
     * @param parentIsInterfaceDeclaration a boolean set true iff
     * the parent is an InterfaceDeclaration 
     */

    public ParameterDeclaration(ExtList children, 
				boolean parentIsInterfaceDeclaration) {
        super(children,parentIsInterfaceDeclaration);
	this.varSpec=new ArrayOfVariableSpecification(new
	    VariableSpecification[]{(VariableSpecification)
			    children.get(VariableSpecification.class)});
    }


    public VariableSpecification getVariableSpecification() {
        return varSpec.getVariableSpecification(0);
    }

    public ArrayOfVariableSpecification getVariables() {
        return varSpec;
    }


    /**
 *      Returns the number of children of this node.
 *      @return an int giving the number of children of this node
    */

    public int getChildCount() {
        int result = 0;
        if (modArray != null) result += modArray.size();
        if (typeReference != null) result++;
        if (varSpec  != null) result++;
        return result;
    }

    /**
 *      Returns the child at the specified index in this node's "virtual"
 *      child array
 *      @param index an index into this node's "virtual" child array
 *      @return the program element at the given position
 *      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
 *                 of bounds
    */

    public ProgramElement getChildAt(int index) {
        int len;
        if (modArray != null) {
            len = modArray.size();
            if (len > index) {
                return modArray.getModifier(index);
            }
            index -= len;
        }
        if (typeReference != null) {
            if (index == 0) return typeReference;
            index--;
        }
        if (varSpec != null) {
            if (index == 0) return varSpec.getVariableSpecification(0);
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    public int getChildPositionCode(ProgramElement child) {
        // role 0 (IDX): modifier
        // role 1: type reference
        // role 2 (IDX): var specs
        if (modArray != null) {
            int index = getArrayPos(modArray,child);
            if (index >= 0) {
                return (index << 4) | 0;
            }
        }
        if (typeReference == child) {
            return 1;
        }
        if (varSpec.getVariableSpecification(0) == child) {
            return 2;
        }
        return -1;
    }



    /**
     * Parameters are never private.
     */

    public boolean isPrivate() {
        return false;
    }

    /**
     * Parameters are never protected..
     */

    public boolean isProtected() {
        return false;
    }

    /**
     * Parameters are never "public".
     */

    public boolean isPublic() {
        return false;
    }

    /**
     * Parameters are never static.
     */

    public boolean isStatic() {
        return false;
    }

    /**
     * Parameters are never transient.
     */

    public boolean isTransient() {
        return false;
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnParameterDeclaration(this);
    }


}
