// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

// This file is part of the COMPOST library, (c) 1999 Universitaet Karlsruhe.

package de.uka.ilkd.key.java.expression;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.java.visitor.Visitor;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;


/**
 *  An ArrayInitializer is a valid expression exclusively for initializing
 *  ArrayTypes. Any other expressions are suited for any expression node.
 *  These rules could have been expressed by appropriate types, but these
 *  solutions would require a couple of new interfaces which did not seem
 *  adequate.
 *  The parent expression is either another ArrayInitializer (nested blocks)
 *  or a VariableDeclaration.
 */

public class ArrayInitializer
 extends JavaNonTerminalProgramElement
 implements Expression, ExpressionContainer {

    /**
 *      Children.
     */

    protected final ArrayOfExpression children;


    /**
     *  Array initializer.
     *  @param list with all children.
     * 	 May contain: 
     * 		several of Expression (as the initializing expression)
     * 		Comments
     */

    public ArrayInitializer(ExtList list) {
	super(list);
	this.children = 
	    new ArrayOfExpression((Expression[])list.collect(Expression.class));
    }


    /**
 *      Returns the number of children of this node.
 *      @return an int giving the number of children of this node
    */

    public int getChildCount() {
        return (children != null) ? children.size() : 0;
    }

    /**
 *      Returns the child at the specified index in this node's "virtual"
 *      child array
 *      @param index an index into this node's "virtual" child array
 *      @return the program element at the given position
 *      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
 *                 of bounds
    */

    public ProgramElement getChildAt(int index) {
        if (children != null) {
            return children.getExpression(index);
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    public int getChildPositionCode(ProgramElement child) {
        // role 0 (IDX): subexpression
        if (children != null) {
            int index = getArrayPos(children,child);
            if (index >= 0) {
                return (index << 4) | 0;
            }
        }
        return -1;
    }

    /**
 *      Get the number of expressions in this container.
 *      @return the number of expressions.
     */

    public int getExpressionCount() {
        return (children != null) ? children.size() : 0;
    }

    /*
      Return the expression at the specified index in this node's
      "virtual" expression array.
      @param index an index for an expression.
      @return the expression with the given index.
      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
      of bounds.
    */

    public Expression getExpressionAt(int index) {
        if (children != null) {
            return children.getExpression(index);
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnArrayInitializer(this);
    }

    public void prettyPrint(PrettyPrinter p) throws java.io.IOException {
        p.printArrayInitializer(this);
    }

    /**
     *      Get arguments.
     *      @return the wrapped argument array
     */
    public ArrayOfExpression getArguments() {
        return children;
    }

    public KeYJavaType getKeYJavaType(Services javaServ) {
	Expression i=this;
	int n=0;
	KeYJavaType kjt;
	Sort s=null;
	for(;i instanceof ArrayInitializer && i!=null &&
		((ArrayInitializer)i).getChildCount()!=0 ;i=((ArrayInitializer)i).getExpressionAt(0)){
	    n++;
	}
	if(!(i instanceof ArrayInitializer)) s = i.getKeYJavaType(javaServ).getSort(); 
	for(int k=0; k<n;k++){
	    s=ArraySortImpl.getArraySort(s);
	}
	kjt = javaServ.getJavaInfo().getKeYJavaType(s);
	if(kjt==null) kjt=new KeYJavaType(s);
	return kjt;
	//	throw new IllegalStateException("Not implemented: ArrayInit did "
	//				+"not know how to "
	//				+"determine its KeYJavaType");
    }


}
