package de.uka.ilkd.key.java.recoderext;

import de.uka.ilkd.key.util.Debug;
import recoder.*;
import recoder.abstraction.*;
import recoder.java.*;
import recoder.java.declaration.*;
import recoder.java.declaration.modifier.*;
import recoder.java.expression.operator.*;
import recoder.java.expression.literal.*;
import recoder.java.reference.*;
import recoder.java.statement.*;
import recoder.kit.TwoPassTransformation;
import recoder.list.ClassTypeList;
import recoder.list.FieldList;

import recoder.list.CompilationUnitMutableList;
import recoder.list.FieldSpecificationList;
import recoder.list.MemberDeclarationMutableList;
import recoder.list.ModifierMutableList;
import recoder.list.ModifierArrayList;
import recoder.list.ParameterDeclarationArrayList;
import recoder.list.StatementArrayList;
import recoder.list.StatementMutableList;
import java.util.*;

/**
 * If an allocation expression <code>new Class(...)</code> occurs, a new object
 * has to be created, in KeY this is quite similar to take it out of a list of
 * objects and setting the implicit flag <code> &lt;created&gt; </code> to
 * <code>true</code> as well as setting all fields of the object to their
 * default values. For the complete procedure, the method creates the 
 * implicit method <code>&lt;createObject$gt;</code> which on its part calls
 * another implicit method <code>lt;prepare&gt;</code> for setting the fields
 * default values.
 */
public class CreateObjectBuilder extends PrepareObjectBuilder {


    public CreateObjectBuilder
	(CrossReferenceServiceConfiguration services) {	
	super(services);
    }


    /**
     * creates the methods <code>&lt;createObject</code> and
     * <&lt;prepareParent&gt;</code> for the given type
     * @return the method declaration of <code>&lt;createObject</code>
     *  <&lt;prepareParent&gt;</code> is added implicit
     */
    public MethodDeclaration createMethod(recoder.kit.TwoPassTransformation pass,
					  TypeDeclaration type) {

 	((RecoderModelTransformer)pass).attach
	    (createPrepareParentMethod(type), type, 0);	

	ModifierMutableList modifiers = new ModifierArrayList(2);
	modifiers.add(new Private());	
	modifiers.add(new Static());	
	return new MethodDeclaration
	    (modifiers, null, new ImplicitIdentifier("<createObject>"), 
	     new ParameterDeclarationArrayList(0), null,
	     createBody(type));

    }


    /**
     * collects all implicit fields of the given type declaration
     * @param td the TypeDeclaration whose implicit fields have to be
     * collected 
     * @return the implicit fields of the given type declaration
     */
    protected FieldSpecification[] getImplicitFields(TypeDeclaration td) {
	LinkedList result = new LinkedList();
	FieldSpecificationList specs = td.getFieldsInScope();	
	for (int i = 0; i<specs.size(); i++) {
	    if (specs.getFieldSpecification(i).getName().charAt(0)=='<') {
		result.add(specs.getFieldSpecification(i));
	    }
	}
	return (FieldSpecification[])result.toArray
	(new FieldSpecification[result.size()]);
    }


    /**
     * retrieves the field specification of the field with given name
     * out of a given list
     * @param name a String with the name of the field to be looked for
     * @param specs an array of FieldSpecification where we have to look
     * for the field
     * @return the field specification of the field with given name or
     * null if not found.
     */
    protected FieldSpecification find(String name, FieldSpecification[] specs) {
	for (int i = 0; i<specs.length; i++) {
	    if (specs[i].getName().toString().equals(name)) {
		return specs[i];
	    }
	}
	return null;
    }
    

    /**
     * creates statements which take the next object out of the list of
     * available objects
     * @return statements taking the next object out of the list of
     * available objects
     */
    private LinkedList createCreateObject(LocalVariableDeclaration local, 
					  TypeDeclaration classType) {

	LinkedList result = new LinkedList();

	FieldSpecification[] implicitFields = getImplicitFields(classType);
	VariableSpecification newObject = local.
	    getVariableSpecifications().getVariableSpecification(0);
	Identifier newObjectId = newObject.getIdentifier();
	// retrieves implicit fields; optimized for readability not for speed
	// if necessary find all attributes at once
	FieldSpecification nextToCreate = find("<nextToCreate>", implicitFields);
	FieldSpecification created      = find("<created>", implicitFields);
	FieldSpecification initialized  = find("<initialized>", implicitFields);
	FieldSpecification next         = find("<next>", implicitFields);

	result.addLast(local);
	result.addLast(assign(new VariableReference(newObjectId), 
			      new FieldReference
				  (new TypeReference(new Identifier(classType.getName())), 
				   nextToCreate.getIdentifier())));
	result.addLast(assign(attribute(new VariableReference(newObjectId), created), 
			      new BooleanLiteral(true)));
	result.addLast(assign
		       (new FieldReference(new TypeReference
			   (new Identifier(classType.getName())),
					   nextToCreate.getIdentifier()), 
			attribute(new VariableReference(newObjectId), next)));
	result.addLast(assign(attribute(new VariableReference(newObjectId), initialized), 
			      new BooleanLiteral(false)));
	return result;
    }


    /**
     * Creates the <code>&lt;prepareParent&gt;</code> method. The
     * purpose of this method is to have an initial call to the
     * <code>&lt;prepare&gt;</code> method. We could just call
     * <code>prepare()</code> on the new created object itself, which
     * would lead to a split in case distinction in order to decide
     * which of the <code>prepare</code> methods have to be taken. This
     * holds only for the first call as the succeeding calls are
     * performed using <code>super.&lt;prepare()&gt;</code> and
     * therefore are statically bound. 
     */
    protected MethodDeclaration createPrepareParentMethod(TypeDeclaration classType) {
	ModifierMutableList modifiers = new ModifierArrayList(1);
	modifiers.add(new Protected());	
	StatementMutableList body = new StatementArrayList(1);
	if (!"java.lang.Object".equals(classType.getFullName())) {
	    // we can access the implementation	    	    
	    body.add(new MethodReference(new SuperReference(), 
					 new ImplicitIdentifier("<prepare>")));
	}	
	return new MethodDeclaration
	    (modifiers, null, new ImplicitIdentifier("<prepareParent>"), 
	     new ParameterDeclarationArrayList(0), null,
	     new StatementBlock(body));	
    }
    
    
    /**
     * creates the body for the implicit <code>&lt;createObject&gt;</code> method
     * @param classType the TypeDeclaration representing the type of the object to be created
     * @return the body of the <code>&lt;createObject&gt;</code> method
     */
    private StatementBlock createBody(TypeDeclaration classType) {
	LocalVariableDeclaration local = declare("newObject", classType);	
	VariableSpecification newObject = local.
	    getVariableSpecifications().getVariableSpecification(0);
	LinkedList body = createCreateObject(local, classType);

	StatementBlock prepareBlock = createPrepareBody
	    (new VariableReference(new Identifier(newObject.getName())), 
				   classType);
	StatementMutableList prepareBody = prepareBlock.getBody();

	prepareBody.remove(0);
	prepareBody.insert
	(0, new MethodReference(new VariableReference(newObject.getIdentifier()), 
				new ImplicitIdentifier("<prepareParent>")));
	body.addLast(prepareBlock);
	
	/* would be nicer, but causes unnecessary case distinctions
	   new MethodReference(null, prepareName, newObject, 
					 createPrepareMethod(classType, newObject, 
					 javaInfo)));
	*/
	body.addLast(new Return
	    (new VariableReference(new Identifier(newObject.getName()))));
	
	StatementMutableList content = new StatementArrayList
	    (((Statement[])body.toArray(new Statement[body.size()])));

	return new StatementBlock(content);
    }


    // Java construction helper methods for recoder data structures
    
    protected FieldReference attribute
	(ReferencePrefix prefix, VariableSpecification spec) {
	return new FieldReference(prefix, spec.getIdentifier());
    }

    protected CopyAssignment assign(Expression lhs, Expression rhs) {
	return new CopyAssignment(lhs, rhs);
    }

    protected LocalVariableDeclaration declare
	(String name, ClassType type) {
	return new LocalVariableDeclaration
	    (new TypeReference(new Identifier(type.getName())), 
	     new Identifier(name));
    }


}
