package de.uka.ilkd.key.java.recoderext;

import de.uka.ilkd.key.util.Debug;

import recoder.*;
import recoder.abstraction.*;
import recoder.java.*;
import recoder.java.declaration.*;
import recoder.java.declaration.modifier.*;
import recoder.java.reference.*;
import recoder.java.statement.*;

import recoder.list.CompilationUnitMutableList;
import recoder.list.MemberDeclarationMutableList;
import recoder.list.ModifierMutableList;
import recoder.list.ModifierArrayList;

/**
 * The Java DL requires some implicit fields and methods, that are
 * available in each Java class. The name of the implicit fields/methods
 * is usually enclosed between two angle brackets. To access them in a
 * uniform way, they are added as usual fields to the classes, in
 * particular this makes it possible to parse them in a natural way.
 *   The ImplicitFieldAdder is responsible to add all implicit fields to
 * the type declarations of the model. As the implicit methods and only
 * them will access these fields, this transformer has to be executed
 * before the other transformers are called.
 * @see ImplicitMethodAdder
 */
public class ImplicitFieldAdder extends RecoderModelTransformer {

    /**
     * creates a transformation that adds all implicit fields,
     * for example <code>&lt;created&gt;</code>,
     * <code>&lt;initialized&gt;</code> and
     * <code>&lt;nextToCreate&gt;</code> etc. 
     * @param services the CrossReferenceServiceConfiguration to access 
     * model information
     * @param units the array of CompilationUnits describing the model
     * to be transformed
     */
    public ImplicitFieldAdder
	(CrossReferenceServiceConfiguration services, 
	 CompilationUnitMutableList units) {	
	super(services, units);
    }

    /**
     * creates an implicit field of the given type and name
     * @param typeName the name of the type of the new field to create
     * @param fieldName the name of the field
     * @param isStatic a boolean that is true if the field has to be
     * created as static (class) field
     * @return the new created field declaration
     */
    private FieldDeclaration createImplicitRecoderField
	(String typeName, String fieldName, boolean isStatic) {
	
	ModifierMutableList modifiers = new ModifierArrayList
	    (1 + (isStatic ? 1 : 0));

	modifiers.add(new Private());
	if (isStatic) {
	    modifiers.add(new Static());
	}

	FieldDeclaration fd = new FieldDeclaration
	    (modifiers, new TypeReference(new Identifier(typeName)), 
	     new ImplicitIdentifier(fieldName), null);
	return fd;
    }


    /** 
     * adds implicit fields to the given type declaration
     * @param td the recoder.java.TypeDeclaration to be enriched with
     * implicit fields 
     */
    private void addImplicitRecoderFields
    (recoder.java.declaration.TypeDeclaration td) {

	MemberDeclarationMutableList members = td.getMembers();
	// static
	String className = td.getName(); 
	if (className == null) {
	    Debug.out("makeImplicitMembersExplicit: anonymous class will not add" +
		      "implicit fields");
	    return;
	}
	attach(createImplicitRecoderField("boolean", "<classInitialized>", true), td, 0);
	attach(createImplicitRecoderField(className, "<first>", true), td, 0);
	attach(createImplicitRecoderField(className, "<nextToCreate>", true), td, 0);
	if (!td.isInterface()) {
	    // instance
	    attach(createImplicitRecoderField("boolean", "<created>", false), td, 0);
	    attach(createImplicitRecoderField("boolean", "<initialized>", false), td, 0);
	    attach(createImplicitRecoderField(className, "<next>", false), td, 0);
	}	
	td.setMembers(members);
    }

    
    protected void makeExplicit(TypeDeclaration td) {
	addImplicitRecoderFields(td);
	int typeCount = td.getTypeDeclarationCount();
	for (int j = 0; j<typeCount; j++) {
	    makeExplicit(td.getTypeDeclarationAt(j));
	}
	td.makeAllParentRolesValid();
    }
}
