// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.java.recoderext;

import recoder.abstraction.*;
import recoder.abstraction.Package;
import recoder.java.*;
import recoder.java.declaration.*;
import recoder.java.expression.*;
import recoder.java.expression.operator.*;
import recoder.java.reference.*;
import recoder.list.*;
import recoder.*;
import recoder.service.*;

public class KeYCrossReferenceSourceInfo implements CrossReferenceSourceInfo{

    
    CrossReferenceSourceInfo cinf;

    public KeYCrossReferenceSourceInfo(CrossReferenceSourceInfo cinf) {
	this.cinf=cinf;
    }
    /**
       Strict checking. Does not allow "broken links" during reference
       resolution.
     */
    public int STRICT = 0;
    
    /**
       Sloppy checking. Allows "broken links" during reference resolution.
     */
    public int SLOPPY = 1;

    /**
       Sets the check mode behavior. Can be set to {@link #SLOPPY} thereby 
       suppressing most 
       {@link recoder.service.UnresolvedReferenceException}s.
       The settings apply from now on for all program elements that are
       analyzed by this object.
       Default is {@link #STRICT}.
       @param strategy the check mode strategy.
     */
    public void setCheckMode(int strategy){
	cinf.setCheckMode(strategy);
    }
    
    /**
       Returns the current check mode behavior.       
       @return the check mode strategy.
       @see #setCheckMode
    */
    public int getCheckMode(){
	return cinf.getCheckMode();	
    }

    /**
       Returns the list of references to a given method (or constructor).
       The references stem from all known compilation units.
       @param m a method.
       @return the possibly empty list of references to the given method.
     */
    public MemberReferenceList getReferences(Method m){
	return cinf.getReferences(m);
    }

    /**
       Returns the list of references to a given constructor. The references
       stem from all known compilation units.
       @param m a constructor.
       @return the possibly empty list of references to the given constructor.
     */
    public ConstructorReferenceList getReferences(Constructor m){
	return cinf.getReferences(m);
    }

    /**
       Returns the list of references to a given variable. The references
       stem from all known compilation units.
       @param v a variable.
       @return the possibly empty list of references to the given variable.
     */
    public VariableReferenceList getReferences(Variable v){
	return cinf.getReferences(v);
    }

    /**
       Returns the list of references to a given field. The references
       stem from all known compilation units.
       @param f a field.
       @return the possibly empty list of references to the given field.
     */
    public FieldReferenceList getReferences(Field f){
	return cinf.getReferences(f);
    }

    /**
       Returns the list of references to a given type. The references
       stem from all known compilation units.
       @param t a type.
       @return the possibly empty list of references to the given type.
     */
    public TypeReferenceList getReferences(Type t){
	return cinf.getReferences(t);
    }

    /**
       Returns the list of references to a given package. The references
       stem from all known compilation units.
       @param p a package.
       @return the possibly empty list of references to the given package.
     */
    public PackageReferenceList getReferences(Package p){
	return cinf.getReferences(p);
    }

    /**
       Returns the list of type declarations that are direct subtypes of the
       given class type.
       @param ct a class type.
       @return the list of type declarations that are direct subtypes of the
       given class type.
     */
    public TypeDeclarationList getSubtypes(ClassType ct){
	return cinf.getSubtypes(ct);
    }

    /**
       Returns the list of type declarations that are transitive subtypes
       of the given class type. The list is in topological order of the
       inheritance hierarchy. Note that intermediate types that are no
       type declarations (e.g. byte code class files) interrupt the hierarchy.
       @param ct a class type.
       @return the list of type declarations that are subtypes of the
       given class type, in topological order.
     */
    public TypeDeclarationList getAllSubtypes(ClassType ct){
	return cinf.getAllSubtypes(ct);
    }

    /**
	Returns the class type that contains the given program element.
	@param  context a program element.
	@return the type to which the given program element belongs
	(may be <CODE>null</CODE>).
    */
    public ClassType getContainingClassType(ProgramElement context){
	return cinf.getContainingClassType(context);
    }

    /**
       Returns the referred package.
       @param r a package reference.
       @return the referred package (may be <CODE>null</CODE>).
     */
    public Package getPackage(PackageReference r){
	return cinf.getPackage(r);
    }
    
    /** 
	Returns the type for the given program element.
	@param  pe the program element to compute the type for.
	@return the type for that object (may be <CODE>null</CODE>).
    */
    public Type getType(ProgramElement pe){
	return cinf.getType(pe);
    }

    /** 
	Returns the referred type.
	@param  tr a type reference.
	@return the referred type (may be <CODE>null</CODE>).
    */
    public Type getType(TypeReference tr){
	return cinf.getType(tr);
    }

    /**
       Returns the declared type.
       @param td a type declaration.
       @return the corresponding type.
     */
    public ClassType getType(TypeDeclaration td){
	return cinf.getType(td);
    }

    /** 
	Returns the type for the variable specification.
	@param  vs a variable specification.
	@return the type for that variable (may be <CODE>null</CODE>).
    */
    public Type getType(VariableSpecification vs){
	Type t= cinf.getType(vs);
	return t;
    }

    /** 
	Tries to find a type with the given name using the given
	program element as context. Useful to check for name clashes
	when introducing a new identifier.
	Neither name nor context may be <CODE>null</CODE>.
	@param  name the name for the type to be looked up{
	
    }
	may or may not be qualified.
	@param  context a program element defining the lookup context (scope).
	@return the corresponding type (may be <CODE>null</CODE>).
    */
    public Type getType(String name, ProgramElement context){
	return cinf.getType(name, context);
    }

    /**
       Returns the type of the given expression (may be <CODE>null</CODE>).
       @param expr an expression.
       @return the type of the expression, or <CODE>null</CODE> if the
       type could not be computed.
    */
    public Type getType(Expression expr){
	return cinf.getType(expr);
    }

    /**
       Checks automatic narrowing from compile-time constant ints
       to byte, char, or short. Returns false if the primitive type
       is not one of byte, char, short, or if the expression is
       not a compile-time constant int as defined in the Java language
       specification.
       @param expr the expression that might be a compile-time constant 
       fitting into the given type.
       @param to the type that the expression might fit into.
       @return <CODE>true</CODE> if the expression value is "constant" and
       would fit into a variable of the given type without loss of information,
       <CODE>false</CODE> in any other case.
     */
    public boolean isNarrowingTo(Expression expr, PrimitiveType to){
	return cinf.isNarrowingTo(expr, to);
    }
    
    /** 
	Tries to find a variable with the given name using the given 
	program element as context. The variable may be a local variable,
	a parameter or a field. Useful to check for name clashes
	when introducing a new identifier.
	@param  name the name of the variable to be looked up.
	@param  context a program element defining the lookup context.
	@return the corresponding variable (may be <CODE>null</CODE>).
    */
    public Variable getVariable(String name, ProgramElement context){
	return cinf.getVariable(name, context);
    }

    /**
       Returns the declared variable.
       @param vs a variable specification.
       @return the corresponding variable.
     */
    public Variable getVariable(VariableSpecification vs){
	return cinf.getVariable(vs);
    }

    /** 
	Returns the referred variable.
	@param  vr a variable reference.
	@return the referred variable (may be <CODE>null</CODE>).
    */
    public Variable getVariable(VariableReference vr){
	return cinf.getVariable(vr);
    }

    /** 
	Returns the referred field.
	@param  fr a field reference.
	@return the referred field (may be <CODE>null</CODE>).
    */
    public Field getField(FieldReference fr){
	return cinf.getField(fr);
    }

    /** 
	Returns the locally defined fields of the given type declaration.
	This method does not report inherited fields.
	@param  td a type declaration.
	@return a list of fields that are members of the declaration.
	@see ProgramModelInfo#getFields
	@see ProgramModelInfo#getAllFields	
    */
    public FieldList getFields(TypeDeclaration td){
	return cinf.getFields(td);
    }

    /** 
	Returns the locally defined inner types of the given type declaration.
	This method does not report inherited types.
	@param  td a type declaration.
	@return a list of inner types that are members of the declaration.
	@see ProgramModelInfo#getTypes
	@see ProgramModelInfo#getAllTypes
    */
    public ClassTypeList getTypes(TypeDeclaration td){
	return cinf.getTypes(td);
    }

    /**
       Returns the declared method.
       @param md a method declaration.
       @return the corresponding method.
     */
    public Method getMethod(MethodDeclaration md){
	return cinf.getMethod(md);
    }

    /** 
	Returns the locally defined methods of the given type declaration.
	This method does not report inherited methods.
	@param  td a type declaration.
	@return a list of methods that are members of the declaration.
	@see ProgramModelInfo#getMethods
	@see ProgramModelInfo#getAllMethods
    */
    public MethodList getMethods(TypeDeclaration td){
	return cinf.getMethods(td);
    }

    /**
       Returns the referred method, if there is an unambiguous one.
       @param mr the method reference to collate.
       @return the referred method.
       @exception AmbiguousReferenceException if there are no single most 
       specific applicable method.
       @exception UnresolvedReferenceException if there are no applicable 
       method.
       @see #getMethods(MethodReference)
     */
    public Method getMethod(MethodReference mr){
	return cinf.getMethod(mr);
    }

    /**
       Returns all methods that are applicable with respect to the given
       reference. Only the most specific methods are returned{
	
    } the program
       is not semantically correct if there is more than one such method.
       @param mr the method reference to collate.
       @return the list of applicable methods.
       @see #getMethod(MethodReference)
     */
    public MethodList getMethods(MethodReference mr){
	return cinf.getMethods(mr);
    }
    
    /**
       Returns the declared constructor.
       @param cd a constructor declaration.
       @return the corresponding constructor.
     */
    public Constructor getConstructor(ConstructorDeclaration cd){
	return cinf.getConstructor(cd);
    }

    /** 
	Returns the locally defined constructors of the given type declaration.
	This method always returns a non-empty list, at least containing
	a {@link recoder.abstraction.DefaultConstructor}.
	@param  td a type declaration.
	@return a list of constructors that are members of the declaration.
	@see ProgramModelInfo#getConstructors
    */
    public ConstructorList getConstructors(TypeDeclaration td){
	return cinf.getConstructors(td);
    }

    /**
       Returns the referred constructor, if there is an unambiguous one.
       @param cr the constructor reference to collate.
       @return the constructor referred.
       @exception AmbiguousReferenceException if there are is no most specific
       applicable constructor.
       @exception UnresolvedReferenceException if there are no applicable 
       constructor.
       @see #getConstructors(ConstructorReference)
     */
    public Constructor getConstructor(ConstructorReference cr){
	return cinf.getConstructor(cr);
    }

    /**
       Returns all constructors that are applicable with respect to the given
       reference. Only the most specific constructors are returned{
	
    } the program
       is not semantically correct if there is more than one such constructor.
       @param cr the constructor reference to collate.
       @return the list of applicable constructors.
       @see #getConstructor(ConstructorReference)
     */
    public ConstructorList getConstructors(ConstructorReference cr){
	return cinf.getConstructors(cr);
    }
    
    /**
       Creates a signature by resolving the types of the given expression list.
       An empty or <CODE>null</CODE> expression list results in an empty
       type list.
       @param args a list of expressions (may be <CODE>null</CODE>).
       @return a list of types that correspond to the respective expressions.
     */
    public TypeList makeSignature(ExpressionList args){
	return cinf.makeSignature(args);
    }

    /**
       Replaces the uncollated qualifier by its proper counterpart.
       Returns <CODE>null</CODE>, if this was not possible, the
       replaced reference otherwise. Note that this method does not
       notify the change history of the change!
       @param urq an uncollated reference.
       @return a resolved reference, or <CODE>null</CODE> if the reference
       could not be resolved.       
     */
    public Reference resolveURQ(UncollatedReferenceQualifier urq){
	return cinf.resolveURQ(urq);
    }

    /**
       Check in a program element and built up all scopes. 
       @param pe the root of a syntax tree that shall be analyzed.
       @deprecated this method will not be public in future - use ChangeHistory.attach instead
     */
    public void register(ProgramElement pe){
	cinf.register(pe);
    }

    /**
       Returns the syntactical counterpart of the given classtype.
       Returns <CODE>null</CODE>, if the given type is not a type declaration.
       @param ct a class type.
       @return the corresponding type declaration, or <CODE>null</CODE>, if 
       the given type has no syntactical representation.
     */
    public TypeDeclaration getTypeDeclaration(ClassType ct){
	return cinf.getTypeDeclaration(ct);
    }

    /**
       Returns the syntactical counterpart of the given method.
       Returns <CODE>null</CODE>, if the given method is not a method 
       declaration.
       @param m a method.
       @return the corresponding method declaration, or <CODE>null</CODE>, if
       the given method has no syntactical representation.
     */
    public MethodDeclaration getMethodDeclaration(Method m){
	return cinf.getMethodDeclaration(m);
    }

    /**
       Returns the syntactical counterpart of the given method.
       Returns <CODE>null</CODE>, if the given method is not a method 
       declaration.
       @param m a method.
       @return the corresponding method declaration, or <CODE>null</CODE>, if
       the given method has no syntactical representation.
     */
    public ConstructorDeclaration getConstructorDeclaration(Constructor c){
	return cinf.getConstructorDeclaration(c);
    }

    /**
       Returns the syntactical counterpart of the given variable.
       Returns <CODE>null</CODE>, if the given variable is not a variable 
       specification.
       @param m a variable.
       @return the corresponding variable specification, or <CODE>null</CODE>, 
       if the given variable has no syntactical representation.
     */
    public VariableSpecification getVariableSpecification(Variable v){
	return cinf.getVariableSpecification(v);
    }



 /**
       Returns the type of the given program model element. For types this
       is the model element itself, for variables it is their defined type
       and for methods it is their statically defined return type.
       @param  pme the program model element to analyze.
       @return the type of the program element or <tt>null</tt> if the type
       is unknown or unavailable.
    */
    public Type getType(ProgramModelElement pme){
	return cinf.getType(pme);
    }

    /** 
	Returns the package of the given program model element.
	@param  pme a program model element.
	@return the package of that object (may be <CODE>null</CODE>).
    */
    public Package getPackage(ProgramModelElement pme){
	return cinf.getPackage(pme);
    }

    /** 
	Returns the class types locally defined within the given container.
	Returns inner types when the container itself is a class type.
	@param ctc a class type container.
	@return a list of contained class types.
    */
    public ClassTypeList getTypes(ClassTypeContainer ctc){
	return cinf.getTypes(ctc);
    }

    /** 
	Returns all class types that are inner types of the given class types,
	including visible inherited types.
	@param ct a class type.
	@return a list of class types that are members of the given type
	or any of its supertypes.
	@see #getAllSupertypes
    */
    public ClassTypeList getAllTypes(ClassType ct){
	return cinf.getAllTypes(ct);
    }

    /** 
	Returns the class type container for the given class type.
	@param ct a class type.
	@return the container of the given type.
    */
    public ClassTypeContainer getClassTypeContainer(ClassType ct){
	return cinf.getClassTypeContainer(ct);
    }

     /** 
 	Returns the list of locally declared supertypes of the given class 
 	type.
 	@param ct a class type.
 	@return the list of locally defined supertypes of the given type.
     */
     public ClassTypeList getSupertypes(ClassType ct){
 	return cinf.getSupertypes(ct);
     }

     /** 
 	Returns the list of all supertypes of the given class type,
 	in topological order, including the class type isself as first element.
 	The order allows to resolve member overloading or overloading.
 	@param ct a class type.
 	@return the list of all supertypes of the given type in topological 
 	order.
     */
     public ClassTypeList getAllSupertypes(ClassType ct){
 	return cinf.getAllSupertypes(ct);
     }

     /** 
 	Returns the fields locally defined within the given class type.
 	@param ct a class type.
 	@return the list of field members of the given type.
     */
     public FieldList getFields(ClassType ct){
 	return cinf.getFields(ct);
     }

    /** 
	Returns all visible fields that are defined in the given class type
	or any of its supertypes. The fields are in topological order
	with respect to the inheritance hierarchy.
	@param ct a class type.
	@return the list of visible field members of the given type and its
	supertypes.
    */
    public FieldList getAllFields(ClassType ct){
	return cinf.getAllFields(ct);
    }

    /** 
	Returns the constructors locally defined within the given class type.
	@param ct a class type.
	@return the list of constructors of the given type.
    */
    public ConstructorList getConstructors(ClassType ct){
	return cinf.getConstructors(ct);
    }

    /** 
	Returns the methods locally defined within the given class type.
	@param ct a class type.
	@return the list of methods of the given type.
    */
    public MethodList getMethods(ClassType ct){
	return cinf.getMethods(ct);
    }

    /** 
	Returns all visible methods that are defined in the given class type
	or any of its supertypes. The methods are in topological order
	with respect to the inheritance hierarchy.
	@param ct a class type.
	@return the list of visible methods of the given type and its
	supertypes.
    */
    public MethodList getAllMethods(ClassType ct){
	return cinf.getAllMethods(ct);
    }

    /** 
	Returns the logical parent class of the given member.
	@param m a member.
	@return the class type containing the given member.
    */
    public ClassType getContainingClassType(Member m){
	return cinf.getContainingClassType(m);
    }

    /** 
	Returns the signature of the given method or constructor.
	@param m a method (or constructor).
	@return the signature of the given method.
    */
    public TypeList getSignature(Method m){
	return cinf.getSignature(m);
    }

    /** 
	Returns the exceptions of the given method or constructor.
	@param m a method (or constructor).
	@return the exceptions of the given method.
    */
    public ClassTypeList getExceptions(Method m){
	return cinf.getExceptions(m);
    }

    /** 
	Returns the return type of the given method.
	@param m a method.
	@return the return type of the given method.
    */
    public Type getReturnType(Method m){
	return cinf.getReturnType(m);
    }


    /**
       Returns the promoted type for binary operations between the the given 
       primitive types. The rules are defined in the Java language 
       specification{
	
    } in short, the promoted type is is the "bigger"
       type of the two.
       @param a a primitive type.
       @param b a primitive type.
       @return the promoted type for the pair of types.
       @exception TypingException if one of the types represents the boolean
       type, while the other does not.
     */
    public PrimitiveType getPromotedType(PrimitiveType a, PrimitiveType b){
	return cinf.getPromotedType(a,b);
    }

    /**
       Checks if the first given primitive type is widening to the second one.
       The rules are defined in the Java language specification{
	
    } in short, 
       a "smaller" type is widening to a "bigger" one.
       @param from a primitive type that might be widening into another type.
       @param to a primitive type that is the destination of a widening 
       operation.
       @return <CODE>true</CODE> if the first type widens into the second
       one, <CODE>false</CODE> otherwise.
     */
    public boolean isWidening(PrimitiveType from, PrimitiveType to){
	return cinf.isWidening(from, to);
    }

    /**
       Checks if the first given class type is widening to the second one.
       This is the case iff the first type is a subtype of the second.
       @param from a class type that might be widening into another type.
       @param to a class type that is the destination of a widening 
       operation.
       @return <CODE>true</CODE> if the first type widens into the second
       one, <CODE>false</CODE> otherwise.
     */
    public boolean isWidening(ClassType from, ClassType to){
	return cinf.isWidening(from, to);
    }

    /**
       Checks if the first given array type is widening to the second one.
       The rules are defined in the Java language specification.
       @param from an array type that might be widening into another type.
       @param to an array type that is the destination of a widening 
       operation.
       @return <CODE>true</CODE> if the first type widens into the second
       one, <CODE>false</CODE> otherwise.
     */
    public boolean isWidening(ArrayType from, ArrayType to){
	return cinf.isWidening(from, to);
    }

    /**
       Checks if the first given type is widening to the second one.
       This method makes no assumptions between the corrsponding types
       and handles combinations of different meta types (e.g.
       array types to class types such as "Cloneable").
       @param from a type that might be widening into another type.
       @param to a type that is the destination of a widening operation.
       @return <CODE>true</CODE> if the first type widens into the second
       one, <CODE>false</CODE> otherwise.
     */
    public boolean isWidening(Type from, Type to){
	return cinf.isWidening(from, to);
    }

    /** 
	Checks if the first given class type is a subtype of the second.
        @param a class type.
        @param b class type.
        @return <CODE>true</CODE> if the first class type is a subtype of 
	the second or if both types are equal, <CODE>false</CODE> otherwise.
    */
    public boolean isSubtype(ClassType a, ClassType b){
	return cinf.isSubtype(a,b);
    }

    /** 
	Checks if the first given class type is a supertype of the second.
        @param a class type.
        @param b class type.
        @return <CODE>true</CODE> if the first class type is a supertype of 
	the second or if both types are equal, <CODE>false</CODE> otherwise.
    */
    public boolean isSupertype(ClassType a, ClassType b){
	return cinf.isSupertype(a,b);
    }

    /**
       Checks visibility of the given member from the point of view of
       the given class type.
       @param m a member that might be referred to from within the given
       classtype.
       @param t the class type that might refer to the given member.
       @return <CODE>true</CODE> if the member is visible for the class
       type, <CODE>false</CODE> otherwise.
    */
    public boolean isVisibleFor(Member m, ClassType t){
	return cinf.isVisibleFor(m,t);
    }

    /**
       Checks if the first signature is compatible to the second one.
       A signature is compatible if all types are widing to their counterparts.
       @param a a signature.
       @param b a signature.
       @return <CODE>true</CODE> if the first signature is compatible with
       the second one, <CODE>false</CODE> otherwise.
    */
    public boolean isCompatibleSignature(TypeList a, TypeList b){
	return cinf.isCompatibleSignature(a,b);
    }

    /**
       Retains methods that are applicable for a given call.
       All other methods are removed from the list.
       @param list the list of candidate methods.
       @param name the name of the method as used by the caller.
       @param signature the types of the argument expressions of the call.
       @param context the context where the call takes place.      
     */
    public void filterApplicableMethods(MethodMutableList list,
				 String name, 
				 TypeList signature,
				 ClassType context){
	cinf.filterApplicableMethods(list, name, signature, context);
    }

    /**
       Retains methods with signatures that are not compatible to each other.
       If used upon methods that apply to a certain method reference,
       the most specific methods are returned. This handles method overloading,
       while ambiguous lists are considered semantical errors (list.size() > 1
       after return).
       @param list the list methods to compare.
       @see #filterApplicableMethods
    */
    public void filterMostSpecificMethods(MethodMutableList list){
	cinf.filterMostSpecificMethods(list);
    }

    /**
       Returns the list of most specific methods with the given name that 
       are defined in the given type or in a supertype where they are visible 
       for the given type, and have a signature that is compatible to the 
       given one.
       If used to resolve a method call, the result should be defined and
       unambiguous.
       @param ct the class type to get methods from.
       @param name the name of the methods in question.
       @param signature the statical type signature of a callee.
       @return the methods that correspond best to the given constraints.
     */
    public MethodList getMethods(ClassType ct, String name, TypeList signature){
	return cinf.getMethods(ct, name, signature);
    }

    /**
       Returns the list of most specific constructorsthat are defined in 
       the given type and have a signature that is compatible to the 
       given one.
       If used to resolve a constructor call, the result should be defined and
       unambiguous.
       @param ct the class type to get constructors from.
       @param signature the statical type signature of a callee.
       @return the constructors that correspond best to the given constraints.
     */
    public ConstructorList getConstructors(ClassType ct, TypeList signature){
	return cinf.getConstructors(ct, signature);
    }   


    /**
       Called by the service configuration indicating that all services
       are known. Services may now start communicating or linking among
       their configuration partners. The service configuration can be
       memorized if it has not been passed in by a constructor already.
       @param cfg the service configuration this services has been assigned to.
     */
    public void initialize(ServiceConfiguration cfg){
	cinf.initialize(cfg);
    }

    /**
       Returns the service configuration this service is a part of.
       @return the configuration of this service.
     */
    public ServiceConfiguration getServiceConfiguration(){
	return cinf.getServiceConfiguration();
    }



}
