// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001 Universitaet Karlsruhe, Germany
//                    and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General 
// Public License. See LICENSE.TXT for details.
//

// This file is partially taken from the RECODER library, which is protected by 
// the LGPL, and modified.


package de.uka.ilkd.key.java.recoderext;

import recoder.*;
import recoder.java.*;
import recoder.list.*;
import recoder.convenience.*;
import recoder.java.SourceElement.Position;
import recoder.java.reference.*;
import recoder.java.declaration.*;
import de.uka.ilkd.key.java.recoderext.*;
import de.uka.ilkd.key.parser.schemajava.*;
import java.io.*;
import java.util.*;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Named;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;

public class SchemaJavaProgramFactory extends JavaProgramFactory {

    protected Namespace svns;

    /**
     Protected constructor - use {@link #getInstance} instead.
     */
    protected SchemaJavaProgramFactory() {}

    /** 
     The singleton instance of the program factory.
     */
    private static SchemaJavaProgramFactory theFactory 
	= new SchemaJavaProgramFactory();

//     public static SchemaJavaProgramFactory getSchemaJavaInstance() {
//         return theFactory;
//     }

    /** 
     Returns the single instance of this class.
     */
    public static JavaProgramFactory getInstance() {
        return (JavaProgramFactory)theFactory;
    }

    public RMethodCallStatement createRMethodCallStatement(ProgramVariableSVWrapper resVar,
						       Statement st) {
	return new RMethodCallStatement(resVar, st);
    }

    public RKeYMetaConstruct createRKeYMetaConstruct() {
	return new RKeYMetaConstruct();
    }

    public RKeYMetaConstructExpression createRKeYMetaConstructExpression() {
	return new RKeYMetaConstructExpression();
    }


    public RKeYMetaConstructType createRKeYMetaConstructType() {
	return new RKeYMetaConstructType();
    }

    public ContextStatementBlock createContextStatementBlock() {
	return new ContextStatementBlock();
    }

    public static void throwSortInvalid(SchemaVariable sv, String s) 
	throws ParseException {
	throw new ParseException("Sort of declared schema variable "
				  +sv.name().toString()+" "
				  +sv.sort().name().toString()
				  +" does not comply with expected type "+s
				  +" in Java program.");
    }


    public boolean lookupSchemaVariableType(String s, ProgramSVSort sort){
	SchemaVariable sv=null;
	if (svns==null) return false;
	Named n=svns.lookup(new Name(s));
	if (n!=null && n instanceof SchemaVariable) {
	    return ((SchemaVariable) n).sort()==sort;
	} 
	return false;
    }


    public SchemaVariable lookupSchemaVariable(String s) throws ParseException{
	SchemaVariable sv=null;
	Named n=svns.lookup(new Name(s));
	if (n!=null && n instanceof SchemaVariable) {
	    sv=(SchemaVariable) n;
	} else {
	    throw new ParseException("Schema variable not declared: "+s);
	}
	return sv;
    }

    public StatementSVWrapper getStatementSV(String s) throws ParseException{
	SchemaVariable sv=lookupSchemaVariable(s);
	if (!sv.isProgramSV()) {   
	    throwSortInvalid(sv, "Statement");
	}

	return new StatementSVWrapper(sv);
    }

    public ExpressionSVWrapper getExpressionSV(String s) throws ParseException{
	SchemaVariable sv=lookupSchemaVariable(s);
	if (!sv.isProgramSV()) {
	    throwSortInvalid(sv, "Expression");
	}
	return new ExpressionSVWrapper(sv);
    }


    public LabelSVWrapper getLabelSV(String s) throws ParseException{
	SchemaVariable sv=lookupSchemaVariable(s);
	if (!sv.isProgramSV()) {   
	    throwSortInvalid(sv, "Label");
	}
	return new LabelSVWrapper(sv);
    }

    public TypeSVWrapper getTypeSV(String s) throws ParseException{
	SchemaVariable sv=lookupSchemaVariable(s);
	if (!sv.isProgramSV()) {   
	    throwSortInvalid(sv, "Type");
	}
	return new TypeSVWrapper(sv);
    }


    public ProgramVariableSVWrapper getProgramVariableSV(String s) 
	throws ParseException{
	SchemaVariable sv=lookupSchemaVariable(s);
	if (!sv.isProgramSV()) {   
	    throwSortInvalid(sv, "Program Variable");
	}
	return new ProgramVariableSVWrapper(sv);
    }

    public CatchSVWrapper getCatchSV(String s) 
	throws ParseException {
	SchemaVariable sv=lookupSchemaVariable(s);
	if (!sv.isProgramSV()) {   
	    throwSortInvalid(sv, "Catch");
	}
	return new CatchSVWrapper(sv);
    }




    /** 
     For internal reuse and synchronization.
     */
    private static SchemaJavaParser parser = new SchemaJavaParser(System.in);

    private final static Position ZERO_POSITION = new Position(0, 0);

    private static void attachComment(Comment c, ProgramElement pe) {
        ProgramElement dest = pe;
        if (!c.isPrefixed()) {
            NonTerminalProgramElement ppe = dest.getASTParent();
            int i = 0;
            if (ppe != null) {
                for (; ppe.getChildAt(i) != dest; i++);
            }
            if (i == 0) { // before syntactical parent
		c.setPrefixed(true);
            } else {
                dest = ppe.getChildAt(i - 1);
                while (dest instanceof NonTerminalProgramElement) {
                    ppe = (NonTerminalProgramElement)dest;
                    i = ppe.getChildCount();
                    if (i == 0) {
                        break;
                    }
                    dest = ppe.getChildAt(i - 1);
                }
            }
        }
        if (c instanceof SingleLineComment && c.isPrefixed()) {
            Position p = dest.getFirstElement().getRelativePosition();
            if (p.getLine() < 1) {
                p.setLine(1);
                dest.getFirstElement().setRelativePosition(p);
            }
        }
        CommentMutableList cml = dest.getComments();
        if (cml == null) {
            dest.setComments(cml = new CommentArrayList());
        }
        cml.add(c);
    }

    /**
       Perform post work on the created element. Creates parent links
       and assigns comments.
     */
    private static void postWork(ProgramElement pe) {
        CommentMutableList comments = parser.getComments();
        int commentIndex = 0;
        int commentCount = comments.size();
        Position cpos = ZERO_POSITION;
        Comment current = null;
        if (commentIndex < commentCount) {
            current = comments.getComment(commentIndex);
            cpos = current.getFirstElement().getStartPosition();
        }
        TreeWalker tw = new TreeWalker(pe);
        while (tw.next()) {
            pe = tw.getProgramElement();
            if (pe instanceof NonTerminalProgramElement) {
                ((NonTerminalProgramElement)pe).makeParentRoleValid();
            }
	    if (pe.getFirstElement()!=null) {
		Position pos = pe.getFirstElement().getStartPosition();
		while ((commentIndex < commentCount) 
		       && pos.compareTo(cpos) > 0) {
		    attachComment(current, pe);
		    commentIndex += 1;
		    if (commentIndex < commentCount) {
			current = comments.getComment(commentIndex);
			cpos = current.getFirstElement().getStartPosition();
		    }
		}
            }
        }
        if (commentIndex < commentCount) {
            while (pe.getASTParent() != null) {
                pe = pe.getASTParent();
            }
            CommentMutableList cml = pe.getComments();
            if (cml == null) {
                pe.setComments(cml = new CommentArrayList());
            }
            do {
                current = comments.getComment(commentIndex);
                current.setPrefixed(false);
                cml.add(current);
                commentIndex += 1;
            } while (commentIndex < commentCount);
        }
    }

    /**
     Parse a {@link CompilationUnit} from the given reader.
     */
    public CompilationUnit parseCompilationUnit(Reader in) throws IOException, ParserException {
        synchronized(parser) {
	    try {
		parser.initialize(in);
		CompilationUnit res = parser.CompilationUnit();
		postWork(res);
		return res;
	    } catch (ParseException e) {
		throw new ParserException();
	    }
        }
    }

    /**
     Parse a {@link TypeDeclaration} from the given reader.
     */
    public TypeDeclaration parseTypeDeclaration(Reader in) throws IOException, ParserException {
        synchronized(parser) {
	    try{
		parser.initialize(in);
		TypeDeclaration res = parser.TypeDeclaration();
		postWork(res);
		return res;
	    } catch (ParseException e) {
		throw new ParserException();
	    }

        }
    }

    /**
     Parse a {@link FieldDeclaration} from the given reader.
     */
    public FieldDeclaration parseFieldDeclaration(Reader in) throws IOException, ParserException {
        synchronized(parser) {
	    try{
		parser.initialize(in);
		FieldDeclaration res = parser.FieldDeclaration();
		postWork(res);
		return res;
	    } catch (ParseException e) {
		throw new ParserException();
	    }

        }
    }

    /**
     Parse a {@link MethodDeclaration} from the given reader.
     */
    public MethodDeclaration parseMethodDeclaration(Reader in) throws IOException, ParserException {
        synchronized(parser) {
	    try{
		parser.initialize(in);
		MethodDeclaration res = parser.MethodDeclaration();
		postWork(res);
		return res;
	    } catch (ParseException e) {
		throw new ParserException();
	    }

        }
    }

    /**
     Parse a {@link MemberDeclaration} from the given reader.
     */
    public MemberDeclaration parseMemberDeclaration(Reader in) throws IOException, ParserException {
        synchronized(parser) {
	    try{
		parser.initialize(in);
		MemberDeclaration res = parser.ClassBodyDeclaration();
		postWork(res);
		return res;
	    } catch (ParseException e) {
		throw new ParserException();
	    }

        }
    }

    /**
     Parse a {@link ParameterDeclaration} from the given reader.
     */
    public ParameterDeclaration parseParameterDeclaration(Reader in) throws IOException, ParserException {
        synchronized(parser) {
	    try{
		parser.initialize(in);
		ParameterDeclaration res = parser.FormalParameter();
		postWork(res);
		return res;
	    } catch (ParseException e) {
		throw new ParserException();
	    }

        }
    }

    /**
     Parse a {@link ConstructorDeclaration} from the given reader.
     */
    public ConstructorDeclaration parseConstructorDeclaration(Reader in) throws IOException, ParserException {
        synchronized(parser) {
	    try{
		parser.initialize(in);
		ConstructorDeclaration res = parser.ConstructorDeclaration();
		postWork(res);
		return res;
	    } catch (ParseException e) {
		throw new ParserException();
	    }

        }
    }

    /**
     Parse a {@link TypeReference} from the given reader.
     */
    public TypeReference parseTypeReference(Reader in) throws IOException, ParserException {
        synchronized(parser) {
	    try{
		parser.initialize(in);
		TypeReference res = parser.ResultType();
		postWork(res);
		return res;
	    } catch (ParseException e) {
		throw new ParserException();
	    }

        }
    }

    /**
     Parse an {@link Expression} from the given reader.
     */
    public Expression parseExpression(Reader in) throws IOException, ParserException {
        synchronized(parser) {
	    try{
		parser.initialize(in);
		Expression res = parser.Expression();
		postWork(res);
		return res;
	    } catch (ParseException e) {
		throw new ParserException();
	    }

        }
    }

    /**
     Parse some {@link Statement}s from the given reader.
     */
    public StatementMutableList parseStatements(Reader in) throws IOException, ParserException {
        synchronized(parser) {
	    try{
		parser.initialize(in);
		StatementMutableList res = parser.GeneralizedStatements();
		for (int i = 0; i < res.size(); i += 1) {
		    postWork(res.getStatement(i));
		}
		return res;
	    } catch (ParseException e) {
		throw new ParserException();
	    }

        }
    }

    /**
     * Parse a {@link StatementBlock} from the given string.
     */
    public StatementBlock parseStatementBlock(Reader in) 
	throws IOException, ParserException {
	synchronized(parser) {
	    try{
		parser.initialize(in);
		StatementBlock res = parser.StartBlock();
		postWork(res);
		return res;
	    } catch (ParseException e) {
		e.printStackTrace();
		throw new ParserException();
	    }

        }
    }

    public void setSVNamespace(Namespace ns) {
	svns=ns;
    }

}
