// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.java.reference;

import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.visitor.Visitor;
import de.uka.ilkd.key.util.ExtList;


public class ExecutionContext
 extends JavaNonTerminalProgramElement implements Reference {

    /**
     * the execution context 
     */
    protected final TypeReference executionContext;
    
    /**
     * the reference to the active object
     */
    protected final ReferencePrefix runtimeInstance;


    /**
     * creates an execution context reference
     * @param executionContext the TypeReference refering to the next enclosing
     * class 
     * @param runtimeInstance a ReferencePrefix to the object that
     * is currently active/executed
     */
    public ExecutionContext(TypeReference executionContext, 
			    ReferencePrefix runtimeInstance) {
	this.executionContext = executionContext;
	this.runtimeInstance = runtimeInstance;
    }

    /**
     * creates an execution context reference
     * @param children an ExtList with the required children of the execution
     * context
     */
    public ExecutionContext(ExtList children) {
	this.executionContext = (TypeReference) children.get(TypeReference.class);	
	children.remove(this.executionContext);
	this.runtimeInstance = (ReferencePrefix) children.get(ReferencePrefix.class);
    }



    /**
     * Returns the number of children of this node.
     * @return an int giving the number of children of this node
     */
    public int getChildCount() {
	int count = 0;
	if (runtimeInstance != null) count++;
	if (executionContext != null) count++;
	return count;
    }

    /**
     * Returns the child at the specified index in this node's "virtual"
     * child array.
     * @param index an index into this node's "virtual" child array
     * @return the program element at the given position
     * @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *    of bounds
     */
    public ProgramElement getChildAt(int index) {
	if (executionContext != null) {
	    if (index == 0) return executionContext;
	    index--;
	}
	if (runtimeInstance != null) {
	    if (index == 0) return runtimeInstance;
	    index--;
	}
	throw new ArrayIndexOutOfBoundsException();
    }

    /**
     * Returns the positional code of the given child
     * @param child the exact child to look for.
     * @return the positional code of the given child, or <CODE>-1</CODE>.
     */
    public int getChildPositionCode(ProgramElement child) {
	if (child != null) {
	    if (child == executionContext) return 0;
	}
	if (runtimeInstance != null) {
	    if (child == runtimeInstance) return (1 << 4 | 1);
	}
	return -1;
    }

    /**
     * returns the type reference to the next enclosing class
     * @return the type reference to the next enclosing class
     */
    public TypeReference getTypeReference() {
	return executionContext;
    }

    /**
     * returns the runtime instance object
     * @return the runtime instance object
     */
    public ReferencePrefix getRuntimeInstance() {
	return runtimeInstance;
    }


    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnExecutionContext(this);
    }

    public void prettyPrint(PrettyPrinter p) throws java.io.IOException {
        p.printExecutionContext(this);
    }


}
