// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

// This file is part of the COMPOST library, (c) 1999 Universitaet Karlsruhe.

package de.uka.ilkd.key.java.reference;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.expression.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.logic.op.ProgramMethod;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.visitor.Visitor;
import de.uka.ilkd.key.util.Debug;

/**
 *  Method reference.
 *  @author <TT>AutoDoc</TT>
 */
public class MethodReference extends JavaNonTerminalProgramElement
                             implements MemberReference, ReferencePrefix, 
                             ReferenceSuffix, ExpressionStatement, 
                             TypeReferenceContainer, NameReference {
 
    /**
       Access path.
    */
    protected ReferencePrefix prefix;

    /**
     *      Name.
     */
    protected final ProgramElementName name;

    /**
     *      Arguments.
     */
    protected final ArrayOfExpression arguments;
    
    /** KeY Method representant */
    protected final ProgramMethod method;


    /**
     * Constructor for the transformation of COMPOST ASTs to KeY.
     * @param children the children of this AST element as KeY classes.
     *  May contain: 
     * 	several of Expression (as initializers of the array),
     *  a ProgramElementName (as the name of the method reference), 
     * 	Comments.
     *  MUST NOT CONTAIN: the ReferencePrefix for the accessPath because
     *    Expression and ReferencePrefix might not be disjunct.
     * @param accessPath a ReferencePrefix of the array reference. 
     */
    public MethodReference(ExtList children, 
                           ReferencePrefix accessPath, ProgramMethod method) {
	super(children);
	this.prefix=accessPath;
	ProgramElementName n
	    =(ProgramElementName)children.get(ProgramElementName.class);
	if (n==null) {
	    name=(ProgramElementName)method.name();
	} else {
	    name=n;
	}
	this.arguments=new
	    ArrayOfExpression((Expression[])
			      children.collect(Expression.class));  
	this.method=method;
    }


    public MethodReference(ArrayOfExpression args, ProgramElementName n, 
			   ReferencePrefix p, ProgramMethod meth) {
	this.prefix=p;
	this.method = meth;
	if (n==null) {
	    name=(ProgramElementName)method.name();
	} else {
	    name=n;
	}
	this.arguments=args;
    }


    public SourceElement getFirstElement() {
        return (prefix == null) 
	    ? getChildAt(0).getFirstElement() : prefix.getFirstElement();
    }


    /**
     *      Get reference prefix.
     *      @return the reference prefix.
     */
    public ReferencePrefix getReferencePrefix() {
        return prefix;
    }

    public MethodReference setReferencePrefix(ReferencePrefix p) {
        return new MethodReference(arguments, name, p, null);
    }

    /**
     *      Returns the number of children of this node.
     *      @return an int giving the number of children of this node
     */

    public int getChildCount() {
        int result = 0;
        if (prefix != null) result++;
        if (name       != null) result++;
        if (arguments  != null) result += arguments.size();
        return result;
    }

    /**
     *      Returns the child at the specified index in this node's "virtual"
     *      child array
     *      @param index an index into this node's "virtual" child array
     *      @return the program element at the given position
     *      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *                 of bounds
     */
    public ProgramElement getChildAt(int index) {
        if (prefix != null) {
            if (index == 0) return prefix;
            index--;
        }
        if (name != null) {
            if (index == 0) return name;
            index--;
        }
        if (arguments != null) {
	    return arguments.getExpression(index);
        }
        throw new ArrayIndexOutOfBoundsException();
    }


    public int getChildPositionCode(ProgramElement child) {
        // role 0: prefix
        // role 1: name
        // role 3 (IDX): parameters
        if (prefix == child) {
            return 0;
        }
        if (name == child) {
            return 1;
        }
        if (arguments != null) {
            int index = getArrayPos(arguments,child);
            if (index >= 0) {
                return ((index << 4) | 2);
            }
        }
        return -1;
    }


    /**
     *      Get the number of type references in this container.
     *      @return the number of type references.
     */
    public int getTypeReferenceCount() {
        return (prefix instanceof TypeReference) ? 1 : 0;
    }

    /*
      Return the type reference at the specified index in this node's
      "virtual" type reference array.
      @param index an index for a type reference.
      @return the type reference with the given index.
      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
      of bounds.
    */
    public TypeReference getTypeReferenceAt(int index) {
        if (prefix instanceof TypeReference && index == 0) {
            return (TypeReference)prefix;
        }
        throw new ArrayIndexOutOfBoundsException();
    }


    /**
     *      Get the number of expressions in this container.
     *      @return the number of expressions.
     */
    public int getExpressionCount() {
        int result = 0;
        if (prefix instanceof Expression) result += 1;
        if (arguments != null) {
            result += arguments.size();
        }
        return result;
    }

    /*
      Return the expression at the specified index in this node's
      "virtual" expression array.
      @param index an index for an expression.
      @return the expression with the given index.
      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
      of bounds.
    */
    public Expression getExpressionAt(int index) {
        if (prefix instanceof Expression) {
            if (index == 0) {
                return (Expression)prefix;
            }
            index -= 1;
        }
        if (arguments != null) {
            return arguments.getExpression(index);
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    /**
     *      Get name.
     *      @return the string.
     */
    public final String getName() {
        return (name == null) ? null : name.toString();
    }

    /**
     *      Get identifier.
     *      @return the identifier.
     */
    public ProgramElementName getProgramElementName() {
        return name;
    }

    /**
     *      Get arguments.
     *      @return the expression array wrapper.
     */
    public ArrayOfExpression getArguments() {
        return arguments;
    }

    /**
     *      Gets index-th argument    
     *      @return the expression 
     */
    public Expression getArgumentAt(int index) {
        if (arguments != null) {
            return arguments.getExpression(index);
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    /**
     * returns the KeY representation of the method
     */
    public ProgramMethod method(JavaInfo javaInfo) {
	if (method != null) {
	    return method;
	}
	KeYJavaType referencePrefixType=null;
	if (getReferencePrefix() instanceof Expression) {
	    //%%optimize:could be cached!!!
	    referencePrefixType
		= javaInfo.getTypeConverter().getKeYJavaType
		((Expression)getReferencePrefix());
	} else if (getReferencePrefix() instanceof TypeReference) {
	    referencePrefixType 
		= ((TypeReference)getReferencePrefix()).getKeYJavaType();
	} else {
	    Debug.assertTrue(false,"Unknown prefix of method reference");	    
	}	
	return method(javaInfo, referencePrefixType);
    }

    /**
     * determines the arguments types and constructs a signature of the current
     * method
     */
    public ListOfType getMethodSignature(JavaInfo ji) {
	ListOfType signature = SLListOfType.EMPTY_LIST;
	if (arguments != null) {
	    for (int i = arguments.size()-1; i>=0; i--) {
		signature = signature.prepend(ji.getTypeConverter().
					      getKeYJavaType(getArgumentAt(i)));
	    }
	}
	return signature;
    }

    /**
     * returns the KeY representation of the method.
     * Returns null if there is no implementation
     */
    public ProgramMethod method(JavaInfo javaInfo, 
				KeYJavaType refPrefixType) {
	if (method != null) {
	    return method;
	}
	ProgramMethod m = javaInfo.getProgramMethod
	    (refPrefixType, name.toString(), getMethodSignature(javaInfo));
	return m;
    }




    /**
     * returns true if the method is unambigious
     */
    public boolean isResolved() {
	return (method != null);
    }


    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnMethodReference(this);
    }

    public void prettyPrint(PrettyPrinter p) throws java.io.IOException {
        p.printMethodReference(this);
    }


    public KeYJavaType getKeYJavaType(Services javaServ) {	
	return method(javaServ.getJavaInfo())
		.getMethodDeclaration().getTypeReference().getKeYJavaType();
    }

}
