// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.java.reference;

import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.expression.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.abstraction.Constructor;
import de.uka.ilkd.key.util.ExtList;

/**
 *  Occurs in a constructor declaration as the first statement
 *  as this(...) or super(...) reference.
 *  The Reference knows the constructor declaration it refers to.
 */

public abstract class SpecialConstructorReference
 extends JavaNonTerminalProgramElement implements ConstructorReference {

    
    /**
     *      Arguments
     */
    protected final ArrayOfExpression arguments;


    /**
     * the execution context references the class where the super(...) call
     * occurs and it may contain a reference to the 'this' object.
     * For the sake of unambiguousness this information had to be encapsulated
     * in a separate class.
     */
    protected final ExecutionContext executionContext;

    
    /**
     * Special constructor reference.
     */
    public SpecialConstructorReference() {
	this.arguments = null;
	this.executionContext = null;
    }

    /**
     * Special constructor reference.
     * @param arguments an expression mutable list.
     * @param executionContext the ExecutionContext 
     *    where this reference is placed
     */
    public SpecialConstructorReference(ExecutionContext executionContext) {
	this.arguments = null;
	this.executionContext  = executionContext;
    }

    /**
     * Special constructor reference.
     * @param arguments an expression mutable list.
     * @param executionContext the ExecutionContext where this reference is placed
     */
    public SpecialConstructorReference(Expression[] arguments,
				       ExecutionContext executionContext) {
	this.arguments = new ArrayOfExpression((Expression[])
					       copyArray(arguments));
	this.executionContext  = executionContext;
    }


    /**
     * Special constructor reference.
     * @param arguments an expression mutable list.
     * @param executionContext the ExecutionContext to the execution
     *        context where this reference is placed
     */
    public SpecialConstructorReference(ArrayOfExpression arguments,
				       ExecutionContext executionContext) {
	this.arguments = arguments;
	this.executionContext  = executionContext;
    }

    /**
     * Special constructor reference.
     * @param arguments an expression mutable list.
     */
    public SpecialConstructorReference(Expression[] arguments) {
	this(arguments, null);
    }
    
    /**
     * Constructor for the transformation of COMPOST ASTs to KeY.
     * @param children the children of this AST element as KeY classes.
     * May contain: 
     * 	several of Expression (as initializers of the array),
     *  Comments
     */ 
    public SpecialConstructorReference(ExtList children) {
	super(children);	
	this.arguments = new ArrayOfExpression
	    ((Expression[])children.collect(Expression.class)); 
	this.executionContext = (ExecutionContext)
	    children.get(ExecutionContext.class); 
    }

    /**
     * Constructor for the transformation of COMPOST ASTs to KeY.
     * @param children the children of this AST element as KeY classes.
     * May contain: 
     * 	several of Expression (as initializers of the array),
     *  Comments
     */ 
    public SpecialConstructorReference(ExtList children, 
				       ExecutionContext executionContext) {
	super(children);	
	this.arguments = new ArrayOfExpression
	    ((Expression[])children.collect(Expression.class)); 
	this.executionContext = executionContext;
    }

    /**
     *      Returns the number of children of this node.
     *      @return an int giving the number of children of this node
     */
    public int getChildCount() {
	int result = 0;
	if (executionContext != null) result++;	    
        return result + ((arguments != null) ? arguments.size() : 0);
    }

    /**
     *      Returns the child at the specified index in this node's "virtual"
     *      child array
     *      @param index an index into this node's "virtual" child array
     *      @return the program element at the given position
     *      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *                 of bounds
     */
    public ProgramElement getChildAt(int index) {
	if (executionContext != null) {
	    if (index == 0) return executionContext;
	    index --;
	}
        if (arguments != null) {
            return arguments.getExpression(index);
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    /**
     *      Get the number of expressions in this container.
     *      @return the number of expressions.
     */

    public int getExpressionCount() {
        return (arguments != null) ? arguments.size() : 0;
    }

    /*
      Return the expression at the specified index in this node's
      "virtual" expression array.
      @param index an index for an expression.
      @return the expression with the given index.
      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
      of bounds.
    */
    public Expression getExpressionAt(int index) {
        if (arguments != null) {
            return arguments.getExpression(index);
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    /**
     *      Get arguments.
     *      @return the expression mutable list.
     */

    public ArrayOfExpression getArguments() {
        return arguments;
    }


    /**
     * returns the execution context
     * @return the execution context
     */
    public ExecutionContext getExecutionContext() {
	return executionContext;
    }
    
}
