// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.java.statement;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.op.IProgramVariable;
import de.uka.ilkd.key.logic.op.ProgramMethod;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.java.abstraction.*;
import de.uka.ilkd.key.java.reference.TypeReference;
import de.uka.ilkd.key.java.reference.ReferencePrefix;
import de.uka.ilkd.key.java.reference.MethodReference;
import de.uka.ilkd.key.java.visitor.Visitor;

/**
 *  A shortcut-statement for a method body.
 */
public class MethodBodyStatement extends JavaNonTerminalProgramElement
    implements Statement, NonTerminalProgramElement {

    /**
     * Exact ClassType with the body
     */
    private final KeYJavaType bodySource;

    protected final MethodReference methRef;

    private final ReferencePrefix invocationTarget;
    
    
    /**
     *      Construct a method body shortcut
     *      @param bodySource exact class where the body is declared
     *      @param pm called ProgramMethod 
     *      @param methRef MethodReference that represents the call
     */
    public MethodBodyStatement(KeYJavaType bodySource, 
                               MethodReference methRef, 
                               ReferencePrefix invocationTarget) {

        this.bodySource  = bodySource;
        this.methRef     = methRef;
	this.invocationTarget = invocationTarget;
    }

    /**
     *      Labeled statement.
     *      @param list the ExtList with the children
     */
    public MethodBodyStatement(ExtList list) {
	//FIXME
	System.err.println("Is this constructor safe?");

        this.bodySource  = (KeYJavaType) list.get(KeYJavaType.class);	
        this.methRef     = (MethodReference)list.get(MethodReference.class);
	list.remove(this.methRef);
        this.invocationTarget = (ReferencePrefix) list.get(ReferencePrefix.class);
    }


    /**
     *      Get method body.
     *      @return the Statement
     */
    public Statement getBody(JavaInfo ji) {
        Statement s=methRef.method(ji, bodySource).getBody();
	return s;
    }

    /**
     *      Returns the number of children of this node.
     *      @return an int giving the number of children of this node
     */
    public int getChildCount() {
 	return 1;
    }

    /**
     *      Returns the child at the specified index in this node's "virtual"
     *      child array.
     *      @param index an index into this node's "virtual" child array
     *      @return the program element at the given position
     *      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *                 of bounds
     */
    public ProgramElement getChildAt(int index) {
	if (index == 0) {
	    return methRef;
 	} 
 	throw new ArrayIndexOutOfBoundsException();
    }
    
    /**
     *        Returns the index of the given child, or <CODE>-1</CODE> if there is 
     *        no such child. The child is searched for by identity:
     *        <CODE>getChildAt(getIndexOfChild(x)) == x</CODE>.
     *        @param child the exact child to look for.
     *        @return the index of the given child, or <CODE>-1</CODE>.
     */
    public int getIndexOfChild(ProgramElement child) {
 	if (child == methRef) return 0;
 	return -1;
    }

    /**
     *        Returns the positional code of the given child, or <CODE>-1</CODE> if 
     *        there is no such child. The result contains an encoding of the
     *        relative position of the child as well as the role it has been
     *        playing in this parent element. This information is required for
     *        proper undo of transformations and is to be delivered to the
     *        detached method of the ChangeHistory.
     *        @see recoder.service.ChangeHistory#detached
     *        @param child the exact child to look for.
     *        @return the positional code of the given child, or <CODE>-1</CODE>.
     */
    public int getChildPositionCode(ProgramElement child) {
	// needs to be implemented
 	return getIndexOfChild(child);
    }


    /**
     *      Get the methodReference that cause this call
     *      @return the MethodReference
     */
    public MethodReference getMethodReference() {
        return methRef;
    }
    
    
    public boolean isStatic(JavaInfo ji) {
        return methRef.method(ji).isStatic();
    }


    /**
     * The symbolic execution context of this method call frame
     * @return prefix
     */
    public ReferencePrefix getDesignatedContext() {
        return invocationTarget;
    }

    public KeYJavaType getDesignatedContextType(JavaInfo ji) {
	return ji.getTypeConverter().getKeYJavaType
	    ((Expression)getDesignatedContext());
    }

    /**
     *	The exact Class the denoted method body is from
     */
    public KeYJavaType getBodySource() {
	return bodySource;
    }


    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnMethodBodyStatement(this);
    }

    public void prettyPrint(PrettyPrinter p) throws java.io.IOException {
	p.printMethodBodyStatement(this);
    }
    
}
