// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.java.statement;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.op.IProgramVariable;
import de.uka.ilkd.key.java.reference.MethodReference;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.java.visitor.Visitor;
/**
 *  The statement inserted by KeY if a method call is executed.
 */

public class MethodCallStatement extends JavaStatement implements
    Statement, StatementContainer {

    /**
     * result
     */

    protected final IProgramVariable resultVar;

    /**
     *      Body.
     */

    protected final Statement body;

    private final MethodReference methRef;
    
    
    /**
     *      Labeled statement.
     *      @param resultVar the ProgramVariable the return value is assigned to
     *      @param body a Statement containing the method body of
     *      @param methRef the MethodReference 
     *      the called method
     */

    public MethodCallStatement(IProgramVariable resultVar, 
			       StatementBlock body, 
			       MethodReference methRef) {
        this.resultVar = resultVar;
        this.body      = body;
	this.methRef   = methRef;
    }

    /**
     *      Labeled statement.
     *      @param list the ExtList with the children
     */

    public MethodCallStatement(ExtList list, MethodReference methRef) {
        this.resultVar = (IProgramVariable) list.get(IProgramVariable.class);	
        this.body      = (StatementBlock)list.get(StatementBlock.class);
	this.methRef   = methRef;
    }

    public SourceElement getFirstElement() {
        return getChildAt(0).getFirstElement();
    }

    public SourceElement getLastElement() {
        return body.getLastElement();
    }

    /**
     *      Get the method call header.
     *      @return the MethodCallHeader
     */
    public IProgramVariable getProgramVariable() {
	return resultVar;
    }

    /**
     *      Get body.
     *      @return the Statement
     */
    public Statement getBody() {
        return body;
    }

    /**
     *      Get the methodReference that cause this call
     *      @return the MethodReference
     */
    public MethodReference getMethodReference() {
        return methRef;
    }

    /**
     *      Returns the number of children of this node.
     *      @return an int giving the number of children of this node
     */

    public int getChildCount() {
        int result = 0;
        if (resultVar != null) result++;
        if (body != null) result++;
        return result;
    }

    /**
     *      Returns the child at the specified index in this node's "virtual"
     *      child array
     *      @param index an index into this node's "virtual" child array
     *      @return the program element at the given position
     *      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *                 of bounds
     */

    public ProgramElement getChildAt(int index) {
        if (resultVar != null) {
            if (index == 0) return resultVar;
            index--;
        }
        if (body != null) {
            if (index == 0) return body;
            index--;
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    public int getChildPositionCode(ProgramElement child) {
        // role 0: resultVar
        // role 1: body
        if (resultVar == child) {
            return 0;
        }
        if (body == child) {
            return 1;
        }
        return -1;
    }

    /**
     *      Get the number of statements in this container.
     *      @return the number of statements.
     */

    public int getStatementCount() {
        return (body != null) ? 1 : 0;
    }

    /**
     *       Return the statement at the specified index in this node's
     *       "virtual" statement array.
     *       @param index an index for a statement.
     *       @return the statement with the given index.
     *       @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     *       of bounds.
     */

    public Statement getStatementAt(int index) {
        if (body != null && index == 0) {
            return body;
        }
        throw new ArrayIndexOutOfBoundsException();
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnMethodCallStatement(this);
    }

    public void prettyPrint(PrettyPrinter p) throws java.io.IOException {
	p.printMethodCallStatement(this);
    }
}
