// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.java.visitor;

import java.util.Stack;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.expression.*;
import de.uka.ilkd.key.java.expression.operator.*;
import de.uka.ilkd.key.java.expression.literal.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.declaration.modifier.*;
import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.PosInProgram;

/** Walks through a java AST in depth-left-fist-order. 
 * This walker is used to transform a program according to the given 
 * SVInstantiations.
 */
public class ContextAddVisitor extends CreatingASTVisitor {


    private Stack line = new Stack();

    private ProgramElement result=null;
    
    private int depth;

    private int width;
    
    private ProgramElement putIn;
    
    private PosInProgram prefix;

    private int depthCount;

    private int widthCount;

    /** create the  ContextAddVisitor
     * @param root the ProgramElement where to begin
     */
    public ContextAddVisitor(PosInProgram prefix,
			     int depth, int width, ProgramElement context,
			     StatementBlock putIn) {
	super(context);
	this.depth = depth;
	this.width = width;
	if (putIn.getChildCount() == 0) {
	    this.putIn = null;
	} else {//if (putIn.getChildCount() == 1) {
	    //	    this.putIn = putIn.getChildAt(0);
	    //	} else {
	    this.putIn = putIn;
	}
    }


    /** the action that is performed just before leaving the node the
     * last time 
     */
    protected void doAction(ProgramElement node) {	
	node.visit(this);	
    }
    
    /** starts the walker*/
    public void start() {	
	depthCount=0;
	widthCount=-1;
	stack.push(new ExtList());		
	walk(root());
	ExtList el=(ExtList)stack.peek();
	int i=0;
	while (!(el.get(i) instanceof ProgramElement)) {
	    i++;
	}
	result = (ProgramElement) ((ExtList)stack.peek()).get(i);	
    }

    public ProgramElement result() { 
	return result;
    }

    
    /** walks through the AST. While keeping track of the current node
     * @param node the JavaProgramElement the walker is at 
     */
    protected void walk(ProgramElement node) {	       
	line.push(node);

	if (widthCount>=0) {
	    widthCount++;
	    if (width==widthCount) {
		widthCount=-1;
		changed();
	    }	    
	} 
	if (widthCount==-1) {
	    stack.push(new ExtList());
	    if (depth==depthCount) {
		if (putIn!=null) {
		    addChildCarefully(putIn);
		}  else {
		    if (((ExtList)stack.peek()).size()==0) {
			stack.pop();
		    } else {
			Debug.out("Unexpected behavior at ContextAddVisitor"+
				  "::walk");
		    }
		}
		depthCount++;
		widthCount++;
		changed();
	    } else {		
		if (!(node instanceof LabeledStatement) &&
		    !(node instanceof MethodCallStatement &&
		      ((MethodCallStatement)node).getProgramVariable() != null)) {
		    depthCount++; 
		} 
		super.walk(node);
	    }
	}
	line.pop();
    }

    public String toString() {
	return stack.peek().toString();
    }

    /** the implemented default action is called if a program element is,
     * and if it has children all its children too are left unchanged
     */
    protected void doDefaultAction(SourceElement x) {
	addChild(x);
    }


    protected void changed() {
	ExtList list = (ExtList)stack.peek();
	if (list.isEmpty() || list.getFirst()!=CHANGED) {
	    list.addFirst(CHANGED);
	}
    }

   public void performActionOnStatementBlock(StatementBlock x) {
	DefaultAction def=new DefaultAction() {		
		ProgramElement createNewElement(ExtList changeList) {
		    return new StatementBlock(changeList);
		}
	    };
	def.doAction(x);
    }

    public void performActionOnWhile(While x)     {
	ExtList changeList = (ExtList)stack.peek();
	if (changeList.getFirst() == CHANGED) {	    
	    changeList.removeFirst();	    	    
	    Expression guard = (Expression) changeList.removeFirst();
	    Statement body = (Statement) (changeList.isEmpty() ? 
					  null :
					  changeList.removeFirst());
	    addChild(new While(guard, body));	  

	    changed();
	} else {
	    doDefaultAction(x);
	}
    }



    public void performActionOnCase(Case x)     {
	Expression e = null;
	ExtList changeList = (ExtList)stack.peek();
	if (changeList.getFirst() == CHANGED) {	    
	    changeList.removeFirst();	    
	    if (x.getExpression() != null) {
		e = (Expression) changeList.removeFirst();
	    }
	    addChild(new Case(changeList, e));
	    changed();
	} else {
	    doDefaultAction(x);
	}    
    }




    public void performActionOnLabeledStatement(LabeledStatement x) {
	Label l = x.getLabel();
	ExtList changeList = (ExtList)stack.peek();
	if (changeList.getFirst() == CHANGED) {	    
	    changeList.removeFirst();	    
	    if (x.getLabel() != null) {
		if (changeList.size()==0) {
		    changed();
		    return;
		}
		if (changeList.getFirst() instanceof Label) {
		    l = (Label) changeList.removeFirst();
		}
	    }
	    addChild(new LabeledStatement(changeList, l));
	    changed();
	} else {
	    doDefaultAction(x);
	}    
    } 

    public void performActionOnMethodCallStatement(MethodCallStatement x) {
	ExtList changeList = (ExtList)stack.peek();
	if (!changeList.isEmpty() && changeList.getFirst() == CHANGED) {	  
	    changeList.removeFirst();	
	    addChild(new MethodCallStatement(changeList, x.getMethodReference()));
	    changed();
	} else {
	    doDefaultAction(x);
	}	    
    } 


    /**
     * adds a child considering if block parentheses are needed
     * @param x the child to add
     */
    private void addChildCarefully(SourceElement x) {
	if (x instanceof StatementBlock) {
	    StatementBlock sb = (StatementBlock) x;
	    stack.pop();	    
	    ExtList list = (ExtList) stack.pop();
	    ProgramElement parent;
	    if (line.size()>1) {	
		parent = (ProgramElement)line.elementAt(line.size()-2);
	    } else {
		parent=null;
	    }
	    if (!(parent instanceof StatementBlock) || (parent == null)) {
		if (sb.getChildCount() == 1 
		    && !(parent instanceof Try || parent instanceof MethodCallStatement)) {
		    for (int i=0; i<sb.getChildCount(); i++) {
			list.add(sb.getChildAt(i));
		    }		    
		} else {
		    list.add(sb);
		}
	    } else {	    
		for (int i=0; i<sb.getChildCount(); i++) {
		    list.add(sb.getChildAt(i));
		}
	    }
	    stack.push(list);
	}
    }
}
