// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.java.visitor;

import java.util.Stack;
import java.util.Iterator;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.NonTerminalProgramElement;
import de.uka.ilkd.key.logic.PosInProgram;

public class JavaIterator {
    

    private ProgramElement root;
    private PosTracker rootPos;

    /** the top element of the stack is the position tracker with the 
     * node where the iterator stands and the element returned by
     * method element()
     */
    private Stack posStack;

    public JavaIterator(ProgramElement root) {
	this.root = root;
	rootPos   = new PosTracker(root, -1);
	posStack = new Stack();
	posStack.push(rootPos);
    }

    /** increments postracker that is at the top of the stack */
    private void incPos() {
	PosTracker pos = (PosTracker) posStack.peek();
	if (pos != null) {
	    pos.next();
	}
    }

    /** decrements postracker that is at the top of the stack */
    private void decPos() {
	PosTracker pos = (PosTracker) posStack.peek();
	if (pos != null) {
	    pos.prev();
	}
    }

    /** return the current child position of the position tracker */ 
    private int childPos() {
	PosTracker pos = (PosTracker) posStack.peek();
	return pos.pos();
    }

    /** */
    private int getChildCount(ProgramElement pe) {
	if (pe instanceof NonTerminalProgramElement) {
	    return ((NonTerminalProgramElement)pe).getChildCount();
	}
	return 0;
    }

    /** returns the enclosing parent of the current element 
     * @return   the enclosing parent of the current element 
    */
    public ProgramElement parent() {
	ProgramElement result = null;
	ProgramElement current = element();
	PosTracker pos;
	if (current != null && current != root) {
	    pos = (PosTracker)posStack.pop();
	    result = ((PosTracker)posStack.peek()).node();
	    posStack.push(pos);
	}
	return result;
    }

    /** skips the subtree starting at this element and goes to the
     * next sibling or if not available to the next possible sibling
     * of one of the ancestors of this node     
     */
    public ProgramElement skip() {
	if (hasNextSibling()) {
	    // there is a next sibling
	    return nextSibling();
	} else {
	    if (element() != root) {
		posStack.pop();
	    } else if (hasNext()) {
		    ((PosTracker)posStack.peek()).setLast();
	    }	    
	    return next();	
	}	           
    }
    
    /** moves the iterator to the next node (depth-left-first)  */
    public ProgramElement next() {	
	if (hasNext()) {
	    ProgramElement current = element();
	    incPos();
	    if (childPos() < getChildCount(current)) {
		posStack.push
		    (new PosTracker(((NonTerminalProgramElement)current).getChildAt(childPos()), -1));
	    } else if (current != root) { // do not remove root from
					  // stack 
		posStack.pop();
		next();
	    }
	}
	return element();
    }

    /** moves the iterator to the next sibling of the node. Iterator
     * stays at the current node if no next sibling exists
     */
    public ProgramElement nextSibling() {	
	if (hasNextSibling()) {
	    posStack.pop();
	    return next();
	}       	
	return element();
    }

    /** returns true if next sibling exists */
    public boolean hasNextSibling() {
	if (element() != root) {
	    PosTracker pos = (PosTracker) posStack.pop();
	    if (childPos() < getChildCount(element())-1) {
		posStack.push(pos);		
		return true;
	    }
	    posStack.push(pos);		
	}
	return false;
    }

    /** moves the iterator to the previous node (depth-left-first)  */
    public ProgramElement back() {	
	ProgramElement current = element();
	if (current == null) { // the end had been reached
	    decPos();
	    goToRightLeave();
	    return element();
	} else if (!atTop()) {
	    decPos();
	    if (childPos() < -1) {
		posStack.pop();
		back();
	    } else if (childPos() > -1) {
		// go to the most right leave
		posStack.push
		    (new PosTracker(((NonTerminalProgramElement)current).
				    getChildAt(childPos()), -1));
		goToRightLeave();
	    }
	}
	return element();
    }

    /** goes to the rightest leave in the current subtree */
    private void goToRightLeave() {	
	ProgramElement current = element();
	while (childPos() < getChildCount(current)-1) {
	    incPos();
	}	
	if (childPos() != -1) {
	    ProgramElement nextChild = ((NonTerminalProgramElement)current).
		getChildAt(childPos()); 
	    posStack.push
		(new PosTracker(nextChild, -1));
	    if (getChildCount(nextChild) > 0) {
		goToRightLeave();
	    }
	}
    }
    
    /** returns the position in program of the current element. It is
     * null if the end of the program has been reached. */
    public PosInProgram position() {
	PosInProgram pos = PosInProgram.TOP;
	if (hasNext()) {
	    Iterator it = posStack.iterator();	
	    while (it.hasNext()) {
		PosTracker pT = (PosTracker) it.next();
		if (pT.pos() >= 0) {
		    pos = pos.down(pT.pos());
		}
	    }
	    return pos;
	} else {
	    return null;
	}
    }

    /** returns the current node where the iterator stand or null if
     * the end of the programelement to iterate through has been
     * reached 
     */
    public ProgramElement element() {
	if (hasNext()) {
	    return ((PosTracker)posStack.peek()).node();
	} else {
	    return null;
	}
    }

    /** returns true if there is a non visited element */
    public boolean hasNext() {
	return rootPos.pos() < getChildCount(root);
    }

    /** returns true if the iterator is back at root */
    public boolean atTop() {
	return rootPos.pos() == -1;
    }

    


    private static class PosTracker {
	private ProgramElement pe;
	private int currentChildPos;

	
	/** the program element and the number of the child that is
	 *  visited 
	 */
	public PosTracker(ProgramElement pe, 
			  int currentChildPos) {
	    this.pe = pe;
	    this.currentChildPos = currentChildPos;
	}

	ProgramElement node() {
	    return pe;
	}

	int pos() {
	    return currentChildPos;
	}

	void setLast() {
	    if (pe instanceof NonTerminalProgramElement) {
		currentChildPos = ((NonTerminalProgramElement)pe).getChildCount();
	    } else {
		currentChildPos = -1;
	    }
	}
	
	void next() {
	    currentChildPos++;
	}

	void prev() {
	    currentChildPos--;
	}
	
	public String toString() {
	    return "PE:"+pe+"|| Pos:"+currentChildPos;
	}

    }


}
