package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.java.visitor.ProgramVariableCollector;
import de.uka.ilkd.key.util.Debug;

import java.util.Stack;
import java.util.Vector;
import java.util.HashSet;
import java.util.Iterator;

/**
 * This class describes the deletion strategy used by the simultaneous update
 * simplifier. The deletion strategy removes updates (or parts of a simultaneouseous
 * one) if e.g. there exists an update like {x:=x} or the location in the update
 * is not refered to during the rest of the term behind the update.
 */
public class DefaultDeletionStrategy extends Visitor 
    implements DeletionStrategy {

    private final static TermFactory tf = TermFactory.DEFAULT;

    private Stack termStack = new Stack();

    private boolean del_Unused  = true;
    private boolean del_Simple  = true;
            
    
    public DefaultDeletionStrategy() {
	termStack.push(new StackElement());		
    }

    public DefaultDeletionStrategy(boolean del_Simple, 
				   boolean del_Unused) {

	this.del_Simple  = del_Simple;
	this.del_Unused  = del_Unused;

	termStack.push(new StackElement());		
    }

    public DeletionStrategy copy() {
	return new DefaultDeletionStrategy(del_Simple, 
					   del_Unused);
    }

    public Object clone() {
	return copy();
    }


    /** 
     * returns true if the subformula contains a CTRL operator, necessary for
     * specification computing 
     */
    protected boolean deletionForbidden() {
	return containsCtrl();
    }

    protected boolean isUnusedUpdate(Term loc) {	
	return !usesVar(loc.op());
    }

    protected boolean isSimpleUpdate(Term loc, Term ass) {
	return loc.equals(ass);
    }

    protected boolean 
	isSimpleUpdateDeletionSafe(HashSet accessed, Term location) {
	if (location.op() == Op.ATTR) {
	    return !accessed.contains(location.sub(1).op());
	} else if (location.op() == Op.ARRAY) {
	    Sort locSort = location.sort();
	    Iterator it = accessed.iterator();
	    while (it.hasNext()) {
		Object key = it.next();
		if (key instanceof Sort &&
		    locSort.extendsTrans((Sort)key)) {
		    return false;
		}
	    }
	}
	return true;
    }

    protected boolean delete() {
	StackElement se  = peek();
	boolean deleted  = false;
	HashSet accessed = new HashSet();
	int max = se.size() - 1;
	for (int i = 0; i < max ; i+=2) {
	    if (!deletionForbidden()) {		
		Term location = se.getSubTerm(i);
		Term value    = se.getSubTerm(i+1);
		if ((del_Simple && isSimpleUpdate(location, value) &&
		     isSimpleUpdateDeletionSafe(accessed, location)) || 
		    (del_Unused && isUnusedUpdate(location))) {		    
		    se.removeSubTerm(i+1);
		    se.removeSubTerm(i);
		    deleted = true;
		    // keep track of removed elements
		    i   -= 2;
		    max -= 2;
		} else if (location.op() == Op.ATTR) {
		    accessed.add(location.sub(1).op());
		} else if (location.op() == Op.ARRAY) {
		    accessed.add(location.sort());
		}			    
	    }
	}
	return deleted;	
    }

    protected void collectUsedVar(Term visited) {
	if (visited.op() instanceof ProgramVariable && 
	    (parent() == null ||
	    !(parent().op() instanceof SimultaneousUpdateOp) ||	     
	     relativePos() % 2 != 0 || 
	     relativePos() == parent().arity()-1))  {	    
	    	    
	    addVar((ProgramVariable)visited.op());	    

	} else if (visited.javaBlock() != JavaBlock.EMPTY_JAVABLOCK) {
	    ProgramVariableCollector pvc = 
		new ProgramVariableCollector(visited.javaBlock().program());
	    pvc.start();
	    addVar(pvc.result());

	}
    }


    // VISITOR RELATED METHODS

    public void subtreeEntered(Term t) {
	// put subterm list on stack
	termStack.push(new StackElement(peek(), t));
	// collect vars in t
	collectUsedVar(t);
    }


    public void visit(Term visited) {
	Term result = visited;	

	if (visited.op() instanceof SimultaneousUpdateOp) {
	    if (delete()) {
		markChanged();
	    }
	} 

	StackElement se = pop();
	if (se.changed()) { //rebuild
	    Term[] subs = new Term[se.size()];

	    ArrayOfQuantifiableVariable[] vars = new 
		ArrayOfQuantifiableVariable[subs.length];

	    if (visited.op() instanceof SimultaneousUpdateOp) {
		for (int i = 0; i < subs.length; i++) {
		    subs[i] = se.getSubTerm(i);
		}		
		result = (subs.length == 1 ? subs[0] : 
			  tf.createUpdateTerm(subs));	    
	    } else {
		// default behaviour, except SimultaneousUpdateOp
		for (int i = 0; i < subs.length; i++) {
		    subs[i] = se.getSubTerm(i);
		    vars[i] = visited.varsBoundHere(i);
		}
	    		
		result = tf.createTerm(visited.op(), subs, vars, 
				       visited.javaBlock(), visited.ruleVar());
	    }
	    markChanged();	    
	}
	pushOnStack(result);
    }    

    // ENTRANCE METHOD
    public Term delete(Term t) {
	t.execPostOrder(this);
	return peek().getSubTerm(0);
    }

    // TERMSTACK OPERATIONS
     
    private StackElement pop() {
	return (StackElement) termStack.pop();
    }

    private StackElement peek() {
	return (StackElement) termStack.peek();
    }

    private void pushOnStack(Term t) {
	peek().addSubTerm(t);
    }

    private void markChanged() {
	peek().changed(true);
    }

    public boolean usesVar(Operator var) {
	// at the moment no attributes are deleted, a partial decision is
	// possible, but later when needed 
	// arrays are also not deleted, but here once they are implemented we can be stronger 
	return !(var instanceof ProgramVariable) ? true : peek().usesVar(var);
    }

    public boolean containsCtrl() {
	return peek().containsCtrl(); 
    }

    public void addVar(ProgramVariable var) {
	peek().addVar(var);
    }

    public void addVar(HashSet vars) {
	peek().addVar(vars);
    }

    public Term parent() {
	return peek().parent();
    }

    public Term attached() {
	return peek().attached();
    }

    public int relativePos() {
	StackElement se = peek().parentSE();
	return  se != null ? se.size() : -1;
    }

    // inner container class
    
    static class StackElement {	

	private StackElement parentSE;
	private Term attached;

	private Vector subterms = new Vector(10);
	private HashSet usedVar = new HashSet(10);
	private boolean changed = false;
	private boolean containsCtrl = false;
	

	public StackElement() {
	}

	public StackElement(StackElement parentSE, Term attached) {
	    this.parentSE = parentSE;
	    this.attached = attached;	    

	    if (attached.op() == Op.CTRL) {
		propagateContainsCtrl();
	    }
	}

	public StackElement parentSE() {
	    return parentSE;
	}

	public Term parent() {
	    return parentSE != null ? parentSE.attached() : null;
	}

	public Term attached() {
	    return attached;
	}
	
	public Term getSubTerm(int idx) {
	    return (Term)subterms.get(idx);
	}

	public void addSubTerm(Term term) {
	    subterms.add(term);
	}

	public void removeSubTerm(int idx) {
	    subterms.remove(idx);
	}

	public int size() {
	    return subterms.size();
	}

	public boolean changed() {
	    return changed;
	}

	public void changed(boolean b) {
	    changed = b;
	}

	public boolean containsCtrl() {
	    return containsCtrl;
	}

	public boolean usesVar(Operator var) {
	    return usedVar.contains(var);
	}

	public void addVar(HashSet vars) {
	    Iterator it = vars.iterator();
	    while (it.hasNext()) {
		addVar((ProgramVariable) it.next());
	    }
	}
	
	private void propagateContainsCtrl() {
	    containsCtrl = true;
	    if ( parentSE != null ) {
		parentSE.propagateContainsCtrl();
	    }
	}
	
	private void addVarHelp(ProgramVariable var) {
	    if (attached != null && 
		attached.op() instanceof SimultaneousUpdateOp &&
		attached.arity() - 1 == size()) {
		usedVar.add(var);
	    }	    
	}

	public void addVar(ProgramVariable var) {	    
	    if (parentSE != null) {
		parentSE.addVarHelp(var);
		parentSE.addVar(var);		
	    }
	}

    }
    

}
