// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.java.NameAbstractionTable;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.SyntacticalReplaceVisitor;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Iterator;


/** this class implements a persistent constraint. The constraint
 * contains pairs of Metavariables X and Terms t meaning X=t. It
 * offers services like joining two Constraint objects, adding new
 * constraints to this one by unfying two terms and creating all
 * necessary Metavariable - Term pairs. There are no public
 * constructors to build up a new Constraint use the BOTTOM constraint
 * of the Constraint interface (final static class variable) and add
 * the needed constraints. If a constraint would not be satisfiable
 * (cycles, unification failed) the Constraint TOP of interface
 * Constraint is returned.
 */

public class EqualityConstraint implements Constraint {
    
    /** contains a boolean value */
    private final static BooleanContainer
	CONSTRAINTBOOLEANCONTAINER=new BooleanContainer();

    /** stores constraint content as a mapping from Metavariable to
     * Term */
    private HashMap map;

    /** cache for return values of getInstantiation */
    private HashMap instantiationCache        = null;

    /** cache elements that needs to be updated, as the constraint has
     * become stronger */
    private HashMap instantiationCacheInvalid = null;

    /** Don't use this constructor, use Constraint.BOTTOM instead */
    public EqualityConstraint() {
	this ( new HashMap () );
    }

    private EqualityConstraint( HashMap map ) {
	this.map = map;
    }

    private synchronized Object writeReplace() 
	throws java.io.ObjectStreamException {
	return new EqualityConstraint ( map );
    }
    
    private Object readResolve() 
	throws java.io.ObjectStreamException {
	if ( isBottom () )
	    return Constraint.BOTTOM;
	return this;
    }

    protected synchronized Object clone () {
	EqualityConstraint res        =
	    new EqualityConstraint ( (HashMap)map.clone () );
	res.instantiationCache        =
	    instantiationCache == null ?
	    null :
	    (HashMap)instantiationCache       .clone ();
	res.instantiationCacheInvalid =
	    instantiationCacheInvalid == null ?
	    null :
	    (HashMap)instantiationCacheInvalid.clone ();
	return res;
    }
    
    /** returns true if Bottom
     * @return true if Bottom 
     */
    public boolean isBottom() {
	return map.isEmpty ();
    }

    /** a constraint being instance of this class is satisfiable. If a
     * method realizes that an unsatisfiable Constraint would be built
     * because of failed unification, cycle or s.th. similar it
     * returns the singleton TOP being instance of the subclass Top
     * @return true always 
     */
    public boolean isSatisfiable() {
	return true;
    }


    /**
     * @return list of metavariables, instantiations of which may
     * restricted by this constraint
     */
    public IteratorOfMetavariable restrictedMetavariables () {
	return new MVIterator ( map.keySet ().iterator () );
    }

    private static class MVIterator implements IteratorOfMetavariable {
	private Iterator it;

	public MVIterator ( Iterator p_it ) {
	    it = p_it;
	}

	public boolean hasNext () {
	    return it.hasNext ();
	}

	public Metavariable next () {
	    return (Metavariable)it.next ();
	}
    }


    /**
     * @return The most general known term that is more defining than
     * p_mv itself by which p_mv can be replaced if the constraint is
     * valid (or null if the constraint allows arbitrary
     * instantiations of p_mv). This is just the entry of map.
     */
    public Term getDirectInstantiation ( Metavariable p_mv ) {
	return (Term)map.get ( p_mv );
    }


    /**
     * @return the term by which p_mv is instantiated by the most
     * general substitution satisfying the constraint
     */
    public synchronized Term getInstantiation ( Metavariable p_mv ) {
	Term t = null;
	if ( instantiationCache == null )
	    instantiationCache = new HashMap ();
	else
	    t = (Term)instantiationCache.get ( p_mv );

	if ( t == null ) {
	    if ( instantiationCacheInvalid != null )
		t = (Term)instantiationCacheInvalid.remove ( p_mv );

	    if ( t != null && t.op () == p_mv )
		t = null;

	    if ( t == null ) {
		t = (Term)map.get ( p_mv );
		if ( t == null )
		    t = TermFactory.DEFAULT.createFunctionTerm ( p_mv );
		else	
		    t = instantiate ( t );
	    }

	    instantiationCache.put ( p_mv, t );
	}

	return t;
    }
    

    private Term instantiate ( Term p ) {
	// Empty JavaInfo can be used, because metavariable
	// instantiations must not contain elements for which
	// JavaInfo is needed
	SyntacticalReplaceVisitor srVisitor =
	    new SyntacticalReplaceVisitor ( new JavaInfo (), this );
	p.execPostOrder ( srVisitor );
	return srVisitor.getTerm ();
    }


    /** executes unification for terms t1 and t2
     * @param t1 Term to be unified 
     * @param t2 term to be unified
     * @return TOP if not possible, else a new constraint with after
     * unification of t1 and t2
     */
    public Constraint unify ( Term t1, Term t2 ) {
	return unify(t1, t2, CONSTRAINTBOOLEANCONTAINER);
    }

        
    /** executes unification for terms t1 and t2. 
     * @param t1 Term to be unfied 
     * @param t1 Term to be unfied
     * @param unchanged true iff the new constraint equals this one
     * @return TOP if not possible, else a new constraint unifying t1
     * and t2 ( == this iff this subsumes the unification )
     */
    public Constraint unify ( Term             t1,
			      Term             t2,
			      BooleanContainer unchanged ) {
	EqualityConstraint newConstraint;

	try {
	    newConstraint = unifyHelp ( t1, t2, null );
	} catch ( UnsatisfiableException e ) {
	    unchanged.setVal ( false );
	    return Constraint.TOP;
	}

	if ( newConstraint == null ) {
	    unchanged.setVal ( true );
	    return this;
	}

	unchanged.setVal ( false );
	newConstraint.instantiationCacheInvalid =
	    buildInvalidCache ();
	return newConstraint;
    }

    /**
     * Make a copy of the instantiation cache, moving all previously
     * valid cache elements to the invalid section
     */
    private synchronized HashMap buildInvalidCache () {
	HashMap res = null;
	if ( instantiationCache != null ) {
	    res = (HashMap)instantiationCache.clone ();
	    if ( instantiationCacheInvalid != null )
		res.putAll ( instantiationCacheInvalid );
	} else if ( instantiationCacheInvalid != null )
	    res = (HashMap)instantiationCacheInvalid.clone ();	
	return res;
    }

    /**
     * compare two quantifiable variables if they are equal modulo renaming
     * @param ownVar first QuantifiableVariable to be compared
     * @param cmpVar second QuantifiableVariable to be compared
     * @param ownVarEnum the enumeration of ownVar to abstract from
     * variable objects 
     * @param cmpVarEnum the enumeration of cmpVar to abstract from
     * variable objects 
     */
    private boolean compareBoundVariables
	(QuantifiableVariable ownVar,
	 QuantifiableVariable cmpVar,
	 MapFromQuantifiableVariableToInteger ownVarEnum, 
	 MapFromQuantifiableVariableToInteger cmpVarEnum) {	
	
	Integer ownVarNo=ownVarEnum.get(ownVar);
	Integer cmpVarNo=cmpVarEnum.get(cmpVar); 
	if (cmpVarNo == null && ownVarNo == null) {
	    // if both are null the variables have to be the
	    // same object
	    if (!(ownVar == cmpVar)) {
		return false; 
	    } else {
		return true;
	    }
	} else if (cmpVarNo == null || ownVarNo == null) {
	    // exact one is null
	    return false;
	}
	if (!cmpVarNo.equals(ownVarNo)) {
	    // both are non-null so the enumeration has to be
	    // equal 
	    return false;
	}	
	return true;
    }


    /**
     * compares two terms modulo bound renaming, changing
     * this as needed to make the two terms equal
     * @param t0 the first term
     * @param t1 the second term
     * @param ownVarEnum the MapFromQuantifiableVariableToInteger containing
     * the enumeration of the own bound vars in order to abstract from
     * the variables
     * @param cmpVarEnum the MapFromQuantifiableVariableToInteger containing
     * the enumeration of the bound vars of the Term to be compared
     * with, in order to abstract from the variables  
     * @param changed this has been changed during the unification
     * @return a constraint that subsumes newConstraint under which
     * t0, t1 are equal modulo bound renaming. newConstraint (the same
     * object) returned iff the terms are equal modulo bound renaming
     * under newConstraint
     * @throws UnsatisfiableException iff unification is not possible
     */
    private EqualityConstraint unifyHelp
	( Term                                 t0,
	  Term                                 t1,
	  MapFromQuantifiableVariableToInteger ownVarEnum, 
	  MapFromQuantifiableVariableToInteger cmpVarEnum,
	  NameAbstractionTable                 nat,
	  EqualityConstraint                   newConstraint )
	throws UnsatisfiableException {

	if ( t0 == t1 && ownVarEnum.equals ( cmpVarEnum ) )
	    return newConstraint;
	if ( t0.op() instanceof QuantifiableVariable ) {
	    if ( !( (t1.op() instanceof QuantifiableVariable) &&
		    compareBoundVariables ( (QuantifiableVariable) t0.op(),
					    (QuantifiableVariable) t1.op(),
					    ownVarEnum,
					    cmpVarEnum) ) )
		throw UNSATISFIABLE_EXCEPTION;
	    return newConstraint;
	} else {
	    if ( t1.op () instanceof Metavariable ) {
		if ( t0.op () == t1.op () )
		    return newConstraint;
		if ( t0.op () instanceof Metavariable &&
		     t1.sort ().extendsTrans ( t0.sort () ) ) {
		    if ( t0.sort ().extendsTrans ( t1.sort () ) ) {
			// sorts are equal
			// use Metavariable-order
			// to choose the left MV
			if ( ((Metavariable)t0.op ())
			     .compareTo ( t1.op () ) >= 0 )
			    return normalize
				( (Metavariable)t0.op (), t1, newConstraint );
			else
			    return normalize
				( (Metavariable)t1.op (), t0, newConstraint );
		    } else
		        return normalize
			    ( (Metavariable)t0.op (), t1, newConstraint );
		} else if ( t0.sort ().extendsTrans ( t1.sort () ) )
		    return normalize
			( (Metavariable)t1.op (), t0, newConstraint );
		else
		    throw UNSATISFIABLE_EXCEPTION;
	    } else if ( t0.op () instanceof Metavariable ) {
		if ( t1.sort ().extendsTrans ( t0.sort () ) )
		    return normalize
			( (Metavariable)t0.op (), t1, newConstraint );
		else
		    throw UNSATISFIABLE_EXCEPTION;
	    }

	    if ( !( t0.sort().equals(t1.sort()) && t0.arity() == t1.arity() ) )
		throw UNSATISFIABLE_EXCEPTION;

	    if ( t0.javaBlock () != JavaBlock.EMPTY_JAVABLOCK ||
		 t1.javaBlock () != JavaBlock.EMPTY_JAVABLOCK ) {
		if ( nat == null )
		    nat = new NameAbstractionTable ();
		if (!(t0.javaBlock().equalsModRenaming(t1.javaBlock(), nat)))
		    throw UNSATISFIABLE_EXCEPTION;
	    }

	    if ( !(t0.op() instanceof SchemaVariable) && 
		 t0.op() instanceof ProgramVariable ) {
		if (!(t1.op() instanceof ProgramVariable))
		    throw UNSATISFIABLE_EXCEPTION;
		if ( nat == null )
		    nat = new NameAbstractionTable ();
		if (!((ProgramVariable)t0.op()).
		    equalsModRenaming((ProgramVariable)t1.op(), nat))
		    throw UNSATISFIABLE_EXCEPTION;
	    } else if (!(t0.op().equals(t1.op())))
		throw UNSATISFIABLE_EXCEPTION;

	    for (int i=0;i<t0.arity();i++) {
		if (t0.varsBoundHere(i).size()!=t1.varsBoundHere(i).size())
		    throw UNSATISFIABLE_EXCEPTION;
		for (int j=0; j<t0.varsBoundHere(i).size(); j++) {
		    QuantifiableVariable ownVar =
			t0.varsBoundHere(i).getQuantifiableVariable(j);
		    QuantifiableVariable cmpVar = 
			t1.varsBoundHere(i).getQuantifiableVariable(j);
		    ownVarEnum = ownVarEnum.put(ownVar, 
						new Integer(ownVarEnum.size()));
		    cmpVarEnum = cmpVarEnum.put(cmpVar,
						new Integer(cmpVarEnum.size()));
		}

		newConstraint = unifyHelp(t0.sub(i), t1.sub(i), 
					  ownVarEnum, cmpVarEnum,
					  nat, newConstraint);
	    }
	    return newConstraint;
	}
    }


    /**
     * Unify t1 and t2, modifying this
     * @throws UnsatisfiableException iff unification is not possible
     */
    private EqualityConstraint unifyHelp ( Term t1, Term t2,
					   EqualityConstraint newConstraint )
	throws UnsatisfiableException {
	return unifyHelp ( t1, t2,
			   MapAsListFromQuantifiableVariableToInteger.EMPTY_MAP, 
			   MapAsListFromQuantifiableVariableToInteger.EMPTY_MAP,
			   null,
			   newConstraint );
    }


    /**
     * Exception thrown by unifyHelp in case of ununifiable terms
     */
    private static class UnsatisfiableException extends Exception {}	

    private static final UnsatisfiableException UNSATISFIABLE_EXCEPTION =
	new UnsatisfiableException ();


    /** checks for cycles and adds additional constraints if necessary 
     *
     * PRECONDITION: the sorts of mv and t match; if t is also a
     * metavariable with same sort as mv, the order of mv and t is
     * correct (using Metavariable.compareTo)
     *
     * @param mv Metavariable asked to be mapped to the term t
     * @param t the Term the metavariable should be mapped to (if there
     * are no cycles )
     * @return the resulting Constraint ( == this iff this subsumes
     * the new constraint )
     */ 
    private EqualityConstraint normalize(Metavariable mv, Term t,
					 EqualityConstraint newConstraint)
	throws UnsatisfiableException {
	// MV cycles are impossible if the orders of MV pairs are
	// correct

	if ( !t.isRigid () )
	    throw UNSATISFIABLE_EXCEPTION;	    

	if ( newConstraint == null ) {
	    // metavariable instantiations must not contain free variables
	    if ( t.freeVars ().size () != 0 ||
		 hasCycle ( mv, t ) ) // cycle
		throw UNSATISFIABLE_EXCEPTION;
	    else if ( isDefined ( mv ) )
		return unifyHelp ( valueOf ( mv ), t, newConstraint );
	    else {
		newConstraint = new EqualityConstraint ( (HashMap)map.clone () );
		newConstraint.map.put ( mv, t );
		return newConstraint;
	    }
	} else {
	    // metavariable instantiations must not contain free variables
	    if ( t.freeVars ().size () != 0 ||
		 newConstraint.hasCycle ( mv, t ) ) // cycle
		throw UNSATISFIABLE_EXCEPTION;
	    else if ( newConstraint.isDefined ( mv ) )
		return unifyHelp ( newConstraint.valueOf ( mv ), t,
				   newConstraint );
	    else {
		newConstraint.map.put ( mv, t );
		return newConstraint;
	    }
	}
    }

    /** checks equality of constraints by subsuming relation
     */
    public boolean equals ( Object obj ) {
	if ( obj instanceof Constraint ) {
	    Constraint c = (Constraint)obj;
	    if ( c instanceof EqualityConstraint )
		return map.keySet ().equals
		    ( ((EqualityConstraint)c).map.keySet () ) &&
		    join ( c ) == this && c.join ( this ) == c;
	    return isAsStrongAs ( c ) && isAsWeakAs ( c );
	}
	return false;
    }

    /**
     * @return true iff this constraint is as strong as "co",
     * i.e. every instantiation satisfying "this" also satisfies "co".
     */
    public boolean isAsStrongAs ( Constraint co ) {
	if ( co instanceof EqualityConstraint )
	    // use necessary condition for this relation: key set of
	    // this is superset of key set of co
	    return
		map.keySet ().containsAll
		( ((EqualityConstraint)co).map.keySet () ) &&
		join ( co ) == this;
	return co.isAsWeakAs ( this );
    }

    /**
     * @return true iff this constraint is as weak as "co",
     * i.e. every instantiation satisfying "co" also satisfies "this".
     */
    public boolean isAsWeakAs ( Constraint co ) {
	if ( co instanceof EqualityConstraint )
	    // use necessary condition for this relation: key set of
	    // co is superset of key set of this
	    return
		((EqualityConstraint)co).map.keySet ().containsAll
		( map.keySet () ) &&
		co.join ( this ) == co;
	return co.isAsStrongAs ( this );	
    }

    /** joins the given constraint with this constraint
     * and returns the joint new constraint. 
     * @param co Constraint to be joined with this one
     * @return the joined constraint 
     */	
    public Constraint join(Constraint co) { 
	return join(co, CONSTRAINTBOOLEANCONTAINER); 
    }
    

    /** joins constraint co with this constraint
     * and returns the joint new constraint. The BooleanContainer is
     * used to wrap a second return value and indicates a subsumption
     * of co by this constraint. 
     * @param co Constraint to be joined with this one
     * @param unchanged the BooleanContainers value set true, if this
     * constraint is as strong as co
     * @return the joined constraint     
     */
    public Constraint join ( Constraint co, BooleanContainer unchanged ) {

	if ( co.isBottom () || co == this ) {
	    unchanged.setVal ( true );
	    return this;
	} else if ( isBottom () ) {
	    unchanged.setVal ( false );
	    return co;
	}

	if ( !( co instanceof EqualityConstraint ) ) {
	    // BUG: Don't know how to set p_subsumed (at least not
	    // efficiently)
	    unchanged.setVal ( false );
	    return co.join ( this );
	} else {

	    Iterator           it            =
		((EqualityConstraint)co).map.entrySet ().iterator ();
	    EqualityConstraint newConstraint = null;
	    Map.Entry          entry;

	    try {
		while ( it.hasNext () ) {
		    entry         = (Map.Entry)it.next ();
		    newConstraint = normalize ( (Metavariable)entry.getKey   (),
						(Term)        entry.getValue (),
						newConstraint );
		}
	    } catch ( UnsatisfiableException e ) {
		unchanged.setVal ( false );
		return Constraint.TOP;
	    }

	    if ( newConstraint == null ) {
		unchanged.setVal ( true );
		return this;
	    }

	    unchanged.setVal ( false );
	    HashMap i1 = buildInvalidCache ();
	    HashMap i2 = ((EqualityConstraint)co).buildInvalidCache ();
	    if ( i1 != null ) {
		if ( i2 != null )
		    i1.putAll ( i2 );
		newConstraint.instantiationCacheInvalid = i1;
	    } else
		newConstraint.instantiationCacheInvalid = i2;
	    return newConstraint;

	}
    }
    
    /**
     * @return a constraint derived from this one by removing all
     * constraints on the given variable, which may therefore have any
     * value according to the new constraint (the possible values of
     * other variables are not modified)
     */
    public Constraint removeVariables ( SetOfMetavariable mvs ) {
	if ( mvs != SetAsListOfMetavariable.EMPTY_SET && !isBottom () ) {
	    Iterator           it               = map.entrySet ().iterator ();
	    Map.Entry          entry;
	    EqualityConstraint removeConstraint = new EqualityConstraint ();
	    EqualityConstraint newConstraint    = new EqualityConstraint ();

	    // Find equalities with removed metavariable as left side
	    while ( it.hasNext () ) {
		entry = (Map.Entry)it.next ();
		if ( mvs.contains ( (Metavariable)entry.getKey () ) )
		    removeConstraint.map.put
			( entry.getKey (), entry.getValue () );
		else
		    newConstraint   .map.put
			( entry.getKey (), entry.getValue () );
	    }

	    // Find equalities with removed metavariable as right side
	    it = newConstraint.map.entrySet ().iterator ();
	    while ( it.hasNext () ) {
		entry = (Map.Entry)it.next ();
		if ( ((Term)entry.getValue ()).op () instanceof Metavariable &&
		     ( ((Metavariable)entry.getKey   ()).sort () ==
		       ((Term)        entry.getValue ()).sort () ) &&
		     ( mvs.contains
		       ( (Metavariable)((Term)entry.getValue ()).op () ) ) &&
		     !( removeConstraint.map.containsKey
			( ((Term)entry.getValue ()).op () ) ) ) {
		    removeConstraint.map.put
			( ((Term)entry.getValue ()).op (),
			  TermFactory.DEFAULT.createFunctionTerm
			  ( (Metavariable)entry.getKey () ) );
		    it.remove ();
		}
	    }

	    if ( !removeConstraint.map.isEmpty () ) {
		// Substitute removed variables occurring within right
		// sides of the remaining equalities

		// Usually at this point removeConstraint is not a
		// well-formed constraint, as the order of MV may be
		// wrong. However, "SyntacticalReplaceVisitor" doesn't
		// care about that.
		if ( newConstraint.map.isEmpty () )
		    return Constraint.BOTTOM;
		else {
		    it            = newConstraint.map.entrySet ().iterator ();
		    newConstraint = new EqualityConstraint ();

		    while ( it.hasNext () ) {
			entry = (Map.Entry)it.next ();
			newConstraint.map.put
			    ( entry.getKey (),
			      removeConstraint.instantiate
			      ( (Term)entry.getValue () ) );
		    }

		    return newConstraint;
		}
	    }
	}

	return this;
    }

    /** checks if there is a cycle if the metavariable mv and Term
     * term would be added to this constraint e.g. X=g(Y), Y=f(X) 
     * @param mv the Metavariable   
     * @param term The Term 
     * @return a boolean that is true iff. adding a mapping (mv,term)
     * would cause a cycle 
     */

    boolean hasCycle ( Metavariable mv, Term term ) {
	HashSet                body          = new HashSet ();
	Term                   checkForCycle = term;
	ListOfTerm             fringe        = SLListOfTerm.EMPTY_LIST;
	Metavariable           termMV;
	MVCollector            coll;
	IteratorOfMetavariable it;

	fringe = fringe.prepend ( checkForCycle );

	while ( fringe != SLListOfTerm.EMPTY_LIST ) {
	    checkForCycle = fringe.head ();
	    fringe        = fringe.tail ();
	    coll          = new MVCollector();
	    checkForCycle.execPostOrder(coll);
	    it            = coll.mv ();
	    while ( it.hasNext () ) {
		termMV = it.next ();
		if ( !body.contains ( termMV ) ) {
		    body.add ( termMV );
		    if ( map.containsKey ( termMV ) )
			fringe = fringe.prepend ( (Term)map.get (termMV ) );
		    if ( termMV.equals ( mv ) )
			return true;
		}
	    }
	}

	return false;	
    }
    
    /**ONLY FOR TESTS DONT USE THEM IN ANOTHER WAY
     * @return true if metavar is contained as key */
    boolean isDefined ( Metavariable mv ) {
	return map.containsKey ( mv );
    }
    
    /** ONLY FOR TESTS DONT USE THEM IN ANOTHER WAY 
     * @return mapping to mv */
    Term valueOf ( Metavariable mv ) {
	return (Term)map.get ( mv );
    }

    /** @return String representation of the constraint */
    public String toString () {
	return map.toString ();
    }

}
