// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/**
 * this class describes a position in an occurrence of a term. A
 * ConstrainedFormula and a PosInTerm determine an object of this class exactly.
 */
package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.proof.IteratorOfNode;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.op.Metavariable;

public class PosInOccurrence implements java.io.Serializable {


    private ConstrainedFormula cfma;
    private PosInTerm posInTerm;

    // Descend into metavariables instantiations
    private Term      metaTerm      = null;
    private PosInTerm metaPosInTerm = null;

    private Sequent seq;
    private Proof proof;

    /**
     * is true iff there had been a computation of the isAntec
     * bool previously. 
     */    
    private boolean inAntecDetermined=false;

    /**
     * is true iff the position is in the antecedent of a sequent. Only valid
     * if isAntecDetermined is true
     */
    private boolean inAntec=false;


    /**
     * constructs a new PosInOccurrence object with the given
     * ConstrainedFormula and the given PosInTerm determining the position in
     * an occurrence. The Proof is given in order to be able to determine if
     * the ConstrainedFormula occurrs in the antecedent or in the succedent.
     * @param posInTerm a position in a term
     * @param cfma a formula with constraint
     * @param proof a Proof with a sequent containing the cfma.
     */
    public PosInOccurrence(ConstrainedFormula cfma, PosInTerm
			   posInTerm, Proof proof) {
	this.cfma=cfma;
	this.posInTerm=posInTerm;
	this.proof=proof;
    }

    /**
     * constructs a new PosInOccurrence object with the given
     * ConstrainedFormula and the given PosInTerm determining the position in
     * an occurrence. A Sequent is given to facilitate the query
     * determining if the ConstrainedFormula is in the antecedent or
     * the succedent. 
     * @param posInTerm a position in a term
     * @param cfma a formula with constraint
     */
    public PosInOccurrence(ConstrainedFormula cfma, PosInTerm
			   posInTerm, Sequent seq) {
	this.cfma=cfma;
	this.posInTerm=posInTerm;
	this.seq=seq;
    }  

    public PosInOccurrence(ConstrainedFormula cfma, PosInTerm
			   posInTerm, Sequent seq,
			   boolean inAntec) {
	this ( cfma, posInTerm, seq );
	this.inAntec           = inAntec;
	this.inAntecDetermined = true;
    }

    public PosInOccurrence(ConstrainedFormula cfma, PosInTerm
			   posInTerm, Sequent seq,
			   Term metaTerm, PosInTerm metaPosInTerm) {
	this.cfma=cfma;
	this.posInTerm=posInTerm;
	this.seq=seq;
	this.metaTerm=metaTerm;
	this.metaPosInTerm=metaPosInTerm;
    }  

    private PosInOccurrence(ConstrainedFormula cfma, PosInTerm 
			    posInTerm, Sequent seq, Proof proof,
			    boolean inAntec, boolean 
			    inAntecDetermined) {
	this.inAntecDetermined=inAntecDetermined;
	this.inAntec=inAntec;
	this.cfma=cfma;
	this.proof=proof;
	this.seq=seq;
	this.posInTerm=posInTerm;
    }

    private PosInOccurrence(ConstrainedFormula cfma, PosInTerm 
			    posInTerm,
			    Term metaTerm, PosInTerm metaPosInTerm,
			    Sequent seq, Proof proof,
			    boolean inAntec, boolean 
			    inAntecDetermined) {
	this.inAntecDetermined=inAntecDetermined;
	this.inAntec=inAntec;
	this.cfma=cfma;
	this.proof=proof;
	this.seq=seq;
	this.posInTerm=posInTerm;
	this.metaTerm=metaTerm;
	this.metaPosInTerm=metaPosInTerm;
    }

    /**
     * creates a new PosInOccrrence that has exactly the same state as this
     * object except the PosInTerm is new and walked down the specified
     * subterm, as specified in method down of 
     * {@link de.uka.ilkd.key.logic.PosInTerm}.
     */
    public PosInOccurrence down(int i) {
	if ( metaTerm == null )
	    return new PosInOccurrence(cfma, posInTerm.down(i), seq,
				       proof, inAntec, inAntecDetermined);
	else
	    return new PosInOccurrence(cfma, posInTerm, metaTerm, metaPosInTerm.down(i), seq,
				       proof, inAntec, inAntecDetermined);
    }

    /**
     * If "this" points to a metavariable, a term can be inserted at
     * this position, allowing to walk down the term further "inside"
     * the metavariable.
     */
    public PosInOccurrence setTermBelowMetavariable ( Term p_metaTerm ) {
	Debug.assertTrue ( metaTerm == null &&
			   ( constrainedFormula().formula().subAt(posInTerm).op()
			     instanceof Metavariable ) &&
			   p_metaTerm.sort().extendsTrans
			   ( constrainedFormula().formula().subAt(posInTerm).sort() ),
			   "Illegal position (no metavariable) or incompatible sorts" );

	return new PosInOccurrence(cfma, posInTerm, p_metaTerm, PosInTerm.TOP_LEVEL,
				   seq, proof, inAntec, inAntecDetermined);
    }

    /**
     * returns the ConstrainedFormula that determines the occurrence of
     * this PosInOccurrence 
     */
    public ConstrainedFormula constrainedFormula() {
	return cfma;
    }

    /**
     * returns the subterm this object points to
     */
    public Term subTerm () {
	if ( metaTerm == null )
	    return constrainedFormula().formula().subAt(posInTerm);
	else
	    return metaTerm.subAt ( metaPosInTerm );
    }

    /**
     * returns the position in the formula of the ConstrainedFormula of
     * this PosInOccurrence. 
     */
    public PosInTerm posInTerm() {
	return posInTerm;
    }


    public Term termBelowMetavariable () {
	return metaTerm;
    }


    public PosInTerm posInTermBelowMetavariable () {
	return metaPosInTerm;
    }


    public Proof proof () {
	return this.proof;
    }


    /**
     * returns true iff the occurrence is in the
     * antecedent of a sequent. 
     */
    public boolean isInAntec() {
	if (!inAntecDetermined) {
	    determineAntecSucc();
	}
	return inAntec;
    }

    /** compares this PosInOccurrence with another PosInOccurrence
     * and returns true if both describe the same occurrence */
    public boolean equals(Object obj) {
	if (!(obj instanceof PosInOccurrence)) {
	    return false;
	}
	PosInOccurrence cmp = (PosInOccurrence) obj;
	if (constrainedFormula()!=cmp.constrainedFormula()) {
	    return false;
	}
	if (isInAntec() == cmp.isInAntec()) {
	    return posInTerm().equals(cmp.posInTerm());
	}
	return false;
    }
     

    /**
     * determines the 'inAntec' information depending on what has been given
     * when constructing the PosInOccurrence: Sequence or Proof.
     */
    private void determineAntecSucc() {
	if (seq==null) {
	    determineAntecSucc(proof);
	} else {
	    determineAntecSucc(seq);
	}
    }

    /**
     * sets the attributes 'isAntec' and 'isAntecDetermined': The given
     * sequent is searched for the ConstrainedFormula of this
     * PosInOccurrence. If it is found in the antecedent isAntec is set to
     * true; if it is found in the succedent isAntec is set to false. If one
     * of these two cases hold isAntecDetermined is set to true, otherwise
     * to false.
     * @param seq the sequent that determines if the ConstrainedFormula is in
     * the succedent or in the antecedent.
     */
    private void determineAntecSucc(Sequent seq) {
	if (seq.antecedent().contains(cfma)) {
	    inAntec=true;
	    inAntecDetermined=true;
	} else if (seq.succedent().contains(cfma)) {
	    inAntec=false;
	    inAntecDetermined=true;
	} else {
	    inAntecDetermined=false;
	}
    }

    /**
     * sets the attributes 'isAntec' and 'isAntecDetermined': The given
     * proof is searched for the ConstrainedFormula of this
     * PosInOccurrence. If it is found in the antecedent of a sequent in the
     * proof isAntec is set to
     * true; if it is found in the succedent isAntec is set to false. If one
     * of these two cases hold isAntecDetermined is set to true, otherwise
     * to false.
     * @param proof the proof that determines if the ConstrainedFormula is in
     * the succedent or in the antecedent of a sequent.
     */
    private void determineAntecSucc(Proof proof) {
	determineAntecSuccAfterNode(proof.root());
    }

    /**
     * calls determineAntecSucc for each sequence of the nodes after (and
     * inclusive) the given node until all nodes have been visited or the
     * ConstrainedFormula could be found.
     * @param node the start node of the search
     */
    private void determineAntecSuccAfterNode(Node node) {
	determineAntecSucc(node.sequent());
	if (!inAntecDetermined) {
	    IteratorOfNode it=node.childrenIterator();
	    while (it.hasNext() && !inAntecDetermined) {
		determineAntecSuccAfterNode(it.next());
	    }
	}
    }
    
    
    public static PosInOccurrence findInSequent(Sequent seq, 
                                                int formulaNumber, 
                                                PosInTerm pos) {
       return new PosInOccurrence(seq.getFormulabyNr(formulaNumber),  pos, seq);
    }
    
    
    public int formulaNumberInSequent() {
       return seq.formulaNumberInSequent(isInAntec(), cfma);
    }
    
    
    /** toString */
    public String toString() {
	return "Term "+
	    posInTerm()+" of "+
	    constrainedFormula();
//            formulaNumberInSequent()+" th formula";

    }
	
}
