// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.collection.*;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.JavaNonTerminalProgramElement;
import de.uka.ilkd.key.java.StatementContainer;

/** this class describes the position of a statement in a program.
 */
public class PosInProgram implements java.io.Serializable {
  
    /** list of Integers encoding the exact position */
    private ListOfInteger positionList=SLListOfInteger.EMPTY_LIST;
    /** list of Integers encoding the exact position */
    private ListOfInteger reversePositionList=SLListOfInteger.EMPTY_LIST;
    /** pos at the beginning of the program */
    public final static PosInProgram TOP = new PosInProgram();

    
    /** returns the ProgramElement at the given position 
     * @param pos the PosInProgram 
     * @param program the ProgramElement we walk through
     * @return the ProgramElement at the given position 
     */
    public static ProgramElement getProgramAt(PosInProgram pos,
					      ProgramElement prg)
    {
	ProgramElement result = prg;
	JavaNonTerminalProgramElement container = null;
	IteratorOfInteger it = pos.iterator();	
	while (it.hasNext()) {	    
	    if (result instanceof JavaNonTerminalProgramElement) {
		container = (JavaNonTerminalProgramElement) result;
	    } else {
		throw new RuntimeException("Unknown Position.");
	    }
	    result = container.getChildAt(it.next().intValue());
	}
	return result;
    }

    /** returns the ProgramElement at the given position 
     * @param pos the PosInProgram 
     * @param program the ProgramElement we walk through
     * @return the ProgramElement at the given position 
     */
    public static ProgramElement getProgramAt(PosInProgram pos,
					      JavaBlock block)
    {
	return getProgramAt(pos, block.program());
    }

    /**
     * creates a new PosInProgram 
     * @param positionList ListOfInteger describing the exact
     * position. 
     */
    private PosInProgram(ListOfInteger positionList) {
	this.positionList = positionList;
    }   

    /**
     * creates a new PosInProgram 
     * position. 
     */
    private PosInProgram() {
    }   

    /** size of the position list */
    public int depth() {
	return positionList.size();
    }
      
    /** descending downwards choosing the n'th statement of the program 
     * @param nr int describes the position of the statement in the block
     * @return position of the statement
     */    
    public PosInProgram down(int n) {
	return new PosInProgram(positionList.append(new Integer(n)));
    }

    /**
     * creates a new PosInProgram representing a position going down from
     * this PosInProgram with the given PosInProgram. Example: is this
     * PosInProgram representin position [x,y,z] and the given PosInProgram
     * pp representing position [u,v,w] then the new position is [x,y,z,u,v,w].
     * @param pp the position to be appended.
     * @return a new position where the append is done
     */
    public PosInProgram append(PosInProgram pp) {
	return new PosInProgram(positionList.append(pp.positionList));
    }
    
    /** compares this PosInProgram with another PosInProgram
     * and returns true if both describe the same position 
     */
    public boolean equals(Object obj) {
	if (!(obj instanceof PosInProgram)) {
	    return false;
	}
	PosInProgram cmp = (PosInProgram) obj;
	IteratorOfInteger it = iterator();
	IteratorOfInteger cmpIt = cmp.iterator();
	while (it.hasNext()) {
	    if (cmpIt.hasNext()) {
		if (it.next().intValue() != 
		    cmpIt.next().intValue()) {
		    return false;
		}
	    } else {
		return false;
	    }
	}
	return !cmpIt.hasNext();
    }


    /** reverse position list head element becomes last element 
     * @return reversed list */
    private ListOfInteger reversePositionList() {
	ListOfInteger reverse=SLListOfInteger.EMPTY_LIST;
	IteratorOfInteger it=positionList.iterator();
	while(it.hasNext()) {
	    reverse=reverse.prepend(it.next());	    
	}
	return reverse;
    }

    
    public boolean leq(PosInProgram pos) {
	IteratorOfInteger longerIt = iterator();	
	IteratorOfInteger shorterIt = pos.iterator();
	
	boolean result = false;
	while (shorterIt.hasNext() && longerIt.hasNext()) {
	    if (shorterIt.next().intValue()<longerIt.next().intValue()) {
		return false;
	    }
	}

	return true;
    }


    /**
     * returns an IteratorOfInteger that iterates through the subterms
     * of a sequent in the reverse order as the PosInProgram has been defined.
     * @return IteratorOfInteger that iterates through the subterms
     * of a sequent in the reverse order as the PosInProgram has been defined.
     */ 

    public IteratorOfInteger reverseIterator() {
	if (positionList!=SLListOfInteger.EMPTY_LIST &&
	    reversePositionList==SLListOfInteger.EMPTY_LIST) {
		reversePositionList=reversePositionList();	    
	}       
	return reversePositionList.iterator();
    }

    public Integer get(int i) {
	if (i>=positionList.size()) {
	    throw new ArrayIndexOutOfBoundsException();
	}
	IteratorOfInteger it=iterator();
	for (int k=0; k<i; k++) {
	    it.next();
	}
	return it.next();
    }

    /**
     * returns an iterator over the list defining the position in a term.
     * @return an iterator over the list defining the position in a term.
     */
    public IteratorOfInteger iterator() {
	return positionList.iterator();
    }

    public ListOfInteger list() {
	return positionList;
    }

    /** toString */
    public String toString() {
	return "PosInProgram: "+positionList;
    }
    
}













