// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/** this class describes a position in a term 
 */
package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.collection.*;

public class PosInTerm implements java.io.Serializable {
  
    /** list of Integers encoding the exact position */
    private ListOfInteger positionList = SLListOfInteger.EMPTY_LIST;
    /** list of Integers encoding the exact position */
    private ListOfInteger reversePositionList = SLListOfInteger.EMPTY_LIST;
    /** top level term position */
    public static PosInTerm TOP_LEVEL=new PosInTerm();

    /** creates new position */
    private PosInTerm() {
    }   

    /**
     * creates a new PosInTerm that determines a subterm in a term.
     * @param positionList ListOfInteger describing the exact
     * position. A list containing x1,...,xn as elements describes the
     * xn'th subterm of the x(n-1)'th subterm of ... of the x1'th subterm
     * of a term.
     */
    private PosInTerm(ListOfInteger positionList) {
	this.positionList=positionList;
    }   
    
  
    /** descending downwards choosing the n'th subterm, subformula 
     * @param nr int describs the chosen subterm 
     * @return Position like old Position but with a deeper
     * subterm.
     */    
    public PosInTerm down(int n) {
	return new PosInTerm(positionList.append(new Integer(n)));
    }


    /** 
     * compares this PosInTerm with another PosInTerm
     * and returns true if both describe the same position 
     */
    public boolean equals(Object obj) {
	if (!(obj instanceof PosInTerm)) {
	    return false;
	}
	PosInTerm cmp = (PosInTerm) obj;
	IteratorOfInteger it = iterator();
	IteratorOfInteger cmpIt = cmp.iterator();
	while (it.hasNext()) {
	    if (cmpIt.hasNext()) {
		if (it.next().intValue() != 
		    cmpIt.next().intValue()) {
		    return false;
		}
	    } else {
		return false;
	    }
	}
	return !cmpIt.hasNext();
    }


    /** 
     * reverse position list head element becomes last element 
     * @return reversed list 
     */
    public ListOfInteger reversePositionList() {
	ListOfInteger reverse = SLListOfInteger.EMPTY_LIST;
	IteratorOfInteger it  = positionList.iterator();
	while (it.hasNext()) {
	    reverse = reverse.prepend(it.next());	    
	}
	return reverse;
    }

    /**
     * get subterm/-formula information.  A PosInTerm defining the xn'th
     * subterm of the x(n-1)'th subterm of ... of the x1'th subterm of a
     * term returns an iterator subsequently resulting in the list xn...x1.
     * @return IteratorOfInteger that iterates through the subterms
     * of a sequent in the reverse order as the PosInTerm has been defined.
     */ 
    public IteratorOfInteger reverseIterator() {
	if (positionList != SLListOfInteger.EMPTY_LIST &&
	    reversePositionList == SLListOfInteger.EMPTY_LIST) {
	    reversePositionList = reversePositionList();	    
	}       
	return reversePositionList.iterator();
    }

    /**
     * returns an iterator defining the position in a term according to
     * subterm relation. A PosInTerm defining the xn'th subterm of the
     * x(n-1)'th subterm of ... of the x1'th subterm of a term returns an
     * iterator subsequently resulting in the list x1...xn.
     * @return an iterator over the list defining the position in a term.
     */
    public IteratorOfInteger iterator() {
	return positionList.iterator();
    }
    
    
    public static PosInTerm parseReverseString(String s) {
       if ("".equals(s)) return TOP_LEVEL;
       ListOfInteger list=SLListOfInteger.EMPTY_LIST;
       java.util.StringTokenizer tker = 
          new java.util.StringTokenizer(s, ",", false);
       while (tker.hasMoreTokens()) 
	   list = list.prepend(Integer.decode(tker.nextToken()));
       return new PosInTerm(list);       
    }

    
    /** toString */
    public String toString() {
	return (this!=TOP_LEVEL) ? "subterm "+positionList : "top level";
    }

}
