// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/** represents a name that is part of a program
 */
package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.visitor.Visitor;
import de.uka.ilkd.key.java.reference.ReferenceSuffix;

public class ProgramElementName extends Name 
    implements TerminalProgramElement, Label, ReferenceSuffix {

    /** create a new name
     * @param name the name
     */
    public ProgramElementName(String name){
	super(name);
    }

    /** @return comments */
    public Comment[] getComments() {
	return null;
    }

    /**
     * to be compatible to a ProgramElement
     */
    public SourceElement getFirstElement() {
	return this;
    }


    /**
     * to be compatible to a ProgramElement
     */
    public SourceElement getLastElement() {
	return this;
    }

    public void prettyPrint(PrettyPrinter w) throws java.io.IOException {
	w.printProgramElementName(this);
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnProgramElementName(this);
    }


    /**
       Returns the start position of the primary token of this element.
       To get the start position of the syntactical first token,
       call the corresponding method of <CODE>getFirstElement()</CODE>.
       @return the start position of the primary token.
    */
    public Position getStartPosition(){
	return Position.UNDEFINED;
    }

    /**
       Returns the end position of the primary token of this element.
       To get the end position of the syntactical first token,
       call the corresponding method of <CODE>getLastElement()</CODE>.
       @return the end position of the primary token.
     */
    public Position getEndPosition(){
	return Position.UNDEFINED;
    }

    /**
       Returns the relative position (number of blank heading lines and 
       columns) of the primary token of this element.
       To get the relative position of the syntactical first token,
       call the corresponding method of <CODE>getFirstElement()</CODE>.
       @return the relative position of the primary token.
     */
    public Position getRelativePosition(){
	return  Position.UNDEFINED;
    }

    /** equals modulo renaming is described in the corresponding
     * comment in class SourceElement. The ProgramElementName has to
     * check if an abstract name has been assigned and if, if both
     * elements are assigned to the same name, otherwise the names
     * have to be equal
     */
    public boolean equalsModRenaming(SourceElement se, 
				     NameAbstractionTable nat) {
	if (!(se instanceof ProgramElementName)) {
	    return false;
	}
	if (nat.contains(this)) {
	    return nat.sameAbstractName(this, se);
	}
	return this.equals(se);
    }
}
