// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.logic;

/**
 * Records the changes made to a sequent. Keeps track of added and removed
 * formula to one of the semisequents. The intersection of added and removed
 * formulas of the same semisequent may not be empty, in particular this means
 * that a formula added and removed afterwards will occur in both lists. The
 * situation where this can happen is that a list of formulas had to be added to
 * the sequent and the list has not been redundance free.  
 */
public class SequentChangeInfo implements java.io.Serializable {

  /** change information related to the antecedent, this means the
   * there added and removed formulas*/
  private SemisequentChangeInfo antecedent;
  
  /** change information related to the antecedent, this means the
   * there added and removed formulas*/
  private SemisequentChangeInfo succedent;

  /**
   * the sequent after the changes
   */
  private Sequent sequent;
         
  /**
   * creates a new sequent change info whose semisequent described by position
   * pos has changed. The made changes are stored in semiCI and the resulting
   * sequent is given by result
   * 
   * @param pos the PosInOccurrence describing the semisequent where
   * the changes took place
   * @param semiCI the SemisequentChangeInfo describing the changes in
   * detail (which formulas have been added/removed)
   * @return the sequent change information object describing the
   * complete changes made to the sequent together with the operations
   * result.  
   */
  public static SequentChangeInfo createSequentChangeInfo
  (PosInOccurrence pos, SemisequentChangeInfo semiCI, Sequent result) {
    return createSequentChangeInfo(pos.isInAntec(), semiCI, result);
  }

  /**
   * creates a new sequent change info whose semisequent described by the
   * value of the selector antec (true selects antecedent; false selects
   * succedent) has changed. The made changes are stored in semiCI and the
   * resulting sequent is given by result
   *
   * @param pos the PosInOccurrence describing the semisequent where
   * the changes took place 
   * @param semiCI the SemisequentChangeInfo describing the
   * changes in detail (which formulas have been added/removed)
   * @return the sequent change information object describing the
   * complete changes made to the sequent together with the operations
   * result.  */
  public static SequentChangeInfo createSequentChangeInfo
  (boolean antec, SemisequentChangeInfo semiCI, Sequent result) {
    if (antec) {
      return new SequentChangeInfo(semiCI, null, result);
    } else {
      return new SequentChangeInfo(null, semiCI, result);
    }
  }

    
  /**
   * creates a new sequent change information object. Therefore it
   * combines the changes to the semisequents of the sequent.
   * @param antecedent the SemisequentChangeInfo describing the changes of
   * the antecedent
   * @param succedent the SemisequentChangeInfo describing the changes of
   * the succedent
   * @param sequent the Sequent being the result of the changes 
   */
  private SequentChangeInfo(SemisequentChangeInfo antecedent, 
			    SemisequentChangeInfo succedent,
			    Sequent sequent) {       
    this.antecedent = antecedent;
    this.succedent = succedent;
    this.sequent = sequent;
  }

  /** 
   * returns true iff the sequent has been changed by the operation 
   * @return true iff the sequent has been changed by the operation 
   */    
  public boolean hasChanged() {
    return antecedent.hasChanged() || succedent.hasChanged();
  }

  /** 
   * returns true if the selected part of sequent has changed. Thereby the
   * flag 'antec' specifies the selection: true selects the antecedent and
   * false the succedent of the sequent.
   *
   * @return true iff the sequent has been changed by the operation */
  public boolean hasChanged(boolean antec) {
    return antec ? (antecedent != null && antecedent.hasChanged()) :
      (succedent != null && succedent.hasChanged());
  }

  /** 
   * The formulas added to one of the semisequents are returned.  The selected
   * semisequent depends on the value of selector 'antec' which is the
   * antecedent if 'antec' is true and the succedent otherwise.
   *
   * @param antec a boolean used to select one of the two semisequents
   * of a sequent (true means antecedent; false means succedent)
   * @return list of formulas added to the selected semisequent 
   */
  public ListOfConstrainedFormula addedFormulas(boolean antec) {
    return antec ? (antecedent != null ? antecedent.addedFormulas()
		    : SLListOfConstrainedFormula.EMPTY_LIST) :
      (succedent != null ? succedent.addedFormulas() : 
       SLListOfConstrainedFormula.EMPTY_LIST);
  }

  /** 
   * The formulas added to the sequent are returned as a concetenated list of
   * the formulas added to each semisequent.
   * @return list of formulas added to sequent 
   */
  public ListOfConstrainedFormula addedFormulas() {
    return addedFormulas(true).size() > addedFormulas(false).size() ?
	addedFormulas(false).prepend(addedFormulas(true)) : 
	addedFormulas(true).prepend(addedFormulas(false));
  }

  /** 
   * The formulas removed from one of the semisequents are returned.  The
   * selected semisequent depends on the value of selector 'antec' which is
   * the antecedent if 'antec' is true and the succedent otherwise.
   *
   * @param antec a boolean used to select one of the two semisequents
   * of a sequent (true means antecedent; false means succedent)
   * @return list of formulas removed from the selected semisequent 
   */
  public ListOfConstrainedFormula removedFormulas(boolean antec) {
    return antec ? (antecedent != null ? antecedent.removedFormulas()
		    : SLListOfConstrainedFormula.EMPTY_LIST) :
      (succedent != null ? succedent.removedFormulas() : 
       SLListOfConstrainedFormula.EMPTY_LIST);
  }

  /** 
   * The formulas removed from the sequent are returned as a
   * concetenated list of the formulas removed from each semisequent.
   * @return list of formulas removed from the sequent 
   */
  public ListOfConstrainedFormula removedFormulas() {
    return removedFormulas(true).size() > removedFormulas(false).size() ?
	removedFormulas(false).prepend(removedFormulas(true)) : 
	removedFormulas(true).prepend(removedFormulas(false));
  }

  /**
   * returns the resulting sequent 
   * @return the resulting sequent
   */
  public Sequent sequent() {
    return sequent;
  }

  /**
   * toString helper 
   */
  private String toStringHelp(boolean antec) {
    String result = "";
    if (hasChanged(antec)) {
      result += "\t added:" +addedFormulas(antec);
      result += "\t removed:" +removedFormulas(antec);
    }
    return result;
  }

  /**
   * toString
   */
  public String toString() {
    String result = "antecedent: "+hasChanged(true);
    result += toStringHelp(true);

    result += "\n succedent: "+hasChanged(false);
    result += toStringHelp(false);       

    return result;
  }

}
	
