package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.util.Debug;

import java.util.HashSet;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;

/**
 * Performs simplification of updates. It is parametrised with a
 * deletion strategy and a boolean flag that indicates if simultaneous updates
 * shall be created. As default simultaneous updates and the default deletion
 * strategy are used.  
*/
public class SimultaneousUpdateSimplifier {

    /**
     * shortcut to access the default termfactory
     */
    private final TermFactory tf = TermFactory.DEFAULT;

    /** 
     * The deletion strategy performs simplifications that delete parts of
     * existing updates, making use of static analysis.
     */
    private DeletionStrategy deletionStrategy;

    /** 
     * if true then consecutive updates are merged to a single simultaneous
     * update
     */
    private boolean useSimultaneousUpdates = true;


    /** 
     * the contraint under which the formula is simplified 
     */
    private Constraint formulaConstraint = Constraint.BOTTOM;

    /** 
     * creates a new simplifier that makes use of simultaneous updates and the
     * DefaultDeletionStrategy 
     * 
     */
    public SimultaneousUpdateSimplifier() {
	this.deletionStrategy = new DefaultDeletionStrategy();
    }

    /**
     * creates a new simplifier 
     * @param deletionStrategy the DeletionStrategy to use 
     * @param useSimultaneousUpdates a boolean that allows the use of simultaneous
     * updates
     */
    public SimultaneousUpdateSimplifier(DeletionStrategy deletionStrategy,
					boolean useSimultaneousUpdates) {

	this.useSimultaneousUpdates = useSimultaneousUpdates;
	this.deletionStrategy = deletionStrategy;	
    }

    // helpers

    /**
     * creates an if-else term, if necessary or resolves the term
     * if possible
     */
    private final Term createConditionalTerm(Operator op, Term[] subs) {
	if (op == Op.IF_ELSE) {
	    return createIfElseTerm(subs);
	} else if (op instanceof ConjCond) {
	    return createConjCondTerm(subs);
	}
	Debug.fail();
	return null;
    }

    /**
     * creates an if-else term, if necessary or resolves the term
     * if possible
     */
    private final Term createIfElseTerm(Term[] subs) {
	if (subs[0].equals(subs[1])) {
	    // return then part
	    return subs[2];
	}
	if (subs[2].equals(subs[3])) {
	    return subs[2];
	}
	return tf.createIfElseTerm(Op.IF_ELSE, subs);
    }

    /**
     * creates a conj. conditional term, if necessary or resolves the
     * term as far as possible
     */
    private final Term createConjCondTerm(Term[] subs) {
	Term[] result = subs;
	boolean sameBranches = true;
	for (int i = 3; i < subs.length; i++) {
	    if (subs[0].equals(subs[i]) && 
		subs[1].equals(subs[i+1])) {
		if (i < 3) {
		    return subs[2];
		} else {
		    result = new Term[i];
		    System.arraycopy(subs, 0, result, 0, i);
		    result[2] = subs[i+2];		    
		    break;
		}     
	    }
	    sameBranches = sameBranches && 
		subs[2].equals(subs[i+2]);
	}
	if (sameBranches) {
	    return subs[2];
	} else {
	    return tf.createConjCondTerm(subs);
	}
    }

    

    // Basic Simplification Rules

    /**
     * an update is applied on a programm variable      
     * @preconditions var.op().oclIsTypeOf(ProgramVariable)    
     */
    private Term applyOnProgramVariable(Term update, 
					Term var) {
	
	for (int i = update.arity()-3; i >= 0; i-=2) {
	    if (update.sub(i).op() == var.op()) {
		return update.sub(i+1);
	    }
	}
	return var;
    }


    /**
     * a list of updates is applied on a programm variable      
     * @preconditions var.op().oclIsTypeOf(ProgramVariable)
     */
    private Term applyOnProgramVariable(ListOfTerm updates, Term var) {
	ListOfTerm unusedUpdates = updates;	
	Term result = var;	
	while (unusedUpdates != SLListOfTerm.EMPTY_LIST &&
	       result.op() instanceof ProgramVariable) {	    
	    result        = applyOnProgramVariable(unusedUpdates.head(), 
						   result);
	    unusedUpdates = unusedUpdates.tail();
	}
	return unusedUpdates == SLListOfTerm.EMPTY_LIST ? 
	    result : simplify(unusedUpdates, result);
    }


    /**
     * an update is applied on a static attribute
     * @preconditions attribute.op().oclIsTypeOf(Attribute)
     */
    private Term applyOnStaticAttribute(Term update,
					Term attr) {
	TermSymbol attribute = Op.ATTR.attribute(attr);

	for (int i = update.arity()-3; i >=0; i-=2) {
	    if (update.sub(i).op() instanceof AttributeOp &&
		Op.ATTR.attribute(update.sub(i)) == attribute) {
		return update.sub(i+1);
	    }
	}       

	Term simplifiedReference = simplify(update, attr.sub(0));
	return (simplifiedReference == attr.sub(0) ? attr :
		tf.createAttributeTerm(simplifiedReference, attr.sub(1)));
    }


    /**
     * an update is applied on an instance attribute
     * @preconditions attribute.op().oclIsTypeOf(Attribute)    
     */
    private Term applyOnNoneStaticAttribute(Term update,
					    Term attr) {
	TermSymbol attribute = Op.ATTR.attribute(attr);
	// simplified reference: Be U={l1:=t1,...,ln:=tn}. U o.a --> then (U o)
	// is the simplified reference
	Term simplifiedReference = simplify(update, attr.sub(0));
	Term result = simplifiedReference == attr.sub(0) ? attr : 
	    tf.createAttributeTerm(simplifiedReference, attr.sub(1));
		
	// collect all attributes with same attribute variable as the given
	// attribute (lefthandside) and the terms they are updated with
	// (righthandside).
	for (int i = 0; i < update.arity()-1 ; i+=2) {
	    if (update.sub(i).op() instanceof AttributeOp) {
		if (update.sub(i).sub(1).op() == attribute) {
		    // %%% not so nice (may create useless things if o?=o occurs
		    // and deletes them afterwards
		    result = simplifiedReference.equals(update.sub(i).sub(0)) ?
			update.sub(i+1) :
			createIfElseTerm(new Term[]{
			    simplifiedReference,
			    update.sub(i).sub(0),
			    update.sub(i+1), result});
		}
	    }
	}

	return result;
    }

    /**
     * an update is applied on an instance attribute
     * @preconditions attribute.op().oclIsTypeOf(Attribute)    
     */
    protected Term applyOnAttribute(Term update,
				    Term attr) {

	// type cast should be save. SimultaneousUpdateSimplifier should not run
	// on terms containing schemavariables 
	ProgramVariable attribute = (ProgramVariable)Op.ATTR.attribute(attr);

	if (attribute.isStatic()) {
	    return applyOnStaticAttribute(update, attr);
	} else {
	    return applyOnNoneStaticAttribute(update, attr);
	}
    }


    

    /**
     * a list of updates is applied on an attribute
     * @preconditions var.op().oclIsTypeOf(ProgramVariable)
     */
    private Term applyOnAttribute(ListOfTerm updates,
				  Term attr) {
	Term result = attr;
	ListOfTerm unusedUpdates = updates;	
	while (unusedUpdates != SLListOfTerm.EMPTY_LIST && 
	       result.op() instanceof AttributeOp) {
	    result = applyOnAttribute(unusedUpdates.head(), result);
	    unusedUpdates = unusedUpdates.tail();
	}

	return unusedUpdates == SLListOfTerm.EMPTY_LIST ? 
	    result : simplify(unusedUpdates, result);
    }


    
    /**
     * creates a conjunction conditional out of a list of arrays
     * therefore the conditional part is split up into the conjunctions,
     * e.g. (a[0], a[j], a[4], a[0], 3, 4) becomes to
     * a*0 : a[0]?= (a * j:3 , a * 4 : 4 ) 
     */ 
    private Term createConjCond(List cond, List val) {
	Term result = (Term) val.get(0);
	Term arrayReference = ((Term)cond.get(0)).sub(0);
	Term arrayIndex = ((Term)cond.get(0)).sub(1);
	for (int i = 1; i<cond.size(); i++) {
	    Term candidate = (Term) cond.get(i);
	    result = createConjCondTerm
		(new Term[]{arrayReference, arrayIndex,
			    result, 
			    candidate.sub(0), candidate.sub(1), 
			    (Term)val.get(i)});
	}
        return result;
    }


    protected Term applyOnArray(ListOfTerm updates, Term array) {
	ListOfTerm updatesToApply = updates;
	Term result = array;
	while (updatesToApply != SLListOfTerm.EMPTY_LIST &&
	       result.op() == Op.ARRAY) {
	    result = applyOnArray(updatesToApply.head(), result);
	    updatesToApply = updatesToApply.tail();
	}
	return updatesToApply == SLListOfTerm.EMPTY_LIST ? result : 
	    simplify(updatesToApply, result);
    }


    protected Term applyOnArray(Term update, Term array) {
	// in the following comments term array is referenced as arr[I]

	// compute simplified references
	Term arrayReference = simplify(update, array.sub(0));
	Term arrayIndex     = simplify(update, array.sub(1));

	Term result = arrayIndex != array.sub(1) || arrayReference != array.sub(0) ?
	    tf.createArrayTerm(Op.ARRAY, arrayReference, arrayIndex) : array;


	// list of potential candidates
	List candidates = new ArrayList(update.arity()<<1);	

	// Collect potential location candidates. A potential candidate has the
	// form r_i[J_i], where r is compatible to arr
	Term location;
	for (int i = 0; i < update.arity() - 1; i += 2) {	  	    
	    location = update.sub(i);
	    if (location.op() == Op.ARRAY && 
		compatible(location.sub(0).sort(), arrayReference.sort())) {
		// add assignment pair (location, value)
		candidates.add(update.sub(i)); 
		candidates.add(update.sub(i+1));		    
	    }
	}

	// do we have candidates?

        if (candidates.size()==0) { // no, so we can leave here
	    return result;
	}

	// build conditional term; we have to iterate backwards according to the
	// last one wins semantic
	List conditionals = new ArrayList(1+candidates.size());
	List values       = new ArrayList(1+conditionals.size());

	Term candidateArray;
	Term value;	
	conditionals.add(result);
	values.add(result);
	for (int i = candidates.size()-2; i>=0; i-=2) {
	    candidateArray = (Term)candidates.get(i);
	    if (equalsWrtFormulaConstraint(arrayReference, candidateArray.sub(0)) && 
		equalsWrtFormulaConstraint(arrayIndex, candidateArray.sub(1))) {
		values.set(0, (Term)candidates.get(i+1));
		break; // we can stop collecting candidates here
	    } 
	    conditionals.add(candidateArray);
	    values.add((Term)candidates.get(i+1));
	}

	// create new array reference		       
	return createConjCond(conditionals, values);
    }
 
    // rules for complex term structure 


    /**
     * application of an update on a conditional (ifelse) term.
     * If possible the conditional term is simplified, too.
     */
    protected Term applyOnConditionalTerm(Term update, Term target) {
		
	boolean changed = false;
	Term[] subs = new Term[target.arity()];
	for (int i = 0; i<subs.length; i++) {
	    subs[i] = simplify(update, target.sub(i));
	    changed = changed || (subs[i] != target.sub(i));
	}
	if (changed) {
	    return createConditionalTerm(target.op(), subs);
	} else {
	    return target;
	}
    }

    /**
     * application of a list of updates on a conditional term. Therefore the
     * updates are succesively applied on the target term.
     */
    protected Term applyOnConditionalTerm(ListOfTerm updates, Term target) {
	IteratorOfTerm it = updates.iterator();
	ListOfTerm unusedUpdates = updates;
	Term result = target;

	while (unusedUpdates != SLListOfTerm.EMPTY_LIST && 
	       result.op() == Op.IF_ELSE) {	    
	    result = applyOnConditionalTerm(unusedUpdates.head(), result);
	    unusedUpdates = unusedUpdates.tail();
	}

	return unusedUpdates == SLListOfTerm.EMPTY_LIST ? result : 
	    simplify(unusedUpdates, result);
    }


    private Term createOperatorTerm(Term oldTerm, 
				    Term[] subs) {

	ArrayOfQuantifiableVariable[] vbh =
	    new ArrayOfQuantifiableVariable[oldTerm.arity()];
	
	for (int i = 0; i < oldTerm.arity(); i++) {
	    vbh[i] = oldTerm.varsBoundHere(i);
	}
	
	return tf.createTerm(oldTerm.op(), subs, vbh, 
			    JavaBlock.EMPTY_JAVABLOCK, null);
	
    }


    
    /**
     * application of a list of updates on a target term. updates are handed
     * over to the subterms.
     */
    protected Term applyOnOperatorTerm(ListOfTerm updates, Term target) {
	Term[] subs = new Term[target.arity()];
	boolean changed = false;
	for (int i = 0; i < target.arity(); i++) {
	    subs[i] = simplify(updates, target.sub(i));
 	    changed = changed || (subs[i] != target.sub(i));
	}
	return (changed ? createOperatorTerm(target, subs) : target);					
    }

    /**
     * concatenation of updates in front of a given term
     */
    private Term concatenate(ListOfTerm updates, Term target) {
	Term result = target;
	IteratorOfTerm it = updates.iterator();
	while ( it.hasNext() ) {		
	    Term upd = it.next();
  	    if (upd.sub(upd.arity()-1) != result) {
		Term[] subs = new Term[upd.arity()];
		    for (int i = 0; i < upd.arity()-1; i++) {
			subs[i] = upd.sub(i);
		    }
		    subs[subs.length-1] = result;
		    result = tf.createUpdateTerm(subs);
  	    } else {
  		result = upd;
  	    }
	}
	return result;
    }

    /**
     * application on a control term results in a concatenation in front 
     * of this term  (no deletion, no simplification beyond the term)
     */
    private Term applyOnControlTerm(ListOfTerm updates, Term target) {
	Term result = target;
	if (useSimultaneousUpdates) {
	    result = merge(updates, result);
	} else {
	    result = concatenate(updates, target);
	}
	return result;    	
    }


    /**
     * application of a list of updates on a program term. That is the point
     * where updates are merged and the deletion strategy is used.
     * @preconditions target.op().oclIsTypeOf(Modality)     
     */
    protected Term applyOnProgramTerm(ListOfTerm updates, 
				      Term target) {


	Term result = simplify(target.sub(0));
	result = (result != target.sub(0)) ?
	    tf.createProgramTerm((Modality)target.op(), target.javaBlock(), 
				 result) : target;
	if (useSimultaneousUpdates) {
	    result = merge(updates, result);
	} else {
	    result = concatenate(updates, target);
	}
	return delete(result);
    }
            
    /**
     * @preconditions target.op().oclIsTypeOf(SimultaneousUpdateOp)
     */
    protected Term applyOnSimultaneousUpdateTerm(ListOfTerm updates, 
					  Term target) {
	return simplify(updates.prepend(target), target.sub(target.arity()-1));
    }

    
    protected Term delete(Term t) {
	return deletionStrategy.copy().delete(t);
    }

    protected Term merge(ListOfTerm updates, Term target) {

	IteratorOfTerm updateIterator = updates.iterator();
	Term result = updateIterator.next();

	if (result == null) {
	    return target;
	} else if (updates.size() == 1 && 
	     result.sub(result.arity()-1) == target) {
	    return result;
	}

	Term update = null;
	while (updateIterator.hasNext()) {
	    update = updateIterator.next();
	    result = mergeSimultaneousUpdates(update, result);
	}
	
	Term[] subs = new Term[result == null ? 1 : result.arity()];

	for (int i = 0; i < subs.length-1; i++) {
	    subs[i] = result.sub(i);
	    
	}
	subs[subs.length-1] = target;

	return subs.length == 1 ? subs[0] : tf.createUpdateTerm(subs) ;
    }

    /**
     */
    protected Term applyUpdateOnLocation (Term update, Term loc) {
	Term result = loc;
	Term simplifiedReference;
	Term simplifiedIndex;
	if (loc.op() instanceof AccessOp) {
	    simplifiedReference = simplify(update, loc.sub(0));
	    if (loc.op() instanceof AttributeOp) {	    
		if (simplifiedReference != loc.sub(0)) {
		    result = tf.createAttributeTerm(simplifiedReference, 
						    loc.sub(1));	    
		}
	    } else {
		simplifiedIndex = simplify(update, loc.sub(1));
		if (simplifiedIndex != loc.sub(1) ||
		    simplifiedReference != loc.sub(0)) {
		    result = tf.createArrayTerm(Op.ARRAY, 
						simplifiedReference, 
						simplifiedIndex);
		}
	    }
	}
	return result;
    }


    /**
     * merging to simultaneous updates takes place here
     * @preconditions target.op().oclIsTypeOf(SimultaneousUpdateOp)
     */
    protected Term mergeSimultaneousUpdates(Term update, 
					    Term target) {

	/*
	  The simultaneouseous update in Term update be refered as 
	       U={l_1:=t_1,...,l_n:=t_n}, 
	  the simultaneouseous update in Term target be refered as 
	       U'={l'_1:=t'_1,...,l'_s:=t'_s}
	  U U' Phi ~~~~> {U*, l''_1:=U t'_1,...,U l''_s:= U t'_s} Phi
	  where U* = U\{l_i: ex j : l_i=l'_j} and l''_i = l'_i, 
          if l'i = program variable or  (U r'_i).a if (l'_i = r'_i.a) or
	  (Ur'_i)[Uj] if l'_i:=r'_i[j]  
	 */
	
	ListOfTerm subs = SLListOfTerm.EMPTY_LIST;

	HashSet pv_locs = new HashSet(target.arity());
	HashMap locsWithAccessors = new HashMap(target.arity());

	boolean changed = false;

	for (int i = 0; i < target.arity()-1; i += 2) {
	    
	    // compute U l'_i
	    Term loc_i = applyUpdateOnLocation(update, target.sub(i));
	    if (loc_i.op() instanceof ProgramVariable) {
		pv_locs.add(loc_i.op());
	    } else {
		Object accessorKey = (loc_i.op() == Op.ATTR ?
				      (Object) loc_i.sub(1) : (Object) loc_i.sub(0).sort());
		
		ListOfTerm sameAccessors = 
		    (ListOfTerm)locsWithAccessors.get(accessorKey);
		if (sameAccessors == null) { 
		    sameAccessors = SLListOfTerm.EMPTY_LIST;
		}
		locsWithAccessors.put(accessorKey, sameAccessors.prepend(loc_i));
	    }
	    // compute t'_i
	    Term val_i = simplify(update, target.sub(i+1));
	    changed = changed || (val_i != target.sub(i+1));
	    subs = subs.append(loc_i).append(val_i);
	}

	// compute U*
	ListOfTerm u_star = SLListOfTerm.EMPTY_LIST;

	for (int i = 0; i < update.arity()-1; i+=2) {

	    Term location = update.sub(i);
	    Term value    = update.sub(i+1);
	    if (location.op() instanceof ProgramVariable) {
		value = pv_locs.contains(location.op()) ? null : value;
	    } else {
		ListOfTerm potentialEqualLocations = SLListOfTerm.EMPTY_LIST;
		AccessOp accessOp = (AccessOp) location.op();
		if (accessOp instanceof AttributeOp) {
		    if (locsWithAccessors.containsKey(location.sub(1))) {
			potentialEqualLocations = 
			    (ListOfTerm)locsWithAccessors.get(location.sub(1));
		    }
		} else { // attribute operator
		    Sort referenceSort = location.sub(0).sort();
		    if (locsWithAccessors.containsKey(referenceSort)) {
			potentialEqualLocations = 
			    (ListOfTerm)locsWithAccessors.get(referenceSort);
		    }
		}

		IteratorOfTerm it = potentialEqualLocations.iterator();
		// reference, i.e. o in o.a and r in r[j]
		Term locationReference = location.sub(0);
		// accessor, i.e. a in o.a and j in r[j]		
		Term locationAccessor  = location.sub(1);
		
		while (it.hasNext()) {
		    Term locInTarget = it.next();		    
		    if (locInTarget.sub(0).equals(locationReference) && 
			(locInTarget.op() != Op.ARRAY || 
			 locInTarget.sub(1).equals(locationAccessor))) {
			value = null;
			break;				
		    } else if (locationReference.sort().extendsTrans
			       (locInTarget.sub(0).sort()) ||
			       locInTarget.sub(0).sort().extendsTrans
			       (locationReference.sort())) {
				// if sorts are in one line the terms may refer
				// to same objects 
			if (locInTarget.op() == Op.ARRAY) {
			    value = createConjCondTerm
				(new Term[]{ 
				    locationReference, locationAccessor, 
				    value,
				    locInTarget.sub(0), locInTarget.sub(1), 
				    location});
			} else {
			    value = createIfElseTerm(new Term[]{
				locationReference, locInTarget.sub(0), 
				location, value});
			}
		    } 
		}
	    }    	   
	    changed = changed || (value != null);
	    u_star = (value != null ? 
		      u_star.append(location).append(value) :
		      u_star);
	    
	} 
	Term t = changed ? tf.createUpdateTerm
	    (subs.append(target.sub(target.arity()-1)).prepend(u_star)) : target;
	return t;
    }

    
    protected Term applyOnMetavariable(ListOfTerm updates, Term target) {
	return target;
    }


    // static
    public static int depthWithoutUpdates(Term t) {
	if (t.arity()==0) return 0;
	if (t.op() instanceof SimultaneousUpdateOp) {
	    return depthWithoutUpdates(t.sub(1));
	}
	int m=0;
	for (int i=0; i<t.arity(); i++) {	   
	    int k=depthWithoutUpdates(t.sub(i))+1;
	    if (k>m) {
		m=k;
	    }
	}
	return m;
    }


    
    /**
     * 
     */
    protected Term simplify(ListOfTerm updates, Term target) {
	if (target.op() instanceof SimultaneousUpdateOp) {
	    return applyOnSimultaneousUpdateTerm(updates, target);
	} else if (target.op() instanceof Modality) {
	    return applyOnProgramTerm(updates, target);
	} else if (target.op() instanceof ProgramVariable) {
	    return applyOnProgramVariable(updates, target);
	} else if (target.op() == Op.ARRAY) {
	    return applyOnArray(updates, target);
	} else if (target.op() == Op.ATTR) {
	    return applyOnAttribute(updates, target);
	} else if (target.op() == Op.IF_ELSE) {
	    return applyOnConditionalTerm(updates, target);
	} else if (target.op() instanceof Metavariable) {
	    return applyOnMetavariable(updates, target);
	} else if (target.op() == Op.CTRL) {
	    return applyOnControlTerm(updates, target);
	} else if (target.op() instanceof TermSymbol || 
		   target.op() instanceof Op) {
	    return applyOnOperatorTerm(updates, target);
	} 
        /* Only print an error message if 'updates' is not empty.
         * e.g. the ASM-KeY version introduces a lot of new kinds of
         * terms, but doesn't use simultaneous updates at all, -hs. */
        if (updates.size() > 0) {
            Debug.assertTrue(false, "simultaneousupdatesimplifier: unrecognized term");
            Debug.out("simultaneousupdatesimplifier: unrecognized term:", target.op());
        }
        return target;
    }


    public Term simplify(Term update, Term target) {
	return simplify(SLListOfTerm.EMPTY_LIST.prepend(update), target);
    }

    /**
     * starts simplification
     */
    public Term simplify(Term target, int dummy) {
	return simplify(target);
    }

    /**
     * starts simplification
     */
    public Term simplify(Term target) {
	formulaConstraint = Constraint.BOTTOM;
	if (target.op() instanceof SimultaneousUpdateOp) {
	    return simplify(target, target.sub(target.arity()-1));
	}
	return simplify(SLListOfTerm.EMPTY_LIST, target);
    }

    /**
     * starts simplification for a contrained formula
     */
    public ConstrainedFormula simplify(ConstrainedFormula target, 
				       Constraint userConstraint) {
	
	Term result;
	Term targetTerm = target.formula();
	formulaConstraint = target.constraint();
	if (targetTerm.op() instanceof SimultaneousUpdateOp) {
	    result = simplify(targetTerm, targetTerm.sub(targetTerm.arity()-1));
	    
	} else {
	    result = simplify(SLListOfTerm.EMPTY_LIST, targetTerm);
	}
	return target.constraint()==formulaConstraint && 
	    result == targetTerm ? target : 
	    new ConstrainedFormula(result, formulaConstraint);
    }


    // some helpers
    /** returns true iff the two given sorts are compatible this means if one of
     * them is a subsort of the other one 
     * @param s1 first Sort to be tested for compatibility with the second sort s2
     * @param s2 second Sort to be tested for compatibility with the first sort s1
     * @return true iff s1 is subsort of s2 or vice versa
     */
    private boolean compatible(Sort s1, Sort s2) {
	return s1.extendsTrans(s2) || s2.extendsTrans(s1);
    }
    

    /**
     * compares both terms if they are equal with respect to the instantiations
     * of the formula (not user) constraint. This is done by unifying both terms
     * and returning true if the received constrained is subsumed by the formula
     * constraint
     * @param t1 the Term compared to t2
     * @param t2 the Term compared to t1
     * @return true if both terms are equal wrt to the formula constrained
     */
    private boolean equalsWrtFormulaConstraint(Term t1, Term t2) {
	return Constraint.BOTTOM.unify(t1, t2).isAsWeakAs(formulaConstraint);
    }


}
