// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.logic;

import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.util.Debug;
import java.util.*;

/** The represented term is a term with a simultaneous update as top level
 * operator. The arity of the term has to be odd, the last subterm is the
 * term/formula on which the update is applied to.
 * The syntax is similar to <code>{x_1:=t_1, ..., x_n:=t_n} Phi </code>.
 * 
 * Two simultaneous update terms u, u' are equal if Phi equals Phi' and if there
 * is a permutation perm such that x_perm(i)=x'_j and t_perm(i)=t'_j.
 */

class SimultaneousUpdateTerm extends OpTerm {


    /** creates a SimultaneousUpdateTerm
     * @param op the SimultaneousUpdateOperator
     * @param subs an array of Term. The last array element is the term the
     * update is applied to.
     */
    SimultaneousUpdateTerm(SimultaneousUpdateOp op, 
			   Term[] subs) {
	super(op, normalform(subs));
    }

    /** creates a SimultaneousUpdateTerm for the special case that only 
     * one variable is recorded by the update.
     * @param op the SimultaneousUpdateOperator
     * @param updatedVar the Term containing the reference to the variable whose
     * value is the one given by updateTerm
     * @param updateTerm a Term 
     * @param applicationTerm the Term the update is applied on
     */
    SimultaneousUpdateTerm(SimultaneousUpdateOp op, 
			   Term updateVar, 
			   Term updateTerm,
			   Term applicationTerm) {
	this(op, new Term[]{updateVar, updateTerm, applicationTerm});
    }  
    
    

    public static Term[] normalform(Term[] subs) {

	TreeMap pvLocs = new TreeMap(Classifier.DEFAULT_CLASSIFIER);
	TreeMap accessLocs = new TreeMap(Classifier.DEFAULT_CLASSIFIER);
	
	int size = 0;
	for (int i = 0; i < subs.length-1; i+=2) {
	    if (subs[i].op() instanceof ProgramVariable) {
		pvLocs.put(subs[i].op(), new Integer(i));
	    } else if (subs[i].op() instanceof SchemaVariable) {
		return subs;
	    } else {	
		if (subs[i].op() == Op.ATTR) {
		    TermSymbol attr = (TermSymbol) subs[i].sub(1).op();
		    if (attr instanceof SchemaVariable) {
			return subs;
		    } else {
			LinkedList list = (LinkedList) accessLocs.get(attr);
			if (list == null) {
			    list = new LinkedList();
			}
			list.add(new Integer(i));
			size+=2;
			accessLocs.put(attr, list);
		    }		
		} else {
		    Term ref = subs[i].sub(0);
		    if (ref.op() instanceof SchemaVariable) {
			return subs;
		    } else {
			LinkedList list = (LinkedList) accessLocs.get(ref.sort());
			if (list == null) {
			    list = new LinkedList();
			}
			list.add(new Integer(i));
			size+=2;
			accessLocs.put(ref.sort(), list);
		    }		
		    
		}
	    }
	}
	
	size += 2*pvLocs.size();
	
	Term[] result = new Term[size+1];
	int count = 0;
	Iterator it = pvLocs.values().iterator();
	while (it.hasNext()) {
	    int pos = ((Integer) it.next()).intValue();
	    result[count] = subs[pos];
	    result[count+1] = subs[pos+1];
	    count+=2;
	}

	Iterator attr_it = accessLocs.values().iterator();
	while (attr_it.hasNext()) {
	    Iterator positionIterator = ((LinkedList) attr_it.next()).iterator();
	    while (positionIterator.hasNext()) {
		int pos = ((Integer) positionIterator.next()).intValue();
		result[count] = subs[pos];
		result[count+1] = subs[pos+1];
		count+=2;
	    }
	}

	result[count] = subs[subs.length-1];

	//	for (int i = 0; i < result.length; i++) {
	    //	    System.out.println(""+result[i]);
	//	}
	//System.out.println(count+":size:"+size);	

	return result;
	
    }



    /** toString */
    public String toString() {
	StringBuffer result=new StringBuffer("{");
	for (int i=0; i<arity()-1; i++) {
	    if (i % 2 == 0 && i>0) {
		result.append(", ");
	    }
	    result.append(sub(i).toString());
	}
	result.append("}");
	result.append(sub(arity()-1).toString());
	return result.toString();
    }

    // inner class that sorts the updates in a safe way (kind of normal form)
    static class Classifier implements Comparator{

	final static Classifier DEFAULT_CLASSIFIER = new Classifier();

	private int primitiveLongCompare(long i, long j) {
	    return (i < j ? -1 : (i == j ? 0 : 1));
	}

	public int compare(Object o1, Object o2) {	    
	    if (o1 == o2) return 0;
	    Named pv1 = (Named) o1;
	    Named pv2 = (Named) o2;
	    int cmpResult= pv1.name().toString().
		compareTo(""+pv2.name());
	    if (cmpResult == 0) {
		boolean bPV1 = (pv1 instanceof ProgramVariable);
		boolean bPV2 = (pv2 instanceof ProgramVariable);
		cmpResult = bPV1 ? ( bPV2 ? primitiveLongCompare
				     (((ProgramVariable)pv1).id(), 
				      ((ProgramVariable)pv2).id()) : -1)
		    : ( bPV2 ? 1 : 0);
	    }	    
	    return cmpResult;
	}
    }
}
