// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.logic;

import junit.framework.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.StatementBlock;
import de.uka.ilkd.key.java.abstraction.PrimitiveType;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.declaration.ClassDeclaration;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.rule.*;
import java.io.*;

public class TestSimultaneousUpdateSimplifier extends TestCase {

    private Namespace variables = new Namespace();
    private Namespace functions = new Namespace();
    private Namespace sorts     = new Namespace();
    
    private ProgramVariable[] pv;
    private ProgramVariable spv;

    private TermFactory tf = TermFactory.DEFAULT;
	

    public TestSimultaneousUpdateSimplifier(String s) {
	super(s);
    }

    public void setUp() {
	Sort sort=new ClassInstanceSortImpl(new Name("testSort1"));
	KeYJavaType kjt=new KeYJavaType(ClassDeclaration.objectClassDecl, sort);
	sorts.add(sort);

	pv = new ProgramVariable[7];
	for (int i=0; i<pv.length; i++) {	    
	    pv[i]=new ProgramVariable(new ProgramElementName("pv"+i), kjt);
	    variables.add(pv[i]);	    
	}
	
	spv = new ProgramVariable(new ProgramElementName("spv"), kjt, true);
    }


    private Term parseTerm(String termstr) {
	return TacletForTests.parseTerm(termstr, 
					new NamespaceSet
					    (variables, functions, 
					     sorts, new Namespace()));
// 	JavaInfo ji = new JavaInfo();	
// 	if (termstr.equals("")) return null;
// 	try {	    
// 	    StringReader br=new StringReader(termstr);
// 	    TermParser parser = new TermParser(new KeYLexer(br), 
// 					       "No file. " +
// 					       "TacletForTests.parseTerm("
// 					       +termstr+")",
// 					       TermFactory.DEFAULT, 
// 					       ji,variables,
// 					       functions, sorts, new de.uka.ilkd.key.gui.AbbrevMap());
// 	    return parser.term();
// 	} catch (Exception e) {
// 	    System.err.println("Exception during parsing!");
// 	    e.printStackTrace();
// 	    return null;
// 	}
    }
    
    public void testBasicRules() {

	SimultaneousUpdateSimplifier simply=new SimultaneousUpdateSimplifier
	    (new DefaultDeletionStrategy(false, false), true);

	Term parsed = parseTerm("{pv1:=pv2} pv3");
	assertTrue(simply.simplify(parsed) == parsed.sub(parsed.arity()-1));

	parsed=parseTerm("{pv1:=pv2} pv1");
	assertTrue(simply.simplify(parsed)==parsed.sub(parsed.arity()-2));
	

	parsed=parseTerm("{pv1:=pv2, pv2:=pv3} pv1");
	assertTrue(simply.simplify(parsed).op()==pv[2]);
	

	Term[] subs=new Term[parsed.arity()];
	for (int i=0;i<parsed.arity()-1;i++) {
	    subs[i]=parsed.sub(i);
	}
	subs[parsed.arity()-1]=tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[4]));
		    
	// {pv1:=pv2, pv2:=pv3} pv4.pv5 ~~~> pv4.pv5
	Term constr=tf.createUpdateTerm(subs);
	Term simplified=simply.simplify(constr);
	assertTrue(simplified==constr.sub(constr.arity()-1));
	

	// {pv2:=pv3, pv1.pv5:=pv2} pv4.pv5 ~~~>(pv4?=pv1).pv5/->pv2
	subs[0] = tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[1]));
	constr  = tf.createUpdateTerm(subs);
	simplified = simply.simplify(constr);
	assertTrue("If-Else Operator with attribute pv5 in 'then' part expected, but is: "+
	       simplified, simplified.op() == Op.IF_ELSE &&
	       simplified.sub(3).sub(1).op() == pv[5] &&
	       simplified.sub(3).sub(0).op() == pv[4]);
	assertTrue("Expected subterms pv4, pv1, pv2, pv4.pv5. But is:"+
	       simplified.sub(0).op()+", "+simplified.sub(1).op()+", "+
	       simplified.sub(2)+", "+simplified.sub(3), 
	       simplified.sub(0).op() == pv[4] &&
	       simplified.sub(1).op() == pv[1] && 
	       simplified.sub(2).op() == pv[2] &&
	       simplified.sub(3).sub(1).op() == pv[5] && 
	       simplified.sub(3).sub(0).op() == pv[4]);

	// {pv1.pv5:=pv2, pv2.pv5:=pv3} pv4.pv5 ~~~>(pv4?=pv2) pv3 : ((pv4?=pv1).pv5/->pv2)
	subs[0] = tf.createAttributeTerm(pv[5],  tf.createFunctionTerm(pv[1]));
	subs[2] = tf.createAttributeTerm(pv[5],  tf.createFunctionTerm(pv[2]));
	constr  = tf.createUpdateTerm(subs);
	simplified = simply.simplify(constr);


	String expectedString=
	    simplified.sub(0) + ", " +
	    simplified.sub(1) + ", " +
	    simplified.sub(2) + ", " +
	    simplified.sub(3).sub(0) + ", " +
	    simplified.sub(3).sub(1) + ", " +
	    simplified.sub(3).sub(2) + ", " +
	    simplified.sub(3).sub(3);
 	assertTrue("Expected subterms pv[4, 2, 3, 4, 1, 2, 4.5]. But is:"+
 	       expectedString, 
	       simplified.op() == Op.IF_ELSE &&
	       simplified.sub(3).op() == Op.IF_ELSE &&
 	       simplified.sub(0).op()==pv[4] &&
 	       simplified.sub(1).op()==pv[2] && 
	       simplified.sub(2).op()==pv[3] &&

 	       simplified.sub(3).sub(0).op() == pv[4] &&
 	       simplified.sub(3).sub(1).op() == pv[1] && 
	       simplified.sub(3).sub(2).op() == pv[2] &&
 	       simplified.sub(3).sub(3).sub(0).op() == pv[4] && 
 	       simplified.sub(3).sub(3).sub(1).op() == pv[5]);

	// {pv1:=pv2} {pv2:=pv3, pv3:=pv1}<>true ~~> {pv1:=pv2, pv2:=pv3, pv3:=pv2}<>true
	parsed=parseTerm("{pv1:=pv2} {pv2:=pv3, pv3:=pv1} <{}>true");
	Term expected=parseTerm("{pv1:=pv2, pv2:=pv3, pv3:=pv2} <{}>true");
	simplified = simply.simplify(parsed);
	assertTrue("Expected {pv1:=pv2, pv2:=pv3, pv3:=pv2} <> true. But is:"+simplified,
	       simplified.equals(expected));


	// {pv1:=pv2} {pv1:=pv3, pv3:=pv1}<>true ~~> {pv1:=pv3, pv3:=pv2}<>true
	parsed = parseTerm("{pv1:=pv2} {pv1:=pv3, pv3:=pv1} <{}>true");
	expected = parseTerm("{pv1:=pv3, pv3:=pv2} <{}>true");
	simplified = simply.simplify(parsed);
	assertTrue("Expected {pv1:=pv3, pv3:=pv2}<>true. But is:"+simplified,
	       simplified.equals(expected));

	
	// {pv2.pv5:=pv1, pv1.pv5:=pv2, pv4.pv5:=pv3} pv1.pv5 = pv1 ~~> 
	// ((pv1 ?= pv4) pv3 : pv2) = pv1
	subs = new Term[7];
	subs[0] = tf.createAttributeTerm(pv[5], 
					 tf.createFunctionTerm(pv[2])); 
	subs[1] = tf.createFunctionTerm(pv[1]); 
	subs[2] = tf.createAttributeTerm(pv[5], 
					 tf.createFunctionTerm(pv[1])); 
	subs[3] = tf.createFunctionTerm(pv[2]); 
	subs[4] = tf.createAttributeTerm(pv[5], 
					 tf.createFunctionTerm(pv[4])); 
	subs[5] = tf.createFunctionTerm(pv[3]); 
	subs[6] = tf.createEqualityTerm
	    (tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[1])),
	     tf.createFunctionTerm(pv[1]));

	constr  = tf.createUpdateTerm(subs);
        expected = parseTerm("(pv1 ?= pv4) pv3 : pv2");

	Term pv35 = tf.createAttributeTerm(pv[5], 
					   tf.createFunctionTerm(pv[3]));
	Term[] e_subs = new Term[expected.arity()];
	for (int i = 0; i < expected.arity(); i++) {
	    e_subs[i] = expected.sub(i);
	}	

	expected = tf.createEqualityTerm
	    (tf.createIfElseTerm(Op.IF_ELSE, e_subs),
	     tf.createFunctionTerm(pv[1]));
					 

	simplified = simply.simplify(constr);


	assertTrue("Expected:" + expected + ", but is:" + simplified,
		   simplified.equals(expected));
	

    }


    public void testApplyOnAttribute() {
	SimultaneousUpdateSimplifier simply=new SimultaneousUpdateSimplifier
	    (new DefaultDeletionStrategy(false, false), true);
	Term parsed = parseTerm("{pv1:=pv6, pv2:=pv2} pv1");

	Term[] subs=new Term[parsed.arity()];
	for (int i=0;i<parsed.arity()-1;i++) {
	    subs[i]=parsed.sub(i);
	}
	subs[parsed.arity()-1]=tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[4]));
		    
	// none-static
	// {pv3.pv5:=pv6, pv1.pv5:=pv2} pv4.pv5 ~~~>(pv4?=pv1) pv2 : (pv4 ?= pv3) pv6 : pv4.pv5
	subs[0] = tf.createAttributeTerm(pv[5], 
					 tf.createFunctionTerm(pv[3]));
	subs[2] = tf.createAttributeTerm(pv[5], 
					 tf.createFunctionTerm(pv[1]));
	Term constr  = tf.createUpdateTerm(subs);
	Term simplified = simply.simplify(constr);


	assertTrue("If-Else Operator expected, but is "+simplified.op(), 
		   simplified.op() == Op.IF_ELSE);

	assertTrue("Expected pv4, but is: "+simplified.sub(0).op(),
		   simplified.sub(0).op()==pv[4]);

	assertTrue("Expected pv1, but is: "+simplified.sub(1).op(),
		   simplified.sub(1).op()==pv[1]);

	assertTrue("Expected pv2, but is: "+simplified.sub(2).op(),
		   simplified.sub(2).op()==pv[2]);

	assertTrue("If-Else Operator expected, but is "+simplified.sub(3).op(), 
		   simplified.sub(3).op() == Op.IF_ELSE);
	
	assertTrue("Expected pv4, but is: "+simplified.sub(3).sub(0).op(),
		   simplified.sub(3).sub(0).op()==pv[4]);

	assertTrue("Expected pv3, but is: "+simplified.sub(3).sub(1).op(),
		   simplified.sub(3).sub(1).op()==pv[3]);

	assertTrue("Expected pv6, but is: "+simplified.sub(3).sub(2).op(),
		   simplified.sub(3).sub(2).op()==pv[6]);

	assertTrue("Expected pv4.pv5, but is: "+simplified.sub(3).sub(3),
		   simplified.sub(3).sub(3).op() instanceof AttributeOp &&
		   Op.ATTR.attribute(simplified.sub(3).sub(3)) == pv[5] &&
		   simplified.sub(3).sub(3).sub(0).op()==pv[4]);

	// static
	// {pv3.spv:=pv6, pv1.spv:=pv2} pv4.spv ~~~>pv2
	subs[0] = tf.createAttributeTerm(spv, 
					 tf.createFunctionTerm(pv[3]));
	subs[2] = tf.createAttributeTerm(spv, 
					 tf.createFunctionTerm(pv[1]));
	subs[parsed.arity()-1]=tf.createAttributeTerm(spv, tf.createFunctionTerm(pv[4]));

	constr  = tf.createUpdateTerm(subs);
	simplified = simply.simplify(constr);


	assertTrue("Expected pv2, but is "+simplified.op(), 
		   simplified.op() == pv[2]);



    }
    


    public void testDeletionStrategy() {
	DefaultDeletionStrategy strat = new DefaultDeletionStrategy();
	
	Term parsed = parseTerm("{pv1:=pv2} pv3");
	Term expected = parseTerm("pv3");
	Term result = strat.delete(parsed);
	assertTrue("Expected:" + expected + "\n Is:" + result, 
	       result == parsed.sub(parsed.arity()-1));

	strat = new DefaultDeletionStrategy();
	parsed = parseTerm("{pv1:=pv2, pv3:=pv5} pv1");	
	result = strat.delete(parsed);
	expected = parseTerm("{pv1:=pv2} pv1");
	assertTrue("Expected: " + expected + "\n Is: " + result, 
	       result.equals(expected));


	strat = new DefaultDeletionStrategy();
	parsed = parseTerm("{pv1:=pv2, pv3:=pv5, pv4:=pv4} pv1");	
	result = strat.delete(parsed);
	expected = parseTerm("{pv1:=pv2} pv1");
	assertTrue("Expected: " + expected + "\n Is: " + result, 
	       result.equals(expected));


	strat = new DefaultDeletionStrategy();
	parsed = parseTerm("{pv1:=pv2, pv3:=pv5, pv4:=pv4} pv4");	
	result = strat.delete(parsed);
	expected = parseTerm("pv4");
	assertTrue("Expected: " + expected + "\n Is: " + result, 
	       result.equals(expected));

	strat = new DefaultDeletionStrategy();
	parsed = parseTerm("{pv1:=pv2, pv3:=pv5, pv4:=pv4} pv1");	

        // {pv1.pv5:=pv2, pv3:=pv5, pv4:=pv4} pv1
	Term[] subs=new Term[parsed.arity()];
	for (int i = 0; i < parsed.arity(); i++) {
	    subs[i]=parsed.sub(i);
	}
	subs[0] = tf.createAttributeTerm(pv[5],  tf.createFunctionTerm(pv[1]));
	Term constr = tf.createUpdateTerm(subs);
	result = strat.delete(constr);


	assertTrue("Expected: {pv1.pv5:=pv2} pv1 \n Is: " + result, 
	       result.sub(0) == subs[0] && 
	       result.sub(1).op() == pv[2] &&
	       result.sub(2).op() == pv[1]);

	strat = new DefaultDeletionStrategy();
	parsed = parseTerm("{pv1:=pv2, pv3:=pv5, pv4:=pv4}<{ pv3 = pv3; }> pv1=pv1");	
 	result = strat.delete(parsed);
 	assertTrue("Expected: {pv1:=pv2, pv3:=pv5} <{ pv3=pv3; }> pv1=pv1 \n Is: " + result, 
 	       result.sub(0).op() == pv[1] && 
 	       result.sub(1).op() == pv[2] &&
 	       result.sub(2).op() == pv[3] &&
	       result.sub(3).op() == pv[5] &&
	       result.sub(4) == parsed.sub(6));

	strat = new DefaultDeletionStrategy();
	parsed = parseTerm("{pv1:=pv2, pv3:=pv5, pv4:=pv3} pv1=pv4");	
 	result = strat.delete(parsed);
 	assertTrue("Expected: {pv1:=pv2, pv4:=pv3} pv1=pv4 \n Is: " + result, 
 	       result.sub(0).op() == pv[1] && 
 	       result.sub(1).op() == pv[2] &&
 	       result.sub(2).op() == pv[4] &&
	       result.sub(3).op() == pv[3] &&
	       result.sub(4) == parsed.sub(6));
	
    }

    public void testSimultaneousUpdateEquality() {
	Term cmp1 = parseTerm("{pv1:=pv2, pv3:=pv5, pv4:=pv3} true");	
	Term cmp2 = parseTerm("{pv3:=pv5, pv4:=pv3, pv1:=pv2} true");	

	assertTrue("ProgramVariables commute.", cmp1.equals(cmp2));


	cmp1 = parseTerm("{pv1:=pv2, pv5:=pv2, pv2:=pv3,"+
			 " pv4:=pv2, pv6:=pv2} true");	

	Term[] subs = new Term[cmp1.arity()];
	for (int i = 0; i < cmp1.arity(); i++) {
	    subs[i] = cmp1.sub(i);
	}

	// {pv1:=pv2, pv3.pv5:=pv2, pv2:=pv3, 
	//  pv1.pv5:=pv2, pv3.pv6:=pv2} pv4.pv5}
	subs[2] = tf.createAttributeTerm
	    (pv[5], tf.createFunctionTerm(pv[3]));
	subs[6] = tf.createAttributeTerm
	    (pv[5],  tf.createFunctionTerm(pv[1]));
	subs[8] = tf.createAttributeTerm
	    (pv[6],  tf.createFunctionTerm(pv[3]));

	
	cmp1 = tf.createUpdateTerm(subs);       	

	assertTrue("Normalform defect: "+cmp1, 
		   cmp1.sub(4).sub(0).op() == pv[3] &&
		   cmp1.sub(6).sub(0).op() == pv[1] && 
		   cmp1.sub(8).sub(0).op() == pv[3]);


	Term[] subs2 = new Term[cmp1.arity()];
	subs2[0] = subs[2];
	subs2[1] = subs[3];
	subs2[2] = subs[4];
	subs2[3] = subs[5];
	subs2[4] = subs[6];
	subs2[5] = subs[7];
	subs2[6] = subs[0];
	subs2[7] = subs[1];
	subs2[8] = subs[8];
	subs2[9] = subs[9];
	subs2[10] = subs[10];
	cmp2 = tf.createUpdateTerm(subs2);
	
	assertTrue("Normalform defect.", cmp2.equals(cmp1));

	subs2 = new Term[cmp1.arity()];
	subs2[0] = subs[0];
	subs2[1] = subs[1];
	subs2[2] = subs[8];
	subs2[3] = subs[9];
	subs2[4] = subs[6];
	subs2[5] = subs[7];
	subs2[6] = subs[4];
	subs2[7] = subs[5];
	subs2[8] = subs[2];
	subs2[9] = subs[3];
	subs2[10] = subs[10];
	cmp2 = tf.createUpdateTerm(subs2);

	assertTrue("Normalform defect.", !cmp2.equals(cmp1));
	


    }


    public void testApplicationOnAttributeNoneSim() {

	SimultaneousUpdateSimplifier simply=new SimultaneousUpdateSimplifier
	    (new DefaultDeletionStrategy(false, false), false);

	Term loc1 = tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[4]));
	Term loc2 = tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[3]));
	Term val  = tf.createFunctionTerm(pv[2]);

	// {pv4.pv5 := pv2} {p3.pv5 := pv2} (p4.p5 = p3.pv5)

	Term constr     = tf.createUpdateTerm
	    (new Term[]{loc1, val,
	     tf.createUpdateTerm
	    (new Term[]{loc2, val, tf.createEqualityTerm(loc2, loc1)})});
	Term simplified = simply.simplify(constr);
	Term expected   = tf.createEqualityTerm(val, val);						
	assertEquals("Error applying non-simultaneous updates on attributes.", 
		     expected, simplified);
    }

    public void testApplicationOnAttributeSim() {

	SimultaneousUpdateSimplifier simply=new SimultaneousUpdateSimplifier
	    (new DefaultDeletionStrategy(false, false), true);

	Term loc1 = tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[4]));
	Term loc2 = tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[3]));
	Term val  = tf.createFunctionTerm(pv[2]);

	// {pv4.pv5 := pv2} {p3.pv5 := pv2} (p4.p5 = p3.pv5)

	Term constr     = tf.createUpdateTerm
	    (new Term[]{loc1, val, loc2, val, tf.createEqualityTerm(loc2, loc1)});
	Term simplified = simply.simplify(constr);
	Term expected   = tf.createEqualityTerm(val, val);						
	assertEquals("Error applying simultaneous update on attributes.", 
		     expected, simplified);
    }

    public void testBugInUStarComputation() {
	SimultaneousUpdateSimplifier simply=new SimultaneousUpdateSimplifier
	    (new DefaultDeletionStrategy(false, true), true);

	Term p4p5 = tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[4]));
	Term p1   = tf.createFunctionTerm(pv[1]);
	Term p2   = tf.createFunctionTerm(pv[2]);

	// {pv1:=pv4.pv5, pv4.pv5 := pv2} {pv4.pv5 := pv1} <>(pv1=pv2)

	Term constr     = tf.createUpdateTerm
	    (new Term[]{p1, p4p5, p4p5, p2,  
			tf.createUpdateTerm
			(new Term[]{p4p5, p1,
			tf.createDiamondTerm(JavaBlock.createJavaBlock(new StatementBlock()), 
					     tf.createEqualityTerm(p2, p2))})});
	Term simplified = simply.simplify(constr);
	Term expected   = tf.createUpdateTerm
	    (new Term[]{p4p5, p4p5,
			tf.createDiamondTerm(JavaBlock.createJavaBlock(new StatementBlock()), 
					     tf.createEqualityTerm(p2, p2))});
	assertEquals("Error when merging updates.", expected, simplified);	
    }

    public void testBugInDeleteTrivialUpdates() {
	// deletion of updates has been wrong for the folowing case:
	//   {o1.a:=c; o2.a:=o2.a} phi
        // previously o2.a:=o2.a has been deleted, but that is wrong, if o1=o2

	SimultaneousUpdateSimplifier simply=new SimultaneousUpdateSimplifier
	    (new DefaultDeletionStrategy(true, false), true);

	Term p3p5 = tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[3]));
	Term p4p5 = tf.createAttributeTerm(pv[5], tf.createFunctionTerm(pv[4]));
	Term p1   = tf.createFunctionTerm(pv[1]);
	Term p2   = tf.createFunctionTerm(pv[2]);

	// {pv3.pv5:=pv1, pv4.pv5 := pv4.pv5} <>(pv1=pv2)

	Term constr     = tf.createUpdateTerm
	    (new Term[]{p3p5, p1, p4p5, p4p5,  
			tf.createDiamondTerm(JavaBlock.createJavaBlock(new StatementBlock()), 
					     tf.createEqualityTerm(p2, p2))});

	Term simplified = simply.simplify(constr);
	Term expected   = constr;
	assertEquals("Trivial updates may only be deleted if it is safe.", expected, simplified);	
    }
    


    public static void main(String[] args) {
	TestSimultaneousUpdateSimplifier tsus=new TestSimultaneousUpdateSimplifier("t");
	tsus.setUp();
	tsus.testBasicRules();
	tsus.testDeletionStrategy();
	tsus.testSimultaneousUpdateEquality();
	tsus.testApplyOnAttribute();
    }

   
}
