// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.logic.ldt;

import de.uka.ilkd.key.java.abstraction.Type;
import de.uka.ilkd.key.java.abstraction.ListOfType;
import de.uka.ilkd.key.java.expression.Operator;
import de.uka.ilkd.key.java.expression.operator.*;
import de.uka.ilkd.key.java.expression.Literal;
import de.uka.ilkd.key.java.expression.literal.IntLiteral;
import de.uka.ilkd.key.java.expression.literal.LongLiteral;
import de.uka.ilkd.key.util.Debug;
import java.io.File;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.proof.RuleSource;

/** This class inherits from LDT and implements all method that are
 * necessary to handle integers, shorts and bytes.
 * It loads the needed rule file and offers method to convert java
 * number types to their logic counterpart
 */
public class ArithIntLDT extends NumberLDT {
    
    public ArithIntLDT() {
	super(new Name("arithmeticInteger"));
        ruleSource = RuleSource.initRuleFile("intRulesArith.key");
    }


    /** returns the name of the target sort of the LDT 
     * is needed to find the right Sort when constructing the LDT
     * @return the Name of the target sort
     */
    public Name getSortName() {
	return targetSort() == null ? new Name("arithInt") : targetSort().name();
    }

    public void init(Namespace func_ns, Sort targetSort, ListOfType type) {
	super.init(func_ns, targetSort, type);	 

 	Namespace functions = functions();

	// the operators
	greaterThan     = (Function)functions.lookup(new Name("jgt"));
	greaterOrEquals = (Function)functions.lookup(new Name("jgeq"));
	lessThan        = (Function)functions.lookup(new Name("jlt"));
	lessOrEquals    = (Function)functions.lookup(new Name("jleq"));
	equals          = (Function)functions.lookup(new Name("eq"));
	plus            = (Function)functions.lookup(new Name("+"));
	minus           = (Function)functions.lookup(new Name("~d"));
	negative        = (Function)functions.lookup(new Name("~m"));
	negativeNumber  = (Function)functions.lookup(new Name("neg"));
	times           = (Function)functions.lookup(new Name("mul"));
	divide          = (Function)functions.lookup(new Name("div"));
	modulo          = (Function)functions.lookup(new Name("modulo"));	
    }

    
    /** returns true if the LDT offers an operation for the given java
     * operator and the logic subterms 
     * @param op the de.uka.ilkd.key.java.expression.Operator to
     * translate
     * @param subs the logic subterms of the java operator
     * @return  true if the LDT offers an operation for the given java
     * operator and the subterms 
     */
    public boolean isResponsible
	(Operator op, Term[] subs) {
	if (subs.length == 1) {
	    return isResponsible(op, subs[0]);
	} else if (subs.length == 2) {
	    return isResponsible(op, subs[0], subs[1]);	
	}
	return false;	
    }

    /** returns true if the LDT offers an operation for the given
     * binary java operator and the logic subterms 
     * @param op the de.uka.ilkd.key.java.expression.Operator to
     * translate
     * @param left the left subterm of the java operator
     * @param right the right subterm of the java operator
     * @return  true if the LDT offers an operation for the given java
     * operator and the subterms 
     */
    public boolean isResponsible
	(Operator op, Term left, Term right) {
	if (left.sort() == targetSort() && right.sort() == targetSort()) {
	    if (getFunctionFor(op) != null) {
		return true;
	    }
	}
	return false;
    }

    /** returns true if the LDT offers an operation for the given
     * unary java operator and the logic subterms 
     * @param op the de.uka.ilkd.key.java.expression.Operator to
     * translate
     * @param sub the logic subterms of the java operator
     * @return  true if the LDT offers an operation for the given java
     * operator and the subterm
     */
    public boolean isResponsible
	(Operator op, Term sub) {
	if (sub.sort() == targetSort()) {
	    if (op instanceof Negative) {
		return true;
	    }
	}
	return false;
    }


} 
