// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.java.reference.FieldReference;
import de.uka.ilkd.key.java.reference.ReferencePrefix;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.logic.ProgramInLogic;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.ldt.*;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.gui.ProofSettings;

/** this class is used to represent the reference operator '.' of
 * Java, e.g.>> int[] a = new int[]{1,2,3};  int i = a.length; <<<
 * where the '.' in a.length points to field called length
 * ATTENTION: At this moment only variables are adressed to this may
 * change in future
 */
public class AttributeOp extends AccessOp {

    /** creates a new attribute operator.
     */
    AttributeOp() {
	super(new Name("."));
    }


    /**
     * checks whether the top level structure of the given {@link Term}
     * is syntactically valid, given the assumption that the top level
     * operator of the term is the same as this Operator. The
     * assumption that the top level operator and the term are equal
     * is NOT checked.  
     * @return true iff the top level structure of
     * the @link Term is valid.
     */
    public boolean validTopLevel(Term term) {
	Operator attribute = term.sub(0).op();
	if (attribute instanceof SchemaVariable || 
            term.sub(1).op() instanceof SchemaVariable) {
	    return true;
	}
	if (term.arity() != arity()) {
	    return false;
	}
	if (!(term.sub(0).sort() instanceof ObjectSort)
	    && !(term.sub(0).sort()==Sort.TYPE)
	    && !((term.sub(0).sort() instanceof ArraySort)
	    && ((ProgramVariable)term.sub(1).op()).name().toString().equals("length"))){
	    return false;
	}
	return true;
    }

    /** returns the attribute of the operator 
     * @return  the attribute of the operator 
     */
    public TermSymbol attribute(Term t) {
	return (TermSymbol)t.sub(1).op();
    }

    public KeYJavaType getKeYJavaType(Term t) {
	return ((ProgramVariable)attribute(t)).getKeYJavaType();
    }

    public Expression convertToProgram(Term t, de.uka.ilkd.key.util.ExtList l) {
	return new FieldReference((ProgramVariable)attribute(t), 
				  (ReferencePrefix)l.get(0));
    }


    /**
     * determines the sort of the {@link Term} if it would be created using this
     * Operator as top level operator and the given terms as sub terms. The
     * assumption that the constructed term would be allowed is not checked.
     * @param term an array of Term containing the subterms of a (potential)
     * term with this operator as top level operator
     * @return sort of the term with this operator as top level operator of the
     * given substerms
     */
    public Sort sort(Term[] term) {
	return term[1].sort();
    }

    /** returns the name of the operator 
     * @return the name of the operator */
    public Name name() {
	return name;
    }


    /** toString */
    public String toString() {
	return ".";
    }

    public boolean equals(Object o) {
	return (o instanceof AttributeOp);
    }

    public int hashCode() {
	return AttributeOp.class.hashCode();
    }
    
} 
