// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/** This file represents a Java method in the logic. It is part of the
 * AST of a java program 
 */
package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.reference.MethodReference;
import de.uka.ilkd.key.java.reference.ReferencePrefix;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.logic.sort.Sort;
import java.io.IOException;
import de.uka.ilkd.key.java.visitor.Visitor;

public class ProgramMethod extends TermSymbol 
    implements SourceElement, ProgramElement, MemberDeclaration, ProgramInLogic{

    private final MethodDeclaration method; 
    private final KeYJavaType kjt;
    private final KeYJavaType contKJT;

    public ProgramMethod(MethodDeclaration method, 
			 KeYJavaType contKJT, 
			 KeYJavaType kjt) {

	super(method.getProgramElementName(), null);
	this.method  = method;
	this.contKJT = contKJT;
	this.kjt     = kjt;
    }

   /** @return arity of the method as int.  This is one more than
    * the number of Java parameters because of the this pointer.
    */
    public int arity() {
	return method.getParameterDeclarationCount()+1;
    }

    /** @return Sort of the TermSymbol */
    public Sort sort() {
	return kjt.getSort();
    }

    /** @return Sort of the TermSymbol */
    public Sort sort(Term[] term) {
	return sort();
    }

    /** 
     * @return true iff number of subterms of term is equal to its own arity
     */
    public boolean validTopLevel(Term term){
        return term.arity()==this.arity(); //%%% needs more checking!!!!
    }

    // java related methods
    
    public MethodDeclaration getMethodDeclaration() {
	return method;
    }

    public StatementBlock getBody() {
        return getMethodDeclaration().getBody();
    }

    /** toString */
    public String toString() {
	return name().toString();
    }

    public SourceElement getFirstElement(){
	return method.getFirstElement();
    }

    public SourceElement getLastElement(){
	return method.getLastElement();
    }

    public Comment[] getComments() {
	return null;
    }

    public void prettyPrint(PrettyPrinter w) throws IOException {
	method.prettyPrint(w);
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnProgramMethod(this);
    }

    /**
 *        Returns the start position of the primary token of this element.
 *        To get the start position of the syntactical first token,
 *        call the corresponding method of <CODE>getFirstElement()</CODE>.
 *        @return the start position of the primary token.
    */
    public Position getStartPosition(){
	return Position.UNDEFINED;
    }

    /**
 *        Returns the end position of the primary token of this element.
 *        To get the end position of the syntactical first token,
 *        call the corresponding method of <CODE>getLastElement()</CODE>.
 *        @return the end position of the primary token.
     */
    public Position getEndPosition(){
	return Position.UNDEFINED;
    }

    /**
 *        Returns the relative position (number of blank heading lines and 
 *        columns) of the primary token of this element.
 *        To get the relative position of the syntactical first token,
 *        call the corresponding method of <CODE>getFirstElement()</CODE>.
 *        @return the relative position of the primary token.
     */
    public Position getRelativePosition(){
	return  Position.UNDEFINED;
    }
    
    /**
     * Test whether the declaration is private.
     */
    public boolean isPrivate(){
	return method.isPrivate();
    }

    /**
     * Test whether the declaration is protected.
     */
    public boolean isProtected(){
	return method.isProtected();
    }

    /**
     * Test whether the declaration is public.
     */
    public boolean isPublic(){
	return method.isPublic();
    }

    /**
     * Test whether the declaration is static.
     */
    public boolean isStatic(){
	return method.isStatic();
    }

    /**
     * Test whether the declaration is strictfp.
     */
    public boolean isStrictFp(){
	return method.isStrictFp();
    }
    
    public ArrayOfModifier getModifiers(){
	return method.getModifiers();
    }

    public int getChildCount() {
	return method.getChildCount();
    }

    public ProgramElement getChildAt(int i){
	return method.getChildAt(i);
    }
    public int getIndexOfChild(ProgramElement pe) {
	return method.getIndexOfChild(pe);
    }

    public int getChildPositionCode(ProgramElement pe) {
	return method.getChildPositionCode(pe);
    }
    
    public int getRoleOfChild(int i) {
	return method.getRoleOfChild(i);
    }
    
    public int getIndexOfChild(int i) {
	return method.getIndexOfChild(i);
    }

    /** equals modulo renaming is described in class
     * SourceElement.
     */
     public boolean equalsModRenaming(SourceElement se, 
				      NameAbstractionTable nat) {
	 if (se == null || !(se instanceof ProgramMethod)) {
	     return false;
	 }

	 return method==((ProgramMethod)se).getMethodDeclaration();
     }

    public KeYJavaType getKeYJavaType(Term t) {
	return kjt;
    }

    public KeYJavaType getKeYJavaType() {
	return kjt;
    }

    public KeYJavaType getContainedInKeYJavaType() {
	return contKJT;
    }

    public Expression convertToProgram(Term t, ExtList l) {
	//!!! assume non-static %%%%%%
	ProgramElement called=(ProgramElement)l.get(0);
	l.remove(0);
	return new MethodReference(l, (ReferencePrefix) called, this);
    }

    /**
     * Get the method signature in Together format
     * e.g.: callM#(#SuperA#C#)#
     * @author VK
     */
    public String togetherSignature() {
       String s = name().toString()+"#(#";
       for (int i=0; i<arity()+1; i++) {
          if (i>0) s = s+"#";
          s=s+method.getParameterDeclarationAt(i).
                     getVariableSpecification().
                     getType().
                     getFullName();
       }
       s = s + "#)#";
System.err.println("togetherSig="+s);       
       return s;
    }
    
    /**
     * @return true if the value of "term" having this operator as
     * top-level operator and may not be changed by modalities
     */
    public boolean isRigid (Term term) {
	return false;
    }

}
