// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.gui.SequentPrinter;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.reference.ReferencePrefix;
import de.uka.ilkd.key.java.reference.ReferenceSuffix;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.logic.ProgramInLogic;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.Visitor;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.util.ExtList;

import java.io.IOException;

public class ProgramVariable extends TermSymbol 
    implements SourceElement, ProgramElement, 
    Expression, ReferencePrefix, IProgramVariable, 
    ParsableVariable, ReferenceSuffix, ProgramInLogic {

    // attention: this counter is used to get a unique variable name, once the
    // names are unique the counter should be removed %%%%
    private static long COUNTER = 0;
    private long id;
    
    private final boolean isStatic;
    private final KeYJavaType type;


    private ProgramVariable(ProgramElementName name, 
			    Sort               s,
			    KeYJavaType        t, 
			    boolean            isStatic) {

	super(name, s);
	this.type = t;
	this.isStatic = isStatic;
	
	// remove this as soon as possible %%%
	id = COUNTER;
	COUNTER++;
	
    }

    public ProgramVariable(ProgramElementName name, 
			   KeYJavaType        t, 
			   boolean            isStatic) {
	this(name, t.getSort(), t, isStatic);
    }

    public ProgramVariable(ProgramElementName name, KeYJavaType t) {
	this(name, t.getSort(), t, false);
    }

    public ProgramVariable(ProgramElementName name, Sort s) {
	this(name, s, null, false);
    }

    /** returns unique id %%%% HACK */
    public long id() {
	return id;
    }

    /** returns sort */
    public Sort sort() {
	return super.sort() == null ? type.getSort() : super.sort();
    }

   /** @return arity of the Variable as int */
    public int arity() {
	return 0;
    }

    /** @return name of the ProgramVariable */
    public ProgramElementName getProgramElementName() {
	return (ProgramElementName) name();
    }
    
    /** toString */
    public String toString() {
	return name()+((sort() == Sort.FORMULA)? "" : ":"+sort());
    }
    
    /**
     * returns true if the program variable has been declared as static
     */
    public boolean isStatic() {
	return isStatic;
    }

    public SourceElement getFirstElement(){
	return this;
    }

    public SourceElement getLastElement(){
	return this;
    }

    public Comment[] getComments() {
	return null;
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(de.uka.ilkd.key.java.visitor.Visitor v) {
	v.performActionOnProgramVariable(this);
    }


    /** the recoder pretty printer */
    public void prettyPrint(PrettyPrinter w) throws IOException {
	w.printProgramVariable(this);
    }

    /**
     * Returns the start position of the primary token of this element.
     * To get the start position of the syntactical first token,
     * call the corresponding method of <CODE>getFirstElement()</CODE>.
     * @return the start position of the primary token.
     */
    public Position getStartPosition(){
	return Position.UNDEFINED;
    }

    /**
     * Returns the end position of the primary token of this element.
     * To get the end position of the syntactical first token,
     * call the corresponding method of <CODE>getLastElement()</CODE>.
     * @return the end position of the primary token.
     */
    public Position getEndPosition(){
	return Position.UNDEFINED;
    }

    /**
     * Returns the relative position (number of blank heading lines and 
     * columns) of the primary token of this element.
     * To get the relative position of the syntactical first token,
     * call the corresponding method of <CODE>getFirstElement()</CODE>.
     * 
     * @return the relative position of the primary token.
     */
    public Position getRelativePosition(){
	return  Position.UNDEFINED;
    }

    public KeYJavaType getKeYJavaType(Term t) {
	return getKeYJavaType();
    }
    

    public KeYJavaType getKeYJavaType() {
	return type;
    }

    public KeYJavaType getKeYJavaType(Services javaServ) {
	return getKeYJavaType();
    }

    
    /**
     * We do not have a prefix, so fake it!
     * This way we implement ReferencePrefix
     * @author VK
     */
    public ReferencePrefix getReferencePrefix() {
	return null;
    }

    public ReferencePrefix setReferencePrefix(ReferencePrefix r) {
	return this;
    }
    
    
    /** equals modulo renaming is described in the corresponding
     * comment in class SourceElement. In this case two
     * programvariables are considered to be equal if they are
     * assigned to the same abstract name or if they are the same
     * object.
     */
    public boolean equalsModRenaming(SourceElement se, 
				     NameAbstractionTable nat) {
	if (nat.contains(this)) {
	    return nat.sameAbstractName(this, se);
	}	    
	return (se.equals(this));
    }

    public Expression convertToProgram(Term t, ExtList l) {
	return this;
    }
    
    public String proofToString() {
       return sort()+" "+name()+";\n";
    }

    /**
     * @return true if the value of "term" having this operator as
     * top-level operator and may not be changed by modalities
     */
    public boolean isRigid (Term term) {
	return false;
    }
    
}
