// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.declaration.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.gui.SequentPrinter;
import java.io.IOException;
import de.uka.ilkd.key.logic.ProgramConstruct;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.java.visitor.Visitor;

/**
 * This class represents the root of a schema variable hierarchy to be
 * express termstructures that match on logical terms. Schema variables
 * are used in Taclets where they act as placeholders for other
 * TermSymbols. The TermSymbols a SchemaVariable is allowed to match
 * is specified by their type and sort. If a match fits these
 * conditions can be checked using the method isCompatible(TermSymbol t)
 */
public abstract class SchemaVariable extends TermSymbol
    implements ParsableVariable, QuantifiableVariable, ProgramConstruct {

    /** 
     * SchemaVariables are used as placeholder for other
     * TermSymbols. The allowed type of the symbols is stored here.
     */
    private Class matchType = TermSymbol.class;

    /**
     * this flag indicates if the schemavariable matches a list of syntactical
     * elements 
     */
    private boolean listSV = false;


    /**
     * creates a SchemaVariable representing quantified variables 
     * @param name the Name of the SchemaVariable
     * @param sort the Sort of the variable the SchemaVariable will be
     * used to represent
     * @param listSV a boolean which is true iff the schemavariable is allowed
     * to match a list of quantified variables
     * @return the SchemaVariable
     */
    public static SchemaVariable createVariableSV(Name name,
						  Sort sort,
						  boolean listSV) {
	return new VariableSV(name, sort, listSV);
    }

    /** 
     * creates a SchemaVariable representing a term but not a formula
     * @param name the Name of the SchemaVariable
     * @param sort the Sort of the term the SchemaVariable will be
     * used to represent
     * @param rigidness true iff this SV may only match rigid
     * terms/formulas
     * @param listSV a boolean which is true iff the schemavariable is allowed
     * to match a list of terms
     * @return the SchemaVariable
     */
    public static SchemaVariable createTermSV(Name    name,
					      Sort    sort,
					      boolean listSV,
					      boolean rigidness) {
	return new TermSV(name, sort, listSV, rigidness);
    }

    public static SchemaVariable createTermSV(Name    name,
					      Sort    sort,
					      boolean listSV) {
	return createTermSV(name, sort, listSV, false);
    }

    /** creates a SchemaVariable representing a formula 
     * @param name the Name of the SchemaVariable
     * @param rigidness true iff this SV may only match rigid
     * terms/formulas
     * @param listSV a boolean which is true iff the schemavariable is allowed
     * to match a list of formulas
     * @return the SchemaVariable
     */
    public static SchemaVariable createFormulaSV(Name    name,
						 boolean listSV,
						 boolean rigidness) {
	return new FormulaSV(name, listSV, rigidness);
    }

    public static SchemaVariable createFormulaSV(Name    name,
						 boolean listSV) {
	return createFormulaSV(name, listSV, false);
    }

    /** creates a SchemaVariable representing a program construct
     * @param name the ProgramElementName of the SchemaVariable
     * @param listSV a boolean which is true iff the schemavariable is allowed
     * to match a list of program constructs
     * @return the SchemaVariable
     */
    public static SchemaVariable createProgramSV(ProgramElementName name, 
						 ProgramSVSort s,
						 boolean listSV){
	return new ProgramSV(name, s, listSV);
    }


    /** 
     * creates a SchemaVariable representing a skolem term
     * @param name the Name of the SchemaVariable
     * @param listSV a boolean which is true iff the schemavariable is allowed
     * to match a list of skolem terms
     * @return the SchemaVariable
     */
    public static SchemaVariable createDependingSV(Name name, 
						   Sort s,
						   boolean listSV){
	return new DependingSV(name, s, listSV);
    }


    /** 
     * creates a new SchemaVariable. That is used as placeholder for
     * TermSymbols of a certain type.
     * @param name the Name of the SchemaVariable
     * @param matchType Class representing the type of symbols that
     * can be matched
     * @param sort the Sort of the SchemaVariable and the matched type     
     * @param listSV a boolean which is true iff the schemavariable is allowed
     * to match a list of syntactical elements
     */    
    private SchemaVariable(Name name,
			   Class matchType,
			   Sort sort,
			   boolean listSV) {
	super(name, sort);
	this.matchType = matchType;
	this.listSV    = listSV;
    }

//     never used, though possibly useful
//     /** checks if the given TermSymbol fits to the set of possible
//      * allowed matches
//      * @param symbol TermSymbol to be checked if the matching conditions
//      * (type and sort compatibility) are fulfilled
//      * @return true iff symbol fits the conditions
//      */
//     public boolean isCompatible(TermSymbol symbol) {
// 	return (symbol.sort() == sort() &&
// 		matchType.isInstance(symbol));
//     }

    /** 
     * this method tests on object identity
     */
    public boolean equalsModRenaming(SourceElement se, 
				     NameAbstractionTable nat) {
	return se == this;
    }
    
    
    public Class matchType() {
	return matchType;
    }

    /** returns true iff this SchemaVariable is used to match
     * bound (quantifiable) variables 
     * @return true iff this SchemaVariable is used to match
     * bound (quantifiable) variables 
     */
    public boolean isVariableSV() {
	return false;
    }

    /** returns true iff this SchemaVariable is used to match
     * a term but not a formula
     * @return true iff this SchemaVariable is used to match
     * a term but not a formula
     */
    public  boolean isTermSV() {
	return false;
    }


    /** returns true iff this SchemaVariable is used to match
     * a formula 
     * @return true iff this SchemaVariable is used to match
     * a formula
     */
    public boolean isFormulaSV() {
	return false;
    }

    /** returns true iff this SchemaVariable is used to match
     * a part of a program 
     * @return true iff this SchemaVariable is used to match
     * a part of a program
     */
    public boolean isProgramSV() {
	return false;
    }

    /**
     * returns true if the schemavariable can match a list of syntactical
     * elements 
     * @return true if the schemavariable can match a list of syntactical
     * elements 
     */
    public boolean isListSV() {
	return listSV;
    }


    /** returns true iff this SchemaVariable is used to match (create)
     * a skolem term
     * @return true iff this SchemaVariable is used to match (create)
     * a skolem term
     */
    public boolean isDependingSV() {
	return false;
    }


   /** @return arity of the Variable as int */
    public int arity() {
	return 0;
    }

    /** toString */
    public String toString(String sortSpec) {
	return name()+" ("+sortSpec+")"; 
    }
    
    /**
     * @return true if the value of "term" having this operator as
     * top-level operator and may not be changed by modalities
     */
    public boolean isRigid (Term term) {
	return isRigid ();
    }

    public boolean isRigid () {
	return false;
    }

    // Stuff needed for ProgramSV 
    /** @returns comments if the schemavariable stands for programm
     * construct and has comments attached to it (not supported yet)
     */
    public Comment[] getComments() {
	return null;
    }

    public KeYJavaType getKeYJavaType() {
	return null;
    }

    public SourceElement getFirstElement(){
	return null;
    }

    public SourceElement getLastElement(){
	return null;
    }

    /**
 *        Returns the start position of the primary token of this element.
 *        To get the start position of the syntactical first token,
 *        call the corresponding method of <CODE>getFirstElement()</CODE>.
 *        @return the start position of the primary token.
    */
    public Position getStartPosition(){
	return Position.UNDEFINED;
    }

    /**
 *        Returns the end position of the primary token of this element.
 *        To get the end position of the syntactical first token,
 *        call the corresponding method of <CODE>getLastElement()</CODE>.
 *        @return the end position of the primary token.
     */
    public Position getEndPosition(){
	return Position.UNDEFINED;
    }

    /**
 *        Returns the relative position (number of blank heading lines and 
 *        columns) of the primary token of this element.
 *        To get the relative position of the syntactical first token,
 *        call the corresponding method of <CODE>getFirstElement()</CODE>.
 *        @return the relative position of the primary token.
     */
    public Position getRelativePosition(){
	return Position.UNDEFINED;
    }

    public ReferencePrefix getReferencePrefix() {
	return null;
    }

    public ReferencePrefix setReferencePrefix(ReferencePrefix r) {
	return this;
    }

    public int getDimensions(){
	return 0;
    }

    public int getTypeReferenceCount(){
	return 0;
    }

    public TypeReference getTypeReferenceAt(int index) {
	return this;
    }

    public PackageReference getPackageReference() {
	return null;
    }

    public int getExpressionCount() {
	return 0;
    }

    public Expression getExpressionAt(int index) {
	return null;
    }

    public int getChildCount() {
	return 0;
    }

    public ProgramElement getChildAt(int index) {
	return this;
    }
    
    public int getIndexOfChild(ProgramElement child) {
	return 0;
    }
    
    public int getChildPositionCode(ProgramElement child) {
	return 0;
    }

    public int getIndexOfChild(int positionCode) {
	return 0;
    }

    public int getRoleOfChild(int positionCode) {
	return 0;
    }

    public int getStatementCount() {
	return 0;
    }

    public int size() {
	return 0;
    }

    public ArrayOfExpression getUpdates() {
        return null;
    }

    public ArrayOfLoopInitializer getInits() {
	return null;
    }


    public Statement getStatementAt(int i) {
	return this;
    }

    public ProgramElementName getProgramElementName() {
	return new ProgramElementName(toString());
    }    

    public String getName() {
	return name().toString();
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnSchemaVariable(this);
    }

    /** this pretty printer method is for the program pretty printer
     * and needs not to be overwritten by ProgramSV but at the moment
     * it is not 
     */
    public void prettyPrint(PrettyPrinter w) throws IOException {	
	w.printSchemaVariable(this);
    }


    public KeYJavaType getKeYJavaType(Services javaServ) {
	return null;
    }

////////////////////////////////////////////////////////////////////////

    /** inner class that represents only schemavariables that are
     * bound and therefore only match other variables and not terms
     */
    static private class VariableSV extends SchemaVariable {
	/** creates a new SchemaVariable. That is used as placeholder for
	 * bound(quantified) variables.
	 * @param name the Name of the SchemaVariable
	 * @param sort the Sort of the SchemaVariable and the matched type     
	 * @param listSV a boolean which is true iff the schemavariable is allowed
	 * to match a list of quantified variables
	 */    
	private VariableSV(Name name,
			   Sort sort,
			   boolean listSV) {
	    super(name, QuantifiableVariable.class, sort, listSV); 	
	}

	/** returns true iff this SchemaVariable is used to match
	 * bound (quantifiable) variables 
	 * @return true iff this SchemaVariable is used to match
	 * bound (quantifiable) variables 
	 */
	public boolean isVariableSV() {
	    return true;
	}

	public boolean isRigid () {
	    return true;
	}

	/** toString */
	public String toString() {
	    return toString("variable");
	}
    }

    /** inner class that represents schemavariables that are
     * used to match logical terms and formulas.
     */
    static private abstract class AbstractTermSV extends SchemaVariable {
	private boolean rigidness;

	/** creates a new SchemaVariable
	 * @param name the Name of the SchemaVariable
	 * @param sort the Sort of the SchemaVariable and the matched type     
	 * @param rigidness true iff this SV may only match rigid
	 * terms/formulas
	 * @param listSV a boolean which is true iff the schemavariable is allowed
	 * to match a list of terms
	 */    
	protected AbstractTermSV(Name    name,
				 Sort    sort,
				 boolean listSV,
				 boolean rigidness) {
	    super(name, Term.class, sort, listSV);
	    this.rigidness = rigidness;
	}
	
	/**
	 * @return true if the value of "term" having this operator as
	 * top-level operator and may not be changed by modalities
	 */
	public boolean isRigid () {
	    return rigidness;
	}
    }

    /** inner class that represents schemavariables that are
     * used to match logical terms but not a formula.
     */
    static private class TermSV extends AbstractTermSV {
	/** creates a new SchemaVariable. That is used as placeholder for
	 * terms.
	 * @param name the Name of the SchemaVariable
	 * @param sort the Sort of the SchemaVariable and the matched type     
	 * @param rigidness true iff this SV may only match rigid
	 * terms/formulas
	 * @param listSV a boolean which is true iff the schemavariable is allowed
	 * to match a list of terms
	 */    
	private TermSV(Name    name,
		       Sort    sort,
		       boolean listSV,
		       boolean rigidness) {
	    super(name, sort, listSV, rigidness);
	    if (sort == Sort.FORMULA) {
		throw new RuntimeException("A TermSV is not allowed to"
					   +" have the sort "+sort);
	    }

	}
	

	/** returns true iff this SchemaVariable is used to match
	 * a term but not a formula
	 * @return true iff this SchemaVariable is used to match
	 * a term but not a formula
	 */
	public boolean isTermSV() {
	    return true;
	}

	/** toString */
	public String toString() {
	    return toString(sort().toString()+" term");
	}
	
    }

    /** inner class that represents schemavariables that are
     * used to match a formula.
     */
    static private class FormulaSV extends AbstractTermSV {
	/** creates a new SchemaVariable. That is used as placeholder for
	 * formulae.
	 * @param name the Name of the SchemaVariable
	 * @param rigidness true iff this SV may only match rigid
	 * terms/formulas
	 * @param listSV a boolean which is true iff the schemavariable is allowed
	 * to match a list of formulas
	 */    
	private FormulaSV(Name    name,
			  boolean listSV,
			  boolean rigidness) {
	    super(name, Sort.FORMULA, listSV, rigidness);
	}

	/** returns true iff this SchemaVariable is used to match
	 * a formula 
	 * @return true iff this SchemaVariable is used to match
	 * a formula
	 */
	public boolean isFormulaSV() {
	    return true;
	}

	/** toString */
	public String toString() {
	    return toString("formula");
	}	
    }

    /** 
     * inner class that represents schemavariables that are
     * used to match a part of a program
     */
    static private class ProgramSV extends SchemaVariable {

	/** 
	 * creates a new SchemaVariable used as a placeholder for program
	 * constructs
	 * @param name the Name of the SchemaVariable
	 * @param listSV a boolean which is true iff the schemavariable is
	 * allowed to match a list of program constructs
	 */    
	private ProgramSV(Name name, ProgramSVSort s, boolean listSV) {
	    super(name, ProgramConstruct.class, s, listSV);	
	}

	/** returns true iff this SchemaVariable is used to match
	 * a formula 
	 * @return true iff this SchemaVariable is used to match
	 * a formula
	 */
	public boolean isProgramSV() {
	    return true;
	}

	/** @returns comments if the schemavariable stands for programm
	 * construct and has comments attached to it (not supported yet)
	 */
	public Comment[] getComments() {
	    return null;
	}

	public SourceElement getFirstElement(){
	    return this;
	}

	public SourceElement getLastElement(){
	    return this;
	}

	/**
	 * Returns the start position of the primary token of this element.
	 * To get the start position of the syntactical first token,
	 * call the corresponding method of <CODE>getFirstElement()</CODE>.
	 * @return the start position of the primary token.
	 */
	public Position getStartPosition(){
	    return Position.UNDEFINED;
	}

	/**
	 * Returns the end position of the primary token of this element.
	 * To get the end position of the syntactical first token,
	 * call the corresponding method of <CODE>getLastElement()</CODE>.
	 * @return the end position of the primary token.
	 */
	public Position getEndPosition(){
	    return Position.UNDEFINED;
	}

	/**
	 * Returns the relative position (number of blank heading lines and 
	 * columns) of the primary token of this element.
	 * To get the relative position of the syntactical first token,
	 * call the corresponding method of <CODE>getFirstElement()</CODE>.
	 * @return the relative position of the primary token.
	 */
	public Position getRelativePosition(){
	    return  Position.UNDEFINED;
	}

	/** toString */
	public String toString() {
	    return toString("program "+sort().name());
	}	
    }

    /** 
     * inner class that represents schemavariables that are
     * used to create skolem terms
     */
    static private class DependingSV extends SchemaVariable {

	/** 
	 * creates a new SchemaVariable. That is used as placeholder for
	 * skolem terms.
	 * @param name the Name of the SchemaVariable
	 * @param sort the Sort of the SchemaVariable and the matched type     
	 * @param listSV a boolean which is true iff the schemavariable is
	 * allowed to match a list of program constructs
	 */    
	private DependingSV(Name name, Sort sort, boolean listSV) {
	    super(name, Term.class, sort, listSV);	
	    if (sort == Sort.FORMULA) {
		throw new RuntimeException("A DependingSV is not allowed to"
					   +" have the sort "+sort);
	    }

	}
	

	/** returns true iff this SchemaVariable is used to match
	 * a term but not a formula
	 * @return true iff this SchemaVariable is used to match
	 * a term but not a formula
	 */
	public boolean isDependingSV() {
	    return true;
	}

	public boolean isRigid () {
	    return true;
	}

	/** toString */
	public String toString() {
	    return toString(sort().toString()+" skolem term");
	}
	
    }

}
