// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.logic.op;

import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.util.Debug;

// to be replaced by HashMapFromIntegerToSimultaneousUpdateOp
import java.util.HashMap;

/** 
 * This class implements the simultaneous update operator. There exists exact one
 * operator for each arity. Allowed arities are of kind 2*n+1 (n>=1).
 * The operator syntax is similar to <code>{x_1:=t_1, ..., x_n:=t_n} Phi </code>
 */
public class SimultaneousUpdateOp implements Operator {

    /** Caches existing operators to fulfill the singleton assertion. 
     * The key is the arity of the operator, the value the operator.
     */
    private static final HashMap operatorHash = new HashMap(10);

    /** name of this operator */
    protected final Name name = new Name("simultaneousUpdateOp");

    /** arity of the operator */
    private int arity;
    

    /** returns the simultaneous update operator of given arity. 
     * If no such exists a new one is created and returned.
     */
    public static SimultaneousUpdateOp createSimultaneousUpdateOp(int arity) {	
	Integer key = new Integer(arity);

	SimultaneousUpdateOp result = (SimultaneousUpdateOp) 
	    operatorHash.get(key);
	if (result == null) {
	    result = new SimultaneousUpdateOp(arity);    
	    operatorHash.put(key, result);
	}

	return result;
    }

    /** 
     * creates a new update operator of the given _arity and adds it to the
     * set of already existing simultaneous update operators
     * @preconditions operatorHash->forAll(u:SimultaneousUpdateOp| u.arity<>_arity)
     * @postconditions operatorHash->includes(self) 
     */
    private SimultaneousUpdateOp(int _arity) {
	this.arity = _arity;

	Debug.assertTrue(!operatorHash.containsKey(new Integer(arity)), 
			 "simultaneousupdate: constructor precondition "+
			 "hurt (singleton)");
    }

    /** @return name of the Operator */
    public Name name() {
	return name;
    }

    /** @return true if the given term has  
     * is equal to the arity of this operator
     */
    public boolean validTopLevel(Term term) {
	if (term.arity() == arity && arity >= 3) {
	    for (int i=0; i<term.arity()-1; i+=2) {
		Operator op = term.sub(i).op();
		if (!(op  instanceof ProgramVariable || 
		      op instanceof AccessOp || 
		      (op instanceof SchemaVariable && 
		       ((SchemaVariable)op).isProgramSV()))) {
		    return false;
		}
	    }
	    return true;
	}
	return false;
    }

    /** @return arity of the update operator as int */
    public int arity() {
	return arity;
    }
    
    /** 
     * returns the sort a term would have if constructed with this operator and
     * the given sunterms. It is assumed that the term would be allowed if
     * constructed. 
     * @param term an array of Term containing the subterms of a (potential)
     * simultaneous update term 
     * @return sort of the simultaneous update term consisting of the given
     * subterms, if this op would be the top symbol
     */
     public Sort sort(Term[] term) {
        return term[arity-1].sort();
    }            

    /**
     * return true iff the given position describes the position of location 
     * @returns true iff the given position describes the position of location 
     */
    public boolean isLocation(int pos) {
	return (pos<arity-1 && pos%2==0);
    }

    /**
     * toString
     */
    public String toString() {
	return "simUpd("+arity()+")";
    }

    /**
     * @return true if the value of "term" having this operator as
     * top-level operator and may not be changed by modalities
     */
    public boolean isRigid (Term term) {
	return term.sub ( arity () - 1 ).isRigid ();
    }

}
