// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General Public License.
// See LICENSE.TXT for details.
//
// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001 Universitaet Karlsruhe, Germany
//                    and Chalmers University of Technology, Sweden
//
// The KeY system is protected by the GNU General
// Public License. See LICENSE.TXT for details.
//

/** @author Kristofer Johannisson */

package de.uka.ilkd.key.ocl.gf;

import java.util.List;
import java.util.Iterator;
import java.util.Arrays;
import java.util.HashMap;
import java.util.TreeSet;
import java.util.Collection;

import java.net.URL;
import java.io.*;
import java.util.*;
import de.uka.ilkd.key.util.Debug;

import de.uka.ilkd.key.ocl.tudresden.*;
import de.uka.ilkd.key.casetool.*;
import de.uka.ilkd.key.casetool.together.*;
import com.togethersoft.openapi.ide.project.IdeProjectManagerAccess;
import tudresden.ocl.check.types.Type;
import de.uka.ilkd.key.util.KeYResourceManager;

/** control object interfacing KeY with GF */
public class GFinterface
{
    public static String gfBinaryPath = "gf";

    private static TempGrammarFiles tgf;


    boolean PRINTDEBUG = false; // print a lot of information on System.out?
    boolean AUTO_GENERATE = true; // automatically generate grammars before starting editor
    UMLOCLTogetherModel model;
    Collection fromObject;
    Collection inOCL;
    Collection unknownAdded;

    /** path to the Together project */
    String projectRoot;

    public GFinterface()
    {
        // get the UML model
        String tprFile = IdeProjectManagerAccess
                          .getProjectManager().getActiveProject().getFileName();
        // xxx TODO: "/"-test not windows compliant
        projectRoot = tprFile.substring(0, tprFile.lastIndexOf("/"));
        model = new UMLOCLTogetherModel(projectRoot);

        // set up the filter (list) for features inherited from the java class "Object"
        String[] names = {"clone", "equals", "finalize", "getClass", "hashCode",
            "notify", "notifyAll", "toString", "wait"};
        fromObject = Arrays.asList(names);

        // set up list of OCL types
        String[] types = {"OclType", "OclAny", "OclState", "OclExpression", "Real", "Integer", "String",
            "Boolean", "Enumeration"};
        inOCL = Arrays.asList(types);

        unknownAdded = new TreeSet();
    }


    private void copyGrammars()
    {
        if (tgf == null) { // only copy if necessary
            try {
                tgf = new TempGrammarFiles();
            }
            catch (IOException e) {
                System.err.println("GFinterface: could not copy GF OCL grammars.");
                System.err.println(
                                   "e.getMessage(): " + e.getMessage());
                System.err.println(
                                   "e.getLocalizedMessage(): " +
                                   e.getLocalizedMessage());
                System.err.println("e.toString(): " + e);
                System.err.println("e.printStackTrace():");
                e.printStackTrace(System.err);
                //		Debug.out(e);
            }
        }
    }

    /** we may want to remove features inherited from java class "Object" */
    private boolean isInheritedFromObject(String oper)
    {
        return fromObject.contains(GrammarCreator.stripParams(oper));
    }

    /** make GF judgements for a class given that it is not in the model and has not
        * already been taken care of */
    private void addIfNotInModel(String clss, HashMap classMap, GrammarCreator gc)
    {
        if  (!unknownAdded.contains(clss) && !inOCL.contains(clss) && classMap.get(clss) == null) {
            unknownAdded.add(clss);
            gc.mkClass(clss);
        }
    }

    /** Start GF editor for editing a class invariant
        * @param classname The name of the class whose invariant should be edited
        * @param inv The current invariant in OCL as a String (currently ignored)
        * @param abs string containing GF abstract syntax representation of invariant
        * @return message to Together (only if something went wrong?)
        */
    public String editClassInvariant(String classname, String inv, String abs, CallbackClassInv cci)
    {
        copyGrammars();
        if (AUTO_GENERATE)
            createGrammar();

        //System.out.println("editClassInvariant received class " + classname +
        //		   " and invariant " + inv);
        // now find TJClassifier for classname
        //TJClassifier tjc = (TJClassifier) model.getClasses().get(classname);
        //if (tjc == null)
        //	return ("Could not find class " + classname + " in model.");

        // create a CallbackClassInv containing this TJClassifier
        // CallbackClassInv cci = new CallbackClassInv(tjc);

        String gf = gfBinaryPath + " +java " + //}; // +
            projectRoot+"/specs.OCL.gf " +
            projectRoot+"/specs.Eng.gf";

        GrammarCreator gc = new GrammarCreator();
        GFEditor13.mainClassInv(gf, cci, abs, gc.classConstraint(classname)+"HC");
        return ""; // all is well
}

/** Start GF editor for editing pre-/postcondition of a method/operation
* @param the name of class of the operation
* @param oper signature of the operation
* @param pre string containg OCL precondition (currently ignored)
* @param post string containg OCL postcondition (currently ignored)
* @param abs string containing GF abstract syntax representation of pre-/postconditions
* @return message to Together (if something went wrong)
*/
public String editPrePost(String clss, String oper, String pre, String post, String abs, CallbackPrePost cpp)
{
    GrammarCreator gc = new GrammarCreator();
    String gf = gfBinaryPath + " +java " + //}; // +
        projectRoot+"/specs.OCL.gf " +
        projectRoot+"/specs.Eng.gf";

    copyGrammars();
    if (AUTO_GENERATE)
        createGrammar();

    // convert operation signature into GF identifier
    String dfault = gc.mkOpConstrSyntaxTree(clss, oper);

    if (PRINTDEBUG) {
        System.out.println("editPrePost sending to GFeditor:");
        System.out.println("    abs = " + abs);
        System.out.println("    dfault = " + dfault);
    }
    GFEditor13.mainPrePost(gf, cpp, abs, dfault);
    return ""; // everything should be OK
    }

/** start the GF syntax editor with the grammars for the current project
* (it is assumed that these grammars have already been created) */
public void startEditor()
{
    String [] args = { gfBinaryPath + " +java " + //}; // +
        projectRoot+"/specs.OCL.gf " +
        projectRoot+"/specs.Eng.gf"} ;
    GFEditor13.main(args);
    }

/** create the necessary GF grammars from the information in the UML model of
* the current project */
public void createGrammar()
{
    GrammarCreator gc = new GrammarCreator();

    Iterator it = model.getUMLOCLClassifiers().keySet().iterator();

    // iterate through all classes in the model:
    while (it.hasNext()) {
        String className = it.next().toString();
        gc.mkClass(className);

        UMLOCLClassifier tjc = model.getUMLOCLClassifiers().getClassifier(className);
        // System.out.println("Invariant for " + tjc.getName() + " :" + tjc.getInvariant());

        // add information about inheritance (supertypes of this class)
        if (tjc.getSupertypes() == null) {System.out.println("supertypes NULL");};
        if (tjc.getSupertypes().keySet() == null) {System.out.println("keyset NULL");};
      
        //	    String[] supertypes = (String[])(tjc.getSupertypes().keySet().toArray());
        Set supertypes = tjc.getSupertypes().keySet();
        Iterator itst = supertypes.iterator();
        /* 	    if (supertypes.length > 0) { */
        /* 		for (int i=0; i < supertypes.length; i++)  { */
        /* 		    String sclass = supertypes[i]; */
        /* 		    // the suptertype might not be in the model: */
        /* 		    addIfNotInModel(sclass, model.getUMLOCLClassifiers(), gc);  */
        /* 		    if (! (sclass.equals("Object") || sclass.equals("java.lang.Object"))) { */
        /* 			// don't bother about java class Object */
        /* 			gc.mkSuperType(className, sclass); */
        /* 		    } */
        /* 		} */
        /* 	    } else if (PRINTDEBUG) { */
        /* 		System.out.println("No supertypes of "+ className); */
        /* 	    } */
        while (itst.hasNext()) {
            String sclass = (String) itst.next();
            // the suptertype might not be in the model:
            addIfNotInModel(sclass, model.getUMLOCLClassifiers(), gc);
            if (! (sclass.equals("Object") || sclass.equals("java.lang.Object"))) {
                // don't bother about java class Object
                gc.mkSuperType(className, sclass);
            }
        }
        // iterate through all features of one class:
        HashMapOfFeatures featureMap = tjc.features();
        Iterator it2 = featureMap.keySet().iterator();
        while (it2.hasNext()) {
            UMLOCLFeature f = (UMLOCLFeature) featureMap.get(it2.next());
            boolean inhFromObj = isInheritedFromObject(f.getName());
            if (f instanceof UMLOCLStructuralFeature && !inhFromObj) {
                // Attributes (is this all?)
                gc.mkAttribute(className, f.getName(), f.getType().toString());
                // make GF judg. for the attr. type if it is not in the model
                addIfNotInModel(f.getType().toString(), model.getUMLOCLClassifiers(), gc);
            } else if (f instanceof UMLOCLBehaviouralFeature && !inhFromObj) {
                // operations

                if (f.getType() != null)
                    addIfNotInModel(GrammarCreator.stripParams(f.getType().toString()),
                                    model.getUMLOCLClassifiers(), gc);

                // we convert the list of parameter types to a list of strings
                Type[] ps = ((UMLOCLBehaviouralFeature) f).getParameters();
                String[] params = new String[ps.length];

                for (int i=0; i < ps.length; i++) {
                    params[i] = ps[i].toString();
                    // make GF judg. for the param type if it is not in the model
                    addIfNotInModel(params[i], model.getUMLOCLClassifiers(), gc);
                }

                if (((UMLOCLBehaviouralFeature) f).isQuery()) {
                    if (f.getType() == null) {
                        // this wasn't a query after all...
                        // we should probably throw some nice exception
                        // but we instead just assume it is a non-query operation:
                        gc.mkNonQuery(className, f.getName(),
                                      null, params);
                    } else
                        gc.mkQuery(className, f.getName(),
                                   f.getType().toString(), params);
                } else {
                    // this operation is not a query
                    if (f.getType() != null)
                        gc.mkNonQuery(className, f.getName(),
                                      f.getType().toString(), params);
                    else
                        gc.mkNonQuery(className, f.getName(),
                                      null, params);
                }
            }
        }

        // iterate through all association ends of one class:
        // this is currently not working
        HashMapOfAssociations assocMap = tjc.getAssociations();
        Iterator it3 = assocMap.keySet().iterator();

        // System.out.println("Assoc. exists for class " + className + ": " + it3.hasNext());
        while (it3.hasNext()) {
            UMLOCLAssociation ae = (UMLOCLAssociation)(assocMap.get(it3.next()));
            // System.out.println("Association ends for the class " + className);
            // System.out.println(ae.getName() + ":" + ae.getMultiplicity() + ", " +
            //	ae.isOrdered() + ", " + ae.getType());
            gc.mkAssoc(className, ae.getTargetRole(), ae.getTarget().getName(),
                       ae.getTargetMultiplicity(), false);
        }
    }

    // 	URL grammarsURL = KeYResourceManager.getManager().
    // 	    getResourceFile(GFinterface.class, "grammars");
    // 	if (grammarsURL == null) {
    // 	    System.err.println("GFinterface: could not find path to GF OCL grammars.");
    // 	} else {
    // 	    // use File object to find the absolute path:
    // 	    File dummy = new File(grammarsURL.getPath());
    // 	    String grammarsPath = dummy.getAbsolutePath();
    // 	    if (PRINTDEBUG)
    // 		System.out.println("GFinterface got this grammar path: " +
    // 				   grammarsPath);


    gc.flush(projectRoot, tgf); //  "/Users/kihaa/OCLGF/");
                                // startEditor();
    
    }
}
