// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.ocl.ocldltranslator;

/**
 * A base class for all transformers of ocl expressions (occuring
 * in an ocl constraint)  into logical formulas
 * It especially provides basic functionality for the
 * management of the node-nodetranslations-mapping.
 * It also extends DepthFirstAdapter an therefore implements a basic tree
 * walking procedure.
 */
import tudresden.ocl.*;
import tudresden.ocl.parser.analysis.*;
import java.util.Set;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Vector;
import tudresden.ocl.parser.node.*;
import tudresden.ocl.check.types.*;

import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.heuristics.*;
import de.uka.ilkd.key.ocl.ocldltranslator.control.*;
import de.uka.ilkd.key.ocl.ocldltranslator.normalize.*;
import de.uka.ilkd.key.ocl.ocldltranslator.transform.*;



public abstract class BasicOclExpressionTranslator extends DepthFirstAdapter {

    protected AExpression expression;
    protected OclTree tree;

    // This map maps nodes in the AST to their logical translationsgenerated when executing
    // the translation.
    // It uses Nodes as keys and delivers translation result objects representing the translation ...
    protected Map nodeTranslationResultMap;
    protected Vector predefinedVarNames;
    protected Vector predefinedVarTypes;


    // Maps variable(names) to terms representing them ...
    protected Map varMap;
 
    public BasicOclExpressionTranslator(OclTree constraintTree){
	// initialize state of translation ...

	nodeTranslationResultMap = new HashMap();
        varMap = new HashMap();

	this.tree = constraintTree;
	this.predefinedVarNames = new Vector();
	this.predefinedVarTypes = new Vector();

	requireTreeInvariants();
    }

    public BasicOclExpressionTranslator(OclTree constraintTree, Vector predefinedVarNames,
				 Vector predefinedVarTypes){
	this(constraintTree);

	this.predefinedVarNames = predefinedVarNames;
	this.predefinedVarTypes = predefinedVarTypes;
    }
    

    // basic operations for the handling of translation of nodes ...
    
    protected TranslationResult getTranslationForNode(Node n){
	
	if(n == null){ 
	    error("Trying to get a translation for a 'null' node !");
	}
	
	if (!nodeTranslationResultMap.containsKey(n)) {
	   throw new RuntimeException("Translation for node ("+ n +") required but not found !!!");
	    
	}
	return (TranslationResult) nodeTranslationResultMap.get(n); 
    }

    protected void setTranslationForNode(Node n, TranslationResult translation){
	if (translation != null) {
	    nodeTranslationResultMap.put(n, translation);
	} else {
	    error("Trying to set 'null' as the translation of the node :" + n );
	}
    }



    // basic operations for the handling of translation of variables ...
    
    TranslationResult getTranslationForVariable(String varName, String typeName){
	String varKey = varName + "$$$$" + typeName;

	if (!varMap.containsKey(varKey)) {
	    throw new RuntimeException("Translation for variable ("+ varName + ":" +
				       typeName+") required but not found !!!");
	    
	}
	return (TranslationResult) varMap.get(varKey); 
    }

    void setTranslationForVariable(String varName, String typeName, TranslationResult translation){
	String varKey = varName + "$$$$" + typeName;
	varMap.put(varKey, translation);
	if (typeName.equals("byte") ||
	    typeName.equals("short") ||
	    typeName.equals("int") ||
	    typeName.equals("long")) {
	    varKey = varName + "$$$$" + "Integer";
	    varMap.put(varKey, translation);
	} else if (typeName.equals("boolean")) {
	    varKey = varName + "$$$$" + "Boolean";
	    varMap.put(varKey, translation);
	}

	log("Set translation for variable:  " + varName + ":" + typeName);
    }


    void removeTranslationForVariable(String varName, String typeName){
	String varKey = varName + "$$$$" + typeName;
	varMap.remove(varKey);
    }
    
    /**
     * this method is used to reach through the result of a (given)nodes
     * 'from' as the result of another given node 'to' ...
     */
    
    protected void reachThrough(Node to, Node from) {
	TranslationResult translation = getTranslationForNode(from);
	setTranslationForNode(to,translation);

    }




    // =========== tree traversal operations ===================

   
    public void caseAExpression(AExpression e) {
	// We have to change the order of evaluation of the subnodes ...

	log("Case AExpr for  == " + e);

	// First call translateLetVariableDeclaration for every Let Expression ....
	ALetExpression currentLetExpression = null;
	ListIterator iter = e.getLetExpression().listIterator();

	AExpression ae;
	String letVarName;
	Type letVarType;

	while (iter.hasNext()){
	    currentLetExpression = (ALetExpression) iter.next();
	    ae = (AExpression) currentLetExpression.getExpression();
	    
	    log("Translation let-expr == " + ae);
	    
	    // Translate the specified value for the let variable ...
	    this.caseAExpression(ae);

	    log("Finished translation let-expr == " + ae);

	    letVarName = currentLetExpression.getName().toString().trim();
	    letVarType = tree.getNodeType(ae);
	    translateLetVariableDeclaration(currentLetExpression, letVarName, letVarType);
	}

	// Then translate the logical expression ...
	ALogicalExpression alo = (ALogicalExpression) e.getLogicalExpression();
	super.caseALogicalExpression(alo);
	
	if (!e.getLetExpression().isEmpty()) {
	    // Then translate the let expressions in reversed order ...
	    iter = e.getLetExpression().listIterator();
	
	    while (iter.hasNext()){
		currentLetExpression =(ALetExpression) iter.next();
	    }
	    // Now we have the last Let expression ...
	    this.caseALetExpression(currentLetExpression);
	    while (iter.hasPrevious()){
		currentLetExpression = (ALetExpression)iter.previous();
		this.caseALetExpression(currentLetExpression);
	    }
	    
	}
	
	if(e.getLetExpression().isEmpty()){
	    // There are no Let-Expressions ...
	    log("Set Translation  for Expr == " + e);
	    reachThrough(e, e.getLogicalExpression());
	} else {
	    log("Set Translation  for Expr == " + e);
	    // Take the translation of the first
	    reachThrough(e,currentLetExpression);
	}
    }




    
    /** 
     * if a then b else c endif    
     */
    public void outAIfExpression(AIfExpression ie) {
	PExpression ifExpression = ie.getIfBranch();
	PExpression thenExpression = ie.getThenBranch();
	PExpression elseExpression = ie.getElseBranch();
	
	translateIfThenElseExpression(ie,ifExpression, thenExpression, elseExpression);	
    }

    /** 
     *  AND, OR and XOR have precedence
     *  over IMPLIES.
     *  AND, OR and XOR have the same precedende.
     */
  public void outALogicalExpression(ALogicalExpression le) {
    if (le.getLogicalExpressionTail().isEmpty()) {
	// we have no 'real' logical expression ...
	reachThrough( le, le.getRelationalExpression() );
    } else {
	// we have a logical expression ...
	// Take care of the operator precedence rules !!!!

	int tailElements=le.getLogicalExpressionTail().size(); // this is >= 1 ...
	PRelationalExpression[] relExprs=new PRelationalExpression[tailElements+1];
	PLogicalOperator[] logOps=new PLogicalOperator[tailElements];
	// logOps[i] is the operator AFTER relExprs[i]
	Iterator iter=le.getLogicalExpressionTail().iterator();
	int index=0;
	relExprs[0]=le.getRelationalExpression();
	while (iter.hasNext()) {
	    ALogicalExpressionTail let=(ALogicalExpressionTail)iter.next();
	    logOps[index]=let.getLogicalOperator();
	    index++;
	    relExprs[index]=let.getRelationalExpression();
	}
	
	TranslationResult res = handleLogicalExpression(le, relExprs, logOps, 0, tailElements - 1 );
	setTranslationForNode(le,res);
    }

  }
    
    protected TranslationResult handleLogicalExpression(ALogicalExpression le, PRelationalExpression[] relExprs, PLogicalOperator[] logOps, int from, int to){
	 
	TranslationResult res = null;

	// First find the leftmost IMPLIES operator ...
	int leftMostImplies = -1; // index of operand after last implies
	for (int i= from; (i <= to) && (leftMostImplies == -1) ; i++){
	    if(logOps[i] instanceof AImpliesLogicalOperator){
		leftMostImplies = i;
	    }
	}

	if (leftMostImplies != -1){
	    // There is an IMPLIES operator => take care about precedence ...
	     // First translate left and right logical subexpressions ...

	    TranslationResult leftTranslationResult = null;
	    TranslationResult rightTranslationResult = null; 
	    

	    if (from == leftMostImplies) {
		leftTranslationResult = getTranslationForNode(relExprs[from]);
	    } else {
		leftTranslationResult = handleLogicalExpression(le, relExprs, logOps,
								from, leftMostImplies - 1);
	    }
	    
	    if (to == leftMostImplies) {
		rightTranslationResult = getTranslationForNode(relExprs[to + 1]);
	    } else {
		rightTranslationResult = handleLogicalExpression(le, relExprs, logOps,
								 leftMostImplies + 1, to);
	    }

	    
	    // Now translate the IMPLIES operator ...
	    
	    res = translateLogicalExpression(le, logOps[leftMostImplies].toString().trim(),
					     leftTranslationResult,
					     rightTranslationResult);
	    

	} else {
	    // There is no IMPLIES operator, therefore all logical operators have
	    // the same precedence -> Build term from left to right !
	    
	    TranslationResult leftOperand, rightOperand;
	    
	    leftOperand = getTranslationForNode(relExprs[from]);

	    for (int i= from; (i <= to) ; i++){
		PLogicalOperator currentOperator = logOps[i];
		rightOperand = getTranslationForNode(relExprs[i+1]);
		
		res = translateLogicalExpression(le, currentOperator.toString().trim(),
						 leftOperand, rightOperand);
		leftOperand = res;
	    
	    }

	}
	    
	// Now return the result ...
	return res;
    }
	    
       
    

    
    public void outARelationalExpression(ARelationalExpression re) {
	if (re.getRelationalExpressionTail()==null) {
	    // we have no 'real' relational expression ...
	    reachThrough(re, re.getAdditiveExpression());
	} else {
	    ARelationalExpressionTail ret=
		(ARelationalExpressionTail) re.getRelationalExpressionTail();
	    PRelationalOperator relationalOp=ret.getRelationalOperator();
	    
	    AAdditiveExpression leftOperand = (AAdditiveExpression) re.getAdditiveExpression();
	    AAdditiveExpression rightOperand = (AAdditiveExpression) ret.getAdditiveExpression();
	    String opName = relationalOp.toString().trim(); 
	    
	    translateRelationalExpression(re, opName, leftOperand, rightOperand);
	    
	}
    }


    public void outAAdditiveExpression(AAdditiveExpression ae) {
	if (ae.getAdditiveExpressionTail().isEmpty()) {
	    // we have no 'real' additive expression ...
	    reachThrough(ae, ae.getMultiplicativeExpression());
	} else {
	    // We have + or - which have both the same precedence.
	    // Therefore translate from left to right ...
	    
	    AMultiplicativeExpression leftOperand = (AMultiplicativeExpression) ae.getMultiplicativeExpression();
	    AMultiplicativeExpression rightOperand;
	    PAddOperator currentOperator;

	    TranslationResult leftResult, rightResult, res;
	    res = null;
	    leftResult = getTranslationForNode(leftOperand); 

	    Iterator iter = ae.getAdditiveExpressionTail().iterator();
	    while (iter.hasNext()) {
		AAdditiveExpressionTail aet = (AAdditiveExpressionTail) iter.next();
		rightOperand = (AMultiplicativeExpression) aet.getMultiplicativeExpression();
		rightResult = getTranslationForNode(rightOperand);
		currentOperator = aet.getAddOperator();

		res = translateAdditiveExpression(ae, currentOperator.toString().trim(), leftResult,
						  rightResult);
		leftResult = res;
	    }
	    
	    setTranslationForNode(ae,res);
	}
    }
    
    
    public void outAMultiplicativeExpression(AMultiplicativeExpression me) {
	if (me.getMultiplicativeExpressionTail().isEmpty()) {
	    // we have no 'real' multiplicative expression ...
	    reachThrough(me, me.getUnaryExpression());
	} else {

	    // We have * or / which have both the same precedence.
	    // Therefore translate from left to right ...
	    
	    PUnaryExpression leftOperand = me.getUnaryExpression();
	    PUnaryExpression rightOperand;
	    PMultiplyOperator currentOperator;

	    TranslationResult leftResult, rightResult, res;
	    res = null;
	    leftResult = getTranslationForNode(leftOperand); 

	    Iterator iter = me.getMultiplicativeExpressionTail().iterator();
	    while (iter.hasNext()) {
		AMultiplicativeExpressionTail met = (AMultiplicativeExpressionTail) iter.next();
		rightOperand = met.getUnaryExpression();
		rightResult = getTranslationForNode(rightOperand);
		currentOperator = met.getMultiplyOperator();

		res = translateMultiplicativeExpression(me, currentOperator.toString().trim(),
							leftResult,
							rightResult);
		leftResult = res;
	    }
	    
	    setTranslationForNode(me,res);
	}
    }	
	/**
	 * not , unary - ...
	 */
	public void outAUnaryUnaryExpression(AUnaryUnaryExpression uue) {
	    PUnaryOperator currentOperator = uue.getUnaryOperator();
	    PPostfixExpression pe = uue.getPostfixExpression();
	    TranslationResult operand = getTranslationForNode(pe);
	    TranslationResult res;
	    res = translateUnaryExpression(uue, currentOperator.toString().trim(),
					   operand);
	    setTranslationForNode(uue,res);
	}


	public void outAPostfixUnaryExpression(APostfixUnaryExpression pue) {
	    reachThrough(pue, pue.getPostfixExpression());
	}


    

    // Here we have a navigational OCL expression (feature calls) ...
    
    // checked !
    public void outAPostfixExpression(APostfixExpression pe) {
	if (pe.getPostfixExpressionTail().isEmpty()) {
	    reachThrough(pe, pe.getPrimaryExpression());
	    
	} else {
	    APostfixExpressionTail lastPET = (APostfixExpressionTail)
		pe.getPostfixExpressionTail().getLast();
	    
	    reachThrough(pe, lastPET);
	    
	    
	}
    }


    public void outAFeaturePrimaryExpression(AFeaturePrimaryExpression pe) {
	Type oclType=tree.getNodeType(pe);
	if (oclType instanceof OclType) {
	    // We have an OclType literal ...
	    String lit=pe.toString().trim();
	    translateOclTypeLiteral(pe, lit);
	} else {
	    String varName = pe.toString().trim();
	    Type varType = oclType;
	    translateVariable(pe , varName , varType);
	}
    }

 
    public void inAPostfixExpressionTail(APostfixExpressionTail pet) {
	// Do the iteratorVariable declaration ...
	Vector parameterNodes = new Vector();
	Vector iteratorNames = new Vector();
	
	Type iteratorType = null;
	
	AFeatureCall fc=(AFeatureCall)pet.getFeatureCall();

	AFeatureCallParameters fcp=(AFeatureCallParameters)fc.getFeatureCallParameters();
	AActualParameterList apl=null;
	if (fcp!=null) {
	    apl=(AActualParameterList)fcp.getActualParameterList();
	}
	
	String featureName=fc.getPathName().toString().trim();

	ListIterator iter;
	
	if (featureName.equals("iterate")) {
	    // The iterate case is a bit more complicated ...

	    // Do declarations for the iterator variables ...
	    AIterateDeclarator decl = (AIterateDeclarator) fcp.getDeclarator();
	    String iterName = decl.getIterator().toString().trim();
	    iteratorType = tree.getTypeFor(iterName , apl); 

	    log("Determined type of it-variable '"+iterName+"' as :"+iteratorType.toString());

	    translateVariableDeclaration(iterName, iteratorType);

	    // Do declarations for the accumulator variables ...
	   
	    String accName = decl.getAccumulator().toString().trim();
	    Type accumulatorType = tree.getTypeFor(accName , apl); 

	    log("Determined type of acc-variable '"+accName+"' as :"+accumulatorType.toString());

	    translateVariableDeclaration(accName, accumulatorType);

       

	    
	} else if (tudresden.ocl.check.TypeChecker.setOfIteratingMethodNames.contains(featureName)) {
	    
	    // Do declarations for all iterator variables ...
	    AStandardDeclarator decl = (AStandardDeclarator) fcp.getDeclarator();
	    String iterName = decl.getName().toString().trim();
	    iteratorType = tree.getTypeFor(iterName , apl); 

	    log("Determined type of it-variable '"+iterName+"' as :"+iteratorType.toString());

	    translateVariableDeclaration(iterName, iteratorType);
	    
	    
	    // If there are several iterators then also add their names ...
	    iter=decl.getDeclaratorTail().listIterator();
	    ADeclaratorTail dtail;
	    while (iter.hasNext()) {
		dtail=(ADeclaratorTail) iter.next();
		iterName = dtail.getName().toString().trim();
		translateVariableDeclaration(iterName, iteratorType);
	    }
	} 
    }

    public void outAPostfixExpressionTail(APostfixExpressionTail pet) {
	Node appliedTo=getPreviousNode(pet);
	Type appliedType=tree.getNodeType(appliedTo);
	
	
	AFeatureCall fc=(AFeatureCall)pet.getFeatureCall();
	boolean featureCallWithAtPre = false;
	if (fc.getTimeExpression()!=null) {
	    // we have a time expression (@pre was set for the feature) ...
	    featureCallWithAtPre = true;
	}
	
	boolean featureCallWithArrow = (pet.getPostfixExpressionTailBegin() instanceof
					AArrowPostfixExpressionTailBegin);
	
	Vector parameterNodes = new Vector();
	Vector iteratorNames = new Vector();
	String accumulatorName;
	
	boolean isMethodCall = false;
	
	String featureName=fc.getPathName().toString().trim();
	Type iteratorType = null;
	
      
	
	AFeatureCallParameters fcp=(AFeatureCallParameters)fc.getFeatureCallParameters();
	AActualParameterList apl=null;
	if (fcp!=null) {
	    apl=(AActualParameterList)fcp.getActualParameterList();
	    isMethodCall = true;
	}
	
	ListIterator iter;
	
	if (featureName.equals("iterate")) {
	    // The iterate case is a bit more complicated ...

	    // Do declarations for the iterator variables ...
	    AIterateDeclarator decl = (AIterateDeclarator) fcp.getDeclarator();
	    String iterName = decl.getIterator().toString().trim();
	    iteratorType = tree.getTypeFor(iterName , apl); 

	    log("Determined type of it-variable '"+iterName+"' as :"+iteratorType.toString());
	    iteratorNames.add(iterName);
	    
	    // Do declarations for the accumulator variables ...
	   
	    String accName = decl.getAccumulator().toString().trim();
	    Type accumulatorType = tree.getTypeFor(accName , apl); 

	    log("Determined type of acc-variable '"+accName+"' as :"+accumulatorType.toString());
	    iteratorNames.add(accName);
	
	   
	    
	} else if (tudresden.ocl.check.TypeChecker.setOfIteratingMethodNames.contains(featureName)) {
	    
	    
	    // Get all names of iterator variables ...
	    AStandardDeclarator decl = (AStandardDeclarator) fcp.getDeclarator();
	    String iterName = decl.getName().toString().trim();
	    iteratorType = tree.getTypeFor(iterName , apl); 

	    log("Determined type of it-variable '"+iterName+"' as :"+iteratorType.toString());
	    iteratorNames.add(iterName);


	    // If there are several iterators then also add their names ...
	    iter=decl.getDeclaratorTail().listIterator();
	    ADeclaratorTail dtail;
	    while (iter.hasNext()) {
		dtail=(ADeclaratorTail) iter.next();
		iterName = dtail.getName().toString().trim();
		iteratorNames.add(iterName);
		
	    }
	} 
	
	// Now get all parameters for the feature call ...

	if (featureName.equals("iterate")) {
	    // The iterate case is a bit more complicated ...
	    // Use the initialization expression as the first 'parameter'.

	    AIterateDeclarator declarator = (AIterateDeclarator) fcp.getDeclarator();
	    Node initExpression = declarator.getExpression();
	    parameterNodes.add(initExpression);
	} 

	if (apl != null){
	    // there are some parameters ...
	    Node parameter = apl.getExpression();
	    parameterNodes.add(parameter);
	    // If there are several parameters then also add their nodes ...
	    iter =apl.getActualParameterListTail().listIterator();
	    AActualParameterListTail apltail;
	    while (iter.hasNext()) {
		apltail=(AActualParameterListTail) iter.next();
		parameter = apltail.getExpression();
		parameterNodes.add(parameter);
	    }
	}

	
	Type t = appliedType;
	// Distinguish the different cases ...
	if (t instanceof Basic) {
	    if (t==Basic.INTEGER) {
		translateFeatureCallonInteger(pet, featureName, appliedTo, parameterNodes,
					      iteratorNames, featureCallWithAtPre,featureCallWithArrow );
	    }
	    else if (t==Basic.REAL) {
		translateFeatureCallonReal(pet, featureName, appliedTo,  parameterNodes,
					   iteratorNames, featureCallWithAtPre,featureCallWithArrow );
	    }
	    else if (t==Basic.BOOLEAN) {
		translateFeatureCallonBoolean(pet, featureName, appliedTo,  parameterNodes,
					      iteratorNames, featureCallWithAtPre,featureCallWithArrow );
	    }
	    else if (t==Basic.STRING) {
		translateFeatureCallonString(pet, featureName, appliedTo,  parameterNodes,
					     iteratorNames, featureCallWithAtPre,featureCallWithArrow );
	    }
	    else throw new RuntimeException("illegal basic type");
	} else if (t instanceof Collection) {
	    Collection c=(Collection) t;
	    switch (c.getCollectionKind()) {
	    case Collection.SET:         	
		translateFeatureCallonSet(pet, featureName,  appliedTo, parameterNodes,
					  iteratorNames,iteratorType , featureCallWithAtPre);
		break;
		
	    case Collection.BAG:
		translateFeatureCallonBag(pet, featureName, appliedTo,  parameterNodes,
					  iteratorNames,iteratorType, featureCallWithAtPre);
		break;
		
	    case Collection.SEQUENCE:
		translateFeatureCallonSequence(pet, featureName, appliedTo,
					       parameterNodes, iteratorNames,
					       iteratorType,featureCallWithAtPre);
		break;
		
		//   case Collection.COLLECTION: 
		//  	  translateFeatureCallonCollection(pet, featureName, appliedTo,  parameterNodes, iteratorNames,iteratorType );
		//  	  break;
		
	    default: throw new RuntimeException("illegal collection type");
	    }
	} else if (t instanceof OclType) {
	    translateFeatureCallonOclType(pet, featureName, appliedTo,  parameterNodes,
					  iteratorNames, featureCallWithAtPre, featureCallWithArrow , isMethodCall);
	} else {
	    // application type
	    translateFeatureCallonModelType(pet, featureName, appliedTo,  parameterNodes,
					    iteratorNames, featureCallWithAtPre, featureCallWithArrow, isMethodCall);
	}

	if (featureName.equals("iterate")) {
	    // Now we have to remove the local variable definitions ...

	    AIterateDeclarator decl = (AIterateDeclarator) fcp.getDeclarator();
	    String iterName = decl.getIterator().toString().trim();
	    iteratorType = tree.getTypeFor(iterName , apl); 
	    removeVariableDeclaration(iterName, iteratorType);
	   
	    String accName = decl.getAccumulator().toString().trim();
	    Type accumulatorType = tree.getTypeFor(accName , apl); 
	    removeVariableDeclaration(accName, accumulatorType);
	    	    

	} else if (tudresden.ocl.check.TypeChecker.setOfIteratingMethodNames.contains(featureName)) {

	    
	    // Now we have to remove the local variable definitions ...
	    AStandardDeclarator decl = (AStandardDeclarator) fcp.getDeclarator();
	    
	    String iterName = decl.getName().toString().trim();
	    // We have to call for the deletion of the localy defined iterator variables ...
	    removeVariableDeclaration(iterName, iteratorType);
	    
	    
	    // If there are several iterators then also remove their declarations ...
	    iter=decl.getDeclaratorTail().listIterator();
	    while (iter.hasNext()) {
		ADeclaratorTail dtail=(ADeclaratorTail) iter.next();
		iterName = dtail.getName().toString().trim();
		
		// We have to call for the deletion of the localy defined iterator variables ...
		removeVariableDeclaration(iterName, iteratorType);
	    }
	} 
    
	
	
    }
    

    
    public void caseALetExpression(ALetExpression le) {
	AExpression e = (AExpression) le.getExpression();
	
	 // simulates the node which represents the in-Expression ...
	Node inExpression = getFollowingNode(le);

	String letVarName = le.getName().toString().trim();
	Type letVarType = tree.getNodeType(e);
	translateLetExpression(le, letVarName, letVarType, e, inExpression);
	
	// We have to call for the deletion of the localy defined iterator variables ...
	removeLetVariableDeclaration(letVarName, letVarType);
	
    }
    
    

    public void outALitColPrimaryExpression(ALitColPrimaryExpression lcpe) {
	reachThrough(lcpe, lcpe.getLiteralCollection());
    }
    
    public void outALiteralPrimaryExpression(ALiteralPrimaryExpression lpe) {
	reachThrough(lpe, lpe.getLiteral());
    }
    
    public void outAParenthesesPrimaryExpression(AParenthesesPrimaryExpression lpe) {
	reachThrough(lpe, lpe.getExpression());
    }
    
    public void outAIfPrimaryExpression(AIfPrimaryExpression ipe) {
	reachThrough(ipe, ipe.getIfExpression());
    }
    
    

    public void outAStringLiteral(AStringLiteral sl) {
	String lit=sl.toString().trim();
	lit=lit.substring(1, lit.length()-1); // remove "'"
	translateStringLiteral(sl, lit);
    }
    
    public void outARealLiteral(ARealLiteral rl) {
	String lit=rl.toString().trim();
	translateRealLiteral(rl, lit);
    }
    
    public void outAIntegerLiteral(AIntegerLiteral il) {
	String lit=il.toString().trim();
	translateIntegerLiteral(il, lit);
    }
    
    public void outABooleanLiteral(ABooleanLiteral bl) {
	String lit = bl.toString().trim();
	translateBooleanLiteral(bl, lit);
    }
    
    
    public void outAEnumLiteral(AEnumLiteral el) {
	error("OCL-Transformation does not support Enumeration type !");
    }

    public void outAEnumerationType(AEnumerationType et) {
	error("OCL-Transformation does not support Enumeration type !");
    }


    public void outALiteralCollection(ALiteralCollection lc) {
	Type collType = tree.getNodeType(lc);
	AExpressionListOrRange elor = (AExpressionListOrRange) lc.getExpressionListOrRange();
	Vector elements = new Vector();

	if (elor == null){
	    // we have an empty collection definition ...
	    translateCollectionLiteralWithExpression(lc, collType, elements);
	} else {
	    AExpression firstExpression, nextExpression;
	    firstExpression = (AExpression) elor.getExpression();
	    elements.add(firstExpression);

	    PExpressionListOrRangeTail elort = elor.getExpressionListOrRangeTail();
	    if (elort == null){
		// we have a collection with one element ...
		translateCollectionLiteralWithExpression(lc, collType, elements);
	    } else {
		if (elort instanceof AListExpressionListOrRangeTail){
		    // we have a list of elements ...
		    AListExpressionListOrRangeTail list = (AListExpressionListOrRangeTail) elort;
		    Iterator iter=list.getExpressionListTail().iterator();
		    AExpressionListTail elt;
		    while (iter.hasNext()) {
			elt = (AExpressionListTail) iter.next();
			nextExpression = (AExpression) elt.getExpression();
			elements.add(nextExpression);
		    }
		    
		    translateCollectionLiteralWithExpression(lc, collType, elements);

		} else {
		    // we have a specified range ...
		    ARangeExpressionListOrRangeTail range = (ARangeExpressionListOrRangeTail) elort;
		    nextExpression = (AExpression) range.getExpression();
		    translateCollectionLiteralWithRange(lc, collType, firstExpression, nextExpression);
		}
		
	    }
	       
	}
	
    }
    
    public void outAQualifiers(AQualifiers q) {
	error("QULIFICATION NOT SUPPORTED by OCL-Transformation");
    }
    


    




    /** @return the node that the APostfixExpressionTail is applied to, which must
     *          be either an PPrimaryExpression or another APostfixExpressionTail
     *          node
     */
    protected Node getPreviousNode(APostfixExpressionTail pet) {
	Node ret;
	APostfixExpression pe=(APostfixExpression)pet.parent();
	ListIterator iter=pe.getPostfixExpressionTail().listIterator();
	while (iter.next()!=pet){}
	iter.previous();
	if (iter.hasPrevious()) {
	    ret=(APostfixExpressionTail) iter.previous();
	} else {
	    ret=pe.getPrimaryExpression();
	}

	return ret;
    }

    
    protected Node getFollowingNode(ALetExpression ale) {
	Node ret;
	AExpression ae=(AExpression)ale.parent();
	ListIterator iter=ae.getLetExpression().listIterator();
	while (iter.next()!= ale){}
	if (iter.hasNext()) {
	    ret = (Node) iter.next();
	} else {
	    ret = (Node) ae.getLogicalExpression();
	}
	return ret;
    }
    



    // ================== Some methods for errors ================================
    
    // a method for constructing error messages ...
    protected void error(String msg) throws TranslationException {
	throw (new TranslationException(msg));
    }
    
    void error(String msg, Node node) throws TranslationException {
	throw(new TranslationException(msg,node));
    }

    // ================ Start the translation ====================================
    
    
    public TranslationResult getTranslationFor(AExpression expression){
	this.expression = expression;
	
	TranslationResult result;
	
	expression.apply(this);
	result = getTranslationForNode(expression);
	return result;
    }

    // ============== Some checks on the given AST tree (when necessary) ===================
    
    protected void requireTreeInvariants(){
	log("REQUIRING invariants on the given AST.");
    }
     


    // =============== Translation methods for the several features and constructs ==========


    // the translations for the features of OCL types ...
        
    protected void translateFeatureCallonInteger(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow ){
	log("Feature call on Integer : " + featureName);
	log("Applied on node :" + n.toString());
    }
    
    protected void translateFeatureCallonReal(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow ){
	log("Feature call on Real : " + featureName);
	log("Applied on node :" + n.toString());
    }

    protected void translateFeatureCallonString(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow ){
	log("Feature call on String : " + featureName);
	log("Applied on node :" + n.toString());
    }

    protected void translateFeatureCallonBoolean(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow ){
	log("Feature call on Boolean : " + featureName);
	log("Applied on node :" + n.toString());
    }
    

    protected void translateFeatureCallonSet(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames,Type iteratorType, boolean  featureCallWithAtPre){
	log("Feature call on Set : " + featureName);
	log("Applied on node :" + n.toString());
    }

    protected void translateFeatureCallonBag(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames,Type iteratorType , boolean  featureCallWithAtPre){
	log("Feature call on Bag : " + featureName);
	log("Applied on node :" + n.toString());
    }

    protected void translateFeatureCallonSequence(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames,Type iteratorType, boolean  featureCallWithAtPre){
	log("Feature call on Sequence : " + featureName);
	log("Applied on node :" + n.toString());
    }
    
    protected void translateFeatureCallonOclType(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow,  boolean isMethodCall  ){
	log("Feature call on OclType : " + featureName);
	log("Applied on node :" + n.toString());
    }
    
    protected void translateFeatureCallonModelType(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow,  boolean isMethodCall ){
	log("Feature call on UML Model type : " + featureName);
	log("Applied on node :" + appliedTo.toString());
	
    }


    protected void translateVariable(Node n, String varName , Type varType){
	String varTypeString = varType.toString();
	log("Translating variable '"+ varName +"' of type '"+ varTypeString +"'.");
	
    }
 
    protected void translateVariableDeclaration(String varName, Type varType){
    	log("Translating Variable definition for Var: '" + varName + "'.");
    } 

    protected void removeVariableDeclaration(String varName, Type varType){
    	log("Removing Variable definition for Var: '" + varName + "'.");
    } 

    // and, or, xor, implies
    protected TranslationResult translateLogicalExpression(Node le, String logicalOperatorName, TranslationResult leftOperand, TranslationResult rightOperand){
	log("Translating logical expression with Operator '"+ logicalOperatorName  +"'.");
	return null;
    }
    
    // =, <>, <, <=, >, >=
    protected void translateRelationalExpression(Node re, String relationalOperatorName, Node leftOperand, Node rightOperand){
	log("Translating relational expression with Operator '"+ relationalOperatorName  +"'.");
    }
    
    // +, -
    protected TranslationResult translateAdditiveExpression(Node ae, String additiveOperatorName,
							    TranslationResult leftOperand,
							    TranslationResult rightOperand){
	log("Translating additive expression with Operator '"+ additiveOperatorName  +"'.");
	return null;
    }

    // *, /
    protected TranslationResult translateMultiplicativeExpression(Node me,
								  String multiplicativeOperatorName,
								  TranslationResult leftOperand,
								  TranslationResult rightOperand){
	log("Translating multiplicative expression with Operator '"+ multiplicativeOperatorName  +"'.");
	return null;
    }

    // -(...), not
    protected TranslationResult translateUnaryExpression(Node ue,
					    String unaryOperatorName,
					    TranslationResult operand){
	log("Translating unary expression with Operator '"+ unaryOperatorName  +"'.");
	return null;
    }
    


    protected void translateLetExpression(Node le, String letVarName,Type letVarType, Node e, Node inExpression ){
	log("Translating LET-construct with Variable '" + letVarName + "'.");
    };

    protected void translateLetVariableDeclaration(Node currentLetExpression, String letVarName, Type letVarType){
    	log("Translating LET-Variable definition for let-Var: '" + letVarName + "'.");
    } 


    protected void removeLetVariableDeclaration(String letVarName, Type letVarType){
    	log("Removing LET-Variable definition for Var: '" + letVarName + "'.");
    }

    protected void translateIfThenElseExpression(Node ie, Node ifExpression, Node thenExpression,
						 Node elseExpression){
	log("Translating if-then-else-construct.");
    };
    
    protected void translateStringLiteral(Node sl, String lit){
	log("Translating String literal: " + lit);
    }

    protected void translateIntegerLiteral(Node sl, String lit){
	log("Translating Integer literal: " + lit);
    }

    protected void translateRealLiteral(Node sl, String lit){
	log("Translating Real literal: " + lit);
    }

    protected void translateBooleanLiteral(Node sl, String lit){
	log("Translating Boolean literal: " + lit);
    }
    
     protected void translateOclTypeLiteral(Node sl, String lit){
	log("Translating OclType literal: " + lit);
    }

    protected void translateCollectionLiteralWithExpression(Node cl, Type collType, Vector expressions){
	log("Translating Collection literal with expressions for type: " + collType.toString());
    }
    
    
    protected void translateCollectionLiteralWithRange(Node cl, Type collType, Node lowerBound,
						       Node upperBound){
	log("Translating Collection literal with range for type: " + collType.toString());
    }
    


    
    protected void log(String text){
	System.err.println(text);
	System.err.println("--------------------------------------------------------------");
    }

   
    //  public void defaultIn(Node node)
//      {
//  	log("Entering node :" + node.toString() + "("+node.getClass().toString()+")");
//      }

//      public void defaultOut(Node node)
//      {
//  	log("Leaving node :" + node.toString() + "("+node.getClass().toString()+")");
//      }

    

}






