// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.ocl.ocldltranslator.control;

import tudresden.ocl.*;
import tudresden.ocl.parser.analysis.*;
import java.util.Set;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Vector;
import tudresden.ocl.parser.node.*;
import tudresden.ocl.check.types.*;

import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.heuristics.*;

/**
 * A base class for all classes implementing a translation control service for
 * the translation of an arbitrary OCL expression into some logical language.
 *
 * It changes the behaviour of the getTranslationType() method in order to
 * get a DEFAULT type when the corresponding node was not marked at all and
 * adds an optional logging mechanism.
 * Additionally, it is possible to activate and deactivate heuristics at runtime.
 * Only active heuristics will be applied on the expression to be translated 
 * when executing all heuristics.
 */

public abstract class DefaultTranslationControl extends BasicTranslationControl {
   
    private final Object ACTIVE = new Object();
    private final Object INACTIVE = new Object();

    boolean logEnabled = false;

    Vector activationStates;

    
    public DefaultTranslationControl(){
	super();
	activationStates = new Vector();
    }
    

    /** 
     * Adds a new heuristic to the translation control if it isn't already present.
     * The element will be added as the last element in the list and
     * set active.
     */
    public void  addHeuristic(BasicTranslationHeuristic th){
	addHeuristic(th, true);
    }
    
    /** 
     * Adds a new heuristic to the translation control if it isn't already present.
     * The element will be added as the last element in the list and
     * set active.
     * If the element is already present the activation state will be updated 
     * accordingly.
     */
    public void  addHeuristic(BasicTranslationHeuristic th, boolean activationState){
	if (!registeredHeuristics.contains(th)){
	    registeredHeuristics.add(th);
	    activationStates.add((activationState ? ACTIVE : INACTIVE));   
	} else {
	    int index = registeredHeuristics.indexOf(th);
	    activationStates.setElementAt( (activationState ? ACTIVE : INACTIVE) , index);
	}
    }

    /** 
     * Removes a heuristic from the translation control if it has been registered before.
     */
    public void removeHeuristic(BasicTranslationHeuristic th){
	if (registeredHeuristics.contains(th)){
	    int index = registeredHeuristics.indexOf(th);
	    registeredHeuristics.removeElementAt(index);
	    activationStates.removeElementAt(index);
	}
    }

    /** 
     * @return the list of all registered heuristics.
     * Each such object is an instance of BasicTranslationHeuristic or some sublcass.
     */
    public Vector getRegisteredHeuristics(){
	return registeredHeuristics;
    }

    /** 
     * @return the list of all registered heuristics that are currently active.
     * Each such object is an instance of BasicTranslationHeuristic or some sublcass.
     */
    public Vector getActiveHeuristics(){
	Vector result = new Vector();
	BasicTranslationHeuristic currentHeuristic;
	boolean isActive;
	for(int i=0; i < registeredHeuristics.size(); i++){
	    currentHeuristic = (BasicTranslationHeuristic) registeredHeuristics.elementAt(i);
	    isActive = (activationStates.elementAt(i) == ACTIVE);
	    if (isActive) {
		result.add(currentHeuristic);
	    }
	}
	return result;
    }

    /**
     * Tests whether the given heuristic is an active registered translation
     * heuristic for this translation control instance.
     * @return true iff the given heuristic is a registered translation heuristic
     *                  that currently is active.
     */
     public boolean isActiveHeuristic(BasicTranslationHeuristic th){
	 boolean result;
	if (registeredHeuristics.contains(th)){
	    int index = registeredHeuristics.indexOf(th);
	    result = (activationStates.elementAt(index) == ACTIVE);
	} else {
	    result = false;
	}
	return result;
    }

    /**
     * Changes the activation state of some registered translation
     * heuristic for this translation control instance.
     * @param th - the translation heuristic whose activation state should be changed.
     * @param newActivationState - specifies the new activation state of the given heuristic
     *                             if this heuristic has been registered before.
     */
     public void setHeuristicActivation(BasicTranslationHeuristic th, boolean newActivationState){
	if (registeredHeuristics.contains(th)){
	    int index = registeredHeuristics.indexOf(th);
	    activationStates.setElementAt((newActivationState ? ACTIVE : INACTIVE), index );
	} 
    }
    

    // basic operations for the handling of translation type for nodes ...
   

    /**
     * Returns the translation type specification for the given node.
     */
     public TranslationType getTranslationTypeForNode(Node n){
	
	if(n == null){ 
	    error("Trying to get a translation type for a 'null' node !");
	}
	
	if (!nodeTranslationTypeMap.containsKey(n)) {
	    return ((new TranslationType()).addMark(TranslationMark.DEFAULT));
	    
	}
	return (TranslationType) nodeTranslationTypeMap.get(n); 
    }


    /**
     * Executes all the registered heuristics in the order they have been
     * registered on the given OCL expression that belongs the the specified
     * constraint tree.
     * @param constraintTree - the tree representation of the OCL constraint to be handled.
     * @param expression - the OCL expression to be translated and therefore marked by the
     *                     registered heuristics.
     * @param exprTranslator - the translator object to be used for the overall translation.
     */
    public void executeHeuristics(OclTree constraintTree, AExpression expression, 
				  KeyOclExpressionTranslator exprTranslator){
	this.nodeTranslationTypeMap = new HashMap();
	
	log("TranslationControl: Executing heuristics on OCL expression: " + expression.toString());

	BasicTranslationHeuristic currentHeuristic;
	boolean isActive;

	for(int i=0; i < registeredHeuristics.size(); i++){
	    currentHeuristic = (BasicTranslationHeuristic) registeredHeuristics.elementAt(i);
	    isActive = (activationStates.elementAt(i) == ACTIVE);
	    if (isActive) {
		log("TranslationControl -- Apply active heuristic : " + currentHeuristic.getHeuristicName());
		if (currentHeuristic instanceof KeyTranslationHeuristic) {
		    ((KeyTranslationHeuristic)currentHeuristic).setLoggingEnabled(this.logEnabled);
		}

		// Start current heuristic ...
		currentHeuristic.determineTranslationTypesFor(constraintTree, expression, 
							      nodeTranslationTypeMap, exprTranslator);
		
		// Get new markings of the constraint tree ...
		nodeTranslationTypeMap = currentHeuristic.getMarkingOfConstraintTree();
	    } else {
		log("TranslationControl -- Omitting inactive heuristic : " + currentHeuristic.getHeuristicName());
	    }
	}

	log("TranslationControl: Finished executing heuristics!");

    }



    /**
     * Enable or disable logging.
     */
    public void setLoggingEnabled(boolean log){
	logEnabled = log;
    }

    
    /**
     * Logs the given text if logging has been enabled.
     */

    protected void log(String text){
	if (logEnabled) {
	    super.log(text);
	}
    }

}

















