// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.ocl.ocldltranslator.heuristics;

/**
 * A base class for all classes implementing a heuristic for
 * a suitable control of the translation of an arbitrary OCL expression.
 *
 * It especially provides basic functionality for the
 * management of the node-nodetranslationmark-mapping.
 *
 * It also extends DepthFirstAdapter an therefore implements a basic tree
 * walking procedure.
 */
import tudresden.ocl.*;
import tudresden.ocl.parser.analysis.*;
import java.util.Set;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Vector;
import tudresden.ocl.parser.node.*;
import tudresden.ocl.check.types.*;

import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.control.*;

/**
 * This class is the base class for all classes that represent and implement some
 * heuristic that affects the way some fragment of OCL will be treated by the translation.
 * It basically implements the tree walk through the tree representation of the OCL constraint
 * to be translated.
 * 
 * Subclasses have to overwrite the necessary translate...()- Methods!
 */ 

public abstract class BasicTranslationHeuristic extends DepthFirstAdapter {

    protected AExpression expression;
    protected OclTree tree;
    protected KeyOclExpressionTranslator exprTranslator;

    // This map maps nodes in the AST to a translation type which specifies the way the logical translation
    // of that node will be generated when executing
    // the translation.
    // It uses Nodes as keys and delivers translation type objects representing the kind of translation 
    // to be performed for that node translation ...
    protected Map nodeTranslationTypeMap;
    

    // Maps variable(names) to types to be used for representing them or expressions definied for them 
    // (definied expressions in let-construct)  ...
    protected Map varMap;
 
   
    // basic operations for the handling of translation type for nodes ...
    
    /**
     * @return the TranslationType object that describes how the given node
     *         should be processed when translating the corresponding OCL expression.
     */
    protected TranslationType getTranslationTypeForNode(Node n){
	
	if(n == null){ 
	    error("Trying to get a translation type for a 'null' node !");
	}
	
	if (!nodeTranslationTypeMap.containsKey(n)) {
	   throw new RuntimeException("Translation type for node ("+ n +") required but not found !!!");
	    
	}
	return (TranslationType) nodeTranslationTypeMap.get(n); 
    }


    protected void setTranslationTypeForNode(Node n, TranslationType translation){
	if (translation != null) {
	    nodeTranslationTypeMap.put(n, translation);
	} else {
	    error("Trying to set 'null' as the translation of the node :" + n );
	}
    }

    
    
    // basic operations for the handling of translation of variables ...
    
    protected TranslationType getTranslationTypeForVariable(String varName, String typeName){
	String varKey = varName + "$$$" + typeName;

	if (!varMap.containsKey(varKey)) {
	    throw new RuntimeException("Translation type for variable ("+ varName + ":" +
				       typeName+") required but not found !!!");
	    
	}
	return (TranslationType) varMap.get(varKey); 
    }

    protected void setTranslationTypeForVariable(String varName, String typeName, TranslationType translationType){
	String varKey = varName + "$$$" + typeName;
	varMap.put(varKey, translationType);
    }


    protected void removeTranslationTypeForVariable(String varName, String typeName){
	String varKey = varName + "$$$" + typeName;
	varMap.remove(varKey);
    }
    
    
    
    // public Map getTranslationTypeMap(){
    // return nodeTranslationTypeMap;
    // }

    /**
     * @return the name of this translation heuristic.
     */
    public abstract String getHeuristicName();

    // =========== tree traversal operations ===================

     public void defaultIn(Node node) {
	 
     }


    

    public void caseAExpression(AExpression e) {
	// We have to change the order of evaluation of the subnodes ...

	log("Case AExpr in BasicTranslationControl for  == " + e);
	ALetExpression currentLetExpression = null;
	
	// Then translate the logical expression ...
	ALogicalExpression alo = (ALogicalExpression) e.getLogicalExpression();
	super.caseALogicalExpression(alo);
	
	if (!e.getLetExpression().isEmpty()) {
	    // Then translate the let expressions in reversed order ...
	    ListIterator iter = e.getLetExpression().listIterator();
	
	    while (iter.hasNext()){
		currentLetExpression =(ALetExpression) iter.next();
	    }
	    // Now we have the last Let expression ...
	    super.caseALetExpression(currentLetExpression);
	    while (iter.hasPrevious()){
		currentLetExpression = (ALetExpression)iter.previous();
		this.caseALetExpression(currentLetExpression);
	    }
	    
	}
	
    }

    
    /** 
     * if a then b else c endif    
     */
    public void inAIfExpression(AIfExpression ie) {
	PExpression ifExpression = ie.getIfBranch();
	PExpression thenExpression = ie.getThenBranch();
	PExpression elseExpression = ie.getElseBranch();
	
	translateIfThenElseExpression(ie,ifExpression, thenExpression, elseExpression);	

	defaultIn(ie);

    }

    /** 
     *  AND, OR and XOR have precedence
     *  over IMPLIES.
     *  AND, OR and XOR have the same precedende.
     */
  public void inALogicalExpression(ALogicalExpression le) {
    if (le.getLogicalExpressionTail().isEmpty()) {
	// we have no 'real' logical expression ...
	
    } else {
	// we have a logical expression ...
	// Take care of the operator precedence rules !!!!

	int tailElements=le.getLogicalExpressionTail().size(); // this is >= 1 ...
	PRelationalExpression[] relExprs=new PRelationalExpression[tailElements+1];
	PLogicalOperator[] logOps=new PLogicalOperator[tailElements];
	// logOps[i] is the operator AFTER relExprs[i]
	Iterator iter=le.getLogicalExpressionTail().iterator();
	int index=0;
	relExprs[0]=le.getRelationalExpression();
	while (iter.hasNext()) {
	    ALogicalExpressionTail let=(ALogicalExpressionTail)iter.next();
	    logOps[index]=let.getLogicalOperator();
	    index++;
	    relExprs[index]=let.getRelationalExpression();
	}
	
	handleLogicalExpression(le, relExprs, logOps, 0, tailElements - 1 );
	
    }

    defaultIn(le);

  }
    
    protected void handleLogicalExpression(ALogicalExpression le, PRelationalExpression[] relExprs, PLogicalOperator[] logOps, int from, int to){
	 
	// First find the leftmost IMPLIES operator ...
	int leftMostImplies = -1; // index of operand after last implies
	for (int i= from; (i <= to) && (leftMostImplies == -1) ; i++){
	    if(logOps[i] instanceof AImpliesLogicalOperator){
		leftMostImplies = i;
	    }
	}

	if (leftMostImplies != -1){
	    // There is an IMPLIES operator => take care about precedence ...
	     // First translate left and right logical subexpressions ...

	    Node leftNode  = null;
	    Node rightNode = null; 
	    

	    if (from == leftMostImplies) {
		leftNode = relExprs[from];
	    } else {
		handleLogicalExpression(le, relExprs, logOps,
					from, leftMostImplies - 1);
	    }
	    
	    if (to == leftMostImplies) {
		rightNode = relExprs[to + 1];
	    } else {
		handleLogicalExpression(le, relExprs, logOps,
					leftMostImplies + 1, to);
	    }

	    
	    // Now translate the IMPLIES operator ...
	    
	    translateLogicalExpression(le, logOps[leftMostImplies].toString().trim(),
				       leftNode,
				       rightNode);
	    

	} else {
	    // There is no IMPLIES operator, therefore all logical operators have
	    // the same precedence -> Build term from left to right !
	    
	    Node  leftOperand, rightOperand;
	    
	    leftOperand = relExprs[from];

	    for (int i= from; (i <= to) ; i++){
		PLogicalOperator currentOperator = logOps[i];
		rightOperand = relExprs[i+1];
		
		translateLogicalExpression(le, currentOperator.toString().trim(),
					   leftOperand, rightOperand);
		leftOperand = null;
	    
	    }

	}
	 
    }
	    
       
    
    public void inARelationalExpression(ARelationalExpression re) {
	if (re.getRelationalExpressionTail()==null) {
	    // we have no 'real' relational expression ...
	   
	} else {
	    ARelationalExpressionTail ret=
		(ARelationalExpressionTail) re.getRelationalExpressionTail();
	    PRelationalOperator relationalOp=ret.getRelationalOperator();
	    
	    AAdditiveExpression leftOperand = (AAdditiveExpression) re.getAdditiveExpression();
	    AAdditiveExpression rightOperand = (AAdditiveExpression) ret.getAdditiveExpression();
	    String opName = relationalOp.toString().trim(); 
	    
	    translateRelationalExpression(re, opName, leftOperand, rightOperand);
	    
	}

	defaultIn(re);
    }


    public void inAAdditiveExpression(AAdditiveExpression ae) {
	if (ae.getAdditiveExpressionTail().isEmpty()) {
	    // we have no 'real' additive expression ...
	   
	} else {
	    // We have + or - which have both the same precedence.
	    // Therefore translate from left to right ...
	    
	    AMultiplicativeExpression leftOperand = (AMultiplicativeExpression) ae.getMultiplicativeExpression();
	    AMultiplicativeExpression rightOperand;
	    PAddOperator currentOperator;

	    Node leftNode, rightNode;
	    
	    leftNode = leftOperand; 

	    Iterator iter = ae.getAdditiveExpressionTail().iterator();
	    while (iter.hasNext()) {
		AAdditiveExpressionTail aet = (AAdditiveExpressionTail) iter.next();
		rightOperand = (AMultiplicativeExpression) aet.getMultiplicativeExpression();
		rightNode = rightOperand;
		currentOperator = aet.getAddOperator();
		
		translateAdditiveExpression(ae, currentOperator.toString().trim(), leftNode,
					    rightNode);
		leftNode = null;
	    }
	    
	}

	defaultIn(ae);
    }
    
    
    public void inAMultiplicativeExpression(AMultiplicativeExpression me) {
	if (me.getMultiplicativeExpressionTail().isEmpty()) {
	    // we have no 'real' multiplicative expression ...
	   
	} else {

	    // We have * or / which have both the same precedence.
	    // Therefore translate from left to right ...
	    
	    PUnaryExpression leftOperand = me.getUnaryExpression();
	    PUnaryExpression rightOperand;
	    PMultiplyOperator currentOperator;

	    Node leftNode, rightNode;
	    
	    leftNode = leftOperand; 

	    Iterator iter = me.getMultiplicativeExpressionTail().iterator();
	    while (iter.hasNext()) {
		AMultiplicativeExpressionTail met = (AMultiplicativeExpressionTail) iter.next();
		rightOperand = met.getUnaryExpression();
		rightNode = rightOperand;
		currentOperator = met.getMultiplyOperator();

		translateMultiplicativeExpression(me, currentOperator.toString().trim(),
						  leftNode,
						  rightNode);
		leftNode = null;
	    }
	   
	}
	
	defaultIn(me);

    }	
	/**
	 * not , unary - ...
	 */
	public void inAUnaryUnaryExpression(AUnaryUnaryExpression uue) {
	    PUnaryOperator currentOperator = uue.getUnaryOperator();
	    PPostfixExpression pe = uue.getPostfixExpression();
	    Node operand = pe;
	    
	    translateUnaryExpression(uue, currentOperator.toString().trim(),
				     operand);
	   
	    defaultIn(uue);
	}



    public void inAFeaturePrimaryExpression(AFeaturePrimaryExpression pe) {
	Type oclType=tree.getNodeType(pe);
	if (oclType instanceof OclType) {
	    // We have an OclType literal ...
	    String lit=pe.toString().trim();
	    translateOclTypeLiteral(pe, lit);
	} else {
	    String varName = pe.toString().trim();
	    Type varType = oclType;
	    translateVariable(pe , varName , varType);
	}

	defaultIn(pe);

    }

 

    public void inAPostfixExpressionTail(APostfixExpressionTail pet) {
	Node appliedTo=getPreviousNode(pet);
	Type appliedType=tree.getNodeType(appliedTo);
	
	
	AFeatureCall fc=(AFeatureCall)pet.getFeatureCall();
	boolean featureCallWithAtPre = false;
	if (fc.getTimeExpression()!=null) {
	    // we have a time expression (@pre was set for the feature) ...
	    featureCallWithAtPre = true;
	}
	
	boolean featureCallWithArrow = (pet.getPostfixExpressionTailBegin() instanceof
					AArrowPostfixExpressionTailBegin);
	
	Vector parameterNodes = new Vector();
	Vector iteratorNames = new Vector();
	String accumulatorName;

	boolean isMethodCall = false;
	
	String featureName=fc.getPathName().toString().trim();
	Type iteratorType = null;
	
	AFeatureCallParameters fcp=(AFeatureCallParameters)fc.getFeatureCallParameters();
	AActualParameterList apl=null;
	if (fcp!=null) {
	    apl=(AActualParameterList)fcp.getActualParameterList();
	    isMethodCall = true;
	}
	
	ListIterator iter;
	
	if (featureName.equals("iterate")) {
	    // The iterate case is a bit more complicated ...

	    // Do declarations for the iterator variables ...
	    AIterateDeclarator decl = (AIterateDeclarator) fcp.getDeclarator();
	    String iterName = decl.getIterator().toString().trim();
	    iteratorType = tree.getTypeFor(iterName , apl); 

	    log("Determined type of it-variable '"+iterName+"' as :"+iteratorType.toString());
	    iteratorNames.add(iterName);
	    
	    // Do declarations for the accumulator variables ...
	   
	    String accName = decl.getAccumulator().toString().trim();
	    Type accumulatorType = tree.getTypeFor(accName , apl); 

	    log("Determined type of acc-variable '"+accName+"' as :"+accumulatorType.toString());
	    iteratorNames.add(accName);
	
	   
	} else if (tudresden.ocl.check.TypeChecker.setOfIteratingMethodNames.contains(featureName)) {

	    // Get all names of iterator variables ...
	    AStandardDeclarator decl = (AStandardDeclarator) fcp.getDeclarator();
	    String iterName = decl.getName().toString().trim();
	    iteratorType = tree.getTypeFor(iterName , apl); 

	    iteratorNames.add(iterName);


	    // If there are several iterators then also add their names ...
	    iter=decl.getDeclaratorTail().listIterator();
	    ADeclaratorTail dtail;
	    while (iter.hasNext()) {
		dtail=(ADeclaratorTail) iter.next();
		iterName = dtail.getName().toString().trim();
		iteratorNames.add(iterName);
		
	    }
	} 
	
	// Now get all parameters for the feature call ...
	
	if (apl != null){
	    // there are some parameters ...
	    Node parameter = apl.getExpression();
	    parameterNodes.add(parameter);
	    // If there are several parameters then also add their nodes ...
	    iter = apl.getActualParameterListTail().listIterator();
	    AActualParameterListTail apltail;
	    while (iter.hasNext()) {
		apltail=(AActualParameterListTail) iter.next();
		parameter = apltail.getExpression();
		parameterNodes.add(parameter);
	    }
	}

	
	Type t = appliedType;
	// Distinguish the different cases ...
	if (t instanceof Basic) {
	    if (t==Basic.INTEGER) {
		translateFeatureCallonInteger(pet, featureName, appliedTo, parameterNodes,
					      iteratorNames, featureCallWithAtPre,featureCallWithArrow );
	    }
	    else if (t==Basic.REAL) {
		translateFeatureCallonReal(pet, featureName, appliedTo,  parameterNodes,
					   iteratorNames, featureCallWithAtPre,featureCallWithArrow );
	    }
	    else if (t==Basic.BOOLEAN) {
		translateFeatureCallonBoolean(pet, featureName, appliedTo,  parameterNodes,
					      iteratorNames, featureCallWithAtPre,featureCallWithArrow );
	    }
	    else if (t==Basic.STRING) {
		translateFeatureCallonString(pet, featureName, appliedTo,  parameterNodes,
					     iteratorNames, featureCallWithAtPre,featureCallWithArrow );
	    }
	    else throw new RuntimeException("illegal basic type");
	} else if (t instanceof Collection) {
	    Collection c=(Collection) t;
	    switch (c.getCollectionKind()) {
	    case Collection.SET:         	
		translateFeatureCallonSet(pet, featureName,  appliedTo, parameterNodes,
					  iteratorNames,iteratorType , featureCallWithAtPre);
		break;
		
	    case Collection.BAG:
		translateFeatureCallonBag(pet, featureName, appliedTo,  parameterNodes,
					  iteratorNames,iteratorType, featureCallWithAtPre);
		break;
		
	    case Collection.SEQUENCE:
		translateFeatureCallonSequence(pet, featureName, appliedTo,
					       parameterNodes, iteratorNames,
					       iteratorType,featureCallWithAtPre);
		break;
		
		//   case Collection.COLLECTION: 
		//  	  translateFeatureCallonCollection(pet, featureName, appliedTo,  parameterNodes, iteratorNames,iteratorType );
		//  	  break;
		
	    default: throw new RuntimeException("illegal collection type");
	    }
	} else if (t instanceof OclType) {
	    translateFeatureCallonOclType(pet, featureName, appliedTo,  parameterNodes,
					  iteratorNames, featureCallWithAtPre, featureCallWithArrow, isMethodCall);
	} else {
	    // application type
	    translateFeatureCallonModelType(pet, featureName, appliedTo,  parameterNodes,
					    iteratorNames, featureCallWithAtPre, featureCallWithArrow, isMethodCall);
	}

	
	defaultIn(pet);
	
    }
    

    public void inALetExpression(ALetExpression le) {
	AExpression e = (AExpression) le.getExpression();
	
	 // simulates the node which represents the in-Expression ...
	Node inExpression = getFollowingNode(le);

	String letVarName = le.getName().toString().trim();
	Type letVarType = tree.getNodeType(e);
	translateLetExpression(le, letVarName, letVarType, e, inExpression);

	
    }
    
    

    public void inAStringLiteral(AStringLiteral sl) {
	String lit=sl.toString().trim();
	lit=lit.substring(1, lit.length()-1); // remove "'"
	translateStringLiteral(sl, lit);
	
	defaultIn(sl);
    }
    
    public void inARealLiteral(ARealLiteral rl) {
	String lit=rl.toString().trim();
	translateRealLiteral(rl, lit);

	defaultIn(rl);
    }
    
    public void inAIntegerLiteral(AIntegerLiteral il) {
	String lit=il.toString().trim();
	translateIntegerLiteral(il, lit);

	defaultIn(il);
    }
    
    public void inABooleanLiteral(ABooleanLiteral bl) {
	String lit = bl.toString().trim();
	translateBooleanLiteral(bl, lit);
    
	defaultIn(bl);
    }
    
    
    public void inAEnumLiteral(AEnumLiteral el) {
	error("OCL-Transformation does not support Enumeration type !");
    }

    public void inAEnumerationType(AEnumerationType et) {
	error("OCL-Transformation does not support Enumeration type !");
    }


    public void inALiteralCollection(ALiteralCollection lc) {
	Type collType = tree.getNodeType(lc);
	AExpressionListOrRange elor = (AExpressionListOrRange) lc.getExpressionListOrRange();
	Vector elements = new Vector();

	if (elor == null){
	    // we have an empty collection definition ...
	    translateCollectionLiteralWithExpression(lc, collType, elements);
	} else {
	    AExpression firstExpression, nextExpression;
	    firstExpression = (AExpression) elor.getExpression();
	    elements.add(firstExpression);

	    PExpressionListOrRangeTail elort = elor.getExpressionListOrRangeTail();
	    if (elort == null){
		// we have a collection with one element ...
		translateCollectionLiteralWithExpression(lc, collType, elements);
	    } else {
		if (elort instanceof AListExpressionListOrRangeTail){
		    // we have a list of elements ...
		    AListExpressionListOrRangeTail list = (AListExpressionListOrRangeTail) elort;
		    Iterator iter=list.getExpressionListTail().iterator();
		    AExpressionListTail elt;
		    while (iter.hasNext()) {
			elt = (AExpressionListTail) iter.next();
			nextExpression = (AExpression) elt.getExpression();
			elements.add(nextExpression);
		    }
		    
		    translateCollectionLiteralWithExpression(lc, collType, elements);

		} else {
		    // we have a specified range ...
		    ARangeExpressionListOrRangeTail range = (ARangeExpressionListOrRangeTail) elort;
		    nextExpression = (AExpression) range.getExpression();
		    translateCollectionLiteralWithRange(lc, collType, firstExpression, nextExpression);
		}
		
	    }
	       
	}
	
	defaultIn(lc);

    }
    
    public void inAQualifiers(AQualifiers q) {
	error("QULIFICATION NOT SUPPORTED by OCL-Transformation");
    }
    


    

    /** @return the node that the APostfixExpressionTail is applied to, which must
     *          be either an PPrimaryExpression or another APostfixExpressionTail
     *          node
     */
    protected Node getPreviousNode(APostfixExpressionTail pet) {
	Node ret;
	APostfixExpression pe=(APostfixExpression)pet.parent();
	ListIterator iter=pe.getPostfixExpressionTail().listIterator();
	while (iter.next()!=pet){}
	iter.previous();
	if (iter.hasPrevious()) {
	    ret=(APostfixExpressionTail) iter.previous();
	} else {
	    ret=pe.getPrimaryExpression();
	}

	return ret;
    }

    
    protected Node getFollowingNode(ALetExpression ale) {
	Node ret;
	AExpression ae=(AExpression)ale.parent();
	ListIterator iter=ae.getLetExpression().listIterator();
	while (iter.next()!= ale){}
	if (iter.hasNext()) {
	    ret = (Node) iter.next();
	} else {
	    ret = (Node) ae.getLogicalExpression();
	}
	return ret;
    }
    



    // ================== Some methods for errors ================================
    
    // a method for constructing error messages ...
    void error(String msg) throws TranslationException {
	throw (new TranslationException(msg));
    }
    
    void error(String msg, Node node) throws TranslationException {
	throw(new TranslationException(msg,node));
    }



    // ================ Start the computation of the translation types =====================
    
    /**
     * Starts the heuristic on some expression in some OCL constraint whereby the given 
     * marks (or colors) have already been applied to the tree.
     * @param constraintTree - the constraint tree that contains the expression to be marked.
     * @param expression - the expression (in the whole constraint) to be marked. Usually this is
     *                     some invariant condition or a pre- or post condition.
     * @param currentMarking - the marking that has already been applied to the given constraint tree.
     *                         The map relates nodes in the tree with their corresponding translation type
     *                         if such a type has already been specified.
     */ 
    
    public void determineTranslationTypesFor(OclTree constraintTree, 
					     AExpression expression, Map currentMarking,
					     KeyOclExpressionTranslator exprTranslator){
	// initialize state of translation ...
	this.expression = expression;
	this.tree = constraintTree;
	this.exprTranslator = exprTranslator;

	this.varMap = new HashMap();

	nodeTranslationTypeMap = currentMarking;
       
	expression.apply(this);
	
    }

    /**
     * @return a map the relates the nodes in the constraint tree with their translation type to
     *         be used these nodes.
     */
    public Map getMarkingOfConstraintTree(){
	return nodeTranslationTypeMap; 
    }



    // =============== Translation methods for the several features and constructs ==========
    // REMARK : These methods are similar BUT NOT equal to the methods in BasicOclExpressionTranslator !!!
    // But their intention is the same; especially they are called in a top-down manner whereas in 
    // BasicOclExpressionTranslator those methods are called in a bottom up manner !

    // the translations for the features of OCL types ...
        
    protected void translateFeatureCallonInteger(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow ){
	log("Feature call on Integer : " + featureName);
	log("Applied on node :" + n.toString());
    }
    
    protected void translateFeatureCallonReal(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow ){
	log("Feature call on Real : " + featureName);
	log("Applied on node :" + n.toString());
    }

    protected void translateFeatureCallonString(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow ){
	log("Feature call on String : " + featureName);
	log("Applied on node :" + n.toString());
    }

    protected void translateFeatureCallonBoolean(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow ){
	log("Feature call on Boolean : " + featureName);
	log("Applied on node :" + n.toString());
    }
    

    protected void translateFeatureCallonSet(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames,Type iteratorType, boolean  featureCallWithAtPre){
	log("Feature call on Set : " + featureName);
	log("Applied on node :" + n.toString());
    }

    protected void translateFeatureCallonBag(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames,Type iteratorType , boolean  featureCallWithAtPre){
	log("Feature call on Bag : " + featureName);
	log("Applied on node :" + n.toString());
    }

    protected void translateFeatureCallonSequence(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames,Type iteratorType, boolean  featureCallWithAtPre){
	log("Feature call on Sequence : " + featureName);
	log("Applied on node :" + n.toString());
    }
    
    protected void translateFeatureCallonOclType(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow, boolean isMethodCall){
	log("Feature call on OclType : " + featureName);
	log("Applied on node :" + n.toString());
    }
    
    protected void translateFeatureCallonModelType(Node n, String featureName, Node appliedTo, Vector parameterNodes, Vector iteratorNames, boolean  featureCallWithAtPre, boolean featureCallWithArrow, boolean isMethodCall){
	log("Feature call on UML Model type : " + featureName);
	log("Applied on node :" + appliedTo.toString());
	
    }


    protected void translateVariable(Node n, String varName , Type varType){
	String varTypeString = varType.toString();
	log("Translating variable '"+ varName +"' of type '"+ varTypeString +"'.");
	
    }
 
   
    // and, or, xor, implies
    protected void translateLogicalExpression(Node le, String logicalOperatorName, Node leftOperand, 
					      Node rightOperand){
	log("Translating logical expression with Operator '"+ logicalOperatorName  +"'.");
    }
    
    // =, <>, <, <=, >, >=
    protected void translateRelationalExpression(Node re, String relationalOperatorName, Node leftOperand, Node rightOperand){
	log("Translating relational expression with Operator '"+ relationalOperatorName  +"'.");
    }
    
    // +, -
    protected void translateAdditiveExpression(Node ae, String additiveOperatorName,
					       Node leftOperand,
					       Node rightOperand){
	log("Translating additive expression with Operator '"+ additiveOperatorName  +"'.");
    }

    // *, /
    protected void  translateMultiplicativeExpression(Node me,
						      String multiplicativeOperatorName,
						      Node leftOperand,
						      Node rightOperand){
	log("Translating multiplicative expression with Operator '"+ multiplicativeOperatorName  +"'.");
       
    }

    // -(...), not
    protected void translateUnaryExpression(Node ue,
					    String unaryOperatorName,
					    Node operand){
	log("Translating unary expression with Operator '"+ unaryOperatorName  +"'.");
       
    }
    


    protected void translateLetExpression(Node le, String letVarName,Type letVarType, Node e, Node inExpression ){
	log("Translating LET-construct with Variable '" + letVarName + "'.");
    };

    
    protected void translateIfThenElseExpression(Node ie, Node ifExpression, Node thenExpression,
						 Node elseExpression){
	log("Translating if-then-else-construct.");
    };
    
    protected void translateStringLiteral(Node sl, String lit){
	log("Translating String literal: " + lit);
    }

    protected void translateIntegerLiteral(Node sl, String lit){
	log("Translating Integer literal: " + lit);
    }

    protected void translateRealLiteral(Node sl, String lit){
	log("Translating Real literal: " + lit);
    }

    protected void translateBooleanLiteral(Node sl, String lit){
	log("Translating Boolean literal: " + lit);
    }
    
     protected void translateOclTypeLiteral(Node sl, String lit){
	log("Translating OclType literal: " + lit);
    }

    protected void translateCollectionLiteralWithExpression(Node cl, Type collType, Vector expressions){
	log("Translating Collection literal with expressions for type: " + collType.toString());
    }
    
    
    protected void translateCollectionLiteralWithRange(Node cl, Type collType, Node lowerBound,
						       Node upperBound){
	log("Translating Collection literal with range for type: " + collType.toString());
    }
    



    
    protected void log(String text){
	System.err.println(text);
	System.err.println("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~");
    }


    /**
     * Determines wheter the given object equals this instance.
     * Two heuristics are concerned equal iff they belong to the same (runtime) class.
     */
    public boolean equals(Object o){
	boolean result;

	if (o instanceof BasicTranslationHeuristic){
	    if (o != null){
		result = (this.getClass() == o.getClass());
	    } else {
		result = false;
	    }
	} else {
	    result = false;
	}
	
	return result;
    }


   
    //  public void defaultIn(Node node)
//      {
//  	log("Entering node :" + node.toString() + "("+node.getClass().toString()+")");
//      }

//      public void defaultOut(Node node)
//      {
//  	log("Leaving node :" + node.toString() + "("+node.getClass().toString()+")");
//      }

    

}






