// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.parser;

import junit.framework.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.java.JavaInfo;

import java.io.*;


public class TestDeclParser extends TestCase {

    Namespace sort_ns,var_ns,func_ns,heur_ns;

    public TestDeclParser(String name) {
	super(name);
    }

    public void setUp() {
	sort_ns = new Namespace();
	func_ns = new Namespace();
	var_ns = new Namespace();
	heur_ns = new Namespace();
    }

    private DeclParser stringParser(String s) {
	return new DeclParser(new KeYLexer(new StringReader(s)),
			      "No file. Call of parser from parser/TestDeclParser.java",
			      var_ns,func_ns,sort_ns,heur_ns, new JavaInfo());
    }

    public void parseDecls(String s) {
	try {
	    DeclParser p = stringParser(s);
	    p.decls();
	} catch (Exception e) {
	    StringWriter sw = new StringWriter();
	    PrintWriter pw = new PrintWriter(sw);
	    e.printStackTrace(pw);
	    throw new RuntimeException("Exc while Parsing:\n" + sw );
	}
    }

    public void testSortDecl() {
	parseDecls("sorts { elem; list; }");
	assertEquals("find sort elem", new Name("elem"),
		     sort_ns.lookup(new Name("elem")).name()); 
	assertEquals("find sort list", new Name("list"),
		     sort_ns.lookup(new Name("list")).name());
    }


    protected GenericSort checkGenericSort ( Named            p_n,
					     SetOfSort        p_ext,
					     SetOfSort        p_oneof ) {
	assertTrue ( "Generic sort does not exist", p_n != null );
	assertTrue ( "Generic sort does not have type GenericSort, but " + p_n.getClass(),
		     p_n instanceof GenericSort );

	GenericSort gs = (GenericSort)p_n;
	
	assertEquals ( "Generic sort has wrong supersorts",
		       p_ext,
		       gs.extendsSorts () );
	assertEquals ( "Generic sort has wrong oneof-list",
		       p_oneof,
		       gs.getOneOf () );
	
	return gs;
    }


    protected Sort checkSort ( Named p_n ) {
	assertTrue ( "Sort does not exist", p_n != null );
	assertTrue ( "Sort does not have type Sort, but " + p_n.getClass(),
		     p_n instanceof Sort );

	return (Sort)p_n;
    }


    public void testGenericSortDecl() {
	Named       n;
	GenericSort G, H;
	Sort        S, T;
	
	sort_ns = new Namespace ();
	parseDecls("sorts { generic G; generic H extends G; }");

	G = checkGenericSort ( sort_ns.lookup(new Name("G")),
			       SetAsListOfSort.EMPTY_SET,
			       SetAsListOfSort.EMPTY_SET );
	H = checkGenericSort ( sort_ns.lookup(new Name("H")),
			       SetAsListOfSort.EMPTY_SET.add ( G ),
			       SetAsListOfSort.EMPTY_SET );


	sort_ns = new Namespace ();
	parseDecls("sorts { S; generic G; generic H extends G, S; }");
	
	S = checkSort        ( sort_ns.lookup(new Name("S")) );
	G = checkGenericSort ( sort_ns.lookup(new Name("G")),
			       SetAsListOfSort.EMPTY_SET,
			       SetAsListOfSort.EMPTY_SET );
	H = checkGenericSort ( sort_ns.lookup(new Name("H")),
			       SetAsListOfSort.EMPTY_SET.add ( S ).add ( G ),
			       SetAsListOfSort.EMPTY_SET );


	sort_ns = new Namespace ();
	parseDecls("sorts { S; T; generic H oneof { S, T }; }");

	S = checkSort        ( sort_ns.lookup(new Name("S")) );
	T = checkSort        ( sort_ns.lookup(new Name("T")) );
	H = checkGenericSort ( sort_ns.lookup(new Name("H")),
			       SetAsListOfSort.EMPTY_SET,
			       SetAsListOfSort.EMPTY_SET.add ( S ).add ( T ) );
	

	sort_ns = new Namespace ();
	parseDecls("sorts { S; T; generic G; generic H oneof { S } extends T, G; }");

	S = checkSort        ( sort_ns.lookup(new Name("S")) );
	T = checkSort        ( sort_ns.lookup(new Name("T")) );
	G = checkGenericSort ( sort_ns.lookup(new Name("G")),
			       SetAsListOfSort.EMPTY_SET,
			       SetAsListOfSort.EMPTY_SET );
	H = checkGenericSort ( sort_ns.lookup(new Name("H")),
			       SetAsListOfSort.EMPTY_SET.add ( T ).add ( G ),
			       SetAsListOfSort.EMPTY_SET.add ( S ) );
	

	sort_ns = new Namespace ();
	parseDecls("sorts { S, T; generic G,G2; generic H,H2 oneof { S } extends T, G; }");

	S = checkSort        ( sort_ns.lookup(new Name("S")) );
	T = checkSort        ( sort_ns.lookup(new Name("T")) );
	G = checkGenericSort ( sort_ns.lookup(new Name("G")),
			       SetAsListOfSort.EMPTY_SET,
			       SetAsListOfSort.EMPTY_SET );
	checkGenericSort     ( sort_ns.lookup(new Name("G2")),
			       SetAsListOfSort.EMPTY_SET,
			       SetAsListOfSort.EMPTY_SET );
	H = checkGenericSort ( sort_ns.lookup(new Name("H")),
			       SetAsListOfSort.EMPTY_SET.add ( T ).add ( G ),
			       SetAsListOfSort.EMPTY_SET.add ( S ) );
	checkGenericSort     ( sort_ns.lookup(new Name("H2")),
			       SetAsListOfSort.EMPTY_SET.add ( T ).add ( G ),
			       SetAsListOfSort.EMPTY_SET.add ( S ) );
	

	sort_ns = new Namespace ();
	String str = "sorts { generic G; generic H oneof { G }; }";
	try {
	    DeclParser p = stringParser(str);
	    p.decls();

	    fail ( "Expected an GenericSortException" );
	} catch ( Exception e ) {
	    assertTrue ( "Expected a GenericSortException",
			 e instanceof de.uka.ilkd.key.parser.GenericSortException );
	}
    }

    /** asserts that the found object is a schemavariable and 
     * that the allowed macthing type is QuantifiableVariable
     */
    public void assertVariableSV(String msg, 
				 Object o) {
	assertTrue("The named object: "+o+" is of type "+o.getClass()+
		   ", but the type SchemaVariable was expected",
		   o instanceof SchemaVariable);

	assertTrue(msg, ((SchemaVariable)o).isVariableSV());
    }

    /** asserts that the SchemaVariable matches to term but not to a
     * formula 
     */
    public void assertTermSV(String msg, 
			     Object o) {

	assertTrue("The named object: "+o+" is of type "+o.getClass()+
		   ", but the type SchemaVariable was expected",
		   o instanceof SchemaVariable);
	assertSame(msg,
		   ((SchemaVariable)o).matchType(), Term.class);
	assertTrue("Schemavariable is not allowed to match a term of sort FORMULA.",
		   ((SchemaVariable)o).sort() != Sort.FORMULA);
    }

    /** asserts that the SchemaVariable matches to a formula 
     * and not to a term (of sort != Sort.FORMULA)
     */
    public void assertFormulaSV(String msg, 
				Object o) {
	assertTrue("The named object: "+o+" is of type "+o.getClass()+
		   ", but the type SchemaVariable was expected",
		   o instanceof SchemaVariable);
	assertSame(msg,
		   ((SchemaVariable)o).matchType(), Term.class);
	assertSame("Only matches to terms of sort FORMULA allowed. "+
		   "But term has sort "+((SchemaVariable)o).sort(), 
		   ((SchemaVariable)o).sort(), Sort.FORMULA);

	
    }

    
    public void testFunctionDecl() {
	parseDecls("sorts { elem; list; }\n" +
		   "functions {\n" +
		   "  elem head(list);\n" +
		   "  list tail(list);\n" +
		   "  elem[] tailarray(elem[]);\n" +
		   "  list nil;\n" +
		   "  list cons(elem,list);\n" +
		   "}\n");
	
	Sort elem = (Sort)sort_ns.lookup(new Name("elem"));
	Sort list = (Sort)sort_ns.lookup(new Name("list"));

	assertEquals("find head function", new Name("head"),
		     func_ns.lookup(new Name("head")).name());
	assertEquals("head arity", 1,
		     ((Function)func_ns.lookup(new Name("head"))).arity());
	assertEquals("head arg sort 0", list,
		     ((Function)func_ns.lookup(new Name("head"))).argSort(0));
	assertEquals("head return sort", elem,
		     ((Function)func_ns.lookup(new Name("head"))).sort());

	assertEquals("find tail function", new Name("tail"),
		     func_ns.lookup(new Name("tail")).name());
	assertEquals("tail arity", 1,
		     ((Function)func_ns.lookup(new Name("tail"))).arity());
	assertEquals("tail arg sort 0", list,
		     ((Function)func_ns.lookup(new Name("tail"))).argSort(0));
	assertEquals("tail return sort", list,
		     ((Function)func_ns.lookup(new Name("tail"))).sort());
	assertEquals("tailarray arg sort 0", ArraySortImpl.getArraySort(elem),
		     ((Function)func_ns.lookup(new Name("tailarray"))).argSort(0));
	assertEquals("tailarray return sort", ArraySortImpl.getArraySort(elem),
		     ((Function)func_ns.lookup(new Name("tailarray"))).sort());

	assertEquals("find nil function", new Name("nil"),
		     func_ns.lookup(new Name("nil")).name());
	assertEquals("nil arity", 0,
		     ((Function)func_ns.lookup(new Name("nil"))).arity());
	assertEquals("nil return sort", list,
		     ((Function)func_ns.lookup(new Name("nil"))).sort());

	assertEquals("find cons function", new Name("cons"),
		     func_ns.lookup(new Name("cons")).name());
	assertEquals("cons arity", 2,
		     ((Function)func_ns.lookup(new Name("cons"))).arity());
 	assertEquals("cons arg sort 0", elem,
		     ((Function)func_ns.lookup(new Name("cons"))).argSort(0));
 	assertEquals("cons arg sort 1", list,
		     ((Function)func_ns.lookup(new Name("cons"))).argSort(1));
	assertEquals("cons return sort", list,
		     ((Function)func_ns.lookup(new Name("cons"))).sort());
    }

    public void testPredicateDecl() {
	parseDecls("sorts { elem; list; }\n" +
		   "predicates {\n" +
		   "  isEmpty(list);\n" +
		   "  contains(list,elem);\n" +
		   "  maybe;\n" +
		   "}\n");
	
	Sort elem = (Sort)sort_ns.lookup(new Name("elem"));
	Sort list = (Sort)sort_ns.lookup(new Name("list"));


	assertEquals("find isEmpty predicate", new Name("isEmpty"),
		     func_ns.lookup(new Name("isEmpty")).name());
	assertEquals("isEmpty arity", 1,
		     ((Function)func_ns.lookup(new Name("isEmpty"))).arity());
	assertEquals("isEmpty arg sort 0", list,
		   ((Function)func_ns.lookup(new Name("isEmpty"))).argSort(0));
	assertEquals("isEmpty return sort", Sort.FORMULA,
		     ((Function)func_ns.lookup(new Name("isEmpty"))).sort());

	assertEquals("find contains predicate", new Name("contains"),
		     func_ns.lookup(new Name("contains")).name());
	assertEquals("contains arity", 2,
		     ((Function)func_ns.lookup(new Name("contains"))).arity());
	assertEquals("contains arg sort 0", list,
		   ((Function)func_ns.lookup(new Name("contains"))).argSort(0));
	assertEquals("contains arg sort 1", elem,
		   ((Function)func_ns.lookup(new Name("contains"))).argSort(1));
	assertEquals("contains return sort", Sort.FORMULA,
		     ((Function)func_ns.lookup(new Name("contains"))).sort());

	assertEquals("find maybe predicate", new Name("maybe"),
		     func_ns.lookup(new Name("maybe")).name());
	assertEquals("maybe arity", 0,
		     ((Function)func_ns.lookup(new Name("maybe"))).arity());
	assertEquals("maybe return sort", Sort.FORMULA,
		     ((Function)func_ns.lookup(new Name("maybe"))).sort());
    }

    /*
	var_ns.add(x = new LogicVariable(new Name("x"),elem));
	var_ns.add(y = new LogicVariable(new Name("y"),elem));
	var_ns.add(z = new LogicVariable(new Name("z"),elem));
	var_ns.add(xs = new LogicVariable(new Name("xs"),list));
	var_ns.add(ys = new LogicVariable(new Name("ys"),list));
	var_ns.add(zs = new LogicVariable(new Name("zs"),list));
    */


    public void testSVDecl() {
	parseDecls("sorts { elem; list; }\n" +
		   "schema variables {\n" +
		   "  program statement  #s; \n"+
		   "  elem x,y;\n" +
		   "  list variables lv;\n" +
		   "  formula b;\n" +
		   "}\n");
	
	
	Sort elem = (Sort)sort_ns.lookup(new Name("elem"));
	Sort list = (Sort)sort_ns.lookup(new Name("list"));

	assertEquals("find SV x", new Name("x"),
		     var_ns.lookup(new Name("x")).name()); 
	assertTermSV("SV x type", 
		     var_ns.lookup(new Name("x"))); 
	assertEquals("SV x sort", elem,
		     ((SchemaVariable)var_ns.lookup(new Name("x"))).sort()); 

	assertEquals("find SV ", new Name("y"),
		     var_ns.lookup(new Name("y")).name()); 
	assertTermSV("SV y type", 
		     var_ns.lookup(new Name("y"))); 
	assertEquals("SV y sort", elem,
		     ((SchemaVariable)var_ns.lookup(new Name("y"))).sort()); 

	assertEquals("find SV ", new Name("lv"),
		     var_ns.lookup(new Name("lv")).name()); 
	assertVariableSV("SV lv type", 
		     var_ns.lookup(new Name("lv"))); 
	assertEquals("SV lv sort", list,
		     ((SchemaVariable)var_ns.lookup(new Name("lv"))).sort()); 
	
	assertEquals("find SV ", new Name("b"),
		     var_ns.lookup(new Name("b")).name()); 
	assertFormulaSV("SV b type", 
		     var_ns.lookup(new Name("b"))); 
	assertEquals("SV b sort", Sort.FORMULA,
		     ((SchemaVariable)var_ns.lookup(new Name("b"))).sort()); 
    }
    


    public void testHeurDecl() {
	parseDecls("heuristics { bool; shoot_foot; }");
	assertEquals("find heuristic bool", new Name("bool"),
		     heur_ns.lookup(new Name("bool")).name()); 
	assertEquals("find heuristic shoot_foot", new Name("shoot_foot"),
		     heur_ns.lookup(new Name("shoot_foot")).name());
    }
    

}
    
