// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/* -*-antlr-*- */
header {

    package de.uka.ilkd.key.parser;

    import java.io.*;
    import antlr.*;
    import de.uka.ilkd.key.logic.*;
    import de.uka.ilkd.key.rule.*;
    import de.uka.ilkd.key.logic.op.*;
    import de.uka.ilkd.key.logic.sort.*;
    import de.uka.ilkd.key.java.JavaInfo;
    import de.uka.ilkd.key.java.visitor.*;
    import java.util.Vector;
    import java.util.HashMap;
    import de.uka.ilkd.key.proof.UseInsteadMap;

}

/** 
 * A parser for signature declarations,
 * i.e. sorts, schema variabls, predicates, functions and heuristics.
 */  

class DeclParser extends Parser;
options {
	importVocab=KeYLexer;
	k = 3;
	defaultErrorHandler=false;
}

{
	static final Sort[] AN_ARRAY_OF_SORTS = new Sort[0];

	private Namespace var_ns;
	private Namespace func_ns;
	private Namespace sort_ns;
	private Namespace heur_ns;
        private JavaInfo javaInfo;
	
	// these variables are set if a file is read in step by
	// step. This used when reading in LDTs because of cyclic
	// dependencies.
	protected boolean skip_schemavariables;
	protected boolean skip_functions;
	protected boolean skip_predicates;
	protected boolean skip_sorts;
	protected boolean skip_heuristics;
	protected boolean minus=false;	
	protected UseInsteadMap useInstead=new UseInsteadMap();


	public DeclParser(TokenStream lexer,
			  String filename,
			  Namespace var_ns,
			  Namespace func_ns,
			  Namespace sort_ns,
			  Namespace heur_ns,
                          JavaInfo ji) {
	    this(lexer);
	    setFilename(filename);
	    this.var_ns = var_ns;
	    this.func_ns = func_ns;
	    this.sort_ns = sort_ns;
	    this.heur_ns = heur_ns;
            this.javaInfo = ji;
	    setSkips(false, false, false, false, false);
	}
		
	public Namespace getSorts() {
	    return sort_ns;
	}

	public Namespace functions() {
	    return func_ns;
	}

	public Namespace heuristics() {
	    return heur_ns;
	}

	public Namespace variables() {
	    return var_ns;
	}


	protected int getLine() {
	    int line = -1;
	    try {
		line = LT(0).getLine();
	    } 
	    catch (TokenStreamException e) {
		System.err.println("No further token in stream");
	    }
	    return line;
	}   

	protected int getColumn() {
	    int col = -1;
	    try {
		col = LT(0).getColumn();
	    } 
	    catch (TokenStreamException e) {
		System.err.println("No further token in stream");
	    }
	    return col;
	}   


        protected void setSkips(boolean svars, boolean funcs, 
				boolean preds, boolean sorts,
				boolean heurs) {
	    skip_schemavariables = svars;
   	    skip_functions  	 = funcs;
	    skip_predicates 	 = preds;
	    skip_sorts 		 = sorts;
	    skip_heuristics 	 = heurs;
	}

	public void parseVariables() 	
	    throws antlr.RecognitionException, antlr.TokenStreamException
	    {
		setSkips(false, true, true, true, true); 
		top();
		setSkips(false, false, false, false, false);	    		
	    }
    	public Term toZNotation(String s, boolean minus){
	Term[] t1=new Term[1];
	Operator v;
	Term a;
	TermFactory tf=TermFactory.DEFAULT;
	v=(Function) func_ns.lookup(new Name("#"));
	t1[0] = tf.createFunctionTerm((Function)v);
	v = (Function) func_ns.lookup(new Name(s.substring(0,1)));
	a = tf.createFunctionTerm((Function)v,t1);
	for(int i=1;i<s.length();i++){
		t1[0]=a;
		v = (Function)func_ns.lookup(new Name(s.substring(i,i+1)));
		a = tf.createFunctionTerm((Function)v,t1);
	}
	if(minus){
		t1[0]=a;
		v = (Function) func_ns.lookup(new Name("neg"));
		a = tf.createFunctionTerm((Function)v,t1);
	}
	t1[0]=a;
	v = (Function) func_ns.lookup(new Name("Z"));
	a = tf.createFunctionTerm((Function)v,t1);
	return a;
	}
	public void parseFunctions()
	    throws antlr.RecognitionException, antlr.TokenStreamException

	    {
		setSkips(true, false, true, true, true); 
		top();
		setSkips(false, false, false, false, false);
	    }

	public void parsePredicates()
	    throws antlr.RecognitionException, antlr.TokenStreamException
	    {
		setSkips(true, true, false, true, true); 
		top();
		setSkips(false, false, false, false, false);	    
	    }

	public void parseSorts() 
	    throws antlr.RecognitionException, antlr.TokenStreamException
	    {
		setSkips(true, true, true, false, true); 
		top();
		setSkips(false, false, false, false, false);
	    }

	public void parseHeuristics()
	    throws antlr.RecognitionException, antlr.TokenStreamException
	    {
		setSkips(true, true, true, true, false); 
		top();
		setSkips(false, false, false, false, false);	    
	    }

	public void parseFuncAndPred()
	    throws antlr.RecognitionException, antlr.TokenStreamException
	    {
		setSkips(true, false, false, true, true); 
		top();
		setSkips(false, false, false, false, false);
	    }

    public NamespaceSet namespaces() {
	NamespaceSet ns = new NamespaceSet();
	ns.setVariables(variables());
	ns.setFunctions(functions());
	ns.setSorts(getSorts());
	ns.setHeuristics(heuristics());
	return ns;
    }

}

top : decls;

decls :	
		(
			sort_decls
		|
			schema_var_decls
		|
                	prog_var_decls
                |
			pred_decls
		|
			func_decls
		|
			heur_decls
		) *
	;

sort_decls {
    boolean isObjectSort=false;
    boolean isGenericSort=false;
    Sort[] sortExt=new Sort [0];
    Sort[] sortOneOf=new Sort [0];
    ListOfName sort_ids;
}:
    "sorts" 
	LBRACE ( 
           { 
	       isObjectSort=false; 
	       isGenericSort=false; 
	       sortExt = new Sort[0];
	       sortOneOf = new Sort[0];
	   }
	  ( "object" {isObjectSort=true;} sort_ids = sort_identifiers
	  | "generic" {isGenericSort=true;} sort_ids = sort_identifiers
	  ( "oneof" sortOneOf = oneof_sorts )? 
	  ("extends" sortExt = extends_sorts )?
	  | sort_ids = sort_identifiers
      )
	  SEMI
	{
	    Name           sort_name;
	    IteratorOfName it         = sort_ids.iterator ();
	    while ( it.hasNext () ) {
	    sort_name = it.next ();

	    if (sort_ns.lookup(sort_name) == null) {
		Sort s;
		if (isObjectSort) {
		    if ("Object".equals(sort_name)) {
			if (sort_ns.lookup(new Name("Object"))
			    ==null) {

			    s = new ClassInstanceSortImpl(sort_name);
			}
			s=(Sort)sort_ns.lookup(new Name("Object"));
		    } else {
			s = new ClassInstanceSortImpl(sort_name, 
						      (Sort)sort_ns.lookup
						      (new Name("Object")));
		    }	
		} else if (isGenericSort) {
		    int        i;
		    SetOfSort  ext   = GenericSort.EMPTY_SORT_SET;
		    SetOfSort  oneOf = GenericSort.EMPTY_SORT_SET;

		    for ( i = 0; i != sortExt.length; ++i )
		        ext = ext.add ( sortExt[i] );

		    for ( i = 0; i != sortOneOf.length; ++i )
		        oneOf = oneOf.add ( sortOneOf[i] );

		    try {
		        s = new GenericSort ( sort_name,
		                              ext,
		                              oneOf );
		    } catch ( GenericSupersortException e ) {
		        throw new GenericSortException ( "sort",
		                   "Illegal sort given", e.getIllegalSort (),
		                   getFilename (), getLine (), getColumn () );
		    }
		} else {
		    s = new PrimitiveSort(sort_name);
		}
		sort_ns.add(s);
		if (s instanceof NonCollectionSort) {
		    NonCollectionSort ns
			= (NonCollectionSort)s;
		    sort_ns.add(ns.getSetSort());
		    sort_ns.add(ns.getSequenceSort());
		    sort_ns.add(ns.getBagSort());
		}
		if ( s instanceof SortDefiningSymbols )
		    ((SortDefiningSymbols)s).createSymbols ( func_ns, sort_ns );
	    }

	    }
	}
			
	)* 
	RBRACE
	;

sort_identifiers returns [ListOfName ids = SLListOfName.EMPTY_LIST]
:
        id:IDENT { ids = ids.append ( new Name(id.getText()) ); }
        (
            COMMA
            id2:IDENT { ids = ids.append ( new Name(id2.getText()) ); }
        ) *
    ;

extends_sorts returns [Sort[] extendsSorts = null] 
{
	Vector args = new Vector();
	Sort s;
}
	:
        s = any_sort_id_check[true] { args.add(s); }
        (
            COMMA s = any_sort_id_check[true] {args.add(s);}
        ) *
		{
			extendsSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
		}
	;
	
oneof_sorts returns [Sort[] oneOfSorts = null] 
{
	Vector args = new Vector();
	Sort s;
}
	:
        LBRACE
        s = sort_id_check[true] { args.add(s); }
        (
            COMMA s = sort_id_check[true] {args.add(s);}
        ) *
        RBRACE
		{
			oneOfSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
		}
	;
	
schema_var_decls
{
	Sort s = null;
	boolean makeVariableSV  = false;
	boolean makeDependingSV = false; // skolem functions
	boolean rigid           = false;
	boolean listModifier    = false;
}
	:
		"schema" "variables"
		LBRACE
		(
			{makeVariableSV = false; makeDependingSV = false; rigid = false; listModifier=false;}
			( ( "program" ("svlist" {listModifier=true;})?
					( "statement"
					{
						s=ProgramSVSort.STATEMENT;
					}
					| "expression"
					{
						s=ProgramSVSort.EXPRESSION;
					}
					| "multiplevardecl"
					{
						s=ProgramSVSort.MULTIPLEVARDECL;
					}
					| "simple" "expression"
					{
						s=ProgramSVSort.SIMPLEEXPRESSION;
					}
					| "simple" "javaboolean" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVABOOLEXPRESSION;
					}
					| "simple" "javabyte" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVABYTEEXPRESSION;
					}
					| "simple" "javachar" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVACHAREXPRESSION;
					}
					| "simple" "javashort" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVASHORTEXPRESSION;
					}
					| "simple" "javaint" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVAINTEXPRESSION;
					}
					| "simple" "javalong" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVALONGEXPRESSION;
					}
					| "simple" "javabyteshortint" "expression"
					{
						s=ProgramSVSort.SIMPLEJAVABYTESHORTINTEXPRESSION;
					}
					| "simple" "anyjavatype" "expression"
					{
						s=ProgramSVSort.SIMPLEANYJAVATYPEEXPRESSION;
					}
					| "classtype" "expression"
					{
						s=ProgramSVSort.CLASSTYPEEXPRESSION;
					}
					| "interfacetype" "expression"
					{
						s=ProgramSVSort.INTERFACETYPEEXPRESSION;
					}
					| "nonsimple" "expression"	
					{
						s=ProgramSVSort.NONSIMPLEEXPRESSION;
					}
					| "javaboolean" "expression"	
					{
						s=ProgramSVSort.BOOLEANEXPRESSION;
					}
					| "nonsimple" "javaboolean" "expression"	
					{
						s=ProgramSVSort.NONSIMPLEBOOLEANEXPRESSION;
					} 
					| "lefthandside"
					{
						s=ProgramSVSort.LEFTHANDSIDE;
					}
					| "loopinit"
					{
						s=ProgramSVSort.LOOPINIT;
					}
					| "guard"
					{
						s=ProgramSVSort.GUARD;
					}
					| "forupdates"
					{
						s=ProgramSVSort.FORUPDATES;
					}
					| "statementexpression"
					{
						s=ProgramSVSort.STATEMENTEXPRESSION;
					}
					| "label"
					{
						s=ProgramSVSort.LABEL;
					}
					| "type"
					{
						s=ProgramSVSort.TYPE;
					}
					| "variable"	
					{
						s=ProgramSVSort.VARIABLE;
					}
					| "variableinitializer"	
					{
						s=ProgramSVSort.VARIABLEINIT;
					}
					| "implicit" "next"	
					{
						s=ProgramSVSort.IMPLICITNEXT;
					}				
					| "implicit" "reference" "field"	
					{
						s=ProgramSVSort.IMPLICITREFERENCE;
					}				
					| "new"	
					{
						s=ProgramSVSort.NEW;
					}
					| "new"	"array"
					{
						s=ProgramSVSort.NEWARRAY;
					}
					| "array" "initializer"
					{
						s=ProgramSVSort.ARRAYINITIALIZER;
					}
					| "array" "length"
					{
						s=ProgramSVSort.ARRAYLENGTH;
					}
                    			| "switch"
                    			{
                        		 	s=ProgramSVSort.SWITCH;
                    			}
					|		
					( "booleanvariable") {
						s=ProgramSVSort.BOOLEANVARIABLE;
					}
					|
					( "catch") {
						s=ProgramSVSort.CATCH;
					}
					|		
					( "methodreference") {
						s=ProgramSVSort.METHODREFERENCE;
					}
					|
					  "special" "constructorreference" {
						s=ProgramSVSort.SPECIALCONSTRUCTORREFERENCE;				
					}	
					|		
					( "classtype") {
						s=ProgramSVSort.CLASSTYPE;
					}
					|		
					( "interfacetype") {
						s=ProgramSVSort.INTERFACETYPE;
					}
					|		
					( "arraytype") {
						s=ProgramSVSort.ARRAYTYPE;
					}
					|		
					( "array") {
						s=ProgramSVSort.ARRAY;
					}
					|
					( "methodbody") {
						s=ProgramSVSort.METHODBODY;
					}
					|
					( "branch") {
						s=ProgramSVSort.BRANCH;
					}
					)
				)
				|
				"formula"
                		( "rigid" { rigid = true; } ) ?
				{ 
					s = Sort.FORMULA;
				}			
				|
				s=any_sort_id_check[true]
				( "variables" { makeVariableSV = true; }
                		| "depending" { makeDependingSV = true; }
                		| "rigid" { rigid = true; } ) ?
				
			)
			one_schema_var_decl[s,makeVariableSV,makeDependingSV,rigid,listModifier]
			( COMMA one_schema_var_decl[s,makeVariableSV,makeDependingSV,rigid,listModifier] ) *
			SEMI
		) *
		RBRACE
	;

one_schema_var_decl[Sort s, 
	            boolean makeVariableSV, 
	            boolean makeDependingSV,
	            boolean rigid, 
	            boolean listModifier] :

		id:IDENT
		{
		    if (!skip_schemavariables) {
			SchemaVariable v;
			if ( s == Sort.FORMULA ) {
			    v = SchemaVariable.createFormulaSV
				(new Name(id.getText()), listModifier, rigid);
			} else if ( s instanceof ProgramSVSort ) {
			    v = SchemaVariable.createProgramSV
			    (new ProgramElementName(id.getText()),(ProgramSVSort) s, listModifier);
			} else {
			    if ( makeVariableSV ) {
				v=SchemaVariable.createVariableSV
				(new Name(id.getText()), s, listModifier);
			    } else if ( makeDependingSV ) {
				v = SchemaVariable.createDependingSV
				(new Name(id.getText()), s, listModifier);
			    } else { v = SchemaVariable.createTermSV
				     (new Name(id.getText()), s, listModifier, rigid);
			    }
			}
			var_ns.add(v);
		    }
		}
	;






pred_decls 
{
	Sort[] argSorts;
	String pred_name;
}
	:
		"predicates"
		LBRACE
		(
			pred_name = id
			argSorts=arg_sorts[!skip_predicates]
			{
			    if (!skip_predicates) {
				Name predicate = new Name(pred_name);
				if (func_ns.lookup(predicate) != null) {
				    // throw new AmbigiousDeclException
				    //("predicate "+pred_name+" ", 
				    //getLine(), getColumn());
				} else {
				    Function p = new Function(predicate,
							      Sort.FORMULA, 
							      argSorts);
				    func_ns.add(p);			
				}
			    } 
			}
			SEMI
		) *
		RBRACE
	;

func_decls 
{
	Sort[] argSorts;
	Sort retSort;
        String func_name;
}
	:
		"functions"
		LBRACE
		(
			retSort = sort_id_check[!skip_functions]
			func_name = id
			argSorts = arg_sorts[!skip_functions]
			{
			    if (!skip_functions) {

				Name fct_name = new Name(func_name);
				if (func_ns.lookup(fct_name) != null) {
				    //throw new AmbigiousDeclException
				    //("function "+func_name, 
				    //getLine(), getColumn());
				} else {
				    Function f=new Function(fct_name,
							    retSort, 
							    argSorts);
				    func_ns.add(f);
				}
			    } 
			}
			SEMI
		) *
		RBRACE
	;

arg_sorts[boolean checkSort] returns [Sort[] argSorts = null] 
{
	Vector args = new Vector();
	Sort s;
}
	:
		(
			LPAREN
			s = sort_id_check[checkSort] { args.add(s); }
			(
				COMMA s = sort_id_check[checkSort] {args.add(s);}
			) *
			RPAREN
		) ?
		{
			argSorts = (Sort[])args.toArray(AN_ARRAY_OF_SORTS);
		}
	;

prog_var_decls 
{
	Sort s;
        String var_name;
}
	:
		"program" "variables"
		LBRACE
		(
			s = sort_id_check[true]
			var_name = id
			{
				ProgramVariable pv = new ProgramVariable(
                                   new ProgramElementName(var_name), 
                                   javaInfo.getKeYJavaType(s));
				javaInfo.programVariables().add(pv);
			}
			SEMI
		) *
		RBRACE
	;


heur_decls :
		"heuristics"
		LBRACE
		(
			id:IDENT SEMI
			{ 
			    if (!skip_heuristics) {
				Heuristic h = new Heuristic(new Name(id.getText()));
				heur_ns.add(h);
			    }
			}
		) *
		RBRACE
	;

sort_id  returns [Sort s = null]
	:
		s = sort_id_check[true]
	;			

// Non-generic sorts, array sorts allowed
sort_id_check [boolean checkSort] returns [Sort s = null]                
{
    Sort t;
}
	:
        t = sort_id_check_help[checkSort]
        s = array_decls[t]
	;

// Generic and non-generic sorts, array sorts allowed
any_sort_id_check [boolean checkSort] returns [Sort s = null]                
{
    Sort t;
}
	:
        t = any_sort_id_check_help[checkSort]
        s = array_decls[t]
	;

// also allow generic sorts
any_sort_id_check_help [boolean checkSort] returns [Sort s = null]
{
    String name;
}
	:
		id:IDENT
                { 	String idtext=(useInstead.get(id.getText()));
			s= (Sort) sort_ns.lookup(new Name(idtext));
			if (checkSort && s == null) {
			    throw 
				new NotDeclException("sort", id, 
						     getFilename());
			}
		}
	;

sort_id_check_help [boolean checkSort] returns [Sort s = null]
    :
        s = any_sort_id_check_help[checkSort]
        {
            // don't allow generic sorts or collection sorts of
            // generic sorts at this point
            Sort t = s;
            while ( t instanceof CollectionSort )
                t = ((CollectionSort)t).elementSort ();
            if ( t instanceof GenericSort )
                throw new GenericSortException ( "sort",
                        "Non-generic sort expected", t,
                        getFilename (), getLine (), getColumn () );
        }
    ;

array_decls [Sort p] returns [Sort s = null]                
{
    int  n=0;
}

	:
		(LBRACKET RBRACKET {n++;})*


	        { s = p;
              while(n>0){
                   s=ArraySortImpl.getArraySort(s);
                   n--;
              }
		 } 
	;

id_declaration returns [ IdDeclaration idd = null ]
{
    Sort s = null;
}
    :
        id:IDENT
        ( COLON s = sort_id_check[true] ) ?
        {
            idd = new IdDeclaration ( id.getText (), s );
        }
    ;

id returns [String name = null]
	:
		(IDENT COLON COLON IDENT) => 
                	classID:IDENT COLON COLON nameID:IDENT 
			{ 
			 name = classID.getText() + "::" + nameID.getText();
			}
                |
		pureID:IDENT { name = pureID.getText(); } 
	;

