// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.proof;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.op.IteratorOfProgramVariable;
import de.uka.ilkd.key.logic.op.SetOfProgramVariable;
import de.uka.ilkd.key.logic.op.Metavariable;
import de.uka.ilkd.key.logic.op.SetOfMetavariable;
import de.uka.ilkd.key.logic.op.SetAsListOfMetavariable;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.proof.incclosure.*;

import java.util.List;
import java.util.ArrayList;

/**
 *  A proof is represented as a tree of nodes containing sequents. The initial
 *  proof consists of just one node -- the root -- that has to be
 *  proved. Therefore it is divided up into several sub goals and so on. A
 *  single goal is not divided into sub goals any longer if the contained
 *  sequent becomes an axiom. A proof is closed if all leaves are closed. As
 *  the calculus works only on the leaves of a tree, the goals are the
 *  additional information needed for the proof is only stored at the leaves
 *  (saves memory) and not in the inner nodes. This class represents now a goal
 *  of the proof, this means a leave whose sequent is not closed. It keeps
 *  track of all applied rule applications on the branch and of the
 *  corresponding rule application index. Furthermore it offers methods for
 *  setting back several proof steps. The sequent has to be changed using the
 *  methods of Goal.  */
public class Goal implements java.io.Serializable {

    private Node node;

    /** the index of possible rule applications at this node
     * @undirected
     * @supplierRole possible rule applications
     * @clientCardinality 1
     * @supplierCardinality 1*/
    private RuleAppIndex ruleAppIndex;

    /** list of all applied rule applications at this branch */
    private ListOfRuleApp appliedRuleApps = SLListOfRuleApp.EMPTY_LIST;

    /** goal listeners  */
    private List listeners = new ArrayList();

    /** creates a new goal referencing the given node */
    public Goal(Node node, 
		RuleAppIndex ruleAppIndex, 
		ListOfRuleApp appliedRuleApps) {  
	this.node = node;
	this.ruleAppIndex = ruleAppIndex;
	this.appliedRuleApps = appliedRuleApps;
	//	node.addNoPosTacletApps(ruleAppIndex.tacletIndex().allNoPosTacletApps());
    }

    /** creates a new goal referencing the given node */
    public Goal(Node node, 
		RuleAppIndex ruleAppIndex) {
	this(node, ruleAppIndex, SLListOfRuleApp.EMPTY_LIST);
    }

    /** returns the simplifier that has to be used */
    public SimultaneousUpdateSimplifier simplifier() {
	return node().proof().simplifier();
    }

    /** returns the referenced node */
    public Node node() {
	return node;
    }

    public SetOfProgramVariable getGlobalProgVars() {
	return node().getGlobalProgVars();
    }

    /** 
     * adds the listener l to the list of goal listeners.
     * Attention: A listener added to this goal will be taken over when
     * splitting into subgoals. 
     */
    public void addGoalListener(GoalListener l) {
	listeners.add(l);
    }

    /**
     * removes the listener l from the list of goal listeners.
     * Attention: The listener is just removed from 'this' goal not from the
     * other goals. (All goals can be accessed via proof openGoals())
     */
    public void removeGoalListener(GoalListener l) {
	listeners.remove(l);
    }

    /** 
     * informs all goal listeners about a change of the sequent
     * to reduce unnecessary object creation the necessary information is passed
     * to the listener as parameters and not through an event object.
     */
    protected void fireSequentChanged(SequentChangeInfo sci) {
	for (int i = 0; i<listeners.size(); i++) {
	    ((GoalListener)listeners.get(i)).sequentChanged(this, sci);
	}
    }

    protected void fireGoalReplaced(Goal       goal,
				    Node       parent,
				    ListOfGoal newGoals) {
	for (int i = 0; i<listeners.size(); i++) {
	    ((GoalListener)listeners.get(i)).goalReplaced(goal,
							  parent,
							  newGoals);
	}
    }

    /**
     * adds the global program variables to a new created variable namespace
     * that contains all the elements of the given namespace.
     */
    public Namespace getVariableNamespace(Namespace exNS) {
	Namespace newNS=exNS;
	IteratorOfProgramVariable it=getGlobalProgVars().iterator();
	if (it.hasNext()) {
	    newNS=newNS.extended(it.next());
	}
	while (it.hasNext()) {
	    newNS.add(it.next());
	}
	return newNS;
    }

    public void setGlobalProgVars(SetOfProgramVariable s) {
	node.setGlobalProgVars(s);
    }

    /** set the node the goal is related to
     * @param node the new Node the goal belongs to
     */
    public void setNode(Node node) {
	this.node = node;
    }

    public void setRuleAppIndex(RuleAppIndex ruleAppIndex) {
	this.ruleAppIndex = ruleAppIndex;
    }

    /** 
     * returns the index of possible rule applications at this node
     */
    public RuleAppIndex ruleAppIndex() {
	return ruleAppIndex;
    }
     
    /**
     * returns the Taclet index for this goal. This is just a shortcut to the
     * Taclet index of the RuleAppIndex
     * @return the Taclet index assigned to this goal
     */
    public TacletIndex indexOfTaclets() {
	return ruleAppIndex.tacletIndex();
    }
   
    /** adds a formula to the sequent before the given position
     * and informs the rule appliccation index about this change
     * @param cf the ConstrainedFormula to be added
     * @param p PosInOccurrence encodes the position 
     */
    public void addFormula(ConstrainedFormula cf, PosInOccurrence p) {
	SequentChangeInfo sci = sequent().addFormula(cf, p);
	setSequent(sci.sequent());	
	ruleAppIndex().sequentChanged(sci); 

	fireSequentChanged(sci);
    }

    /** adds a list of formulas to the sequent before the given position
     * and informs the rule appliccation index about this change
     * @param insertions the ListOfConstrainedFormula to be added
     * @param p PosInOccurrence encodes the position 
     */
    public void addFormula(ListOfConstrainedFormula insertions, PosInOccurrence p) {
	if ( insertions != SLListOfConstrainedFormula.EMPTY_LIST ) {
	    SequentChangeInfo sci = sequent().addFormula(insertions, p);
	    setSequent(sci.sequent());
	    ruleAppIndex().sequentChanged(sci);

	    fireSequentChanged(sci);
	}
    }

    /** adds a list of formulas to the antecedent or succedent of a  
     * sequent. Either at its front or back.
     * and informs the rule appliccation index about this change
     * @param insertions the ListOfConstrainedFormula to be added
     * @param inAntec boolean true(false) if ConstrainedFormula has to be
     * added to antecedent (succedent) 
     * @param first boolean true if at the front, if false then cf is
     * added at the back
     */
    public void addFormula ( ListOfConstrainedFormula insertions, 
			     boolean inAntec, boolean first ) {
	if ( insertions != SLListOfConstrainedFormula.EMPTY_LIST ) {
	    SequentChangeInfo sci = sequent().addFormula(insertions,
							 inAntec,
							 first);
	    setSequent(sci.sequent());       
	    ruleAppIndex().sequentChanged(sci);

	    fireSequentChanged(sci);
	}
    }

    /** adds a formula to the antecedent or succedent of a
     * sequent. Either at its front or back.
     * and informs the rule appliccation index about this change
     * @param cf the ConstrainedFormula to be added
     * @param inAntec boolean true(false) if ConstrainedFormula has to be
     * added to antecedent (succedent) 
     * @param first boolean true if at the front, if false then cf is
     * added at the back
     */
    public void addFormula ( ConstrainedFormula cf, boolean inAntec,
			     boolean first ) {
	SequentChangeInfo sci = sequent().addFormula(cf, inAntec, first);;
	setSequent(sci.sequent());
	ruleAppIndex().sequentChanged(sci); 

	fireSequentChanged(sci);
    }

    /** returns set of rules applied at this branch 
     * @return ListOfRuleApp applied rule applications
     */
    public ListOfRuleApp appliedRuleApps() {
	return appliedRuleApps;
    }
	

    /** returns the sequent of the node 
     * @return the Sequent to be proved
     */
    public Sequent sequent() {
	return node().sequent();
    }

    /** sets the sequent of the node 
     * @param the Sequent to be set
     */
    public void setSequent(Sequent seq) {
	node().setSequent(seq);
    }


    /** 
     * replaces a formula at the given position  
     * and informs the rule appliccation index about this change
     * @param cf the ConstrainedFormula replacing the old one
     * @param p PosInOccurrence encodes the position 
     */
    public void changeFormula(ConstrainedFormula cf, PosInOccurrence p) {
	SequentChangeInfo sci = sequent().changeFormula(cf, p);
	setSequent(sci.sequent());	
	ruleAppIndex().sequentChanged(sci);

	fireSequentChanged(sci);
    }

    /** 
     * replaces a formula at the given position  
     * and informs the rule appliccation index about this change
     * @param cf the ConstrainedFormula replacing the old one
     * @param p PosInOccurrence encodes the position 
     */
    public void changeFormula(ListOfConstrainedFormula replacements, 
			      PosInOccurrence p) {
	SequentChangeInfo sci = sequent().changeFormula(replacements, p);
	setSequent(sci.sequent());
	ruleAppIndex().sequentChanged(sci);

	fireSequentChanged(sci);
    }

    /** removes a formula at the given position from the sequent 
     * and informs the rule appliccation index about this change
     * @param cf the ConstrainedFormula to be removed
     * @param p PosInOccurrence encodes the position 
     */
    public void removeFormula(PosInOccurrence p) {
	SequentChangeInfo sci = sequent().removeFormula(p);
	setSequent(sci.sequent());
	ruleAppIndex().sequentChanged(sci);

	fireSequentChanged(sci);
    }

    /**
     * puts the NoPosTacletApp to the set of TacletApps at the node
     * of the goal and to the current RuleAppIndex.
     * @param app the TacletApp
     */
    public void addNoPosTacletApp(NoPosTacletApp app) {
	node().addNoPosTacletApp(app);
	ruleAppIndex.addNoPosTacletApp(app);	
    }

    /**
     * creates a new TacletApp and puts it to the set of TacletApps at the node
     * of the goal and to the current RuleAppIndex.
     * @param rule the Taclet of the TacletApp to create
     * @param insts the given instantiations of the TacletApp to be created
     * @param constraint the constraint under which the taclet can be applied
     */
    public void addTaclet(Taclet           rule,
			  SVInstantiations insts,
			  Constraint       constraint ) {		
	NoPosTacletApp tacletApp = NoPosTacletApp.createFixedNoPosTacletApp(rule,
									    insts,
									    constraint);
	if (tacletApp != null) {
	    addNoPosTacletApp(tacletApp);
	}
    }

    public void addProgramVariable(ProgramVariable pv) {
	node.setGlobalProgVars(getGlobalProgVars().add(pv));
    }

    public void addProgramVariables(Namespace ns) {
	IteratorOfNamed it=ns.elements().iterator();
	while (it.hasNext()) {
	    addProgramVariable((ProgramVariable)it.next());
	}
    }




    /** clones the goal (with copy of tacletindex and ruleAppIndex)
     * @return Object the clone
     */
    public Object clone() {	
	Goal clone = new Goal(node, ruleAppIndex.copy(), appliedRuleApps);
	clone.listeners = (List) ((ArrayList)listeners).clone();
	return clone;
    }

    /** like the clone method but returns right type
     * @return Goal clone of this Goal
     */
    public Goal copy() {
	return (Goal)clone();
    }

    /**
     * puts a RuleApp to the list of the applied rule apps at this goal
     * and stores it in the node of the goal
     * @param app the applied rule app
     */
    public void addAppliedRuleApp(RuleApp app) {
	// Last app first makes inserting and searching faster
	appliedRuleApps = appliedRuleApps.prepend(app);
	node().setAppliedRuleApp(app);
    }

    /**
     * PRECONDITION: appliedRuleApps.size () > 0
     */
    public void removeAppliedRuleApp () {
	appliedRuleApps = appliedRuleApps.tail ();
	node ().setAppliedRuleApp ( null );
    }

    
    /** creates n new nodes as children of the
     * referenced node and new 
     * n goals that have references to these new nodes.
     * @return the list of new created goals.
     */
    public ListOfGoal split(int n) {
	ListOfGoal goalList=SLListOfGoal.EMPTY_LIST;
	Node parent = node(); // has to be stored because the node
	                      // of this goal will be replaced
        if (n>0) {
	    IteratorOfSink itSinks = parent.reserveSinks ( n );
	    BranchRestricter br;
	    Node newNode = null;
	    Goal newGoal = null;

	    for (int i=0;i<n;i++) {
		if (i==0) { // first new goal is this one 
		    newGoal = this;
		} else { // otherwise it is a copy
		    newGoal = copy();
		}
		// create new node and add to tree
		if ( n > 1 ) {
		    br = new BranchRestricter ( itSinks.next () );
		    newNode = new Node(parent.proof(),
				       parent.sequent(),
				       null,
				       parent,
				       br);
		    br.setNode ( newNode );
		} else
		    newNode = new Node(parent.proof(),
				       parent.sequent(),
				       null,
				       parent,
				       itSinks.next ());

		// newNode.addNoPosTacletApps(parent.getNoPosTacletApps());
		newNode.setGlobalProgVars(parent.getGlobalProgVars());
		parent.add(newNode);

		// make new Goal and add to list
		newGoal.setNode(newNode);

		goalList = goalList.prepend(newGoal);
	    }
	}
	
	fireGoalReplaced ( this, parent, goalList );

	return goalList;
    }

    /**
     * sets back the proof step that led to this goal. This goal is set to
     * the parent node of the node corresponding to this goal. Goals given in
     * the goal list parameter are removed from that list, if their
     * corresponding nodes are leaves of the parent node of this goal.
     * @param goallist the list of goals 
     * @return the new list of goals where goals mapped to the leaves of
     * the parent to this goal are removed from compared to the given list.
     */
    public ListOfGoal setBack(ListOfGoal goalList) {
	Node parent=node.parent();
	IteratorOfNode leavesIt=parent.leavesIterator();
	IteratorOfGoal goalIt;
	while (leavesIt.hasNext()) {
	    Node n=leavesIt.next();
	 
	    goalIt=goalList.iterator();
	    while (goalIt.hasNext()) {
		Goal g=goalIt.next();
	
		if (g.node()==n && g!=this) {
		    goalList=goalList.removeAll(g);
		   
		}
	    }
	}

	//	ruleAppIndex.tacletIndex().setTaclets(parent.getNoPosTacletApps());

	ruleAppIndex.tacletIndex().removeTaclets(node.getNoPosTacletApps());
	setGlobalProgVars(parent.getGlobalProgVars());

	parent.cutChildrenSinks ();

	IteratorOfNode siblings=parent.childrenIterator();
	Node[] sibls=new Node[parent.childrenCount()];
	int i=0;
	while (siblings.hasNext()) {
	    sibls[i]=siblings.next(); 
	    i++;
	}

	for (i=0; i<sibls.length; i++) {
	    sibls[i].remove();
	}
	setNode(parent);
	removeAppliedRuleApp ();
	return goalList;
    }

    
    public Constraint getClosureConstraint () {
	return node ().getClosureConstraint ();
    }

    public void addClosureConstraint ( Constraint c ) {
	node ().addClosureConstraint ( c );
    }

    public void addRestrictedMetavariable ( Metavariable mv ) {
	node ().addRestrictedMetavariable ( mv );
    }
    
    public void setName(String s) {
       node.setName(s);
    }

    /** toString */
    public String toString() {
	String result=(node.sequent().prettyprint().toString());
	return result;
    }
}
