// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof;

import java.io.*;
import de.uka.ilkd.key.gui.AbbrevMap;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.proof.init.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.gui.Main;
import de.uka.ilkd.key.gui.KeYMediator;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.util.Debug;
import javax.swing.JOptionPane;
import java.util.*;

public class ProblemLoader implements Runnable {

   File file;
   Main main;
   KeYMediator mediator;
   
   Proof proof = null;
   IteratorOfNode children = null;

   Node currNode = null;
   Goal currGoal = null;
   String currTacletName = null;
   int currFormula = 0;
   PosInTerm currPosInTerm = PosInTerm.TOP_LEVEL;
   boolean interactive = true;
   ListOfHeuristic activeHeuristics = null;
   

   Stack stack = new Stack();
   LinkedList loadedInsts = null;
   ListOfIfFormulaInstantiation ifSeqFormulaList = 
      SLListOfIfFormulaInstantiation.EMPTY_LIST;
   
   public ProblemLoader(File file, Main main) {
      this.main = main;
      mediator  = main.mediator();
      this.file = file;
      activeHeuristics = SLListOfHeuristic.EMPTY_LIST;
   }

   public void run() {
      loadProblem();
   }


   /** parses the given file
    */
   protected void loadProblem() {
      KeYUserProblemFile problem = 
         new KeYUserProblemFile(file.getName(), file, main);;
      try{
          mediator.stopInterface(false); // just the cursor
	  ProblemInitializer init =
             new ProblemInitializer(main.proofSettings(), main);
          main.setStatusLine("Loading problem");
	  init.setUpProof(problem);
	  proof=mediator.getSelectedProof();


//************************** THE PROOF ******************************


          mediator.stopInterface(true); // now stop the listeners
          main.setStatusLine("Loading proof");
          
          currNode = proof.root(); // initialize loader
          children = currNode.childrenIterator(); // --"--
          init.tryReadProof(this);

          mediator.startInterface(true);
          main.setStandardStatusLine();

      } catch (ProofInputException e) {
          mediator.startInterface(true);
          main.setStatusLine("Failed to load problem/proof");
	  JOptionPane.showMessageDialog(main,"Could not load problem/proof " +
            " due to a parsing error. The file may be corrupted." +
	    "\nCause: \n" + e.toString(), "Parsing Error",	
					JOptionPane.ERROR_MESSAGE);
          main.setStandardStatusLine();
      } catch (IllegalArgumentException e) {
	  JOptionPane.showMessageDialog(main,"Could not load problem/proof " +
					" due to a parsing error. The file may be corrupted." +
					"\nCause: \n" + e.toString(), "Parsing Error",	
					JOptionPane.ERROR_MESSAGE);
          main.setStandardStatusLine();
      }
      finally {
          mediator.startInterface(true);
	  if (problem!=null)
	      problem.close();
      }
   }

// note: Expressions without parameters only emit the endExpr signal
   public void beginExpr(char id, String s) {
//System.out.println("start "+id+"="+s);
      switch (id) {
         case 'b' : 
            stack.push(children);
            if (children.hasNext()) currNode = children.next();
            break;

         case 'r' : 
            if (currNode == null) currNode = children.next();
            // otherwise we already fetched the node at branch point
            currGoal      = proof.getGoal(currNode);
            currTacletName= s;
            // set default state
            currFormula   = 0;
            currPosInTerm = PosInTerm.TOP_LEVEL; 
            loadedInsts   = null;
            interactive   = false;
            break;

         case 'f' :
            currFormula   = Integer.parseInt(s);
            break;

         case 't' :
            currPosInTerm = PosInTerm.parseReverseString(s);
            break;
            
         case 'i' :
            if (loadedInsts == null) loadedInsts = new LinkedList();
            loadedInsts.add(s);
            break;
            
         case 'h' :
            activeHeuristics = SLListOfHeuristic.EMPTY_LIST;
            StringTokenizer st = new StringTokenizer(s, ",");
	    while (st.hasMoreTokens()) {
                Name n = new Name(st.nextToken().trim());
		Heuristic h = (Heuristic) mediator.heur_ns().lookup(n);
                if (h == null) 
                   System.err.println("Unknow heuristic ignored: "+n);
                else
                   activeHeuristics = activeHeuristics.append(h);
	    }
            proof.setActiveHeuristics(activeHeuristics);
            break;
            
         case 'q' : // ifseqformula
//            Namespace varNS = mediator.var_ns();
//            varNS = currGoal.getVariableNamespace(varNS);
//            Term f = parseTerm(s, varNS);
            Sequent seq = currGoal.sequent();
            ifSeqFormulaList = ifSeqFormulaList.append(
               new IfFormulaInstSeq(seq, Integer.parseInt(s)));
            break;
      }
   }


   public void endExpr(char id) {
//System.out.println("end "+id);
      switch (id) {
         case 'b' : 
            children = (IteratorOfNode) stack.pop();
            break;
         case 'a' :
            interactive = true;
            break;
         case 'r' : 
            mediator.apply(constructApp(), currGoal, 
                           interactive, activeHeuristics);

            children = currNode.childrenIterator();
            currNode = null;
            ifSeqFormulaList = SLListOfIfFormulaInstantiation.EMPTY_LIST;

            break;
      }
   }
   
   
   private TacletApp constructApp() {

      TacletApp ourApp = null;

      PosInOccurrence pos = null;
      if (currFormula != 0) // otherwise we have no pos
         pos = PosInOccurrence.findInSequent(currGoal.sequent(),
                                             currFormula,
                                             currPosInTerm);

//System.err.print("Want to apply "+currTacletName+" at "+currGoal);
      SetOfTacletApp applics = 
         mediator.getTacletApplications(currGoal, currTacletName, pos,
                                        interactive, activeHeuristics);
      if (applics.size() != 1) {
         throw new RuntimeException(currTacletName+": found "+
            applics.size()+" applications. Don't know what to do!");
      }
      ourApp = applics.iterator().next();
      ourApp = constructInsts(ourApp);

      ourApp = ourApp.setIfFormulaInstantiations(ifSeqFormulaList,
         mediator.getJavaInfo(), mediator.getUserConstraint().getConstraint ());

      ourApp = ourApp.createDependingFunctions ( mediator.func_ns () );

      return ourApp;
   }
   
   
   private TacletApp constructInsts(TacletApp app) {
      if (loadedInsts == null) return app;
      SetOfSchemaVariable uninsts = app.uninstantiatedVars();

      // first pass: add variables
      Iterator it = loadedInsts.iterator();
      while (it.hasNext()) {
         String s = (String) it.next();
         int eq = s.indexOf('=');
         String varname = s.substring(0, eq);
         String value = s.substring(eq+1, s.length());
         
         SchemaVariable sv = lookupName(uninsts, varname);
         
         if (sv.isVariableSV()) {
	    LogicVariable lv=new LogicVariable(new Name(value),
					       app.getRealSort(sv));
	    Term instance = TermFactory.DEFAULT.createVariableTerm(lv);
	    app = app.addCheckedInstantiation(sv, instance, true);
         }
      }


      // second pass: add everything else
      it = loadedInsts.iterator();
      while (it.hasNext()) {
         String s = (String) it.next();
         int eq = s.indexOf('=');
         String varname = s.substring(0, eq);
         String value = s.substring(eq+1, s.length());
         
         SchemaVariable sv = lookupName(uninsts, varname);
         
         if (sv.isVariableSV()) {
         // ignore -- already done
	 } else if (sv.isProgramSV()) {
	    ProgramElement pe = 
               TacletInstantiationsTableModel.getProgramElement(
                  app, value, sv, mediator.getJavaInfo());
	    app = app.addInstantiation(sv, pe, true);
	 } else if ( sv.isDependingSV() ) {
	     app = app.createDependingConstant ( value,
						 sv,
						 true );
	 } else {
            Namespace varNS = mediator.var_ns();
            varNS = currGoal.getVariableNamespace(varNS);
            varNS = app.extendVarNamespaceForSV(varNS, sv);
            Term instance = parseTerm(value, varNS);
            app = app.addCheckedInstantiation(sv,instance, true); 
         }
      }
      return app;
   }
   
   
   private Term parseTerm(String value, Namespace varNS) {
         try {
            return TermParserFactory.createInstance().parse(
               new StringReader(value), null,
               mediator.getJavaInfo(), varNS, 
               mediator.func_ns(), mediator.sort_ns(),new AbbrevMap());
         } catch(ParserException e) {
              throw new RuntimeException("While parsing "+value+"\n"+
                                         varNS+"\n"+e);         
         }
   }
   
   
   private SchemaVariable lookupName(SetOfSchemaVariable set, String name) {
      IteratorOfSchemaVariable it = set.iterator();
      while (it.hasNext()) {
         SchemaVariable v = it.next();
         if (v.getName().equals(name)) return v;
      }
      return null; // handle this better!
   }
   
   

}
