// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof;

import java.util.Iterator;
import java.util.List;
import java.util.Collections;
import java.util.ArrayList;
import java.util.HashSet;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.gui.ProofSettings;
import de.uka.ilkd.key.rule.strategy.StrategyManager;

/**
 * A proof is represented as a tree whose nodes are created by applying rules on
 * the current (open) goals and dividing them up into several new subgoals. To
 * distinct between open goals (the one we are working on) and closed goals or
 * inner nodes we restrict the use of the word goal to open goals which are a
 * subset of the proof tree's leaves. 
 * This proof class represents a proof and encapsulates its tree
 * structure. Furthermore it offers services that deliver the open goals,
 * namespaces and several other informations about the current state of the
 * proof.  
 */
public class Proof implements Named {

    /** name of the proof */
    private Name name;

    /** the root of the proof */
    private Node root;

    /** list with prooftree listeners of this proof */
    private List listenerList = Collections.synchronizedList(new ArrayList(10));

    /** list with the open goals of the proof */ 
    private ListOfGoal openGoals = SLListOfGoal.EMPTY_LIST;

    /** during closure this can be set by "subTreeCompletelyClosed" */
    private Node closedSubtree = null;
    
    /** declarations &c, read from a problem file or otherwise */
    private String problemHeader = "";

    /** the update simplifier (may be moved to nodes)*/
    private SimultaneousUpdateSimplifier upd_simplifier;

    /** name maps of variables, functions, sorts and heuristics */
    private NamespaceSet namemap;

    /** the java information object: JavaInfo+TypeConverter */
    private Services services;

    /** list with the activated heuristics */
    private ListOfHeuristic activeHeuristics;

    /** User constraint */
    private ConstraintTableModel userConstraint = new ConstraintTableModel ();    

    /** Deliverer for new metavariables */
    private MetavariableDeliverer metavariableDeliverer;

    /** strategy that controls the automatic rule application */
    private StrategyManager heuristicStrategy = null;

    /** counter for generating unique taclet names */
    private int tacletNumber = 0;


    /** constructs a new empty proof with name */
    public Proof(Name name) {
	this.name = name;
	metavariableDeliverer = new MetavariableDeliverer ( this );
    }

    /** constructs a new empty proof */
    public Proof() {
	this(new Name(""));
    }


    /** constructs a new empty proof with name */
    public Proof(String name) {
	this(new Name(name));
    }
    
    public Proof(String name, Term problem, TacletIndex rules,
		 BuiltInRuleIndex builtInRules) {

        this(name);

        ConstrainedFormula cf = new ConstrainedFormula(problem);
        Node rootNode = new Node(this, 
				 Sequent.createSuccSequent
				 (Semisequent.EMPTY_SEMISEQUENT.insert(0,cf).semisequent()));
        setRoot(rootNode);

	Goal firstGoal = new Goal(rootNode, 
                                  new RuleAppIndex(new TacletAppIndex(rules),
						   new BuiltInRuleAppIndex(builtInRules)));
	openGoals = openGoals.prepend(firstGoal);

	if (closed())
	    fireProofClosed();
    }
    
    
    public Proof(String name, Term problem, String header, TacletIndex rules,
		 BuiltInRuleIndex builtInRules) {
        this(name, problem, rules, builtInRules);
        problemHeader = header;
    }

    /** 
     * returns the name of the proof. Describes in short what has to be proved.     
     * @return the name of the proof
     */
    public Name name() {
	return name;
    }
    
    public String header() {
       return problemHeader;
    }
    
    public void setHeader(String s) {
       problemHeader = s;
    }
    
    /** 
     * returns a collection of the namespaces valid for this proof
     */
    public NamespaceSet getNamespaces() {
	return namemap;
    }

    /** returns the JavaInfo with the java type information */
    public JavaInfo getJavaInfo() {
	return services.getJavaInfo();
    }

    /** sets the variable, function, sort, heuristics namespaces */
    public void setNamespaces(NamespaceSet ns) {
        namemap = ns;
	heuristicStrategy = new StrategyManager ( getNamespaces ().heuristics (),
						  this );
    }

    /** sets the JavaServices with the java type information */
    public void setJavaServices(Services serv) {
	this.services = serv;
        if (openGoals().size() > 1) 
            throw new IllegalStateException("Proof: Java Services set too late");
        openGoals().head().addProgramVariables(getJavaInfo().programVariables());
    }

    /** returns the activated heuristics */
    public ListOfHeuristic activeHeuristics() {
        if (activeHeuristics == null) { //default to stored values
	    Namespace h_nsp = namemap.heuristics();
	    activeHeuristics=ProofSettings.THE_SETTINGS
		.getHeuristicSettings().getActiveHeuristics(h_nsp);

// // default to full monty
//             ListOfHeuristic list = SLListOfHeuristic.EMPTY_LIST;
// 	    Namespace h_nsp = namemap.heuristics();
// 	    while (h_nsp != null) {
// 	        IteratorOfNamed it = h_nsp.elements().iterator();
// 	        while (it.hasNext()) list = list.prepend((Heuristic)it.next());
// 	        h_nsp = h_nsp.parent();
// 	    }
//             activeHeuristics = list;
        }
	return activeHeuristics;
    }
    
    /** sets the active heuristics 
     * @param activeHeuristics the ListOfHeuristic containing all
     * activated heuristics 
     */
    public void setActiveHeuristics(ListOfHeuristic activeHeuristics) {
	this.activeHeuristics = activeHeuristics;
    }
    
    /** @return the strategy manager for automatic rule application */
    public StrategyManager getHeuristicStrategy () {
	return heuristicStrategy;
    }

    /**
     * counter required to give the taclets a unique name
     */
    public int getUniqueTacletNr() {
        return tacletNumber++;
    }
    
    /** 
     * returns the default simplifier to be used (may be overwritten by branch
     * specific simplifiers in the future)
     * @return the SimultaneousUpdateSimplifier to be used as default one
     * 
     */
    public SimultaneousUpdateSimplifier simplifier() {
	return upd_simplifier;
    }

    /** 
     * sets the default simplifier
     * @param upd_simplifier the SimultaneousUpdateSimplifier to be used as
     * default (may be overwritten by branch specific simplifiers in the future)
     */
    public void setSimplifier(SimultaneousUpdateSimplifier upd_simplifier) {
	this.upd_simplifier = upd_simplifier;
    }

    /** returns the user constraint (table model)
     * @return the user constraint
     */
    public ConstraintTableModel getUserConstraint() {
	return userConstraint;
    }

    /** @return Deliverer of new metavariables (with unique names)*/
    public MetavariableDeliverer getMetavariableDeliverer () {
	return metavariableDeliverer;
    }

    /**
     * returns the root node of the proof
     */
    public Node root() {
	return root;
    }

    /** sets the root of the proof */
    public void setRoot(Node root) {
	if (this.root != null) {
	    throw new IllegalStateException
		("Tried to reset the root of the proof.");
	} else {
	    this.root = root;
	    fireProofStructureChanged();

	    if (closed())
		fireProofClosed();
	}
    }


    /** 
     * returns the list of open goals
     * @return list with the open goals
     */
    public ListOfGoal openGoals() {
	return openGoals;
    }

    /** 
     * removes the given goal and adds the new goals in list 
     * @param goal the old goal that has to be removed from list
     * @param newGoal the ListOfGoal with the new goals that were
     * result of a rule application on goal
     */
    public void replace(Goal oldGoal, ListOfGoal newGoals) {
	openGoals = openGoals.removeAll(oldGoal);

	if ( closed () )
	    fireProofClosed();
	else {
	    fireProofGoalRemoved(oldGoal);
	    add(newGoals);	
	}
    }

    /**
     * Add the given constraint to the closure constraint of the given
     * goal, i.e. the given goal is closed if p_c is satisfied.
     */
    public void closeGoal ( Goal p_goal, Constraint p_c ) {
	p_goal.addClosureConstraint ( p_c );

	removeClosedSubtree ();

	if ( closed () )
	    fireProofClosed();
    }

    /**
     * This is called by a Node that is the root of a subtree that is
     * closed
     */
    public void subtreeCompletelyClosed ( Node p_node ) {
	// This method will be called for nodes with increasing
	// distance to the root
	if ( closedSubtree == null )
	    closedSubtree = p_node;
    }

    /**
     * Use this information to remove the goals of the closed subtree
     */
    protected void removeClosedSubtree () {
	// give the information that a subtree is closed
	// to all nodes below this node
	if (closedSubtree != null) 
	    closedSubtree.setClosed();

	if ( !closed () && closedSubtree != null ) {

	    boolean        b    = false;
	    IteratorOfNode it   = closedSubtree.leavesIterator ();
	    Goal           goal;

	    while ( it.hasNext () ) {
		goal = getGoal ( it.next () );
		if ( goal != null ) {
		    b = true;
		    remove ( goal );
		}
	    }

	    if ( b )
		// For the moment it is necessary to fire the message ALWAYS
		// in order to detect branch closing.
		fireProofGoalsAdded ( SLListOfGoal.EMPTY_LIST );		
	}

	closedSubtree = null;
    }

    /** removes the given goal from the list of open goals. Take care
     * removing the last goal will fire the proofClosed event
     * @param goal the Goal to be removed
     */
    public void remove(Goal goal) {
	ListOfGoal newOpenGoals = openGoals.removeAll(goal);
	if (newOpenGoals != openGoals) {
	    openGoals = newOpenGoals;	
	    if (closed()) {
		fireProofClosed();
	    } else {
		fireProofGoalRemoved(goal);
	    }
	}
    }

    /** adds a new goal to the list of goals 
     * @param goal the Goal to be added 
     */
    public void add(Goal goal) {
	ListOfGoal newOpenGoals = openGoals.prepend(goal);
	if (openGoals != newOpenGoals) {
	    openGoals = newOpenGoals;
	    fireProofGoalsAdded(goal);
	}
    }

    /** adds a list with new goals to the list of open goals 
     * @param goals the ListOfGoal to be prepended 
     */
    public void add(ListOfGoal goals) {
	ListOfGoal newOpenGoals = openGoals.prepend(goals);
	if (openGoals != newOpenGoals) {
	    openGoals = newOpenGoals;
	}

	// For the moment it is necessary to fire the message ALWAYS
	// in order to detect branch closing.
	fireProofGoalsAdded(goals);
	
    }

    /** Return whether the remaining goals can be closed, i.e. whether
     * the conjunction of the constraints of the open goals is
     * satisfiable. In this case all remaining open goals are removed.
     */
    public boolean closed () {
	if ( root ().getRootSink ().isSatisfiable () ) {
	    Goal goal;
	    while ( openGoals != SLListOfGoal.EMPTY_LIST ) {
		goal      = openGoals.head ();
		openGoals = openGoals.tail ();
		fireProofGoalRemoved ( goal );
	    }
	    return true;
	}

	return false;
    }
    
    /** returns true iff sets back to the step that created the given
     * goal. If the undo operation was succesful true is returned.
     * @param goal the Goal desribing the location where to set back
     * @return true iff undo operation was succesfull.
     */
    public  boolean setBack(Goal goal) {		
	if (goal != null) {
	    Node parent = goal.node().parent();
	    if (parent != null) {
		openGoals = goal.setBack(openGoals);
		fireProofGoalsChanged();
		return true;
	    }
	}
	// root reached or proof closed
	return false;
    }

    /** sets back the subtree of node including the node 
     *	itself if the subtree is not closed yet.
     * @param node the node desribing the location where to set back
     * @return true iff undo operation was succesfull.
     */
    public boolean setBack(Node node) {
	if (node == root) {
	    return false;
	}
	Goal goal = getGoal(node);
	if (goal!=null) {
	    return setBack(goal);
	} else {
	    ListOfGoal goalList = getSubtreeGoals(node);
	    if (goalList != SLListOfGoal.EMPTY_LIST) {
		// get different parents		
		HashSet parentSet = new HashSet();
		ListOfGoal removeList = SLListOfGoal.EMPTY_LIST;
		IteratorOfGoal goalIt = goalList.iterator();
		while (goalIt.hasNext()) {
		    Goal nextGoal = goalIt.next();
		    if (!parentSet.contains(nextGoal.node().parent())) {
			removeList = removeList.prepend(nextGoal);
			parentSet.add(nextGoal.node().parent());
		    }
		}
		// iterate through removeList
		IteratorOfGoal removeIt = removeList.iterator();
		while (removeIt.hasNext()) {
		    setBack(removeIt.next());
		}
		return setBack(node);
	    }
	    return false;
	}
    }

    // ?? seems to be required for presentation uses
    // I think there is a mismatch between what the method does and the method's
    // name. Can the one who implemented this method check the name and write a
    // short comment about its purpose %%RB
    public void updateProof() {
	fireProofGoalsChanged();
    }


    /** fires the event that the proof has been expanded at the given node */
    protected synchronized void fireProofExpanded(Node node) {
	ProofTreeEvent e = new ProofTreeEvent(this, node);
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((ProofTreeListener)it.next()).proofExpanded(e);
	    }
	}
    } 

    /** fires the event that the proof has been pruned at the given node */
    protected synchronized void fireProofPruned(Node node, Node removedNode) {
	ProofTreeEvent e = new ProofTreeRemovedNodeEvent(this, node, removedNode);
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((ProofTreeListener)it.next()).proofPruned(e);
	    }
	}
    } 

    /** fires the event that the proof has been restructured */
    protected synchronized void fireProofStructureChanged() {
	ProofTreeEvent e = new ProofTreeEvent(this);
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((ProofTreeListener)it.next()).proofStructureChanged(e);
	    }
	}
    } 

    /** fires the event that a goal has been removed from the list of goals */
    protected synchronized void fireProofGoalRemoved(Goal goal) {
	ProofTreeEvent e = new ProofTreeEvent(this, goal);
	synchronized(listenerList) {
	    Iterator it  = listenerList.iterator(); 
	    while (it.hasNext()) {
		((ProofTreeListener)it.next()).proofGoalRemoved(e);
	    }	
	}
    }

    /** fires the event that new goals have been added to the list of
     * goals 
     */
    protected synchronized void fireProofGoalsAdded(ListOfGoal goals) {
	ProofTreeEvent e = new ProofTreeEvent(this, goals);
	synchronized(listenerList) {
	    Iterator it  = listenerList.iterator(); 
	    while (it.hasNext()) {
		((ProofTreeListener)it.next()).proofGoalsAdded(e);
	    }	
	}
    }

    /** fires the event that new goals have been added to the list of
     * goals 
     */
    protected synchronized void fireProofGoalsAdded(Goal goal) {
	fireProofGoalsAdded(SLListOfGoal.EMPTY_LIST.prepend(goal));
    }

    /** fires the event that the proof has been restructured */
    protected synchronized void fireProofGoalsChanged() {
	ProofTreeEvent e = new ProofTreeEvent(this);
	synchronized(listenerList) {
	    Iterator it = listenerList.iterator();
	    while (it.hasNext()) {
		((ProofTreeListener)it.next()).proofStructureChanged(e);
	    }
	}
    } 

    /** fires the event that the proof has closed. 
     * This event fired instead of the proofGoalRemoved event when
     * the last goal in list is removed.
     */
    protected synchronized void fireProofClosed() {
	ProofTreeEvent e = new ProofTreeEvent(this);
	synchronized(listenerList) {	    
	    Iterator it  = listenerList.iterator(); 
	    while (it.hasNext()) {
		((ProofTreeListener)it.next()).proofClosed(e);
	    }	
	}
    }

    /**
     * adds a listener to the proof 
     * @param listener the ProofTreeListener to be added
     */
    public synchronized void addProofTreeListener
	(ProofTreeListener listener) {
	synchronized(listenerList) {
	    listenerList.add(listener);
	}
    }

    /**
     * removes a listener from the proof 
     * @param listener the ProofTreeListener to be removed
     */
    public synchronized void removeProofTreeListener
	(ProofTreeListener listener) {
	synchronized(listenerList) {
	    listenerList.remove(listener);
	}
    }

    /** returns true if the given node is part of a Goal 
     * @return  true if the given node is part of a Goal 
     */
    public boolean isGoal(Node node) {	
	return getGoal(node) != null;
    }

    /** returns the goal that belongs to the given node or null if the
     * node is an inner one 
     * @return the goal that belongs to the given node or null if the
     * node is an inner one 
     */
    public Goal getGoal(Node node) {	
	Goal result = null;
	IteratorOfGoal it = openGoals.iterator();
	while (it.hasNext()) {
	    result = it.next();
	    if (result.node() == node) {
		return result;
	    }
	}
	return null;
    }

    /** returns the list of goals of the subtree starting with node 
     * @param node the Node where to start from
     * @return the list of goals of the subtree starting with node 
     */
    public ListOfGoal getSubtreeGoals(Node node) {	
	ListOfGoal result = SLListOfGoal.EMPTY_LIST;
	IteratorOfGoal goalsIt  = openGoals.iterator();
	while (goalsIt.hasNext()) {
	    Goal goal = goalsIt.next();
	    IteratorOfNode leavesIt = node.leavesIterator();
	    while (leavesIt.hasNext()) {
		if (leavesIt.next() == goal.node()) {
		    result = result.prepend(goal);
		}
	    }
	}
	return result;
    }

    /** returns true iff the given node is found in the proof tree 
     *	@param node the Node to search for
     *	@return true iff the given node is found in the proof tree 
    */
    public boolean find(Node node) {
	if (root == null) {
	    return false;
	}
	return root.find(node);
    }

    /** toString */
    public String toString() {
	StringBuffer result = new StringBuffer();
	    result.append("Proof -- ");
	    if (!"".equals(name.toString())) {
		result.append(name.toString());
	    } else {
		result.append("unnamed");
	    }
	result.append("\nProoftree:\n");
	result.append(root.toString());
	return result.toString();
    }
}
