// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.proof;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.rule.*;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Collections;

/**
 * manages the possible application of rules (RuleApps) 
 */
public class RuleAppIndex implements java.io.Serializable {

    private TacletAppIndex tacletAppIndex;
    private BuiltInRuleAppIndex builtInRuleAppIndex;
    private List listenerList = Collections.synchronizedList(new ArrayList(10));
               

    public RuleAppIndex(TacletAppIndex tacletAppIndex, 
			BuiltInRuleAppIndex builtInRuleAppIndex) {

	this.tacletAppIndex = tacletAppIndex;
	this.builtInRuleAppIndex = builtInRuleAppIndex;
    }

    /**
     * returns the TacletApp index for this ruleAppIndex. 
     */
    public TacletAppIndex tacletAppIndex() {
	return tacletAppIndex;
    }

    /**
     * returns the Taclet index for this ruleAppIndex. 
     */
    public TacletIndex tacletIndex() {
	return tacletAppIndex.tacletIndex();
    }

    /**
     * returns the built-in rule application index for this
     * ruleAppIndex. 
     */
    public BuiltInRuleAppIndex builtInRuleAppIndex() {
	return builtInRuleAppIndex;
    }


    /**
     * adds a change listener to the index
     * @param l the AppIndexListener to add
     */
    public void addAppIndexListener(AppIndexListener l) {
	listenerList.add(l);
    }
   
    /**
     * removes a change listener to the index
     * @param l the AppIndexListener to remove
     */
    public void removeAppIndexListener(AppIndexListener l) {
	listenerList.remove(l);
    }
     
    /**returns the set of rule applications for
     * the given heuristics 
     * at the given position of the given sequent.
     * @param heuristics the Heuristics that are enabled
     * @param interactive boolean indicates if the interactive or the
     * heuristic rules are to be selected.
     * @param pos the PosInOccurrence to focus
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     */
    public ListOfTacletApp getTacletAppAt(ListOfHeuristic heuristics,
					  boolean         interactive,
					  PosInOccurrence pos,
					  JavaInfo        javaInfo,
					  Constraint      userConstraint) { 
	return tacletAppIndex.getTacletAppAt(heuristics, interactive, pos,
					 javaInfo, userConstraint); 
    }
    

    /**
     * returns the rule applications at the given PosInOccurrence and at all
     * Positions below this. The method calls getTacletAppAt for all the
     * Positions below.
     * @param heuristics the heuristics that are given to method getTacletApp
     * @param interactive boolean indicates if the interactive or the
     * heuristic rules are to be selected.
     * @param pos the position where to start from
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return the possible rule applications 
     */
    public ListOfTacletApp getTacletAppAtAndBelow(ListOfHeuristic heuristics,
						  boolean         interactive,
						  PosInOccurrence pos,
						  JavaInfo        javaInfo,
						  Constraint      userConstraint) {
	return tacletAppIndex.getTacletAppAtAndBelow(heuristics,
						 interactive, 
						 pos, javaInfo, userConstraint);
    }


    /** 
     * collects all FindTacletInstantiations for the given
     * heuristics and position
     * @param interactive a boolean that indicates wether to  select
     * interactive or non-interactive rules 
     * @param pos the PosInOccurrence to focus
     * @param heuristics the Heuristics that are enabled
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of all possible instantiations
     */
    public ListOfNoPosTacletApp getFindTaclet(boolean         interactive,
					      PosInOccurrence pos,
					      ListOfHeuristic heuristics,
					      JavaInfo        javaInfo,
					      Constraint      userConstraint) { 
	return tacletAppIndex.getFindTaclet(interactive, pos, heuristics, javaInfo, userConstraint);
    }

    /** 
     * collects all NoFindTacletInstantiations for the given
     * heuristics 
     * @param interactive a boolean that indicates wether to  select
     * interactive or non-interactive rules 
     * @param heuristics the Heuristics that are enabled
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of all possible instantiations
     */
    public ListOfNoPosTacletApp getNoFindTaclet(boolean         interactive, 
						ListOfHeuristic heuristics,
						JavaInfo        javaInfo,
						Constraint      userConstraint) { 
	return tacletAppIndex.getNoFindTaclet(interactive, heuristics, 
					  javaInfo, userConstraint);	
    }


    /** 
     * collects all RewriteTacletInstantiations for the given
     * heuristics in a subterm of the constraintformula described by a
     * PosInOccurrence
     * @param interactive a boolean that indicates wether to  select
     * interactive or non-interactive rules 
     * @param pos the PosInOccurrence to focus 
     * @param heuristics the Heuristics that are enabled
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of all possible instantiations
     */
    public ListOfNoPosTacletApp getRewriteTaclet (boolean         interactive,
						  PosInOccurrence pos,
						  ListOfHeuristic heuristics,
						  JavaInfo        javaInfo,
						  Constraint      userConstraint) { 
	return tacletAppIndex.getRewriteTaclet(interactive, pos,
					       heuristics, javaInfo, userConstraint);
    }


    /** 
     * returns a list of built-in rule applications applicable
     * for the given goal, user defined constraint and position
     */
    public ListOfRuleApp getBuiltInRule(Goal            goal, 
					PosInOccurrence pos,
					Constraint      userConstraint) {
	 	 	
	 return builtInRuleAppIndex().getBuiltInRule(goal, pos, 
							   userConstraint);
     }

    /**
     * adds a new Taclet with instantiation information to the Taclet Index 
     * of this TacletAppIndex.
     * @param tacletApp the partially instantiated Taclet to add
     */
    public void addNoPosTacletApp(NoPosTacletApp tacletApp) {
	tacletAppIndex.addNoPosTacletApp(tacletApp);
    }


    /** 
     * called if a formula has been replaced
     * @param sci SequentChangeInfo describing the change of the sequent 
     */  
    public void sequentChanged(SequentChangeInfo sci){
	informAppIndexListener(new AppIndexChangeEvent(sci));
    }

    /** 
     * informs all observers, if a formula has been added, changed or 
     * removed
     */ 
    private void informAppIndexListener(AppIndexChangeEvent ce) {
	Iterator it = listenerList.iterator();
	while (it.hasNext()) {
	    ((AppIndexListener)it.next()).changed(ce);
	}	
    }
    

    /**
     * returns a new RuleAppIndex with a copied Taclet Index.
     * Attention: the listener lists are not copied
     */
    public RuleAppIndex copy() {
	return new RuleAppIndex(tacletAppIndex().copy(),
				builtInRuleAppIndex().copy());
    }


    public String toString() {
	return "RuleAppIndex without indexing, getting Taclets from"
	    +" TacletAppIndex "+tacletAppIndex;
    }

    /**
     * @link aggregation
     * @supplierCardinality 1
     * @clientCardinality 1 
     */
    /*#private TacletAppIndex lnkTacletAppIndex;*/

    /**
     * @clientCardinality 1
     * @clientQualifier possible rule applications
     * @supplierCardinality 1 
     */
    /*#Goal lnkGoal;*/

    /** @link aggregation 
     * @supplierCardinality *
     * @clientCardinality 1
     * @supplierRole managed rule applications*/
    /*#RuleApp lnkRuleApp;*/
}
