// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

/** the class manages the available TacletApps. This index has to be used if one
 * wants to ask for the available taclet application at a specific position (or
 * the whole sequent if taclet is a nofind rule). This means all taclet applications
 * that have a position information are managed by this index. For all others
 * the class TacletIndex is used. This index uses also the TacletIndex to calculate
 * new TacletApps.
 */

package de.uka.ilkd.key.proof;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.op.Operator;
import de.uka.ilkd.key.logic.op.SimultaneousUpdateOp;
import de.uka.ilkd.key.logic.op.Metavariable;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.rule.*;

public class TacletAppIndex implements java.io.Serializable {

    private TacletIndex tacletIndex;

    public TacletAppIndex(TacletIndex tacletIndex) {
	this.tacletIndex = tacletIndex;
    }

   /**
     * returns the Taclet index for this ruleAppIndex. 
     */
    public TacletIndex tacletIndex() {
	return tacletIndex;
    }

    private ListOfTacletApp getFindTacletAppAt(ListOfHeuristic heuristics,
					       boolean         interactive,
					       PosInOccurrence pos, 
					       JavaInfo        javaInfo,
					       Constraint      userConstraint) {
	ListOfNoPosTacletApp tacletInsts=SLListOfNoPosTacletApp.EMPTY_LIST;
	if (pos != null) {
	    tacletInsts = tacletInsts.prepend
		(getFindTaclet(interactive,
			       pos, heuristics, javaInfo, userConstraint));
	}	
	ListOfTacletApp sal = createTacletApps(tacletInsts, pos);
	return sal;
    }
     
    /** returns the set of rule applications for
     * the given heuristics 
     * at the given position of the given sequent.
     * @param heuristics the Heuristics that are enabled
     * @param interactive boolean indicates if the interactive or the
     * heuristic rules are to be selected.
     * @param pos the PosInOccurrence to focus
     */
    public ListOfTacletApp getTacletAppAt(ListOfHeuristic heuristics,
					  boolean         interactive,
					  PosInOccurrence pos,
					  JavaInfo        javaInfo,
					  Constraint      userConstraint) {
	ListOfTacletApp sal = getFindTacletAppAt(heuristics, interactive, 
						 pos, javaInfo, userConstraint);
	sal = prepend(sal, getNoFindTaclet(interactive, heuristics, javaInfo, userConstraint));
	return sal;
    }
    
    /** creates TacletApps out of each single NoPosTacletApp object
     * @param tacletInsts the list of NoPosTacletApps the TacletApps are to
     * be created from
     * @param pos the PosInOccurrence to focus
     * @return list of all created TacletApps
     */
    private ListOfTacletApp createTacletApps(ListOfNoPosTacletApp tacletInsts,
					     PosInOccurrence pos) {
	ListOfTacletApp result = SLListOfTacletApp.EMPTY_LIST;
	IteratorOfNoPosTacletApp it = tacletInsts.iterator();
	while (it.hasNext()) {
	    NoPosTacletApp tacletApp = it.next();
	    if (tacletApp.taclet() instanceof FindTaclet) {
		PosTacletApp newTacletApp = tacletApp.setPosInOccurrence ( pos );
		if (newTacletApp != null) {
		    result = result.prepend(newTacletApp);
		}
	    } else {
		result = result.prepend(tacletApp);
	    }
	}
	return result;
    }

    /** 
     * collects all NoFindTacletInstantiations for the given
     * heuristics 
     * @param interactive a boolean that indicates wether to  select
     * interactive or non-interactive rules 
     * @param heuristics the Heuristics that are enabled
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of all possible instantiations
     */
    public ListOfNoPosTacletApp getNoFindTaclet(boolean         interactive, 
						ListOfHeuristic heuristics,
						JavaInfo        javaInfo,
						Constraint      userConstraint) { 
	return tacletIndex.getNoFindTaclet(interactive, heuristics, javaInfo, userConstraint);	
    }

    /** 
     * collects all RewriteTacletInstantiations for the given
     * heuristics in a subterm of the constrainedFormula described by a
     * PosInOccurrence
     * @param interactive a boolean that indicates wether to  select
     * interactive or non-interactive rules 
     * @param pos the PosInOccurrence to focus 
     * @param heuristics the Heuristics that are enabled
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of all possible instantiations
     */
    public ListOfNoPosTacletApp getRewriteTaclet(boolean         interactive, 
						 PosInOccurrence pos,
						 ListOfHeuristic heuristics,
						 JavaInfo        javaInfo,
						 Constraint      userConstraint) { 

	ListOfNoPosTacletApp result = SLListOfNoPosTacletApp.EMPTY_LIST;

	Constraint c = pos.constrainedFormula().constraint();

	if ( pos.termBelowMetavariable () != null ) {
	    c = c.unify ( pos.constrainedFormula ().formula ().subAt ( pos.posInTerm () ),
			  pos.termBelowMetavariable () );
	    if ( !c.isSatisfiable () )
		return SLListOfNoPosTacletApp.EMPTY_LIST;
	}
	
	// check location conditions
	// rewrite taclets, whose find match on a location are only applicable if
	// and only if they have no rewrite part

	IteratorOfNoPosTacletApp rewriteIterator = 
	    tacletIndex.getRewriteTaclet(interactive, pos, c,
					 heuristics, javaInfo, 
					 userConstraint).iterator();


	while (rewriteIterator.hasNext()) {
	    NoPosTacletApp tacletApp = rewriteIterator.next();
	    if (!(tacletApp.taclet().hasReplaceWith() && 		
		  isLocation(pos.constrainedFormula().formula(), 
			     pos.posInTerm()))) {
		result=result.prepend(tacletApp);
	    }	   
	}
	
	return result;
    }


    /**
     * looks if the enclosing term is a simultaneous update and if the current
     * position describes a location
     * @param t the Term to be looked through
     * @param pos the PosInTerm describing the potential location
     * @return true if the described position is a location on the left side of
     * an assignment pair of an update
     */
    private boolean isLocation(Term t, PosInTerm pos) {
	Term term = t;

	Operator op = null;
	IteratorOfInteger it = pos.iterator();
	int subPos = -1;

	while (it.hasNext()) {
	    op = term.op();
	    subPos = it.next().intValue();
	    term = term.sub(subPos);
	}

	return (op != null && 
		op instanceof SimultaneousUpdateOp && 
		((SimultaneousUpdateOp)op).isLocation(subPos));
    }

    /** 
     * collects all FindTaclets with instantiations for the given
     * heuristics and position
     * @param interactive a boolean that indicates wether to  select
     * interactive or non-interactive rules 
     * @param pos the PosInOccurrence to focus
     * @param heuristics the Heuristics that are enabled
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of all possible instantiations
     */
    public ListOfNoPosTacletApp getFindTaclet(boolean         interactive,
					      PosInOccurrence pos,
					      ListOfHeuristic heuristics,
					      JavaInfo        javaInfo,
					      Constraint      userConstraint) { 
	ListOfNoPosTacletApp
	    tacletInsts=SLListOfNoPosTacletApp.EMPTY_LIST;
	if (pos.posInTerm() == PosInTerm.TOP_LEVEL) {
	    if (pos.isInAntec()) {
		tacletInsts = tacletInsts.prepend
		    (antecTaclet(interactive,
			       pos.constrainedFormula(),
			       heuristics, javaInfo, userConstraint)); 
	    } else {
		tacletInsts = tacletInsts.prepend
		    (succTaclet(interactive,
			      pos.constrainedFormula(),
			      heuristics, javaInfo, userConstraint)); 
	    } 
	} else {	    
	    tacletInsts = tacletInsts.prepend
		(getRewriteTaclet(interactive,pos,heuristics, javaInfo, userConstraint));
	}
	return tacletInsts;
    }


    /** 
     * collects all AntecedentTaclet instantiations for the given
     * heuristics and ConstrainedFormula
     * @param interactive a boolean that indicates wether to  select
     * interactive or non-interactive rules 
     * @param heuristics the Heuristics that are enabled
     * @param conFml the ConstrainedFormula the rules have to be
     * connected to
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of all possible instantiations
     */
    private ListOfNoPosTacletApp antecTaclet(boolean            interactive, 
					     ConstrainedFormula conFml,
					     ListOfHeuristic    heuristics,
					     JavaInfo           javaInfo,
					     Constraint         userConstraint) { 
	return tacletIndex.getAntecedentTaclet(interactive,
					       conFml.formula(),
					       conFml.constraint(),
					       heuristics,
					       javaInfo,
					       userConstraint); 
    }
    
    /** 
     * collects all SuccedentTaclet instantiations for the given
     * heuristics and ConstrainedFormula
     * @param interactive a boolean that indicates whether to  select
     * interactive or non-interactive rules 
     * @param heuristics the Heuristics that are enabled
     * @param conFml the ConstrainedFormula the rules have to be
     * connected to
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return list of all possible instantiations
     */
    private ListOfNoPosTacletApp succTaclet(boolean            interactive, 
					    ConstrainedFormula conFml,
					    ListOfHeuristic    heuristics,
					    JavaInfo           javaInfo,
					    Constraint         userConstraint) { 	
	return tacletIndex.getSuccedentTaclet(interactive,
					      conFml.formula(),
					      conFml.constraint(),
					      heuristics,
					      javaInfo,
					      userConstraint);    
    }



    /**
     * returns the rule applications at the given PosInOccurrence and at all
     * Positions below this. The method calls getTacletAppAt for all the
     * Positions below.
     * @param heuristics the heuristics that are given to method getTacletApp
     * @param interactive boolean indicates if the interactive or the
     * heuristic rules are to be selected.
     * @param pos the position where to start from
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return the possible rule applications 
     */
    public ListOfTacletApp getTacletAppAtAndBelow(ListOfHeuristic heuristics,
						  boolean         interactive,
						  PosInOccurrence pos,
						  JavaInfo        javaInfo,
						  Constraint      userConstraint) {
	return getTacletAppAtAndBelow(heuristics, interactive, pos, javaInfo, userConstraint, 0);

    }


    private ListOfTacletApp getTacletAppAtAndBelow(ListOfHeuristic heuristics,
						   boolean         interactive,
						   PosInOccurrence pos,
						   JavaInfo        javaInfo,
						   Constraint      userConstraint,
						   int             depth) {
	Term            term   = pos.subTerm ();
	ListOfTacletApp result = SLListOfTacletApp.EMPTY_LIST;
	if ( term.op () instanceof Metavariable ) {
	    if ( pos.termBelowMetavariable () == null ) {
		Term metaTerm = userConstraint
		    .getInstantiation ( (Metavariable)term.op () );
		if ( metaTerm.op () != term.op () )
		    result =  getTacletAppAtAndBelow ( heuristics, interactive,
						       pos.setTermBelowMetavariable ( metaTerm ),
						       javaInfo, userConstraint, depth );
	    }
	} else {
	    //	    if (!(term.op() instanceof UpdateOp && depth>0)) {	    
	    result=(getFindTacletAppAt(heuristics, interactive, pos, javaInfo, userConstraint));
	    //	    }
	    for(int i=0; i<term.arity(); i++) {
		result = result.prepend
		    (getTacletAppAtAndBelow(heuristics,interactive,
					    pos.down(i), javaInfo, userConstraint, depth++));
	    }
	    // 	System.out.print("Get Taclets for "+term);
	    // 	print(result);
	    // 	System.out.println("");
	    result=prepend(result,getNoFindTaclet(interactive, heuristics, javaInfo, userConstraint));
	}
	return result;
    }



    /** 
     * called if a formula has been replaced
     * @param sci SequentChangeInfo describing the change of the sequent 
     */  
    public void sequentChanged(SequentChangeInfo sci){
    }

    /**
     * adds a new Taclet with instantiation inforrmation to the Taclet Index 
     * of this TacletAppIndex.
     * @param tacletApp the partially instantiated Taclet to add
     */
    public void addNoPosTacletApp(NoPosTacletApp tacletApp) {
	tacletIndex.add(tacletApp);
    }

    /**
     * adds a new Taclet to the Taclet Index of this RuleAppIndex.
     * @param taclet the Taclet to add
     */
    public void addTaclet(Taclet taclet) {
	tacletIndex.add(taclet);
    }


    /**
     * returns a new RuleAppIndex with a copied Taclet Index
     */
    public TacletAppIndex copy() {	
	return new TacletAppIndex(tacletIndex().copy());
    }


    public String toString() {
	return "TacletAppIndex without indexing, getting Taclets from"
	    +" TacletIndex "+tacletIndex;
    }

    // helper because ListOfNoPosTacletApp is no ListOfTacletApp
    private static ListOfTacletApp prepend(ListOfTacletApp l1, 
					 ListOfNoPosTacletApp l2) {
	IteratorOfNoPosTacletApp it=l2.iterator();
	while (it.hasNext()) {
	    l1=l1.prepend(it.next());
	}
	return l1;
    }


}
