// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof;

import javax.swing.table.AbstractTableModel;
import javax.swing.*;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.gui.AbbrevMap;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.ArraySort;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.Label;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Expression;
import java.io.*;
import java.util.Vector;
import java.util.HashMap;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.logic.op.IteratorOfSchemaVariable;
import de.uka.ilkd.key.logic.op.IteratorOfEntryOfSchemaVariableAndInstantiationEntry;
import de.uka.ilkd.key.logic.op.EntryOfSchemaVariableAndInstantiationEntry;
import de.uka.ilkd.key.rule.inst.InstantiationEntry;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.op.LogicVariable;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.rule.TacletApp;
import de.uka.ilkd.key.rule.NewVarcond;
import de.uka.ilkd.key.rule.TacletVariableCollector;
import de.uka.ilkd.key.rule.inst.SortException;
import de.uka.ilkd.key.rule.inst.RigidnessException;
import de.uka.ilkd.key.java.TypeConverter;

public class TacletInstantiationsTableModel extends AbstractTableModel {
	
    /** the instantiations entries */
    private Vector entries; 
    /** the related rule application */
    private TacletApp app;
    /** the integer defines the row until which no editing is possible */
    private int noEditRow;
    /** universal namespace of variables, minimum for input in a row */
    private Namespace var_ns;
    /** namespace of functions */
    private Namespace func_ns;
    /** namespace of sorts */
    private Namespace sort_ns;
    /** the java information object */
    private JavaInfo javaInfo;
    /** the abbreviationmap */
    private AbbrevMap scm;
    

    // This stuff is for generating unique name proposal for schema variables
    // for labels or program variables ...
    private static final String LABEL_NAME_PREFIX = "_label";
    private static final String VARIABLE_NAME_PREFIX = "_var";
    private static final String DEPENDING_VARIABLE_NAME_POSTFIX = "_";


    // Should be proof-specific    
    public static int globalLabelNameCnt = 0;
    public static int globalVariableNameCnt = 0;
    private static HashMap globalDependingVariableNameCnt = new HashMap ();

	
    /** create new data model for tree 
     * @param app the TacletApp where to get the necessary entries
     */
    public TacletInstantiationsTableModel(TacletApp app, 
					  JavaInfo javaInfo,
					  Namespace var_ns,
					  Namespace func_ns, 
					  Namespace sort_ns,
					  AbbrevMap scm) {
	this.app = app;
	this.javaInfo = javaInfo;
	this.var_ns = var_ns;
	this.func_ns = func_ns;
	this.sort_ns = sort_ns;
	this.scm = scm;

	entries = createEntryArray(app);
    }
	
    /** creates a Vector with the row entries of the table
    */
    private Vector createEntryArray(TacletApp app) {
	Vector rowVec = new Vector();       
	IteratorOfSchemaVariable it = app.instantiations().svIterator();	
	int count = 0;

	while (it.hasNext()) {
	    SchemaVariable sv = it.next();
	    Object[] column = new Object[2];
	    column[0] = sv;
	    column[1] = app.instantiations().getInstantiation(sv);	    
	    rowVec.add(column);
	    count++;
	}
	noEditRow = count - 1;

	IteratorOfSchemaVariable varIt = app.uninstantiatedVars().iterator();

	while (varIt.hasNext()) {
	    Object[] column = new Object[2];	    
	    SchemaVariable var = varIt.next();
	    if (!app.taclet ().getIfFindVariables ().contains(var)) {
		column[0] = var;
		
		// NEW ... : create an appropriate and unique proposal for the name ...

		String nameProposal;

		if ( var.isDependingSV() )
		    nameProposal = getNameProposalForDependingVariable(app, var);
		else
		    nameProposal = getNameProposalForSchemaVariable(var);

		if(nameProposal != null) {
		    // A proposal is available ...
		    column[1] = nameProposal;
		}
		
		rowVec.add(column);	
	    }
	}

	return rowVec;
    }

    /** adds an instantiation of a schemavariable */
    public void addInstantiationEntry(int row, Term instantiation) {
	((Object[])entries.get(row))[1] = instantiation;	
    }

    /** return the rule application which is the table models base
     *  @return the Ruleapp 
     */
    public TacletApp application() {
	return app;
    }

    /** number of colums
     * @return number of colums
     */
    public int getColumnCount() { 
	return 2; 
    }

    /** number of rows
     * @return number of rows
     */
    public int getRowCount() { 
	return entries.size();
    }

    /** returns true iff an instantiation is missing 
     * @return true iff an instantiation is missing
     */
    public boolean isCellEditable(int rowIndex, int columnIndex) {
	return (rowIndex > noEditRow) && (columnIndex > 0);
    }


    /** parses the given string that represents the term (or formula)
     * using the given variable namespace and the default namespaces
     * for functions and sorts
     * @param s the String to parse 
     * @param sort the expected sort
     * @param varNS the variable namespace
     */
    public Term parseTerm(String s, Namespace varNS) 
	throws ParserException
    {
	Term term = TermParserFactory.createInstance().parse(
           new StringReader(s), null, javaInfo, varNS, func_ns, sort_ns, scm);
        return term;
    }
    
    /**
     * Parse the declaration of an identifier (i.e. the declaration of
     * a variable or skolem function)
     */
    public IdDeclaration parseIdDeclaration ( String s )
	throws ParserException {
	try {
	    DeclParser parser =
		new DeclParser ( new KeYLexer ( new StringReader ( s ) ),
				 "",
				 var_ns,           // should not be needed
				 func_ns,          // should not be needed
				 sort_ns,
				 new Namespace (), // should not be needed
				 javaInfo );       // should not be needed
	    return parser.id_declaration ();
        } catch (antlr.RecognitionException re) {
            throw new ParserException(re.getMessage(),
                                      new Location(re.getFilename(),
                                                   re.getLine(),
                                                   re.getColumn()));
        } catch (antlr.TokenStreamException tse) {
            throw new ParserException(tse.getMessage(), null);
        }
    }

    /**
     * throws an exception iff no input in indicated row, and no
     * metavariable instantiation is possible
     *
     */
    private void checkNeededInputAvailable(int irow) 
	throws MissingInstantiationException {

	final int icol = 1;

	if ( ( getValueAt(irow, icol) == null ||
	       ((String)getValueAt(irow, icol)).length() == 0 ) &&
	     !app.canUseMVAPriori ( (SchemaVariable)getValueAt(irow, 0) ) ) {

	    throw new MissingInstantiationException
		("" + getValueAt(irow, 0), irow, 0, false);
	}
    }


    /**
     * @return true iff this row is not empty (i.e. a string of data
     * is available)
     */
    private boolean isInputAvailable(int irow) {
	return getValueAt(irow, 1) != null && ((String)getValueAt(irow,1)).length() != 0;
    }


    /**
     * parses the indicated row and returns a Term corresponding to the
     * entry in the row
     *
     * @param irow the row to be parsed
     * @return the parsed term
     */
    private Term parseRow(int irow, Namespace varNS) 
	throws SVInstantiationParserException,
	       MissingInstantiationException {

	String instantiation = (String) getValueAt(irow, 1); 

	if ( instantiation == null || "".equals(instantiation) ) {
	    throw new MissingInstantiationException("", irow, 0, false);
	}

	try {
	    return parseTerm(instantiation, varNS);
	} catch (ParserException pe) {
	    Location loc = pe.getLocation();
	    if (loc != null) {
		throw new SVInstantiationParserException(instantiation, 
							 irow + (loc.getLine() <= 0 ? 0 
								 : loc.getLine()), 
							 loc.getColumn(), pe.getMessage(), 
							 false);
	    } else {
		throw new SVInstantiationParserException(instantiation, 
							 irow, -1, 
							 pe.getMessage(), 
							 false);
	    }
	} 
    }

    /**
     * parses the indicated row and returns a identifier declaration
     * corresponding to the entry in the row
     *
     * @param irow the row to be parsed
     * @return the parsed declaration
     */
    private IdDeclaration parseIdDeclaration(int irow) 
	throws SVInstantiationParserException,
	       MissingInstantiationException {

	String instantiation = (String) getValueAt(irow, 1); 

	if ( instantiation == null || "".equals(instantiation) ) {
	    throw new MissingInstantiationException("", irow, 0, false);
	}

	try {
	    return parseIdDeclaration(instantiation);
	} catch (ParserException pe) {
	    Location loc = pe.getLocation();
	    if (loc != null) {
		throw new SVInstantiationParserException(instantiation, 
							 irow + (loc.getLine() <= 0 ? 0 
								 : loc.getLine()), 
							 loc.getColumn(), pe.getMessage(), 
							 false);
	    } else {
		throw new SVInstantiationParserException(instantiation, 
							 irow, -1, 
							 pe.getMessage(), 
							 false);
	    }
	} 
    }

    public static ProgramElement getProgramElement(TacletApp app, 
						   String instantiation,
						   SchemaVariable sv,
						   JavaInfo javaInfo) {
	Sort svSort = sv.sort();
	if (svSort == ProgramSVSort.LABEL) {
	    return new ProgramElementName(instantiation);
	} else if (svSort == ProgramSVSort.VARIABLE 
		   || svSort == ProgramSVSort.BOOLEANVARIABLE) {
	    NewVarcond nvc = app.taclet().varDeclaredNew(sv);
	    if (nvc != null) {
		KeYJavaType kjt;
		Object o = nvc.getSortDefiningObject();
		if (o instanceof SchemaVariable) {
	            TypeConverter tc=javaInfo.getTypeConverter();
		    SchemaVariable peerSV=(SchemaVariable)o;
		    Object peerInst = app.instantiations().
		    getInstantiation(peerSV);
		    Expression peerInstExpr;
		    if (peerInst instanceof Term) {
			peerInstExpr=tc.convertToProgramElement((Term)peerInst);
		    } else {
			peerInstExpr=(Expression)peerInst;
		    } 
		    kjt = tc.getKeYJavaType(peerInstExpr);
		    if(nvc.isDefinedByElementSort()){
		        Sort s = kjt.getSort();
			if(s instanceof ArraySort) s = ((ArraySort)s).elementSort();
			kjt = javaInfo.getKeYJavaType(s);
		    }
		} else {
		    kjt = javaInfo.getKeYJavaType(((Sort)o).name().toString());
		}
		return new ProgramVariable
		    (new ProgramElementName(instantiation), kjt); 
	    }	
	}
	return null;
    }


    /**
     * parses the indicated row and returns the ProgramElement 
     * corresponding to the entry in the row
     * @param irow the row to be parsed
     * @return the parsed term
     */
    private ProgramElement parseRow(int irow) 
	throws SVInstantiationParserException {

	String instantiation = (String) getValueAt(irow, 1); 
	SchemaVariable sv= (SchemaVariable)getValueAt(irow, 0);
	ProgramElement pe = getProgramElement(app, instantiation, sv, javaInfo);
	if (pe == null) {
	    throw new SVInstantiationParserException
		(instantiation, irow, -1, "Unexpected sort: " + sv.sort() 
		 + "." + "Label SV or a program variable SV expected" 
		 + " declared as new.", false);
	}
	return pe;
    }
    

    /**
     * creates new rule app with all inserted instantiations in the variable
     * instantiations table
     * @throws SVInstantiationException if the instantiation is incorrect
     */
    public TacletApp createTacletAppFromVarInsts() 
	throws SVInstantiationException {

	TacletApp      result = app;
	SchemaVariable sv     = null;
	Sort           sort   = null;
	int            irow   = 0;

	try {

	    for (irow = noEditRow+1; irow < entries.size(); irow++) {
		checkNeededInputAvailable(irow);
		sv   = (SchemaVariable) getValueAt(irow, 0);
		sort = null;

		if ( sv.isVariableSV () || sv.isDependingSV () ) {
		    IdDeclaration idd = parseIdDeclaration ( irow );

		    sort = idd.getSort ();
		    if ( sort == null ) {
			try {
			    sort = result.getRealSort ( sv );
			} catch ( SortException e ) {
			    throw new MissingSortException ( "" + sv,
							     irow, 0 );
			}
		    }

		    if ( sv.isVariableSV () ) {
			LogicVariable lv = 
			    new LogicVariable ( new Name ( idd.getName () ),
						sort );
			Term instance = TermFactory.DEFAULT
			    .createVariableTerm ( lv );
			result = result.addCheckedInstantiation
			    ( sv, instance, true );
		    } else {
			// sv.isDependingSV ()
			result = result.createDependingConstant
			    ( idd.getName (), sv, sort, true );
		    }
		} else if (sv.isProgramSV()) {
		    ProgramElement pe = parseRow(irow);
		    result = result.addInstantiation(sv, pe, true);
		}
	    }
	    SchemaVariable problemVarSV = result.varSVNameConflict();

	    if (problemVarSV != null) {
		throw new SVInstantiationParserException
		    ( "", getSVRow(problemVarSV), 0, 
		      "Ambiguous instantiation of schema variable " + 
		      problemVarSV, false);
	    }

	    for (irow = noEditRow+1; irow < entries.size(); irow++) {
		if ( !isInputAvailable ( irow ) )
		    continue;

		sv   = (SchemaVariable)getValueAt(irow, 0);

		if ( sv.isVariableSV () || sv.isDependingSV () )
		    continue;

		sort = null;

		if (sv.isProgramSV()) {
		    ProgramElement pe = parseRow(irow);
		    result = result.addInstantiation(sv, pe, true);
		} else {
		    Namespace extVarNS = 
			result.extendVarNamespaceForSV(var_ns, sv);
		    Term instance = parseRow(irow, extVarNS);
		    sort = instance.sort ();
		    // check variable conditions for current schema variable
		    if (!result.taclet().checkVariableConditions(sv, instance, result.instantiations())) {
			throw new SVInstantiationParserException((String) getValueAt(irow, 1), 
								 irow, -1, 
								 "variable condition failed", 
								 false);
		    }
		    try {
			result =
			    result.addCheckedInstantiation(sv,instance, true);
		    } catch ( RigidnessException e ) {
			throw new SVRigidnessException ( "" + sv, irow, 0 );
		    }
		}
	    }

	} catch ( SortException e ) {
	    throw new SortMismatchException ( "" + sv, sort, irow, 0 );
	}

	return result;

    }


    /** sets the Value of the cell */
    public void setValueAt(Object instantiation, int rowIndex, 
			   int columnIndex) { 
	((Object[])entries.get(rowIndex))[columnIndex] = instantiation;
	
    }
   

    /** get value at the specified row and col
     * @return the value
     */
    public Object getValueAt(int row, int col) {
	return 	((Object[])entries.get(row))[col]; 
    }

    /** returns the index of the row the given Schemavariable stands 
     *@return the index of the row the given Schemavariable stands (-1
     * if not found)  
     */
    public int getSVRow(SchemaVariable sv) {
	for (int i = 0; i < entries.size(); i++) {
	    if (getValueAt(i, 0).equals(sv)) {
		return i;
	    }
	}
	return -1;
    }

    

    /**
     * Generates a proposal for the instantiation of the given
     * schema variable.
     * Currently this proposal only serves for the case of names (for labels or 
     * variables) and generates a unique new name.
     * @param var the schema var for which a unique name should be proposed as 
     * its instantiation.
     * @return a new and unquie name that can be used as an instantiation for
     *         the given schema variable.
     */
    public static String getNameProposalForSchemaVariable(SchemaVariable var){
	String result = null;

	Sort svSort = var.sort();
	if (svSort == ProgramSVSort.LABEL) {
	    result = LABEL_NAME_PREFIX + globalLabelNameCnt++;
	} else if (svSort == ProgramSVSort.VARIABLE 
		   || svSort == ProgramSVSort.BOOLEANVARIABLE) {
	    result = VARIABLE_NAME_PREFIX + globalVariableNameCnt++;
	}
	
	return result;
    }


    /**
     * Generates a proposal for the instantiation of the given term
     * schema variable, which is declared as depending (i.e. skolem
     * function).
     * @param var the schema var for which a unique name should be proposed as 
     * its instantiation.
     * @return a new and unquie name that can be used as an instantiation for
     *         the given schema variable.
     */
    public static String getNameProposalForDependingVariable(TacletApp      p_app,
							     SchemaVariable p_var){
	String res;
	
	SchemaVariable v = p_app.taclet ().getNameCorrespondent ( p_var );

	if ( v != null && p_app.instantiations ().isInstantiated ( v ) )
	    res = "" + p_app.instantiations ().getInstantiation ( v );
	else
	    // ... otherwise use the name of the dependingSV
	    res = "" + p_var.name ();
	res += DEPENDING_VARIABLE_NAME_POSTFIX;

	synchronized ( globalDependingVariableNameCnt ) {
	    Integer cnt = (Integer)globalDependingVariableNameCnt.get ( res );
	    
	    if ( cnt == null )
		cnt = new Integer ( 0 );
	    else
		cnt = new Integer ( cnt.intValue () + 1 );

	    globalDependingVariableNameCnt.put ( res, cnt );

	    //	    if ( cnt.intValue () > 1 )
	    res += cnt;
	}

	return res;
    }


    public static String getBasenameProposalForMetavariable(TacletApp      p_app,
							    SchemaVariable p_var){
	String res;
	
	SchemaVariable v = p_app.taclet ().getNameCorrespondent ( p_var );

	if ( v != null && p_app.instantiations ().isInstantiated ( v ) )
	    res = "" + p_app.instantiations ().getInstantiation ( v );
	else
	    // ... otherwise use the name of the SV
	    res = "" + p_var.name ();

	return res.toUpperCase ();
    }


}
