// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof.decproc;

import de.uka.ilkd.key.logic.op.DecisionProcedureSimplifyOp;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.gui.SequentPrinter;
import de.uka.ilkd.key.gui.ProgramPrinter;
import de.uka.ilkd.key.gui.NotationInfo;
import de.uka.ilkd.key.util.Debug;
import java.io.*;
import javax.swing.*;
import java.util.*;
import de.uka.ilkd.key.proof.decproc.ConstraintSet;
import de.uka.ilkd.key.logic.Constraint;

/**
 This class invokes the decision procedure
 "Simplify" which is part of ESC/Java by Compaq.
 */

public class DecisionProcedureSimplify {
    
    private static final String SIMPLIFY_COMMAND = "Simplify";

    private Goal goal;
    
    private Constraint userConstraint;
    
    public static final String SPACER = 
	"\\<---This is just a spacer between "+ 
	"sequent and Simplify input resp. Simplify output--->\\";

    public DecisionProcedureSimplify(Goal goal, 
				     Constraint userConstraint) {
	this.goal = goal;
	this.userConstraint = userConstraint;	
    }
    

    /**
     * checks whether the goal can be handled by
     * decision procedure "Simplify"
     */
    public boolean applicable() {
	return true;
    }

    /** Read the input until end of file and return contents in a
     * single string containing all line breaks. */
    private static String read(InputStream in) 
	throws IOException {
	String lineSeparator = System.getProperty("line.separator");
	BufferedReader reader = new BufferedReader
	    (new InputStreamReader(in));
	StringBuffer sb = new StringBuffer();
	String line;
	while ((line = reader.readLine()) != null) {
	    sb.append(line).append(lineSeparator);
	}
	return sb.toString();
    }

    private static String getCurrentDateString() {
	Calendar c = Calendar.getInstance();
	StringBuffer sb = new StringBuffer();
	sb.append(c.get(Calendar.YEAR))
	    .append(':')
	    .append(c.get(Calendar.MONTH) + 1)
	    .append(':')
	    .append(c.get(Calendar.DATE))
	    .append('_')
	    .append(c.get(Calendar.HOUR_OF_DAY))
	    .append(':')
	    .append(c.get(Calendar.MINUTE))
	    .append(':')
	    .append(c.get(Calendar.SECOND))
	    .append(':')
	    .append(c.get(Calendar.MILLISECOND));
	return sb.toString();
    }

    /**
    * the method responsible for making various calls to 
    * Simplify with different Constraints
    */
    public SimplifyResult run(){
	ConstraintSet constraintSet = new ConstraintSet
	    (goal, userConstraint);
	Constraint internalConstraint = constraintSet.
	    chooseConstraintSet(); 
	SimplifyResult res = runSimplify(constraintSet);
	if (res.getResult()) {
	    return res;
	} else {
	    if (constraintSet.addUserConstraint(userConstraint)) {
		return runSimplify(constraintSet);
	    } else {
		return res;
	    }
	}
      
    }      
    /**
     * runs first the conversion into the Simplify syntax,
     * and runs Simplify via a temporary file.
     */
    protected SimplifyResult runSimplify(ConstraintSet constraintSet) {
	try {
	    SimplifyTranslation st = new SimplifyTranslation
		(goal.sequent(), constraintSet, 
		 goal.node().getRestrictedMetavariables());

	    /* It seems, there are problems with Process and
	     * Standard-IO. Even the simplest program doesn't work
	     * correctly with J2SE 1.4. But I haven't found a bug in
	     * Sun Bug Database. 
	     *
	     * try {
	     *     Process p = Runtime.getRuntime().exec("cat");
	     *     OutputStream out = p.getOutputStream();
	     *     InputStream in = p.getInputStream();
	     *     out.write("Hello World!\n".getBytes());
	     *     out.close();
	     *     BufferedReader rd = new BufferedReader(new InputStreamReader(in));
	     *     String s;
	     *     while ((s = rd.readLine()) != null) {
	     *         System.out.println("[" + s + "]");
	     *     }
	     * } catch (IOException ioe) {
	     *     ioe.printStackTrace();
	     * }
	     * 
	     * Instead of control with Standard-IO, I control the
	     * simplify process with a temporary file. -hs. */

	    File file = File.createTempFile("key-simplify", null);
	    PrintWriter out = new PrintWriter(new FileWriter(file));
	    out.println(st.getText());
	    out.close();
	    
	    Process p = Runtime.getRuntime().exec(new String[]{ SIMPLIFY_COMMAND, file.getPath() });
	    InputStream in = p.getInputStream();

	    String response = read(in);
	    in.close();

	    file.delete();

    
            // read the output from the command
	    System.out.println("Here is what Simplify has to say:\n");
	    System.out.println(response);

	    //This part is responsible for logging
	    String logdir = System.getProperty("key.simplify.logdir");
	    if (logdir == null || logdir.trim().length() == 0) {
		System.err.println("$KEY_SIMPLIFY_LOG_DIR is empty or non-existent");
		System.err.println("Logging of Simplify disabled.");
	    } else {
		try {
		    String logFileName = "simplify-log_" + getCurrentDateString() + ".log";
		    PrintWriter logfile = new PrintWriter(new BufferedWriter(new FileWriter(new File(logdir, logFileName))));
		    SequentPrinter sp = new SequentPrinter(new ProgramPrinter(logfile), 
							   NotationInfo.createInstance(), 
							   goal.sequent());
		    sp.printSequent(goal.sequent());
		    logfile.print(sp.result().toString());
		    logfile.println(SPACER);
		    logfile.println(st.getText());
		    logfile.println(SPACER);
		    logfile.println(response);
		    logfile.close();
		} catch (IOException ioe) {
		    System.out.println("error while trying to log:\n" + ioe);
		}
	    }
	    //End of logging part

	    if (response.indexOf("Valid.") > 0) {
		//System.out.println("Simplify has decided and found the formula to be valid.");
		return new SimplifyResult(true, response, constraintSet.chosenConstraint);
	    } else {
		return new SimplifyResult(false, response, constraintSet.chosenConstraint);
	    }
	} catch (IOException ioe) {
	    JOptionPane.showMessageDialog(new JFrame(), "\"Simplify\" execution failed.\n"+
                                       "Is the binary in your $PATH?", "Error", 
					  JOptionPane.ERROR_MESSAGE);
            throw new RuntimeException("\"Simplify\" execution failed.\n"+
                                       "Is the binary in your $PATH?");
	} catch (SimplifyException se) {
	    return new SimplifyResult(false, se.toString(), constraintSet.chosenConstraint);
        }	
    }

}
