// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.proof.decproc;


import java.util.*;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.proof.decproc.*;
import de.uka.ilkd.key.rule.SyntacticalReplaceVisitor;
import de.uka.ilkd.key.java.JavaInfo;

import de.uka.ilkd.key.util.Debug;


final class SimplifyTranslation {

    /**
     * Just a constructor which starts the conversion to Simplify syntax.
     * The result can be fetched with @see getText-
     * @param sequent The sequent which shall be translated.
     * @param cs The constraints which shall be incorporated.
     * @param localmv The local metavariables, should be the ones introduced
     *    after the last branch.
     */
    SimplifyTranslation(Sequent sequent, ConstraintSet cs, 
			SetOfMetavariable localmv) throws SimplifyException  {
	constraintSet = cs;
	localMetavariables = localmv;
	translate(sequent);
	text = predicate.toString() + produceClosure(sb);
	System.out.println("SimplifyTranslation:\n" + text);
	System.out.println(notes);
    }
    
    /**
     * Returns the sequent given with the constructor in Simplify 
     * syntax (as far as possible)
     * @returns The sequent given with the constructor in Simplify 
     * syntax (as far as possible)
     */
    public String getText() {
	return text;
    }
    
    /**
     * Goes through the collected metavariables and quantifies them with 
     * universal-quantifieres if they are global and with existential 
     * quantifiers if they are local.
     * @param s The StringBuffer the quantifieres shall be pre- and 
     * the trailing parantheses be appended.
     * @returns The modified StringBuffer as a String.
     */
    private String produceClosure(StringBuffer s) {
	StringBuffer tmp = new StringBuffer("(");
	tmp.append(DecisionProcedureSimplifyOp.ALL)
	    .append(" (");
	for (Iterator i = usedGlobalMv.iterator(); i.hasNext(); ) {
	    tmp.append(' ')
		.append(getUniqueVariableName((Metavariable)i.next()));
	}
	tmp.append(")");

	tmp.append("(")
	    .append(DecisionProcedureSimplifyOp.EX)
	    .append(" (");
	for (Iterator i = usedLocalMv.iterator(); i.hasNext(); ) {
	    tmp.append(' ')
		.append(getUniqueVariableName((Metavariable)i.next()));
	}
	tmp.append(") ");

	tmp.append(s);

	tmp.append(" ))");

	return tmp.toString();
    }

    /** Translates the given sequent into "Simplify" input syntax and
     * adds the resulting string to the StringBuffer sb.
     *
     * @param s the Sequent which should be written in Simplify syntax
     */
    private void translate(Sequent sequent) throws SimplifyException  {
	sb.append('(').append(DecisionProcedureSimplifyOp.IMP).append(' ');
	translate(sequent.antecedent(), true);
       	sb.append("\n");
	translate(sequent.succedent(), false);
	sb.append(')');
    }

    /** Translates the given Semisequent into "Simplify" input syntax
     * and adds the resulting string to the StringBuffer sb.
     *
     * @param ss the SemiSequent which should be written in Simplify syntax
     * @param antesucc true for antecedent, false for succedent
     */
    private void translate(Semisequent ss, 
			   boolean isAntecedent) throws SimplifyException  {
	sb.append('(');
	if (isAntecedent) {
	    sb.append(DecisionProcedureSimplifyOp.AND);
	} else {
	    sb.append(DecisionProcedureSimplifyOp.OR);
	}
	for (int i = 0; i < ss.size(); ++i) {
	    sb.append(' ');
	    translate(ss.get(i));
	}
	if (isAntecedent) {
		sb.append(' ');
		sb.append(DecisionProcedureSimplifyOp.LIMIT_FACTS);
	}		
	sb.append(')');
    }

    /** Translates the given ConstrainedFormula into "Simplify" input
     * syntax and adds the resulting string to the StringBuffer sb. 
     * @param cf the ConstrainedFormula which should be written in 
     * Simplify syntax
     */
    private void translate(ConstrainedFormula cf) throws SimplifyException {
	if (constraintSet.used(cf)) {
	    SyntacticalReplaceVisitor srVisitor =
		new SyntacticalReplaceVisitor(new JavaInfo(), 
					      constraintSet.chosenConstraint());
	    cf.formula().execPostOrder ( srVisitor );
	    translate(srVisitor.getTerm());
	}
    }

    /** Translates the given term into "Simplify" input syntax and
     * adds the resulting string to the StringBuffer sb.
     *
     * @param term the Term which should be written in Simplify syntax
     */
    private void translate(Term term) throws SimplifyException {
	Operator op = term.op();

	if (op == Op.NOT) {
	    translateSimpleTerm(term, DecisionProcedureSimplifyOp.NOT);
	} else if (op == Op.AND) {
	    translateSimpleTerm(term, DecisionProcedureSimplifyOp.AND);
	} else if (op == Op.OR) {
	    translateSimpleTerm(term, DecisionProcedureSimplifyOp.OR);
	} else if (op == Op.IMP) {
	    translateSimpleTerm(term, DecisionProcedureSimplifyOp.IMP);
	} else if (op == Op.EQV) {
	    translateSimpleTerm(term, DecisionProcedureSimplifyOp.EQV);
	} else if (op == Op.EQUALS) {
	    translateSimpleTerm(term, DecisionProcedureSimplifyOp.EQUALS);
	} else if (op == Op.ALL || op == Op.EX) {
	    Debug.assertTrue(term.arity() == 1);
	    sb.append('(')
		.append(op == Op.ALL
			? DecisionProcedureSimplifyOp.ALL
			: DecisionProcedureSimplifyOp.EX)
		.append(" (");
	    ArrayOfQuantifiableVariable vars = term.varsBoundHere(0);
	    for (int i = 0; i < vars.size(); ++i) {
		if (i > 0) {
		    sb.append(' ');
		}
		translateVariable(vars.getQuantifiableVariable(i));
	    }
	    sb.append(") ");
	    translate(term.sub(0));
	    sb.append(')');
	} else if (op == Op.TRUE) {
	    sb.append(DecisionProcedureSimplifyOp.TRUE);
	} else if (op == Op.FALSE) {
	    sb.append(DecisionProcedureSimplifyOp.FALSE);
	} else if (op instanceof AttributeOp) {
	    translateSimpleTerm(term, getUniqueVariableName(op).toString());
	} else if (op instanceof LogicVariable || op instanceof ProgramVariable) {
	    translateVariable(op);
	} else if (op instanceof Metavariable) {
	    if (localMetavariables.contains((Metavariable)op)) {
		usedLocalMv.add(op);
	    } else {
		usedGlobalMv.add(op);
	    }
	    translateVariable(op);
	} else if (op instanceof Function) {
	    String name = op.name().toString().intern();
	    if (name.equals("+")) {
		translateSimpleTerm(term, DecisionProcedureSimplifyOp.PLUS);
	    } else if (name.equals("~d"))  {
		translateSimpleTerm(term, DecisionProcedureSimplifyOp.MINUS);
	    } else if (name.equals("~m")) {
		//Beware: This is not really hygienic
		translateUnaryMinus(term, DecisionProcedureSimplifyOp.MINUS);
	    } else if (name.equals("mul")) {
		translateSimpleTerm(term, DecisionProcedureSimplifyOp.MULT);
	    } else if (name.equals("div")) {
		translateSimpleTerm(term, getUniqueVariableName(op).toString());
		notes.append("Division encountered which is not supported by Simplify.")
		    .append("It is treated as an uninterpreted function.\n");
	    } else if (name.equals("lt")) {
		translateSimpleTerm(term, DecisionProcedureSimplifyOp.LT);
	    } else if (name.equals("gt")) {
		translateSimpleTerm(term, DecisionProcedureSimplifyOp.GT);
	    } else if (name.equals("leq")) {
		translateSimpleTerm(term, DecisionProcedureSimplifyOp.LEQ);
	    } else if (name.equals("geq")) {
		translateSimpleTerm(term, DecisionProcedureSimplifyOp.GEQ);
	    } else if (name.equals("Z") || 
		       name.equals("C")) {
		Debug.assertTrue(term.arity() == 1);
		sb.append(NumberTranslation.translate(term.sub(0)));
	    } else if (name.equals("byte_MIN") | name.equals("byte_MAX")
		       | name.equals("byte_RANGE") | name.equals("byte_HALFRANGE")
		       | name.equals("short_MIN") | name.equals("short_MAX")
		       | name.equals("short_RANGE") | name.equals("short_HALFRANGE")
		       | name.equals("int_MIN") | name.equals("int_MAX")
		       | name.equals("int_RANGE") | name.equals("int_HALFRANGE")
		       | name.equals("long_MIN") | name.equals("long_MAX")
		       | name.equals("long_RANGE") | name.equals("long_HALFRANGE") ){
		translateSimpleTerm(term, name);		
	    } else {
		translateSimpleTerm(term, getUniqueVariableName(op).toString());
		if (term.sort() == Sort.FORMULA) {
		    addPredicate(getUniqueVariableName(op).toString(), op.arity());
		}
	    }
	} else if ((op instanceof Modality) | 
		   (op instanceof SimultaneousUpdateOp )) {
	    uninterpretedTerm(term, true);
	} else {
	    System.err.println("Warning: unknown operator while translating "+
			       "into simplify syntax.");
	    System.err.println("I will quit here.");
	    System.out.println("sort=" + term.sort().name() 
			       + ", opName=" + term.op().name()
			       + ", opClass=" + term.op().getClass().getName());
	    sb.setLength(0);
	    throw new SimplifyException(op.name() + " not known by Simplify");
	}
    }

    /** 
     * Used to give a variable (program, logic, meta) a unique name.
     * @param op The variable to be translated/renamed.
     */
    private void translateVariable(Operator op) {
	sb.append(getUniqueVariableName(op));
    }

    /**
     * produces a unique name for the given Variable,
     *    enclosed in '|' and with a unique hashcode.
     * @param op The variable to get a new name.
     */
    private StringBuffer getUniqueVariableName(Operator op) {
	return new StringBuffer("|")
	    .append(op.name())
	    .append("_")
	    .append(getUniqueHashCode(op))
	    .append("|");
    }
    
    private void uninterpretedTerm(Term term, boolean modRenaming) {
	StringBuffer temp = new StringBuffer();
	temp.append('|')
	    .append(term.op().name())
	    .append('_');
	if (modRenaming) {
	    temp.append(getUniqueHashCodeForUninterpretedTerm(term));
	} else {
	    temp.append(getUniqueHashCode(term));
	}
	temp.append('|');
	sb.append('(').append(temp);
	IteratorOfQuantifiableVariable i;
	for ( i = term.freeVars().iterator(); i.hasNext(); ) {
	    sb.append(' ');
	    translateVariable(i.next());
	}
	sb.append(')');
	addPredicate(temp.toString(), term.freeVars().size());
    }

    /**
     * Takes a term and translates it with its argument in the Simplify syntax.
     * @param term The term to be converted.
     * @param name The name that should be used for the term (should be unique,
     *    it should be taken care of that somewhere else (if desired)).
     */
    private void translateSimpleTerm(Term term, 
				     String name) throws SimplifyException {
	sb.append('(').append(name);
	for (int i = 0; i < term.arity(); ++i) {
	    Debug.assertTrue(term.varsBoundHere(i).size() == 0);
	    sb.append(' ');
	    translate(term.sub(i));
	}
	sb.append(')');
    }

    /**
     * Translates the unary minus ~m into a "0 -" construct.
     * Could be solved better with a newly created term!!!
     * @param term The term to be converted.
     * @param name The name that should be used for the term (should be unique,
     *    it should be taken care of that somewhere else (if desired)).
     */
    private void translateUnaryMinus(Term term, 
				     String name) throws SimplifyException {
	sb.append('(').append(name).append(" 0");
	translate(term.sub(0));
	sb.append(')');
    }

    /**
     * Adds a predicate to the internal set and adds a line to declare the 
     *    predicate to the declarator stringbuffer.
     * @param name The name of the predicate (no KeY representation jas to exist).
     * @param arity The arity of the term.
     */
    private void addPredicate(String name, int arity) {
	if (!predicateSet.contains(name)) {
	    predicateSet.add(name);
	    predicate.append("(DEFPRED (").append(name);
	    for (int i = 0; i < arity; ++i) {
		predicate.append(" x").append(i);
	    }
	    predicate.append("))\n");
	}
    }

    /**
     * For some terms (AttributeOps) the order in KeY is different
     * than the order of the user or Simplify expects.
     *
     * @return the simplified version of the Term t in reversed order
     * @param t the Term which should be written in Simplify syntax,
     * but in reverse order compared to the internal order in KeY
     */
    private StringBuffer printlastfirst(Term t) {
	StringBuffer sb = new StringBuffer();
	if (t.op().arity() > 0) {
	    Debug.assertTrue(t.op().arity() == 1);
	    sb.append(printlastfirst(t.sub(0)));
	    //sb.append('.');
	}
	sb.append(t.op().name())
	    .append("\\|")
	    .append(getUniqueHashCode(t.op()));
	return sb;
    }


    /**
     * Returns a unique HashCode for the object qv.
     * Unique means unique for the goal given to the calling class.
     * This function does not depend on .hashcode() to deliver unique 
     * hash codes like the memory address of the object. It uses a 
     * hashMap and compares every new Object in O(n) (n number of 
     * Objects with the same .hashCode()) to all others.
     * @returns a unique hashcode for the variable gv.
     * @param v the Object the hashcode should be returned.
     */
    public int getUniqueHashCode(Object qv) {
	Integer number = (Integer)this.variableMap.get(qv);
	if (number == null) {
	    number = new Integer(this.variableMap.size());
	    this.variableMap.put(qv, number);
	}
	return number.intValue();
    }

    /**
     * Returns a unique HashCode for the term qv.
     * Unique means unique for the goal given to the calling class.
     * This function does not depend on .hashcode() to deliver 
     * unique hash codes like the memory address of the object. 
     * It uses a hashMap and compares
     * every new Object in O(n) (n number of Objects 
     * with the same .hashCode()) to all others.
     * It compares with .equalsModRenaming().
     * @returns a unique hashcode for the term gv.
     * @param term the Term the hashcode should be returned.
     */
    public int getUniqueHashCodeForUninterpretedTerm(Term term) {
	Integer number = (Integer)this.variableMap.get
	    (new UninterpretedTermWrapper(term));

	if (number == null) {
	    number = new Integer(this.variableMap.size());
	    this.variableMap.put
		(new UninterpretedTermWrapper(term), number);
	}

	return number.intValue();
    }

        
    /** The translation result is stored in this variable. */
    private final String text;

    /** StringBuffer is used during translation. */
    private final StringBuffer sb = new StringBuffer();

    /** StringBuffer contains all declared predicate symbols. */
    private final StringBuffer predicate = new StringBuffer();

    /** StringBuffer to store text which could be usefull for the user */
    private final StringBuffer notes = new StringBuffer();

    /** remember all printed predicates */
    private final Set predicateSet = new HashSet();

    /** The Constraint under which the sequent is to be proven*/
    private final ConstraintSet constraintSet;

    /** Some terms should need a unique HashCode (CVC doesn't handle quantors,
       ProgramVariables from different blocks could have the same name), 
       functions and user named skolemfunctions can get the same name.
       The difference to variables is that the have quantified variables
       so they have to be compared moduloRenaming.
       Comparing terms without variables with .equals() instead of 
       .equalsModRenaming() should give a speed improvement.
    */
    private final HashMap variableMap = new HashMap();

    /** 
	To handle local metavariabls we have to quantify them existentially. 
	We do not need to quantify already substituted metavariables.
    */
    private final HashSet usedLocalMv = new HashSet();
    private final HashSet usedGlobalMv = new HashSet();
    private final SetOfMetavariable localMetavariables;
}
