// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.proof.TacletIndex;
import de.uka.ilkd.key.proof.BuiltInRuleIndex;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.casetool.together.TgFacade;

/** 
 * an instance of this class describes the initial configuration of the prover.
 * This includes sorts, functions, heuristics, and variables namespaces, 
 * information on the underlying java model, and a set of rules.
 */
public class InitConfig {

    private NamespaceSet ns=new NamespaceSet();
    private Services serv = new Services();
    private SetOfTaclet taclets=SetAsListOfTaclet.EMPTY_SET;
    private ListOfBuiltInRule builtInRules = 
	SLListOfBuiltInRule.EMPTY_LIST;

    public InitConfig() {
	funcNS().add(Op.NULL);
	sortNS().add(Sort.NULL);
    }

    /** sets the NamespaceSet of this initial configuration
     */
    public void setNamespaces(NamespaceSet ns) {
	this.ns=ns;
    }

    /** sets the information of the underlying Java model of this 
     * initial configuration
     */
    public void setServices(Services serv) {
	this.serv=serv;
    }

    public Services getServices() {
	return serv;
    }

    /** sets the set of taclets of this initial configuration
     */
    public void setTaclets(SetOfTaclet taclets) {
	this.taclets=taclets;
    }

    /** 
     * sets the list of built in rules of this initial configuration
     */
    public void setBuiltInRules(ListOfBuiltInRule builtIn) {
	this.builtInRules = builtIn;
    }

    /** returns the namespaces of this initial configuration
     */
    public NamespaceSet namespaces() {
	return ns;
    }

    /**
     * returns the information on the Java model of this initial configuration
     * @see #getJavaInfo()
     */
    public JavaInfo javaInfo() {
	return serv.getJavaInfo();
    }
    /**
     * Gets the information of the underlying Java model of this 
     * initial configuration.
     * @see #javaInfo()
     * @see de.uka.ilkd.key.gui.KeYMediator#getJavaInfo()
     */
    public JavaInfo getJavaInfo() {
	return javaInfo();
    }

    /** returns the taclets of this initial configuration
     */
    public SetOfTaclet taclets() {
	return taclets;
    }

    /** returns the built-in rules of this initial configuration
     */
    public ListOfBuiltInRule builtInRules() {
	return builtInRules;
    }

    /** returns a newly created taclet index for the set of taclets contained
     * in this initial configuration
     */
    public TacletIndex createTacletIndex() {
	return new TacletIndex(taclets);
    }

    /**
     * returns a new created index for built in rules (at the moment immutable
     * list)
     */
    public BuiltInRuleIndex createBuiltInRuleIndex() {
	return new BuiltInRuleIndex(builtInRules);
    }

    /** returns the function namespace of this initial configuration
     */
    public Namespace funcNS() {
	return ns.functions();
    }

    /** returns the sort namespace of this initial configuration
     */
    public Namespace sortNS() {
	return ns.sorts();
    }

    /** returns the heuristics namespace of this initial configuration
     */
    public Namespace heurNS() {
	return ns.heuristics();
    }

    /** returns the variable namespace of this initial configuration
     */
    public Namespace varNS() {
	return ns.variables();
    }

    /** adds namespaces to the namespaces of this initial configuration
     */
    public void add(NamespaceSet ns, ModStrategy mod) {
	if (mod.modifyFunctions()) funcNS().add(ns.functions());
	if (mod.modifySorts()) sortNS().add(ns.sorts());
	if (mod.modifyVariables()) varNS().add(ns.variables());
	if (mod.modifyHeuristics()) heurNS().add(ns.heuristics());
    }

    /** adds taclets to the set of taclets of this init configuration
     */
    public void add(SetOfTaclet tacs) {
	taclets=taclets.union(tacs);
    }

    /** creates a facade for the OCL-DL-Translation using the information
     * stored in this initial configuration
     */
    public TgFacade createFacadeForTransl() {
	return new TgFacade(javaInfo(), ns);
    }

    /** returns a copy of this initial configuration copying the namespaces,
     * the contained JavaInfo while using the immutable set of taclets in the 
     * copy
     */
    public InitConfig copy() {
	InitConfig ic=new InitConfig();
	ic.setNamespaces(namespaces().copy());
	ic.setServices(serv.copy());
	ic.setTaclets(taclets());
	ic.setBuiltInRules(builtInRules());
	return ic;
    }

}
