// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.java.JavaInfo;
import java.io.*;
import de.uka.ilkd.key.gui.Main;
import java.util.Vector;
import de.uka.ilkd.key.java.Recoder2KeY;
import de.uka.ilkd.key.java.ConvertException;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Services;

/** represents an input from a .key problem file producing environment as
 * well as a proof obligation.
 */
public class KeYFile implements ProofOblInput, EnvInput {

    /** the RuleSource delivering the input stream for the file.
     */
    protected RuleSource file = null;
    
    protected InputStream input = null;

    /** the initial configuration to be used (and modified) while reading.
     */
    protected InitConfig initConfig = null;

    private String name;
    protected ProblemParser problemParser=null;

    /** the graphical entity to notify on the state of reading.
     */
    protected Main main;

    /** creates a new representation for a given file by indicating a name
     * and a file representing the physical source of the .key file.
     */
    public KeYFile(String name, File file) {
	this(name, RuleSource.initRuleFile(file), null);
    }

    /** creates a new representation for a given file by indicating a name
     * and a file representing the physical source of the .key file.
     * @param main the graphical entity to notify on the state of reading
     */
    public KeYFile(String name, File file, Main main) {
	this(name, RuleSource.initRuleFile(file), main);
    }

    /** creates a new representation for a given file by indicating a name
     * and a RuleSource representing the physical source of the .key file.
     * @param main the graphical entity to notify on the state of reading
     */
    public KeYFile(String name, RuleSource file, Main main) {
	this.file=file;
	this.name=name;
	this.main=main;
    }

    /** returns the problem term of the KeY file if already read. 
     * Returns null otherwise
     */
    public Proof[] getPO() {
	return new Proof[] {};
    }

    /** returns the initial configuration that is used to read and
     * that may be modified by reading.
     */
    public InitConfig getInitConfig() {
	return initConfig;
    }

    /** returns the number of chars in the .key file.
     */
    public int getNumberOfChars() {
	return file.getNumberOfChars();
    }
    
    protected InputStream getNewStream() {
        try {
            if (input != null) input.close();
        } catch(IOException ioe) {
            System.err.println("WARNING: Cannot close stream "+file+"\n"+ioe);
        }
        input = file.getNewStream();
        return input;
    }
    

    /** reads the whole .key file and modifies the initial configuration
     * assigned to this object according to the given modification strategy. 
     * Throws an exception if no initial configuration has been set yet.
     */
    public void read(ModStrategy mod) throws ProofInputException {
	if (initConfig==null) {
	    throw new IllegalStateException("KeYFile: InitConfig not set.");
	}
	try {
	    Services copy=initConfig.getServices().copy();
	    CountingBufferedInputStream cinp = 
		new CountingBufferedInputStream
		    (getNewStream(),main,getNumberOfChars()/100);
	    problemParser = new ProblemParser(new KeYLexer(cinp), 
					      copy.getJavaInfo(), 
					      initConfig.namespaces().copy()); 
            problemParser.problem(); 
	    initConfig.add(problemParser.namespaces(), mod);
	    if (mod.modifyTaclets()) initConfig.add(problemParser.getTaclets());
	    if (mod.modifyJavaInfo()) initConfig.setServices(copy);
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
	}
    }

    private DeclParser createDeclParser() {
	return new DeclParser(new KeYLexer(getNewStream()),
			      file.toString(), 
			      initConfig.varNS().copy(), 
			      initConfig.funcNS().copy(), 
			      initConfig.sortNS().copy(), 
			      initConfig.heurNS().copy(),
                              initConfig.javaInfo().copy());
    }

    /** reads the sorts declaration of the .key file only, 
     * modifying the sort namespace
     * of the initial configuration if allowed in the given 
     * modification strategy.
     */
    public void readSorts(ModStrategy mod) throws ProofInputException {
	try {
	    DeclParser p=createDeclParser();
	    p.parseSorts();
	    initConfig.add(p.namespaces(), mod);	    
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
	}
    }
    
    /** reads the functions and predicates declared in the .key file only, 
     * modifying the function namespace
     * of the initial configuration if allowed in the given 
     * modification strategy.
     */
    public void readFuncAndPred(ModStrategy mod) throws ProofInputException {
	try {
	    DeclParser p=createDeclParser();
	    p.parseFuncAndPred();
	    initConfig.add(p.namespaces(), mod);
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
	}
    }

   /** reads the rules and problems declared in the .key file only, 
     * modifying the set of rules 
     * of the initial configuration if allowed in the given 
     * modification strategy.
     */
    public void readRulesAndProblem(ModStrategy mod) throws ProofInputException {
	CountingBufferedInputStream cinp = 
	    new CountingBufferedInputStream
	    (getNewStream(),main,getNumberOfChars()/100);
	problemParser = new ProblemParser(new KeYLexer(cinp),
			      file.toString(), TermFactory.DEFAULT,
			      initConfig.javaInfo());
	try {
	    problemParser.parseProblemGivenNamespaces(initConfig.funcNS().copy(), 
						      initConfig.sortNS().copy(), 
						      initConfig.heurNS().copy());
	    initConfig.add(problemParser.namespaces(), mod);
	    if (mod.modifyTaclets()) initConfig.add(problemParser.getTaclets());
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
	}
    }


    /** defines the initial configuration that is to be used to read the .key
     * file and that the read declarations are added to. The set initial 
     * configuration is overwritten by a call to setConfig().
     */
    public void addTo(InitConfig conf) {
	this.initConfig=conf;
    }

    /** defines the initial configuration that is to be used to read the .key
     * file and that the read declarations are added to. The set initial 
     * configuration is overwritten by a call to addTo().
     */
    public void setConfig(InitConfig conf) {
	this.initConfig=conf;
    }

    /** reads a saved proof of a .key file
     */
    public void readProof(ProblemLoader prl) throws ProofInputException {
        try {
            problemParser.proof(prl);
        } catch(antlr.ANTLRException e) {
            throw new ProofInputException(e);
        }
    }

    /** returns the name of the .key file
     */
    public String name() {
	return name;
    }

    private Vector getClasses(String f) throws ProofInputException  {
	File cfile = new File(f);
	Vector v=new Vector();
	if (!cfile.isDirectory()) { // test relative pathname
	    File parent=file.file().getParentFile();
	    cfile=new File(parent,f);	
	}
	if (cfile.isDirectory()) {
	    String[] list=cfile.list();
	    for (int i=0; i<list.length; i++) {
		File n=new File(list[i]);
		if (n.isDirectory()) {		    
		    v.addAll(getClasses(list[i]));
		} else if (list[i].endsWith(".java")) {
		    v.add(cfile.getPath()+File.separator+n.getName());	
		}
	    }
	    return v;
	} else {
	    throw new ProofInputException("Java model path "+f+" not found.");
	}
	
    }

    private String[] getClassesInPath(String f) throws ProofInputException {
	Vector v=getClasses(f);
	String[] res=new String[v.size()];
	for (int i=0; i<res.length; i++) {
	    res[i]=(String)v.get(i);
	}
	return res;
    }

    public void readJava() throws ProofInputException {
	if (initConfig==null) {
	    throw new IllegalStateException("KeYFile: InitConfig not set.");
	}
	problemParser = new ProblemParser(new KeYLexer(getNewStream()),
					  file.toString(), TermFactory.DEFAULT,
					  initConfig.javaInfo());
	String javadir = null;
        try { 
	    javadir = problemParser.java(); 
	} catch (antlr.ANTLRException e) {
	    throw new ProofInputException(e);
        }
	if (javadir==null) return;
	Recoder2KeY r2k = new Recoder2KeY
	    (initConfig.javaInfo().getKeYProgModelInfo().getServConf(), 
	     null, "",  
	     initConfig.javaInfo().rec2key(), 
	     initConfig.javaInfo().getTypeConverter(),
	     initConfig.namespaces());
	String[] cus = getClassesInPath(javadir);
        try { 
	    r2k.readCompilationUnitsAsFiles(cus);
	} catch (ConvertException ce) {
	    throw new ProofInputException(ce);
	}
    }

    public void close() {
        try {
            if (input != null) { 
		input.close();
	    }
        } catch(IOException ioe) {
            System.err.println("WARNING: Cannot close stream "+file+"\n"+ioe);
        }
    }

    public String toString() {
	return name()+" "+file.toString();
    }
    


}
