// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.ldt.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.TypeConverter;
import de.uka.ilkd.key.gui.LDTSettings;
import de.uka.ilkd.key.util.Debug;
import java.io.*;
import de.uka.ilkd.key.java.abstraction.ListOfType;
import de.uka.ilkd.key.java.abstraction.SLListOfType;
import de.uka.ilkd.key.java.abstraction.IteratorOfType;
import de.uka.ilkd.key.java.abstraction.Type;
import de.uka.ilkd.key.java.abstraction.PrimitiveType;
import de.uka.ilkd.key.java.declaration.ClassDeclaration;
import de.uka.ilkd.key.gui.Main;

/** represents the LDT .key files as a whole. Reading the LDT input means 
 * reading parts of associated .key files in a special order, namely first
 * all sort declarations then all function and predicate declarations and
 * third the rules. This procedure makes it possible to use all declared
 * sorts in all rules, e.g.
 * 
 */
public class LDTInput implements EnvInput {

    private InitConfig initConfig = null;
    private String name = "language data types";
    private LDT[] ldts;
    private Main main;

    /** creates a representation of the LDT files to be used as input
     * to the KeY prover. 
     * @param settings the LDT settings to be used 
     * @param main the main class used to report the progress of reading
     */
    public LDTInput(LDTSettings settings, Main main) {
	ldts=new LDT[]{
	    settings.getBooleanLDT(),
	    settings.getNaturalNumberLDT(),  	   
 	    settings.getObjectLDT()};
	this.main=main;
    }

    /** sets the initial configuration that is used to add the read changes to.
     */
    public void addTo(InitConfig conf) {
	this.initConfig=conf;
    }

    /** returns to number of chars to be read by the LDT input.
     */
    public int getNumberOfChars() {
	int sum=0;
	for (int i=0; i<ldts.length; i++) {
	    sum=sum+ldts[i].getRuleSource().getNumberOfChars();
	}
	return sum;
    }

   /** returns the initial configuration that is used to read and
     * that may be modified by reading.
     */
    public InitConfig getInitConfig() {
	return initConfig;
    }

    /** reads all LDTs, i.e., all associated .key files with respect to 
     * the given modification strategy. Reading is done in a special order: first
     * all sort declarations then all function and predicate declarations and
     * third the rules. This procedure makes it possible to use all declared
     * sorts in all rules, e.g.
     */
    public void read(ModStrategy mod) throws ProofInputException {
	if (initConfig==null) {
	    throw new IllegalStateException("LDTInput: InitConfig not set.");
	}
	KeYFile[] keyFile=new KeYFile[ldts.length];
	for (int i=0; i<keyFile.length; i++) {
	    keyFile[i]=new KeYFile(ldts[i].name().toString(),
				   ldts[i].getRuleSource(), main);
	    keyFile[i].addTo(initConfig);
	}
	for (int i=0; i<keyFile.length; i++) {
	    keyFile[i].readSorts(mod);
	}
	for (int i=0; i<keyFile.length; i++) {
	    keyFile[i].readFuncAndPred(mod);
	}
	initLDTs();
	for (int i=0; i<keyFile.length; i++) {
	    keyFile[i].readRulesAndProblem(mod);
	}
	createLDTModels();
	setUpSorts ( initConfig );
    }


    /** returns a string telling that this is LDT input
     */
    public String name() {
	return name;
    }

    /** initializes the LDTs */
    private void initLDTs() {
	Sort sort=null;
	LDT integerLDT = null;
	LDT booleanLDT = null;		
	LDT objectLDT = null;		
	for (int i = 0; i < ldts.length; i++) {
	    ListOfType type = SLListOfType.EMPTY_LIST;	
	    if (ldts[i] instanceof NumberLDT) {
		type = type.prepend(PrimitiveType.JAVA_INT);
		type = type.prepend(PrimitiveType.JAVA_LONG);
		type = type.prepend(PrimitiveType.JAVA_CHAR);
		type = type.prepend(PrimitiveType.JAVA_BYTE);
		type = type.prepend(PrimitiveType.JAVA_SHORT);
		sort = (Sort)initConfig.sortNS().lookup(ldts[i].getSortName());
		integerLDT = ldts[i];
	    } else if (ldts[i] instanceof BooleanLDT) {
		type = type.prepend(PrimitiveType.JAVA_BOOLEAN);
		sort = (Sort)initConfig.sortNS().lookup(ldts[i].getSortName());
		booleanLDT = ldts[i];
	    } else if (ldts[i] instanceof ObjectLDT) {
		sort = (Sort)initConfig.sortNS().lookup(ldts[i].getSortName());
		type = type.prepend(ClassDeclaration.objectClassDecl);
		objectLDT = ldts[i];
	    } else {
		Debug.out("standardrules: javatype for LDT"
			  +" is not specified:",
			  ldts[i]);
		throw new RuntimeException
		    ("JavaType for LDT not specified.");
	    }		
	    ldts[i].init(initConfig.funcNS(), sort, type);
	}	
	initConfig.javaInfo().getTypeConverter()
	    .init(integerLDT, booleanLDT, objectLDT);
    }


    /** creates the language data type models */    
    private void createLDTModels() {
	TypeConverter converter = initConfig.javaInfo().getTypeConverter();
	LDT intLDT     = null;
	LDT booleanLDT = null;		
	LDT objectLDT = null;		
	LDT[] model    = ldts;
	Debug.out("probleminitializer: initialize LDT/Typeconverter");
	for (int i = 0; i < model.length; i++) {
	    initConfig.sortNS().add(model[i].targetSort());
	    add(initConfig.funcNS(), model[i].functions());
	    if (model[i] instanceof NumberLDT) {
		Debug.out("initialize: integerLDT");
		intLDT = model[i];
	    } else if (model[i] instanceof BooleanLDT) {
		Debug.out("initialize: booleanLDT");
		booleanLDT = model[i];
	    } else if (model[i] instanceof ObjectLDT) {
		Debug.out("initialize: objectLDT");
		objectLDT = model[i];
	    }
	}
	converter.init(intLDT, booleanLDT, objectLDT);
	Debug.out("probleminitializer: initialize TypeConverter end.");
    }

    /** returns the LDT models */
    public LDT[] ldtModels() {
	return ldts;
    }

    private void add(Namespace target, Namespace source) {
	IteratorOfNamed it=source.elements().iterator();
	while (it.hasNext()) {
	    target.add(it.next());
	}
	
    }

    /**
     * Delayed setup of symbols defined by sorts (e.g. functions for
     * collection sorts). This may not have been done for previously
     * defined sorts, as the integer sort was not available.
     */
    protected void setUpSorts ( InitConfig p_initConfig ) {
	NamespaceSet    nss     = new NamespaceSet ();
	Namespace       func_ns = nss.functions ();
	Namespace       sort_ns = nss.sorts ();
	IteratorOfNamed it      =
	    p_initConfig.sortNS ().allElements ().iterator ();
	Sort            sort;

	while ( it.hasNext () ) {
	    sort = (Sort)it.next ();
	    if ( sort instanceof SortDefiningSymbols )
		((SortDefiningSymbols)sort).createSymbols ( func_ns, 
							    p_initConfig.sortNS() );
	}

	p_initConfig.add ( nss, ModStrategy.MOD_ALL );
    }


}
