// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.Recoder2KeY;
import de.uka.ilkd.key.casetool.*;
import de.uka.ilkd.key.TogetherFacade;
import de.uka.ilkd.key.casetool.together.*;
import de.uka.ilkd.key.ocl.tudresden.TUDresdenParserAdapter;
import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.detailedocl.*;
import de.uka.ilkd.key.gui.ProofSettings;
import tudresden.ocl.*;

import java.io.*;

/** represents an OCL proof obligation input to the prover for proving the 
 * behavioral subtyping property.
 */
public class OCLBehavSubPO extends OCLProofOblInput {

    private ReprModelMethod methRepr;
    private String preText;
    private String postText;
    private String preTextParent;
    private String postTextParent;
    private String contextString;
    private String xmifile;
    
    private Term prePO  = null;
    private Term postPO = null;

    /** creates a new OCL proof obligation input representing a behavioral
     * subtyping property proof obligation using 
     * an XMI file name for reading the UML model from, a representation 
     * of the model class the PO is invoked from and five strings, (1)
     * describing the contextString, (2) describing the pre condition text
     * of the parent, (3) describing the pre condition text of the parent class,
     * (4) describing the post condtion string of the subclass, and (5)
     * describing the post condition of the parent class.
     */
    public OCLBehavSubPO(ReprModelMethod aMethRepr, String xmifile, 
			 String contextString, String preTextParent, 
			 String preText, String postTextParent, 
			 String postText) {
	super("BehavioralSubtyping "+aMethRepr.getName());
	this.methRepr=aMethRepr;
	this.xmifile=xmifile;
	this.contextString=contextString;
	this.preTextParent=preTextParent;
	this.preText=preText;
	this.postTextParent=postTextParent;
	this.postText=postText;
    }

    /** reads the Java model for this PO -- do we need it?
     */
    public void readJava() {
	if (initConfig==null) {
	    throw new IllegalStateException("InitConfig not set.");
	}
	Recoder2KeY r2k = new Recoder2KeY
	    (initConfig.javaInfo().getKeYProgModelInfo().getServConf(), 
	     null, "",  
	     initConfig.javaInfo().rec2key(), 
	     initConfig.javaInfo().getTypeConverter(),
	     initConfig.namespaces());
	String[] cus = methRepr.getContainingClass().getClassesInPackage();
	r2k.readCompilationUnitsAsFiles(cus);
    }


    /** reads this OCL proof obligation, thus invoking the OCL-DL-Translation
     */
    public void read(ModStrategy mod) throws ProofInputException {
	if (initConfig==null) {
	    throw new IllegalStateException("InitConfig not set.");
	}
	TUDresdenParserAdapter aAdapter1 = new TUDresdenParserAdapter();
	TUDresdenParserAdapter aAdapter2 = new TUDresdenParserAdapter();


	OclTree prePOAsTree 
	    = aAdapter1.doParse1
	    (contextString + " pre:",  xmifile, "context " 
	     + contextString + " pre: " + "((" + 
	     FunctionalityOnModel.normalizeConstraint(preTextParent) + ")" + 
	     " implies " + "(" 
	     + FunctionalityOnModel.normalizeConstraint(preText) + "))",
	     methRepr.getContainingPackage());
	
	
	OclTree postPOAsTree 
	    = aAdapter2.doParse1
	    (contextString + " post:",  xmifile, "context " 
	     + contextString + " pre: " + "(" + 
	     FunctionalityOnModel.normalizeConstraint(preTextParent) 
	     + " and " + 
	     FunctionalityOnModel.normalizeConstraint(preText) + ")" 
	     + " post: " + 
	     "((" + FunctionalityOnModel.normalizeConstraint(postText) + ")" 
	     + " implies " +
	     "(" + FunctionalityOnModel.normalizeConstraint(postTextParent) 
	     + "))",
	     methRepr.getContainingPackage());

        OCLTreeToDLAggregator aOCLTreeToDLAggregator;
	TogetherFacade po2prover = initConfig.createFacadeForTransl();
	OCLMod[] mods
	    =ProofSettings.THE_SETTINGS
	    .getOCLRefinementSettings().getOCLMods();
	aOCLTreeToDLAggregator =  getAggregator(po2prover, mods, methRepr);
	if ( !(aAdapter1.getErrorOccurred() || aAdapter2.getErrorOccurred()) ) {
	    prePO = aOCLTreeToDLAggregator.treeToTerm(prePOAsTree,
                                                      false, true);
	    postPO = aOCLTreeToDLAggregator.treeToTerm(postPOAsTree,
                                                       false, false);
	} else {
	    throw new ProofInputException(
               "Pre-/Post Conditions are syntactically incorrect");
	}
    }

    public Proof[] getPO() {
        if (prePO == null) throw new IllegalStateException("No prePO term");
        if (postPO == null) throw new IllegalStateException("No postPO term");
        String s = createProofHeader(methRepr.getContainingClass()
				     .getRootDirectory());
        return new Proof[] {
           new Proof("pre conditions", prePO, s,
                     initConfig.createTacletIndex(),
		     initConfig.createBuiltInRuleIndex()),
           new Proof("post conditions", postPO, s,
                     initConfig.createTacletIndex(),
		     initConfig.createBuiltInRuleIndex())
        };
    }

}
