// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//
package de.uka.ilkd.key.proof.init;

import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.Function;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.ocl.ocldltranslator.*;
import de.uka.ilkd.key.ocl.ocldltranslator.detailedocl.*;
import de.uka.ilkd.key.casetool.*;
import de.uka.ilkd.key.TogetherFacade;
import de.uka.ilkd.key.casetool.together.*;
import de.uka.ilkd.key.ocl.tudresden.TUDresdenParserAdapter;
import de.uka.ilkd.key.gui.ProofSettings;
import tudresden.ocl.*;
import java.io.*;

/** Represents the proof obligation input for the prover as an OCL 
 * constraint in textual format.
 */
public abstract class OCLProofOblInput implements ProofOblInput {

    /** The initial configuration for a prover to use for reading and modify*/
    protected InitConfig initConfig = null;

    private String name;

    /** Creates a new representation of an OCL input with a name describing
     * the kind of input.
     */
    public OCLProofOblInput(String name) {
	this.name=name;
    }

    /** returns an OCLTreeToDLAggregator for use in the OCL-DL-Translation.
     * @param mods Modifications to OCL constraints and OCL model used
     * to parse model extensions such as excThrown or null. If this parameter
     * has length 0 a usual OCLTODLAggregator is returned, otherwise a 
     * DetailedAggregator considering the modifications is returned.
     */
    protected OCLTreeToDLAggregator getAggregator(TogetherFacade facade, 
						  OCLMod[] mods,
						  ReprModelMethod aMethRepr) {
	if (mods.length==0) {
	    return new OCLTreeToDLAggregator(facade, aMethRepr);
	}
	return new DetailedAggregator(facade, mods, aMethRepr);
    }

    protected Term generateMethodPO(TogetherFacade po2prover,
                                         ReprModelMethod aMethRepr,
                                         String aTag,
                                         String preText, String postText,
                                         String xmifile) {
	return generateMethodPO(po2prover,
				ProofSettings.THE_SETTINGS
				.getOCLRefinementSettings().getOCLMods(),
				aMethRepr, aTag, preText, postText,
				xmifile);
    }
    /**
     * generate proof obligation as DL text from OCL formulas.
     * @param mods Modifications to OCL constraints and OCL model used
     * to parse model extensions, such as excThrown or null.
     * @param aTag controls the form of the generated text
     *           values for aTag: purepre, prepost, prepostaspl
     * purepre: only preconditions are considered.
     * prepost: is transformed into a DL-form.
     * prepostaspl transformation into PL form where diamond is omitted
     * and @pre constructs in the postconditions are eliminated.
     * @return a proof obligation term
     *  <code>preText -&gt; &lt;m&gt; postText</code>
     * @todo document parameters and check documentation of the effects of different values for aTag.
     */
    protected Term generateMethodPO(TogetherFacade po2prover,
				         OCLMod[] mods,
                                         ReprModelMethod aMethRepr,
                                         String aTag,
                                         String preText, String postText,
                                         String xmifile) {
	Debug.out("generate Method PO");
	String returnvalue = "";
	String contextString = aMethRepr.getContainingClassName() + "::" + 
           aMethRepr.getCallSignature();

	TUDresdenParserAdapter aAdapter = new TUDresdenParserAdapter();
	OclTree methPOAsTree;

	
 	// generating proofobligation: preText -> <m> postText
	methPOAsTree= aAdapter.doParse1
	    (contextString + aTag,  xmifile, 
	     "context " + contextString + 
	     " pre: " + FunctionalityOnModel.normalizeConstraint(preText) +
	     " post: " + FunctionalityOnModel.normalizeConstraint(postText), 
	     aMethRepr.getContainingPackage());
	
        OCLTreeToDLAggregator aOCLTreeToDLAggregator;
        try {
	    aOCLTreeToDLAggregator =  getAggregator(po2prover, mods, aMethRepr);
	    if (methPOAsTree!=null) {
		Term methProofObl;
                // only preconditions are considered
		if (aTag.equals("purepre")){
		    methProofObl 
			= aOCLTreeToDLAggregator.treeToTerm(methPOAsTree, 
							    false, true);
		    // is transformed into a DL-form
		} else if ( aTag.equals("prepost")){
		    methProofObl 
			= aOCLTreeToDLAggregator.treeToTerm(methPOAsTree, 
							    true, false);
		    // transformation into PL form where diamond is omitted
		    // and @pre constructs in the postconditions are eliminated
		} else if ( aTag.equals("prepostaspl")){
		    methProofObl 
			= aOCLTreeToDLAggregator.treeToTerm(methPOAsTree, 
							    false, false);
		} else { 
		    throw new RuntimeException("unexpected tag in "
					       +"generatedMethodPO: " + aTag);
		}
            	return methProofObl;
	    } else {
		throw new RuntimeException("generateMethodPO() failed");
	    }
	} catch(Exception e){
	    System.out.println("KeY-Error: "+ e);
	    e.printStackTrace();
            throw new RuntimeException("Generation of method proof "
				       +"obligation failed: "+e.getMessage());
	}
    }

    /** returns the problems as generated by reading the OCL constraint
     * and translating it to DL formulas.
     */
    public abstract Proof[] getPO();
    
    
    /** creates declarations necessary to save/load proof in textual form */
    String createProofHeader(String javaPath) {
        String s;
        s = "java {\""+javaPath+"\"}\n\n";
        s += "sorts {\n"; // create declaration header for the proof
        IteratorOfNamed it;
        it = initConfig.sortNS().getProtocolled();
        while (it.hasNext()) {
	    String n=it.next().toString();
	    int i;
	    if ((i=n.indexOf("."))<0 || 
		initConfig.sortNS().lookup(new Name(n.substring(0,i)))==null) {
		//the line above is for inner classes.
		//KeY does not want to handle them anyway...
		s = s+"object "+n+";\n";
	    }
	}

        s+="}\n\nprogram variables {\n";
        it = initConfig.javaInfo().programVariables().getProtocolled();
        while (it.hasNext()) 
           s = s+((ProgramVariable)(it.next())).proofToString();

        s+="}\n\nfunctions {\n";
        it = initConfig.funcNS().getProtocolled();
        while (it.hasNext()) s = s+((Function)(it.next())).proofToString();

        s+="}\n\n";
        return s;
    }


    /** returns the initial configuration that is used to read the OCL
     * input and that is used to be modified during reading.
     */
    public InitConfig getInitConfig() {
	return initConfig;
    }

    /** reads the OCL constraint and modifies the initial prover configuration
     * according to the given modification strategy.
     */
    public abstract void read(ModStrategy mod) throws ProofInputException;

    /** sets the initial configuration that is used to read the OCL
     * input and that is used to be modified during reading.
     */
    public void setConfig(InitConfig conf) {
	this.initConfig=conf;
        initConfig.sortNS().startProtocol();
        initConfig.funcNS().startProtocol();
        initConfig.javaInfo().programVariables().startProtocol();
    }

    /** returns the name of the OCL proof obligation input.
     */
    public String name() {
	return name;
    }

}
