// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.proof.init;

import java.io.*;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.parser.*;
import de.uka.ilkd.key.gui.ProofSettings;
import de.uka.ilkd.key.gui.Main;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.ldt.*;
import de.uka.ilkd.key.java.TypeConverter;
import de.uka.ilkd.key.gui.PresentationFeatures;
import de.uka.ilkd.key.logic.*;

public class ProblemInitializer {

    private Main main = null; // for status reporting -- optional

    private ProofOblInput problem = null;
    private EnvInput[] modelIndepInputs=null;

    private static final RuleSource PROP_RULES = 
       RuleSource.initRuleFile("propRule.key"); 

    private static final RuleSource GENERIC_RULES = 
       RuleSource.initRuleFile("genericRules.key");

    private static final RuleSource JAVA_RULES = 
       RuleSource.initRuleFile("javaRules.key"); 

    private static final ListOfBuiltInRule DEFAULT_BUILT_IN_RULES;

    static{
	// update simplifier
	ListOfBuiltInRule list = SLListOfBuiltInRule.EMPTY_LIST.
	    prepend(new UpdateSimplificationRule());
	// check whether "Simplify" is in $PATH. If not,  the 
	// taclet "Run Decision Procedure" won't be available.
	try {
	    
	    File file = File.createTempFile("key-simplify", null);
	    PrintWriter out = new PrintWriter(new FileWriter(file));
	    out.println("TRUE");
	    out.close();
	    Process p = Runtime.getRuntime().exec
		(new String[]{ "Simplify", file.getPath() });
	    list = list.prepend(new SimplifyIntegerRule());
	    file.delete();
	}
	catch (IOException e){} 
	finally {
	    DEFAULT_BUILT_IN_RULES = list;
	}
    }

    public static InitConfig BASE_CONFIG=null;

    private InitConfig initConfig=null;

    private LDTInput ldtIn;

    private SimultaneousUpdateSimplifier simplifier;
    
//    public ProblemInitializer(ProofOblInput problemFile, ProofSettings settings,
//                              Main main) {

    public ProblemInitializer(ProofSettings settings, Main main) 
                                                throws ProofInputException {
       this.main = main;
       modelIndepInputs = new EnvInput[] {
           new LDTInput(settings.getLDTSettings(), main), 
	   new KeYFile("propositional taclets", PROP_RULES, main),
	   new KeYFile("Java taclets", JAVA_RULES, main),
	   new KeYFile("generic taclets", GENERIC_RULES, main)
       };
       ldtIn=(LDTInput)modelIndepInputs[0];
       simplifier = settings.
	   getSimultaneousUpdateSimplifierSettings().getSimplifier();
       if (BASE_CONFIG == null) {
	   initConfig = new InitConfig();
	   initConfig.setBuiltInRules(DEFAULT_BUILT_IN_RULES);
	   readEnvInputs(modelIndepInputs);
           BASE_CONFIG = initConfig;
       }
    }

    //for tests only
    private ProblemInitializer(ProofOblInput problemFile)
                                               throws ProofInputException {
	this.problem=problemFile;
	modelIndepInputs = new EnvInput[] {
	    //            new LDTInput(new de.uka.ilkd.key.gui.LDTSettings(), null)
            new LDTInput(ProofSettings.THE_SETTINGS.getLDTSettings(), null)
        };
	if (BASE_CONFIG==null) {
	   initConfig = new InitConfig();
	   initConfig.setBuiltInRules(DEFAULT_BUILT_IN_RULES);	   
	   readEnvInputs(modelIndepInputs);
           BASE_CONFIG = initConfig;
	}
    }
    
    //for tests only
    public static ProblemInitializer problemInitializerForTests(
	                                 ProofOblInput problemFile)
                                            throws ProofInputException {
	return new ProblemInitializer(problemFile);
    }


    public InitConfig initConfig() {
	return initConfig;
    }


    private void readEnvInputs(EnvInput[] in) throws ProofInputException {
	for (int i=0; i<in.length; i++) {
	    if (main != null) {
		int chars = in[i].getNumberOfChars();
		main.setStatusLine("Reading "+in[i].name(), chars);
	    }
	    in[i].addTo(initConfig);
	    in[i].read(ModStrategy.NO_VARS);
	}
    }
       

    // called by setUp<...>
    public void initializeJavaModel() throws ProofInputException {
	if (main != null) {
	    main.setStatusLine("Reading Java model");
	}
	problem.readJava();
    }

    public Proof[] initializeProblem() throws ProofInputException {
	if (main != null) 
           main.setStatusLine("Loading problem \""+problem.name()+"\"");
	problem.read(ModStrategy.MOD_ALL);
	return problem.getPO();
    }

    public Proof[] setUpTest()  throws ProofInputException {
        initConfig = BASE_CONFIG.copy();
	problem.setConfig(initConfig);
	initializeJavaModel();
	return initializeProblem();
    }

    public void setUpProof(ProofOblInput problemFile) 
                                        throws ProofInputException {
        this.problem = problemFile;
        initConfig = BASE_CONFIG.copy();
	problem.setConfig(initConfig);
	initializeJavaModel();
	Proof[] pos = initializeProblem();
	InitConfig pi=initConfig();
	for (int i=0; i < pos.length; i++) {
	    pos[i].setSimplifier(simplifier);
            pos[i].setNamespaces(pi.namespaces());
            pos[i].setJavaServices(pi.getServices());
	    main.addProblem(pos[i]);
	}	
	PresentationFeatures.initialize(main.mediator().func_ns(), //XXXX: baeh
					main.mediator().getNotationInfo(),
					main.mediator().getSelectedProof());
    }

    public void tryReadProof(ProblemLoader prl) throws ProofInputException {
	if (problem instanceof KeYFile) {
	    ((KeYFile)problem).readProof(prl);
	}
    }

    public LDT[] ldtModels() {
	return ldtIn.ldtModels();
    }

    private void add(Namespace target, Namespace source) {
	IteratorOfNamed it=source.elements().iterator();
	while (it.hasNext()) {
	    target.add(it.next());
	}
	
    }

}
