// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.op.MapFromSchemaVariableToTacletPrefix;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.inst.SVInstantiations;

/** 
 * An AntecTaclet represents a taclet whose find part has to match a top level
 * formula in the antecedent of the sequent. 
 */
public class AntecTaclet extends FindTaclet{

     /**
     * creates a Schematic Theory Specific Rule (Taclet) with the given
     * parameters.  
     * @param name the name of the Taclet 
     * @param applPart contains the application part of an Taclet that is
     * the if-sequent, the variable conditions
     * @param goalTemplates a list of goal descriptions.
     * @param heuristics a list of heurisics for the Taclet
     * @param constraint the Constraint under which the Taclet is valid
     * @param attrs attributes for the Taclet; these are boolean values
     * indicating a noninteractive or recursive use of the Taclet. 
     * @param find the find term of the Taclet
     * @param prefixMap a MapFromSchemaVariableToTacletPrefix that contains the
     * prefix for each SchemaVariable in the Taclet
     */
    public AntecTaclet(Name name, TacletApplPart applPart,  
		     ListOfTacletGoalTemplate goalTemplates, 
		     ListOfHeuristic heuristics,
		     Constraint constraint,
		     TacletAttributes attrs,
		     Term find, MapFromSchemaVariableToTacletPrefix prefixMap){
	super(name, applPart, goalTemplates, heuristics, constraint, 
	      attrs, find, prefixMap);
	cacheMatchInfo();
    }

    
    /** looks if a schema variable occurs in the ifseq or
     * in the sequent of the goal templates(@see
     * de.uka.ilkd.key.rule.TacletGoalTemplate#seq() seq).
     * (this is ifseq, addexp and rwexp)
     * used by lookupInRule)
     * @param var the SchemaVariable to be looked for
     * @return true iff the variable has been found
     */
    protected boolean lookupInIfAddRwExp(SchemaVariable var) { 
	if (!ifSequent().lookup(var)) {
	    IteratorOfTacletGoalTemplate it = goalTemplates().iterator();
	    TacletGoalTemplate tpl;
	    while (it.hasNext()) {
		tpl=it.next();		
		if (tpl.sequent().lookup(var) ||
		    (tpl instanceof AntecSuccTacletGoalTemplate) && 
		     ((AntecSuccTacletGoalTemplate)tpl).replaceWith()
		    .lookup(var)) {
		    return true;
		}
	    }   
	    return false;	    
	}
	return true;	    
    }

    /** this method is used to determine if top level updates are
     * allowed to be ignored. This is the case if we have an Antec or
     * SuccTaclet but not for a RewriteTaclet
     * @return true if top level updates shall be ignored 
     */
    protected boolean ignoreTopLevelUpdates() {
	return true;
    }


    /** CONSTRAINT NOT USED 
     * applies the replacewith part of Taclets
     * @param gt TacletGoalTemplate used to get the replaceexpression in the Taclet
     * @param goal the Goal where the rule is applied
     * @param posOfFind the PosInOccurrence belonging to the find expression
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst a SVInstantiations object containing the necessary
     * instantiations, so that the find term matches 
     */
    protected void applyReplacewith(TacletGoalTemplate gt, Goal goal,
				    PosInOccurrence posOfFind,
				    JavaInfo javaInfo, 
				    MatchConditions matchCond) {
	if (gt instanceof AntecSuccTacletGoalTemplate) {
	    Semisequent ante  = goal.sequent().antecedent();
	    Semisequent succ  = goal.sequent().succedent();
	    Sequent replWith = ((AntecSuccTacletGoalTemplate)gt).replaceWith();

	    replaceAtPos(replWith.antecedent(), goal, 
			 posOfFind, javaInfo, matchCond);	   	    	    
	    addToSucc(replWith.succedent(), goal, 
		      null, javaInfo, matchCond);	   	    	    
	    
	
	} else {
	    // Then there was no replacewith...
	}
    }

    /**
     * adds the sequent of the add part of the Taclet to the goal sequent
     * @param add the Sequent to be added
     * @param goal the Goal to be updated
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param posOfFind the PosInOccurrence describes the place where to add
     * the semisequent 
     * @param svInst the SVInstantiations to be used to instantiate the Taclet
     */
    protected void applyAdd(Sequent add, Goal goal,
			    PosInOccurrence posOfFind,
			    JavaInfo javaInfo,
			    MatchConditions matchCond) {
	addToAntec(add.antecedent(), goal, posOfFind, javaInfo, matchCond);
	addToSucc(add.succedent(), goal, null, javaInfo, matchCond);
    }
        
    /** toString for the find part */
    String toStringFind() {
	return "find " +find()+"==> ";
    }

    protected Taclet setName(String s) {
	AntecTacletBuilder b=new AntecTacletBuilder();
	b.setFind(find());
	return super.setName(s, b);
    }


}
