// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.java.Statement;

import de.uka.ilkd.key.java.*; 
import de.uka.ilkd.key.util.ExtList;
import java.io.IOException;
import de.uka.ilkd.key.java.visitor.Visitor;

/** In the DL-formulae description of Taclets the program part can have
 * the following form < pi alpha;...; omega > Phi where pi is a prefix
 * consisting of open brackets, try's and so on and omega is the rest
 * of the program. Between the prefix pi and the postfix omega there
 * can stand an arbitrary program. This pattern is realized using this
 * class.
 */

public class ContextTerm extends JavaNonTerminalProgramElement {

    /** contains the middle part of the context term */    
    private final StatementBlock body;

    /** is true if the context terms has a prefix and postfix */
    private boolean hasPrefix;   

    /** creates a ContextTerm 
     * @param body an Extlist  representing the middle part of
     * the ContextTerm
     * @param hasPrefix a boolean indicating if the ContextTerm
     * has a pi and an omega
     */
    public ContextTerm(ExtList children) {
	super(children);
	body = new StatementBlock(children); 
    }


    /** returns true if the prefix of the ContextTerm is not empty
     * this means if one exists
     * @return true if the prefix of the ContextTerm is not empty
     * this means if one exists
     */
    public boolean hasPrefix() {
	return hasPrefix;
    }

    /** returns the middle part( the body) of the context term
     * @return the body of the context term
     */
    public StatementBlock getBody() {	
	return body;
    }


    public int getChildPositionCode(ProgramElement child) {
        // role 0 (IDX): statements
        if (body != null) {
	    if (child == body) {
		return 0;
	    }
        }
        return -1;
    }

    /**
     * Returns the number of children of this node.
     * @return an int giving the number of children of this node
    */

    public int getChildCount() {
        return (body != null) ? 1 : 0;
    }

    /**
     * Returns the child at the specified index in this node's "virtual"
     * child array
     * @param index an index into this node's "virtual" child array
     * @return the program element at the given position
     * @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     * of bounds
     */
    public ProgramElement getChildAt(int index) {
	if (index == 0) {
	    return body;
	}
	throw new ArrayIndexOutOfBoundsException();
    }


    /**
     *      Get the number of statements in this container.
     *      @return the number of statements.
     */

    public int getStatementCount() {
        return (body != null) ? 1 : 0;
    }

    /*
      Return the statement at the specified index in this node's
      "virtual" statement array.
      @param index an index for a statement.
      @return the statement with the given index.
      @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
      of bounds.
    */

    public Statement getStatementAt(int index) {
	return body;
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnContextTerm(this);
    }

    public void prettyPrint(PrettyPrinter w) throws IOException {
	w.printContextTerm(this);
    }
    
    /* toString */
    public String toString() {
	StringBuffer result = new StringBuffer();
	result.append((hasPrefix ? ".." : ""));
	result.append(body.toString());
	result.append("\n");
	result.append((hasPrefix ? "..." : ""));
	return result.toString();
    }
}
