// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.logic.op.SetOfSchemaVariable;
import de.uka.ilkd.key.logic.op.MapFromSchemaVariableToTacletPrefix;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.Constraint;
import de.uka.ilkd.key.logic.Term;
import de.uka.ilkd.key.logic.Sequent;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.logic.TermFactory;

/** 
 * Used to implement a Taclet that has no <I>find</I> part. This kind of taclet
 * is not attached to term or formula, but to a complete sequent. A typical
 * representant is the <code>cut</code> rule. 
 */
public class NoFindTaclet extends Taclet {

    /**
     * creates a Schematic Theory Specific Rule (Taclet) with the given
     * parameters.  
     * @param name the name of the Taclet 
     * @param applPart contains the application part of an Taclet that is
     * the if-sequent, the variable conditions
     * @param goals a list of goal descriptions.
     * @param heuristics a list of heurisics for the Taclet
     * @param constraint the Constraint under which the Taclet is valid
     * @param attrs attributes for the Taclet; these are boolean values
     * @param prefixMap a MapFromSchemaVariableToTacletPrefix that contains the
     * prefix for each SchemaVariable in the Taclet
     */
    public NoFindTaclet(Name name, TacletApplPart applPart,  
		      ListOfTacletGoalTemplate goalTemplates, 
		      ListOfHeuristic heuristics, 
		      Constraint constraint,
		      TacletAttributes attrs,
		      MapFromSchemaVariableToTacletPrefix prefixMap){
	super(name, applPart, goalTemplates, heuristics, constraint, attrs, 
	      prefixMap);
	cacheMatchInfo();
    } 

     
    /** checks if a given schema variables is used in the rule; this implies
     * in the ifseq, rwexp, addseq or addrule
     * @param var the SchemaVariable to be looked for
     * @return true iff Variable var occurs in the Taclet
     */
    protected boolean lookup(SchemaVariable var) {
	return lookupImpl(var);
    }

    /** looks if a schema variable occurs in the ifseq or
     * in the sequent of the goal templates(@see
     * de.uka.ilkd.key.rule.TacletGoalTemplate#seq() seq).
     * (this is ifseq, addexp (rwexp does not exist here)
     * used by lookup)
     * @param var the SchemaVariable to be looked for
     * @return true iff the variable has been found
     */
    protected boolean lookupInIfAddRwExp(SchemaVariable var) { 
	if (!ifSequent().lookup(var)) {
	    IteratorOfTacletGoalTemplate it=goalTemplates().iterator();
	    TacletGoalTemplate tpl;
	    while (it.hasNext()) {
		tpl=it.next();		
		if (tpl.sequent().lookup(var)) {		    
		    return true;
		}
	    } 		   
	    return false;	    
	}
	return true;	    
    }


    /** checks if a schema variable occurrs maximal in one instantiation area
     * @param var the SchemaVariable whose instantiation is checked
     * @return true iff instantiation is correct
     */
    protected boolean correctInstantiationAreaOccurrence(SchemaVariable var) {
	// here for findexp == sequent, overwritten in RewriteTaclet
	// ifseq, findexp, all addseq and all rwexp are one area
	Boolean correct=areaCache.get(var);
	if (correct!=null) {
	    return correct.booleanValue();
	}
	
	boolean foundInOneArea=false;
	
	if (lookupInAreaOne(var)) {
	    foundInOneArea=true;
	}

	if  (!lookupInAddRulesArea(var,foundInOneArea)) {
		areaCache.put(var, new Boolean(true));
		return true;
	}

	areaCache.put(var, new Boolean(false));	    
	return false;
    }


    /**
     * adds the sequent of the add part of the Taclet to the goal sequent
     * @param add the Sequent to be added
     * @param goal the Goal to be updated
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst the SVInstantiations object with the neccessary 
     * instantiations information
     */
    protected void applyAdd(Sequent add, Goal goal, 
			    JavaInfo javaInfo,
			    MatchConditions matchCond) {
	addToAntec(add.antecedent(), goal, null, javaInfo, matchCond);
	addToSucc (add.succedent(),  goal, null, javaInfo, matchCond);
    }
    

    /**
     * the rule is applied on the given goal using the
     * information of rule application. 
     * @param goal the goal that the rule application should refer to.
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param ruleApp the taclet application that is executed.
     * @param proveIfSequent a boolean that is true if a goal has to
     * be created in order to prove that the sequent in the if part of
     * the Taclet can be derived and therefore the application of the
     * rule is correct (normally true when the if-sequent does not match)
     */
    public ListOfGoal apply(Goal     goal,
			    JavaInfo javaInfo, 
			    RuleApp  ruleApp) {

	// Number without the if-goal eventually needed
	int                          numberOfNewGoals = goalTemplates().size();

	TacletApp                    tacletApp        = (TacletApp) ruleApp;
	MatchConditions              mc               = tacletApp.matchConditions ();

	// Restrict introduced metavariables to the subtree
	setRestrictedMetavariables ( goal, mc );

	ListOfGoal                   newGoals         =
	    checkIfGoals ( goal,
			   tacletApp.ifFormulaInstantiations (),
			   mc,
			   numberOfNewGoals );
	
	IteratorOfTacletGoalTemplate it               = goalTemplates().iterator();
	IteratorOfGoal               goalIt           = newGoals.iterator();

	while (it.hasNext()) {
	    TacletGoalTemplate gt          = it    .next();
	    Goal               currentGoal = goalIt.next();
	    // add first because we want to use pos information that
	    // is lost applying replacewith
	    applyAdd(         gt.sequent(),
			      currentGoal,
			      javaInfo,
			      mc);

	    applyAddrule(     gt.rules(),
			      currentGoal,
			      javaInfo,
			      mc);

	    applyAddProgVars( gt.addedProgVars(),
			      currentGoal, 
			      mc);
	}
       
	return newGoals;
    }

    protected Taclet setName(String s) {
	NoFindTacletBuilder b=new NoFindTacletBuilder();
	return super.setName(s, b);
    }

    /** checks if term and rule fit in some particular conditions 
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     *
     * PRECONDITION: matchCond.getConstraint ().isSatisfiable ()
     */
    public MatchResult checkInteractive(ListOfHeuristic heuristics,
					MatchConditions matchCond,
					JavaInfo        javaInfo,
					Constraint      userConstraint) {
	if (noninteractive()) {
	    IteratorOfHeuristic heurIt=heuristics(); 	    
	    while (heurIt.hasNext()) {
		if (heuristics.contains(heurIt.next())) {
		    return MatchResult.FAILED;
		}
	    }
	} 	
	
	return new MatchResult(matchCond);
    }

    /** checks if term and rule fit in some particular conditions 
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     *
     * PRECONDITION: matchCond.getConstraint ().isSatisfiable ()
     */
    public MatchResult checkHeuristic(ListOfHeuristic heuristics,
				      MatchConditions matchCond,
				      JavaInfo        javaInfo,
				      Constraint      userConstraint) {
	IteratorOfHeuristic heurIt=heuristics(); 	    
	while (heurIt.hasNext()) {
	    if (heuristics.contains(heurIt.next())) {
		return new MatchResult(matchCond);
	    }
	}	 	
	return MatchResult.FAILED;
    }


    /**
     * @return Set of schemavariables of the if and the (optional)
     * find part
     */
    public SetOfSchemaVariable getIfFindVariables () {
	return getIfVariables ();
    }

}
