// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.inst.SVInstantiations;

/** 
 *  A RewriteTaclet represents a taclet, whose find can be matched against any
 * term in the sequent no matter where it occurs. The only constraint to be
 * fulfilled is that the term matches the structure described by the term of the
 * find-part.
 */
public class RewriteTaclet extends FindTaclet{

    /** If true: The updates above the find position and the updates
     * in front of the if formulas have to be equivalent; there may
     * not be any program modalities above the find position */
    protected boolean sameUpdatePrefix;

    /**
     * creates a Schematic Theory Specific Rule (Taclet) with the given
     * parameters that represents a rewrite rule.  
     * @param name the name of the Taclet 
     * @param applPart contains the application part of an Taclet that is
     * the if-sequent, the variable conditions
     * @param goalTemplates a list of goal descriptions.
     * @param heuristics a list of heurisics for the Taclet
     * @param constraint the Constraint under which the Taclet is valid
     * @param attrs attributes for the Taclet; these are boolean values
     * indicating a noninteractive or recursive use of the Taclet. 
     * @param find the find term of the Taclet
     * @param prefixMap a MapFromSchemaVariableToTacletPrefix that contains the
     * prefix for each SchemaVariable in the Taclet
     * @param replaceWith a boolean indicating if one of the goal descriptions
     * is a replacewith (required for location check)
     */
    public RewriteTaclet(Name name, TacletApplPart applPart,  
			 ListOfTacletGoalTemplate  goalTemplates, 
			 ListOfHeuristic           heuristics,
			 Constraint                constraint,
			 TacletAttributes          attrs,
			 Term                      find, 
			 MapFromSchemaVariableToTacletPrefix prefixMap,
			 boolean                   p_sameUpdatePrefix){
	super(name, applPart, goalTemplates, heuristics, constraint,
	      attrs, find, prefixMap);
	sameUpdatePrefix = p_sameUpdatePrefix;
	
	cacheMatchInfo();
    }	

      
    /** 
     * this method is used to determine if top level updates are
     * allowed to be ignored. This is the case if we have an Antec or
     * SuccTaclet but not for a RewriteTaclet
     * @return true if top level updates shall be ignored 
     */
    protected boolean ignoreTopLevelUpdates() {
	return false;
    }
    
    /**
     * @return true iff: The updates above the find position and the
     * updates in front of the if formulas have to be equivalent;
     * there may not be any program modalities above the find position
     */
    public boolean getSameUpdatePrefix () {
	return sameUpdatePrefix;
    }

    /**
     * For taclets with <code>getSameUpdatePrefix ()</code>, collect
     * the updates above <code>p_pos</code> and add them to the update
     * context of the instantiations object <code>p_mc</code>.
     * @return the new instantiations with the additional updates, or
     * <code>null</code>, if program modalities appear above
     * <code>p_pos</code>
     */
    public MatchConditions checkUpdatePrefix
	( PosInOccurrence p_pos,
	  MatchConditions p_mc,
	  JavaInfo        p_javaInfo,
	  Constraint      p_userConstraint ) {
	if ( !getSameUpdatePrefix () )
	    return p_mc;
	SVInstantiations svi = p_mc.getInstantiations ();
	if ( p_pos.posInTerm () != null ) {
	    svi = collectUpdates ( p_pos.constrainedFormula ().formula (),
				   p_pos.posInTerm (),
				   svi );
	    if ( svi != null && p_pos.termBelowMetavariable () != null )
		svi = collectUpdates ( p_pos.termBelowMetavariable (),
				       p_pos.posInTermBelowMetavariable (),
				       svi );
	}

	if ( svi == null )
	    return null;
	return p_mc.setInstantiations ( svi );
    }

    /**
     * Collect the updates above the given position <code>p_pit</code>
     * within <code>p_term</code>
     */
    private SVInstantiations collectUpdates ( Term             p_term,
					      PosInTerm        p_pit,
					      SVInstantiations p_svi ) {
	Term subTerm = p_term;
	IteratorOfInteger posIt = p_pit.iterator();
	while(posIt.hasNext()) {
	    int subTermPos = posIt.next().intValue();
	    if ( subTerm.op () instanceof SimultaneousUpdateOp )
		p_svi = p_svi.addUpdate ( subTerm );
	    else if ( subTerm.op () instanceof Modality )
		return null;
	    subTerm = subTerm.sub(subTermPos);
	}
	return p_svi;
    }

    /** 
     * looks if a schema variable occurs in the ifseq or
     * in the sequent of the goal templates(@see
     * de.uka.ilkd.key.rule.TacletGoalTemplate#seq() seq).
     * (this is ifseq, addexp and rwexp)
     * used by lookup
     * @param var the SchemaVariable to be looked for
     * @return true iff the variable has been found
     */
    protected boolean lookupInIfAddRwExp(SchemaVariable var) { 
	if (!ifSequent().lookup(var)) {
	    IteratorOfTacletGoalTemplate it=goalTemplates().iterator();
	    TacletGoalTemplate tpl;
	    while (it.hasNext()) {
		tpl=it.next();		
		if (tpl.sequent().lookup(var)
		    || (tpl instanceof RewriteTacletGoalTemplate 
			&& 
			lookupInTerm(var,
				     ((RewriteTacletGoalTemplate)tpl).replaceWith()))		    
		    ) {

		    return true;


		} 
	    }
	    return false;
	}
	return true;	    
    }


    /** 
     * looks if a schema variable occurs in the ifseq or
     * in the sequent of the goal templates(@see
     * de.uka.ilkd.key.rule.TacletGoalTemplate#seq() seq).
     * (this is ifseq, and all addseq )
     * @param var the SchemaVariable to be looked for
     * @return true iff the variable has been found
     */
    protected boolean lookupInAreaOne(SchemaVariable var) { 
	if (!ifSequent().lookup(var)) {
	    IteratorOfTacletGoalTemplate it=goalTemplates().iterator();
	    while (it.hasNext()) {				
		if (it.next().sequent().lookup(var)) {  
		    return true;
		} 
	    }
	    return false;
	}
	return true;	    
    }

    /** 
     * checks if a schema variable occurrs maximal in one instantiation area
     * @param var the SchemaVariable whose instantiation is checked
     * @return true iff instantiation is correct
     */
    protected boolean correctInstantiationAreaOccurrence(SchemaVariable var) {
	Boolean correct = areaCache.get(var);
	if (correct != null) {
	    return correct.booleanValue();
	} else if (var.isVariableSV() || varDeclaredNew(var)!=null) {
	    return super.correctInstantiationAreaOccurrence(var);
	}	
	// var is not new and findeseq is term or formula
	boolean found = lookupInAreaOne(var);
	boolean foundInFind = lookupInTerm(var,find());

	IteratorOfTacletGoalTemplate tplIt = goalTemplates().iterator();
	while (tplIt.hasNext()) {
	    TacletGoalTemplate tpl=tplIt.next();
	    if ((tpl instanceof RewriteTacletGoalTemplate) &&
		(!find().equals(((RewriteTacletGoalTemplate)tpl).replaceWith()))) {  
		if ( foundInFind || 
		     lookupInTerm(var,
			  ((RewriteTacletGoalTemplate)tpl).replaceWith())) {
		    if (found) {
			areaCache.put(var, new Boolean(false));
			return false; 
		    } else {
			found=true;
		    }
		}
	    }
	}

	if (!lookupInAddRulesArea(var,found)) { // does not occur
		areaCache.put(var, new Boolean(true));		
		return true;
	}

	areaCache.put(var, new Boolean(false));
	return false;
    }

    /**
     * does the work for applyReplacewith (wraps recursion) 
     */
    private Term replace(Term term, Term with, IteratorOfInteger it,
			 JavaInfo javaInfo, SVInstantiations svInst) {	
	if (it.hasNext()) {	    
	    int sub=it.next().intValue();
	    Term[] subs=new Term[term.arity()];
	    for (int i=0;i<term.arity();i++) {
		if (i!=sub) {
		    subs[i]=term.sub(i);
		} else {
		    subs[i]=replace(term.sub(i), with, it, javaInfo, svInst);
		}
	    }
	    
	    ArrayOfQuantifiableVariable origvars 
		= (term.varsBoundHere(0).size()==0)
		? term.varsBoundHere(1) 
		: term.varsBoundHere(0);
	    QuantifiableVariable[] vars 
		= new QuantifiableVariable[origvars.size()];
	    for (int i=0; i<origvars.size(); i++) {
		vars[i] = origvars.getQuantifiableVariable(i);
	    }	    	    
	    return TermFactory.DEFAULT.createTerm(term.op(), subs, vars,
						  term.javaBlock(), 
						  term.ruleVar());	    
	} 
	return svInst == SVInstantiations.EMPTY_SVINSTANTIATIONS ?
	    with : syntacticalReplace(with, javaInfo, svInst);
    }


    /**
     * Check if p_pos contains an explicit metavariable instantiation,
     * and creates in this case a new simple term by replacing the
     * metavariable with the instantiation
     */
    private PosInOccurrence flatten ( PosInOccurrence p_pos,
				      JavaInfo        p_javaInfo) {
	if ( p_pos.termBelowMetavariable () != null ) {
	    Term flatTerm = replace ( p_pos.constrainedFormula ().formula (),
				      p_pos.termBelowMetavariable (),
				      p_pos.posInTerm ().iterator (),
				      p_javaInfo,
				      SVInstantiations.EMPTY_SVINSTANTIATIONS );

	    PosInOccurrence pos = new PosInOccurrence
		( new ConstrainedFormula ( flatTerm,
					   p_pos.constrainedFormula ().constraint () ),
		  p_pos.posInTerm (),
		  p_pos.proof     () );

	    IteratorOfInteger it = p_pos.posInTermBelowMetavariable ().iterator ();
	    while ( it.hasNext () )
		pos = pos.down ( it.next ().intValue () );

	    return pos;
	}

	return p_pos;
    }


    /** CONSTRAINT NOT USED 
     * applies the replacewith part of Taclets
     * @param gt TacletGoalTemplate used to get the replaceexpression in the Taclet
     * @param goal the Goal where the rule is applied
     * @param posOfFind the PosInOccurrence belonging to the find expression
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst a SVInstantiations object containing the necessary
     * instantiations
     */
    protected void applyReplacewith(TacletGoalTemplate gt, 
				    Goal               goal,
				    PosInOccurrence    posOfFind,
				    JavaInfo           javaInfo,
				    MatchConditions    matchCond) {
	if ( gt instanceof RewriteTacletGoalTemplate ) {
	    PosInOccurrence flatPos = flatten ( posOfFind, javaInfo );

	    Term term = flatPos.constrainedFormula().formula();
	    IteratorOfInteger it = flatPos.posInTerm().iterator();
	    Term rwTemplate=((RewriteTacletGoalTemplate)gt).replaceWith();
	    Term result = replace(term, rwTemplate, it, javaInfo, matchCond.getInstantiations ());
	    Constraint c = matchCond.getConstraint ();
	    if ( c.isAsWeakAs ( posOfFind.constrainedFormula ().constraint () ) )
		goal.changeFormula(new ConstrainedFormula(result, c), posOfFind);
	    else
		goal.addFormula(new ConstrainedFormula(result, c), posOfFind);
	} else {
	    // Then there was no replacewith...
	    // This is strange in a RewriteTaclet, but who knows...
	}
    }

    /**
     * adds the sequent of the add part of the Taclet to the goal sequent
     * @param add the Sequent to be added
     * @param goal the Goal to be updated
     * @param posOfFind the PosInOccurrence describes the place where to add
     * the semisequent
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst a SVInstantiations object containing the necessary
     * instantiations
     */
    protected void applyAdd(Sequent         add, 
			    Goal            goal,
			    PosInOccurrence posOfFind,
			    JavaInfo        javaInfo,
			    MatchConditions matchCond) {
	if (posOfFind.isInAntec()) {
	    addToAntec(add.antecedent(), goal, posOfFind, javaInfo, matchCond);
	    addToSucc(add.succedent(), goal, null, javaInfo, matchCond);
	} else {
	    addToAntec(add.antecedent(), goal, null, javaInfo, matchCond);	
	    addToSucc(add.succedent(), goal, posOfFind, javaInfo, matchCond);
	}
    }
    
    protected Taclet setName(String s) {
	RewriteTacletBuilder b = new RewriteTacletBuilder();
	b.setFind(find());
	b.setSameUpdatePrefix ( getSameUpdatePrefix () );
	return super.setName(s, b);
    }


    StringBuffer toStringFind(StringBuffer sb) {
	StringBuffer res = super.toStringFind ( sb );
	if ( getSameUpdatePrefix () )
	    res.append ( "sameUpdateLevel\n" );
	return res;
    }

}
