// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.expression.Literal;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.util.ExtList;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;

import java.util.*;
import java.text.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.java.visitor.JavaIterator;

// There used to be methods in this class to calculate proof obligations
// from taclets.  Since v0.82 however, the logic data structures do not
// match those methods.  They can be retreived from PRCS version 0.81 of
// Feb 5th 2001. They have been removed since version 0.632, if you have to look
// on them, checkout an elder version

/** 
 * Taclets are the DL-extension of schematic theory specific rules. They are
 * used to describe rules of a logic (sequent) calculus. A typical taclet
 * definition looks similar to <br></br>
 * <code> 
 *    taclet_name { if ( ... ) find ( ... ) goal_descriptions }
 * </code> <br></br>
 * where the if-part must and the find-part can contain a sequent arrow, that
 * indicates, if a term has to occur at the top level and if so, on which side of
 * the sequent. The goal descriptions consists of lists of add and replacewith
 * constructs. They describe, how to construct a new goal out of the old one by
 * adding or replacing parts of the sequent. Each of these lists describe a new
 * goal, whereas if no such list exists, means that the goal is closed. <p>
 *   The find part of a taclet is used to attached the rule to a term in the
 * sequent of the current goal. Therefore the term of the sequent has to match
 * the schema as found in the taclet's find part. The taclet is then attached to
 * this term, more precise not the taclet itself, but an application object of
 * this taclet (see {@link de.uka.ilkd.key.rule.TacletApp TacletApp}. When
 * this attached taclet application object is applied, the new goals are
 * constructed as described by the goal descriptions. For example <br></br>
 * <code> 
 *    find (A | B ==>) replacewith ( A ==> ); replacewith(B ==>) 
 * </code> <br></br>creates 
 * two new goals, where the first has been built by replacing <code> A | B </code>
 * with <code>A</code> and the second one by replacing <code>A | B</code> with
 * <code>B</code>. For a complete description of the syntax and semantics of
 * taclets consult the KeY-Manual.  The objects of this class serve different
 * purposes: First they represent the syntactical structure of a taclet, but 
 * they also include the taclet interpreter isself. The taclet interpreter
 * knows two modes: the match and the execution mode. The match mode tries to
 * find a a mapping from schemavariables to a given term or formula. In the
 * execution mode, a given goal is manipulated in the manner as described by the
 * goal descriptions. </p>
 * <p>
 *   But an object of this class neither copies or split the goal, nor it
 * iterates through a sequent looking where it can be applied, these tasks have
 * to be done in advance. For example by one of the following classes 
 * {@link de.uka.ilkd.key.proof.RuleAppIndex RuleAppIndex} or 
 * {@link de.uka.ilkd.key.proof.TacletAppIndex TacletAppIndex} or 
 * {@link de.uka.ilkd.key.rule.TacletApp TacletApp} </p>
 */

public abstract class Taclet implements Rule, Named {

    private final String AUTONAME = "taclet";

    /** name of the taclet */
    private Name name;
    /** name displayed by the pretty printer */
    private String displayName = null;
    /** contains useful text explaining the taclet */
    private String helpText = null;

    /**
     * @label ifSequent
     */
    private Sequent ifSequent;
    /** 
     * Variables that have to be created each time the taclet is applied. 
     * Those variables occur in the varcond part of a taclet description.
     */
    private ListOfNewVarcond varsNew;
    /** 
     * variables with a "x not free in y" variable condition. This means the
     * instantiation of VariableSV x must not occur free in the instantiation of
     * TermSV y.
     */
    private ListOfNotFreeIn varsNotFreeIn;
    /** 
     * variable conditions used to express that a termsv depends on the free
     * variables of a given formula(SV)
     * Used by skolemization rules.
     */
    private ListOfNewDependingOn varsNewDependingOn;
    /** Additional generic conditions for schema variable instantiations. */
    private final List variableConditions;

    /**
     * (currently unused) indicates if a the same taclet has to be
     * applied recursively on its results
     */
    private boolean recursive;
    /**
     * @link aggregation
     * @supplierCardinality *
     */
    /*#private TacletGoalTemplate lnkTacletGoalTemplate;*/
    private ListOfTacletGoalTemplate goalTemplates;
    /**
     * @link aggregation
     * @supplierCardinality *
     * @label Rule Heuristics
     */
    /*#private Heuristic lnkHeuristic;*/
    protected ListOfHeuristic heuristics;
    /** 
     * should taclet be applied by heuristics only 
     */
    private boolean noninteractive;
    /** 
     * constraint under which the Taclet is valid 
     */
    private Constraint constraint;
    /** 
     * caches variable to variable occurs in less than two
     * instantiation areas
     */
    protected HashMapFromSchemaVariableToBoolean areaCache;

    /**
     * map from a schemavariable to its prefix. The prefix is used to test
     * correct instantiations of the schemavariables by resolving/avoiding
     * collisions. Mainly the prefix consists of a list of all variables that
     * may appear free in the instantiation of the schemavariable (a bit more
     * complicated for rewrite taclets, see paper of M:Giese)
     */
    protected MapFromSchemaVariableToTacletPrefix prefixMap;
    
    /** for a variable pair (x,y) in this list: x occurs in the scope
     * of a variable y in this Taclet. At least the pair (y,y) occurs if
     * variable y is bound in the Taclet, i.e. the ys are the bound
     * variables in the Taclet.
     */
    protected SetOfInScopePair varInScope
	=SetAsListOfInScopePair.EMPTY_SET;


    /** contains set of all bound variables */
    protected SetOfQuantifiableVariable boundVars
	=SetAsListOfQuantifiableVariable.EMPTY_SET;
    
    /** tracks state of pre-computation */
    private boolean contextInfoComputed=false;
    private boolean contextIsInPrefix=false;
    /** true if one of the goal descriptions is a replacewith */
    private boolean hasReplaceWith = false;
     
    
    protected String tacletAsString;

    /** Set of schemavariables of the if part */
    private SetOfSchemaVariable ifVariables = null;

    /** This map contains (a, b) if there is a substitution {b a}
     * somewhere in this taclet */
    private MapFromSchemaVariableToSchemaVariable
	svNameCorrespondences = null;


    /** creates an empty Taclet */
    Taclet() {
	this("");
    }

    /** 
     * creates an empty Taclet with the given name 
     */
    Taclet(String nameStr) {
	name               = new Name(nameStr);
	ifSequent          = Sequent.EMPTY_SEQUENT;
	varsNew            = SLListOfNewVarcond.EMPTY_LIST;
	varsNotFreeIn      = SLListOfNotFreeIn.EMPTY_LIST;
	varsNewDependingOn = SLListOfNewDependingOn.EMPTY_LIST;
	variableConditions = Collections.EMPTY_LIST;
	recursive          = false;
	goalTemplates      = SLListOfTacletGoalTemplate.EMPTY_LIST;
	heuristics         = SLListOfHeuristic.EMPTY_LIST;
	noninteractive     = false;	
	constraint         = Constraint.BOTTOM;
	areaCache          = new HashMapFromSchemaVariableToBoolean();
	prefixMap          =
	    MapAsListFromSchemaVariableToTacletPrefix.EMPTY_MAP;
	if (this.displayName == null)  this.displayName = name.toString();
    }

    /**
     * creates a Schematic Theory Specific Rule (Taclet) with the given
     * parameters.  
     * @param name the name of the Taclet 
     * @param applPart contains the application part of an Taclet that is
     * the if-sequence, the variable conditions
     * @param goalTemplates a list of goal descriptions.
     * @param heuristics a list of heurisics for the Taclet
     * @param constraint the Constraint under which the Taclet is valid
     * @param attrs attributes for the Taclet; these are boolean values
     * indicating a noninteractive or recursive use of the Taclet. 
     *
     */
    Taclet(Name                     name, 
	   TacletApplPart           applPart,  
	   ListOfTacletGoalTemplate goalTemplates, 
	   ListOfHeuristic          heuristics,
	   Constraint               constraint, 
	   TacletAttributes         attrs,
	   MapFromSchemaVariableToTacletPrefix prefixMap){

	this.name          = name;
	ifSequent          = applPart.ifSequent();
	varsNew            = applPart.varsNew();
	varsNotFreeIn      = applPart.varsNotFreeIn();
	varsNewDependingOn = applPart.varsNewDependingOn();
	variableConditions = applPart.getVariableConditions();
	this.goalTemplates = goalTemplates;
	this.heuristics    = heuristics;
	recursive          = attrs.recursive();
	noninteractive     = attrs.noninteractive();
	this.constraint    = constraint;
	areaCache          = 
	    new HashMapFromSchemaVariableToBoolean();
	this.prefixMap     = prefixMap;
        this.displayName   = attrs.displayName();
        if (this.displayName == null)  this.displayName = name.toString();
    }

    
    protected void cacheMatchInfo() {
	boundVars = SetAsListOfQuantifiableVariable.EMPTY_SET;
	varInScope = getVarInScope();

	IteratorOfInScopePair it = varInScope.iterator();

	while (it.hasNext()) {
	    boundVars = 
		boundVars.add((QuantifiableVariable)it.next().second());
	}
	
	IteratorOfTacletGoalTemplate goalDescriptions = 
	    goalTemplates.iterator();

	
	while (!hasReplaceWith && goalDescriptions.hasNext()) {
	    if (goalDescriptions.next().
		replaceWithExpressionAsObject() != null) {
		hasReplaceWith = true;
	    }
	}
	
    }

    /** looks for a Schemavariable var in Term term
     * @param var the SchemaVariable to be looked for
     * @param term the Term to be looked through
     * @return true iff variable has been found in the term 
     */    
    protected boolean lookupInTerm(SchemaVariable var, Term term) {
	TacletVariableCollector vcv=new TacletVariableCollector();
	term.execPostOrder(vcv);
	return vcv.contains(var);
    }
    
    /** checks if a given variables is used in the rule this implies
     * in the ifSequent, findexp, rwexp, addseq or addrule
     * @param var the Variable to be looked for
     * @return true iff Variable var occurs in the Taclet
     */
    protected abstract boolean lookup(SchemaVariable var);

    /** partial implementation of lookup. But lookup
     * must also check the findexp for FindTaclets. 
     * checks if a given variables is used in the rule this implies
     * the ifSequent, addseq, rwexp and addrule. 
     * @param var the Variable to be looked for
     * @return true iff Variable var occurs in the Taclet
     */    
    protected boolean lookupImpl(SchemaVariable var) {
	if (!lookupInIfAddRwExp(var)) {
	    IteratorOfTacletGoalTemplate tplIt = goalTemplates().iterator();
	    while (tplIt.hasNext()) {
		IteratorOfTaclet ruleIt=tplIt.next().rules().iterator();
		while (ruleIt.hasNext()) {
		    if (ruleIt.next().lookup(var)) {
			return true;
		    }
		}    
	    }
	    return false;
	}
	return true;
    }	


    /** looks if a schema variable occurs in the ifSequent or
     * in the sequent of the goal templates(@see
     * de.uka.ilkd.key.rule.TacletGoalTemplate#seq() seq).
     * (this is ifSequent, addexp and rwexp, rwexp only if not RewriteTaclet
     * (overwrites this method)).
     * @param var the SchemaVariable to be looked for
     * @return true iff the variable has been found
     */
    protected boolean lookupInAreaOne(SchemaVariable var) {
	return lookupInIfAddRwExp(var);
    }


    /** looks if a variable occurs in the ifSequent or
     * in the sequent of the goal templates(@see
     * de.uka.ilkd.key.rule.TacletGoalTemplate#seq() seq).
     * (this is ifSequent, addexp and rwexp)
     * used by lookup)
     * @param var the SchemaVariable to be looked for
     * @return true iff the variable has been found
     */
    protected abstract boolean lookupInIfAddRwExp(SchemaVariable var);
    

    /** checks if a variable occurs in
     * the addrules part of the Taclet. The method returns true if the
     * variable is found in two areas (not necessarily in two addrule
     * areas)
     * @param var the SchemaVariable to be looked for
     * @param foundBefore a boolean that has to be set true iff the
     * variable to be looked for has already been found in a former area
     * @return true iff. variable is found in more than two areas
     */
    protected boolean lookupInAddRulesArea(SchemaVariable var, 
					   boolean foundBefore) {
	boolean found = foundBefore;
	IteratorOfTacletGoalTemplate
	    goalTplIt=goalTemplates().iterator();
	while (goalTplIt.hasNext()) {
	    IteratorOfTaclet tacletIt=goalTplIt.next().rules().iterator();
		while (tacletIt.hasNext()) {
		    Taclet s=tacletIt.next();
		    if (s.lookup(var)) {
			if (found) { 
			    return true;
			}
			found = true;
		    }
		}
	}
	return false;
    }

    /** checks if a variable occurrs maximal in one instantiation area
     * @param var the SchemaVariable whose instantiation is checked
     * @return true iff instantiation is correct
     */
    protected abstract boolean
	correctInstantiationAreaOccurrence(SchemaVariable var); 

    /**
     * looks if a variable is declared as not free in
     * @param var the SchemaVariable to look for
     * @return true iff declared not free
     */
    private boolean varDeclaredNotFree(SchemaVariable var) {
	IteratorOfNotFreeIn it = varsNotFreeIn();
	while (it.hasNext()) {
	    if (it.next().first() == var) {
		return true;
	    }
	}
	return false;
    }

    /**
     * returns true iff the taclet contains a "close goal"-statement
     * @return true iff the taclet contains a "close goal"-statement
     */
    public boolean closeGoal () {
	return goalTemplates == SLListOfTacletGoalTemplate.EMPTY_LIST;
    }

    /**
     * looks if a variable is declared as new and returns its sort to match
     * with or the schema variable it shares the match-sort with. Returns
     * null if the SV is not declared to as new.
     * @param var the SchemaVariable to look for
     * @return the sort of the SV to match or the SV it shares the same
     * match-sort with
     */
    public NewVarcond varDeclaredNew(SchemaVariable var) {
	IteratorOfNewVarcond it=varsNew.iterator();
	while (it.hasNext()) {
	    NewVarcond nv=it.next();
	    if (nv.getSchemaVariable()==var) {
		return nv;
	    }
	}
	return null;
    }


    /**
     * tests if the schema variable var has to fulfil the variable condition 
     * "new depending on" and if positive returns true
     * @return true if the schemavariable has to fulfil the variable condition
     * "new depending on"
     */
    public boolean varDeclaredNewDependingOn(SchemaVariable var) {
	IteratorOfNewDependingOn it = varsNewDependingOn();
	while (it.hasNext()) {
	    if (it.next().first() == var) {
		return true;
	    }
	}
	return false;
    }

    /**
     * returns true iff the given variable is bound either in the
     * ifSequent or in 
     * any part of the TacletGoalTemplates
     * @param v the bound variable to be searched 
     */
    protected boolean varIsBound(SchemaVariable v) {
	return (v instanceof QuantifiableVariable) 
	    ? boundVars.contains((QuantifiableVariable)v)
	    : false;
    }

    /**
     * returns pairs of variables (x,y) where y is bound in this Taclet and x
     * occurs in the scope of y; (y,y) is part of the set for all bound ys
     * with no x bound in the scope of y.
     * @return a set of variable pairs as specified above
     */
    protected SetOfInScopePair getVarInScope() {
	IteratorOfTacletGoalTemplate it=goalTemplates().iterator();
	SetOfInScopePair result
	    = SetAsListOfInScopePair.EMPTY_SET;
	while (it.hasNext()) {
	    result = result.union(it.next().getVarInScope());
	}
	return result.union(ifSequent.getVarInScope());	
    }

    /** checks if the program element can potentially belong to the
     * prefix
     * @param pe the ProgramElement to check
     * @return true iff prefix compatible
     */
    public static boolean isPrefix(ProgramElement pe) {
	if (pe instanceof LabeledStatement || 
	    pe instanceof StatementBlock ||
	    pe instanceof Try || 
	    pe instanceof MethodCallStatement) {
	    return true;
	}
	return false;
    }

    /** returns a list of program elements whose first element is the
     * statement that is not in the prefix. Sideeffect: the position
     * of the given iterator will be the first element after the prefix as
     * result (or null if no such exists). The iterator will return 'null' in
     * cases like this: <{ {{{{}}}} }>  in cases like <{ { {} int i = 0; } }>
     * the iterator will return '{}'.
     * @param it the JavaIterator to use
     * @return a list of program elements who build the prefix      
     */
    private ListOfProgramElement getPrefix(JavaIterator it) {
	ListOfProgramElement prefixList = SLListOfProgramElement.EMPTY_LIST;

	ProgramElement lastElement = null;	
	ProgramElement el = it.element();	
	boolean labeled         = false;	
	boolean was_empty_block = false;
	boolean is_empty_block  = false;

	while (it.hasNext()) {
	    // the empty block thing catches the following to
	    // situations in case of {} stmnt; 
	    // the last thing not in prefix is the closing brace '}'
	    // in a situation like l1:{} we want to only l1:{ to be
	    // once inside the prefix 
	    if (isEmptyBlock(el)) {
		is_empty_block = true;		
	    } else {
		is_empty_block = false;
	    }

	    Debug.assertTrue(el != null, "JavaIterator must have a bug.");

	    if (labeled && is_empty_block) {
		// we have a labeled empty block
		// next() method is called the empty block will be returned
		labeled = false;
		return prefixList;
	    } else if (is_empty_block) {
		return prefixList.prepend(el);
	    } else if (!isPrefix(el)) {		
		// Debug.out("taclet: Not prefix compatible program element:", el);
		labeled = false;
 		if (was_empty_block) {
 		    // the last thing not in prefix is the closing
 		    // brace '}' so set back twice		    
 		    it.back();
 		}
		return prefixList;
	    } else {		
		if (el instanceof StatementBlock) {
		    if (!labeled || 
			lastElement instanceof LabeledStatement) {
			prefixList = prefixList.prepend(el);
		    }
		} else {
		    prefixList = prefixList.prepend(el);
		}
		labeled = false;
	    }
	    if (el instanceof MethodCallStatement) {
		labeled = true;
		if (((MethodCallStatement)el).getProgramVariable() != null) {
		    // skip result var
		    it.next();
		}
	    } else if (el instanceof LabeledStatement) {
		labeled = true;
		// skip name
		it.next();	       
	    }
	    if (el instanceof Try) {
		labeled = true;
		// do not skip block because it may be the empty one
	    }
	    was_empty_block = is_empty_block;
	    lastElement = el;
	    el = it.next();
	}	
	return prefixList;
    }

    /** 
     * returns true iff the program element is StatementBlock with no
     * children 
     * @return true iff the program element is StatementBlock with no
     * children 
     */
    private boolean isEmptyBlock(ProgramElement pe) {
	if (pe != null && pe instanceof StatementBlock) {
	    if (((StatementBlock)pe).getChildCount() == 0) {
		return true;
	    }
	}
	return false;
    }

    /** returns the position in the program of the prefix if it can be
     * matched, null if no match has been found. Sideeffect: 
     * the iterator seqPE delivers after the next next() call the
     * first element that has to be matched 
     * @param seqPE the JavaIterator through the program element
     * part of a the term in the sequent for which we look if this
     * Taclet fits
     * @param tacletPE the JavaIterator through the program element
     * that is part of the find- or if-section of the Taclet
     * @return the position in the program of the prefix if it can be
     * matched, null if no match has been found
     */
    private PosInProgram matchPrefix( JavaIterator seqPE,
				      JavaIterator tacletPE ) {
	PosInProgram prefixPos = null;
      	ListOfProgramElement prefixList = getPrefix(seqPE); 
	// list of elements at the beginning of the program elements
	// of the Taclet (normally inside a context term) that are
	// potential prefix elements but that are not in the prefix
	// because the term in the find- or if-section of this sequent
	// that has to be matched starts with context compatible elements.
	// As every context term starts with a statement block as
	// wrapper, the wrapper appears in the prefix list and has to
	// be ignored. As we are only interested in the length of the
	// list, this is done by substracting 1 from the lists size.	
	int notInPrefixDepth = getPrefix(tacletPE).size() - 1;
	if (notInPrefixDepth > prefixList.size()) {
	    // the term we try to match with the taclet has no context
	    // compatible elements but some are required by the taclet
	    // term, so a matching is impossible and we return null
 	    return null;
 	} else if (notInPrefixDepth != 0) {
	    // we have now a maximal prefix. Now we have to go backwards
	    // so that 'notInPrefixDepth' elements are left that have to
	    // be checked if they match the taclet ones.
	    Debug.out("taclet: Prefixelements to match(nr):", notInPrefixDepth);
	    ProgramElement seqPEback = seqPE.element();
	    for (int i = 0; i < notInPrefixDepth; i++) { 		
		if (i == 1 && isEmptyBlock(seqPEback)) {
		    // in a situation where we match s.th. like
		    // {....{}...} the first iteration may not set
		    // back the iterator as we are at the correct
		    // position, but if we have to set back more
		    // elements than just one, we have to set the
		    // iterator back at the next pass. Afterwards a
		    // similar situation cannot happen
		    seqPEback = seqPE.back();
		}
		if (!isEmptyBlock(seqPEback)) {		   
		    seqPEback = seqPE.back();
		    if (seqPEback instanceof StatementBlock) {
			// take try { and methodcall..:{ as one
			seqPEback = seqPE.back();
			if (!(seqPEback instanceof Try ||
			    seqPEback instanceof ProgramVariable ||
			      seqPEback instanceof MethodCallStatement)) {
			    seqPEback = seqPE.next();
			} 
		    }
		} else {
		    // skip labeled empty block as we skip it when
		    // calculating the prefix 
		    ProgramElement lookAhead = seqPE.back();		    
		    if (lookAhead instanceof Label || 
			lookAhead instanceof Try || 
			lookAhead instanceof MethodCallStatement) {
			seqPEback = lookAhead;
		    } else if (lookAhead instanceof ProgramVariable) {
			// can happen if we have a methodcall construct
			ProgramElement lookAhead2 = seqPE.back();
			if (lookAhead2 instanceof MethodCallStatement) {
			    seqPEback = lookAhead2;
			} else {
			    seqPE.next();
			    seqPE.next();			    
			}		    
		    } else {
			seqPE.next();
		    }
		}
		if (seqPEback != null) {
		    if (seqPEback instanceof Label) {
			seqPEback = seqPE.back();		
			if (!(seqPEback instanceof LabeledStatement)) {
			    return null;
			} 
		    } else if (seqPEback instanceof ProgramVariable) {
			seqPEback = seqPE.back();
			if (!(seqPEback instanceof MethodCallStatement)) {
			    return null;
			} 			
		    }
		}
	    }
	}
	// get position of the prefix
	ProgramElement last = seqPE.element();
	if (last == null) {
 	    // at the end of the program ? can happen 
	    // e.g. if <{ {{{{{}}}}} }> has to be matched
	    // as all program elements disappear in the prefix
	    last = seqPE.back();
 	}
	// catch the case that we have a labeled empty block in these
	// case we want to have returned the labeled block and not
	// just the empty one
	if (isEmptyBlock(last)) {
	    last = seqPE.back();
	    if (last instanceof Label || 
		last instanceof ProgramVariable) {
		last = seqPE.back(); // back to labeled/method call statement
	    } else if (!(last instanceof Try || 
			 last instanceof MethodCallStatement)) {
		last = seqPE.next(); // no labeled or try block 
	    }
	}
	return seqPE.position();
    }



    /** this method is responsible for recognizing closing braces, as
     * we match a tree using a kind of iterator that works in a depth
     * first search mode, it is not this trivial to decide whether a
     * match has to fail because the pattern has a closing brace but
     * the program not, e.g. the template
     * <code> 
     * { {return #se;} #s  } 
     * </code> 
     * where <code>#se</code> is the schemavariable for simple
     * expression and <code>#s</code> a schemavariable for a
     * statemen. We have to avoid that this matches to a program like
     * <code> 
     *   { return 2; int i = 0; }
     * </code>
     *  that could easily happen because after matching
     * <code>#se </code> to <code> 2 </code> the iterator returns as
     * next element <code> int i = 0; </code> that matches with 
     * <code> #s </code> returned by the iterator of the template tree.
     */
    private boolean matchClosingBraces(PosInProgram prefixPos,
				       PosInProgram programPos,
				       PosInProgram lastProgramPos,		       
				       PosInProgram templatePos,
				       PosInProgram lastTemplatePos,				       
				       int          templateOffset,
				       boolean templateHasNext) {


	int prefixDepth   = prefixPos.depth() - 1;

	/* If the depth of the template is null, then the end of the
	   template that has to be matched is reached. As there is an
	   enclosing wrapper around the template the depth of it is
	   equal to one (not zero)
	 */
	int templateDepth = (templatePos == null ? 
			     templateOffset : templatePos.depth()); 
	
	int programDepth  = programPos == null ? 0 : programPos.depth();



	/* if the first element that does not belong to the prefix is
	   at a deeper position than the element of the program that
	   has to be matched, then the match fails because we have
	   left the part of the program to be matched and entered
	   parts of the enclosing context, e.g. 
	   <code> <{ {{int i = 0; }} int j=2;  }> </code>
	   is tried to be matched by
	   <code> <{ pi #t #v = #se; #s omega }> </code>
	   The schemavariable <code> #s </code> would be instantiated by
	   <code> int j=2; </code> but int j = 2; is in the context.
	*/	
	if (templatePos != null && 
	    (prefixDepth > programDepth || 
	     (prefixDepth == programDepth && 
	      !(programPos == null || programPos.leq(prefixPos)))))
	    {
		Debug.out("taclet: MATCH FAILED. Entered context but there "+
			  "are templates to be matched.");
		return false; // FAILED
	    }

	/* We calculate the difference of the depth of the position of
	   the last template element ans the current element that has
	   to be matched next. The difference is a measure for the
	   number of enclosing braces. (Same for the program element)
	*/
	int templateDiff = (templateDepth - lastTemplatePos.depth());	
	int matchDiff    = (programDepth  - lastProgramPos.depth());
	if (matchDiff != templateDiff &&
	    (matchDiff >= 0 || templateHasNext || 
	     templateDiff - matchDiff > prefixDepth+1)) {
	    Debug.out("taclet: MATCH FAILED. One has left the "+
		      "container the other not.");
	    Debug.out("taclet: Program position  (current, last): ", 
		      programPos, lastProgramPos);
	    Debug.out("taclet: Template position (current, last):" , 
		      templatePos, lastTemplatePos);
	    return false;
	}
	return true;
    }


    /** 
     * returns true, if the given SchemaVariable can stand for the
     * ProgramElement 
     * @param sv the SchemaVariable
     * @param match the ProgramElement to be matched
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return true if the SchemaVariable canstand for the given element
     */
    private boolean check(SchemaVariable sv, ProgramElement match,
			  JavaInfo javaInfo) {
	Debug.out("taclet: Check: SchemaVariable (name, sort): ",sv, sv.sort());
	Debug.out("taclet: With: ", match.getClass());
	if (match == null) {
	    Debug.out("taclet: MATCH FAILED. 'null' cannot be "+
		      "matched to a SchemaVariable.");
	    return false;
	}
	if (!((ProgramSVSort)sv.sort()).canStandFor(match, javaInfo)) {
	    Debug.out("taclet: MATCH FAILED. Sort mismatch.", match.getClass());	
	    return false; // FAILED
	}
	return true;
    }

    /** 
     * the following method is used to match schemavariables that can
     * stand for lists of program elements, e.g. list of statements or
     * catch-finally expressions
     * @param schemavar the SchemaVariable tried to be instantiated
     * @param programIterator the iterator through the program to be
     * matched
     * @param instantiations the SVInstantiations that has been found
     * until now
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return SVInstantiations with the new found matchings or null
     * if match failed
     */
    private SVInstantiations
	matchProgramElementList(SchemaVariable schemavar,
				boolean templateBlockHasNextSibling,
				JavaIterator programIterator,
				SVInstantiations instantiations,
				JavaInfo javaInfo) {
	SVInstantiations result = instantiations;
	// match all statements on this level
	// if the schemavariable has been matched we need to
	// check if the matching is equal
	ArrayOfProgramElement plist = (ArrayOfProgramElement)
	    result.getInstantiation(schemavar);
	ProgramElement program = programIterator.element();
	Debug.out("taclet: Start list match at element:", program);
	NonTerminalProgramElement parent =
	    (NonTerminalProgramElement) programIterator.parent();
	// length of statements to be matched is number of
	// child - current position
	int childCount = parent.getChildCount();
	int childPos   = parent.getIndexOfChild(programIterator.element());
	int size       = childCount-childPos;
	int count      = 0;
	boolean hasNextSibling = false;
	ExtList matchedStmnts = new ExtList ();
	do {
	    if (plist != null) {
		// schemavar has been matched before so we just
		// have to test if the instantiations are
		// equal
		if ( (count >= plist.size() && !(count==0 && plist.size()==0)) ||
		    !(plist.getProgramElement(count).equals(program))) {
		    Debug.out("taclet: The SchemaVariable has been matched "+
			      "before with a different block");
		    return null; // FAILED
		}
	    } else if (!check(schemavar, program, javaInfo)) {
		if ( templateBlockHasNextSibling ) {
		    programIterator.back();
		    break;
		} else {
		    // otherwise the list SV has to match the complete
		    // tail of block
		    if (count == 0) {
			programIterator.back();
			break;
		    } else {
			Debug.out("taclet: The SchemaVariable has been matched "+
				  "before with a different block");
			return null; // FAILED
		    }
		}
	    }
	    matchedStmnts.add ( program );
	    hasNextSibling   = programIterator.hasNextSibling();
	    program          = programIterator.nextSibling();
            count ++;
	} while (hasNextSibling);
	if ( !templateBlockHasNextSibling && size != count ) {
	    Debug.out("The list schema variable has to match all" +
		      " following elements");
	    return null; // FAILED
	}
	if (plist == null) {
	    Debug.out("taclet: Match Success. Length of list ",
		      matchedStmnts.size ());
	    ArrayOfProgramElement array = new ArrayOfProgramElement
		( (ProgramElement[])matchedStmnts
		  .collect ( ProgramElement.class ) );
	    result = result.add(schemavar, array);
	}
	return result;
    }

    /** 
     * Performs a pre-check if two non-terminal program elements might 
     * match. Therefore it compares mainly their sizes, as the template part
     * may contain list schema variables the check has to pay respect that the
     * size of the template's body can differ from the size of the program's
     * body. Let <emph>nr</emph> be the number of list schema variables in the
     * body of the template, then the template body is allowed to have at most
     * <emph>nr</emph> more children (lists may be empty), if <emph>nr=0</emph>
     * both non-terminal elements must have the same size otherwise the program
     * body's size is allowed to be greater or equal to the template's body
     * size.
     * @param template the NonTerminalProgramElement describing the schema in
     * which the program element "program" has to fit
     * @param program the NonTerminalProgramElement that is tried to be matched
     * @return false means no match is possible, true means the match may be
     * possible
     */
    private boolean checkBlockSize(NonTerminalProgramElement template,
				   NonTerminalProgramElement program) {

	int tplSize     = template.getChildCount();
	int programSize = program.getChildCount();

	// count list sv
	int count = 0;
	for (int i = 0; i<tplSize; i++) {
	    if (template.getChildAt(i) instanceof SchemaVariable &&
		((SchemaVariable)template.getChildAt(i)).isListSV()) {
		count++;
	    }		
	}

	if (tplSize == programSize || 
	    (count != 0 && tplSize-count<=programSize)) {
	    return true;
	}
	
	Debug.out("taclet: MATCH FAILED. Incompatible sizes "
		  +" of non terminal program elements");
	return false; //FAILED. Size incompatible
    }


    /** 
     * matches java block when the taclet term is a context term. Therefore it
     * first determines the program prefix/suffix and then tries to match the
     * program fragment starting at the prefix's end up to the suffix start
     * point
     * @param block the JavaBlock conatining the program fragment to be matched
     * against the context term
     * @param ctxt the ContextTerm describing the schema the given java program
     * has to fit for a successful match 
     * @param inst the SVInstantiations conatining the matching information
     * until this point, in particular all already matched schema variables
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.     
     * @return an extended version of the given SVInstantiations enriched with 
     * additional information about the required instantiation of schema
     * variables so that the java program can be matched or null if no match is
     * possible.
     */
    protected SVInstantiations matchJavaProgram(JavaBlock block,
						ContextTerm ctxt,
						SVInstantiations inst,
						JavaInfo javaInfo) {
	Debug.out("taclet: Starting program match of rule: ", this);
	Debug.out("taclet: Match (program, template):", block.program(), ctxt.getBody());
	
	JavaIterator programIterator = new JavaIterator(block.program());
	JavaIterator ctxtIt = new JavaIterator(ctxt.getBody());
	PosInProgram prefixPos = matchPrefix(programIterator, ctxtIt);

	if (prefixPos == null) {
	    Debug.out("taclet: Failed to match prefix.");
	    return null;  //FAILED prefix matching
	}

	JavaIterator templateIt = new JavaIterator(ctxt.getBody());
	// skip wrapping block
	templateIt.next();

	return matchJavaProgram(block, programIterator,
				prefixPos, templateIt,
				inst, javaInfo);
    }

    /** 
     * Matches the java program starting behind the the prefix (includes cases
     * where the prefix is empty or non-existant). 
     * @param programBlock the java fragment to be matched
     * @param programIterator a JavaIterator through the program to be matched
     * whose position is at the first element to match (if a prefix exist this
     * means the first statement that is not inside the prefix)
     * @param prefixPos the PosInProgram describing the position of the first
     * statement after the prefix
     * @param templateIt a JavaIterator through the template program starts at
     * the first element 
     * @param inst the SVInstantiations conatining the matching information
     * until this point, in particular all already matched schema variables
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return an extended version of the given SVInstantiations enriched with 
     * additional information about the required instantiation of schema
     * variables so that the java program can be matched or null if no match is
     * possible.
     */
    protected SVInstantiations matchJavaProgram(JavaBlock programBlock, 
                                                JavaIterator programIterator, 
                                                PosInProgram prefixPos,
	                                        JavaIterator templateIt, 
                                                SVInstantiations inst, 
                                                JavaInfo javaInfo) {
	SVInstantiations result      = inst;
	// the position before the last iteration
	ProgramElement program       = programIterator.element();
	PosInProgram lastProgramPos  = prefixPos;

	ProgramElement template      = templateIt.element();
	PosInProgram lastTemplatePos = templateIt.position();
	// if we have a context term the depth is here 1
	int templateStartDepth       = lastTemplatePos.depth();


	boolean forceEmptyList = false;

	do {
	    lastProgramPos = programIterator.position();

	    if (program == null && template != null &&
		template instanceof StatementBlock) {
		// may be the empty block match
		break;
	    } else if (program == null || template == null) {
		Debug.out("taclet:: MATCH FAILED. No program or no template.");
		return null;
	    }			
	    Debug.out("taclet: Match (program, class) ", program, program.getClass());
	    Debug.out("taclet: with (template, class) ", template, template.getClass());
	    if (template instanceof SchemaVariable) {
		SchemaVariable sv = (SchemaVariable) template;
		if (sv.isListSV()) {
		    result = matchProgramElementList
			(sv, templateIt.hasNextSibling (),
			 programIterator, result, javaInfo);		    
		    if (result == null) {
			Debug.out("taclet: MATCH FAILED. List match failed.");
			return null; //FAILED
		    }		
		} else { // schemavariables that are not matching on lists
		    if (!check(sv, program, javaInfo)) {
			Debug.out("taclet: MATCH FAILED. Sort of SchemaVariable cannot "+
				  "stand for the program");
			return null; //FAILED
		    }
		    Object instant=result.getInstantiation(sv);
		    if ( instant == null ||
			 instant.equals(program) ||
			 ((instant instanceof Term) 
			      && ((Term)instant).op().equals(program))) {
			Debug.out("taclet: Add instantiation (sv, program): ",sv, program);
			result = result.add(sv, program);
		    } else {
			Debug.out("taclet: MATCH FAILED 3. Former match of "+
				  " SchemaVariable incompatible with "+
				  " the current match.");
			return null; //FAILED mismatch
		    }
		}
	    } else if (template instanceof SchematicFieldReference) {
                if (!(program instanceof FieldReference)) return null;
		if (((FieldReference)program).getChildCount()<2) return null;
		if(ProgramSVSort.VARIABLE.canStandFor(program,javaInfo)) return null;
                SchemaVariable prefix = (SchemaVariable) 
                    ((SchematicFieldReference)template).getChildAt(0);		
                result = result.add(prefix, 
                                    ((FieldReference)program).getChildAt(0));		

                SchemaVariable suffix = (SchemaVariable) 
                    ((SchematicFieldReference)template).getChildAt(1);
		
		result = result.add((SchemaVariable)suffix,
					((FieldReference)program).getChildAt(1)); 
            } else if (template.getClass() != program.getClass()) {
		// if this match does not fail we can ssume in the
		// following cases that the type of the template and
		// the program to be matched are equal. Allows to
		// restrict instanceof checking to the template.
		Debug.out("taclet: MATCH FAILED. Different classes.");
		return null;  //FAILED
 	    } else if (template instanceof NonTerminalProgramElement) {
		if (!checkBlockSize
		    ((NonTerminalProgramElement)template,
		     (NonTerminalProgramElement)program)) {
		    Debug.out("taclet: MATCH FAILED. Incompatible non-terminal sizes.");
		    return null; //FAILED
		}		
 	    } else if (template instanceof Literal ||
		       template instanceof ProgramElementName) {
		if (!template.equals(program)) {
		    Debug.out("taclet: MATCH FAILED. Literal/Name mismatch.");
		    return null;  //FAILED
		}
	    }	    

	    // if the template is a schemavariable it has matched
	    // at least at the current subtree, so we have to skip
	    // it and not to descend into it.
	    program = (template instanceof SchemaVariable ?
		       programIterator.skip() : programIterator.next() );
	    lastTemplatePos = templateIt.position();
	    template = templateIt.next();
	    Debug.out("taclet: Previous position ", lastProgramPos);	
	    Debug.out("taclet: Next position is ", programIterator.position());	
	    while (!matchClosingBraces(prefixPos,
				       programIterator.position(),
				       lastProgramPos,
				       templateIt.position(),
				       lastTemplatePos,
				       templateStartDepth,
				       templateIt.hasNext())) {
		if (template instanceof SchemaVariable &&
		    ((SchemaVariable)template).isListSV()) {
		    Debug.out("taclet: Template variable is a list schemavariable."+
			      "Give another chance and force empty list.");
		    result = result.add((SchemaVariable)template, 
					new ArrayOfProgramElement(new ProgramElement[0]));
		    
		} else {
		    return null; //FAILED
		}	    
		/*
		  do not save "lastTemplatePos = templateIt.position();"
		  as we asume that the list is empty we have to emulate
		  the case that there has been no schema variable
		 */
		if (templateIt.hasNextSibling()) {
		    template = templateIt.nextSibling();
		} else {
		    template = templateIt.next();		    
		    Debug.out("taclet: Next template element is ", template);	
		    //		    if (!matchClosingBraces(prefixPos,
		    //					    programIterator.position(),
		    //					    lastProgramPos,
		    //					    templateIt.position(),
		    //					    lastTemplatePos,
		    //					    templateStartDepth,
		    //					    templateIt.hasNext())) {
		    //			return null; //FAILED
		    //		    }
		}
	    }
	} while (programIterator.hasNext() && templateIt.hasNext());
	if (templateIt.hasNext()) {
 	    Debug.out("taclet: MATCH FAILED. Template has more unmatched elements.");
	    return null;
	} else if (prefixPos.depth()==0  && programIterator.hasNext()) {
 	    Debug.out("taclet: MATCH FAILED. Suffix without prefix.");	    
	    return null;
	} else {    
	    result = result.add(prefixPos, lastProgramPos, programBlock.program());
	    Debug.out("taclet: :-))) RULE MATCHED :-)))");
	    Debug.out("taclet: (EndPos, result)", lastProgramPos, result);
	    return result;
	}
    }

		
    /**
     * returns a SVInstantiations object iff the given Term
     * template can be instantiated to 
     * match the given Term term using the known instantiations stored in
     * svInst.  If a
     * matching cannot be found null is returned.
     * The not free in condition is checked in TacletApp. Collisions are resolved
     * there as well, if necessary.
     * @param term the Term that has to be matched
     * @param template the Term that is checked if it can match term
     * @param ignoreUpdates a boolean if set to true updates will be ignored as 
     * e.g. wanted if an if-sequent is matched
     * @param matchCond the SVInstantiations/Constraint that are
     * required because of formerly matchings
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @param userConstraint current user constraint (which is in some
     * situations used to find instantiations)
     * @return the new MatchConditions needed to match template with
     * term , if possible, null otherwise
     */
    protected MatchConditions match(Term            term,
				    Term            template,
				    boolean         ignoreUpdates,
				    MatchConditions matchCond,
				    JavaInfo        javaInfo,
				    Constraint      userConstraint) {
	return matchHelp(term, template, ignoreUpdates, matchCond, javaInfo, userConstraint);
    }



    /**
     * same as the method above but with ignoreUpdates always false
     * @param term the Term that has to be matched
     * @param template the Term that is checked if it can match term
     * @param matchCond the SVInstantiations/Constraint that are
     * required because of formerly matchings
     * @param javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @param userConstraint current user constraint (which is in some
     * situations used to find instantiations)
     * @return the new MatchConditions needed to match template with
     * term , if possible, null otherwise
     */
    protected MatchConditions match(Term            term,
				    Term            template,
				    MatchConditions matchCond,
				    JavaInfo        javaInfo,
				    Constraint      userConstraint) {
	return match(term, template, false, matchCond, javaInfo, userConstraint);
    }


    /**
     * checks if the conditions for a correct instantiation are fulfilled
     * @param var the template Variable to be instantiated
     * @param term the Term to be mapped to var
     * @param svInst the SVInstantiations that are already known to be needed 
     * @param boundVars the ListOfQuantifiableVariable in whose
     * bounding area var is  
     * @return true iff conditions are fulfilled
     */
    public boolean checkVariableConditions(SchemaVariable var, Term term,
					   SVInstantiations svInst) {

	if (!(term.op() instanceof QuantifiableVariable)) {
	    if (var.isVariableSV() || varIsBound(var) ||
		varDeclaredNotFree(var) ||
		(term.freeVars() != SetAsListOfQuantifiableVariable.EMPTY_SET 
		 && !correctInstantiationAreaOccurrence(var))
		) {
		// match(x) is not a variable, but the
		// corresponding template variable is bound
		// or declared non free (so it has to be
		// matched to a variable) or match(x) contains
		// free variables and the corresponding
		// template variable occurs in more than one
		// instantiation area			
		return false;
	    }
	} else {
	    if (varIsBound(var)) {
		IteratorOfInScopePair boundInIt = varInScope.iterator();
		while (boundInIt.hasNext()) {
		    InScopePair varpair = boundInIt.next();
		    if (varpair.second() == var) {
			if (term.equals(svInst.getInstantiation
					((SchemaVariable)varpair.first()))) {

			    return false;			    
			}
		    }
		}
	    }
	    if (!correctInstantiationAreaOccurrence(var)) {
		return false;

	    }
	}
	
	// check generic conditions
	for (Iterator i = variableConditions.iterator(); i.hasNext(); ) {
	    VariableCondition vc = (VariableCondition) i.next();
	    if (!vc.check(var, term, svInst)) {
		return false;
	    }
	}

	return true;	
    }


    private MatchConditions addInstantiation(Term term, SchemaVariable sv, 
					     MatchConditions matchCond) {   
	Term t = (Term) matchCond.getInstantiations ().getInstantiation(sv);	   		
	if (t != null) {
	    Constraint c = matchCond.getConstraint ().unify ( t, term );
	    if ( !c.isSatisfiable () ) {
		return null; //FAILED;
	    } else {
		return matchCond.setConstraint ( c );
	    }
	}	

	// no former matching found
	if (!checkVariableConditions(sv, term, 
				     matchCond.getInstantiations ())) {
	    return null; //FAILED;
	}
	try {
	    return matchCond.setInstantiations ( matchCond.getInstantiations ()
						 .add(sv, term) );
	} catch ( IllegalInstantiationException e ) {}

	return null;
    }
    

    /** 
     * Tests if the given SchemaVariable sv found in the logic part of a formula
     * (not inside a java block) matches the given term and returns an updated
     * version of the matchconditions if a match is possible, otherwise null is
     * returned 
     * (marked as final to help the compiler inlining methods)
     * @param term the Term that should match the given SchemaVariable sv
     * @param sv the SchemaVariable that is tested if it can stand for the given
     * term 
     * @param matchCond the MatchConditions that have to be fulfilled
     * @return null if no match is possible (e.g. already matched with a
     * different term or sorts do not fit), otherwise the necessary match
     * conditions are returned (usually the pair (sv, term) has been added to
     * the match conditions.
     */
    private final MatchConditions matchSVInLogic(Term term, SchemaVariable sv, 
						 MatchConditions matchCond) {
	if (sv.isProgramSV()) {
	    if (((ProgramSVSort)sv.sort()).canStandFor(term)) {
		return addInstantiation(term, sv, matchCond);
	    } else {
		return null; //FAILED
	    }
	} else {
	    return addInstantiation(term, sv, matchCond);
	}	 
    }

    /**
     * tries to match the bound variables of the given term against the one
     * described by the template
     * @param term the Term whose bound variables are matched against the
     * JavaBlock of the template
     * (marked as final to help the compiler inlining methods)
     * @param template the Term whose bound variables are the template that have
     * to be matched
     * @param matchCond the MatchConditions that has to be paid respect when
     * trying to match
     * @return the new matchconditions if a match is possible, otherwise null
     */
    private final MatchConditions matchBoundVariables(Term term, Term template, 
						      MatchConditions matchCond) {
	for (int j=0; j<term.arity(); j++) {		
	    
	    ArrayOfQuantifiableVariable bound    = term.varsBoundHere(j);
	    ArrayOfQuantifiableVariable tplBound = template.varsBoundHere(j); 
	    
	    if (bound.size() != tplBound.size()) {
		return null; //FAILED
	    }
	    
	    for (int i=0; i<bound.size(); i++) {		
		if (tplBound.getQuantifiableVariable(i) instanceof SchemaVariable) {			
		    //			    if (!tplBound.getQuantifiableVariable(i)
		    //				.sort().extendsTrans
		    //				(bound.getQuantifiableVariable(i).sort())) {
		    //			    return null;  //FAILED
		    //            		}
		    Term foundMapping 
			= (Term)matchCond.getInstantiations ().getInstantiation
			((SchemaVariable)
			 tplBound.getQuantifiableVariable(i));
		    if (foundMapping != null && 
			((QuantifiableVariable)foundMapping.op()) 
			!= bound.getQuantifiableVariable(i)) {
			return null; //FAILED;			    

		    } else if (foundMapping == null) {
			Term varTerm
			    = TermFactory.DEFAULT.createVariableTerm
			    ((LogicVariable)bound
			     .getQuantifiableVariable(i));
			if (checkVariableConditions((SchemaVariable)
						    tplBound.getQuantifiableVariable(i),
						    varTerm,
						    matchCond.getInstantiations ())) {
			    try {
				matchCond = matchCond
				    .setInstantiations ( matchCond.getInstantiations ()
							 .add((SchemaVariable)tplBound
							      .getQuantifiableVariable(i), 
							      varTerm) );
			    } catch ( IllegalInstantiationException e ) {
				return null; //FAILED;
			    }
			}
		    } 
		} else {
		    if (tplBound.getQuantifiableVariable(i) 
			!= bound.getQuantifiableVariable(i)) {
			return null; //FAILED;
		    }
		}		    
	    }
	}
	return matchCond;
    }

    /**
     * returns the matchconditions that are required if the java block of the
     * given term matches the schema given by the template term or null if no
     * match is possible
     * (marked as final to help the compiler inlining methods)
     * @param term the Term whose JavaBlock is matched against the JavaBlock of
     * the template
     * @param template the Term whose JavaBlock is the template that has to
     * be matched
     * @param matchCond the MatchConditions that has to be paid respect when
     * trying to match the JavaBlocks
     * @param javaInfo the JavaInfo object encapsulating information about the
     * program context
     * @return the new matchconditions if a match is possible, otherwise null
     */
    private final MatchConditions matchJavaBlock(Term term, Term template, 
						 MatchConditions matchCond,
						 JavaInfo javaInfo) {

	if (term.javaBlock().isEmpty()) {
	    if (!template.javaBlock().isEmpty()){
		return null; //FAILED
	    }
	} else { //both javablocks!=null
	    SVInstantiations newSVInst;
	    if (!(template.javaBlock().program() instanceof ContextTerm)) {	 
		newSVInst = matchJavaProgram (term.javaBlock(), 
					      new JavaIterator(term.javaBlock().program()), 
					      PosInProgram.TOP,  
					      new JavaIterator(template.javaBlock().program()),
					      matchCond.getInstantiations (),
					      javaInfo);
	    } else {
		newSVInst = matchJavaProgram(term.javaBlock(),
					     (ContextTerm)
					     template.javaBlock().program(),
					     matchCond.getInstantiations (),
					     javaInfo);
	    }
	    if (newSVInst == null) {
		return null; //FAILED
	    } else {
		return matchCond.setInstantiations ( newSVInst );
	    }
	}

	return matchCond;
    }
    
    
    /** returns a SVInstantiations object with the needed SchemaVariable to Term
     * mappings to match the given Term template to the Term term or
     * null if no matching is possible.
     * (marked as final to help the compiler inlining methods)
     * @param term the Term the Template should match
     * @param template the Term tried to be instantiated so that it matches term
     * @param ignoreUpdates a boolean if set to true updates will be ignored as 
     * e.g. wanted if an if-sequent is matched
     * @param matchCond the SVInstantiations/Constraint that are
     * required because of formerly matchings
     * @param userConstraint current user constraint (which is in some
     * situations used to find instantiations)
     * @return the new MatchConditions needed to match template with
     * term , if possible, null otherwise
     *
     * PRECONDITION: matchCond.getConstraint ().isSatisfiable ()
     */
    private MatchConditions matchHelp(Term             term,
				      Term             template, 
				      boolean          ignoreUpdates,
				      MatchConditions  matchCond,
				      JavaInfo         javaInfo,
				      Constraint       userConstraint) {

	if (template.op() instanceof SchemaVariable) {
	    return matchSVInLogic(term, (SchemaVariable)template.op(), matchCond);
	} else {
	    if ( template.op () instanceof Metavariable ) {

		Constraint c = matchCond.getConstraint ().unify ( term, template );
		return c.isSatisfiable () ? matchCond.setConstraint ( c ) : null;

	    } else if ( term.op () instanceof Metavariable ) {

		// "term" is a metavariable, "template" neither a
		// schemavariable nor a metavariable
		Term t;

		// try to instantiate "term" according to the current constraint
		if ( ( matchCond.getConstraint () instanceof EqualityConstraint ) &&
		     ( t = ((EqualityConstraint)matchCond.getConstraint ())
		       .getDirectInstantiation ( (Metavariable)term.op () ) ) != null )

		    return matchHelp ( t, template, ignoreUpdates, matchCond, javaInfo, userConstraint );

		else { 

		    // try to instantiate "term" according to the current user constraint
		    if ( ( userConstraint instanceof EqualityConstraint ) &&
			 ( t = ((EqualityConstraint)userConstraint)
			   .getDirectInstantiation ( (Metavariable)term.op () ) ) != null ) {

			Constraint c  = matchCond.getConstraint ().unify ( term, t );
			if ( c.isSatisfiable () ) {
			    MatchConditions mc =
				matchHelp ( t, template, ignoreUpdates,
					    matchCond.setConstraint ( c ),
					    javaInfo, userConstraint );
			    if ( mc != null )
				return mc;
			}
		    }
		    
		    // try to instantiate uninstantiated SVs by
		    // creating new metavariables or bound
		    // logicvariables
		    SyntacticalReplaceVisitor srVisitor;
		    try {
			srVisitor = new SyntacticalReplaceVisitor
			    ( javaInfo,
			      matchCond.getInstantiations (),
			      matchCond.getConstraint (),
			      true,
			      "" + term.op ().name () );
			template.execPostOrder ( srVisitor );
		    } catch ( IllegalInstantiationException e ) {
			return null; //FAILED;
		    }

		    if ( srVisitor.getNewMetavariables () == null )
			return null; //FAILED;

		    matchCond = matchCond.setInstantiations ( srVisitor.getSVInstantiations () )
			.setNewMetavariables ( matchCond.getNewMetavariables ()
					       .union ( srVisitor.getNewMetavariables () ) );

		    Constraint c = matchCond.getConstraint ().unify ( term, srVisitor.getTerm () );

		    return c.isSatisfiable () ? matchCond.setConstraint ( c ) : null;
		}

	    }   else if (ignoreUpdates && term.op() instanceof SimultaneousUpdateOp && 
			 !(template.op() instanceof SimultaneousUpdateOp)) {
		// updates can be ignored
		matchCond = matchCond
		    .setInstantiations ( matchCond.getInstantiations ().
					 addUpdate (term) );
		return matchHelp(term.sub(term.arity()-1), template,
				 true, matchCond, javaInfo, userConstraint);
	    }  else if ( term    .op () instanceof SortDependingSymbol &&
			 template.op () instanceof SortDependingSymbol ) {
		if ( template.op () != term.op () ) {
		    GenericSortCondition c = GenericSortCondition.createCondition
			( (SortDependingSymbol)template.op (),
			  (SortDependingSymbol)term    .op () );

		    if ( c == null ) {
			return null; //FAILED;
		    } else {
			try {
			    matchCond = matchCond
				.setInstantiations ( matchCond.getInstantiations ().
						     add ( c ) );
			} catch ( SortException e ) {
			    return null; //FAILED;
			}
		    }
		}
	    } else if (term.op() != template.op()) {	       		
		return null; //FAILED;
	    }
    
	    //match java blocks:
	    matchCond = matchJavaBlock(term, template, matchCond, javaInfo);
	    if (matchCond == null) { 
		return null;  //FAILED
	    }


	    //match bound variables:
	    matchCond = matchBoundVariables(term ,template, matchCond);
	    if (matchCond == null) { 
		return null;  //FAILED
	    }

	    
	    for (int i = 0; i < term.arity(); i++) {
		matchCond = matchHelp(term.sub(i), template.sub(i), false,
				      matchCond, javaInfo, userConstraint);
		if (matchCond == null) {
		    return null; //FAILED
		} 
	    }	    
	    return matchCond;	
	}
    }
    

    /**
     * Match the given template (which is probably a formula of the if
     * sequence) against a list of constraint formulas (probably the
     * formulas of the antecedent or the succedent), starting with the
     * given instantiations and constraint p_matchCond.
     * @param p_toMatch list of constraint formulas to match p_template to
     * @param p_template template formula as in "match"
     * @param p_matchCond already performed instantiations
     * @param p_javaInfo the JavaInfo object encapsulating information
     * about the java datastructures like (static)types etc.
     * @return Two lists (in an IfMatchResult object), containing the
     * the elements of p_toMatch that could successfully be matched
     * against p_template, and the corresponding MatchConditions.
     */
    public IfMatchResult matchIf ( IteratorOfIfFormulaInstantiation p_toMatch,
				   Term                             p_template,
				   MatchConditions                  p_matchCond,
				   JavaInfo                         p_javaInfo,
				   Constraint                       p_userConstraint ) {
	ListOfIfFormulaInstantiation     resFormulas =
	    SLListOfIfFormulaInstantiation.EMPTY_LIST;
	ListOfMatchConditions            resMC       =
	    SLListOfMatchConditions       .EMPTY_LIST;

	Term                             updateFormula;
	if ( p_matchCond.getInstantiations ().getUpdateContext() ==
	     SLListOfUpdatePair.EMPTY_LIST )
	    updateFormula = p_template;
	else
	    updateFormula = TermFactory.DEFAULT
		.createUpdateTerm(p_matchCond.getInstantiations ()
				  .getUpdateContext(),
				  p_template);

	IfFormulaInstantiation           cf;
	Constraint                       newConstraint;
	MatchConditions                  newMC;

	while ( p_toMatch.hasNext () ) {
	    cf            = p_toMatch.next ();
	    newConstraint = p_matchCond.getConstraint ()
		.join ( cf.getConstrainedFormula ().constraint () );

	    if ( newConstraint.isSatisfiable () ) {
		newMC = match ( cf.getConstrainedFormula ().formula (),
				updateFormula,
				false,
				p_matchCond.setConstraint ( newConstraint ),
				p_javaInfo,
				p_userConstraint );
		if ( newMC != null ) {
		    resFormulas = resFormulas.prepend ( cf );
		    resMC       = resMC      .prepend ( newMC );
		}
	    }
	}

	return new IfMatchResult ( resFormulas, resMC );
    }


    /**
     * Match the whole if sequent using the given list of
     * instantiations of all if sequent formulas, starting with the
     * instantiations given by p_matchCond.
     * PRECONDITION: p_toMatch.size () == ifSequent ().size ()
     * @return resulting MatchConditions or null if the given list
     * p_toMatch does not match
     */
    public MatchConditions matchIf ( IteratorOfIfFormulaInstantiation p_toMatch,
				     MatchConditions                  p_matchCond,
				     JavaInfo                         p_javaInfo,
				     Constraint                       p_userConstraint ) {

	IteratorOfConstrainedFormula     itIfSequent   = ifSequent () .iterator ();

	ListOfMatchConditions            newMC;

	while ( itIfSequent.hasNext () ) {
	    newMC = matchIf ( ( SLListOfIfFormulaInstantiation.EMPTY_LIST
				.prepend ( p_toMatch.next () ).iterator () ),
			      itIfSequent.next ().formula (),
			      p_matchCond,
			      p_javaInfo,
			      p_userConstraint ).getMatchConditions ();

	    if ( newMC == SLListOfMatchConditions.EMPTY_LIST )
		return null;

	    p_matchCond = newMC.head ();
	}

	return p_matchCond;

    }


    /** returns the name of the Taclet
     */
    public Name name() {
	return name;
    } 
    

    /** returns the display name of the taclet, or, if not specified -- 
     *  the canonical name
     */
    public String displayName() {
	return displayName;
    }
    
    
    public String helpText() {
       return helpText;
    }


   /** returns the if-sequence of the application part of the Taclet.
     */
    public Sequent ifSequent() {
	return ifSequent;
    } 
    
    /** returns an iterator over the variables that are new in the Taclet. 
     */
    public ListOfNewVarcond varsNew() {
	return varsNew;
    } 

    
    /** returns an iterator over the variable pairs that indicate that are 
     * new in the Taclet. 
     */
    public IteratorOfNotFreeIn varsNotFreeIn() { 
	return varsNotFreeIn.iterator();
    } 

    public IteratorOfNewDependingOn varsNewDependingOn() { 
	return varsNewDependingOn.iterator();
    } 

    /** returns true iff the Taclet is to be applied recursive
     */
    public boolean recursive() {
	return recursive;
    } 
    
    /** returns the Constraint under which the Taclet is
     * valid */
    public Constraint constraint() {
	return constraint;
    }


    /** returns an iterator over the goal descriptions.
     */
    public ListOfTacletGoalTemplate goalTemplates() {
	return goalTemplates;
    } 

    /** returns an iterator over the heuristics. */
    public IteratorOfHeuristic heuristics() {
	return heuristics.iterator();
    } 

    public ListOfHeuristic getHeuristics() {
	return heuristics;
    }

    /** returns true iff the Taclet is to be applied only noninteractive
     */
    public boolean noninteractive() {
	return noninteractive;
    }


    public MapFromSchemaVariableToTacletPrefix prefixMap() {
	return prefixMap;
    }

    /** 
     * returns true if one of the goal templates is a replacewith. Already
     * computed and cached by method cacheMatchInfo
     */
    public boolean hasReplaceWith() {
	return hasReplaceWith;
    }

    /**
     * returns the computed prefix for the given schemavariable. The prefix of a
     * schemavariable is used to determine if an instantiation is correct, in
     * more detail: it mainly contains all variables that can appear free in an
     * instantiation of the schemvariable sv (rewrite taclets have some special
     * handling, see paper of M. Giese and comment of method isContextInPrefix). 
     * @param sv the Schemavariable 
     * @return prefix of schema variable sv
     */
    public TacletPrefix getPrefix(SchemaVariable sv) {
	return prefixMap.get(sv);
    }

    /** 

    /**
     * returns true iff a context flag is set in one of the entries in
     * the prefix map. Is cached after having been called
     * once. __OPTIMIZE__ is caching really necessary here?
     *
     * @return true iff a context flag is set in one of the entries in
     * the prefix map.
     */
    public boolean isContextInPrefix() {
	if (contextInfoComputed) {
	    return contextIsInPrefix;
	}
	contextInfoComputed=true;
	IteratorOfTacletPrefix it=prefixMap().valueIterator();
	while (it.hasNext()) {
	    if (it.next().context()) {
		contextIsInPrefix=true;
		return true;
	    }
	}
	contextIsInPrefix=false;
	return false;
    }

    /** 
     * return true if o is a taclet of the same name 
     */
    public boolean equals(Object o) {
	if ( ! ( o instanceof Taclet ) )
	    return false;

	return name.equals(((Taclet) o).name);
    }


    /**
     * creates the initial instantiation of the SchemaVariables 
     * @return the initial SVInstantiations 
     */
    public SVInstantiations createInitialInstantiation() {
	return SVInstantiations.EMPTY_SVINSTANTIATIONS;
    }


    /** 
     * a new term is created by replacing variables of term whose replacement is
     * found in the given SVInstantiations 
     * @param term the Term the syntactical replacement is performed on
     * @param svInst the SVInstantiations containing the instantiations needed
     * for the replacement
     * @return the created Term 
     */
    protected Term syntacticalReplace(Term term,
				      JavaInfo javaInfo,
				      SVInstantiations svInst) {	

	//	Debug.out("taclet: SYNTACTICALREPLACE  (term, svInst):", term, svInst);
	SyntacticalReplaceVisitor srVisitor = 
	    new SyntacticalReplaceVisitor(javaInfo, svInst);
	term.execPostOrder(srVisitor);

	return srVisitor.getTerm();
    }
    

    /**
     * adds ConstrainedFormula to antecedent or succedent depending on
     * position information or the boolean antec 
     * contrary to "addToPos" frm will not be modified
     * @param frm the formula that should be added
     * @param goal the Goal that knows the node the formulae have to be added
     * @param pos the PosInOccurrence describing the place in the sequent
     * @param antec boolean true(false) if elements have to be added to the
     * antecedent(succedent) (only looked at if pos == null)
     */
    private void addToPosWithoutInst(ConstrainedFormula frm,
				     Goal goal,			  
				     PosInOccurrence pos,
				     boolean antec) {
	if (pos != null) {
	    goal.addFormula(frm, pos);
	} else {
	    // cf : formula to be added , 1. true/false: antec/succ,
	    // 2. true: at head 
	    goal.addFormula(frm, antec, true);		
	}	    
    }


    /** 
     * the given constrained formula is instantiated and then
     * the result (usually a complete instantiated formula) is returned.
     * @param schemaFormula the ConstrainedFormula to be instantiated
     * @param javaInfo the JavaInfo object carrying ja related information
     * @param matchCond the MatchConditions object with the instantiations of
     * the schemavariables, constraints etc.
     * @return the as far as possible instantiated ConstrainedFormula
     */
    private ConstrainedFormula 
	instantiateReplacement(ConstrainedFormula schemaFormula,
			       JavaInfo           javaInfo,
			       MatchConditions    matchCond) { 

	ConstrainedFormula cf;

	if (matchCond.getInstantiations ().getUpdateContext() == SLListOfUpdatePair.EMPTY_LIST) {
	    cf = new ConstrainedFormula
		(syntacticalReplace(schemaFormula.formula(), 
				    javaInfo,
				    matchCond.getInstantiations ()),
		 matchCond.getConstraint() );
	} else {
	    Term formula  = schemaFormula.formula();
	    Term replaced = syntacticalReplace(formula, javaInfo,
					       matchCond.getInstantiations ());

	    Term updated  = TermFactory.DEFAULT
		.createUpdateTerm(matchCond.getInstantiations ().getUpdateContext(), replaced);
		
	    cf = new ConstrainedFormula(updated, matchCond.getConstraint());

	}
	
	return cf;
    }
						      
    

    /**
     * replaces the constrained formula at the given position with the first
     * formula in the given semisequent and adds possible other formulas of the
     * semisequent starting at the position
     * @param semi the Semisequent with the the ConstrainedFormulae to be added
     * @param goal the Goal that knows the node the formulae have to be added
     * @param pos the PosInOccurrence describing the place in the sequent
     * @param antec boolean true(false) if elements have to be added to the
     * antecedent(succedent) (only looked at if pos == null)
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst the necessary instantiations 
     */
    protected void replaceAtPos(Semisequent semi,
				Goal goal,
				PosInOccurrence pos,
				JavaInfo javaInfo, 
				MatchConditions matchCond) {

	ListOfConstrainedFormula replacements = SLListOfConstrainedFormula.EMPTY_LIST;
	IteratorOfConstrainedFormula it = semi.iterator();	
	
	while (it.hasNext()) {
	    replacements = replacements.append
		(instantiateReplacement(it.next(), javaInfo, matchCond));	    
	}
		
	goal.changeFormula(replacements, pos);
    }

    /**
     * adds ConstrainedFormula to antecedent or succedent depending on
     * position information or the boolean antec 
     * @param semi the Semisequent with the the ConstrainedFormulae to be added
     * @param goal the Goal that knows the node the formulae have to be added
     * @param pos the PosInOccurrence describing the place in the sequent
     * @param antec boolean true(false) if elements have to be added to the
     * antecedent(succedent) (only looked at if pos == null)
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst the necessary instantiations 
     */
    private void addToPos ( Semisequent semi,
			    Goal goal,			  
			    PosInOccurrence pos,
			    boolean antec,
			    JavaInfo javaInfo, 
			    MatchConditions matchCond ) 
    {
	ListOfConstrainedFormula replacements = SLListOfConstrainedFormula.EMPTY_LIST;
	IteratorOfConstrainedFormula it = semi.iterator();	
	
	while (it.hasNext()) {
	    replacements = replacements.append
		(instantiateReplacement(it.next(), javaInfo, matchCond));	    
	}
	
	if (pos!=null) {
	    goal.addFormula(replacements, pos);
	} else {
	    goal.addFormula(replacements, antec, true);
	}
    }

    /**
     * adds ConstrainedFormula to antecedent depending on
     * position information (if none is handed over it is added at the
     * head of the antecedent). Of course it has to be ensured that
     * the position information describes one occurrence in the
     * antecedent of the sequent.
     * @param semi the Semisequent with the the ConstrainedFormulae to be added
     * @param goal the Goal that knows the node the formulae have to be added
     * @param pos the PosInOccurrence describing the place in the
     * sequent or null for head of antecedent
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst the necessary instantiations 
     */
    protected void addToAntec(Semisequent semi,
			      Goal goal,
			      PosInOccurrence pos,
			      JavaInfo javaInfo, 
			      MatchConditions matchCond) { 
	addToPos(semi, goal, pos, true, javaInfo, matchCond);
    }

    /**
     * adds ConstrainedFormula to succedent depending on
     * position information (if none is handed over it is added at the
     * head of the succedent). Of course it has to be ensured that
     * the position information describes one occurrence in the
     * succedent of the sequent.
     * @param semi the Semisequent with the the ConstrainedFormulae to be added
     * @param goal the Goal that knows the node the formulae have to be added
     * @param pos the PosInOccurrence describing the place in the
     * sequent or null for head of antecedent
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst the necessary instantiations 
     */
    protected void addToSucc(Semisequent semi,
			     Goal goal,
			     PosInOccurrence pos,
			     JavaInfo javaInfo, 
			     MatchConditions matchCond) {
	addToPos(semi, goal, pos, false, javaInfo, matchCond);
    }

    protected abstract Taclet setName(String s);
    
    protected Taclet setName(String s, TacletBuilder b) {	
	b.setTacletGoalTemplates(goalTemplates());
	b.setHeuristics(getHeuristics());
	b.setName(new Name(s));
	b.setDisplayName(displayName());
	b.setIfSequent(ifSequent());
	b.setConstraint(constraint());
	b.addVarsNew(varsNew());
	b.addVarsNotFreeIn(varsNotFreeIn);
	return b.getTaclet();
    }

    /**
     * adds the given rules (i.e. the rules to add according to the Taclet goal
     * template to the node of the given goal
     * @param rules the rules to be added
     * @param goal the goal describing the node where the rules should be added
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param svInst the necessary instantiations of the SchemaVariables
     */
    protected void applyAddrule(ListOfTaclet rules, Goal goal, 
				JavaInfo javaInfo,
				MatchConditions matchCond) {
                                
	IteratorOfTaclet it = rules.iterator();
	while (it.hasNext()) {
	    Taclet tacletToAdd = it.next(); 
	    String uniqueTail=""; // we need to name the new taclet uniquely
/*
            TacletGoalTemplate replacewithCandidate = null;
	    IteratorOfTacletGoalTemplate actions = 
               tacletToAdd.goalTemplates().iterator();
            while (actions.hasNext()) {
               replacewithCandidate = actions.next();
               if (replacewithCandidate instanceof RewriteTacletGoalTemplate)
                  break;
            }
            if ((replacewithCandidate instanceof RewriteTacletGoalTemplate) &&
                (tacletToAdd instanceof FindTaclet)) {
                // we have _both_ FIND and REPLACEWITH
                Term find = ((FindTaclet)tacletToAdd).find();
                Term replwith = 
                   ((RewriteTacletGoalTemplate)
                      replacewithCandidate).replaceWith();
                      
                SyntacticalReplaceVisitor visitor = // now instantiate them!
                   new SyntacticalReplaceVisitor(javaInfo, matchCond.getInstantiations ());
                visitor.visit(find);
                uniqueTail = "_" + visitor.getTerm();
                visitor = new SyntacticalReplaceVisitor(javaInfo, matchCond.getInstantiations ());
                visitor.visit(replwith);
                uniqueTail += "_" + visitor.getTerm();
	    }
*/
            if ("".equals(uniqueTail)) { // otherwise just number it
               uniqueTail = "_"+AUTONAME+
                  goal.node().proof().getUniqueTacletNr();
            }

            tacletToAdd=tacletToAdd.setName(tacletToAdd.name()+uniqueTail);


	    // the new Taclet may contain variables with a known
	    // instantiation. These must be used by the new Taclet and all
	    // further rules it contains in the addrules-sections. Therefore all
	    // appearing (including the addrules) SchemaVariables have to be
	    // collected, then it is looked if an instantiation is known and if
	    // positive the instantiation is memorized. At last the Taclet with
	    // its required instantiations is handed over to the goal, where a
	    // new TacletApp should be built including the necessary instantiation
	    // information

	    SVInstantiations neededInstances = SVInstantiations.
		EMPTY_SVINSTANTIATIONS.addUpdateList
		(matchCond.getInstantiations ().getUpdateContext());
	    TacletVariableCollector collector = new
		TacletVariableCollector(); 
	    collector.visit(tacletToAdd, true);// true, because
	                                     // descend into
					     // addrules
	    IteratorOfSchemaVariable svIt = collector.varIterator();
	    while (svIt.hasNext()) {
		SchemaVariable sv = svIt.next();
		if (matchCond.getInstantiations ().isInstantiated(sv)) {
		    neededInstances = neededInstances.add
			(sv, matchCond.getInstantiations ().getInstantiationEntry(sv));
		} 
	    }

	    {
		ListOfGenericSortCondition     cs  =
		    matchCond.getInstantiations ()
		    .getGenericSortInstantiations ().toConditions ();
		IteratorOfGenericSortCondition cit =
		    cs.iterator ();

		while ( cit.hasNext () )
		    neededInstances = neededInstances.add ( cit.next () );
	    }

	    goal.addTaclet(tacletToAdd, neededInstances, matchCond.getConstraint ());
	}
    }

    protected void applyAddProgVars(SetOfSchemaVariable pvs, Goal goal,
				    MatchConditions matchCond) {
	IteratorOfSchemaVariable it = pvs.iterator();
	while (it.hasNext()) {
	    ProgramVariable inst
		= (ProgramVariable)matchCond.getInstantiations ().getInstantiation(it.next());
	    goal.addProgramVariable(inst);
	}
    }
    

    /** the rule is applied to the given goal using the
     * information of rule application. 
     * @param goal the goal that the rule application should refer to.
     * @param javaInfo the JavaInfo encapsulating all java information
     * @param tacletApp the rule application that is executed.
     * @param proveIfSequent a boolean that is true if a goal has to
     * be created in order to prove that the sequent in the if part of
     * the Taclet can be derived and therefore the application of the
     * rule is correct (normally true when the if-sequent does not match)
     * @return List of the goals created by the rule which have to be
     * proved. If this is a close-goal-taclet ( this.closeGoal () ),
     * the first goal of the return list is the goal that should be
     * closed (with the constraint this taclet is applied under).
     */
    public abstract ListOfGoal apply(Goal goal, JavaInfo javaInfo, 
				     RuleApp tacletApp);


    /**
     * Search for formulas within p_list that have to be proved by an
     * explicit if goal, i.e. elements of type IfFormulaInstDirect.
     * @return an array with two goals if such formulas exist (the
     * second goal is the if goal), otherwise an array consisting of
     * the single element p_goal
     */
    protected ListOfGoal checkIfGoals ( Goal                         p_goal,
					ListOfIfFormulaInstantiation p_list,
					MatchConditions              p_matchCond,
					int                          p_numberOfNewGoals ) {
	ListOfGoal     res    = null;
	IteratorOfGoal itGoal;

	// proof obligation for the if formulas
	Term           ifObl  = null;

	// always create at least one new goal
	if ( p_numberOfNewGoals == 0 )
	    p_numberOfNewGoals = 1;

	if ( p_list != null ) {
	    IteratorOfIfFormulaInstantiation it     = p_list.iterator ();
	    IfFormulaInstantiation           inst;
	    int                              i      = ifSequent ().antecedent ().size ();
	    Term                             ifPart;

	    while ( it.hasNext () ) {
		inst = it.next ();
		if ( !( inst instanceof IfFormulaInstSeq ) ) {
		    // build the if obligation formula
		    ifPart = inst.getConstrainedFormula ().formula ();

		    // negate formulas of the if succedent
		    if ( i <= 0 )
			ifPart = TermFactory.DEFAULT.createJunctorTerm
			    ( Op.NOT, ifPart );		    

		    if ( res == null ) {
			res   = p_goal.split( p_numberOfNewGoals + 1 );
			ifObl = ifPart;
		    } else
			ifObl = TermFactory.DEFAULT.createJunctorTerm
			    ( Op.AND, ifObl, ifPart );
		    
		    // UGLY: We create a flat structure of the new
		    // goals, thus the if formulas have to be added to
		    // every new goal
		    itGoal = res.iterator ();
		    p_goal = itGoal.next ();
		    while ( itGoal.hasNext () ) {
			addToPosWithoutInst ( inst.getConstrainedFormula (),
					      p_goal,
					      null,
				   ( i > 0 ) ); // ( i > 0 ) iff inst is formula
			                        // of the antecedent
			p_goal = itGoal.next ();
		    }
		}

		--i;
	    }
	}

	if ( res == null )
	    res = p_goal.split ( p_numberOfNewGoals );
	else {
	    // find the sequent the if obligation has to be added to
	    itGoal = res.iterator ();
	    p_goal = itGoal.next ();
	    while ( itGoal.hasNext () )
		p_goal = itGoal.next ();

	    addToPosWithoutInst ( new ConstrainedFormula ( ifObl,
							   Constraint.BOTTOM ),
				  p_goal,
				  null,
				  false );
	}
	
	return res;
    }

    /**
     * Restrict introduced metavariables to the subtree
     */
    protected void setRestrictedMetavariables ( Goal            p_goal,
						MatchConditions p_matchCond ) {
	IteratorOfMetavariable it =
	    p_matchCond.getNewMetavariables ().iterator ();
	while ( it.hasNext () )
	    p_goal.addRestrictedMetavariable ( it.next () );
    }

    /**
     * returns the set of schemavariables of the taclet's if-part
     * @return Set of schemavariables of the if part
     */
    protected SetOfSchemaVariable getIfVariables () {
	// should be synchronized
	if ( ifVariables == null ) {
	    TacletVariableCollector svc = new TacletVariableCollector ();
	    svc.visit( ifSequent () );
	    
	    ifVariables                 = SetAsListOfSchemaVariable.EMPTY_SET;
	    IteratorOfSchemaVariable it = svc.varIterator ();
	    while ( it.hasNext () )
		ifVariables = ifVariables.add ( it.next () );
	}

	return ifVariables;
    }


    /**
     * @return set of schemavariables of the if and the (optional)
     * find part
     */
    public abstract SetOfSchemaVariable getIfFindVariables ();


    /**
     * Find a schema variable that could be used to choose a name for
     * an instantiation (a new variable or constant) of "p"
     * @return a schema variable that is substituted by "p" somewhere
     * in this taclet (that is, these schema variables occur as
     * arguments of a substitution operator)
     */
    public SchemaVariable getNameCorrespondent ( SchemaVariable p ) {
	// should be synchronized
	if ( svNameCorrespondences == null ) {
	    SVNameCorrespondenceCollector c =
		new SVNameCorrespondenceCollector ();
	    c.visit ( this, true );
	    svNameCorrespondences = c.getCorrespondences ();
	}

	return svNameCorrespondences.get ( p );
    }


    StringBuffer toStringIf(StringBuffer sb) {
	if (!ifSequent.empty()) {
	    sb = sb.append("if (").append(ifSequent).append(") ");
	    sb = sb.append("\n");
	}
	return sb;
    }

    StringBuffer toStringVarCond(StringBuffer sb) {
	IteratorOfNewVarcond itVarsNew=varsNew().iterator();
	IteratorOfNotFreeIn itVarsNotFreeIn=varsNotFreeIn();
	if (itVarsNew.hasNext() || itVarsNotFreeIn.hasNext()) {
	    sb = sb.append("varcond(");
	    while (itVarsNew.hasNext()) {
		sb=sb.append(itVarsNew.next());
		if (itVarsNew.hasNext() || itVarsNotFreeIn.hasNext())
		    sb=sb.append(", "); 
	    }
	    while (itVarsNotFreeIn.hasNext()) {
		NotFreeIn pair=itVarsNotFreeIn.next();
		sb=sb.append(pair.first());
		sb=sb.append(" not free in ").append(pair.second());	 
		if (itVarsNotFreeIn.hasNext()) sb=sb.append(", ");
	    }	    
	    sb=sb.append(")\n");	    
	}
	return sb;
    }

    StringBuffer toStringGoalTemplates(StringBuffer sb) {
	IteratorOfTacletGoalTemplate itGoalTempl=goalTemplates().iterator();
	while (itGoalTempl.hasNext()) {
	    sb=sb.append(itGoalTempl.next());
	    if (itGoalTempl.hasNext()) sb = sb.append(";");
	    sb = sb.append("\n");
	}
	return sb;
    }

    StringBuffer toStringHeuristics(StringBuffer sb) {
	IteratorOfHeuristic itHeu=heuristics();
	if (itHeu.hasNext()) {
	    sb=sb.append("heuristics(");
	    while (itHeu.hasNext()) {
		sb=sb.append(itHeu.next());
		if (itHeu.hasNext()) sb=sb.append(", ");
	    }
	    sb=sb.append(")");
	}
	return sb;
    }

    StringBuffer toStringAttribs(StringBuffer sb) {
	if (noninteractive()) sb = sb.append(" noninteractive");
	if (recursive()) sb = sb.append(" recursive");
	return sb;
    }
    
    /**
     * returns a representation of the Taclet as String
     * @return string representation
     */
    public String toString() {
	if (tacletAsString == null) {
	    StringBuffer sb=new StringBuffer();
	    sb = sb.append(name()).append(" {\n");
	    sb = toStringIf(sb);
	    sb = toStringVarCond(sb);
	    sb = toStringGoalTemplates(sb);
	    sb = toStringHeuristics(sb);
	    sb = toStringAttribs(sb); 
	    tacletAsString = sb.append("}").toString();
	}
	return tacletAsString;
    }
}
