// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.SLListOfGoal;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.proof.TacletInstantiationsTableModel;
import java.util.HashSet;
import java.util.Iterator;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.gui.KeYMediator;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.sort.Sort;
import de.uka.ilkd.key.logic.sort.GenericSort;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.inst.*;

/** 
 * A TacletApp object contains information required for a concrete application. 
 * These information may consist of
 * <ul>
 * <li> instantiations of the schemavariables </li>
 * <li> position of the find term </li>
 * </ul>
 * Its methods <code>complete</code> or <code>sufficientlyComplete</code> are
 * used to determine if the information is complete or at least sufficient
 * (can be completed using meta variables) complete, so that is can be
 * applied. 
 */
public abstract class TacletApp implements RuleApp {

    /** the taclet for which the application information is collected */
    private Taclet taclet;
    /** contains the instantiations of the schemavarioables of the
     * Taclet 
     */
    protected SVInstantiations instantiations;

    /**
     * constraint under which the taclet application can be performed
     */
    protected Constraint       matchConstraint;
	
    /**
     * metavariables that have been created during the matching
     */
    protected SetOfMetavariable matchNewMetavariables;

    /**
     * choosen instantiations for the if sequent formulas
     */
    ListOfIfFormulaInstantiation ifInstantiations = null;

    /** set of schemavariables that appear in the Taclet and need to be
     * instantiated but are not instantiated yet. This means
     * SchemaVariables in addrule-sections have to be ignored
     */
    private SetOfSchemaVariable missingVars = null;

    /** subset of missingVars that cannot be instantiated using
     * metavariables
     */
    private SetOfSchemaVariable neededMissingVars = null;

    /**
     * the instantiations of the following SVs must not be changed;
     * they must not be instantiated with metavariables either
     */
    protected SetOfSchemaVariable fixedVars = SetAsListOfSchemaVariable.EMPTY_SET;
    /**
     * the update context given by the current instantiations must not
     * be changed
     */
    protected boolean updateContextFixed = false;
    
    /** constructs a TacletApp for the given taclet, with an empty instantiation 
     *	map 
     */
    TacletApp(Taclet taclet) {
	this ( taclet, SVInstantiations.EMPTY_SVINSTANTIATIONS, Constraint.BOTTOM,
	       SetAsListOfMetavariable.EMPTY_SET, null );
    }

    TacletApp ( Taclet                       taclet,
		SVInstantiations             instantiations,
		Constraint                   matchConstraint,
		SetOfMetavariable            matchNewMetavariables,
		ListOfIfFormulaInstantiation ifInstantiations ) {
	this.taclet                = taclet;
	this.instantiations        = instantiations;
	this.matchConstraint       = matchConstraint;
	this.matchNewMetavariables = matchNewMetavariables;
	this.ifInstantiations      = ifInstantiations;
    }

    /** constructs a TacletApp for the given taclet. With some information about
     * required instantiations.
     */
    TacletApp(Taclet taclet, SVInstantiations instantiations) {
	this ( taclet, instantiations, Constraint.BOTTOM,
	       SetAsListOfMetavariable.EMPTY_SET, null );
    }

    /** collects all bound variables above the occurrence of the schemavariable
     * whose prefix is given
     * @param prefix the TacletPrefix of the schemavariable
     * @param instantiations the SVInstantiations so that the
     * find(if)-expression matches 
     * @return set of the bound variables 
     */
    protected static SetOfQuantifiableVariable 
	boundAtOccurrenceSet(TacletPrefix prefix,
			     SVInstantiations instantiations) {	
	return collectPrefixInstantiations(prefix, 
					   instantiations);
    }


    /** collects all those logic variable that are instances of the variableSV
     * of the given prefix
     * @param pre the TacletPrefix of a SchemaVariable that is bound
     * @param instantiations the SVInstantiations to look at
     * @return the set of the logic variables whose elements are the
     * instantiations of a bound SchemaVariable appearing in the TacletPrefix
     */ 
    private static SetOfQuantifiableVariable collectPrefixInstantiations
	(TacletPrefix pre, SVInstantiations instantiations) {

	SetOfQuantifiableVariable instanceSet 
	    = SetAsListOfQuantifiableVariable.EMPTY_SET;
	IteratorOfSchemaVariable it = pre.prefix().iterator();
	while (it.hasNext()) {
	    SchemaVariable var = it.next();
	    instanceSet = 
		instanceSet.add((LogicVariable)
				((Term)instantiations.getInstantiation(var)).op());
	}
	return instanceSet;
    }


    /** returns the taclet the application information is collected for
     * @return the Taclet the application information is collected for
     */
    public Taclet taclet() {
	return taclet;
    }

    /** returns the rule the application information is collected for
     * @return the Rule the application information is collected for
     */
    public Rule rule() {
	return taclet;
    }

    /**
     * returns the instantiations for the application of the Taclet at the
     * specified position.
     * @return the SVInstantiations needed to instantiate the Taclet
     */
    public SVInstantiations instantiations() {
	return instantiations;
    }


    public Constraint       constraint () {
	return matchConstraint;
    }


    public SetOfMetavariable newMetavariables () {
	return matchNewMetavariables;
    }


    public MatchConditions  matchConditions () {
	return new MatchConditions ( instantiations   (),
				     constraint       (),
				     newMetavariables () );
    }


    public ListOfIfFormulaInstantiation ifFormulaInstantiations () {
	return ifInstantiations;
    }


    /**
     * resolves collisions between bound SchemaVariables in an SVInstantiation
     *
     * @param insts the original SVInstantiations
     * @return the resolved SVInstantiations
     */
    protected static SVInstantiations resolveCollisionVarSV
	(Taclet taclet, SVInstantiations insts) {

	HashMapFromLogicVariableToSchemaVariable collMap = 
	    new HashMapFromLogicVariableToSchemaVariable();
	IteratorOfEntryOfSchemaVariableAndInstantiationEntry it = 
	    insts.pairIterator();
	while (it.hasNext()) {
	    EntryOfSchemaVariableAndInstantiationEntry pair = it.next();
	    if (pair.key().isVariableSV()) {
		SchemaVariable varSV = pair.key();		
		Term value = ((TermInstantiation)pair.value()).getTerm();
		if (!collMap.containsKey((LogicVariable)
					 value.op())) {
		    collMap.put((LogicVariable)value.op(), 
				varSV);
		} else {
		    insts = replaceInstantiation(taclet, insts, varSV);	
		}
	    }
	}
	return insts;
    }

    /**
     * delivers the term below the (unique) quantifier of a bound
     * SchemaVariable in the given term.
     *
     * @param varSV the searched Bound Schemavariable
     * @param term the term to be searched in
     * @return the term below the given quantifier in the given term
     */
    private static Term getTermBelowQuantifier(SchemaVariable varSV, Term term) {
	for (int i=0; i<term.arity(); i++) {
	    for (int j=0; j<term.varsBoundHere(i).size(); j++) {
		if (term.varsBoundHere(i).getQuantifiableVariable(j)==varSV) {
		    return term.sub(i);
		} 
	    }
	    Term rec=getTermBelowQuantifier(varSV, term.sub(i));
	    if (rec!=null) {
		return rec;
	    }
	}
	return null;
    }

    /**
     * delivers the term below the (unique) quantifier of 
     * a bound SchemaVariable varSV in the find and if-parts of the Taclet 
     *
     * @param varSV the searched bound SchemaVariable 
     * @return the term below the given quantifier in the find and if-parts of 
     * the Taclet
     */
    private static Term getTermBelowQuantifier(Taclet taclet, SchemaVariable varSV) {
	IteratorOfConstrainedFormula it=taclet.ifSequent().iterator();
	while (it.hasNext()) {
	    Term result=getTermBelowQuantifier(varSV, it.next().formula());
	    if (result!=null) {
		return result;
	    }
	}

	if (taclet instanceof FindTaclet) {
	    return getTermBelowQuantifier(varSV, ((FindTaclet)taclet).find());
	} else	
	    return null;
    }


    /** returns true iff the instantiation of a bound SchemaVariable contains the given
    * Logicvariable 
    * @param boundVars ArrayOfQuantifiableVariable with the
    * bound SchemaVariables
    * @param x the LogicVariable that is looked for
    * @param insts the SVInstantiations where to get the necessary
    * instantiations of the bound SchemaVariables
    * @return true iff the instantiation of a Bound Schemavariable contains the given
    * Logicvariable 
    */
    static private boolean contains(ArrayOfQuantifiableVariable boundVars,
			     LogicVariable x,
			     SVInstantiations insts) {
	for (int i=0; i<boundVars.size(); i++) {
	    Term instance = (Term) insts.getInstantiation
		((SchemaVariable)boundVars.getQuantifiableVariable(i));
	    if (instance.op() == x) {
		return true;
	    } 
	}
	
	return false;
    }

    /**
     * returns a new SVInstantiations that modifies the given
     * SVInstantiations insts at the bound SchemaVariable varSV to a new LogicVariable.
     */
    protected static SVInstantiations replaceInstantiation
	(Taclet taclet, SVInstantiations insts, SchemaVariable varSV){
	Term term=getTermBelowQuantifier(taclet, varSV);
	LogicVariable newVariable = new LogicVariable 
	    (new Name(((Term)insts.getInstantiation(varSV)).op()
		      .name().toString()+"0"),
	     ((Term)insts.getInstantiation(varSV)).sort());
	// __CHANGE__ How to name the new variable?	
	Term newVariableTerm 
	    = TermFactory.DEFAULT.createVariableTerm(newVariable);
	return replaceInstantiation(insts, term, varSV, newVariableTerm);
    }

    /**
     * returns a new SVInstantiations that modifies the given
     * SVInstantiations insts at the bound SchemaVariable u to the Term (that is
     * a LogicVariable) y.
     */
    static private SVInstantiations replaceInstantiation
	(SVInstantiations insts, Term t,  SchemaVariable u, Term y){

	SVInstantiations result = insts;
	LogicVariable x = (LogicVariable)
	    ((Term)insts.getInstantiation(u)).op();
	if (t.op() instanceof SchemaVariable) {
	    if (!((SchemaVariable)t.op()).isVariableSV()) {
		SchemaVariable sv=(SchemaVariable)t.op();
		ClashFreeSubst cfSubst
		    = new ClashFreeSubst(x,y); 
		result = result.replace
		    (sv, 
		     cfSubst.apply((Term)insts.getInstantiation(sv)));
	    }
	} else {
	    for (int i=0; i<t.arity(); i++) {
		if (!contains(t.varsBoundHere(i), x, insts)) {	       
		    result = replaceInstantiation(result, t.sub(i),
						  u, y);
		}
	    }
		
	}

	result=result.replace(u, y);
	return result;
    }

    /** applies the specified rule at the specified position 
     * if all schema variables have been instantiated
     * @param goal the Goal at which the Taclet is applied
     * @param javaInfo the JavaInfo encapsulating all java information
     * @return list of new created goals 
     */
    public ListOfGoal execute(Goal goal, JavaInfo javaInfo) {

	if (!complete()) {	  
	    throw new IllegalStateException("Tried to apply rule \n"+taclet
					    +"\nthat is not complete.");
	}
	goal.addAppliedRuleApp(this);	
	return taclet().apply(goal, javaInfo, this);
    }    

    /** applies the specified rule at the specified position 
     * if all schema variables have been instantiated
     * and creates a new goal if the if-sequent does not match.
     * The new goal proves that the if sequent can be derived.
     * @param goal the Goal where to apply the rule
     * @param javaInfo the JavaInfo encapsulating all java information
     * @return list of new created goals 
     */
    /*
    public ListOfGoal executeForceIf(Goal goal, JavaInfo javaInfo) {
	if (!complete()) {
	    throw new IllegalStateException("Applied rule not complete.");
	}
	goal.addAppliedRuleApp(this);	
	return taclet().apply(goal, javaInfo, this, true);
    }    
    */

    /** calculate needed SchemaVariables that have not been
     * instantiated yet. This means to ignore SchemaVariables that
     * appear only in addrule-sections of Taclets
     * @return SetOfSchemaVariable that need to be instantiated but
     * are not
     */
    protected SetOfSchemaVariable calculateNonInstantiatedSV() {
	if (missingVars == null) {
	    missingVars = SetAsListOfSchemaVariable.EMPTY_SET;
	    TacletVariableCollector coll =
		new TacletVariableCollector(instantiations());
	    coll.visitWithoutAddrule(taclet());	    
	    IteratorOfSchemaVariable it = coll.varIterator();
	    while (it.hasNext()) {
		SchemaVariable var=it.next();
		if (!instantiations().isInstantiated(var)) {
		    missingVars = missingVars.add(var);	
		}
	    }	    
	}
	return missingVars;
    }

    /** Calculate needed SchemaVariables that have not been
     * instantiated yet and cannot be instantiated using
     * metavariables. This is a subset of
     * calculateNonInstantiatedSV();
     * @return SetOfSchemaVariable that need to be instantiated but
     * are not
     */
    protected SetOfSchemaVariable calculateNeededNonInstantiatedSV() {
	if (neededMissingVars == null) {
	    IteratorOfSchemaVariable it  =
		calculateNonInstantiatedSV ().iterator ();
	    SchemaVariable           var;
	    SVInstantiations         svi = instantiations ();

	    neededMissingVars = SetAsListOfSchemaVariable.EMPTY_SET;

	    while ( it.hasNext () ) {
		var = it.next ();

		if ( canUseMVAPriori ( var ) ) {
		    GenericSortCondition c =
			GenericSortCondition.forceInstantiation ( var.sort (),
								  true );
		    if ( c == null )
			continue; // then the sort is not generic
		    else {
			try {
			    svi = svi.add ( c );
			    continue; // then the sort can be guessed
			} catch ( GenericSortException e ) {}
		    }
		}

		neededMissingVars = neededMissingVars.add ( var );
	    }
	}
	
	return neededMissingVars;
    }

    /** creates a new Tacletapp where the SchemaVariable sv is
     * instantiated with the the given Term term. Sort equality is
     * checked. If the check fails an IllegalArgumentException is
     * thrown
     * @param sv the SchemaVariable to be instantiated
     * @param term the Term the SchemaVariable is instantiated with
     * @return the new TacletApp
     */
    public TacletApp addCheckedInstantiation(SchemaVariable sv, 
					     Term term,
                                             boolean interesting) 
	throws IllegalArgumentException {

	// This check is done by the SVInstantiations object
	/*
	if (!term.sort().extendsTrans(sv.sort())) {
	    throw new IllegalArgumentException
		("Could not add the instantiation of "+sv
		 +" (sort: "+sv.sort()+") because "
		 +"the sort of "+
		 term+" (sort: "+term.sort()+") is different.");
	} else
	*/
	if (sv.isVariableSV() && 
	    !(term.op() instanceof LogicVariable)) {
	    throw new IllegalArgumentException("Could not add "+
	       "the instantiation of "+sv+" because "+
		term+" is no variable.");
	} 
	return addInstantiation(sv, term, interesting);    
    }


    /**
     * returns the variables that have not yet been instantiated and
     * need to be instantiated to apply the Taclet. (These are not all
     * SchemaVariables like the one that appear only in the addrule
     * sections)
     * @return SetOfSchemaVariable with SchemaVariables that have not
     * been instantiated yet 
     */
    public SetOfSchemaVariable uninstantiatedVars() {
	return calculateNonInstantiatedSV();
    }
    
    /**
     * returns the variables that are currently uninstantiated and
     * cannot be instantiated automatically using metavariables
     * @return SetOfSchemaVariable with SchemaVariables that have not
     * been instantiated yet 
     */
    public SetOfSchemaVariable neededUninstantiatedVars() {
	return calculateNeededNonInstantiatedSV();
    }
    

    /**
     * @return A TacletApp with this.sufficientlyComplete() or null
     */    
    public TacletApp tryToInstantiate(JavaInfo javaInfo) {
        TacletApp app = this;
	IteratorOfSchemaVariable it = uninstantiatedVars().iterator();
	while (it.hasNext()) {
	    SchemaVariable sv=it.next();
	    if (sv.sort()==ProgramSVSort.VARIABLE ||
		sv.sort()==ProgramSVSort.BOOLEANVARIABLE ||
		sv.sort()==ProgramSVSort.LABEL) {
		String proposal = TacletInstantiationsTableModel.
		   getNameProposalForSchemaVariable(sv);
		ProgramElement pe = TacletInstantiationsTableModel.
		   getProgramElement(app, proposal, sv, javaInfo);
		app = app.addInstantiation(sv, pe, true);
	    } else if ( sv.isDependingSV () ) {
		// if the sort of the schema variable is generic,
		// ensure that it is instantiated
		GenericSortCondition c =
		    GenericSortCondition.forceInstantiation ( sv.sort (),
							      false );
		if ( c != null ) {
		    try {
			app = app.setInstantiation
			    ( instantiations ().add ( c ) );
		    } catch ( GenericSortException e ) {
			return null;
		    }
		}
		String proposal = TacletInstantiationsTableModel
		    .getNameProposalForDependingVariable(app, sv);
		app = app.createDependingConstant ( proposal,
						    sv,
						    true );
	    } else if ( !( sv.isTermSV () && canUseMVAPriori ( sv ) ) ) {
		return null;
	    }
	}
	return app;
    }
    

    /**
     * Make the instantiation complete using metavariables and convert
     * plain instantiations to metavariables and user constraint
     * entries if possible.
     * Pre-condition: this.sufficientlyComplete()
     */
    public TacletApp instantiateWithMV ( KeYMediator p_mediator ) {
	Debug.assertTrue ( this.sufficientlyComplete (),
			   "TacletApp cannot be made complete" );

	SVInstantiations  insts   = instantiations   ();
	SetOfMetavariable newVars = newMetavariables ();
	Term              t;
	Metavariable      mv;
	SchemaVariable    sv;

	{
	    // Replace plain instantiations of termSV

	    IteratorOfEntryOfSchemaVariableAndInstantiationEntry it    =
		insts.pairIterator ();
	    EntryOfSchemaVariableAndInstantiationEntry           entry;

	    while ( it.hasNext () ) {
		entry = it.next ();
		sv    = entry.key ();

		if ( (sv.isTermSV ())
		     && !taclet ().getIfFindVariables ().contains ( sv ) &&
		     canUseMVAPosteriori ( sv,
					   ((TermInstantiation)entry.value ())
					   .getTerm () ) ) {
		    mv = getMVFor ( sv,
				    insts.getGenericSortInstantiations ()
				    .getRealSort ( sv ),
				    p_mediator );
		    if ( mv != null ) {
			newVars = newVars.add ( mv );
			t       = TermFactory.DEFAULT.createFunctionTerm ( mv );
			p_mediator.getUserConstraint ().addEquality
			    ( t, ((TermInstantiation)entry.value ()).getTerm () );
			insts   = insts.replace ( sv, t );
		    }
		}
	    }
	}

	{
	    // Instantiate uninstantiated termSV using metavariables

	    SetOfSchemaVariable      svs = uninstantiatedVars ();
	    IteratorOfSchemaVariable it  = svs.iterator ();
	    GenericSortCondition     c;

	    // force all generic sorts to be instantiated
	    try {
		while ( it.hasNext () ) {
		    sv = it.next ();
		    c  = GenericSortCondition.forceInstantiation ( sv.sort (),
								   true );
		    if ( c != null )
			insts = insts.add ( c );
		}
	    } catch ( GenericSortException e ) {
		Debug.out ( "TacletApp cannot be made complete" );		
	    }

	    it = svs.iterator ();
	    while ( it.hasNext () ) {
		sv = it.next ();
		Debug.assertTrue ( canUseMVAPriori ( sv ),
				   "Should be able to instantiate " + sv +
				   " using metavariables, but am not" );

		mv      = getMVFor ( sv,
				     insts.getGenericSortInstantiations ()
				     .getRealSort ( sv ),
				     p_mediator );
		newVars = newVars.add ( mv );
		t       = TermFactory.DEFAULT.createFunctionTerm ( mv );
		insts   = insts.add ( sv, t );
	    }
	}

	return setMatchConditions ( new MatchConditions ( insts,
							  constraint (),
							  newVars ) );
    }


    /**
     * Create a Metavariable the given SchemaVariable can be
     * instantiated with
     * @return an appropriate mv, or null if for some reason the
     * creation failed
     */
    private Metavariable getMVFor ( SchemaVariable p_sv,
				    Sort           p_sort,
				    KeYMediator    p_mediator ) {
	return p_mediator.getMetavariableDeliverer ().createNewVariable
	    ( TacletInstantiationsTableModel
	      .getBasenameProposalForMetavariable ( this, p_sv ), p_sort );
    }


    /**
     * @return p_s iff p_s is not a generic sort, the concrete sort
     * p_s is instantiated with currently otherwise 
     * @throws GenericSortException iff p_s is a generic sort which is
     * not yet instantiated
     */
    public Sort getRealSort ( SchemaVariable p_sv )
	throws GenericSortException {
	return instantiations ().getGenericSortInstantiations ()
	    .getRealSort ( p_sv );
    }

    public Sort getRealSort ( Sort p_s )
	throws GenericSortException {
	return instantiations ().getGenericSortInstantiations ()
	    .getRealSort ( p_s );
    }


    /**
     * Create a new constant named "instantiation" and instantiate
     * "sv" with. This constant will later (by
     * "createDependingFunctions") be replaced by a function having
     * the occurring metavariables as arguments
     */
    public TacletApp createDependingConstant ( String         instantiation,
					       SchemaVariable sv,
					       boolean        interesting ) {
	return createDependingConstant ( instantiation,
					 sv,
					 getRealSort ( sv ),
					 interesting );
    }
    
    public TacletApp createDependingConstant ( String         instantiation,
					       SchemaVariable sv,
					       Sort           sort,
					       boolean        interesting ) {
	Function c = new Function ( new Name ( instantiation ),
				    sort,
				    new Sort [0] );
	return addInstantiation ( sv,
				  TermFactory.DEFAULT.createFunctionTerm ( c ),
				  interesting );
    }
    

    /**
     * Create skolem functions (for variables declared via "c new
     * depending on phi") to replace previously created constants with
     */
    public TacletApp createDependingFunctions ( Namespace p_func_ns ) {
	IteratorOfNewDependingOn it    = taclet ().varsNewDependingOn ();

	if ( !it.hasNext () )
	    return this;

	NewDependingOn           entry;
	SVInstantiations         insts = instantiations   ();
	Term                     t;
	Term                     phi;
	MVCollector              coll;
	IteratorOfMetavariable   itMV;
	Metavariable             mv;
	HashSet                  mvs;
	Iterator                 itSet;
	int                      i;
	Sort[]                   argSorts;
	Term[]                   argTerms;
	Function                 skolemFunc;

	while ( it.hasNext () ) {
	    entry      = it.next ();
	    t          = (Term)insts.getInstantiation ( entry.first () );

	    Debug.assertTrue ( t != null,
			       "Name for depending variable missing" );

	    phi        = (Term)insts.getInstantiation ( entry.second () );

	    Debug.assertTrue ( phi != null,
			       "Variable depends on uninstantiated variable" );

	    // Collect variables the schema variable should depend on
	    coll       = new MVCollector ();
	    phi.execPostOrder ( coll );

	    itMV       = coll.mv ();
	    
	    if ( !itMV.hasNext () ) {
		p_func_ns.add ( t.op () );
		continue;
	    }
	    
	    // Collect metavariables, remove doubles
	    mvs        = new HashSet ();
	    while ( itMV.hasNext () )
		mvs.add ( itMV.next () );

	    // Create function with correct arguments
	    argSorts   = new Sort [ mvs.size () ];
	    argTerms   = new Term [ mvs.size () ];
	    itSet      = mvs.iterator ();
	    i          = 0;
	    while ( itSet.hasNext () ) {
		mv          = (Metavariable)itSet.next ();
		argSorts[i] = mv.sort ();
		argTerms[i] = TermFactory.DEFAULT.createFunctionTerm ( mv );
		++i;
	    }

	    skolemFunc = new Function ( t.op ().name (),
					t.sort (),
					argSorts );
	    t          = TermFactory.DEFAULT.createFunctionTerm ( skolemFunc,
								  argTerms );
	    insts      = insts.replace ( entry.first (), t );

	    p_func_ns.add ( skolemFunc );
	}

	return setInstantiation ( insts );
    }

    
    /** adds a new instantiation to this TacletApp 
     * @param sv the SchemaVariable to be instantiated
     * @param term the Term the SchemaVariable is instantiated with
     * @return the new TacletApp
     */
    public abstract TacletApp addInstantiation(SchemaVariable sv, Term term,
                                               boolean interesting);


    /** adds a new instantiation to this TacletApp 
     * @param sv the SchemaVariable to be instantiated
     * @param term the ProgramElement the SV is instantiated with
     * @return the new TacletApp
     */
    public abstract TacletApp addInstantiation(SchemaVariable sv, 
					       ProgramElement pe,
                                               boolean interesting);


    /**
     * creates a new Taclet application containing all the
     * instantiations given 
     * by the SVInstantiations
     * @param svi the SVInstantiations whose entries are the needed
     * instantiations 
     * @return the new Taclet application
     */
    public abstract TacletApp addInstantiation(SVInstantiations svi);


    /**
     * creates a new Taclet application containing all the
     * instantiations given 
     * by the SVInstantiations and forget the old ones
     * @param svi the SVInstantiations whose entries are the needed
     * instantiations 
     * @return the new Taclet application
     */
    protected abstract TacletApp setInstantiation(SVInstantiations svi);


    /**
     * creates a new Taclet application containing all the
     * instantiations, constraints and new metavariables given 
     * by the mc object and forget the old ones
     */
    protected abstract TacletApp setMatchConditions ( MatchConditions mc );


    /**
     * creates a new Taclet application containing all the
     * instantiations, constraints, new metavariables and if formula
     * instantiations given and forget the old ones
     */
    protected abstract TacletApp
	setAllInstantiations ( MatchConditions              mc,
			       ListOfIfFormulaInstantiation ifInstantiations );


    /**
     * Creates a new Taclet application by matching the given formulas
     * against the formulas of the if sequent, adding SV
     * instantiations, constraints and metavariables as needed. This
     * will fail if the if formulas have already been instantiated.
     */
    public TacletApp setIfFormulaInstantiations
	( ListOfIfFormulaInstantiation p_list,
	  JavaInfo                     p_javaInfo,
	  Constraint                   p_userConstraint ) {
	Debug.assertTrue ( p_list != null &&
			   ifInstsCorrectSize ( taclet, ifInstantiations ) &&
			   ifInstantiations == null,
			   "If instantiations list has wrong size or is null " +
			   "or the if formulas have already been instantiated" );

	MatchConditions  mc = taclet ().matchIf ( p_list.iterator (),
						  matchConditions (),
						  p_javaInfo,
						  p_userConstraint );
	
	return mc == null ? null : setAllInstantiations ( mc, p_list );
    }


    /**
     * Find all possible instantiations of the if sequent formulas
     * within the sequent "p_seq".
     * @return a list of tacletapps with the found if formula
     * instantiations
     */
    public ListOfTacletApp
	findIfFormulaInstantiations ( Sequent     p_seq,
				      JavaInfo    p_javaInfo,
				      Constraint  p_userConstraint ) {
	
	Debug.assertTrue ( ifInstantiations == null,
			   "The if formulas have already been instantiated" );

	if ( taclet ().ifSequent () == Sequent.EMPTY_SEQUENT )
	    return SLListOfTacletApp.EMPTY_LIST.prepend ( this );

	return
	    findIfFormulaInstantiationsHelp
	    ( createSemisequentList ( taclet ().ifSequent ()  // Matching starting
				      .succedent () ),    // with the last formula
	      createSemisequentList ( taclet ().ifSequent ()
				      .antecedent () ),
	      IfFormulaInstSeq.createList(p_seq, false),
	      IfFormulaInstSeq.createList(p_seq, true),
	      SLListOfIfFormulaInstantiation.EMPTY_LIST,
	      matchConditions (),
	      p_javaInfo,
	      p_userConstraint );
	
    }

    
    /**
     * Recursive function for matching the remaining tail of an if sequent
     * @param p_ifSeqTail tail of the current semisequent as list
     * @param p_ifSeqTail2nd the following semisequent
     * (i.e. antecedent) or null
     * @param p_toMatch list of the formulas to match the current if
     * semisequent formulas with
     * @param p_toMatch2nd list of the formulas of the succedent
     * @param p_matchCond match conditions until now, i.e. after
     * matching the first formulas of the if sequent
     */
    private ListOfTacletApp findIfFormulaInstantiationsHelp
	( ListOfConstrainedFormula      p_ifSeqTail,
	  ListOfConstrainedFormula      p_ifSeqTail2nd,
	  ListOfIfFormulaInstantiation  p_toMatch,
	  ListOfIfFormulaInstantiation  p_toMatch2nd,
	  ListOfIfFormulaInstantiation  p_alreadyMatched,
	  MatchConditions               p_matchCond,
	  JavaInfo                      p_javaInfo,
	  Constraint                    p_userConstraint ) {

	while ( p_ifSeqTail == SLListOfConstrainedFormula.EMPTY_LIST ) {
	    if ( p_ifSeqTail2nd == null ) {
		// All formulas have been matched, collect the results
		TacletApp res = setAllInstantiations ( p_matchCond,
						       p_alreadyMatched );
		if ( res != null )
		    return SLListOfTacletApp.EMPTY_LIST.prepend ( res );
		return SLListOfTacletApp.EMPTY_LIST;
	    } else {
		// Change from succedent to antecedent
		p_ifSeqTail    = p_ifSeqTail2nd;
		p_ifSeqTail2nd = null;
		p_toMatch      = p_toMatch2nd;
	    }
	}

	// Match the current formula
	IfMatchResult mr        = taclet ().matchIf ( p_toMatch.iterator (),
						      p_ifSeqTail.head ().formula (),
						      p_matchCond,
						      p_javaInfo,
						      p_userConstraint );

	// For each matching formula call the method again to match
	// the remaining terms
	IfFormulaInstantiation           candidate;
	ListOfTacletApp                  res    = SLListOfTacletApp.EMPTY_LIST;
	IteratorOfIfFormulaInstantiation itCand = mr.getFormulas        ().iterator ();
	IteratorOfMatchConditions        itMC   = mr.getMatchConditions ().iterator ();
	p_ifSeqTail                             = p_ifSeqTail.tail ();
	while ( itCand.hasNext () ) {
	    res = res.prepend
		( findIfFormulaInstantiationsHelp ( p_ifSeqTail,
						    p_ifSeqTail2nd,
						    p_toMatch,
						    p_toMatch2nd,
						    p_alreadyMatched
						    .prepend ( itCand.next () ),
						    itMC.next (),
						    p_javaInfo,
						    p_userConstraint ) );
	}

	return res;
    }

    private ListOfConstrainedFormula createSemisequentList ( Semisequent p_ss ) {
	ListOfConstrainedFormula res = SLListOfConstrainedFormula.EMPTY_LIST;

	IteratorOfConstrainedFormula it  = p_ss.iterator ();
	while ( it.hasNext () )
	    res = res.prepend ( it.next () );

	return res;	
    }


    /**
     * returns a new PosTacletApp that is equal to this TacletApp except
     * that the position is set to the given PosInOccurrence
     * @param pos the PosInOccurrence of the newl created PosTacletApp
     * @return the new TacletApp
     */
    public PosTacletApp setPosInOccurrence(PosInOccurrence pos) {
	if (taclet() instanceof NoFindTaclet) {
	    throw new IllegalStateException("Cannot add position to an taclet"
					    +" without find");
	}
	return PosTacletApp.createPosTacletApp((FindTaclet)taclet      (), 
					       instantiations          (),
					       constraint              (),
					       newMetavariables        (),
					       ifFormulaInstantiations (),
					       pos);
    }


    /** returns true iff all necessary informations are collected, so
     * that the Taclet can be applied.
     * @return true iff all necessary informations are collected, so
     * that the Taclet can be applied.
     */
    public abstract boolean complete();


    /**
     * @return true iff the taclet instantiation can be made complete
     * using metavariables
     */
    public abstract boolean sufficientlyComplete();


    /**
     * @return true iff the SV instantiations can be made complete
     * using metavariables
     */
    protected boolean instsSufficientlyComplete() {
	return neededUninstantiatedVars () == SetAsListOfSchemaVariable.EMPTY_SET;
    }


    /**
     * @return true iff the if instantiation list is not null or no if
     * sequent is needed
     */
    protected boolean ifInstsComplete() {
	return ifInstantiations != null ||
	    ( taclet ().ifSequent () == Sequent.EMPTY_SEQUENT );
    }


    /**
     * returns the PositionInOccurrence (representing a ConstrainedFormula and
     * a position in the corresponding formula)
     * @return the PosInOccurrence
     */
    public abstract PosInOccurrence posInOccurrence();


    /** compares the given Object with this one and returns true iff both are
     * from type TacletApp with equal taclets, instantiations and positions.
     */
    public boolean equals(Object o) {
	if (o==this) return true;
	if (!(o instanceof TacletApp)) return false;
	TacletApp s=(TacletApp)o;
	return (s.taclet().equals(taclet()) 
		&& s.instantiations().equals(instantiations()));
    }

    public String toString() {
	return "Application of Taclet " +taclet()+ " with "
	    +instantiations();
    }


    /**
     * checks if there are name conflicts (i.e. there are
     * two matched bound SchemaVariable that are matched to variables with an
     * equal name); if yes a new TacletApp is returned that equals this
     * TacletApp except that the name conflict is resolved by replacing the
     * instantiation of one of the conflict-causing SchemaVariables by a 
     * bound SchemaVariable with a new name; 
     * if the check is negative, the same TacletApp is returned.
     *
     * @return a conflict resolved TacletApp, remainder equal to this TacletApp
     */
    public TacletApp prepareUserInstantiation() {
	TacletApp result = this;
	SchemaVariable varSV = varSVNameConflict();
	while (varSV!=null) {
	    result = setInstantiation(replaceInstantiation
				    (taclet, result.instantiations(), varSV));
	    varSV = result.varSVNameConflict();
	}
	return result;
    }

    protected abstract SetOfQuantifiableVariable contextVars(SchemaVariable sv);

    /**
     * creates a new variable namespace by adding names of the instantiations
     * of the schema variables in the context of the given schema
     * variable and (if the TacletApp's prefix has the context flag set)
     * by adding names of the logic variables of the context.
     *
     * @param sv the schema variable to be considered
     * @param var_ns the old variable namespace
     * @return the new created variable namespace
     */
    public Namespace extendVarNamespaceForSV(Namespace var_ns, 
					     SchemaVariable sv) {
	Namespace ns=new Namespace(var_ns);
	IteratorOfSchemaVariable it=taclet().getPrefix(sv).prefix().iterator();
	while (it.hasNext()) {
	    LogicVariable var = (LogicVariable) 
		((Term)instantiations().getInstantiation(it.next())).op();
	    ns.add(var);
	}
	if (taclet().getPrefix(sv).context()) {	    
	    IteratorOfQuantifiableVariable lit
		= contextVars(sv).iterator();
	    while (lit.hasNext()) {
		ns.add(lit.next());
	    }
	}
	return ns;
    }


    /**
     * returns the bound SchemaVariable that causes a name conflict (i.e. there are
     * two matched bound SchemaVariables that are matched to variables with an
     * equal name, the returned SchemaVariable is (an arbitrary) one of these
     * SchemaVariables) or null if there is no such conflict
     *
     * @return null iff there is no conflict and one of the
     * conflict-causing bound SchemaVariables
     */
    public SchemaVariable varSVNameConflict() {
	IteratorOfSchemaVariable svIt=uninstantiatedVars().iterator();
	while (svIt.hasNext()) { 
	    SchemaVariable sv=svIt.next();
	    if (!(sv.isVariableSV())&& !(sv.isProgramSV())) {
		TacletPrefix prefix=taclet().getPrefix(sv);
		HashSet names=new HashSet();	    
		if (prefix.context()) {
		    IteratorOfQuantifiableVariable contextIt
			= contextVars(sv).iterator();
		    while (contextIt.hasNext()) {
			names.add(contextIt.next().name());
		    }
		}
		IteratorOfSchemaVariable varSVIt=prefix.iterator();
		while(varSVIt.hasNext()) {
		    SchemaVariable varSV=varSVIt.next();
		    Term inst = (Term)
			instantiations().getInstantiation(varSV);
		    if (inst!=null) {
			Name name = inst.op().name();
			if (!names.contains(name)) {
			    names.add(name);
			} else {
			    return varSV;
			}
		    }
		}
	    }
	}
	return null;
    }


    /**
     * For a given SV of the taclet decide a priori whether an
     * instantiation via metavariable is possible
     * @return true iff p_var is a termSV with empty prefix
     */
    public boolean canUseMVAPriori ( SchemaVariable p_var ) {
	if ( !p_var.isTermSV () ||
	     fixedVars.contains ( p_var ) )
	    return false;
	else {
	    TacletPrefix prefix = taclet ().getPrefix ( p_var );
	    return (prefix.prefix ().size () == 0 && !prefix.context ());
	}
    }


    /**
     * For a given SV of the taclet decide a posteriori whether an
     * instantiation via metavariable is possible
     * @param p_term Term to instantiate the schemavariable with
     */
    protected boolean canUseMVAPosteriori ( SchemaVariable p_var, Term p_term ) {
	return canUseMVAPriori ( p_var ) && p_term.isRigid ();
    }    


    /**
     * @return true iff the list of if instantiations has the correct
     * size or is null
     */
    protected static boolean ifInstsCorrectSize ( Taclet                       p_taclet,
						  ListOfIfFormulaInstantiation p_list ) {
	return p_list == null ||
	    p_list.size () == ( p_taclet.ifSequent ().antecedent ().size () +
				p_taclet.ifSequent ().succedent  ().size () );
    }

}
