// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.logic.op.*;
import java.io.Serializable;
import de.uka.ilkd.key.logic.Sequent;

/** 
 * this class contains the goals of the schematic theory specific
 * rules (Taclet). There are new sequents that have to be added, new
 * rules and rule variables. The replacewith-goal is implemented in
 * subclasses 
 */
public class TacletGoalTemplate implements Serializable{

    /** stores sequent that is one of the new goals */
    private Sequent addedSeq = Sequent.EMPTY_SEQUENT;     

    /** stores list of Taclet which are introduced*/
    private ListOfTaclet addedRules = SLListOfTaclet.EMPTY_LIST;

    /** used rulevariables */
    private ListOfRulevariable ruleVars=SLListOfRulevariable.EMPTY_LIST;

    private SetOfSchemaVariable addedProgVars
	= SetAsListOfSchemaVariable.EMPTY_SET;
        
    private String name = null;
        

    /** creates new Goaldescription 
     *@param addedSeq new Sequent to be added
     *@param addedRules ListOfTaclet contains the new allowed rules
     * at this branch 
     *@param ruleVars ListOfRulevariables contains used rulevariables 
     */
    public TacletGoalTemplate(Sequent addedSeq,
			      ListOfTaclet addedRules,
			      ListOfRulevariable ruleVars,
			      SetOfSchemaVariable addedProgVars) {
	TacletBuilder.checkContainsFreeVarSV(addedSeq, "add sequent");
	   
	this.addedRules=addedRules;
	this.ruleVars=ruleVars;
	this.addedSeq=addedSeq;
	this.addedProgVars = addedProgVars;
    }

    public TacletGoalTemplate(Sequent addedSeq,
			      ListOfTaclet addedRules,
			      ListOfRulevariable ruleVars) {		
	this(addedSeq, addedRules, ruleVars, 
	     SetAsListOfSchemaVariable.EMPTY_SET);
    }


    /** a Taclet may replace parts of sequent.
     * @return term (or sequent) to be placed instead of the findexp-term.
     * REMARK: returns 'null' if there is no replace-with part !
     * Overwritten in subclasses !
     */
    Object replaceWithExpressionAsObject() {
	return null;
    }


    /** a Taclet may add a new Sequent as Goal. Use this method to get
     * this Sequent
     * @return Sequent to be added as Goal or Sequent.EMPTY_SEQUENT if
     * no such Sequent exists
     */
    public Sequent sequent() {
	return addedSeq;
    }

    /** the goal of a Taclet may introduce new rules. Call this method
     * to get them 
     * @return ListOfTaclet contains new introduced rules
     */
    public ListOfTaclet rules() {
	return addedRules;
    }

    /** the goal of a Taclet may use ruleVars. Call this method
     * to get them.
     * @return ListOfRuleVariable contains new introduced rules
     */
    public ListOfRulevariable ruleVars() {
	return ruleVars;
    }

    public SetOfSchemaVariable addedProgVars() {
	return addedProgVars;
    }


    /**
     * returns true iff the given schema variable is bound in the sequent 
     * part of
     * this goal sequent or in an added rule
     * @param v the bound variable to be searched
     */
    boolean varIsBound(SchemaVariable v) {
	if (!v.isVariableSV()) {
	    return false;
	}
	IteratorOfTaclet tacletIt=rules().iterator();
	while (tacletIt.hasNext()) {
	    if (tacletIt.next().varIsBound(v)) return true;
	}	    
	return (sequent().varIsBound(v));
    }


    /**
     * returns pairs of schema variables (x,y) where y is bound in this
     * TacletGoalTemplate and x 
     * occurs in the scope of y; (y,y) is part of the set for all bound ys
     * with no x bound in the scope of y. x is a bound SchemaVariable 
     * (as itself is bound)
     * while y is a common Schema Variable.
     * @return a set of variable pairs as specified above
     */
    protected SetOfInScopePair getVarInScope() {
	SetOfInScopePair result
	    =SetAsListOfInScopePair.EMPTY_SET;
	IteratorOfTaclet tacletIt=rules().iterator();
	while (tacletIt.hasNext()) {
	    result=result.union(tacletIt.next().getVarInScope());
	}	    
	return result.union(sequent().getVarInScope());
	 
    }
    
    
    public void setName(String s) {
       name = s;
    }
    
    public String name() {
       return name;
    }

    public boolean equals(Object o) {
	if ( ! ( o instanceof TacletGoalTemplate ) )
	    return false;

	TacletGoalTemplate other = (TacletGoalTemplate) o;
	
	return 
	    addedSeq.equals(other.addedSeq)
	    && addedRules.equals(other.addedRules)
	    && ruleVars.equals(other.ruleVars);
    }

    
    /** toString */
    public String toString() {
	String result="";
	if (sequent()!=Sequent.EMPTY_SEQUENT) result+="add "+sequent()+" "; 
	if (rules()!=SLListOfTaclet.EMPTY_LIST) result+="addrules "+rules()+" ";
	if (addedProgVars().size()>0) result+="addprogvar "+addedProgVars()+" ";
	if (ruleVars()!=SLListOfRulevariable.EMPTY_LIST) 
	    result+="(rulevariables: "+ruleVars()+") ";
	return result;
    }

}
