// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//


package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.java.visitor.ProgramSVCollector;

public class TacletPrefixBuilder {

    /** set of all schemavariables that are only allowed to be matched
     * with quantifiable variables. 
     */
    private SetOfSchemaVariable currentlyBoundVars=SetAsListOfSchemaVariable.EMPTY_SET;
    private TacletBuilder tacletBuilder;

    protected MapFromSchemaVariableToTacletPrefix prefixMap
	=MapAsListFromSchemaVariableToTacletPrefix.EMPTY_MAP;

    TacletPrefixBuilder(TacletBuilder tacletBuilder) {
	this.tacletBuilder=tacletBuilder;
    }
    

    private void removeVarsBoundHere(Term visited, int subTerm) {
	ArrayOfQuantifiableVariable bdVars=visited.varsBoundHere(subTerm);
	for (int i=0; i<bdVars.size(); i++) {
	    QuantifiableVariable boundVar = bdVars.getQuantifiableVariable(i);
	    if ( boundVar instanceof SchemaVariable &&
		 ((SchemaVariable)boundVar).isVariableSV() ) {
		currentlyBoundVars=currentlyBoundVars.remove
		    ((SchemaVariable)boundVar);
	    }
	}
    }


    private void addVarsBoundHere(Term visited, int subTerm) {
	ArrayOfQuantifiableVariable bdVars=visited.varsBoundHere(subTerm);
	for (int i=0; i<bdVars.size(); i++) {
	    QuantifiableVariable boundVar = bdVars.getQuantifiableVariable(i);
	    if ( boundVar instanceof SchemaVariable &&
		 ((SchemaVariable)boundVar).isVariableSV() ) {
		currentlyBoundVars = currentlyBoundVars.add
		    ((SchemaVariable)boundVar);
	    }
	}
    }

    private void setPrefixOfOccurrence(SchemaVariable sv, 
				       SetOfSchemaVariable relevantBoundVars) {

	prefixMap = 
	    prefixMap.put(sv, new TacletPrefix(relevantBoundVars, false));
    }

    /**
     * removes all variables x that are declared as x not free in sv from the
     * currently bound vars set.
     */
    private SetOfSchemaVariable removeNotFreeIn(SchemaVariable sv) {
	SetOfSchemaVariable result=currentlyBoundVars;
	IteratorOfNotFreeIn it=tacletBuilder.varsNotFreeIn();
	while (it.hasNext()) {
	    NotFreeIn notFreeIn=it.next();
	    if (notFreeIn.second()==sv) {
		result=result.remove(notFreeIn.first());
	    }
	}
	return result;
    }

    private void visit(Term t) {
	if (t.op() instanceof SchemaVariable && 
	    !((SchemaVariable)t.op()).isVariableSV() &&
	    !((SchemaVariable)t.op()).isProgramSV()) {	    // see below /AR
	    SchemaVariable sv=(SchemaVariable)t.op();
	    SetOfSchemaVariable relevantBoundVars = 
		removeNotFreeIn(sv);
	    TacletPrefix prefix=prefixMap.get(sv);
	    if (prefix==null || prefix.prefix().equals(relevantBoundVars)) {
		setPrefixOfOccurrence(sv, relevantBoundVars);
	    } else {
	       throw new InvalidPrefixException(sv);
	    }
	} 
//      IT SEEMS THAT THIS IS UNNECESSARY BECAUSE PROG SVS DON'T HAVE RESTRICTIONS
//      ON THEIR PREFIXES. /AR
// 	else if (t.op() instanceof Modality) {
// 	    ProgramSVCollector prgSVColl = new 
// 		ProgramSVCollector(t.javaBlock().program(), 
// 				   SLListOfSchemaVariable.EMPTY_LIST);
// 	    prgSVColl.start();
// 	    IteratorOfSchemaVariable it = prgSVColl.
// 		getSchemaVariables().iterator();
// 	    while (it.hasNext()) {
// 		SchemaVariable sv = it.next();
// 		SetOfSchemaVariable relevantBoundVars = 
// 		    removeNotFreeIn(sv);
// 		TacletPrefix prefix = prefixMap.get(sv);
// 		if (prefix==null || prefix.prefix().equals(relevantBoundVars)) {
// 		    setPrefixOfOccurrence(sv, relevantBoundVars);
// 		} else {
// 		    throw new InvalidPrefixException(sv);
// 		}
// 	    }
// 	}
	for (int i=0; i<t.arity(); i++) {
	    SetOfSchemaVariable oldBounds=currentlyBoundVars;
	    addVarsBoundHere(t, i);
	    visit(t.sub(i));
	    currentlyBoundVars=oldBounds;
	}
    }
    

    private void visit(Sequent s) {
	IteratorOfConstrainedFormula it=s.iterator();
	while (it.hasNext()) {
	    visit(it.next().formula());
	}
    }

    private void visit(TacletGoalTemplate templ) {
	visit(templ.sequent());
	if (templ instanceof RewriteTacletGoalTemplate) {
	    visit(((RewriteTacletGoalTemplate)templ).replaceWith());
	}
	if (templ instanceof AntecSuccTacletGoalTemplate) {
	    visit(((AntecSuccTacletGoalTemplate)templ).replaceWith());
	}
    }

    public void build() {
	visit(tacletBuilder.ifSequent());
	if (tacletBuilder instanceof FindTacletBuilder) {
	    visit(((FindTacletBuilder)tacletBuilder).getFind());
	}
	IteratorOfTacletGoalTemplate itGoalTempl
	    =tacletBuilder.goalTemplates().iterator();
	while (itGoalTempl.hasNext()) {
	    visit(itGoalTempl.next());
	}
    }
    

    private boolean atMostOneRepl() {
	RewriteTacletBuilder rwtacletBuilder=(RewriteTacletBuilder)tacletBuilder;
	IteratorOfTacletGoalTemplate it
	    =rwtacletBuilder.goalTemplates().iterator();
	int count=0;
	while (it.hasNext()) {
	    TacletGoalTemplate tmpl = it.next();
	    if (tmpl instanceof RewriteTacletGoalTemplate) {
		if (((RewriteTacletGoalTemplate)tmpl).replaceWith()!=null) {
		    count++;
		}
	    }
	    if (count>1) return false;
	}
	return true;
    }

    private boolean occurrsOnlyInFindOrRepl(SchemaVariable sv) {
	RewriteTacletBuilder rwtacletBuilder=(RewriteTacletBuilder)tacletBuilder;
	TacletVariableCollector svc=new TacletVariableCollector();
	svc.visit(rwtacletBuilder.ifSequent());
	IteratorOfTacletGoalTemplate it
	    = rwtacletBuilder.goalTemplates().iterator();
	while (it.hasNext()) {
	    TacletGoalTemplate tmpl = it.next();
//	    if (tmpl instanceof RewriteTacletGoalTemplate) {	    
//		RewriteTacletGoalTemplate
//		    gt=(RewriteTacletGoalTemplate)tmpl; 
		svc.visit(tmpl.sequent());   //add rule ommited!!!!
//	    }
	}
	return !svc.contains(sv);
    }

    private void considerContext() {
	if (!(tacletBuilder instanceof RewriteTacletBuilder) || !atMostOneRepl()) {
	    return;
	}
	IteratorOfEntryOfSchemaVariableAndTacletPrefix 
	    it=prefixMap.entryIterator();
	while (it.hasNext()) {
	    EntryOfSchemaVariableAndTacletPrefix entry=it.next();
	    if (occurrsOnlyInFindOrRepl(entry.key())) {
		prefixMap = prefixMap.put(entry.key(), 
					entry.value().setContext(true));
	    }
	}
    }

    public MapFromSchemaVariableToTacletPrefix getPrefixMap() { 
	considerContext();
	return prefixMap;
    }
    
    class InvalidPrefixException extends IllegalStateException {
	InvalidPrefixException(SchemaVariable sv) {
	    super("Schema Variable "+sv+" has"
		  +" different Prefixes at"
		  +" different occurrences.");
	}
    }

}


