// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule;
import de.uka.ilkd.key.proof.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.java.visitor.ProgramSVCollector;
/** This class is used to collect all appearing schemavariables in a
 * Taclet. 
 * Duplicates are not removed because the use of persistent
 * datastructure and up to now we just have a SetAsList-implementaion
 * causing to have O(sqr(n)) if it would used. 
 * The class is used by RuleApps to compute all non-instantiated variables.
*/

public class TacletVariableCollector extends Visitor {

    /** collects all found variables */
    public ListOfSchemaVariable varList; //!!!!!
    /** the instantiations needed for unwind loop constructs */
    private SVInstantiations instantiations = 
	SVInstantiations.EMPTY_SVINSTANTIATIONS;

    /** creates the Variable collector.
     */
    public TacletVariableCollector() {
	varList = SLListOfSchemaVariable.EMPTY_LIST;
    }

    /** creates the Variable collector.
     * @param insts the SVInstantiations that have been already found
     * (needed by unwind loop constructs to determine which labels are needed)
     */
    public TacletVariableCollector(SVInstantiations svInsts) {
	varList = SLListOfSchemaVariable.EMPTY_LIST;
	instantiations = svInsts;
    }

    /** collects all SchemVariables that occur in the JavaBlock
     * @param jb the JavaBlock where to look for Schemavariables
     * @param vars the ListOfSchemaVariable where to add the found
     * SchemaVariables 
     * @return the extended list of found schemavariables 
     */
    private ListOfSchemaVariable collectSVInProgram
	(JavaBlock jb, ListOfSchemaVariable vars) {

	ProgramSVCollector prgSVColl = new 
	    ProgramSVCollector(jb.program(), vars, instantiations);
	prgSVColl.start();
	return prgSVColl.getSchemaVariables();
    }

    /** is called by the execPostOrder-method of a term 
     * @param Term to checked if it is a Variable and if true the
     * Variable is added to the list of found Variables
     */  
    public void visit(Term t) {	
	if (t.op() instanceof SchemaVariable) {
	    varList=varList.prepend((SchemaVariable)t.op());
	} else {
	    if (t.op() instanceof Modality) {
		varList = collectSVInProgram(t.javaBlock(),
					     varList);
	    }
	    for (int j=0; j<t.arity(); j++) {
		for (int i=0;i<t.varsBoundHere(j).size();i++) {
		    if (t.varsBoundHere(j).getQuantifiableVariable(i) 
			instanceof SchemaVariable) {
			varList = varList.prepend
			    ((SchemaVariable)t.varsBoundHere(j)
			     .getQuantifiableVariable(i));
		    }
		}
	    }
	}
    }
    
    /** @return iterator of the found Variables */
    public IteratorOfSchemaVariable varIterator() {
	return varList.iterator();
    }

    /** @return number of the found variables */
    public int size() {
	return varList.size();
    }

    /** @return true iff term contains the given variable */
    public boolean contains(SchemaVariable var) {
	return varList.contains(var);
    }

   
    
    /** collects all variables in a Semisequent 
     * @param semiseq the Semisequent to visit
     */
    private void visit(Semisequent semiseq) {
	IteratorOfConstrainedFormula it=semiseq.iterator();
	while(it.hasNext()) {
	    it.next().formula().execPostOrder(this);
	}
    }

    /** goes through the given sequent an collects all vars found
     * @param seq the Sequent to visit
     */
    public void visit(Sequent seq) {
	visit(seq.antecedent());
	visit(seq.succedent());
    }

    /** collects all variables in a Taclet
     * @param taclet the Taclet where the variables have to be collected to
     * @param visitAddRules a boolean that contols if the addrule sections are
     * to be ignored (iff false) or if the visitor descends into them (iff true) 
     */
    public void visit(Taclet taclet, boolean visitAddrules) {
	visit(taclet.ifSequent());
	if (taclet instanceof FindTaclet) {
	    (((FindTaclet)taclet).find()).execPostOrder(this);
	}	
	IteratorOfTacletGoalTemplate it = taclet.goalTemplates().iterator();
	while (it.hasNext()) {
	    TacletGoalTemplate gt=it.next();
	    visit(gt.sequent());
	    if (gt instanceof RewriteTacletGoalTemplate) {
		((RewriteTacletGoalTemplate)gt).replaceWith().execPostOrder(this);
	    } else {
		if(gt instanceof AntecSuccTacletGoalTemplate) {
		    visit(((AntecSuccTacletGoalTemplate)gt).replaceWith());
		}
	    }
	    if (visitAddrules) {
		IteratorOfTaclet addruleIt = gt.rules().iterator();
		while (addruleIt.hasNext()) {
		    visit(addruleIt.next(), true);		    
		}
	    }
	}
    }


    /** collects all variables in a Taclet but ignores the variables that appear
     * only in the addrule sections of the Taclet
     * @param taclet the Taclet where the variables have to be collected to
     */
    public void visitWithoutAddrule(Taclet taclet) {
	visit(taclet, false);
    }




}
