// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule;

import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.SLListOfGoal;
import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.java.expression.Literal;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.java.JavaInfo;

import java.util.*;
import java.text.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.rule.inst.*;
import de.uka.ilkd.key.java.visitor.JavaIterator;

/**
 * Built in rule that starts the update simplifier on the sequent or
 * selected term. The rule is useful if one has changes the simultaneous
 * update simplifier settings. 
 */
public class UpdateSimplificationRule implements BuiltInRule {
    

    private final Name name = new Name("Update Simplification");


    public UpdateSimplificationRule() {
    }

    
    private Term replace(Term term, Term with, 
			 IteratorOfInteger it) {	

	if (it.hasNext()) {	    
	    int sub=it.next().intValue();
	    Term[] subs=new Term[term.arity()];
	    for (int i=0;i<term.arity();i++) {
		if (i!=sub) {
		    subs[i]=term.sub(i);
		} else {
		    subs[i]=replace(term.sub(i), with, it);
		}
	    }
	    
	    ArrayOfQuantifiableVariable origvars 
		= (term.varsBoundHere(0).size()==0)
		? term.varsBoundHere(1) 
		: term.varsBoundHere(0);
	    QuantifiableVariable[] vars 
		= new QuantifiableVariable[origvars.size()];
	    for (int i=0; i<origvars.size(); i++) {
		vars[i] = origvars.getQuantifiableVariable(i);
	    }	    	    
	    return TermFactory.DEFAULT.createTerm(term.op(), subs, vars,
						  term.javaBlock(), 
						  term.ruleVar());	    
	} 

	return with;
    }

    private LinkedList applyOnTerm(SimultaneousUpdateSimplifier sus,
				   PosInOccurrence pio) {

	LinkedList result = new LinkedList();
	
	ConstrainedFormula cf = pio.constrainedFormula();
	Term fml = cf.formula();

	Term simplified = sus.simplify(fml.subAt(pio.posInTerm()));
	if (simplified != fml) {
	    ConstrainedFormula new_cf = new ConstrainedFormula
		(replace(fml, simplified, pio.posInTerm().iterator()), 
		 cf.constraint()); 
	    result.add(new ConstrainedFormulaContainer(cf, new_cf));
	}

	return result;
    }



    private LinkedList applyOnSequent(Sequent seq, 
				      SimultaneousUpdateSimplifier sus) {			

	LinkedList result = new LinkedList();

	IteratorOfConstrainedFormula it = seq.iterator();
	Term simplified;
	while (it.hasNext()) {
	    ConstrainedFormula cf = it.next();
	    simplified = sus.simplify(cf.formula());
	    if (simplified != cf.formula()) {
		result.add
		    (new ConstrainedFormulaContainer
			(cf, 
			 new ConstrainedFormula(simplified, cf.constraint()))
			);		
	    }
	}

	return result;
    }

    /** the rule is applied on the given goal using the
     * information of rule application. As applications of this rule kind may
     * not be successful in each case one has to ensure that the goal split is
     * done only iff the application was successful.
     * @param goal the goal that the rule application should refer to.
     * @param javaInfo the JavaInfo with the necessary information
     * about the java programs 
     * @param the rule application that is executed.
     */
    public ListOfGoal apply(Goal     goal, 
			    JavaInfo javaInfo, 
			    RuleApp  ruleApp) {

	ListOfGoal result = null;
	LinkedList applicationResult = null;
	
	PosInOccurrence pio = ruleApp.posInOccurrence();	    	
	if (pio == null) {		
	    applicationResult = applyOnSequent(goal.sequent(), goal.simplifier());
	    
	} else {
	    applicationResult = applyOnTerm(goal.simplifier(), pio);
	}	

	if (applicationResult != null && applicationResult.size() >0) { 
	    result = goal.split(1);
	    Goal newGoal = result.head();
	    Iterator it = applicationResult.iterator();
	    while (it.hasNext()) {
		ConstrainedFormulaContainer cfc = 
		    (ConstrainedFormulaContainer) it.next();
		
		newGoal.changeFormula(cfc.replacement(), 
				      new PosInOccurrence(cfc.old(), 
							  PosInTerm.TOP_LEVEL,
							  goal.node().proof()));		
	    }
	}
	return result;
    }

    /**
     * tests if the rule is applicable at the given position
     */
    public boolean isApplicable(Goal            goal, 
				PosInOccurrence pio,
				Constraint      userConstraint) {

	if (pio != null) {
	    Term fml = pio.constrainedFormula().formula().
		subAt(pio.posInTerm());
	    Term simplified = goal.simplifier().simplify(fml);
	    if (simplified != fml) {
		return true;
	    }
	} else { // this may be too slow; may be return true always
	    IteratorOfConstrainedFormula it = goal.sequent().iterator();
	    Term simplified;
	    while (it.hasNext()) {
		ConstrainedFormula cf = it.next();
		simplified = goal.simplifier().simplify(cf.formula());
		if (simplified != cf.formula()) {
		    return true;
		}
	    }
	}
	return false;
    }
    
    public Name name() {
	return name;
    }

    public String displayName() {
	return toString();
    }


    public String toString() {
	return name().toString();
    }

    // container for old constrained formula and the one used to replace it
    static class ConstrainedFormulaContainer {

	private ConstrainedFormula old;
	private ConstrainedFormula replacement;
	
	public ConstrainedFormulaContainer
	    (ConstrainedFormula old, ConstrainedFormula replacement) {
	    this.old = old;
	    this.replacement = replacement;
	}

	public ConstrainedFormula old() {
	    return old;
	}

	public ConstrainedFormula replacement() {
	    return replacement;
	}

    }
        
}
