// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule.inst;

import de.uka.ilkd.key.logic.sort.*;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.util.Debug;
import java.util.HashSet;
import java.util.Iterator;
import java.io.Serializable;
import java.io.ObjectStreamException;


/**
 * This class handles the instantiation of generic sorts (used for
 * generic taclets), i.e. the class tries to find a solution for the
 * conditions on the generic sorts given by the type hierarchy and the
 * chosen SV instantiations
 *
 * this class is immutable
 */

public class GenericSortInstantiations implements java.io.Serializable {

    public static final GenericSortInstantiations EMPTY_INSTANTIATIONS =
	new GenericSortInstantiations ( MapAsListFromGenericSortToSort.EMPTY_MAP );


    protected MapFromGenericSortToSort insts;


    private GenericSortInstantiations ( MapFromGenericSortToSort p_insts ) {
	insts = p_insts;
    }


    private Object readResolve() 
	throws ObjectStreamException {
	if (insts == MapAsListFromGenericSortToSort.EMPTY_MAP) {
	    return EMPTY_INSTANTIATIONS;
	}
	return this;
    }


    /**
     * Create an object that solves the conditions given by the
     * instantiation iterator, i.e. instantiations of the generic
     * sorts used within "p_instantiations" are sought for which are
     * compatible with the instantiations of the SVs
     * @param p_instantiations list of SV instantiations
     * @throws GenericSortException iff the conditions could not be
     * solved
     */
    public static GenericSortInstantiations create
	( IteratorOfEntryOfSchemaVariableAndInstantiationEntry p_instantiations )
	throws GenericSortException {
	return create ( p_instantiations,
			SLListOfGenericSortCondition.EMPTY_LIST );
    }


    /**
     * Create an object that solves the conditions given by the
     * instantiation iterator, i.e. instantiations of the generic
     * sorts used within "p_instantiations" are sought for which are
     * compatible with the instantiations of the SVs
     * @param p_instantiations list of SV instantiations
     * @param p_conditions additional conditions for sort instantiations
     * @throws GenericSortException iff the conditions could not be
     * solved
     */
    public static GenericSortInstantiations create
	( IteratorOfEntryOfSchemaVariableAndInstantiationEntry p_instantiations,
	  ListOfGenericSortCondition                           p_conditions )
	throws GenericSortException {

	ListOfGenericSort                          sorts      =
	    SLListOfGenericSort.EMPTY_LIST;
	GenericSortCondition                       c;
	IteratorOfGenericSortCondition             it;

	// Find the generic sorts within "p_instantiations" and
	// "p_conditions", create the conditions

	for ( it = p_conditions.iterator ();
	      it.hasNext (); )
	    sorts = sorts.prepend ( it.next ().getGenericSort () );

	while ( p_instantiations.hasNext () ) {
	    c = GenericSortCondition.createCondition
		( p_instantiations.next ().value () );
	    if ( c != null ) {
		p_conditions = p_conditions.prepend ( c );
		sorts        = sorts       .prepend ( c.getGenericSort () );
	    }
	}

	return create ( sorts, p_conditions );
    }


    /**
     * Create an object that holds instantiations of the generic sorts
     * "p_sorts" satisfying the conditions "p_conditions"
     * @param p_sorts generic sorts to instantiate
     * @param p_conditions conditions the instantiations have to
     * satisfy
     * @throws GenericSortException if no instantiations has been
     * found
     */
    public static GenericSortInstantiations create
	( ListOfGenericSort          p_sorts,
	  ListOfGenericSortCondition p_conditions )
	throws GenericSortException {

	if ( p_sorts == SLListOfEntryOfGenericSortAndSort.EMPTY_LIST )
	    return EMPTY_INSTANTIATIONS;

	return new GenericSortInstantiations ( solve ( p_sorts,
						       p_conditions ) );
    }


    /**
     * @return Boolean.TRUE if the sorts used within "p_entry" are
     * correct, Boolean.FALSE if the sorts are definitely incorrect,
     * null if the sorts could (perhaps) be made correct by choosing
     * the right generic sort instantiations
     */
    public Boolean checkSorts ( InstantiationEntry p_entry ) {
	if ( !( p_entry instanceof TermInstantiation ) ||
	     p_entry.getSchemaVariable ().isProgramSV () )
	    return Boolean.TRUE;

	GenericSortCondition c = GenericSortCondition
	    .createCondition ( p_entry );

	if ( c != null )
	    return checkCondition ( c );

	return new Boolean ( p_entry.getSchemaVariable ().sort () ==
			     ((TermInstantiation)p_entry).getTerm ().sort () );
    }


    /**
     * @return Boolean.TRUE if the generic sort instantiations within
     * "this" satisfy "p_condition", null otherwise (this means,
     * "p_condition" could be satisfied by create a new
     * "GenericSortInstantiations"-object)
     */
    public Boolean checkCondition ( GenericSortCondition p_condition ) {
	Sort s = insts.get ( p_condition.getGenericSort () );

	if ( s == null )
	    return null;

	if ( p_condition instanceof GenericSortCondition.GSCSupersort ) {
	    if ( ((GenericSortCondition.GSCSupersort)p_condition)
		 .getSubsort ().extendsTrans ( s ) )
		return Boolean.TRUE;
	} else if ( p_condition instanceof GenericSortCondition.GSCIdentity ) {
	    if ( ((GenericSortCondition.GSCIdentity)p_condition).getSort () == s )
		return Boolean.TRUE;
	} else if ( p_condition instanceof
		    GenericSortCondition.GSCForceInstantiation ) {
	    if ( isInstantiated ( p_condition.getGenericSort () ) )
		return Boolean.TRUE;
	}

	return null;
    }


    public boolean isInstantiated   ( GenericSort p_gs ) {
	return insts.containsKey ( p_gs );
    }


    public Sort    getInstantiation ( GenericSort p_gs ) {
	return insts.get ( p_gs );
    }


    /**
     * Create a list of conditions establishing the instantiations
     * stored by this object (not saying anything about further
     * generic sorts)
     */
    public ListOfGenericSortCondition toConditions () {
	ListOfGenericSortCondition          res  =
	    SLListOfGenericSortCondition.EMPTY_LIST;
	IteratorOfEntryOfGenericSortAndSort it   =
	    insts.entryIterator ();
	EntryOfGenericSortAndSort           entry;
	
	while ( it.hasNext () ) {
	    entry = it.next ();
	    res   = res.prepend ( GenericSortCondition.createIdentityCondition
				  ( entry.key (), entry.value () ) );
	}

	return res;
    }


    /**
     * @return p_s iff p_s is not a generic sort, the concrete sort
     * p_s is instantiated with currently otherwise 
     * @throws GenericSortException iff p_s is a generic sort which is
     * not yet instantiated
     */
    public Sort getRealSort ( SchemaVariable p_sv )
	throws GenericSortException {
	return getRealSort ( p_sv.sort () );
    }

    public Sort getRealSort ( Sort p_s )
	throws GenericSortException {
	if ( p_s instanceof GenericSort ) {
	    p_s = getInstantiation ( (GenericSort)p_s );
	    if ( p_s == null )
		throw new GenericSortException
		    ( "Generic sort is not yet instantiated" );
	} else if ( p_s instanceof CollectionSort ) {
	    Sort s = getRealSort ( ((CollectionSort)p_s).elementSort () );

	    s = ((CollectionSort)p_s).cloneFor ( s );

	    if ( s == null )
		throw new GenericSortException
		    ( "Generic collection sort is instantiated with a sort" +
		      " without collections" );

	    return s;
	}

	return p_s;
    }


    /**
     * Really solve the conditions given
     * @param p_sorts generic sorts that must be instantiated
     * @param p_conditions conditions to be solved
     * @throws GenericSortException no solution could be found
     * @return the/a found solution
     */
    protected static MapFromGenericSortToSort solve
	( ListOfGenericSort          p_sorts,
	  ListOfGenericSortCondition p_conditions )
	throws GenericSortException {

	MapFromGenericSortToSort res;

	// the generic sorts are sorted topologically, i.e. if sort A
	// is a supersort of sort B, then A appears behind B within
	// "topologicalSorts"
	ListOfGenericSort topologicalSorts = topology ( p_sorts );

	res = solveHelp ( topologicalSorts,
			  MapAsListFromGenericSortToSort.EMPTY_MAP,
			  p_conditions,
			  SLListOfGenericSort.EMPTY_LIST );

	if ( res == null )
	    throw new GenericSortException
		( "Conditions for generic sorts could not be solved" );

	return res;
    }


    /**
     * Method which is called recursively and tries to instantiate one
     * (the first) generic sort from the "p_remainingSorts"-list
     * @param p_remainingSorts generic sorts which needs to be
     * instantiated (topologically sorted)
     * @param p_curRes instantiations so far
     * @param p_conditions conditions (see above)
     * @return a solution if one could be found, null otherwise
     */
    protected static MapFromGenericSortToSort solveHelp
	( ListOfGenericSort          p_remainingSorts,
	  MapFromGenericSortToSort   p_curRes,
	  ListOfGenericSortCondition p_conditions,
	  ListOfGenericSort          p_pushedBack ) {

	if ( p_remainingSorts == SLListOfGenericSort.EMPTY_LIST ) {
	    return solveForcedInst ( p_pushedBack, p_curRes, p_conditions );
	} else {
	    // next generic sort to seek an instantiation for
	    GenericSort gs   = p_remainingSorts.head ();
	    p_remainingSorts = p_remainingSorts.tail ();

	    // Find the sorts "gs" has to be a supersort of and the
	    // identity conditions
	    ListOfSort                     subsorts     = SLListOfSort.EMPTY_LIST;
	    ListOfGenericSortCondition     idConditions =
		SLListOfGenericSortCondition.EMPTY_LIST;
	    GenericSortCondition           c;
	    IteratorOfGenericSortCondition itC          = p_conditions.iterator ();
	
	    // subsorts given by the conditions (could be made faster
	    // by using a hash map for storing the conditions)
	    while ( itC.hasNext () ) {
		c = itC.next ();
		if ( c.getGenericSort () == gs ) {
		    if ( c instanceof GenericSortCondition.GSCSupersort )
			subsorts = subsorts.prepend
			    ( ((GenericSortCondition.GSCSupersort)c)
			      .getSubsort () );
		    else if ( c instanceof GenericSortCondition.GSCIdentity )
			idConditions = idConditions.prepend ( c );
		}
	    }

	    // add the subsorts given by the already instantiated
	    // generic sorts
            IteratorOfEntryOfGenericSortAndSort itEntry = p_curRes.entryIterator ();
            EntryOfGenericSortAndSort           entry;
	    while ( itEntry.hasNext () ) {
		entry = itEntry.next ();
		if ( entry.key ().extendsTrans ( gs ) )
		    subsorts = subsorts.prepend ( entry.value () );
	    }

	    // Solve the conditions
	    IteratorOfSort           itChosen = null;
	    Sort                     chosen;
	    MapFromGenericSortToSort res;

	    if ( idConditions != SLListOfGenericSortCondition.EMPTY_LIST ) {
		// then the instantiation is completely determined by
		// an identity condition
		itC    = idConditions.iterator ();
		chosen =
		    ((GenericSortCondition.GSCIdentity)itC.next ()).getSort ();
		
		// other identity conditions must lead to the same
		// instantiation
		while ( itC.hasNext () ) {
		    if ( chosen !=
			 ( ((GenericSortCondition.GSCIdentity)itC.next ())
			   .getSort () ) )
			return null;
		}

		itChosen = SLListOfSort.EMPTY_LIST.prepend ( chosen ).iterator ();
	    } else {
		// if a list of possible instantiations of the generic
		// sort has been given, use it
		itChosen = gs.getOneOf ().iterator ();

		if ( !itChosen.hasNext () )
		    itChosen = null;
	    }

	    if ( itChosen != null ) {
		while ( itChosen.hasNext () ) {
		    chosen = itChosen.next ();
		    if ( gs.isPossibleInstantiation ( chosen ) &&
			 isSupersortOf ( chosen, subsorts ) ) {
			res = solveHelp ( p_remainingSorts,
					  p_curRes.put ( gs, chosen ),
					  p_conditions,
					  p_pushedBack );
			if ( res != null )
			    return res;
		    }
		}
	    } else if ( subsorts != SLListOfSort.EMPTY_LIST ) {
		// if anything else has failed, construct minimal
		// supersorts of the found subsorts and try them
		itChosen = minimalSupersorts ( subsorts.iterator () );
		while ( itChosen.hasNext () ) {
		    chosen = itChosen.next ();

		    // should not be necessary
		    Debug.assertFalse ( itChosen.hasNext (),
					"Common supersorts are not unique: " +
					subsorts );

		    if ( gs.isPossibleInstantiation ( chosen ) ) {
			res = solveHelp ( p_remainingSorts,
					  p_curRes.put ( gs, chosen ),
					  p_conditions,
					  p_pushedBack );
			if ( res != null )
			    return res;
		    }
		}
	    } else {
		// and if even that did not work, remove the generic
		// sort from the list and try again later
		return solveHelp ( p_remainingSorts,
				   p_curRes,
				   p_conditions,
				   p_pushedBack.prepend ( gs ) );
	    }
	}

	return null;
    }


    /**
     * Method which is called by solveHelp and tries to instantiate
     * the generic sorts for which GSCForceInstantiation-conditions
     * (with "maximum" parameter) are contained within "p_conditions"
     * @param p_curRes instantiations so far
     * @param p_conditions conditions (see above)
     * @return a solution if one could be found, null otherwise
     */
    protected static MapFromGenericSortToSort solveForcedInst
	( ListOfGenericSort          p_remainingSorts,
	  MapFromGenericSortToSort   p_curRes,
	  ListOfGenericSortCondition p_conditions ) {

	if ( p_remainingSorts == SLListOfGenericSort.EMPTY_LIST )
	    return p_curRes; // nothing further to be done

	IteratorOfGenericSort it = topology ( p_remainingSorts ).iterator ();
	p_remainingSorts         = SLListOfGenericSort.EMPTY_LIST;

	// reverse the order of the sorts, to start with the most
	// general one
	while ( it.hasNext () )
	    p_remainingSorts = p_remainingSorts.prepend ( it.next () );

	return solveForcedInstHelp ( p_remainingSorts,
				     p_curRes );
    }


    /**
     * Method which is called recursively and tries to instantiate one
     * (the first) generic sort from the "p_remainingSorts"-list
     * @param p_remainingSorts generic sorts which needs to be
     * instantiated (topologically sorted, starting with the most
     * general sort)
     * @param p_curRes instantiations so far
     * @return a solution if one could be found, null otherwise
     */
    protected static MapFromGenericSortToSort solveForcedInstHelp
	( ListOfGenericSort        p_remainingSorts,
	  MapFromGenericSortToSort p_curRes ) {
	if ( p_remainingSorts == SLListOfGenericSort.EMPTY_LIST ) {
	    // we're done
	    return p_curRes;
	} else {
	    // next generic sort to seek an instantiation for
	    GenericSort gs   = p_remainingSorts.head ();
	    p_remainingSorts = p_remainingSorts.tail ();

	    IteratorOfSort           it;
	    Sort                     cur;
	    MapFromGenericSortToSort res;
	    HashSet                  todo   = new HashSet ();
	    HashSet                  done   = new HashSet ();
	    Sort                     cand;

	    // search for instantiated supersorts of gs (the method
	    // below is neither fast nor complete, but should find a
	    // solution for relevant situations)

	    // insert into the working list (set)
	    it = gs.extendsSorts ().iterator ();
	    while ( it.hasNext () )
		todo.add ( it.next () );

	    while ( !todo.isEmpty () ) {
		it  = null;
		cur = (Sort)todo.iterator ().next ();
		todo.remove ( cur );
		done.add    ( cur );

		if ( cur instanceof GenericSort ) {
		    cand = p_curRes.get ( (GenericSort)cur );
		    if ( cand == null )
			it = cur.extendsSorts ().iterator ();
		} else {
		    it   = cur.extendsSorts ().iterator ();
		    cand = cur;
		}

		if ( cand != null &&
		     isPossibleInstantiation ( gs, cand, p_curRes ) ) {
		    res = solveForcedInstHelp ( p_remainingSorts,
						p_curRes.put ( gs, cand ) );
		    if ( res != null )
			return res;
		}

		if ( it != null ) {
		    while ( it.hasNext () ) {
			cur = it.next ();
			if ( !done.contains ( cur ) )
			    todo.add ( cur );
		    }
		}
	    }
	}

	return null;
    }


    /**
     * Sort generic sorts topologically, i.e. if sort A is a supersort
     * of sort B, then A appears behind B within the return value
     * @return sorted sorts
     */
    protected static ListOfGenericSort topology ( ListOfGenericSort p_sorts ) {
	ListOfGenericSort     res     = SLListOfGenericSort.EMPTY_LIST;
	IteratorOfGenericSort it;
	GenericSort           curMax;
	GenericSort           tMax;
	ListOfGenericSort     tList;

	while ( p_sorts != SLListOfGenericSort.EMPTY_LIST ) {
	    // search for a maximal element
	    it      = p_sorts.iterator ();
	    curMax  = it.next ();

	    if ( res.contains ( curMax ) ) {
		p_sorts = p_sorts.tail ();
		continue;
	    }

	    tList   = SLListOfGenericSort.EMPTY_LIST;

	    while ( it.hasNext () ) {
		tMax = it.next ();
		if ( !res.contains ( tMax ) ) {
		    if ( curMax.extendsTrans ( tMax ) ) {
			tList  = tList.prepend ( curMax );
			curMax = tMax;
		    } else
			tList  = tList.prepend ( tMax );
		}
	    }

	    res     = res.prepend ( curMax );
	    p_sorts = tList;
	}

	return res;
    }

    /**
     * Find all minimal common supersorts of "p_itSorts"
     *
     * PRECONDITION: p_itSorts.hasNext ()
     */
    protected static IteratorOfSort minimalSupersorts
	( IteratorOfSort p_itSorts ) {
	HashSet        inside    = new HashSet ();
	HashSet        outside   = new HashSet ();
	HashSet        todo      = new HashSet ();
	HashSet        t;
	Sort           condition;
	Sort           candidate;
	IteratorOfSort itSort;
	boolean        checkNULL = false;

	// the null sort has to be handled separately, as it doesn't
	// provide a list of direct supersorts
	inside.add ( candidate = p_itSorts.next () );
	while ( candidate == Sort.NULL && p_itSorts.hasNext () ) {
	    inside.clear ();
	    inside.add ( candidate = p_itSorts.next () );
	    checkNULL = true;
	}

	// Find a set "inside" of common supersorts of "p_itSorts"
	// that contains all minimal common supersorts
	while ( !inside.isEmpty () && p_itSorts.hasNext () ) {
	    condition = p_itSorts.next ();

	    // At this point todo.isEmpty ()
	    t         = todo;
	    todo      = inside;
	    inside    = t;

	    while ( !todo.isEmpty () ) {
		candidate = (Sort)todo.iterator ().next ();
		todo.remove ( candidate );

		if ( !inside .contains ( candidate ) &&
		     !outside.contains ( candidate ) ) {
		    if ( condition.extendsTrans ( candidate ) )
			inside .add ( candidate );
		    else {
			outside.add ( candidate );			
			itSort = candidate.extendsSorts ().iterator ();
			while ( itSort.hasNext () )
			    todo.add ( itSort.next () );
		    }
		}
	    }
	}

	if ( checkNULL ) {
	    Iterator it = inside.iterator ();
	    // At this point todo.isEmpty ()
	    inside      = todo;

	    while ( it.hasNext () ) {
		candidate = (Sort)it.next ();
		if ( Sort.NULL.extendsTrans ( candidate ) )
		    inside.add ( candidate );
	    }
	}

	// Find the minimal elements of "inside"
	ListOfSort     res       = SLListOfSort.EMPTY_LIST;
	
	if ( inside.size () == 1 )
	    res = res.prepend ( (Sort)inside.iterator ().next () );
	else {
	    ListOfSort     res2      = SLListOfSort.EMPTY_LIST;
	    Iterator       it;

	    it = inside.iterator ();
	    while ( it.hasNext () ) {
		candidate = (Sort)it.next ();

		res2      = SLListOfSort.EMPTY_LIST;
		itSort    = res.iterator ();
		while ( itSort.hasNext () ) {
		    condition = itSort.next ();

		    if ( condition.extendsTrans ( candidate ) ) {
			candidate = null;
			break;
		    } else if ( !candidate.extendsTrans ( condition ) )
			res2 = res2.prepend ( condition );
		}

		if ( candidate != null )
		    res = res2.prepend ( candidate );
	    }
	}

	return res.iterator ();
    }


    /**
     * @return true iff "p_s" is supersort of every sort of
     * "p_subsorts"
     */
    protected static boolean isSupersortOf ( Sort       p_s,
					     ListOfSort p_subsorts ) {
	IteratorOfSort it = p_subsorts.iterator ();

	while ( it.hasNext () ) {
	    if ( !it.next ().extendsTrans ( p_s ) )
		return false;
	}

	return true;
    }


    /**
     * @return true iff "p_s" is a valid instantiation of the generic
     * sort "p_gs", and this instantiation is consistent with
     * previously chosen instantiations
     */
    protected static boolean isPossibleInstantiation
	( GenericSort                p_gs,
	  Sort                       p_s,
	  MapFromGenericSortToSort   p_curRes) {
	if ( !p_gs.isPossibleInstantiation ( p_s ) )
	    return false;

	// check whether the new instantiation is consistent with the
	// already chosen instantiations
	IteratorOfEntryOfGenericSortAndSort itEntry = p_curRes.entryIterator ();
	EntryOfGenericSortAndSort           entry;
	while ( itEntry.hasNext () ) {
	    entry = itEntry.next ();
	    if ( entry .key   ().extendsTrans ( p_gs ) &&
		 !entry.value ().extendsTrans ( p_s  ) ||
		 p_gs.extendsTrans ( entry.key   () ) &&
		 !p_s.extendsTrans ( entry.value () ) )
		return false;
	}
	
	return true;
    }


    /** toString */
    public String toString () {
	IteratorOfEntryOfGenericSortAndSort it  = insts.entryIterator ();
	EntryOfGenericSortAndSort           entry;
	String                              res = "";

	while ( it.hasNext () ) {
	    if ( !"".equals ( res ) )
		res += ", ";
	    entry  = it.next ();
	    res   += entry.key () + "=" + entry.value ();
	}

	return res;
    }


    /**
     * ONLY FOR JUNIT TESTS
     */

    public MapFromGenericSortToSort getAllInstantiations () {
	return insts;
    }

}
