package de.uka.ilkd.key.rule.metaconstruct;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.abstraction.ClassType;
import de.uka.ilkd.key.java.abstraction.Constructor;
import de.uka.ilkd.key.java.abstraction.DefaultConstructor;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.abstraction.ListOfKeYJavaType;
import de.uka.ilkd.key.java.abstraction.SLListOfKeYJavaType;
import de.uka.ilkd.key.java.declaration.ArrayOfFieldSpecification;
import de.uka.ilkd.key.java.declaration.ArrayOfMemberDeclaration;
import de.uka.ilkd.key.java.declaration.ArrayOfParameterDeclaration;
import de.uka.ilkd.key.java.declaration.ClassDeclaration;
import de.uka.ilkd.key.java.declaration.ClassInitializer;
import de.uka.ilkd.key.java.declaration.ConstructorDeclaration;
import de.uka.ilkd.key.java.declaration.FieldDeclaration;
import de.uka.ilkd.key.java.declaration.FieldSpecification;
import de.uka.ilkd.key.java.declaration.MemberDeclaration;
import de.uka.ilkd.key.java.declaration.MethodDeclaration;
import de.uka.ilkd.key.java.declaration.Modifier;
import de.uka.ilkd.key.java.declaration.ParameterDeclaration;
import de.uka.ilkd.key.java.declaration.modifier.*;
import de.uka.ilkd.key.java.reference.*;
import de.uka.ilkd.key.java.statement.*;

import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.op.ProgramMethod;

import java.util.List;
import java.util.LinkedList;
import de.uka.ilkd.key.util.Debug;

/**
 * The constructor body normalform meta construct is used to
 * insert the body of a constructor, for an implicit or explicit
 * constructor call. Therefore the body has to be transformed in its
 * normalform, so that the creation and initialization of the object
 * is made explicit and executed as described in the JLS. 
 */
public abstract class ConstructorBodyNormalform extends ProgramMetaConstruct {
    
    private final ProgramElementName initName = new ProgramElementName("<init>");

    public ConstructorBodyNormalform(Name name, ProgramElement consRef) {
	super(name, consRef);
    }

    /**
     * in a given list of statements, which represent the body of the current
     * constructor, an eventually existing <code>super(..)</code> or
     * <code>this(..)</code> constructor reference is replaced by its resolved
     * version. If no such call exists, and the enclosing class is not Object
     * the resolved default constructor call <code> super(); </code>
     */
    private Statement getResolvedConstructorReference
	(StatementBlock body, ExecutionContext executionContext, 
	 JavaInfo javaInfo) {
	
	Statement firstStatement = (body.getChildCount()>0 ? 
				    (Statement) body.getChildAt(0) : null);
      	Statement resolvedReference = firstStatement;

	if (firstStatement instanceof SpecialConstructorReference) {
	    SpecialConstructorReference reference =
		(SpecialConstructorReference) firstStatement;	    
	    	    
	    if (firstStatement instanceof ThisConstructorReference) {		
		resolvedReference = new ThisConstructorReference
		    (reference.getArguments(), executionContext);
	    } else {
		resolvedReference = new SuperConstructorReference
		    (((SuperConstructorReference)reference).getReferencePrefix(), 
		     reference.getArguments(), executionContext);
	    }
	} else if (executionContext.getTypeReference().getKeYJavaType() != 
		   javaInfo.getObjectKeYJavaType()) {
	    // first statement is not super(..) or this(..)
	    // so we have to create the default constructor call super()
	    resolvedReference = new SuperConstructorReference(executionContext);
	}	
	
	return resolvedReference;
    }

    /**
     * retrieves the object initializers (including field initializers) with
     * respect to their occurrence in the source code
     * @return the object initializers (including field initializers) with
     * respect to their occurrence in the source code
     */
    public LinkedList getObjectInitializers(ClassDeclaration parent) {
	LinkedList initializers = new LinkedList();
	ArrayOfMemberDeclaration classMembers = parent.getMembers();
	// collect initializers (field and non-static class initializers)
	for (int i = 0; i<classMembers.size(); i++) {
	    MemberDeclaration member = classMembers.getMemberDeclaration(i);
	    if (member instanceof ClassInitializer) {
		initializers.addLast(((ClassInitializer)member).getBody());
	    } else if (member instanceof FieldDeclaration) {
		ArrayOfFieldSpecification specs = 
		    ((FieldDeclaration)member).getFieldSpecifications();
		for (int j = 0; j<specs.size(); j++) {
		    FieldSpecification field = specs.getFieldSpecification(j); 
		    if (!field.isStatic() && field.getInitializer()!=null) {
			initializers.addLast
			    (assign(new FieldReference
				((ProgramVariable)field.getProgramVariable(), 
				 new ThisReference()), 
				    field.getInitializer()));
		    }
		}
	    }
	}
	return initializers;
    }

    /** 
     * inserts the given statements at the beginning of the block and
     * remove first nr statements 
     * @param stmnt array of Statement those have to be inserted
     * @param b the Statementblock where to insert 
     * @param nr int number of statements to be removed
     * @return statement block starting with statements of stmnt followed by
     * those of the given statement block b starting at its second statement
     */
    private StatementBlock replaceFirstStatementInBlock(Statement[] stmnt, 
							StatementBlock b,
							int nr) {
	
	int blockSize = b.getStatementCount();
	Statement[] block = new Statement
	    [(blockSize>0 ? blockSize-nr : 0)+stmnt.length];

	for (int j = 0; j < stmnt.length; j++) {
	    block[j] = stmnt[j];
	}

	for (int i = nr; i < b.getStatementCount(); i++) {
	    block[(i-nr)+stmnt.length] = b.getStatementAt(i);
	}

	return new StatementBlock(new ArrayOfStatement(block));	
    }

    /** 
     * retrieves the most specific constructor wrt. the arguments.
     * The body of the constructor is transformed into its normal form and then
     * the new body is packed as a new implict method <code>&lt;init&gt</code> and a
     * reference to this method is returned.
     * @param arguments an ArrayOfExpression with the arguments handed over to
     * the constructor
     * @param executionContext the ExecutionContext of the constructor reference
     * @param javaInfo the JavaInfo object with useful information about the
     * java model
     * @return the method reference to the new created implicit method
     */
    public MethodReference createConstructorReference
	(ArrayOfExpression arguments, ExecutionContext executionContext, 
	 JavaInfo javaInfo) {

	TypeReference executionContextType = executionContext.getTypeReference();
	KeYJavaType classType = executionContextType.getKeYJavaType();

	Constructor constructor = javaInfo.getKeYProgModelInfo().getConstructor
	    (classType, javaInfo.createSignature(arguments));
	
	ArrayOfParameterDeclaration parameters;
	StatementBlock constructorBody;

	if (constructor instanceof ConstructorDeclaration) {
	    constructorBody = ((ConstructorDeclaration)constructor).getBody();
	    parameters = ((ConstructorDeclaration)constructor).getParameters();
	} else if (constructor instanceof DefaultConstructor) {
	    constructorBody = new StatementBlock
		(new ArrayOfStatement(new Statement[0]));
	    parameters = new ArrayOfParameterDeclaration(new ParameterDeclaration[0]);
	} else {
	    // no implementation available
	    // but perhaps Object default constructor?
	    if (classType == javaInfo.getObjectKeYJavaType()) {
		constructorBody = new StatementBlock
		    (new ArrayOfStatement(new Statement[0]));
		parameters = new ArrayOfParameterDeclaration(new ParameterDeclaration[0]);
		
		//wrap the new constructor in a method
		MethodDeclaration md = new MethodDeclaration
		    (new Modifier[]{new Private()}, 
		     executionContextType, initName, parameters, null, 
		     constructorBody, false);
		
		return new MethodReference(arguments, initName, 
					   executionContext.getRuntimeInstance(), 
					   new ProgramMethod(md, classType, classType));
		
	    } else {
		return null;
	    }
	}

	// special constructor reference + object initializers
	ClassDeclaration classDecl = 
	    (ClassDeclaration)classType.getJavaType();
	Statement consRef = 
	    getResolvedConstructorReference(constructorBody, 
					    executionContext, javaInfo);
	
	LinkedList preambleList;
	if (consRef instanceof ThisConstructorReference) {
	    preambleList = new LinkedList();
	} else {
	    preambleList = getObjectInitializers(classDecl);
	}

	if (consRef != null) {
	    preambleList.addFirst(consRef);
	}

	// put initializers in front of the constructors body
	if (constructorBody.getChildCount()>0 && 
	    constructorBody.getChildAt(0) instanceof SpecialConstructorReference) {
	    constructorBody = replaceFirstStatementInBlock
		((Statement[])preambleList.toArray(new Statement[preambleList.size()]), 
		 constructorBody, 1);
	} else {
	    constructorBody = replaceFirstStatementInBlock
		((Statement[])preambleList.toArray(new Statement[preambleList.size()]), 
		 constructorBody, 0);		
	}
	

	
	// wrap the new constructor in a method
       	MethodDeclaration md = new MethodDeclaration
	    (new Modifier[]{new Private()}, executionContextType, initName, 
	    parameters, null, constructorBody, false);
	
	return new MethodReference(arguments, initName, 
				   executionContext.getRuntimeInstance(), 
				   new ProgramMethod(md, classType, classType));
    }
    

}

