
package de.uka.ilkd.key.rule.metaconstruct;

import de.uka.ilkd.key.java.*;
import de.uka.ilkd.key.java.abstraction.ArrayType;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.abstraction.SLListOfType;
import de.uka.ilkd.key.java.abstraction.SLListOfField;
import de.uka.ilkd.key.java.declaration.LocalVariableDeclaration;
import de.uka.ilkd.key.java.expression.ArrayInitializer;
import de.uka.ilkd.key.java.expression.literal.BooleanLiteral;
import de.uka.ilkd.key.java.expression.literal.IntLiteral;
import de.uka.ilkd.key.java.expression.operator.LessThan;
import de.uka.ilkd.key.java.expression.operator.LogicalOr;
import de.uka.ilkd.key.java.expression.operator.New;
import de.uka.ilkd.key.java.expression.operator.NewArray;
import de.uka.ilkd.key.java.expression.operator.PostIncrement;
import de.uka.ilkd.key.java.reference.ArrayReference;
import de.uka.ilkd.key.java.reference.MethodReference;
import de.uka.ilkd.key.java.reference.ReferencePrefix;
import de.uka.ilkd.key.java.reference.TypeRef;
import de.uka.ilkd.key.java.reference.TypeReference;
import de.uka.ilkd.key.java.statement.*;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import de.uka.ilkd.key.logic.op.SchemaVariable;
import de.uka.ilkd.key.rule.inst.SVInstantiations;

import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.util.ExtList;
import java.util.LinkedList;

/**
 * Split an array creation expression with explicit array initializer,
 * creating a creation expression with dimension expression and a list
 * of assignments (-> Java language specification, 15.10)
 *
 * This meta construct delivers the creation expression
 */
public class InitArrayCreation extends InitArray {

    private final SchemaVariable newObjectSV;

    public InitArrayCreation(SchemaVariable newObjectSV,
			     ProgramElement newExpr) {	
	super("init-array-creation", newExpr); 
	this.newObjectSV = newObjectSV;
    }

    private If checkNegativeDimension(Expression cond, 
				      JavaInfo javaInfo) {
	Then then = new Then
	    (new Throw(new New(new Expression[0], 
			       new TypeRef 
				   (javaInfo.getKeYJavaType
				    ("java.lang.NegativeArraySizeException")), 
			       null)));
			     
	return new If(cond, then);
    }


    /**
     * creates statements for 
     *   <ol> 
     *     <li> evaluation of the dimension expressions, <li>
     *     <li> check if a dimension is non-negative </li> 
     *  </ol>
     * and adds them to given list of statements. Further more the new
     * declared program variables initialised with the evaluated
     * dimension expressions are returned
     * @param bodyStmnts the LinkedList of statements where the new
     * statements are inserted
     * @param dimExpr the ArrayOfExpression which describe the array's
     * dimensions
     * @param javaInfo the JavaInfo object
     */
    private ProgramVariable[] evaluateAndCheckDimensionExpressions
	(LinkedList bodyStmnts, ArrayOfExpression dimExpr, 
	 JavaInfo javaInfo) {
	KeYJavaType integerType = javaInfo.getPrimitiveKeYJavaType("int");

	IntLiteral zero = new IntLiteral(0);
	Expression checkDimensions = BooleanLiteral.FALSE;
	ProgramVariable[] pvars = new ProgramVariable[dimExpr.size()];
	for (int i = 0; i<pvars.length; i++) {
	    LocalVariableDeclaration argDecl = 
		declare("dim_"+(TEMPVARCOUNTER++)+"_"+i, 
			dimExpr.getExpression(i), integerType);
	    pvars[i] = (ProgramVariable)argDecl.getVariables().
		getVariableSpecification(0).getProgramVariable();
	    bodyStmnts.add(argDecl);	    
	    LessThan negativeDimension = new LessThan(pvars[i], zero);
	    if (i == 0) {
		checkDimensions = negativeDimension;
	    } else {
		checkDimensions = new LogicalOr(checkDimensions, 
						negativeDimension);
	    }
	}

	bodyStmnts.add(checkNegativeDimension(checkDimensions, javaInfo));

	return pvars;
    }

    private void createNDimensionalArray(LinkedList bodyStmnts,
					 Expression resultVar,
					 KeYJavaType arrayType,
					 ProgramVariable[] dimensions, 
					 JavaInfo javaInfo) {
	KeYJavaType integerType = javaInfo.getPrimitiveKeYJavaType("int");
 	MethodReference mr = new MethodReference
	    (new ArrayOfExpression(new Expression[]{dimensions[0]}), 
	     new ProgramElementName("<createArray>"), 
	     new TypeRef(arrayType), 
	     javaInfo.getKeYProgModelInfo().getProgramMethod
	     (arrayType, "<createArray>", 
	      SLListOfType.EMPTY_LIST.prepend(integerType))); 

	bodyStmnts.add(assign(resultVar, mr));

	if (dimensions.length > 1) {
	    Expression[] baseDim = new Expression[dimensions.length-1];
	    System.arraycopy(dimensions, 1, baseDim, 
			     0, dimensions.length-1);
	    IntLiteral zero = new IntLiteral(0);	    
	    LocalVariableDeclaration forInit = declare("i"+(TEMPVARCOUNTER++), 
						       zero, integerType);		

	    ProgramVariable pv = (ProgramVariable)forInit.getVariables().
		    getVariableSpecification(0).getProgramVariable();	    	    
	    TypeReference baseTypeRef = 
		((ArrayType)arrayType.getJavaType()).getBaseType();
	    KeYJavaType baseType = baseTypeRef.getKeYJavaType();
	    For forLoop = new For
		(new LoopInitializer[]{forInit},
		 new LessThan(pv, dimensions[0]),
		 new Expression[]{new PostIncrement(pv)},
		 (Statement)assign
		 (new ArrayReference
		     ((ReferencePrefix)resultVar,
		      new Expression[]{pv}), 
		  new NewArray(baseDim, baseTypeRef, 
			       baseType, null, 
			       ((ArrayType)baseType.getJavaType()).
			       getDimension())));
	    bodyStmnts.add(forLoop);
	}

    }

    /**
     * executes an array creation without initializers involved 
     */
    private ProgramElement arrayCreationWithoutInitializers
	(Expression newObject, NewArray na, JavaInfo javaInfo) {
	
	LinkedList bodyStmnts = new LinkedList();	

	ProgramVariable[] dimensions = 
	    evaluateAndCheckDimensionExpressions
	    (bodyStmnts, na.getArguments(), javaInfo);

 	KeYJavaType arrayType = 
	    na.getKeYJavaType(javaInfo.getServices());

	createNDimensionalArray(bodyStmnts, 
				newObject,
				arrayType, 
				dimensions, 
				javaInfo);

 	return new StatementBlock((Statement[])bodyStmnts.toArray
				  (new Statement[bodyStmnts.size()]));
	
    }
    
    public ProgramElement symbolicExecution(ProgramElement pe, 
					    JavaInfo javaInfo,
					    SVInstantiations svInst) {


	Expression array = (Expression)
	    svInst.getInstantiation ( newObjectSV );

	NewArray na = null;

	if ( pe instanceof NewArray ) {
	    na = (NewArray) pe;
	    if (na.getArrayInitializer() == null) {
		return arrayCreationWithoutInitializers
		    (array, na, javaInfo);
	    }
	} else if (pe instanceof ArrayInitializer) {
	    ArrayInitializer init = (ArrayInitializer)pe;
	    KeYJavaType kjt = array.getKeYJavaType(javaInfo.getServices());	    
	    na = new NewArray(new Expression[0], 
			      ((ArrayType)kjt.getJavaType()).getBaseType(),
			      kjt, init, 0);
	} else {
	    return pe;
	}    
	    
	int arrayLength = na.getArrayInitializer().getArguments().size();
	Statement[] body = new Statement[2*arrayLength+1];	    
	
	ProgramVariable[] vars = evaluateInitializers ( body, na );
	
	body[arrayLength] = 
	    assign ( array, createArrayCreation ( na ));
	
	createArrayAssignments ( arrayLength + 1, 
				 body, 
				 vars,  
				 array,
				 na );

	return new StatementBlock ( body );
    }
    

}
