// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule.metaconstruct;

import de.uka.ilkd.key.java.Expression;
import de.uka.ilkd.key.java.JavaNonTerminalProgramElement;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.JavaProgramElement;
import de.uka.ilkd.key.java.PrettyPrinter;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.Services;
import de.uka.ilkd.key.java.SourceElement;
import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.java.StatementContainer;
import de.uka.ilkd.key.java.abstraction.KeYJavaType;
import de.uka.ilkd.key.java.declaration.LocalVariableDeclaration;
import de.uka.ilkd.key.java.declaration.VariableSpecification;
import de.uka.ilkd.key.java.expression.operator.CopyAssignment;
import de.uka.ilkd.key.java.reference.PackageReference;
import de.uka.ilkd.key.java.reference.FieldReference;
import de.uka.ilkd.key.java.reference.ReferencePrefix;
import de.uka.ilkd.key.java.reference.TypeReference;
import de.uka.ilkd.key.java.reference.TypeRef;
import de.uka.ilkd.key.java.statement.JavaStatement;
import de.uka.ilkd.key.java.visitor.Visitor;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.logic.Name;
import de.uka.ilkd.key.logic.ProgramElementName;
import de.uka.ilkd.key.logic.op.ProgramVariable;
import java.io.IOException;

/** This class is used to perform program transformations needed in
 * Taclets to be able to do symbolic program execution. The actual
 * implementation of these transformations is done by subclasses.
 * The general structure is the one of statement container that
 * contains exact one statement, e.g. unwind-loop( loop ) or exact one
 * expression.  
 */
public abstract class ProgramMetaConstruct extends JavaNonTerminalProgramElement
    implements StatementContainer, Statement, Expression, TypeReference {

    /** the name of the meta construct */
    private Name name;
    /** the encapsulated program element */
    private ProgramElement body;

    /** creates a ProgramMetaConstruct 
     * @param name the Name of the meta construct 
     * @param body the ProgramElement contained by the meta construct 
     */
    public ProgramMetaConstruct(Name name, ProgramElement body) {
	this.name = name;
	this.body = body;
    }

    /** creates a ProgramMetaConstruct 
     * @param name the String with the name of the meta construct
     * @param body the ProgramElement contained by the meta construct 
     */
    public ProgramMetaConstruct(String name, ProgramElement body) {
	this(new Name(name), body);
    }

    /** performs the program transformation needed for symbolic
     * program transformation 
     * @param pe the ProgramElement on which the execution is performed
     * @param javaInfo the JavaInfo with all necessary information 
     * about the java programs
     * @param svInst the instantiations of the schemavariables 
     * @return the transformated program
     */
    public abstract ProgramElement symbolicExecution
	(ProgramElement pe, JavaInfo javaInfo,
	 SVInstantiations svInst);
    
    /** returns the name of the meta construct 
     * @return the name of the meta construct 
     */
    public Name name() {
	return name;
    }

    /** returns the body of the meta construct 
     * @return the body of the meta construct 
     */
    public ProgramElement body() {
	return body;
    }

    /**
     * Finds the source element that occurs last in the source.
     * @return the last source element in the syntactical representation of
     * this element, may be equals to this element.
    */
    public SourceElement getLastElement() {
        return (body != null) ? body : (SourceElement)this;
    }

    /**
     * Get the number of statements in this container.
     * @return the number of statements.
     */
    public int getStatementCount() {
	return (body instanceof Statement ? 1 : 0);
    }


    /*
     * Return the statement at the specified index in this node's
     * "virtual" statement array.
     * @param index an index for a statement.
     * @return the statement with the given index.
     * @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     * of bounds.
     */
    public Statement getStatementAt(int index) {
	if (index == 0 && body instanceof Statement) {
	    return (Statement)body;
	} else if (!(body instanceof Statement)) {
	    return null;
	} else {
	    throw new ArrayIndexOutOfBoundsException
		("A ProgramMetaConstruct contains only one statement ");
	}
    }

    /**
     * Returns the number of children of this node.
     * @return an int giving the number of children of this node
     */
    public int getChildCount() {
	return 1;
    }

    /**
     * Returns the child at the specified index in this node's "virtual"
     * child array.
     * @param index an index into this node's "virtual" child array
     * @return the program element at the given position
     * @exception ArrayIndexOutOfBoundsException if <tt>index</tt> is out
     * of bounds
     */
    public ProgramElement getChildAt(int index) {
	return body;
    }
    
    /**
     * Returns the index of the given child, or <CODE>-1</CODE> if there is 
     * no such child. The child is searched for by identity:
     * <CODE>getChildAt(getIndexOfChild(x)) == x</CODE>.
     * @param child the exact child to look for.
     * @return the index of the given child, or <CODE>-1</CODE>.
     */
    public int getIndexOfChild(ProgramElement child) {
	if (body != null && body.equals(child)) {
	    return 0;
	} 
	return -1;
    }
    
    /**
     * Returns the positional code of the given child, or <CODE>-1</CODE> if 
     * there is no such child. The result contains an encoding of the
     * relative position of the child as well as the role it has been
     * playing in this parent element. 
     * 
     * @param child the exact child to look for.
     * @return the positional code of the given child, or <CODE>-1</CODE>.
     */
    public int getChildPositionCode(ProgramElement child) {
        if (body != null && body.equals(child)) {
	    return 0 | 0;
        }
        return -1;
    }

    /**
     * Extracts the index of a child from its position code.
     * @param positionCode the position code.
     * @return the index of the given position code.
     * @see NonTerminalProgramElement#getChildPositionCode(ProgramElement)
     */
    public int getIndexOfChild(int positionCode) {
	return positionCode >> 4;
    }


    /**
     * Extracts the role of a child from its position code.
     * @param positionCode the position code.
     * @return the role code of the given position code.
     * @see NonTerminalProgramElement#getChildPositionCode(ProgramElement)
     */
    public int getRoleOfChild(int positionCode) {
	return positionCode & 15;
    }

    //-------------some methods to pretend being a type reference --------

    public KeYJavaType getKeYJavaType() {
	return null;
    }

    public ReferencePrefix getReferencePrefix() {
	return null;
    }

    public ReferencePrefix setReferencePrefix(ReferencePrefix r) {
	return this;
    }


    public int getDimensions() {
	return 0;
    }


    public int getTypeReferenceCount(){
	return 0;
    }

    public TypeReference getTypeReferenceAt(int index) {
	return this;
    }

    public PackageReference getPackageReference() {
	return null;
    }

    public int getExpressionCount() {
	return 0;
    }

    public Expression getExpressionAt(int index) {
	return null;
    }


    public ProgramElementName getProgramElementName() {
	return new ProgramElementName(toString());
    }    

    public String getName() {
	return toString();
    }

    /** calls the corresponding method of a visitor in order to
     * perform some action/transformation on this element
     * @param v the Visitor
     */
    public void visit(Visitor v) {
	v.performActionOnProgramMetaConstruct(this);
    }

    public void prettyPrint(PrettyPrinter p) throws IOException {
        p.printProgramMetaConstruct(this);
    }

    /** to String */
    public String toString() {
	return name+"( "+body+ ");";
    }

    public KeYJavaType getKeYJavaType(Services javaServ) {
	return getKeYJavaType();
    }
    
    // helpers
    
    /** 
     * creates an assignment <code> lhs:=rhs </code>
     */
    protected final Statement assign(Expression lhs, Expression rhs) {
	return new CopyAssignment(lhs, rhs);
    }

    /**
     * creates an attribute access
     */
    protected final Expression attribute(ReferencePrefix prefix, 
					 ProgramVariable field) {
	return new FieldReference(field, prefix);
    }

    /**
     * create a local variable declaration
     */
    protected final LocalVariableDeclaration declare(String name, KeYJavaType type) {
	return new LocalVariableDeclaration
	    (new TypeRef(type), 
	     new VariableSpecification
		 (new ProgramVariable(new ProgramElementName(name), type)));
    }

    /**
     * create a local variable declaration
     */
    protected final LocalVariableDeclaration declare
	(String name, Expression init, KeYJavaType type) {
	return new LocalVariableDeclaration
	    (new TypeRef(type), 
	     new VariableSpecification
		 (new ProgramVariable(new ProgramElementName(name), type), 
		  init, type));
    }



}
