// This file is part of KeY - Integrated Deductive Software Design 
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule.metaconstruct;
import de.uka.ilkd.key.logic.sort.ProgramSVSort;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.java.ProgramElement;
import de.uka.ilkd.key.java.Label;
import de.uka.ilkd.key.java.Statement;
import de.uka.ilkd.key.java.statement.LoopStatement;
import de.uka.ilkd.key.util.NotSupported;
import de.uka.ilkd.key.logic.op.*;
import de.uka.ilkd.key.rule.inst.SVInstantiations;
import de.uka.ilkd.key.logic.*;

/** 
 * This class is used to perform program transformations needed 
 * for the symbolic execution of a loop. It unwinds the loop:
 * e.g. 
 * <code>
 * while ( i<10 ) {
 *   i++
 * }  
 * </code> becomes 
 * <code>
 * l1: { 
 *   if (i<10) { i++} 
 *        else break l1;
 * l2: {
 *   while ( i<10 ) {
 *     i++
 *   }
 * }}  
 * </code> becomes          
 */
public class UnwindLoop extends  ProgramMetaConstruct {

    /** the outer label that is used to leave the while loop ('l1') */
    private final SchemaVariable outerLabel = 
	SchemaVariable.createProgramSV(new ProgramElementName("outer_label"),
				       ProgramSVSort.LABEL, false);
    /** the inner label ('l2') */
    private final SchemaVariable innerLabel =
	SchemaVariable.createProgramSV(new ProgramElementName("inner_label"),
				       ProgramSVSort.LABEL, false);
    /** list of the labels */
    private ListOfSchemaVariable instantiations;

    /** creates an unwind-loop ProgramMetaConstruct 
     * @param loop the LoopStatement contained by the meta construct 
     */
    public UnwindLoop(LoopStatement loop) {
	super("unwind-loop", loop); 
	instantiations = SLListOfSchemaVariable.EMPTY_LIST
	    .prepend(innerLabel).prepend(outerLabel);
    }

    /** performs the program transformation needed for symbolic
     * program transformation 
     * @param javaInfo the JavaInfo with all necessary information 
     * about the java programs
     * @param svInst the instantiations esp. of the inner and outer label 
     * @return the transformated program
     */
    public ProgramElement symbolicExecution(ProgramElement pe,
					    JavaInfo javaInfo,
					    SVInstantiations svInst) {
	if (!(pe instanceof LoopStatement)) {
	    return pe;
	}
	LoopStatement originalLoop = (LoopStatement)pe;
	WhileLoopTransformation w = 
	    new WhileLoopTransformation(originalLoop,
					(ProgramElementName)
					svInst.getInstantiation(outerLabel),
					(ProgramElementName)
					svInst.getInstantiation(innerLabel));
	w.start();
	return w.result();
    }

    /** returns the schemavariables that are needed to transform the given
     * loop. The unwind-loop construct may need labels if unlabeled breaks 
     * and/or continues occur in the loop. Often there will be uninstantiated
     * Schemavariables in the loop that is why the found instantiations have to
     * be given.
     */
    public ListOfSchemaVariable neededInstantiations(ProgramElement originalLoop,
						     SVInstantiations svInst) {
	WhileLoopTransformation w = 
	    new WhileLoopTransformation(originalLoop, svInst);
	w.start();
	instantiations = SLListOfSchemaVariable.EMPTY_LIST;
	if (w.innerLabelNeeded()) {
	    instantiations = instantiations.prepend(innerLabel);
	}
	if (w.outerLabelNeeded()) {
	    instantiations = instantiations.prepend(outerLabel);
	}
	return instantiations;
    }
        
}
