// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule.strategy;

import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.java.JavaInfo;


/**
 * Class providing some methods frequently needed by strategy-objects
 */
public abstract class AbstractStrategy implements Strategy {

    /**
     * Search for taclet apps at a given position
     * @param p_below also search for apps below p_pos
     * @param p_new only return new (not yet applied at p_pos or using
     * this if formula instantiations) rule apps
     * @return a list of taclet apps that are "sufficientlyComplete"
     */
    protected IteratorOfTacletApp findTacletApps ( ListOfHeuristic p_heur,
						   Goal            p_goal,
						   PosInOccurrence p_pos,
						   boolean         p_below,
						   boolean         p_new,
						   JavaInfo        p_javaInfo,
						   Constraint      p_userConstraint ) {
	if ( p_below )
	    return new TacletAppsIterator
		( p_goal.ruleAppIndex ()
		  .getTacletAppAtAndBelow ( p_heur, false, p_pos,
					    p_javaInfo, p_userConstraint )
		  .iterator(),
		  p_goal, p_new, p_javaInfo, p_userConstraint );
	else
	    return new TacletAppsIterator
		( p_goal.ruleAppIndex ()
		  .getTacletAppAt ( p_heur, false, p_pos,
				    p_javaInfo, p_userConstraint )
		  .iterator(),
		  p_goal, p_new, p_javaInfo, p_userConstraint );
    }


    protected static boolean isNewTacletApp ( Goal p_goal, TacletApp p_app ) {
	return !actuallyContains ( p_goal.appliedRuleApps (),
				   p_app );
    }


    protected static boolean violatesUserConstraint ( TacletApp  p_app,
						      Constraint p_userConstraint ) {
	return !p_app.constraint ().join ( p_userConstraint ).isSatisfiable ();
    }


    private static boolean actuallyContains ( ListOfRuleApp list, TacletApp rapp ) {

	boolean b;

	IteratorOfRuleApp it = list.iterator ();
	while ( it.hasNext () ) {	    
	    RuleApp cmp = it.next ();
	    if ( (rapp.getClass()==cmp.getClass()) &&
		 (rapp.rule().equals(cmp.rule())) &&
		 ( !( rapp instanceof PosTacletApp ) ||
		   ((PosTacletApp)rapp).posInOccurrence ()
		   .equals ( ((PosTacletApp)cmp).posInOccurrence () ) ) ) {
		if ( ((TacletApp)rapp).ifFormulaInstantiations () != null &&
		     ((TacletApp)cmp) .ifFormulaInstantiations () != null ) {
		    IteratorOfIfFormulaInstantiation it0 =
			((TacletApp)rapp).ifFormulaInstantiations ().iterator ();
		    IteratorOfIfFormulaInstantiation it1 =
			((TacletApp)cmp) .ifFormulaInstantiations ().iterator ();
		    b = true;
		    while ( b && it0.hasNext () ) {
			if ( it0.next ().getConstrainedFormula () !=
			     it1.next ().getConstrainedFormula () )
			    b = false;
		    }
		    if ( b )
			return true;
		} else
		    return true;
	    }
	}
	 
	return false;
    }


    private static class TacletAppsIterator implements IteratorOfTacletApp {
	private IteratorOfTacletApp appsAtPos;
	private IteratorOfTacletApp appsWithIfInst = null;
	
	private Goal                goal;
	private boolean             checkNew;
	private JavaInfo            javaInfo;
	private Constraint          userConstraint;

	private TacletApp           app;
	private boolean             nextReady      = false;

	public TacletAppsIterator ( IteratorOfTacletApp p_appsAtPos,
				    Goal                p_goal,
				    boolean             p_new,
				    JavaInfo            p_javaInfo,
				    Constraint          p_userConstraint ) {
	    appsAtPos      = p_appsAtPos;
	    goal           = p_goal;
	    checkNew       = p_new;
	    javaInfo       = p_javaInfo;
	    userConstraint = p_userConstraint;
	}

	protected void findNext () {
	    app = null;
	    while ( app == null ) {
		if ( appsWithIfInst != null && appsWithIfInst.hasNext () ) {
		    app = appsWithIfInst.next ();
		    if ( !app.complete () )
 			app = app.tryToInstantiate( javaInfo );
		    if ( app != null ) {
			if ( checkNew && !isNewTacletApp ( goal, app ) ||
			     violatesUserConstraint ( app, userConstraint ) )
			    app = null;
		    }
		} else {
		    if ( appsAtPos.hasNext () )
			appsWithIfInst = appsAtPos.next ()
			    .findIfFormulaInstantiations ( goal.node ().sequent (),
							   javaInfo,
							   userConstraint )
			    .iterator ();
		    else {
			appsAtPos = null;
			return;
		    }
		}
	    }
	    nextReady = true;
	}

	/** @return TacletApp the next element of collection */
	public TacletApp next() {
	    if ( !nextReady )
		findNext ();
	    nextReady = false;
	    return app;
	}

	/** @return boolean true iff collection has more unseen elements */
	public boolean hasNext() {
	    if ( !nextReady && appsAtPos != null )
		findNext ();
	    return nextReady;
	}
	
    }

}
