// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule.strategy;

import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.rule.strategy.proofevent.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.op.Modality;
import de.uka.ilkd.key.logic.op.Quantifier;
import de.uka.ilkd.key.logic.op.Metavariable;
import java.util.Hashtable;


/**
 * Apply beta rules, in combination with "ReplaceKnownStrategy" for
 * simplification
 */
public class BetaStrategy extends MarkNodesStrategy {

    /**
     * Informations about a term as cached within "betaCandidates"
     */
    protected static class TermInfo {
	public static final Integer CAND_NEVER = new Integer ( 0 );
	public static final Integer CAND_LEFT  = new Integer ( 1 );
	public static final Integer CAND_RIGHT = new Integer ( 2 );
	public static final Integer CAND_BOTH  = new Integer ( 3 );

	/** formula is positive (not negated) */
	public int     maxPosPath_positive;
	public boolean purePosPath_positive;

	/** formula is negated */
	public int     maxPosPath_negative;
	public boolean purePosPath_negative;

	/** formula contains a quantifier or modality */
	public boolean containsQuantifier;

	/** one of CAND_* */
	public Object  candidate;
    }

    /*
     * Table of formulas which could be splitted using the beta rule
     * This is the cache the method "isBetaCandidate" uses
     *
     *    keys: Term              values: TermInfo
     */
    protected ObjectHashtable betaCandidates = new ObjectHashtable ();

    ListOfHeuristic heur              = SLListOfHeuristic.EMPTY_LIST;

    ListOfHeuristic heurBeta          = SLListOfHeuristic.EMPTY_LIST;


    public BetaStrategy ( Namespace p_heur_ns ) {
	Named h = p_heur_ns.lookup ( new Name ( "beta" ) );

	if ( h != null )
	    heur = heurBeta = heurBeta.prepend ( (Heuristic)h );
	else
	    Debug.out ( "Didn't find the heuristic \"beta\" I am part of" );
    }


    /**
     * @return Iterator of heuristics that include this strategy
     */
    public IteratorOfHeuristic heuristics () {
	return heur.iterator ();
    }


    /**
     * @param p_priority Number of strategies for which
     * "recommendRuleApp" has already been called without result
     * @param p_heuristics List of heuristics chosen by the user
     * @param p_goals List of goals a possible rule application should
     * be searched for
     * @return The RuleApp that should be executed next (this strategy
     * thinks), or null
     */
    public RuleAppRecommendation recommendRuleApp ( int             p_priority,
						    ListOfHeuristic p_heuristics,
						    ListOfGoal      p_goals,
						    JavaInfo        p_javaInfo,
						    Constraint      p_userConstraint ) {
	IteratorOfGoal               it  = p_goals.iterator ();
	IteratorOfConstrainedFormula it2;
	IteratorOfTacletApp          appIt;
	Goal                         goal;
	Node                         node;
	Sequent                      seq;
	ConstrainedFormula           cf;
	PosInOccurrence              pos;
	boolean                      inAntec;
	int                          minPath, i;
	boolean                      purePosPath, b;

	while ( it.hasNext () ) {
	    goal = it.next ();
	    node = goal.node ();

	    if ( !isMarkedNode ( node ) ) {
		seq  = node.sequent ();
		
		// Find formula with most positive literals within a
		// d-path (if possible formula with a d-path
		// consisting only of positive literals)
		it2         = seq.iterator ();
		inAntec     = true;
		pos         = null;
		minPath     = 1000000; // INT_MAX in Java?
		purePosPath = false;
		while ( it2.hasNext () ) {
		    cf = it2.next ();

		    if ( inAntec &&
			 seq.succedent () != Semisequent.EMPTY_SEMISEQUENT &&
			 cf == seq.succedent ().getFirst () )
			inAntec = false;

		    if ( isBetaCandidate ( cf.formula (), inAntec ) ) {
			b = hasPurePosPath ( cf.formula (), !inAntec );
			if ( !purePosPath || b ) {
			    i = maxPosPath ( cf.formula (), !inAntec );
			    if ( purePosPath != b || i < minPath ) {
				minPath     = i;
				purePosPath = b;
				pos         = new PosInOccurrence
				    ( cf, PosInTerm.TOP_LEVEL, seq );			    
			    }
			}
		    }
		}

		if ( pos != null ) {
		    appIt = findTacletApps ( heurBeta, goal, pos, false, false,
					     p_javaInfo, p_userConstraint );
		    if ( appIt.hasNext () )
			return new RuleAppRecommendation ( goal, appIt.next () );
		}

		markNode ( node );
	    }
	    
	}

	return null;
    }


    /**
     * Settings like the set of active heuristics or the user
     * constraint may have changed
     */
    public void settingsChange ( ListOfHeuristic p_heuristics,
				 Proof           p_proof,
				 JavaInfo        p_javaInfo,
				 Constraint      p_userConstraint ) {
	betaCandidates.clear ();
	super.settingsChange ( p_heuristics,
			       p_proof,
			       p_javaInfo,
			       p_userConstraint );
    }


    /**
     * p_t contains a d-path consisting only of positive literals (as
     * a formula of the antecedent)
     */
    protected boolean hasPurePosPath ( Term p_t, boolean p_positive ) {
	TermInfo ti = termInfo ( p_t );
	return
	    p_positive ?
	    ti.purePosPath_positive :
	    ti.purePosPath_negative;
    }


    /**
     * The maximal number of positive literals occurring within a
     * d-path of "p_t" as a formula of the antecedent
     */
    protected int     maxPosPath     ( Term p_t, boolean p_positive ) {
	TermInfo ti = termInfo ( p_t );
	return
	    p_positive ?
	    ti.maxPosPath_positive :
	    ti.maxPosPath_negative;
    }


    /**
     * @return true iff "p_t" contains a quantifier or a modality
     */
    protected boolean containsQuantifier ( Term p_t ) {
	TermInfo ti = termInfo ( p_t );
	return ti.containsQuantifier;
    }


    /**
     * @return true iff the formula p_t could be splitted using the
     * beta rule
     */
    public boolean isBetaCandidate ( Term    p_t,
				     boolean p_inAntec ) {
	TermInfo ti = termInfo ( p_t );
	return
	    ti.candidate == TermInfo.CAND_BOTH ||
	    ti.candidate == ( p_inAntec ?
			      TermInfo.CAND_LEFT :
			      TermInfo.CAND_RIGHT );
    }


    /**
     * Get the informations about a term
     */
    protected TermInfo termInfo ( Term p_t ) {
	TermInfo ti = (TermInfo)betaCandidates.get ( p_t );

	if ( ti == null ) {
	    ti                      = new TermInfo ();

	    ti.purePosPath_positive = hasPurePosPathHelp     ( p_t, true  );
	    ti.purePosPath_negative = hasPurePosPathHelp     ( p_t, false );

	    ti.maxPosPath_positive  = maxPosPathHelp         ( p_t, true  );
	    ti.maxPosPath_negative  = maxPosPathHelp         ( p_t, false );

	    ti.containsQuantifier   = containsQuantifierHelp ( p_t );

	    ti.candidate            = candidateHelp          ( p_t, ti );

	    betaCandidates.put ( p_t, ti );
	}

	return ti;
    }


    protected int maxPosPathHelp ( Term p_t, boolean p_positive ) {
	if ( p_t.op () == ( p_positive ? Op.AND : Op.OR ) )
	    return maxPosPath ( p_t.sub ( 0 ), p_positive ) +
		maxPosPath ( p_t.sub ( 1 ), p_positive );

	else if ( p_t.op () == ( p_positive ? Op.OR : Op.AND ) )
	    return Math.max ( maxPosPath ( p_t.sub ( 0 ), p_positive ),
			      maxPosPath ( p_t.sub ( 1 ), p_positive ) );
	
	else if ( p_t.op () == Op.NOT )
	    return maxPosPath ( p_t.sub ( 0 ), !p_positive );

	else if ( p_positive && p_t.op () == Op.IMP )
	    return Math.max ( maxPosPath ( p_t.sub ( 0 ), !p_positive ),
			      maxPosPath ( p_t.sub ( 1 ),  p_positive ) );

	else if ( !p_positive && p_t.op () == Op.IMP )
	    return maxPosPath ( p_t.sub ( 0 ), !p_positive ) +
		maxPosPath ( p_t.sub ( 1 ), p_positive );

	else if ( p_positive && p_t.op () == Op.EQV )
	    return Math.max ( maxPosPath ( p_t.sub ( 0 ),  p_positive ) +
			      maxPosPath ( p_t.sub ( 1 ),  p_positive ),
			      maxPosPath ( p_t.sub ( 0 ), !p_positive ) +
			      maxPosPath ( p_t.sub ( 1 ), !p_positive ) );

	else if ( !p_positive && p_t.op () == Op.EQV )
	    return Math.max ( maxPosPath ( p_t.sub ( 0 ), !p_positive ) +
			      maxPosPath ( p_t.sub ( 1 ),  p_positive ),
			      maxPosPath ( p_t.sub ( 0 ),  p_positive ) +
			      maxPosPath ( p_t.sub ( 1 ), !p_positive ) );

	else if ( alwaysReplace ( p_t ) )
	    return 1;

	return p_positive ? 0 : 1;
    }


    protected boolean hasPurePosPathHelp ( Term p_t, boolean p_positive ) {
	if ( p_t.op () == ( p_positive ? Op.AND : Op.OR ) )
	    return hasPurePosPath ( p_t.sub ( 0 ), p_positive ) &&
		hasPurePosPath ( p_t.sub ( 1 ), p_positive );

	else if ( p_t.op () == ( p_positive ? Op.OR : Op.AND ) )
	    return hasPurePosPath ( p_t.sub ( 0 ), p_positive ) ||
		hasPurePosPath ( p_t.sub ( 1 ), p_positive );
	
	else if ( p_t.op () == Op.NOT )
	    return hasPurePosPath ( p_t.sub ( 0 ), !p_positive );

	else if ( p_positive && p_t.op () == Op.IMP )
	    return hasPurePosPath ( p_t.sub ( 0 ), !p_positive ) ||
		hasPurePosPath ( p_t.sub ( 1 ), p_positive );

	else if ( !p_positive && p_t.op () == Op.IMP )
	    return hasPurePosPath ( p_t.sub ( 0 ), !p_positive ) &&
		hasPurePosPath ( p_t.sub ( 1 ), p_positive );

	else if ( p_positive && p_t.op () == Op.EQV )
	    return ( hasPurePosPath ( p_t.sub ( 0 ), p_positive ) &&
		     hasPurePosPath ( p_t.sub ( 1 ), p_positive ) ) ||
		( hasPurePosPath ( p_t.sub ( 0 ), !p_positive ) &&
		  hasPurePosPath ( p_t.sub ( 1 ), !p_positive ) );

	else if ( !p_positive && p_t.op () == Op.EQV )
	    return ( hasPurePosPath ( p_t.sub ( 0 ), !p_positive ) &&
		     hasPurePosPath ( p_t.sub ( 1 ), p_positive ) ) ||
		( hasPurePosPath ( p_t.sub ( 0 ), p_positive ) &&
		  hasPurePosPath ( p_t.sub ( 1 ), !p_positive ) );

	else if ( alwaysReplace ( p_t ) )
	    return true;

	return !p_positive;
    }

    protected Object candidateHelp  ( Term     p_t,
				      TermInfo p_ti ) {
	if ( p_t.op () == Op.IMP ||
	     p_t.op () == Op.OR )
	    return
		isBetaCandidateHelp ( p_ti, false ) ?
		TermInfo.CAND_LEFT  :
		TermInfo.CAND_NEVER;
	else if ( p_t.op () == Op.AND )
	    return
		isBetaCandidateHelp ( p_ti, true  ) ?
		TermInfo.CAND_RIGHT :
		TermInfo.CAND_NEVER;
	else if ( p_t.op () == Op.EQV ) {
	    if ( isBetaCandidateHelp ( p_ti, true ) )
		return
		    isBetaCandidateHelp ( p_ti, false ) ?
		    TermInfo.CAND_BOTH  :
		    TermInfo.CAND_RIGHT;
	    else
		return
		    isBetaCandidateHelp ( p_ti, false ) ?
		    TermInfo.CAND_LEFT  :
		    TermInfo.CAND_NEVER;
	}

	return TermInfo.CAND_NEVER;
    }

    protected boolean isBetaCandidateHelp ( TermInfo p_ti,
					    boolean  p_positive ) {
	return
	    p_ti.containsQuantifier ||
	    ( p_positive ?
	      p_ti.maxPosPath_positive > 1 :
	      p_ti.maxPosPath_negative > 1 );
    }

    protected boolean containsQuantifierHelp ( Term p_t ) {
	if ( p_t.op () == Op.AND ||
	     p_t.op () == Op.OR ||
	     p_t.op () == Op.IMP ||
	     p_t.op () == Op.EQV )
	    return
		containsQuantifier ( p_t.sub ( 0 ) ) ||
		containsQuantifier ( p_t.sub ( 1 ) );
	else if ( p_t.op () == Op.NOT )
	    return
		containsQuantifier ( p_t.sub ( 0 ) );
	else
	    return alwaysReplace ( p_t );
    }

    /**
     * @return true iff the sign of "p_t" is not relevant (quantifiers
     * etc. could be positive or negative)
     */
    public static boolean alwaysReplace ( Term p_t ) {
	return
	    p_t.op () instanceof Modality ||
	    p_t.op () instanceof Quantifier;
    }

}
