// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule.strategy;

import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.util.Debug;


/**
 * Strategy to apply closure rules as soon as possible
 *
 * This does also handle alpha rules and simple simplify rules (rules
 * replacing "true" and "false")
 */
public class CloseStrategy extends TrackChangesStrategy {

    ListOfHeuristic heur              = SLListOfHeuristic.EMPTY_LIST;

    ListOfHeuristic heurClosure       = SLListOfHeuristic.EMPTY_LIST;
    ListOfHeuristic heurAlphaConcrete = SLListOfHeuristic.EMPTY_LIST;


    boolean         closureActive     = false;
    boolean         alphaActive       = false;


    public CloseStrategy ( Namespace p_heur_ns ) {
	Named h = p_heur_ns.lookup ( new Name ( "closure" ) );

	if ( h != null ) {
	    heurClosure = heurClosure.prepend ( (Heuristic)h );
	    heur        = heur       .prepend ( (Heuristic)h );
	} else
	    Debug.out ( "Didn't find the heuristic \"closure\" I am part of" );

	h = p_heur_ns.lookup ( new Name ( "alpha_concrete" ) );

	if ( h != null ) {
	    heurAlphaConcrete = heurAlphaConcrete.prepend ( (Heuristic)h );
	    heur              = heur             .prepend ( (Heuristic)h );
	} else
	    Debug.out ( "Didn't find the heuristic \"alpha_concrete\" I am part of" );
    }


    /**
     * @return Iterator of heuristics that include this strategy
     */
    public IteratorOfHeuristic heuristics () {
	return heur.iterator ();
    }


    /**
     * @param p_priority Number of strategies for which
     * "recommendRuleApp" has already been called without result
     * @param p_heuristics List of heuristics chosen by the user
     * @param p_goals List of goals a possible rule application should
     * be searched for
     * @return The RuleApp that should be executed next (this strategy
     * thinks), or null
     */
    public RuleAppRecommendation recommendRuleApp ( int             p_priority,
						    ListOfHeuristic p_heuristics,
						    ListOfGoal      p_goals,
						    JavaInfo        p_javaInfo,
						    Constraint      p_userConstraint ) {
	IteratorOfGoal           it = p_goals.iterator ();
	Goal                     goal;
	Node                     node;
	IteratorOfTacletApp      appIt;
	PosInOccurrence          pos;

	while ( it.hasNext () ) {
	    goal = it.next ();
	    node = goal.node ();

	    while ( ( pos = topNewFormula ( node ) ) != null ) {
		if ( closureActive ) {
		    // Search for closure rules
		    appIt = findTacletApps ( heurClosure, goal, pos,
					     false, true,
					     p_javaInfo, p_userConstraint );
			
		    if ( appIt.hasNext () )
			return new RuleAppRecommendation ( goal, appIt.next () );
		}

		if ( alphaActive ) {
		    // Search for alpha_concrete rules
		    appIt = findTacletApps ( heurAlphaConcrete, goal, pos,
					     true, false,
					     p_javaInfo, p_userConstraint );

		    if ( appIt.hasNext () )
			return new RuleAppRecommendation ( goal, appIt.next () );
		}

		popNewFormula ( node );
	    }
	}

	return null;
    }


    /**
     * Settings like the set of active heuristics or the user
     * constraint may have changed
     */
    public void settingsChange ( ListOfHeuristic p_heuristics,
				 Proof           p_proof,
				 JavaInfo        p_javaInfo,
				 Constraint      p_userConstraint ) {
	super.settingsChange ( p_heuristics,
			       p_proof,
			       p_javaInfo,
			       p_userConstraint );

	closureActive = p_heuristics.contains ( heurClosure      .head () );
	alphaActive   = p_heuristics.contains ( heurAlphaConcrete.head () );
    }

}
