// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule.strategy;

import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.op.ArrayOfQuantifiableVariable;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.proof.ListOfNode;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.gui.ProofEvent;
import de.uka.ilkd.key.rule.strategy.proofevent.*;
import de.uka.ilkd.key.util.Debug;
import java.util.Enumeration;


/**
 * Apply gamma rules (quantifier instantiations using metavariables) fair
 */
public class GammaStrategy extends AbstractStrategy {

    ListOfHeuristic heur            = SLListOfHeuristic.EMPTY_LIST;

    ListOfHeuristic heurCopy        = SLListOfHeuristic.EMPTY_LIST;
    ListOfHeuristic heurDestructive = SLListOfHeuristic.EMPTY_LIST;


    protected static final Name RULE_COPY_ALL        = new Name ( "all_left"        );
    protected static final Name RULE_COPY_EX         = new Name ( "ex_right"        );
    protected static final Name RULE_DESTRUCTIVE_ALL = new Name ( "all_left_delete" );
    protected static final Name RULE_DESTRUCTIVE_EX  = new Name ( "ex_right_delete" );


    /**
     * Holding the state of each node regarding gamma rule
     * applications
     * key: "Node"           value: "NodeState"
     */
    private ObjectHashtable nodes = new ObjectHashtable ();


    public GammaStrategy ( Namespace p_heur_ns ) {
	Named h = p_heur_ns.lookup ( new Name ( "gamma" ) );

	if ( h != null )
	    heurCopy = heurCopy.prepend ( (Heuristic)h );
	else
	    Debug.out ( "Didn't find the heuristic \"gamma\" I am part of" );

	h = p_heur_ns.lookup ( new Name ( "gamma_destructive" ) );

	if ( h != null )
	    heurDestructive = heurDestructive.prepend ( (Heuristic)h );
	else
	    Debug.out ( "Didn't find the heuristic \"gamma_destructive\" I am part of" );

	heur = heurCopy;
    }


    /**
     * @return Iterator of heuristics that include this strategy
     */
    public IteratorOfHeuristic heuristics () {
	return heur.iterator ();
    }



    private int previousPriority = 100000;


    /**
     * @param p_priority Number of strategies for which
     * "recommendRuleApp" has already been called without result
     * @param p_heuristics List of heuristics chosen by the user
     * @param p_goals List of goals a possible rule application should
     * be searched for
     * @return The RuleApp that should be executed next (this strategy
     * thinks), or null
     */
    public RuleAppRecommendation recommendRuleApp ( int             p_priority,
						    ListOfHeuristic p_heuristics,
						    ListOfGoal      p_goals,
						    JavaInfo        p_javaInfo,
						    Constraint      p_userConstraint ) {
	IteratorOfGoal        it   = p_goals.iterator ();
	Goal                  goal;
	Node                  node;
	NodeState             ns;
	RuleAppRecommendation res  = null;

	while ( res == null && it.hasNext () ) {
	    goal = it.next   ();
	    node = goal.node ();
	    ns   = (NodeState)nodes.get ( node );
	    if ( ns == null ) {
		nodeAdded ( goal.node () );
		ns = (NodeState)nodes.get ( node );
	    }

	    if ( previousPriority < p_priority )
		res = ns.recommendRuleAppSecondCall ( p_priority,
						      goal,
						      p_javaInfo,
						      p_userConstraint );
	    else
		res = ns.recommendRuleAppFirstCall ( p_priority,
						     goal,
						     p_javaInfo,
						     p_userConstraint );
	}

	previousPriority = p_priority;

	return res;
    }


    /**
     * Tell the strategy object that a rule has been applied
     */
    public void ruleApplied ( RuleAppInfo p_rai ) {
	Node                      node  = p_rai.getOriginalNode ();
	NodeState                 ns    = null;
	NodeState                 newNs;
	IteratorOfNodeReplacement it    = p_rai.getReplacementNodes ();
	NodeReplacement           nr;

	if ( node != null )
	    ns = (NodeState)nodes.remove ( node );

	if ( ns != null )
	    ns.ruleApplied ( p_rai );

	while ( it.hasNext () ) {
	    nr = it.next ();

	    if ( ns == null )
		newNs = new NodeState ( nr.getNode () );
	    else {
		if ( it.hasNext () )
		    newNs = (NodeState)ns.clone  ();
		else
		    newNs = ns;
		newNs.findNewFormulas ( nr );
	    }

	    nodes.put ( nr.getNode (), newNs );
	}
    }


    /**
     * A new node has been added to the proof (and we don't know how
     * exactly)
     */
    public void nodeAdded ( Node p_node ) {
	if ( nodes.get ( p_node ) == null )
	    nodes.put ( p_node,
			new NodeState ( p_node ) );
    }


    /**
     * Tell the strategy object that a node has been modified
     */
    public void nodeChanged      ( Node              p_node,
				   SequentChangeInfo p_sci ) {}

    /**
     * Tell the strategy object that the proof has been expanded,
     * adding the nodes <code>p_newLeaves</code> as new leaves below
     * <code>p_oldLeaf</code>
     */
    public void proofExpanded    ( Node       p_oldLeaf,
				   ListOfNode p_newLeaves ) {}

    /**
     * Settings like the set of active heuristics or the user
     * constraint may have changed
     */
    public void settingsChange ( ListOfHeuristic p_heuristics,
				 Proof           p_proof,
				 JavaInfo        p_javaInfo,
				 Constraint      p_userConstraint ) {}


    /**
     * Immutable
     * The state of a formula of a node, i.e. latest
     * application of a gamma rule
     */
    private static abstract class FormulaState {}

    private static class QuantifiedFormulaState extends FormulaState {
	// Latest instantiation was done when NodeState.appCount ==
	// latestApp
	private int latestApp = -1;

	// Result of the latest instantiation
	private ConstrainedFormula latestInstantiation = null;

	public int getLatestApp () {
	    return latestApp;
	}

	public ConstrainedFormula getLatestInstantiation () {
	    return latestInstantiation;
	}

	public QuantifiedFormulaState setLatestApp ( int p ) {
	    QuantifiedFormulaState res = new QuantifiedFormulaState ();
	    res.latestApp              = p;
	    res.latestInstantiation    = latestInstantiation;
	    return res;
	}

	public QuantifiedFormulaState setLatestInstantiation
	    ( ConstrainedFormula p ) {
	    QuantifiedFormulaState res = new QuantifiedFormulaState ();
	    res.latestApp              = latestApp;
	    res.latestInstantiation    = p;
	    return res;
	}
    }
    
    private static class InstantiationState extends FormulaState {
	// Formula this one has been derived from
	private ConstrainedFormula origin;
	
	public InstantiationState ( ConstrainedFormula p_origin ) {
	    origin = p_origin;
	}

	public ConstrainedFormula getOrigin () {
	    return origin;
	}
    }


    /**
     * State of a node, i.e. table of the formula states, etc.
     */
    private class NodeState {
	// Number of rule applications for this node
	private int appCount  = 0;

	// Least value of "getLatestApp" within "formulas"
	private int oldestApp = -1;

	/**
	 * Table of all universally quantified formulas of this node
	 * key objects "ConstrainedFormula"     value objects "FormulaState"
	 */
	private ObjectHashtable          formulas;

	/**
	 * Formulas to which the gamma rule has been applied once, but
	 * which may need iterated application (i.e. more than one quantifier)
	 */
	private ListOfConstrainedFormula completionDue;
	private boolean                  completeNextFormula = false;
	private ConstrainedFormula       completeNextFOrigin = null;

	/** Instantiate all quantified formulas of the sequent once */
	private boolean                  instantiationTriggered = false;
	private ListOfConstrainedFormula doneThisTime           =
	    SLListOfConstrainedFormula.EMPTY_LIST;

	private NodeState ( ObjectHashtable          p_table,
			    ListOfConstrainedFormula p_compDue ) {
	    formulas       = p_table;
	    completionDue  = p_compDue;
	}

	public NodeState ( Node p_node ) {
	    IteratorOfConstrainedFormula it;
	    ConstrainedFormula           cf;

	    completionDue  = SLListOfConstrainedFormula.EMPTY_LIST;
	    formulas       = new ObjectHashtable ();

	    it = p_node.sequent ().antecedent ().iterator ();
	    while ( it.hasNext () ) {
		cf = it.next ();
		if ( isQuantified ( cf, true ) )
		    addFormula ( cf );
	    }

	    it = p_node.sequent ().succedent ().iterator ();
	    while ( it.hasNext () ) {
		cf = it.next ();
		if ( isQuantified ( cf, false ) )
		    addFormula ( cf );
	    }
	}


	public Object clone () {
	    NodeState ns = new NodeState ( (ObjectHashtable)formulas.clone (),
					   completionDue );
	    ns.appCount  = appCount;
	    ns.oldestApp = oldestApp;
	    return ns;
	}


	/**
	 * This is called before cloning the "NodeState"-object if a
	 * rule creates more than one new node
	 */
	public void ruleApplied ( RuleAppInfo p_rai ) {
	    ++appCount;

	    RuleApp rApp  = p_rai.getRuleApp ();
	    boolean check = true;

	    if ( rApp != null && rApp instanceof PosTacletApp ) {
		PosTacletApp app  = (PosTacletApp)rApp;
		String       name = app.taclet ().displayName ();

		// If rule has really been applied, remove the
		// corresponding formula from the right stack
		if ( name.equals ( RULE_COPY_ALL.toString () ) ||
		     name.equals ( RULE_COPY_EX .toString () ) ) {

		    ConstrainedFormula cf = app.posInOccurrence ().constrainedFormula ();

		    completeNextFormula = true;
		    completeNextFOrigin = cf;

		    // Update state of the formula
		    updateApp ( cf );

		    doneThisTime.prepend ( cf );

		} else if ( name.equals ( RULE_DESTRUCTIVE_ALL.toString () ) ||
			    name.equals ( RULE_DESTRUCTIVE_EX .toString () ) ) {

		    // Remove from stack
		    ConstrainedFormula cf = app.posInOccurrence ().constrainedFormula ();

		    if ( completionDue != SLListOfConstrainedFormula.EMPTY_LIST &&
			 cf == completionDue.head () )
			completionDue = completionDue.tail ();
		    completeNextFormula = true;
		    completeNextFOrigin = cf;

		    check               = false;

		}

	    }

	    if ( check ) {
		PosInOccurrence pos = rApp.posInOccurrence ();

		if ( pos != null )
		    formulaUsed ( pos.constrainedFormula () );

		if ( rApp instanceof TacletApp &&
		     ((TacletApp)rApp).ifFormulaInstantiations () != null ) {
		    IteratorOfIfFormulaInstantiation it     = ((TacletApp)rApp)
			.ifFormulaInstantiations ().iterator ();
		    IfFormulaInstantiation           ifInst;

		    while ( it.hasNext () ) {
			ifInst = it.next ();
			if ( ifInst instanceof IfFormulaInstSeq )
			    formulaUsed ( ifInst.getConstrainedFormula () );
		    }
		}
	    }
	}


	/**
	 * This is called after cloning the "NodeState"-object if a
	 * rule creates more than one new node
	 */
	public void findNewFormulas ( NodeReplacement p_nr ) {
	    IteratorOfNodeChange it = p_nr.getNodeChanges ();
	    NodeChange           nc;
	    PosInOccurrence      pos;

	    while ( it.hasNext () ) {
		nc = it.next ();
		if ( nc instanceof NodeChangeAddFormula ) {
		    pos = ((NodeChangeAddFormula)nc).getPos ();
		    if ( completeNextFormula ) {
			completionDue =
			    completionDue.prepend ( pos.constrainedFormula () );
			completeNextFormula = false;

			/** Remember the formula the new one is an
			 * instantiation of */
			if ( completeNextFOrigin != null ) {
			    FormulaState fs =
				(FormulaState)formulas.get ( completeNextFOrigin );

			    if ( fs != null ) {
				if ( fs instanceof QuantifiedFormulaState ) {
				    formulas.put    ( pos.constrainedFormula (),
						      new InstantiationState
						      ( completeNextFOrigin ) );
				} else if ( fs instanceof InstantiationState ) {
				    formulas.remove ( completeNextFOrigin );
				    formulas.put    ( pos.constrainedFormula (),
						      fs );
				}
			    }

			    completeNextFOrigin = null;
			}
		    } else if ( isQuantified ( pos ) )
			addFormula ( pos.constrainedFormula () );
		}
	    }
	}

	/**
	 * The first, high priority call of "recommendRuleApp"
	 */
	public RuleAppRecommendation recommendRuleAppFirstCall
	    ( int         p_priority,
	      Goal        p_goal,
	      JavaInfo    p_javaInfo,
	      Constraint  p_userConstraint ) {
	    RuleAppRecommendation res = null;

	    /** Complete applications that have been begun earlier */
	    while ( res == null &&
		    completionDue != SLListOfConstrainedFormula.EMPTY_LIST ) {
		ConstrainedFormula cf  = completionDue.head ();
		Sequent            seq = p_goal.node ().sequent ();

		if ( seq.antecedent ().contains ( cf ) ||
		     seq.succedent  ().contains ( cf ) ) {
		    PosInOccurrence     pos   =
			new PosInOccurrence ( cf,
					      PosInTerm.TOP_LEVEL,
					      seq );
		    IteratorOfTacletApp appIt =
			findTacletApps ( heurDestructive, p_goal, pos, false, false,
					 p_javaInfo, p_userConstraint );

		    while ( appIt.hasNext () )
			res = new RuleAppRecommendation ( p_goal,
							  appIt.next () );

		    if ( res == null )
			instantiationComplete ( cf );
		} else
		    removeFormula ( cf );
		
		if ( res == null )
		    completionDue = completionDue.tail ();		    		
	    }

	    if ( res == null && instantiationTriggered ) {
		/** Instantiate further formulas */

		ConstrainedFormula cf;
		while ( res == null &&
			instantiationTriggered &&
			( cf = findCandidate () ) != null )
		    res = startInstantiation ( cf,
					       p_goal,
					       p_javaInfo,
					       p_userConstraint );

		if ( res == null )
		    stopInstantiation ();
	    }

	    return res;
	}

	/**
	 * The second, low priority call of "recommendRuleApp" (if
	 * nothing else is applicable)
	 */
	public RuleAppRecommendation recommendRuleAppSecondCall
	    ( int         p_priority,
	      Goal        p_goal,
	      JavaInfo    p_javaInfo,
	      Constraint  p_userConstraint ) {
	    RuleAppRecommendation res = null;

	    if ( !formulas.isEmpty () ) {
		ConstrainedFormula cf;

		instantiationTriggered = true;

		while ( res == null &&
			instantiationTriggered &&
			( cf = findCandidate () ) != null ) {
		    res = startInstantiation ( cf,
					       p_goal,
					       p_javaInfo,
					       p_userConstraint );
		}

		//		if ( res == null )
		    stopInstantiation ();
	    }

	    return res;
	}

	
	/**
	 * Start the instantiation of the quantified formula "p_cf"
	 */
	private RuleAppRecommendation startInstantiation
	    ( ConstrainedFormula p_cf,
	      Goal               p_goal,
	      JavaInfo           p_javaInfo,
	      Constraint         p_userConstraint ) {
	    RuleAppRecommendation res = null;

	    if ( !doneThisTime.contains ( p_cf ) ) {
		Sequent seq = p_goal.node ().sequent ();

		if ( seq.antecedent ().contains ( p_cf ) ||
		     seq.succedent  ().contains ( p_cf ) ) {
		    PosInOccurrence     pos   =
			new PosInOccurrence ( p_cf,
					      PosInTerm.TOP_LEVEL,
					      seq );
		    IteratorOfTacletApp appIt =
			findTacletApps ( heurCopy, p_goal, pos, false, false,
					 p_javaInfo, p_userConstraint );

		    while ( appIt.hasNext () )
			res = new RuleAppRecommendation ( p_goal,
							  appIt.next () );
		}

		if ( res == null )
		    removeFormula ( p_cf );
	    } else
		// Something went wrong
		stopInstantiation ();

	    return res;
	}


	private void stopInstantiation () {
	    instantiationTriggered = false;
	    doneThisTime           = SLListOfConstrainedFormula.EMPTY_LIST;
	}


	private void addFormula ( ConstrainedFormula p_cf ) {
	    formulas.put ( p_cf, new QuantifiedFormulaState () );
	    oldestApp = -1;
	}


	private void removeFormula ( ConstrainedFormula p_cf ) {
	    formulas.remove ( p_cf );
	}


	private void updateApp ( ConstrainedFormula p_cf ) {
	    FormulaState fs = (FormulaState)formulas.get ( p_cf );
	    if ( fs == null || !( fs instanceof QuantifiedFormulaState ) )
		addFormula ( p_cf );
	    else
		formulas.put ( p_cf,
			       ((QuantifiedFormulaState)fs)
			       .setLatestApp ( appCount ) );

	    updateOldestApp ();
	}

	
	/**
	 * @param p_cf the result of (possibly) repeated applications
	 * of gamma rules: a new instantiation of a quantified formula
	 */
	private void instantiationComplete ( ConstrainedFormula p_cf ) {
	    FormulaState fs = (FormulaState)formulas.get ( p_cf );
	    if ( fs != null && fs instanceof InstantiationState ) {
		ConstrainedFormula origin = ((InstantiationState)fs).getOrigin ();
		FormulaState       ofs    = (FormulaState)formulas.get ( origin );

		if ( ofs != null && ofs instanceof QuantifiedFormulaState )
		    formulas.put ( origin,
				   ((QuantifiedFormulaState)ofs)
				   .setLatestInstantiation ( p_cf ) );		    
	    }
	}


	private void updateOldestApp () {
	    Enumeration        en = formulas.keys ();
	    ConstrainedFormula cf;
	    FormulaState       fs;

	    oldestApp             = appCount;

	    while ( en.hasMoreElements () ) {
		cf = (ConstrainedFormula)en.nextElement ();
		fs = (FormulaState)formulas.get ( cf );
		if ( fs instanceof QuantifiedFormulaState ) {
		    int t = ((QuantifiedFormulaState)fs).getLatestApp ();
		    if ( t < oldestApp )
			oldestApp = t;
		}
	    }
	}


	/**
	 * Find candidate for gamma rule application with least recent
	 * instantiation
	 */
	private ConstrainedFormula findCandidate () {
	    Enumeration        en           = formulas.keys ();
	    ConstrainedFormula cf;
	    ConstrainedFormula res          = null;
	    int                minLatestApp = appCount;
	    FormulaState       fs;

	    while ( en.hasMoreElements () ) {
		cf = (ConstrainedFormula)en.nextElement ();
		fs = (FormulaState)formulas.get ( cf );
		if ( fs instanceof QuantifiedFormulaState &&
		     ((QuantifiedFormulaState)fs).getLatestApp () < minLatestApp &&
		     isCandidate ( (QuantifiedFormulaState)fs ) ) {
		    res          = cf;
		    minLatestApp = ((QuantifiedFormulaState)fs).getLatestApp ();
		}
	    }

	    return res;
	}


	private boolean isCandidate ( int p_latestApp ) {
	    return true;
	    // || p_latestApp <= appCount - 10;  // To be improved
	}


	private boolean isCandidate ( QuantifiedFormulaState p_fs ) {
	    return p_fs.getLatestInstantiation () == null;
	}


	private void formulaUsed ( ConstrainedFormula p_cf ) {
	    FormulaState       fs = (FormulaState)formulas.get ( p_cf );
	    ConstrainedFormula origin;
	    if ( fs != null &&
		 fs instanceof InstantiationState &&
		 ( origin = ((InstantiationState)fs).getOrigin () ) != null ) {
		FormulaState ofs = (FormulaState)formulas.get ( origin );
		if ( ofs != null &&
		     ofs instanceof QuantifiedFormulaState )
		    formulas.put ( origin,
				   ((QuantifiedFormulaState)ofs)
				   .setLatestInstantiation ( null ) );		    
	    }
	}


	private boolean isQuantified ( ConstrainedFormula p_cf,
				       boolean            p_inAntec ) {
	    return p_inAntec ?
		p_cf.formula ().op () == Op.ALL :
		p_cf.formula ().op () == Op.EX;
	}

	private boolean isQuantified ( PosInOccurrence    p_pos ) {
	    return 
		p_pos.constrainedFormula ().formula ().op () == Op.ALL &&
		p_pos.isInAntec () ||
		p_pos.constrainedFormula ().formula ().op () == Op.EX  &&
		!p_pos.isInAntec ();
	}
    }

}
