// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule.strategy;

import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.proof.ListOfNode;
import de.uka.ilkd.key.proof.IteratorOfNode;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.rule.strategy.proofevent.*;
import de.uka.ilkd.key.util.Debug;
import java.util.HashSet;


/**
 * Strategy base class that observes changed node (without further
 * details)
 */
public abstract class MarkNodesStrategy extends AbstractStrategy {
    
    /**
     * Nodes that have already been checked in some way and haven't
     * been modified since (and need not to be considered anymore)
     * This is a hashset with key objects "Node"
     */
    protected HashSet markedNodes = new HashSet ();

    /**
     * @return true iff "p_node" has been marked and has not been
     * modified since
     */
    protected boolean isMarkedNode   ( Node p_node ) {
	return markedNodes.contains ( p_node );
    }

    /**
     * Mark "p_node"
     */
    protected void    markNode       ( Node p_node ) {
	markedNodes.add ( p_node );
    }

    /**
     * Unmark all nodes
     */
    protected void    removeAllMarks () {
	markedNodes.clear ();
    }


    /**
     * A new node has been added to the proof (and we don't know how
     * exactly)
     */
    public void     nodeAdded       ( Node        p_node ) {
	markedNodes.remove ( p_node );
    }

    /**
     * Tell the strategy object that a node has been modified
     */
    public void                  nodeChanged      ( Node              p_node,
						    SequentChangeInfo p_sci ) {
	markedNodes.remove ( p_node );	
    }

    /**
     * Tell the strategy object that the proof has been expanded,
     * adding the nodes <code>p_newLeaves</code> as new leaves below
     * <code>p_oldLeaf</code>
     */
    public void                  proofExpanded    ( Node       p_oldLeaf,
						    ListOfNode p_newLeaves ) {
	// if the old node was marked, also mark the new ones
	if ( isMarkedNode ( p_oldLeaf ) ) {
	    IteratorOfNode it = p_newLeaves.iterator ();

	    while ( it.hasNext () )
		markNode ( it.next () );
	}
    }

    /**
     * Tell the strategy object that a rule has been applied
     */
    public void     ruleApplied     ( RuleAppInfo p_rai ) {
	/*
	IteratorOfNodeReplacement it = p_rai.getReplacementNodes ();
	
	while ( it.hasNext () )
	    markedNodes.remove ( it.next ().getNode () );	    
	*/
    }

    /**
     * Settings like the set of active heuristics or the user
     * constraint may have changed
     */
    public void settingsChange ( ListOfHeuristic p_heuristics,
				 Proof           p_proof,
				 JavaInfo        p_javaInfo,
				 Constraint      p_userConstraint ) {
	removeAllMarks ();
    }

}
