// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule.strategy;

import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.rule.strategy.proofevent.*;
import de.uka.ilkd.key.util.Debug;
import de.uka.ilkd.key.logic.op.Op;
import de.uka.ilkd.key.logic.op.Metavariable;
import java.util.Hashtable;


/**
 * Use the "hyper tableau strategy" to simplify and split formulas intelligently
 * This will make little sense if the heuristic "split" is activated
 * This class is to be used in combination with "BetaStrategy" for splitting
 */
public class ReplaceKnownStrategy extends TrackChangesStrategy {

    ListOfHeuristic heur              = SLListOfHeuristic.EMPTY_LIST;

    ListOfHeuristic heurReplaceKnown  = SLListOfHeuristic.EMPTY_LIST;


    protected static final Name RULE_SIMP_LEFT  = new Name ( "replace_known_left"      );
    protected static final Name RULE_SIMP_RIGHT = new Name ( "replace_known_right"     );


    protected BetaStrategy beta;


    public ReplaceKnownStrategy ( Namespace    p_heur_ns,
				  BetaStrategy p_beta ) {
	beta = p_beta;

	Named h = p_heur_ns.lookup ( new Name ( "replace_known" ) );

	if ( h != null )
	    heurReplaceKnown = heurReplaceKnown.prepend ( (Heuristic)h );
	else
	    Debug.out ( "Didn't find the heuristic \"replace_known\" I am part of" );

	heur = heurReplaceKnown;
    }


    /**
     * @return Iterator of heuristics that include this strategy
     */
    public IteratorOfHeuristic heuristics () {
	return heur.iterator ();
    }


    /**
     * @param p_priority Number of strategies for which
     * "recommendRuleApp" has already been called without result
     * @param p_heuristics List of heuristics chosen by the user
     * @param p_goals List of goals a possible rule application should
     * be searched for
     * @return The RuleApp that should be executed next (this strategy
     * thinks), or null
     */
    public RuleAppRecommendation recommendRuleApp ( int             p_priority,
						    ListOfHeuristic p_heuristics,
						    ListOfGoal      p_goals,
						    JavaInfo        p_javaInfo,
						    Constraint      p_userConstraint ) {
	IteratorOfGoal               it  = p_goals.iterator ();
	IteratorOfConstrainedFormula it2;
	IteratorOfTacletApp          appIt;
	PosTacletApp                 app;
	Goal                         goal;
	Node                         node;
	Sequent                      seq;
	ConstrainedFormula           cf;
	ConstrainedFormula           cf2;
	PosInOccurrence              pos;
	boolean                      inAntec;

	while ( it.hasNext () ) {
	    goal = it.next ();
	    node = goal.node ();
	    seq  = node.sequent ();

	    // Apply simplify rules
	    while ( ( pos = topNewFormula ( node ) ) != null ) {
		cf      = pos.constrainedFormula ();
		inAntec = pos.isInAntec ();

		// Try to simplify this formulas using the other
		// formulas of the sequent
		{
		    SimplifyWithAll s =
			new SimplifyWithAll ( goal,
					      p_javaInfo,
					      p_userConstraint );
		    try {
			findSimplifyPos ( cf.formula (),
					  pos,
					  new Boolean ( !inAntec ),
					  s,
					  p_userConstraint );
		    } catch ( SimplifyPosHandler.BreakException e ) {
			return new RuleAppRecommendation ( goal, s.result );
		    }
		}

		if ( inAntec ||
		     BetaStrategy.alwaysReplace ( cf.formula () ) ) {
		    // Try to simplify the other formulas of the
		    // sequent using this formula
		    SimplifyWithFormula s =
			new SimplifyWithFormula ( cf,
						  inAntec,
						  goal,
						  p_javaInfo,
						  p_userConstraint );
		    try {
			it2 = seq.iterator ();
			while ( it2.hasNext () ) {
			    cf2 = it2.next ();
			    if ( cf2 != cf ) {
				pos = new PosInOccurrence
				    ( cf2, PosInTerm.TOP_LEVEL, seq );
				findSimplifyPos ( cf2.formula (),
						  pos,
						  new Boolean
						  ( !pos.isInAntec () ),
						  s,
						  p_userConstraint );
			    }
			}
		    } catch ( SimplifyPosHandler.BreakException e ) {
			return new RuleAppRecommendation ( goal, s.result );
		    }
		}
		popNewFormula ( node );
	    }
	}
	
	return null;
    }


    /**
     * Interface used for "findSimplifyPos"
     */
    protected static interface SimplifyPosHandler {
	public static class BreakException extends Exception {
	    public BreakException () {
		super ( "Found something" );
	    }
	}

	public static final BreakException BREAK_EXCEPTION =
	    new BreakException ();

	/**
	 * Decide about a position within a formula
	 * @param p_positive true if the subformula is not negated
	 * (top-level formulas of the antecedent are negated), null if
	 * the subformula appears within an argument of Op.EQV
	 * @throws BreakException if the search should be stopped
	 */
	public void handlePos ( Term            p_t,
			        PosInOccurrence p_pos,
				Boolean         p_positive )
	    throws BreakException;
    }


    /**
     * Search for subformulas of a formula which can be simplified
     * using any formula of the sequent
     */
    protected class SimplifyWithAll implements SimplifyPosHandler {
	public  TacletApp  result;

	private Goal       goal;
	private JavaInfo   javaInfo;
	private Constraint userConstraint;

	/**
	 * @param p_goal sequent within which formulas should be
	 * sought for
	 */
	public SimplifyWithAll ( Goal       p_goal,
				 JavaInfo   p_javaInfo,
				 Constraint p_userConstraint ) {
	    goal           = p_goal;
	    javaInfo       = p_javaInfo;
	    userConstraint = p_userConstraint;
	}

	public void handlePos ( Term            p_t,
				PosInOccurrence p_pos,
				Boolean         p_positive )
	    throws BreakException {

	    IteratorOfTacletApp appIt;

	    appIt = findTacletApps ( heurReplaceKnown, goal, p_pos,
				     false, true,
				     javaInfo, userConstraint );
	    while ( appIt.hasNext () ) {
		result = appIt.next ();

		// only simplify positive subformulas using formulas
		// of the antecedent
		if ( ( p_positive == null ||
		       p_positive.booleanValue () ==
		       result.taclet ().name ().equals ( RULE_SIMP_LEFT ) ) &&
		     ( result.taclet ().name ().equals ( RULE_SIMP_LEFT ) ||
		       BetaStrategy.alwaysReplace ( p_t ) ) &&
		     checkSimplify ( (PosTacletApp)result ) )
		    throw BREAK_EXCEPTION;
	    }
	}
    }


    /**
     * Try to simplify a formula using a specific other formula
     */
    protected class SimplifyWithFormula implements SimplifyPosHandler {
	public  TacletApp                    result;

	private ConstrainedFormula           subst;
	private boolean                      inAntec;

	private ListOfIfFormulaInstantiation ifInst;
	
	private Goal                         goal;
	private JavaInfo                     javaInfo;
	private Constraint                   userConstraint;

	private boolean                      alwaysReplace;

	/**
	 * @param p_subst the formula which can be used to simplify
	 * @param p_inAntec true if p_subst is part of the antecedent
	 */
	public SimplifyWithFormula ( ConstrainedFormula p_subst,
				     boolean            p_inAntec,
				     Goal               p_goal,
				     JavaInfo           p_javaInfo,
				     Constraint         p_userConstraint ) {
	    subst          = p_subst;
	    inAntec        = p_inAntec;
	    goal           = p_goal;
	    javaInfo       = p_javaInfo;
	    userConstraint = p_userConstraint;

	    ifInst         = SLListOfIfFormulaInstantiation.EMPTY_LIST
		.prepend ( new IfFormulaInstSeq ( goal.node ().sequent (),
						  inAntec,
						  subst ) );

	    alwaysReplace  = BetaStrategy.alwaysReplace ( p_subst.formula () );
	}

	public void handlePos ( Term            p_t,
				PosInOccurrence p_pos,
				Boolean         p_positive )
	    throws BreakException {

	    // only simplify positive subformulas using formulas
	    // of the antecedent

	    // already try to unify formulas to decide whether a
	    // subformula can be removed (faster than by trying to
	    // create a TacletApp)
	    if ( ( p_positive == null ||
		   p_positive.booleanValue () == inAntec ) &&
		 Constraint.BOTTOM.unify ( p_t, subst.formula () )
		 .isSatisfiable () ) {
		IteratorOfTacletApp appIt;

		appIt = goal.ruleAppIndex ()
		    .getTacletAppAt ( heurReplaceKnown, false, p_pos,
				      javaInfo, userConstraint )
		    .iterator();

		while ( appIt.hasNext () ) {
		    result = appIt.next ();
		    // "RULE_SIMP_LEFT" if "subst" is within the antecedent
		    // "RULE_SIMP_RIGHT" if "subst" is within the succedent
		    if ( inAntec ==
			 result.taclet ().name ().equals ( RULE_SIMP_LEFT ) ) {
			result = result.setIfFormulaInstantiations
			    ( ifInst, javaInfo, userConstraint );
			if ( result != null && 
			     checkSimplify ( (PosTacletApp)result ) &&
			     isNewTacletApp ( goal, result ) &&
			     !violatesUserConstraint ( result,
						       userConstraint ) )
			    throw BREAK_EXCEPTION;
		    }
		}
	    }
	}
    }


    /**
     * Find all subterms of "p_t" together with their signs
     * @param p_pos position of "p_t"
     * @param p_positive p_pos.isInAntec () ? Boolean.FALSE : Boolean.TRUE
     * @param p_handler method to call for every found subterm
     */
    protected void findSimplifyPos ( Term               p_t,
				     PosInOccurrence    p_pos,
				     Boolean            p_positive,
				     SimplifyPosHandler p_handler,
				     Constraint         p_userConstraint )
	throws SimplifyPosHandler.BreakException {

	/** only for real subterms */
	if ( p_pos.posInTerm () != PosInTerm.TOP_LEVEL )
	    p_handler.handlePos ( p_t, p_pos, p_positive );

	if ( p_t.op () == Op.NOT ) {
	    findSimplifyPos ( p_t  .sub  ( 0 ),
			      p_pos.down ( 0 ),
			      invert ( p_positive ),
			      p_handler,   p_userConstraint );
	} else if ( p_t.op () == Op.IMP ) {
	    findSimplifyPos ( p_t  .sub  ( 0 ),
			      p_pos.down ( 0 ),
			      invert ( p_positive ),
			      p_handler,   p_userConstraint );
	    findSimplifyPos ( p_t  .sub  ( 1 ),
			      p_pos.down ( 1 ),
			      p_positive,
			      p_handler,   p_userConstraint );
	} else if ( p_t.op () == Op.EQV ) {
	    findSimplifyPos ( p_t  .sub  ( 0 ),
			      p_pos.down ( 0 ),
			      null,
			      p_handler,   p_userConstraint );
	    findSimplifyPos ( p_t  .sub  ( 1 ),
			      p_pos.down ( 1 ),
			      null,
			      p_handler,   p_userConstraint );
	} else if ( p_t.op () == Op.AND ||
		    p_t.op () == Op.OR ) {
	    findSimplifyPos ( p_t  .sub  ( 0 ),
			      p_pos.down ( 0 ),
			      p_positive,
			      p_handler,   p_userConstraint );
	    findSimplifyPos ( p_t  .sub  ( 1 ),
			      p_pos.down ( 1 ),
			      p_positive,
			      p_handler,   p_userConstraint );
	} else if ( p_t.op () instanceof Metavariable ) {
	    if ( p_pos.termBelowMetavariable () == null ) {
		Term metaTerm = p_userConstraint
		    .getInstantiation ( (Metavariable)p_t.op () );
		if ( metaTerm.op () != p_t.op () )
		    findSimplifyPos ( metaTerm,
				      p_pos.setTermBelowMetavariable ( metaTerm ),
				      p_positive,
				      p_handler,   p_userConstraint );
	    }
	}
    }

    
    protected Boolean invert ( Boolean p ) {
	return
	    p == null ?
	    null :
	    ( p.booleanValue () ? Boolean.FALSE : Boolean.TRUE );
    }


    protected boolean checkSimplify ( PosTacletApp p_app ) {
	// Don't apply simplification, if this would create a new
	// formula (constraint becomes stronger) and we have to split
	// the formula anyway)
	return
	    !isBetaCandidate ( p_app.posInOccurrence ()
			       .constrainedFormula ().formula (),
			       p_app.posInOccurrence ()
			       .isInAntec () ) ||
	    p_app.posInOccurrence ().constrainedFormula ().constraint ()
	    .isAsStrongAs ( p_app.constraint () );
    }


    /**
     * @return true iff the formula p_t could be splitted using the
     * beta rule
     */
    protected boolean isBetaCandidate ( Term    p_t,
					boolean p_inAntec ) {
	return beta.isBetaCandidate ( p_t, p_inAntec );
    }

}
