// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule.strategy;

import de.uka.ilkd.key.rule.Heuristic;
import de.uka.ilkd.key.rule.ListOfHeuristic;
import de.uka.ilkd.key.rule.IteratorOfHeuristic;
import de.uka.ilkd.key.rule.RuleApp;
import de.uka.ilkd.key.logic.Namespace;
import de.uka.ilkd.key.logic.SequentChangeInfo;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.GoalListener;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.proof.ListOfNode;
import de.uka.ilkd.key.proof.SLListOfNode;
import de.uka.ilkd.key.proof.IteratorOfNode;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.ProofTreeListener;
import de.uka.ilkd.key.proof.ProofTreeEvent;
import de.uka.ilkd.key.logic.Constraint;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.gui.ProofEvent;
import de.uka.ilkd.key.rule.strategy.proofevent.*;
import de.uka.ilkd.key.gui.InteractiveProofListener;
import de.uka.ilkd.key.gui.KeYMediator;


/**
 * Class holding a list of "Strategy"-objects, using them to find
 * possible rule applications
 */
public class StrategyManager {


    final static boolean USE_GAMMA_STRATEGY = false;


    /** Available strategies */
    ListOfStrategy strategies       = SLListOfStrategy.EMPTY_LIST;
    ListOfStrategy strategyOrder    = SLListOfStrategy.EMPTY_LIST;

    /** Strategies that are active with the currently chosen
     * heuristics */
    ListOfStrategy activeStrategies = SLListOfStrategy.EMPTY_LIST;

    Strategy       std;
    Strategy       simplifier;

    boolean        simplifierOnly   = false;

    /** The proof this strategy manager is operating on */
    Proof          proof;

    KeYMediator    mediator         = null;
    ProofListener  proofListener;
    TreeListener   treeListener;
    GoalListener   goalListener;

    public StrategyManager ( Namespace   p_heur_ns,
			     Proof       p_proof ) {
	proof         = p_proof;
	proofListener = new ProofListener ();
	treeListener  = new TreeListener  ();
	goalListener  = new SCIListener   ();

	createStrategies ( p_heur_ns );

	registerAtProof ();
    }

    private void registerAtProof () {
	if ( proof == null )
	    return;

	proof.addProofTreeListener ( treeListener );

	IteratorOfGoal it = proof.openGoals ().iterator ();
	while ( it.hasNext () )
	    it.next ().addGoalListener ( goalListener );

	addGoals ( proof );
    }

    public void setMediator ( KeYMediator p_mediator ) {
	if ( mediator != null )
	    mediator.removeInteractiveProofListener ( proofListener );

	mediator = p_mediator;

	if ( mediator != null )
	    mediator.addInteractiveProofListener ( proofListener );
    }

    /**
     * @param p_heuristics which strategies should be considered
     * @param p_goals List of goals a possible rule application should
     * be searched for
     * @return The RuleApp that should be executed next (one strategy
     * thinks), or null
     */
    public RuleAppRecommendation recommendRuleApp ( ListOfHeuristic p_heuristics,
						    ListOfGoal      p_goals,
						    JavaInfo        p_javaInfo,
						    Constraint      p_userConstraint ) {
	IteratorOfStrategy    it       = activeStrategies.iterator ();
	int                   priority = 0;
	RuleAppRecommendation res      = null;

	while ( res == null && it.hasNext () ) {
	    res = it.next ().recommendRuleApp ( priority,
						p_heuristics,
						p_goals,
						p_javaInfo,
						p_userConstraint );
	    ++priority;
	}
	
	return res;
    }


    private ListOfHeuristic lastHeuristics     = null;
    private Constraint      lastUserConstraint = null;

    /**
     * Settings like the set of active heuristics or the user
     * constraint may have changed
     */
    public void                  settingsChange   ( ListOfHeuristic p_heuristics,
						    Proof           p_proof,
						    JavaInfo        p_javaInfo,
						    Constraint      p_userConstraint ) {
	if ( lastHeuristics != p_heuristics ||
	     lastUserConstraint == null ||
	     !lastUserConstraint.equals ( p_userConstraint ) ) {
	    IteratorOfStrategy it = strategies.iterator ();
	    while ( it.hasNext () )
		it.next ().settingsChange ( p_heuristics,
					    p_proof,
					    p_javaInfo,
					    p_userConstraint );
	    lastHeuristics     = p_heuristics;
	    lastUserConstraint = p_userConstraint;

	    setupActiveStrategies ( p_heuristics );
	}
    }


    protected void createStrategies ( Namespace p_heur_ns ) {
	// The last inserted strategy will get the highest priority

	GammaStrategy gamma;
	if ( USE_GAMMA_STRATEGY )
	    gamma = new GammaStrategy ( p_heur_ns );
	std        = new StdStrategy        ( p_heur_ns );
	simplifier = new SimplifierStrategy ( p_heur_ns );

	// Gamma is the first and the last strategy
	if ( USE_GAMMA_STRATEGY ) {
	    strategies    = strategies   .prepend ( gamma );
	    strategyOrder = strategyOrder.prepend ( strategies.head () );
	}

	strategies    = strategies   .prepend ( std );
	strategyOrder = strategyOrder.prepend ( strategies.head () );

	BetaStrategy beta = new BetaStrategy ( p_heur_ns );
	strategies    = strategies   .prepend ( beta );
	strategyOrder = strategyOrder.prepend ( strategies.head () );

	strategies    = strategies   .prepend ( new ReplaceKnownStrategy ( p_heur_ns,
									   beta ) );
	strategyOrder = strategyOrder.prepend ( strategies.head () );

	strategies    = strategies   .prepend ( simplifier );
	strategyOrder = strategyOrder.prepend ( strategies.head () );

	strategies    = strategies   .prepend ( new CloseStrategy        ( p_heur_ns ) );
	strategyOrder = strategyOrder.prepend ( strategies.head () );

	if ( USE_GAMMA_STRATEGY )
	    strategyOrder = strategyOrder.prepend ( gamma );	
    }

    
    protected void setupActiveStrategies ( ListOfHeuristic p_heuristics ) {
	activeStrategies              = SLListOfStrategy.EMPTY_LIST;

	IteratorOfStrategy    it      = strategyOrder.iterator ();	
	IteratorOfHeuristic   itHeur;
	Strategy              s;

	while ( it.hasNext () ) {
	    s = it.next ();

	    // Only allow the simplifier strategy if simplifierOnly is
	    // set
	    if ( !simplifierOnly || s == simplifier ) {
		if ( s == std )
		    // Always use the standard strategy
		    activeStrategies = activeStrategies.append ( s );
		else {
		    itHeur = s.heuristics ();
		    while ( itHeur.hasNext () ) {
			if ( p_heuristics.contains ( itHeur.next () ) ) {
			    activeStrategies = activeStrategies.append ( s );
			    break;
			}
		    }
		}
	    }
	}
    }


    /**
     * HACK
     * @param p_simplifierOnly determines whether the only strategy
     * that should be applied is the update simplifier strategy
     */
    public void setSimplifierOnly ( boolean p_simplifierOnly ) {
	simplifierOnly = p_simplifierOnly;
	if ( lastHeuristics != null )
	    setupActiveStrategies ( lastHeuristics );
    }
    

    protected void addGoals ( Proof p_proof ) {
	IteratorOfGoal it = p_proof.openGoals ().iterator ();
	while ( it.hasNext () )
	    nodeAdded ( it.next ().node () );
    }


    protected void nodeAdded ( Node p_node ) {
	IteratorOfStrategy it = strategies.iterator ();
	while ( it.hasNext () )
	    it.next ().nodeAdded ( p_node );	
    }


    public void expandProof    ( Node       p_oldLeaf,
				 ListOfNode p_newLeaves ) {
	IteratorOfStrategy it = strategies.iterator ();
	while ( it.hasNext () )
	    it.next ().proofExpanded ( p_oldLeaf, p_newLeaves );
    }


    public void nodeChanged      ( Node              p_node,
				   SequentChangeInfo p_sci ) {
	IteratorOfStrategy it = strategies.iterator ();
	while ( it.hasNext () )
	    it.next ().nodeChanged ( p_node, p_sci );
    }


   private class ProofListener implements InteractiveProofListener {

	/** invoked if automatic execution has started
	 */
	public void autoModeStarted(ProofEvent e) {}

	/** invoked if automatic execution has stopped
	 */
	public void autoModeStopped(ProofEvent e) {}
    
	/** invoked when a rule has been applied */
	public void ruleApplied(ProofEvent e) {
	    if ( e.getSource ().getProof () != proof )
		return;

	    RuleAppInfo rai = e.getRuleAppInfo ();
	    if ( rai == null )
		return;

	    IteratorOfStrategy it = strategies.iterator ();
	    while ( it.hasNext () )
		it.next ().ruleApplied ( rai );
	}
	
    }


    private class TreeListener implements ProofTreeListener {

	/** The node mentioned in the ProofTreeEvent has changed, and/or
	 * there are new descendants of that node.  Any nodes that are not
	 * descendants of that node are unaffected. */
	public void proofExpanded(ProofTreeEvent e) {}

	/** The proof tree has been pruned under the node mentioned in the
	 * ProofTreeEvent.  In other words, that node should no longer
	 * have any children now.  Any nodes that were not descendants of
	 * that node are unaffected.*/
	public void proofPruned(ProofTreeEvent e) {
	    if ( mediator != null ) {
		Proof proof = mediator.getSelectedProof ();
		if ( proof != null )
		    nodeAdded ( e.getNode () );
	    }
	}

	/** The structure of the proof has changed radically.  Any client should
	 * rescan the whole proof tree. */
	public void proofStructureChanged(ProofTreeEvent e) {}

	/** The proof trees has been closed (the list of goals is empty).
	 */
	public void proofClosed(ProofTreeEvent e) {}

	/** The goal mentioned in the ProofEvent has been removed 
	 *from the list of goals.
	 */
	public void proofGoalRemoved(ProofTreeEvent e) {}

	/** The goals mentiones in the list of added goals in the proof
	 * event have been added to the proof
	 */
	public void proofGoalsAdded(ProofTreeEvent e) {}

	/** The goals mentiones in the list of added goals in the proof
	 * event have been added to the proof
	 */
	public void proofGoalsChanged(ProofTreeEvent e) {}

    }

    private class SCIListener implements GoalListener {

	/** 
	 * informs the listener about a change that occured to the sequent of goal
	 */
	public void sequentChanged(Goal              source,
				   SequentChangeInfo sci) {
	    nodeChanged ( source.node (), sci );
	}

	/**
	 * Informs the listener that the given goal <code>source</code>
	 * has been replaced by the goals <code>newGoals</code> (note that
	 * <code>source</code> may be an element of
	 * <code>newGoals</code>). The nodes of <code>newGoals</code> are
	 * children of the node <code>parent</code>
	 */
	public void goalReplaced(Goal source, Node parent, ListOfGoal newGoals) {	    
	    IteratorOfGoal childrenIt = newGoals.iterator ();
	    ListOfNode     children   = SLListOfNode.EMPTY_LIST;

	    while ( childrenIt.hasNext () )
		children = children.prepend ( childrenIt.next ().node () );

	    expandProof ( parent, children );
	}

    }    

}
