// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule.strategy;

import de.uka.ilkd.key.rule.*;
import de.uka.ilkd.key.logic.*;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.proof.ListOfNode;
import de.uka.ilkd.key.proof.IteratorOfNode;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.java.JavaInfo;
import de.uka.ilkd.key.rule.strategy.proofevent.*;
import de.uka.ilkd.key.util.Debug;
import java.util.Hashtable;


/**
 * Base class for strategy-objects providing methods for observing
 * proof tree modifications (currently appearence of new formulas)
 */
public abstract class TrackChangesStrategy extends AbstractStrategy {
    
    /**
     * Formulas that have been added to any node
     * This is a hashtable with key objects "Node" and values
     * "NewFormulasHolder"
     */
    private Hashtable newFormulas = new Hashtable ();


    /**
     * @return the first new formula of "p_node" or null (iff there is
     * nothing new)
     */
    protected PosInOccurrence topNewFormula ( Node p_node ) {
	NewFormulasHolder nfh = (NewFormulasHolder)newFormulas.get ( p_node );
	PosInOccurrence   res = null;

	if ( nfh != null ) {
	    res = nfh.top ( p_node );
	    testEmptyNFH ( p_node, nfh );
	}

	return res;
    }


    /**
     * @return the first new formula of "p_node" and remove it from
     * the list or null (iff there is nothing new)
     */
    protected PosInOccurrence popNewFormula ( Node p_node ) {
	NewFormulasHolder nfh = (NewFormulasHolder)newFormulas.get ( p_node );
	PosInOccurrence   res = null;

	if ( nfh != null ) {
	    res = nfh.pop ( p_node );
	    testEmptyNFH ( p_node, nfh );
	}

	return res;
    }


    /**
     * Remove the map entry for <code>p_node</code> if
     * <code>p_nfh</code> does not hold any formulas
     */
    private void testEmptyNFH ( Node              p_node,
				NewFormulasHolder p_nfh ) {
	if ( p_nfh.isEmpty () )
	    newFormulas.remove ( p_node );	
    }


    /**
     * A new node has been added to the proof (and we don't know how
     * exactly)
     */
    public void                  nodeAdded        ( Node       p_node ) {
	if ( newFormulas.get ( p_node ) == null ) {
	    NewFormulasHolder nfh = new NewFormulasHolder ( p_node.sequent () );
	    if ( !nfh.isEmpty () )
		newFormulas.put ( p_node, nfh );
	}
    }

    /**
     * Tell the strategy object that a node has been modified
     */
    public void                  nodeChanged      ( Node              p_node,
						    SequentChangeInfo p_sci ) {
	NewFormulasHolder nfh = (NewFormulasHolder)newFormulas.get ( p_node );

	if ( nfh == null ) {
	    nfh = new NewFormulasHolder ();
	    newFormulas.put ( p_node, nfh );
	}

	nfh.addChanges ( p_sci );
	testEmptyNFH ( p_node, nfh );
    }

    /**
     * Tell the strategy object that the proof has been expanded,
     * adding the nodes <code>p_newLeaves</code> as new leaves below
     * <code>p_oldLeaf</code>
     */
    public void                  proofExpanded    ( Node       p_oldLeaf,
						    ListOfNode p_newLeaves ) {
	NewFormulasHolder nfh = (NewFormulasHolder)newFormulas.remove ( p_oldLeaf );

	if ( nfh != null ) {
	    IteratorOfNode it = p_newLeaves.iterator ();

	    if ( it.hasNext () ) {
		while ( true ) {
		    newFormulas.put ( it.next (), nfh );
		    if ( !it.hasNext () )
			break;
		    nfh = (NewFormulasHolder)nfh.clone ();
		}
	    }
	}
    }

    /**
     * Tell the strategy object that a rule has been applied
     */
    public void                  ruleApplied      ( RuleAppInfo p_rai ) {}

    /**
     * Settings like the set of active heuristics or the user
     * constraint may have changed
     */
    public void settingsChange ( ListOfHeuristic p_heuristics,
				 Proof           p_proof,
				 JavaInfo        p_javaInfo,
				 Constraint      p_userConstraint ) {
	newFormulas.clear ();

	IteratorOfGoal it = p_proof.openGoals ().iterator ();
	while ( it.hasNext () )
	    nodeAdded ( it.next ().node () );
    }


    /**
     * Simple class to store accumulated modification information for
     * a single node
     */
    private static class NewFormulasHolder {
	public ListOfConstrainedFormula antecedent;
	public ListOfConstrainedFormula succedent;

	public NewFormulasHolder () {
	    this ( SLListOfConstrainedFormula.EMPTY_LIST,
		   SLListOfConstrainedFormula.EMPTY_LIST );
	}

	/**
	 * Add all formulas of <code>p_seq</code> as new formulas
	 */
	public NewFormulasHolder ( Sequent p_seq ) {
	    this ( p_seq.antecedent ().toList (),
		   p_seq.succedent  ().toList () );
	}

	public NewFormulasHolder ( ListOfConstrainedFormula p_antecedent,
				   ListOfConstrainedFormula p_succedent ) {
	    antecedent = p_antecedent;
	    succedent  = p_succedent;
	}

	public void addChanges ( SequentChangeInfo p_sci ) {
	    antecedent =
		p_sci.addedFormulas ( true ).prepend
		( removeFormulas ( antecedent,
				   p_sci.removedFormulas ( true ) ) );
	    succedent  =
		p_sci.addedFormulas ( false ).prepend
		( removeFormulas ( succedent,
				   p_sci.removedFormulas ( false ) ) );
	}

	/**
	 * @return the position of the first new formula that is
	 * stored within this object, or <code>null</code> iff no such
	 * formula exists
	 */
	public PosInOccurrence top ( Node p_node ) {
	    PosInOccurrence    res = null;
	    ConstrainedFormula cf;
	    Sequent            seq = p_node.sequent ();
	    
	    while ( res == null &&
		    antecedent != SLListOfConstrainedFormula.EMPTY_LIST ) {
		cf = antecedent.head ();
		if ( seq.antecedent ().contains ( cf ) )
		    res        = new PosInOccurrence ( cf,
						       PosInTerm.TOP_LEVEL,
						       seq,
						       true );
		else
		    // Remove formulas that no longer exist within the
		    // sequent
		    antecedent = antecedent.tail ();
	    }

	    while ( res == null &&
		    succedent != SLListOfConstrainedFormula.EMPTY_LIST ) {
		cf = succedent.head ();
		if ( seq.succedent ().contains ( cf ) )
		    res        = new PosInOccurrence ( cf,
						       PosInTerm.TOP_LEVEL,
						       seq,
						       false );
		else
		    // Remove formulas that no longer exist within the
		    // sequent
		    succedent  = succedent.tail ();
	    }

	    return res;
	}

	/**
	 * @return the position of the first new formula that is
	 * stored within this object, and remove this formula from
	 * this object, or <code>null</code> iff no such formula
	 * exists
	 */
	public PosInOccurrence pop ( Node p_node ) {
	    PosInOccurrence res = top ( p_node );
	    if ( res != null ) {
		if ( antecedent == SLListOfConstrainedFormula.EMPTY_LIST )
		    succedent  = succedent .tail ();
		else
		    antecedent = antecedent.tail ();
	    }
	    return res;
	}

	/**
	 * @return true iff this object does not know any new formulas
	 */
	public boolean isEmpty () {
	    return
		antecedent == SLListOfConstrainedFormula.EMPTY_LIST &&
		succedent  == SLListOfConstrainedFormula.EMPTY_LIST;
	}

	private ListOfConstrainedFormula
	    removeFormulas ( ListOfConstrainedFormula p_list,
			     ListOfConstrainedFormula p_toRemove ) {
	    IteratorOfConstrainedFormula it = p_toRemove.iterator ();
	    while ( it.hasNext () )
		p_list = p_list.removeAll ( it.next () );
	    return p_list;
	}

	public Object clone () {
	    return new NewFormulasHolder ( antecedent,
					   succedent );
	}
    }
}
