// This file is part of KeY - Integrated Deductive Software Design
// Copyright (C) 2001-2003 Universitaet Karlsruhe, Germany
//                         and Chalmers University of Technology, Sweden          
//
// The KeY system is protected by the GNU General Public License. 
// See LICENSE.TXT for details.
//

package de.uka.ilkd.key.rule.strategy.proofevent;


import de.uka.ilkd.key.rule.strategy.*;
import de.uka.ilkd.key.rule.RuleApp;
import de.uka.ilkd.key.proof.Proof;
import de.uka.ilkd.key.proof.Node;
import de.uka.ilkd.key.proof.Goal;
import de.uka.ilkd.key.proof.ListOfGoal;
import de.uka.ilkd.key.proof.IteratorOfGoal;
import de.uka.ilkd.key.proof.GoalListener;
import de.uka.ilkd.key.proof.ProofTreeListener;
import de.uka.ilkd.key.proof.ProofTreeEvent;
import de.uka.ilkd.key.logic.Sequent;
import de.uka.ilkd.key.logic.SequentChangeInfo;
import de.uka.ilkd.key.logic.ListOfSequentChangeInfo;
import de.uka.ilkd.key.logic.SLListOfSequentChangeInfo;
import de.uka.ilkd.key.logic.Semisequent;
import de.uka.ilkd.key.logic.ConstrainedFormula;
import de.uka.ilkd.key.logic.IteratorOfConstrainedFormula;
import de.uka.ilkd.key.logic.PosInOccurrence;
import de.uka.ilkd.key.logic.PosInTerm;
import java.util.Enumeration;


/**
 * Collect changes applied to a node during a given period of time
 */
public class NodeChangeJournal implements GoalListener {

    Proof           proof;

    /**
     * The original node
     */
    Node            node;

    /**
     * This is a hashtable storing the leaves that are currently below
     * the original node, and all changes applied to each of them
     *
     * Keys: Node          Values: NodeChanges
     */
    ObjectHashtable changes = new ObjectHashtable ();

    /**
     * @param p_goal the original goal/node
     */
    public NodeChangeJournal ( Proof p_proof,
			       Goal  p_goal ) {
	proof = p_proof;
	node  = p_goal.node ();
	changes.put ( node, new NodeChanges () );

	p_goal.addGoalListener ( this );
    }


    /**
     * Create an RuleAppInfo object containing all changes stored
     * within this object; remove all listeners
     */
    public RuleAppInfo getRuleAppInfo ( RuleApp p_ruleApp ) {
	Enumeration             en      = changes.keys ();
	Node                    newNode;
	Goal                    newGoal;
	NodeChanges             nc;
	ListOfNodeReplacement   nrs     = SLListOfNodeReplacement.EMPTY_LIST;

	while ( en.hasMoreElements () ) {
	    newNode = (Node)en.nextElement ();
	    newGoal = proof.getGoal ( newNode );

	    if ( newGoal != null ) {
		nc  = (NodeChanges)changes.get ( newNode );

		nrs = nrs.prepend ( new NodeReplacement
				    ( newNode, node, nc.scis ) );
	    
		newGoal.removeGoalListener ( this );
	    }
	}

	return new RuleAppInfo ( p_ruleApp, node, nrs );
    }


    private static class NodeChanges {
	public ListOfSequentChangeInfo scis;

	public NodeChanges () {
	    this ( SLListOfSequentChangeInfo.EMPTY_LIST );
	}

	public NodeChanges ( ListOfSequentChangeInfo p_scis ) {
	    scis = p_scis;
	}

	public void addSCI ( SequentChangeInfo p_sci ) {
	    scis = scis.prepend ( p_sci );
	}

	public Object clone () {
	    return new NodeChanges ( scis );
	}
    }



    // GoalListener methods

    /** 
     * informs the listener about a change that occured to the sequent
     * of goal
     */
    public void sequentChanged ( Goal              source,
				 SequentChangeInfo sci ) {
	NodeChanges nc = (NodeChanges)changes.get ( source.node () );

	if ( nc != null )
	    nc.addSCI ( sci );
    }

    /**
     * Informs the listener that the given goal <code>source</code>
     * has been replaced by the goals <code>newGoals</code> (note that
     * <code>source</code> may be an element of
     * <code>newGoals</code>). The nodes of <code>newGoals</code> are
     * children of the node <code>parent</code>
     */
    public void goalReplaced(Goal source, Node parent, ListOfGoal newGoals) {
	NodeChanges nc = (NodeChanges)changes.remove ( parent );

	if ( nc != null ) {
	    IteratorOfGoal it = newGoals.iterator ();
	    if ( it.hasNext () ) {
		while ( true ) {
		    changes.put ( it.next ().node (), nc );
		    if ( !it.hasNext () )
			break;
		    nc = (NodeChanges)nc.clone ();
		}
	    }
	}
    }

}
